////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2008, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

//
// FST.h: public interface to FST library.
//

#ifndef __FST_h__
#define __FST_h__

#include "fst/windows.h"
#include "fst/FSTSTL.h"                      // before other FST and STL headers

#include <cstdio>
#include <cstdarg>
#include <cassert>
#include <iostream>
#include <string>

#include "fst/FSTAllocator.h"
#include "fst/FSTStringMap.h"
#include "fst/FSTSemirings.h"
#include "fst/FSTReferenceCount.h"
#include "fst/ReferenceCountedMemory.h"

// Globals ////////////////////////////////////////////////////////////////////


// For mapping strings to indices and vice-a-versa.
//
extern FSTStringMap FSTstring_map;

// Available semirings for weight arithmetic.
//
extern FSTSemiring FSTMinPlus;
extern FSTSemiring FSTMaxPlus;
extern FSTSemiring FSTPlusTimes;
extern FSTSemiring FSTPlusTimesLog;


// Types //////////////////////////////////////////////////////////////////////


enum FSTArcOrder { FSTNone, FSTInput, FSTOutput };

enum FSTArcMode  { FSTEqual, FSTLessEqual, FSTGreaterEqual };

class FSTArc {
public:
    int   input;
    int   output;
    float weight;
    int   next;
    int   prev;
    FSTArc()
        {}
    FSTArc(int p, int q, int i, int o, float w)
        : input(i), output(o), weight(w), next(q), prev(p) {}
    FSTArc(const FSTArc* a)
        : input(a->input), output(a->output), weight(a->weight),
          next(a->next), prev(a->prev) {}
    FSTAllocator_new_delete(FSTArc);
};

typedef fst::vector<FSTArc*> FSTArcVector;

struct FSTArcIterData
    : public FSTReferenceCounted {
    FSTArcVector arcs;
    FSTArcIterData() {}
    virtual ~FSTArcIterData() {
        for (FSTArcVector::iterator i = arcs.begin(); i != arcs.end(); i++)
            delete *i;
    }
    FSTAllocator_new_delete(FSTArcIterData);
};

class FSTArcIter
    : public FSTReferenceCounter<FSTReferenceCounted> {
private:
    FSTArc** _cur;
    FSTArc** _end;
public:
    FSTArcIter()
        : _RC(0), _cur(0), _end(0) {}
    FSTArcIter(FSTArc** cur,
               FSTArc** end,
               FSTReferenceCounted* data=0)
        : _RC(data), _cur(cur), _end(end) {}
    FSTArcIter(const FSTArcVector& arcs)
        : _RC(0) {
            if (arcs.size() == 0) {
                _cur = _end = 0;
            } else {
                _cur = const_cast<FSTArc**>(&arcs[0]);
                _end = _cur + arcs.size();
            }
    }
    FSTArcIter(FSTArcIterData* data)
        : _RC(data) {
            if (data && data->arcs.size() > 0) {
                _cur = const_cast<FSTArc**>(&data->arcs[0]);
                _end = _cur + data->arcs.size();
            } else {
                _cur = _end = 0;
            }
    }
    FSTArcIter(const FSTArcVector& arcs,
               FSTArcVector::const_iterator begin,
               FSTArcVector::const_iterator end)
        : _RC(0) {
            if (begin < end) {
                _cur = const_cast<FSTArc**>(&arcs[begin - arcs.begin()]);
                _end = _cur + (end - begin);
            } else {
                _cur = _end = 0;
            }
    }
    FSTArcIter(const FSTArcIter& x)
        : _RC(x._rep), _cur(x._cur), _end(x._end) {}
    ~FSTArcIter()
        {}
    void operator=(const FSTArcIter& x)
        { _RC::operator=(x); _cur = x._cur; _end = x._end; }
    void operator++()
        { _cur++; }
    void operator++(int)
        { _cur++; }
    void operator+=(int x)
        { _cur += x; }
    int n() const
        { return _end - _cur; }
    bool valid() const
        { return _cur < _end; }
    const FSTArc* operator->() const
        { return *_cur; }
    const FSTArc* operator* () const
        { return *_cur; }
    operator const FSTArc* () const
        { return *_cur; }
    int prev() const
        { return (*_cur)->prev; }
    int next() const
        { return (*_cur)->next; }
    int input() const
        { return (*_cur)->input; }
    int output() const
        { return (*_cur)->output; }
    void set_input_index(int index)
        { (*_cur)->input = index; }
    void set_output_index(int index)
        { (*_cur)->output = index; }
    void set_weight(float weight)
        { (*_cur)->weight = weight; }
    float weight() const
        { return (*_cur)->weight; }
    // experimental
    FSTArc** begin() const
        { return _cur; }
    FSTArc** end() const
        { return _end; }
};

class FSTStateIter {
private:
    int _i;
    const int* _n_states;
public:
    FSTStateIter(const int& n_states)
        : _i(0), _n_states(&n_states) {}
    ~FSTStateIter()
        {}
    void operator = (const FSTStateIter& x)
        { _i = x._i; _n_states = x._n_states; }
    int n() const
        { return *_n_states - _i; }
    bool valid() const
        { return _i < *_n_states; }
    void operator++()
        { _i++; }
    void operator++(int)
        { _i++; }
    int operator*() const
        { return _i; }
    int state() const
        { return _i; }
};

class FSTFinalStateIterRep
    : public FSTReferenceCounted {
public:
    FSTFinalStateIterRep() {}
    virtual ~FSTFinalStateIterRep() {}
    virtual void operator++(int) = 0;
    virtual bool valid() const = 0;
    virtual int state () const = 0;
};

class FSTFinalStateIter
    : public FSTReferenceCounter<FSTFinalStateIterRep> {
public:
    FSTFinalStateIter(FSTFinalStateIterRep* x=0)
        : _RC(x) {}
    FSTFinalStateIter(const FSTFinalStateIter& x)
        : _RC(x._rep) {}
    ~FSTFinalStateIter()
        {}
    void operator++()
        { (*_rep)++; }
    void operator++(int)
        { (*_rep)++; }
    int operator*() const
        { return _rep->state(); }
    bool valid() const
        { return _rep->valid(); }
    int state() const
        { return _rep->state(); }
};

class FSTGeneric;

class FSTGenericInt
    : public FSTReferenceCounted {
public:
    FSTGenericInt();
    virtual ~FSTGenericInt();

    // Methods supported by all FSTs:
    //
    virtual const fst::set<int>& input_alphabet() const = 0;
    virtual const fst::set<int>& output_alphabet() const = 0;
    bool is_mutable() const
        { return _is_mutable; }
    bool is_bidirectional() const
        { return _is_bidirectional; }
    FSTSemiring semiring() const
        { return _semiring; }
    int n_states() const
        { return _n_states; }
    int n_arcs() const
        { return _n_arcs; }
    virtual FSTStateIter states() const
        { return FSTStateIter(_n_states); }
    virtual int initial() const
        { return _initial; }
    virtual int n_arcs_out(int p) const;
    virtual FSTArcIter arcs_out(int p, FSTArcOrder o=FSTNone) const = 0;
    virtual FSTArc     arc_out(int p) const;
    virtual FSTArcIter arcs_out_with_input(int p, int i,
                                           FSTArcMode m=FSTEqual) const = 0;
    virtual FSTArcIter arcs_out_with_output(int p, int o,
                                            FSTArcMode m=FSTEqual) const = 0;
    virtual void arcs_out(FSTArcVector& result,
                          int p, FSTArcOrder o=FSTNone) const;
    virtual bool is_final(int p) const = 0;
    virtual float final_weight(int p) const = 0;
    virtual float potential(int p) const = 0;
    virtual void force() const = 0;
    virtual void reset() const = 0;
            void dump_ascii(const std::string& filename, bool header=true) const;
            void dump_ascii(const char* filename=0, bool header=true) const;
    virtual void dump_ascii(std::ostream& os, bool header=true) const;
            void dump_binary(const std::string& filename) const;
            void dump_binary(const char* filename=0) const;
    virtual void dump_binary(std::ostream& os) const;
            void dump_mmap(const std::string& filename, FSTArcOrder order) const;
    virtual void dump_mmap(const char* filename, FSTArcOrder order) const;
            void dump_dot(const std::string& filename, const char* init=0) const;
            void dump_dot(const char* filename=0, const char* init=0) const;
    virtual void dump_dot(std::ostream& os, const char* init=0) const;
    virtual void dead_states(fst::vector<int>& dead) const;

    virtual FSTGeneric reload(bool& flag);
            void enable_reload(const char* filename, time_t mtime);
            void train(std::istream& is, double floor=0);
            void train_twosided(std::istream& left_is, std::istream& right_is, const char* ABin_file, 
                                const char* Cin_file, double floor, 
                                const char* save_stats_dir, const char* load_stats_dir);
    virtual void restore_input_output_index();

    // Methods supported by mutable FSTs:
    //
    virtual void set_initial(int p);
    virtual void set_final(int p, float weight);
    void set_final(int p)
        { set_final(p, _semiring->one); }
    void clear_final(int p)
        { set_final(p, _semiring->zero); }
    virtual void set_potential(int p, float weight);
    int add_state()
        { return add_states(1); }
    virtual int  add_states(int n);
    virtual void add_arc(int p, int q, int i, int o, float w);
    void add_arc(int p, int q, int i=0, int o=0)
        { add_arc(p, q, i, o, _semiring->one); }
    void add_arc(int p, int q, const char* i, const char* o, float w)
        { add_arc(p, q, FSTstring_map[i], FSTstring_map[o], w); }
    void add_arc(int p, int q, const char* i, const char* o="")
        { add_arc(p, q, FSTstring_map[i], FSTstring_map[o]); }
    virtual void add_arcs(int p, FSTArcVector& arcs,
                          FSTArcOrder order=FSTNone);
    virtual void set_arcs(int p, FSTArcVector& arcs,
                          FSTArcOrder order=FSTNone);
    virtual void delete_state(int p);
    virtual void delete_states(const fst::vector<int>& states);
    void trim();
    void set_semiring(FSTSemiring semiring);
    
    // Methods supported by bidirectional FSTs:
    //
    virtual FSTFinalStateIter final_states() const;
    virtual FSTArcIter arcs_in(int p, FSTArcOrder o=FSTNone) const;
    virtual FSTArcIter arcs_in_with_input(int p, int i,
                                          FSTArcMode m=FSTEqual) const;
    virtual FSTArcIter arcs_in_with_output(int p, int o,
                                           FSTArcMode m=FSTEqual) const;

protected: // members
    std::string          _class_name;
    FSTSemiring     _semiring;
    mutable int     _n_states;
    mutable int     _n_arcs;
    int             _initial;
    bool            _is_mutable;
    bool            _is_bidirectional;
    std::string          _filename;
    time_t          _mtime;

protected: // methods
    static void _sort_arcs(FSTArcVector& arcs, FSTArcOrder order);
    virtual void _train_begin();
    virtual void _train_begin_left();
    virtual void _train_begin_right();
    virtual void _train_end();
    virtual void _train_normalize();
    virtual void _train_normalize_by_outputlabel();
    virtual void _train_update(fst::vector<double>& a_counts,
                               fst::vector<double>& f_counts,
                               double floor);
    void _dead_states_acyclic(fst::vector<int>& dead, const fst::vector<int>& order) const;
    void _dead_states_cyclic(fst::vector<int>& dead) const;
};

// FST Handle that provides reference counting.
//
class FSTGeneric
    : public FSTReferenceCounter<FSTGenericInt> {
public:
    FSTGeneric()
        : _RC(0) {}
    FSTGeneric(const FSTGeneric& x)
        : _RC(x._rep) {}
    FSTGeneric(FSTGenericInt* x)
        : _RC(x) {}
    ~FSTGeneric()
        {}
    bool reload();

    // See FSTGenericInt for the rest of the methods accessible via ->.
};

// Interface specific to composition.
//
class FSTComposeInt
    : public FSTGenericInt {
public:
    virtual ~FSTComposeInt();
    virtual std::pair<int,int> state_pair(int state) const = 0;
};

// Interface specific to ngrams.
//
class FSTNgramInt
    : public FSTGenericInt {
public:
    virtual ~FSTNgramInt();
    virtual bool ngram_weight(int p, int input, float& weight) const = 0;
    virtual int  ngram_next_state() const = 0;
    virtual void determinize() = 0;
    virtual void invert() = 0;
    virtual void subtract(const FSTNgramInt* x) = 0;
    virtual void quantize(int bits) = 0;
    virtual void dump_pngram(const char* vocab_file, const char* out_file, int weight_width) const = 0;
    virtual void dump_pngram(const fst::list<std::string>& vocab_list, FILE* fp, int weight_width) const = 0;
    virtual void add_classed_word(int word, int word_class, float weight) = 0;
};

typedef FSTArcOrder FSTAstarSide;

// For specifying arc predicate to FSTTopologicalSort.
//
class FSTArcPredicate {
public:
    virtual ~FSTArcPredicate() {}
    virtual bool operator()(const FSTArc* arc) const = 0;
};

class FSTMemory : public ReferenceCountedMemory {
    char* _alloc;
public:
    FSTMemory(const void* data, unsigned size, unsigned alignment);
    virtual ~FSTMemory();
    virtual void dump(const char* filename) const;
};

class FSTLetterToSound {

public:

    // Filename can specify a grapheme:phoneme FST or a graphone
    // (graphone/phoneme) n-gram FST.
    //
    FSTLetterToSound(const char* filename, int N = 10, float S = 0.01);

    ~FSTLetterToSound();

    FSTGeneric generateFst(const std::string& word);

    std::string generateString(const std::string& word);

protected:

    FSTGeneric g2gp;
    FSTGeneric gp;
    FSTGeneric gp2p;

    FSTGeneric g2p;

    int N;
    float S;

    void build_g2gp();
    void build_gp2p();
};

class FSTDictionary {
public:
    FSTDictionary(const char* lFile, const char* letterToSoundFiles, int N=5, float S=5);
    ~FSTDictionary();
    FSTGeneric compose(const FSTGeneric& G);

protected:
    FSTGeneric _L;
    fst::list<FSTLetterToSound*> _letterToSoundList;
    int _iwt;
    int _pound;
    int _poundStar;
    int _apostrophe_s;
    int _poundState;
    int _iwtState;

    void _addWords(const fst::set<int>& words);
    void _addWord(int w);
    void _addUnderbarredWord(int w, const char* word);
    void _addSimpleWord(int w, const char* word);
    void _merge(const FSTGeneric& fst, bool addIwt);
    FSTGeneric _stripLabel(const FSTGeneric& fst, int o);
    FSTGeneric _appendPossessive(FSTGeneric fst);
};

class FSTPocketCompiler {
public:
    FSTPocketCompiler(const char* mcpFile, const char* lFile, const char* letterToSoundFiles,
                      int weightBits=5);
    ~FSTPocketCompiler();
    FSTMemory* jsgfCompile(const char* jsgf);
    FSTMemory* fstCompile(const FSTGeneric& g);
    FSTMemory* jsgfCompileClass(const char* className, const char* jsgf);
    FSTMemory* listCompileClass(const char* className, const char* phrases); // comma-separated
    FSTMemory* fstCompileClass(const char* className, const FSTGeneric& lg);

private:
    FSTGeneric    _mcp;
    FSTGeneric    _mcpMain;
    FSTGeneric    _mcpClass;
    FSTDictionary _dict;
    int           _verbosity;
    int           _weightBits;

    FSTMemory* _jsgfCompile(const char* className, const char* jsgf);
    FSTMemory* _compile(const char* className, const FSTGeneric& g, bool isLG=false);
    FSTGeneric _delaySuffix(const FSTGeneric& x);
    FSTGeneric _fixReentryOutputs(const FSTGeneric& x);
};

// Function Prototypes ////////////////////////////////////////////////////////


extern void FSTFatal(const char* fmt, ...);
extern void FSTFatal(const char* fmt, va_list ap);
extern void FSTWarn(const char* fmt, ...);
extern void FSTWarn(const char* fmt, va_list ap);

extern void       FSTLabels(const char* filename);

extern FSTGeneric FSTBasic(FSTSemiring semiring=FSTMinPlus);
extern FSTGeneric FSTBasic(const FSTGeneric& x);
extern FSTGeneric FSTClearWeights(const FSTGeneric& x);
extern FSTGeneric FSTCopy(const FSTGeneric& x);
extern FSTGeneric FSTLoad(const std::string& file_or_exp);
extern FSTGeneric FSTLoad(const char* file_or_exp);
extern FSTGeneric FSTLoadFile(const std::string& filename);
extern FSTGeneric FSTLoadFile(const char* filename);
extern FSTGeneric FSTMutableCopy(const FSTGeneric& x);
extern FSTGeneric FSTProject(const FSTGeneric& x, FSTArcOrder input, FSTArcOrder output);
extern FSTGeneric FSTReverse(const FSTGeneric& x);
extern FSTGeneric FSTConvert_to_PlusTimes(const FSTGeneric& x);
extern FSTGeneric FSTConvert_to_MinPlus(const FSTGeneric& x);
extern FSTGeneric FSTConvert_to_MaxPlus(const FSTGeneric& x);
extern FSTGeneric FSTTopologicalSort(const FSTGeneric& x);
extern FSTGeneric FSTTopologicalSort(const FSTGeneric& x, const fst::set<int>& inputs);
extern FSTGeneric FSTTopologicalSort(const FSTGeneric& x, const FSTArcPredicate* predicate);
extern bool       FSTTopologicalSortFromInitial(fst::vector<int>& result, const FSTGeneric& x);
extern bool       FSTTopologicalSort(fst::vector<int>& result, const FSTGeneric& x);
extern bool       FSTTopologicalSort(fst::vector<int>& result, const FSTGeneric& x,
                                     const FSTArcPredicate* predicate);
extern bool       FSTTopologicalSort(fst::vector<int>& result, const FSTGeneric& x,
                                     const fst::set<int>& inputs);

extern FSTGeneric FSTPotential(FSTGeneric x);
extern void       FSTPotential(fst::vector<float>& potential, const FSTGeneric& x);

// The following are on-the-fly:
//
extern FSTGeneric FSTBackoff(const FSTGeneric& x, bool invert=false);
extern FSTGeneric FSTCache(const FSTGeneric& x);
extern FSTGeneric FSTCache2(const FSTGeneric& x);
extern FSTGeneric FSTCacheLRU(const FSTGeneric& x, int size=1000);
extern FSTGeneric FSTClosure(const FSTGeneric& x, bool plus=false);
extern FSTGeneric FSTCompose(const FSTGeneric& x, const FSTGeneric& y);
extern FSTGeneric FSTConcatenate(const FSTGeneric& x, const FSTGeneric& y);
extern FSTGeneric FSTConcatenate(const FSTGeneric& x, const FSTGeneric& y,
                                 const FSTGeneric& z);
extern FSTGeneric FSTConcatenate(const fst::vector<FSTGeneric>& inputs);
extern FSTGeneric FSTDeterminize(const FSTGeneric& x, int degree=1,
                                 bool fsa=false, float eps=1e-3, bool strict=false);
extern FSTGeneric FSTDeterminizeA(const FSTGeneric& x, int degree=1,
                                  float eps=1e-3, bool strict=false);
extern FSTGeneric FSTDynamic(const FSTGeneric& main_fst);
extern FSTGeneric FSTInvert(const FSTGeneric& x);
extern FSTGeneric FSTNgramDeterminize(FSTGeneric x);
extern FSTGeneric FSTNgramInvert(FSTGeneric x);
extern FSTGeneric FSTPushWeights(const FSTGeneric& x);
extern FSTGeneric FSTRemoveEpsilons(const FSTGeneric& x, bool fsa=false,
                                    float delta=1e-3);
extern FSTGeneric FSTRemoveEpsilonsA(const FSTGeneric& x, float delta=1e-3);
extern FSTGeneric FSTRtn(const FSTGeneric& root_fst,
                         const fst::vector<std::string>& nonterminals,
                         const fst::vector<FSTGeneric>& nonterminal_fsts,
                         bool tagged=false);
extern FSTGeneric FSTUnion(const FSTGeneric& x, const FSTGeneric& y);
extern FSTGeneric FSTUnion(const FSTGeneric& x, const FSTGeneric& y,
                           const FSTGeneric& z);
extern FSTGeneric FSTUnion(const fst::vector<FSTGeneric>& fsts);
extern FSTGeneric FSTUnion(const fst::vector<FSTGeneric>& fsts,
                           const fst::vector<float>& weights);
extern FSTGeneric FSTUnion(const fst::vector<FSTGeneric>& fsts,
                           const fst::vector<int>& input_tags,
                           const fst::vector<int>& output_tags);
extern FSTGeneric FSTUnion(const fst::vector<FSTGeneric>& fsts,
                           const fst::vector<float>& weights,
                           const fst::vector<int>& input_tags,
                           const fst::vector<int>& output_tags);

extern FSTGeneric FSTExpression(const std::string& input);
extern bool       FSTExpression(const std::string& input, FSTGeneric& result,
                                std::string& error);

extern FSTGeneric FSTRegularExpression(const std::string& input);
extern bool       FSTRegularExpression(const std::string& input, FSTGeneric& result,
                                       std::string& error);

// The following are in-place if x is mutable, otherwise operates on and
// returns a copy:
//
extern FSTGeneric FSTMinimize(FSTGeneric x, float eps=1e-3);
extern FSTGeneric FSTTrim(FSTGeneric x);

// Other functions:
//
extern FSTGeneric   FSTDelayOutputs(const FSTGeneric& x);
extern void         FSTDeterminizeStateSet(const FSTGeneric& x, int p, fst::set<int>& result);
extern void         FSTRemoveEpsilonsStateSet(const FSTGeneric& x, int p, fst::set<int>& result);
extern void         FSTFactorRL(const FSTGeneric& x, FSTGeneric& R, FSTGeneric& L);
extern bool         FSTIsDynamic(const FSTGeneric& x);
extern FSTGeneric   FSTDynamicGet(const FSTGeneric& x, const std::string& class_name);
extern FSTGeneric   FSTDynamicGet(const FSTGeneric& x, int class_name);
extern void         FSTDynamicSet(FSTGeneric& x, const std::string& class_name,
                                  const FSTGeneric& class_fst);
extern void         FSTDynamicSet(FSTGeneric& x, int class_name,
                                  const FSTGeneric& class_fst);
extern void         FSTRtnSet(FSTGeneric& rtn, const std::string& name, const FSTGeneric& fst);
extern FSTGeneric   FSTQuantizeWeights(int bits, FSTGeneric& in);
extern FSTGeneric   FSTOptimize(const FSTGeneric& x, int degree=1, float eps=1e-3);
extern FSTGeneric   FSTOptimize2(const FSTGeneric& x, int degree=1, float eps=1e-3);
extern FSTGeneric   FSTOptimizeA(const FSTGeneric& x, int degree=1, float eps=1e-3);
extern FSTGeneric   FSTOptimize2A(const FSTGeneric& x, int degree=1, float eps=1e-3);
extern FSTGeneric   FSTAstarNbest(const FSTGeneric& x, FSTAstarSide unique,
                                  int N, float threshold);
extern FSTGeneric   FSTAstarGraph(const FSTGeneric& x, FSTAstarSide project,
                                  float threshold, int max_pops);
extern FSTGeneric   FSTSearch(const FSTGeneric& x);
extern FSTGeneric   FSTSearch(const FSTGeneric& a, const FSTGeneric& b,
                              float vprune, int vnodes, bool verbose=true, 
                              fst::vector<int>* nodes=NULL);
extern FSTGeneric   FSTSearchGraph(const FSTGeneric& a, const FSTGeneric& b,
                              float vprune, int vnodes, bool verbose=true);

extern void         FSTNgramSubtract(FSTGeneric x, const FSTGeneric& y);
extern void         FSTNgramAddClassedWord(FSTGeneric& ngram, int word, int word_class, float weight);

extern const FSTComposeInt* FSTComposeCast(const FSTGeneric& x);
extern const FSTNgramInt*   FSTNgramCast(const FSTGeneric& x);

extern void         FSTNgramDumpPacked(const FSTGeneric& x,
                                       const char* vocab_file, const char* out_file,
                                       int weight_width=5);
extern void         FSTNgramDumpPacked(const FSTGeneric& x,
                                       const fst::list<std::string>& vocab_list, FILE* fp,
                                       int weight_width=5);

extern void         train_twosided(std::istream& left_is, std::istream& right_is, 
                                   const char* Ain_file, const char* Cin_file, double floor, 
                                   const char* save_stats_dir, const char* load_stats_dir);

extern FSTGeneric FSTDynamicAugmentMain(const FSTGeneric& in);
extern FSTGeneric FSTDynamicAugmentClass(const FSTGeneric& in);

extern FSTGeneric FSTPackToFile(const char* pfst_file, const char* out_label_file,
                                const FSTGeneric& in, int weight_bits, const fst::set<int>& input_alphabet,
                                bool swap, int verbosity, int version=0);

extern FSTMemory* FSTPackToMemory(const FSTGeneric& in, int weight_bits, const fst::set<int>& input_alphabet,
                                  int verbosity, int version=0);

extern bool FSTJsgfParseString(FSTGeneric& result, std::string& error_msg, const char* input,
                               bool ignore_tags, bool ignore_words, bool rtn_parse, bool pause12,
                               bool allow_recursion=true);
extern bool FSTJsgfParseFile(FSTGeneric& result, const std::string& filename,
                             bool ignore_tags, bool ignore_words, bool rtn_parse, bool pause12,
                             bool allow_recursion=true);

// Stream output (like dump_ascii method).
//
extern std::ostream& operator<<(std::ostream& os, const FSTGeneric& x);

// Binary I/O.
//
extern void           FSTDumpShort(short, std::ostream&);
extern void           FSTDumpUnsignedShort(unsigned short, std::ostream&);
extern void           FSTDumpInt(int, std::ostream&);
extern void           FSTDumpUnsignedInt(unsigned int, std::ostream&);
extern void           FSTDumpCompressedInt(int, std::ostream&);
extern void           FSTDumpCompressedUnsigned(unsigned int, std::ostream&);
extern void           FSTDumpFloat(float, std::ostream&);
extern void           FSTDumpDouble(double, std::ostream&);
extern void           FSTDumpString(const std::string&, std::ostream&);
extern void           FSTDumpChars(const char*, std::ostream&);
extern short          FSTLoadShort(std::istream&);
extern unsigned short FSTLoadUnsignedShort(std::istream&);
extern int            FSTLoadInt(std::istream&);
extern unsigned int   FSTLoadUnsignedInt(std::istream&);
extern int            FSTLoadCompressedInt(std::istream&);
extern unsigned int   FSTLoadCompressedUnsigned(std::istream&);
extern float          FSTLoadFloat(std::istream&);
extern double         FSTLoadDouble(std::istream&);
extern void           FSTLoadString(std::istream&, std::string&);
extern std::string    FSTLoadString(std::istream&);
extern char*          FSTLoadChars(std::istream&);

#endif // __FST_h__

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 80
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
