////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

#ifndef __FSTAllocator_h__
#define __FSTAllocator_h__

#include <iostream>
#include <stdlib.h>
#include <assert.h>

#ifdef __INSURE__
#  define FST_ALLOCATOR_DISABLE
#endif

#define FSTAllocator_max_bytes size_t(128)
#define FSTAllocator_word_size sizeof(void*)
#define FSTAllocator_max_words size_t(FSTAllocator_max_bytes / FSTAllocator_word_size)

//
// Note: unlike STL's alloc, this is not thread-safe.
//

class FSTAllocator {

public:
    static void* allocate(size_t bytes)
    {
#ifdef FST_ALLOCATOR_DISABLE
        return new char[bytes];
#else
        size_t words = (bytes + FSTAllocator_word_size-1) / FSTAllocator_word_size;
        if (words == 0)
            return 0;
        else if (words <= FSTAllocator_max_words)
            return _allocate(words);
        else
            return new char[bytes];
#endif
    }
    static void deallocate(void* pointer, size_t bytes)
    {
        if (pointer && bytes > 0) {
#ifdef FST_ALLOCATOR_DISABLE
            delete[] (char*) pointer;
#else
            size_t words = (bytes + FSTAllocator_word_size-1) / FSTAllocator_word_size;
            if (words <= FSTAllocator_max_words)
                _deallocate(pointer, words);
            else
                delete[] (char*) pointer;
#endif
        }
    }
    static void* reallocate(void*, size_t, size_t);
    static void  gc();
    static void  status(std::ostream& os);

public:
#if 0
    static const size_t max_bytes = 128;
    static const size_t word_size = sizeof(void*);
    static const size_t max_words = max_bytes / word_size;
#endif

protected:
    static void* _allocate(size_t words);
    static void  _deallocate(void* pointer, size_t words);
};

// The following uses STL's allocator instead of our own:
// #define FST_ALLOCATOR alloc

#ifdef FST_ALLOCATOR_DISABLE
#define FSTAllocator_new_delete(type)
#else
#define FST_ALLOCATOR FSTAllocator
#define FSTAllocator_new_delete(type)                     \
    void* operator new(size_t size) {                     \
        assert(size == sizeof(type));                     \
        return FST_ALLOCATOR::allocate(sizeof(type));     \
    }                                                     \
    void operator delete(void* pointer) {                 \
        FST_ALLOCATOR::deallocate(pointer, sizeof(type)); \
    }
#endif

#endif // __FSTAllocator_h__

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 80
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
