////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

#ifndef __FSTPronRules_h__
#define __FSTPronRules_h__

#include <iostream>
#include <string>
#include "fst/FST.h"
#include "fst/vector"
#include "fst/map"
#include "fst/set"

class FSTPronRules {

public: // types
    struct _Rule;

private: // types
    typedef fst::vector<_Rule*> _RuleVector;
    typedef fst::map<int, _RuleVector> _RuleMap;

private: // members
    fst::vector<int>     _alphabet;
    fst::vector<int>     _initial;
    fst::vector<int>     _final;
    fst::vector<int>     _ignore;
    _RuleMap             _rule_map;
    fst::vector<_Rule*>  _rules;
    fst::map<int, int>   _links_f;              // final link -> initial link
    fst::map<int, int>   _links_l;              // label -> leftified label
    fst::map<int, int>   _links_r;              // label -> rightified label
    fst::set<int>        _links_s;              // initial links and left links
    fst::set<int>        _links_a;              // all links
    int                  _verbosity;

    // Compilation results
    //
    FSTGeneric      _Pr;                        // right-to-left part of rules
    FSTGeneric      _Pl;                        // left-to-right part of rules
    FSTGeneric      C, E, L, R;                 // intermediate results

public: // methods
    FSTPronRules(const std::string& filename, int verbosity=1);
    ~FSTPronRules();
    bool bad() const;
    void set_alphabet(const fst::set<int>& labels);
    void set_initial(const fst::set<int>& labels);
    void set_final(const fst::set<int>& labels);
    void set_ignore(const fst::set<int>& labels);
    void add_link(int final, int initial);
    int  expr_context(int label, bool left);
    void add_rule(const fst::set<int>& left, int label, const fst::set<int>& right,
                  FSTGeneric fst);
    FSTGeneric apply(FSTGeneric fst) const;
    void dump(const char* a_file, const char* b_file, const char* c_file) const;
    FSTGeneric single_fst() const;
    std::ostream& print(std::ostream& os) const;

private: // methods
    bool _parse(const std::string& filename);
    void _compile();
    FSTGeneric _construct_C();
    FSTGeneric _construct_E();
    void _construct_LR(FSTGeneric& L, FSTGeneric& R);
    void _check_coverage(FSTGeneric CE, FSTGeneric L, FSTGeneric R);
    void _compute_transitions(fst::set< std::pair<int, int> >& transitions, FSTGeneric fsa);
    void _add_ignore_loops(FSTGeneric x);
    FSTGeneric _label_fst(const FSTGeneric& in, int label);
    static const _Rule* _find_first_left_match(const _RuleVector& rules, int l);
    static void _find_left_matches(
        _RuleVector& result, const _RuleVector& rules, int l);
    static void _find_right_matches(
        _RuleVector& result, const _RuleVector& rules, int r);
    static bool _any_right_matches(const _RuleVector& rules, int l);
    static bool _any_matches(const _RuleVector& rules, int l, int r);
    static bool _compatible(const _Rule* a, const _Rule* b);
    inline static bool _contains(const fst::vector<int>& s, int label);
};

#endif // __FSTPronRules_h__

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 79
// indent-tabs-mode: nil
// c-basic-offset: 4
// End:
