////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

#ifndef __FSTReferenceCount_h__
#define __FSTReferenceCount_h__

// The following is intended to be a base class.  It provides incref()
// and decref() public methods.
//
class FSTReferenceCounted {
public:
    FSTReferenceCounted() : _refcnt(0) {}
    virtual ~FSTReferenceCounted() {}
    void incref() const { _refcnt++; }
    void decref() const { if (--_refcnt <= 0) delete this; }
    int  refcnt() const { return _refcnt; }
private:
    mutable int _refcnt;
};

// The following is intended to be a base class.  It provides for
// reference counting of its _rep for copy construction, destruction,
// and assignment operations.  Template parameter T is the type of the
// _rep, most likely a class derived from FSTReferenceCounted; it must
// provide incref() and decref() methods.
//
template<class T>
class FSTReferenceCounter {
public:
    explicit FSTReferenceCounter(T* x=0) : _rep(x) {
    if (_rep) _rep->incref();
    }
    explicit FSTReferenceCounter(const FSTReferenceCounter& x) : _rep(x._rep) {
    if (_rep) _rep->incref();
    }
    ~FSTReferenceCounter() {
    if (_rep) _rep->decref();
    }
    const FSTReferenceCounter& operator = (const FSTReferenceCounter& x) {
        if (x._rep) x._rep->incref();
        if (_rep) _rep->decref();
        _rep = x._rep;
    return *this;
    }
    T* operator->() {
        return _rep;
    }
    const T* operator->() const {
        return _rep;
    }
    T* operator*() {
        return _rep;
    }
    const T* operator*() const {
        return _rep;
    }
    operator bool() const {
        return _rep != 0;
    }
    bool operator!() const {
        return _rep == 0;
    }
    bool operator==(const FSTReferenceCounter& x) const {
        return _rep == x._rep;
    }
    bool operator!=(const FSTReferenceCounter& x) const {
        return _rep != x._rep;
    }
    bool operator<(const FSTReferenceCounter& x) const {
        return _rep < x._rep;
    }
    bool operator<=(const FSTReferenceCounter& x) const {
        return _rep <= x._rep;
    }
    bool operator>(const FSTReferenceCounter& x) const {
        return _rep > x._rep;
    }
    bool operator>=(const FSTReferenceCounter& x) const {
        return _rep >= x._rep;
    }
    T* pointer() {
        return _rep;
    }
    const T* pointer() const {
        return _rep;
    }
    void destroy() {
    if (_rep) _rep->decref();
    _rep = 0;
    }
    int refcnt() const {
    return (_rep) ? _rep->refcnt() : 0;
    }

protected:
    typedef FSTReferenceCounter _RC;
    T* _rep;
};

#endif // __FSTReferenceCount_h__

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 80
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
