////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

#ifndef __FSTRules_h__
#define __FSTRules_h__

#include "fst/list"
#include "fst/map"
#include "fst/set"

using namespace std;

// Defines to encode rule type as integer.  A rule type has three components:
// direction (left-to-right, right-to-left, simultaneous), mandatory, and
// directed.  You bit-wise or (|) the components together.
//
#define DIRECTION_LR        0x01
#define DIRECTION_RL        0x02
#define MANDATORY           0x04
#define DIRECTED            0x08

// Convenient typedefs.
//
typedef fst::set<int> Set;
typedef const Set& SetRef;
typedef FSTGeneric Fst;
typedef const Fst& FstRef;

typedef fst::map<int, Fst> FstMap;
         
// Collection of rules.
//
struct FSTRules {

    //
    // Types used in parse tree of entire rules files.
    //

    // Node: abstract base class for all parse tree nodes.
    //
    class Node {
    public:
        virtual ~Node() {}
        virtual Fst eval(FSTRules* context) = 0;
    };

    typedef Node*       NodePtr;
    typedef list<Node*> NodeList;

    // SymbolNode: a (single symbol, can be epsilon)
    //
    class SymbolNode : public Node {
        int _label;
    public:
        SymbolNode(int label) : _label(label) {}
        virtual Fst eval(FSTRules* context);
    };
    
    // SetNode: {a b c} (set of single symbols)
    //
    class SetNode : public Node {
        Set _set;
    public:
        SetNode() {}
        virtual Fst eval(FSTRules* context);
        void insert(int label) { _set.insert(label); }
        Set::const_iterator begin() { return _set.begin(); }
        Set::const_iterator end() { return _set.end(); }
    };
    
    // DotNode: . (any single symbol)
    //    
    class DotNode : public Node {
    public:
        DotNode() {}
        virtual Fst eval(FSTRules* context);
    };    
    
    // OptionalNode: a ?
    //
    class OptionalNode : public Node {
        Node* _node;
    public:
        OptionalNode(Node* node) : _node(node) {}
        virtual ~OptionalNode() { delete _node; }
        virtual Fst eval(FSTRules* context);
    };
    
    // StarNode: a *
    //
    class StarNode : public Node {
        Node* _node;
    public:
        StarNode(Node* node) : _node(node) {}
        virtual ~StarNode() { delete _node; }
        virtual Fst eval(FSTRules* context);
    };
    
    // PlusNode: a +
    //
    class PlusNode : public Node {
        Node* _node;
    public:
        PlusNode(Node* node) : _node(node) {}
        virtual ~PlusNode() { delete _node; }
        virtual Fst eval(FSTRules* context);
    };
    
    // ComplementNode: a !
    //
    class ComplementNode : public Node {
        Node* _node;
    public:
        ComplementNode(Node* node) : _node(node) {}
        virtual ~ComplementNode() { delete _node; }
        virtual Fst eval(FSTRules* context);
    };
    
    // ConcatenateNode: a b c
    //
    class ConcatenateNode : public Node {
        NodeList _nodes;
    public:
        ConcatenateNode() {}
        virtual ~ConcatenateNode();
        virtual Fst eval(FSTRules* context);
        void prepend(Node* node) { if (node) _nodes.push_front(node); }
    };
    
    // UnionNode: a | b
    //
    class UnionNode : public Node {
        Node* _node_a;
        Node* _node_b;
    public:
        UnionNode(Node* node_a, Node* node_b)
            : _node_a(node_a), _node_b(node_b) {}
        virtual ~UnionNode() { delete _node_a; delete _node_b; }
        virtual Fst eval(FSTRules* context);
    };
    
    // IntersectNode: a & b
    //
    class IntersectNode : public Node {
        Node* _node_a;
        Node* _node_b;
    public:
        IntersectNode(Node* node_a, Node* node_b)
            : _node_a(node_a), _node_b(node_b) {}
        virtual ~IntersectNode() { delete _node_a; delete _node_b; }
        virtual Fst eval(FSTRules* context);
    };
    
    // AssignNode: a = y
    //
    class AssignNode : public Node {
        int _label;
        Node* _node;
    public:
        AssignNode(int label, Node* node) : _label(label), _node(node) {}
        ~AssignNode() { delete _node; }
        virtual Fst eval(FSTRules* context);
    };

    // IgnoreNode: ignore expr
    //
    class IgnoreNode : public Node {
        Node* _ignore_node;
        Node* _series_node;
    public:
        IgnoreNode(Node* ignore_node, Node* series_node)
            : _ignore_node(ignore_node), _series_node(series_node) {}
        ~IgnoreNode() { delete _ignore_node; delete _series_node; }
        virtual Fst eval(FSTRules* context);
    };

    // SeriesNode: rules in series.
    //
    class SeriesNode : public Node {
        NodeList _nodes;
    public:
        SeriesNode() {}
        virtual Fst eval(FSTRules* context);
        void prepend(Node* node) { _nodes.push_front(node); }
        void append(Node* node) { _nodes.push_back(node); }
    };

    // ExprNode: expr
    //
    class ExprNode : public Node {
        int _line;
        Node* _node;
    public:
        ExprNode(int line, Node* node) : _line(line), _node(node) {}
        virtual ~ExprNode() { delete _node; }
        virtual Fst eval(FSTRules* context);
    };

    // RuleNode: phi -> psi or phi -> psi / lambda ___ rho
    //
    class RuleNode : public Node {
        int _line;
        int _type;
        NodePtr _node_phi, _node_psi, _node_lambda, _node_rho;
        Fst _phi, _psi, _lambda, _rho;
    public:
        RuleNode(int line, int type,
                 Node* phi, Node* psi, Node* lambda=0, Node* rho=0);
        virtual ~RuleNode();
        virtual Fst eval(FSTRules* context);
    private:
        Fst _build_s(SetRef sigma, bool mandatory);
        Fst _build_d(SetRef sigma, bool mandatory, bool directed,
                     bool right_to_left);
    };

    //
    // Data Members.
    //

    Set sigma;
    Set ignore;
    FstMap assignments;
    Node* root;
    Fst fst;
    Set warned;
    bool warn;

    //
    // Function Members.
    //

    FSTRules(bool warn=true);
    ~FSTRules();
    void eval();

};

// Functions in FSTRules.y (that bison puts into FSTRulesParser.cpp):
//
bool FSTRulesParseFile(
    Fst& result, const string& filename, bool program=true, bool warn=true);
bool FSTRulesParseString(
    Fst& result, const string& input, bool program=true, bool warn=true);

// Functions in FSTRules.l (that flex puts into FSTRulesLexer.cpp):
//
void FSTRulesSetupLexerFile(const string& filename);
void FSTRulesSetupLexerString(const string& input);

#endif // __FSTRules_h__

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 79
// indent-tabs-mode: nil
// c-basic-offset: 4
// End:
