////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

#ifndef __FSTSemirings_h__
#define __FSTSemirings_h__

#include <cmath>

// Properties of FSTSemirings
//
// PLUS combines scores in parallel.  TIMES combines scores in series.
// ZERO is the neutral element for PLUS: ZERO PLUS x = x and x PLUS
// ZERO = x.  ONE is the neutral element for TIMES: ONE TIMES x = x
// and x TIMES ONE = x.  Additionally, ZERO TIMES x = ZERO and x TIMES
// ZERO = ZERO.  Idempotent means that ONE PLUS x = ONE.

struct FSTSemiringBase {
    FSTSemiringBase(float z, float o, float e, const char* n);
    virtual ~FSTSemiringBase();
    virtual float times    (float x, float y) const = 0;
    virtual float times    (float x, float y, float z) const = 0;
    virtual float plus     (float x, float y) const = 0;
    virtual float plus     (float x, float y, float z) const = 0;
    virtual float closure  (float x) const = 0;
    virtual float divide   (float x, float y) const = 0;
    virtual float invert   (float x) const = 0;
    virtual float better   (float x, float y) const = 0;
    virtual bool  is_better(float x, float y) const = 0;
    virtual bool  is_better(float x, float y, float delta) const = 0;
    virtual bool  is_equal (float x, float y) const = 0;
    virtual bool  is_equal (float x, float y, float delta) const = 0;
    virtual float convert  (float x) const = 0;
    virtual float unconvert(float x) const = 0;
    virtual bool  idempotent() const = 0;
    const float zero;
    const float one;
    const float delta;
    const char* name;
};

typedef const FSTSemiringBase* FSTSemiring;

struct FSTMinPlusSemiring : FSTSemiringBase {
    FSTMinPlusSemiring();
    virtual ~FSTMinPlusSemiring();
    virtual float times    (float x, float y) const;
    virtual float times    (float x, float y, float z) const;
    virtual float plus     (float x, float y) const;
    virtual float plus     (float x, float y, float z) const;
    virtual float closure  (float x) const;
    virtual float divide   (float x, float y) const;
    virtual float invert   (float x) const;
    virtual float better   (float x, float y) const;
    virtual bool  is_better(float x, float y) const;
    virtual bool  is_better(float x, float y, float delta) const;
    virtual bool  is_equal (float x, float y) const;
    virtual bool  is_equal (float x, float y, float delta) const;
    virtual float convert  (float x) const;
    virtual float unconvert(float x) const;
    virtual bool  idempotent() const;
};

struct FSTMaxPlusSemiring : FSTSemiringBase {
    FSTMaxPlusSemiring();
    virtual ~FSTMaxPlusSemiring();
    virtual float times    (float x, float y) const;
    virtual float times    (float x, float y, float z) const;
    virtual float plus     (float x, float y) const;
    virtual float plus     (float x, float y, float z) const;
    virtual float closure  (float x) const;
    virtual float divide   (float x, float y) const;
    virtual float invert   (float x) const;
    virtual float better   (float x, float y) const;
    virtual bool  is_better(float x, float y) const;
    virtual bool  is_better(float x, float y, float delta) const;
    virtual bool  is_equal (float x, float y) const;
    virtual bool  is_equal (float x, float y, float delta) const;
    virtual float convert  (float x) const;
    virtual float unconvert(float x) const;
    virtual bool  idempotent() const;
};

struct FSTPlusTimesSemiring : FSTSemiringBase {
    FSTPlusTimesSemiring();
    virtual ~FSTPlusTimesSemiring();
    virtual float times    (float x, float y) const;
    virtual float times    (float x, float y, float z) const;
    virtual float plus     (float x, float y) const;
    virtual float plus     (float x, float y, float z) const;
    virtual float closure  (float x) const;
    virtual float divide   (float x, float y) const;
    virtual float invert   (float x) const;
    virtual float better   (float x, float y) const;
    virtual bool  is_better(float x, float y) const;
    virtual bool  is_better(float x, float y, float delta) const;
    virtual bool  is_equal (float x, float y) const;
    virtual bool  is_equal (float x, float y, float delta) const;
    virtual float convert  (float x) const;
    virtual float unconvert(float x) const;
    virtual bool  idempotent() const;
};

struct FSTPlusTimesLogSemiring : FSTSemiringBase {
    FSTPlusTimesLogSemiring();
    virtual ~FSTPlusTimesLogSemiring();
    virtual float times    (float x, float y) const;
    virtual float times    (float x, float y, float z) const;
    virtual float plus     (float x, float y) const;
    virtual float plus     (float x, float y, float z) const;
    virtual float closure  (float x) const;
    virtual float divide   (float x, float y) const;
    virtual float invert   (float x) const;
    virtual float better   (float x, float y) const;
    virtual bool  is_better(float x, float y) const;
    virtual bool  is_better(float x, float y, float delta) const;
    virtual bool  is_equal (float x, float y) const;
    virtual bool  is_equal (float x, float y, float delta) const;
    virtual float convert  (float x) const;
    virtual float unconvert(float x) const;
    virtual bool  idempotent() const;
};

#endif // __FSTSemirings_h__

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 80
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
