////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

#ifndef __FSTSymbolString_h__
#define __FSTSymbolString_h__

#include "fst/FSTSTL.h"
#include "fst/FSTAllocator.h"

//
// FixedLengthString<N> - string of output symbols of length <= N
//
template<int N>
class FixedLengthString {

public:
    typedef int* iterator;
    typedef const int* const_iterator;

public:
    FixedLengthString() : n(0) {}
    FixedLengthString(const FixedLengthString& x) : n(x.n) {
    for (int i = n - 1; i >= 0; i--)
        rep[i] = x.rep[i];
    }
    FixedLengthString(int x) : n(0) {
    if (x) rep[n++] = x;
    }
    FixedLengthString& operator = (const FixedLengthString& x) {
    n = x.n;
    for (int i = n - 1; i >= 0; i--)
        rep[i] = x.rep[i];
    return *this;
    }
    FixedLengthString& operator = (int x) {
    n = 0;
    if (x) rep[n++] = x;
    return *this;
    }
    bool operator == (const FixedLengthString& x) const {
    if (n != x.n) return false;
    for (int i = 0; i < n; i++)
        if (rep[i] != x.rep[i])
        return false;
    return true;
    }
    bool operator == (int x) const {
    return (n == 0 && x == 0) || (n == 1 && rep[0] == x);
    }
    bool operator < (const FixedLengthString& x) const {
    for (int i = 0; i < n && i < x.n; i++) {
        if (rep[i] < x.rep[i]) return true;
        if (x.rep[i] < rep[i]) return false;
    }
    return n < x.n;
    }
    bool append(const FixedLengthString& x) {
    if (n + x.n > N)
        return false;
    for (int i = 0; i < x.n; i++)
        rep[n + i] = x.rep[i];
    n += x.n;
    return true;
    }
    bool append(int x) {
    if (x) {
        if (n + 1 > N)
        return false;
        rep[n++] = x;
    }
    return true;
    }
    bool prepend(int x) {
    if (x) {
        if (n + 1 > N)
        return false;
            for (int i = n - 1; i >= 0; i--)
                rep[i + 1] = rep[i];
            rep[0] = x;
            n++;
    }
    return true;
    }
    int front() const {
    return n == 0 ? 0 : rep[0];
    }
    void shift() {
    if (n > 0) {
        n--;
        for (int i = 0; i < n; i++)
        rep[i] = rep[i + 1];
    }
    }
    int size() const {
    return n;
    }
    iterator begin() {
    return rep;
    }
    const_iterator begin() const {
    return rep;
    }
    iterator end() {
    return rep + n;
    }
    const_iterator end() const {
    return rep + n;
    }
    FSTAllocator_new_delete(FixedLengthString<N>);

protected:
    int rep[N];
    int n;

};

//
// Specialized FixedLengthString<1>:
//
template<>
class FixedLengthString<1> {

public:
    typedef int* iterator;
    typedef const int* const_iterator;

public:
    FixedLengthString() : rep(0) {}
    FixedLengthString(const FixedLengthString& x) : rep(x.rep) {}
    FixedLengthString(int x) : rep(x) {}
    FixedLengthString& operator = (FixedLengthString x) {
    rep = x.rep;
    return *this;
    }
    FixedLengthString& operator = (int x) {
    rep = x;
    return *this;
    }
    bool operator == (FixedLengthString x) const {
    return rep == x.rep;
    }
    bool operator == (int x) const {
    return rep == x;
    }
    bool operator < (FixedLengthString x) const {
    return rep < x.rep;
    }
    bool operator < (int x) const {
    return rep < x;
    }
    bool append(FixedLengthString x) {
    if (rep && x.rep)
        return false;
    if (x.rep)
        rep = x.rep;
    return true;
    }
    bool append(int x) {
    if (rep && x)
        return false;
    if (x)
        rep = x;
    return true;
    }
    bool prepend(int x) {
        return append(x);
    }
    int front() const {
    return rep;
    }
    void shift() {
    rep = 0;
    }
    int size() const {
    return rep > 0;
    }
    iterator begin() {
    return &rep;
    }
    const_iterator begin() const {
    return &rep;
    }
    iterator end() {
    return &rep + (rep > 0);
    }
    const_iterator end() const {
    return &rep + (rep > 0);
    }
    FSTAllocator_new_delete(FixedLengthString<1>);

protected:
    int rep;
};

//
// Specialized FixedLengthString<2>:
//
template<>
class FixedLengthString<2> {

public:
    typedef int* iterator;
    typedef const int* const_iterator;

public:
    FixedLengthString() {
    rep[0] = rep[1] = 0;
    }
    FixedLengthString(const FixedLengthString& x) {
    rep[0] = x.rep[0];
    rep[1] = x.rep[1];
    }
    FixedLengthString(int x) {
    rep[0] = x;
    rep[1] = 0;
    }
    FixedLengthString& operator = (FixedLengthString x) {
    rep[0] = x.rep[0];
    rep[1] = x.rep[1];
    return *this;
    }
    FixedLengthString& operator = (int x) {
    rep[0] = x;
    rep[1] = 0;
    return *this;
    }
    bool operator == (FixedLengthString x) const {
    return rep[0] == x.rep[0] && (!rep[0] || rep[1] == x.rep[1]);
    }
    bool operator == (int x) const {
    return rep[0] == x && (!rep[0] || !rep[1]);
    }
    bool operator < (FixedLengthString x) const {
    return rep[0] < x.rep[0] || (rep[0] == x.rep[0] && rep[1] < x.rep[1]);
    }
    bool operator < (int x) const {
    return rep[0] < x;
    }
    bool append(FixedLengthString x) {
    if (x.rep[0]) {
        if (rep[0] && (rep[1] || x.rep[1]))
        return false;
        if (rep[0]) {
        rep[1] = x.rep[0];
        } else {
        rep[0] = x.rep[0];
        rep[1] = x.rep[1];
        }
    }
    return true;
    }
    bool append(int x) {
    if (x) {
        if (rep[0] && rep[1])
        return false;
        if (rep[0]) {
        rep[1] = x;
        } else {
        rep[0] = x;
        rep[1] = 0;
        }
    }
    return true;
    }
    bool prepend(int x) {
    if (x) {
        if (rep[0] && rep[1])
        return false;
        if (rep[0]) {
                rep[1] = rep[0];
                rep[0] = x;
        } else {
        rep[0] = x;
        rep[1] = 0;
        }
    }
    return true;
    }
    int front() const {
    return rep[0];
    }
    void shift() {
    rep[0] = rep[1];
    rep[1] = 0;
    }
    int size() const {
    return rep[0] ? (rep[1] ? 2 : 1) : 0;
    }
    iterator begin() {
    return rep;
    }
    const_iterator begin() const {
    return rep;
    }
    iterator end() {
    return rep + size();
    }
    const_iterator end() const {
    return rep + size();
    }
    FSTAllocator_new_delete(FixedLengthString<2>);

protected:
    int rep[2];
};

//
// VariableLengthString - string of output symbols of unlimited length
//
class VariableLengthString {

    typedef std::vector<int> rep_type;

public:
    typedef rep_type::iterator iterator;
    typedef rep_type::const_iterator const_iterator;

public:
    VariableLengthString() {}
    VariableLengthString(const VariableLengthString& x) : rep(x.rep) {}
    VariableLengthString(int x) {
    if (x)
        rep.push_back(x);
    }
    VariableLengthString& operator = (const VariableLengthString& x) {
    rep = x.rep;
    return *this;
    }
    VariableLengthString& operator = (int x) {
    if (x) {
        rep.erase(rep.begin(), rep.end());
        rep.push_back(x);
    }
    return *this;
    }
    bool operator == (const VariableLengthString& x) const {
    return rep == x.rep;
    }
    bool operator == (int x) const {
    return (x == 0 && rep.size() == 0) || (x == rep[0] && rep.size() == 1);
    }
    bool operator < (const VariableLengthString& x) const {
    return rep < x.rep;
    }
    bool append(const VariableLengthString& x) {
    for (rep_type::const_iterator i = x.rep.begin();
         i != x.rep.end(); i++)
        rep.push_back(*i);
    return true;
    }
    bool append(int x) {
    if (x)
        rep.push_back(x);
    return true;
    }
    int front() const {
    return rep.size() > 0 ? rep[0] : 0;
    }
    void shift() {
    if (rep.size() > 0)
        rep.erase(rep.begin());
    }
    int size() const {
    return rep.size();
    }
    iterator begin() {
    return rep.begin();
    }
    const_iterator begin() const {
    return rep.begin();
    }
    iterator end() {
    return rep.end();
    }
    const_iterator end() const {
    return rep.end();
    }
    FSTAllocator_new_delete(VariableLengthString);

protected:
    rep_type rep;
};

#endif // __FSTSymbolString_h__

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 80
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
