////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

// FST_i.h: Internal interface for FST library.

#ifndef __FST_i_h__
#define __FST_i_h__

#ifdef _MSC_VER
#  include "fst/windows.h"
#endif

#include "fst/FST.h"

template <class T>
inline T _max(T a, T b) {
    return a > b ? a : b;
}

template <class T>
inline T _min(T a, T b) {
    return a < b ? a : b;
}

// The following two classes are for sorting arcs by input and output,
// respectively.
//
struct FSTArcCompareInput
    : std::binary_function<FSTArc*, FSTArc*, bool> {
    bool operator () (const FSTArc* x, const FSTArc* y) const {
        return x->input < y->input;
    }
};
struct FSTArcCompareOutput
    : std::binary_function<FSTArc*, FSTArc*, bool> {
    bool operator () (const FSTArc* x, const FSTArc* y) const {
        return x->output < y->output;
    }
};

// Commonly used assertion.
//
#define FST_ASSERT_VALID_STATE(p) { \
    assert((p) >= 0); assert((p) < _n_states); \
}

// Defines to help implement operator<.
//
#define FST_LESS2(a, b)     ((a) < (b))
#define FST_LESS3(a, b, c)  ((a) < (b) || ((a) == (b) && (c)))

// This class provides mapping from T* to integer indices and vice-a-versa.
//
template<class T>
class FSTPointerMap {
    typedef std::map<T*, int> _Map;
    typedef std::vector<T*> _Vector;
    _Map _m;
    _Vector _v;
public:
    FSTPointerMap() {
        _v.push_back(NULL);
    }
    int operator[](T* ap) {
    int& i = _m[ap];
    if (i == 0 && ap != _v[0]) {
            i = _v.size();            // new entry
            _v.push_back(ap);
    }
    return i;
    }
    T* operator[](int i) const {
    return _v[i];
    }
    int size() const {
    return _v.size();
    }
};

// Function prototypes.
//
extern FSTGeneric FSTLoadBasic(std::istream& is, const char* header);
extern FSTGeneric FSTLoadBinary(std::istream& is, const char* header);
extern FSTGeneric FSTLoadDynamic(std::istream& is, const char* header);
extern FSTGeneric FSTLoadMmap(std::istream& is, const char* header, const char* filename);
extern FSTGeneric FSTLoadNgram(std::istream& is, const char* header);
extern FSTGeneric FSTLoadNgram2(std::istream& is, const char* header);
extern FSTGeneric FSTLoadNgram3(std::istream& is, const char* header);
extern FSTGeneric FSTLoadRtn(std::istream& is, const char* header);

extern std::istream& FSTBinaryCin();
extern std::ostream& FSTBinaryCout();

using namespace std;

#endif // __FST_i_h__

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 80
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
