////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

#ifndef __fst_allocator_h__
#define __fst_allocator_h__

#include <memory>
#include "fst/FSTAllocator.h"

namespace fst {

    template <class T>
    class allocator {
    public:
        typedef T           value_type;
        typedef size_t      size_type;
        typedef ptrdiff_t   difference_type;
        typedef T*          pointer;
        typedef const T*    const_pointer;
        typedef T&          reference;
        typedef const T&    const_reference;
        
        allocator() throw() {
        }

        template <class U>
        allocator(const allocator<U>&) throw() {
        }

        ~allocator() throw() {
        }

        bool operator==(const allocator& other) {
            return true;
        }

        pointer address(reference r) const {
            return &r;
        }
        const_pointer address(const_reference r) const {
            return &r;
        }

        pointer allocate(size_type n, const void* hint=0) {
            return pointer(FSTAllocator::allocate(n * sizeof(value_type)));
        }
        void deallocate(pointer p, size_type n) {
            FSTAllocator::deallocate(p, n * sizeof(value_type));
        }

        void construct(pointer p, const T& val) {
            new(p) T(val);
        }
        void destroy(pointer p) {
            p->~T();
        }

        size_type max_size() const throw() {
            return -1;
        }

        template <class U> struct rebind {
            typedef allocator<U> other;
        };

        bool operator==(const allocator& other) const {
            return true;
        }

        bool operator!=(const allocator& other) const {
            return false;
        }

    };

    template <>
    class allocator<void> {
    public:
        typedef void* pointer;
        typedef const void* const_pointer;
        typedef void value_type;

        template <class U>
        struct rebind {
            typedef allocator<U> other;
        };
    };

}

// The following is an optimization for the STL in GCC 2.95 through 3.3.  It
// informs the base STL containers that fst::allocator<T> is instanceless,
// meaning that the containers do not need to store an allocator object inside
// of them.  There does not appear to be a standard way to do this.  Starting
// with GCC 3.4 I do not believe this is necessary, and as coded below does not
// work.
//
// Note that the allocator_type(). calls do not cause any inefficiencies.
//
// If you get a compilation error about _Alloc_traits not being defined, then
// you likely included this file before including an STL header.
//
//
#if defined(__GNUC__) && (__GNUC__ < 3 || (__GNUC__ == 3 && __GNUC_MINOR__ < 4))

namespace std {

    template <typename T, typename T2>
    struct _Alloc_traits<T, fst::allocator<T2> > {
        static const bool _S_instanceless = true;
        typedef fst::allocator<T> allocator_type;
        class _Alloc_type {
        public:
            static T* allocate(size_t n) {
                return n ? allocator_type().allocate(n) : 0;
             }
            static T* allocate() {
                return allocator_type().allocate(sizeof(T));
            }
            static void deallocate(T* p, size_t n) {
                if (n) allocator_type().deallocate(p, n);
            }
            static void deallocate(T* p) {
                allocator_type().deallocate(p, sizeof(T));
            };
        };
    };

}

#endif

#endif // __fst_allocator_h__

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 80
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
