////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

#include "fst/FSTAllocator.h"
#include "fst/format.h"
#include <iostream>
#include <fstream>
#include <cstdlib>
#include <cstdarg>

#if defined(__APPLE__) && !defined(__osx__)
#  define __osx__
#endif

#if defined(__linux__) || defined(__solaris__) || defined(__cygwin__)
#  include <unistd.h>
#  include <fcntl.h>
#  include <sys/mman.h>
#endif

#if defined(__linux__)
#  include <sys/user.h>
#  include <sys/resource.h>
#endif

#if defined(__osx__)
#  include <unistd.h>
#  include <fcntl.h>
#  include <sys/mman.h>
#  include <sys/user.h>
#  include <sys/resource.h>
#  include <mach/vm_param.h>
#  define MAP_ANONYMOUS MAP_ANON
#endif

#ifdef _WIN32
#  include <windows.h>
#endif

#ifdef __cygwin__
#  include <sys/resource.h>
#  include <windows.h>
#endif

using namespace std;

// Default stack size limit (kB).  This is overruled by "limit stacksize" and
// "setenv FST_STACK_LIMIT".  We put a barrier page that is unreadable,
// unwritable, and unexecutable just below the stack so that the stack does not
// grow into our mmap area and corrupt our data structures.
//
#define DEFAULT_FST_STACK_LIMIT 10240

// Number of empty chunks to keep around for each size (tunable).  Empty chunks
// are discarded in a least-recently used fashion.
//
#define SIZE_EMPTY_CHUNKS 1

//#define INSTRUMENT_ALLOCATIONS

// Define PARANOID to check every single allocation and deallocation.  This
// should check for multiple deallocations of the same pointer, as well as
// deallocating with the wrong size.
//
//#define PARANOID

#ifdef PARANOID

#define PARANOID_ALLOCATE(P, SIZE)   check_allocate(P, SIZE)
#define PARANOID_DEALLOCATE(P, SIZE) check_deallocate(P, SIZE)

#include <hash_map>

struct hash<void*> {
    size_t operator()(const void* x) const {
        return size_t(x);
    }
};

typedef fst::hash_map<void*,size_t> CheckMap;

static CheckMap* check_map;
static bool check_active = true;

static inline void check_allocate(void* p, size_t size) {
    if (check_active) {
        if (!check_map) check_map = new CheckMap(1<<20);
        check_active = false;
        size_t& stored_size = (*check_map)[p];
        assert(stored_size == 0);
        stored_size = size;
        check_active = true;
    }
}

static inline void check_deallocate(void* p, size_t size) {
    if (check_active) {
        assert(check_map);
        check_active = false;
        CheckMap::iterator i = check_map->find(p);
        assert(i != check_map->end());
        assert((*i).second == size);
        check_map->erase(i);
        check_active = true;
    }
}

#else // !PARANOID

#define PARANOID_ALLOCATE(P, SIZE)
#define PARANOID_DEALLOCATE(P, SIZE)

#endif

#define max_words FSTAllocator::max_words
#define word_size FSTAllocator::word_size

extern void FSTFatal(const char* fmt, ...);
extern void FSTFatal(const char* fmt, va_list ap);
extern void FSTWarn(const char* fmt, ...);
extern void FSTWarn(const char* fmt, va_list ap);

// Free items have this overlayed to provide the list link.
//
struct Item {
    Item* next;
};

// Pointers for doubly linked lists of chunks.
//
struct ChunkLink {
    ChunkLink* next;
    ChunkLink* prev;
};

// Each chunk has one of these at its beginning.
//
struct ChunkHead : ChunkLink {
    ChunkHead* chunk;
    Item* free_list;
    size_t free_count;
#ifndef _WIN32
    void (*free_func)(void*);
#endif
};

// Compute number of bytes to skip at the beginning of each page to account for
// the ChunkHead.  However, round up to the nearest multiple of 16 (x86 cache
// line size).
//
#define CHUNK_HEAD_PADDED ((sizeof(ChunkHead) + 15) & ~15)

// We have an array of these, one for each rounded item size.
//
struct SizeEntry {
    ChunkLink free_list;
    ChunkLink full_list;
    ChunkLink empty_list;
    size_t    empty_count;
    size_t    capacity;
};

static SizeEntry size_table[FSTAllocator_max_words + 1];
static size_t page_size, page_mask; // see initialize()
static size_t valloc_chunks;

// Debugging and performance tuning.
//
#ifdef INSTRUMENT_ALLOCATIONS
static size_t chunks_allocated;
static size_t allocations;
static size_t deallocations;
#endif

static inline size_t
_min(size_t x, size_t y)
{
    return x < y ? x : y;
}

#include <cstdio>

static void
initialize()
{
#ifdef _SC_PAGE_SIZE
    page_size = sysconf(_SC_PAGE_SIZE);
#endif

#ifdef _WIN32
    _SYSTEM_INFO info;
    GetSystemInfo(&info);
    page_size = info.dwAllocationGranularity;
#endif    

    page_mask = ~(page_size - 1);

    valloc_chunks = (1 << 20)/page_size;

#ifdef __linux__
    assert(size_t(getpagesize()) == page_size);
#endif

    for (int i = 1; i <= int(FSTAllocator_max_words); i++) {
        SizeEntry* se = &size_table[i];
        se->free_list.next = se->free_list.prev = &se->free_list;
        se->full_list.next = se->full_list.prev = &se->full_list;
        se->empty_list.next = se->empty_list.prev = &se->empty_list;
        se->empty_count = 0;
        se->capacity = (page_size - CHUNK_HEAD_PADDED)/(i*FSTAllocator_word_size);
    }
}

//
// Chunk allocation and deallocation.  This currently uses anonymous mmap which
// allows pages of memory to be returned to system.  This is system-dependent.
//

#ifdef _WIN32

static void*
allocate_chunk()
{
#ifdef INSTRUMENT_ALLOCATIONS
    allocations++;
    chunks_allocated++;
#endif

    return VirtualAlloc(NULL,
                        page_size,
                        MEM_COMMIT | MEM_TOP_DOWN,
                        PAGE_READWRITE);
}

static void
deallocate_chunk(void* p)
{
#ifdef INSTRUMENT_ALLOCATIONS
    deallocations++;
    chunks_allocated--;
#endif

    VirtualFree(p, 0, MEM_RELEASE);
}

#else // !_WIN32

static void
free_mmap_chunk(void* p)
{
    munmap((caddr_t)p, page_size);
}

static void*
get_mmap_chunk()
{
#ifdef MAP_ANONYMOUS
    void* result = mmap(NULL, page_size,
                        PROT_READ|PROT_WRITE, MAP_PRIVATE|MAP_ANONYMOUS, -1, 0);
#else
    static int fd = -1;
    if (fd < 0)
        fd = open("/dev/zero", O_RDWR);

    void* result =mmap(NULL, page_size,
                       PROT_READ|PROT_WRITE, MAP_PRIVATE, fd, 0);
#endif

    if (result != MAP_FAILED)
        reinterpret_cast<ChunkHead*>(result)->free_func = free_mmap_chunk;

    return result;
}

static ChunkLink* valloc_free_list;

static void
free_valloc_chunk(void* p)
{
    ChunkLink* cl = reinterpret_cast<ChunkLink*>(p);
    cl->next = valloc_free_list;
    valloc_free_list = cl;
}

static void*
get_valloc_chunk()
{
    static bool warned = false;
    if (!warned) {
        FSTWarn("FSTAllocator: mmap failed -- using valloc");
        warned = true;
    }

    if (!valloc_free_list) {
        char* chunk = (char*) valloc(page_size * valloc_chunks);
        if (!chunk)
            FSTFatal("FSTAllocator:: valloc failed, completely out of memory");
        for (int i = 0; i < int(valloc_chunks); i++) {
            ChunkHead* ch = reinterpret_cast<ChunkHead*>(chunk);
            ch->free_func = free_valloc_chunk;
            ch->next = valloc_free_list;
            valloc_free_list = ch;
            chunk += page_size;
        }
    }

    ChunkLink* result = valloc_free_list;
    valloc_free_list = result->next;

    return result;
}

static void*
allocate_chunk()
{
#ifdef INSTRUMENT_ALLOCATIONS
    allocations++;
    chunks_allocated++;
#endif

    void* result = get_mmap_chunk();
    if (result != MAP_FAILED)
        return result;

    return get_valloc_chunk();
}

static void
deallocate_chunk(void* p)
{
#ifdef INSTRUMENT_ALLOCATIONS
    deallocations++;
    chunks_allocated--;
#endif

    ChunkHead* ch = reinterpret_cast<ChunkHead*>(p);
    ch->free_func(p);
}

#endif // !_WIN32

//
// Chunk list operations.
//

static inline bool
empty(ChunkLink& l)
{
    return l.next == &l;
}

static inline void
remove(ChunkLink* c)
{
    c->prev->next = c->next;
    c->next->prev = c->prev;
}

static inline void
insert(ChunkLink* c, ChunkLink& l)
{
    c->next = l.next;
    c->prev = &l;
    l.next = c;
    c->next->prev = c;
}

static inline void
move(ChunkLink* c, ChunkLink& l)
{
    remove(c);
    insert(c, l);
}

//
// The allocator.
//

void*
FSTAllocator::_allocate(size_t words)
{
    if (!words)
        return 0;

    if (!page_size) {
        initialize();
    }

    SizeEntry* se = &size_table[words];

    if (!empty(se->free_list)) {

        // Fast path: we have a free item already.
        //
        ChunkHead* ch = static_cast<ChunkHead*>(se->free_list.next);
        Item* result = ch->free_list;
        ch->free_list = result->next;
        if (--ch->free_count == 0)
            move(ch, se->full_list);
        PARANOID_ALLOCATE(result, words);
        return result;

    } else {

        ChunkHead* ch;
        
        if (!empty(se->empty_list)) {

            // We have a chunk ready in the empty list.
            //
            move(se->empty_list.next, se->free_list);
            se->empty_count--;
            ch = static_cast<ChunkHead*>(se->free_list.next);

        } else {

            // Allocate new chunk, put on free list, and set up all the item
            // links.
            //
            char* chunk = reinterpret_cast<char*>(allocate_chunk());
            assert(chunk != NULL);
            ch = reinterpret_cast<ChunkHead*>(chunk);
            insert(ch, se->free_list);
            ch->chunk = ch;
            ch->free_list = reinterpret_cast<Item*>(chunk + CHUNK_HEAD_PADDED);
            ch->free_count = se->capacity;
            size_t bytes = words * FSTAllocator_word_size;
            char* page_start = chunk;
            reinterpret_cast<ChunkHead*>(page_start)->chunk = ch;
            char* end = page_start + page_size - 2*bytes;
            char* p = page_start + CHUNK_HEAD_PADDED;
            for (; p <= end; p += bytes) {
                reinterpret_cast<Item*>(p)->next =
                    reinterpret_cast<Item*>(p + bytes);
            }
            reinterpret_cast<Item*>(p)->next = 0;
        }
        
        Item* result = ch->free_list;
        ch->free_list = result->next;
        ch->free_count--;

        PARANOID_ALLOCATE(result, words);
        return result;

    }
}

//
// The deallocator.
//

void
FSTAllocator::_deallocate(void* pointer, size_t words)
{
    PARANOID_DEALLOCATE(pointer, words);

    SizeEntry* se = &size_table[words];
    ChunkHead* ch =
        reinterpret_cast<ChunkHead*>(size_t(pointer) & page_mask);

    Item* item = reinterpret_cast<Item*>(pointer);
    item->next = ch->free_list;
    ch->free_list = item;
    ch->free_count++;

    if (ch->free_count == 1) {

        // Chunk was full, but is no longer so move to free list.
        //
        move(ch, se->free_list);

    } else if (ch->free_count == se->capacity) {

        // Chunk is now empty.  Deallocate last chunk on empty list if we're at
        // empty list capacity, and then move chunk to head of empty list.
        //
#if SIZE_EMPTY_CHUNKS == 0
        remove(ch);
        deallocate_chunk(ch);
#else
        if (se->empty_count == SIZE_EMPTY_CHUNKS) {
            ChunkLink* last = se->empty_list.prev;
            remove(last);
            deallocate_chunk(last);
            se->empty_count--;
        }
        move(ch, se->empty_list);
        se->empty_count++;
#endif

    }
}

void*
FSTAllocator::reallocate(void*, size_t, size_t)
{
    bool not_implemented = false;
    assert(not_implemented);
    return 0;
}

//
// The garbage collector: return unused chunks back to system.  We already
// return chunks fairly agressively, so this isn't terribly useful.
//

void
FSTAllocator::gc()
{
    // Deallocate chunks on all empty lists.
    //
    for (size_t words = 1; words <= FSTAllocator_max_words; words++) {
        SizeEntry* se = &size_table[words];
        for (ChunkLink* cl = se->empty_list.next; cl != &se->empty_list;) {
            ChunkLink* next = cl->next;
            deallocate_chunk(cl);
            cl = next;
        }
        se->empty_list.next = se->empty_list.prev = &se->empty_list;
        se->empty_count = 0;
    }
#ifdef INSTRUMENT_ALLOCATIONS
    cerr.form("FSTAllocator: %d chunks, %d kB.\n",
              chunks_allocated, chunks_allocated*page_size >> 10);
    cerr.form("              %d allocations, %d deallocations.\n",
              allocations, deallocations);
    allocations = deallocations = 0;
#endif
}

void
FSTAllocator::status(ostream& os)
{
    format(os, "-----------------------------------------\n");
    format(os, "FSTAllocator status\n");
    format(os, "%4s %8s %6s %6s %6s %6s\n",
            "size",
            "used",
            "full",
            "mixed",
            "empty",
            "occup");
    size_t total_pages = 0;
    for (size_t words = 1; words <= FSTAllocator_max_words; words++) {
        size_t n_free = 0, n_full = 0, n_empty = 0, n_entries = 0;
        SizeEntry* se = &size_table[words];
        ChunkLink* cl;
        for (cl = se->free_list.next; cl != &se->free_list; cl = cl->next) {
            n_free++;
            n_entries += se->capacity - static_cast<ChunkHead*>(cl)->free_count;
        }
        for (cl = se->full_list.next; cl != &se->full_list; cl = cl->next)
            n_full++;
        n_entries += n_full * se->capacity;
        n_empty = se->empty_count;

        if (n_free == 0 && n_full == 0 && n_empty == 0)
            continue;

        total_pages += n_free + n_full + n_empty;

        size_t n_full_free = n_full + n_free;
        if (n_full_free == 0)
            n_full_free = 1;

        format(os, "%4d %8d %6d %6d %6d %6.4f\n",
                words*FSTAllocator_word_size,
                n_entries,
                n_full,
                n_free,
                n_empty,
                float(n_entries) / float(n_full_free*se->capacity));
    }
    format(os, "total: %d pages x %d kB/page = %d MB\n",
            total_pages,
            page_size >> 10,
            (total_pages * page_size + (1<<20)-1) >> 20);
    format(os, "-----------------------------------------\n");
}

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 80
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
