////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

// If we want the capability of limiting the size of the priority queue, we will
// probably have to use an ordered set so that we can pop things off the bottom.
// However, this probably has space and speed implications compared to the
// current vector-based heap we're using.  Is there a type of heap that can pop
// off both ends?

//#define DEBUG
//#define INSTRUMENT

#include "fst/FST_i.h"
#include <fstream>
#include <queue>                                // must be after sls/FST.h

// Tree node used to keep track of unique label sequences during N-best search.
//
struct FSTAstarNbestUniqueNode {
    struct Less {
        bool operator()(const FSTAstarNbestUniqueNode* x,
                        const FSTAstarNbestUniqueNode* y) const {
            return x->label < y->label;
        }
    };
    typedef std::set<FSTAstarNbestUniqueNode*, Less> Set;

    const int label;
    Set children;
    FSTAstarNbestUniqueNode(int label_=0)
        : label(label_) {}
    ~FSTAstarNbestUniqueNode() {
        for (Set::iterator i = children.begin(); i != children.end(); ++i)
            delete *i;
    }
    FSTAstarNbestUniqueNode* append(int x) {
        if (!x) return this;
        FSTAstarNbestUniqueNode* n = new FSTAstarNbestUniqueNode(x);
        std::pair<Set::iterator, bool> p = children.insert(n);
        if (!p.second)
            delete n;
        return const_cast<FSTAstarNbestUniqueNode*>(*p.first);
    }
    //FSTAllocator_new_delete(FSTAstarNbestUniqueNode);
};

// Base class for the N-best and graph A* queue entry types.
//
struct FSTAstarQueueEntry {

    float t_score;                              // total score

    FSTAstarQueueEntry(float ts)
        : t_score(ts) {
    }
    struct Compare {
        FSTSemiring semiring;
        Compare() : semiring(0) {}
        Compare(FSTSemiring s) : semiring(s) {}
        bool operator()(const FSTAstarQueueEntry* x,
                        const FSTAstarQueueEntry* y) const {
            return semiring->is_better(y->t_score, x->t_score, 0);
        }
    };
};

// A* queue entry for graph generation.
//
struct FSTAstarGraphQueueEntry
    : FSTAstarQueueEntry {

    float p_score;                              // partial score
    int   i_state;                              // input FST state
    int   a_state;                              // pending arc state (from)
    int   a_input;                              // pending arc input
    int   a_output;                             // pending arc output
    float a_weight;                             // pending arc weight

    FSTAstarGraphQueueEntry(float ts, float ps, int is,
                            int as=-1, int ai=0, int ao=0, float aw=0)
        : FSTAstarQueueEntry(ts), p_score(ps), i_state(is),
          a_state(as), a_input(ai), a_output(ao), a_weight(aw) {
    }
    FSTAllocator_new_delete(FSTAstarGraphQueueEntry);
};

// A* queue entry for N-best list.
//
struct FSTAstarNbestQueueEntry
    : FSTAstarQueueEntry {

    typedef FSTAstarNbestUniqueNode Node;

    float p_score;                              // partial score
    int   i_state;                              // input FST state
    int   s_state;                              // scratch FST state (from)
    Node* u_node;                               // unique sequence node
    int   a_input;                              // pending arc input
    int   a_output;                             // pending arc output
    float a_weight;                             // pending arc weight

    FSTAstarNbestQueueEntry(float ts, float ps, int is, int ss, Node* un,
                            int ai=-1, int ao=0, float aw=0)
        : FSTAstarQueueEntry(ts), p_score(ps), i_state(is), s_state(ss),
          u_node(un), a_input(ai), a_output(ao), a_weight(aw) {
    }
    FSTAllocator_new_delete(FSTAstarNbestQueueEntry);
};

// Heap-based priority queue with quick cleanup of unpopped entries.  This is a
// template so that we can use it with different types of A* queue entries
// depending on the type of A* search we are performing.  We primarily need the
// EntryType so that we call the correct entry delete function.
//
template<class EntryType>
class FSTAstarQueue {
private:
    fst::vector<FSTAstarQueueEntry*> _heap;
    FSTAstarQueueEntry::Compare _compare;
public:
    FSTAstarQueue(FSTSemiring semiring)
        : _compare(semiring){
    }
    ~FSTAstarQueue() {
        fst::vector<FSTAstarQueueEntry*>::iterator i;
        for (i = _heap.begin(); i != _heap.end(); ++i)
            delete (EntryType*)*i;
    }
    unsigned int size() const {
        return _heap.size();
    }
    bool empty() const {
        return _heap.empty();
    }
    void push(EntryType* entry) {
        _heap.push_back(entry);
        push_heap(_heap.begin(), _heap.end(), _compare);
    }
    EntryType* pop() {
        pop_heap(_heap.begin(), _heap.end(), _compare);
        EntryType* result = static_cast<EntryType*>(_heap.back());
        _heap.pop_back();
        return result;
    }
};

FSTGeneric
FSTAstarNbest(const FSTGeneric& x, FSTAstarSide unique, int N, float threshold)
{
    typedef FSTAstarNbestQueueEntry Entry;
    typedef FSTAstarNbestUniqueNode Node;
    typedef pair<int, Node*> Pair;

    float best = x->potential(x->initial());
    threshold += best;

    FSTSemiring semiring = x->semiring();

    // The scratch FST holds the tree associated with all the A* paths, but its
    // arcs are backwards so that we can walk it from leaf to root when we copy
    // a complete path to the result FST.
    //
    FSTGeneric scratch = FSTBasic(semiring);
    scratch->set_initial(scratch->add_state());

    FSTGeneric result = FSTBasic(semiring);
    result->set_initial(result->add_state());

    std::set<Pair> visited_set;

    Node root_node;

    FSTAstarQueue<Entry> q(semiring);
    q.push(new Entry(best, semiring->one,
                     x->initial(), scratch->initial(), &root_node));

    int n = 0;

    // Continue until stopping criteria are met.
    //
    while (n < N && !q.empty()) {

        // Pop the best entry.
        //
        Entry* e = q.pop();

        // There are different types of entries:
        //   initial:    a_input == -1
        //   path-final: a_input == -1 && i_state == -1
        //   normal:     a_input != -1

        // Add pending arc to scratch.  s_state will be changed to the new
        // scratch state.  Note that we make arcs point toward the initial
        // state, the root of the tree.
        //
        if (e->a_input >= 0) {
            int from = e->s_state;
            e->s_state = scratch->add_state();
            scratch->add_arc(e->s_state, from,
                             e->a_input, e->a_output, e->a_weight);
        }

        // Mark (i_state, u_node) visited and note if it was unvisited before,
        // but only if unique is set.
        //
        bool unvisited =
            (unique == FSTNone) ||
            visited_set.insert(Pair(e->i_state, e->u_node)).second;

        // If this (i_state, u_node) was visited before, do not expand it again.
        //
        if (!unvisited) {
            delete e;
            continue;
        }

        // Process a path-final entry and break.
        //
        if (e->i_state < 0) {
            int q = result->add_state();
            result->set_final(q, e->p_score);
            if (e->s_state == 0) {
                result->add_arc(0, q, 0, 0);
            } else {
                for (int p = e->s_state; p != 0;) {
                    FSTArcIter a = scratch->arcs_out(p);
                    assert(a.n() == 1);
                    p = a->next;
                    int r = p ? result->add_state() : 0;
                    result->add_arc(r, q, a->input, a->output, a->weight);
                    q = r;
                }
            }

            // If we wanted to add a hook to call when an N-best entry is ready,
            // this would be the place.  The total path score is e->t_score.

            ++n;
            delete e;
            continue;
        }

        // If we're at a final state, push a path-final entry.  We do not
        // process this N-best path yet because the final weight might not yet
        // been accounted for in the priority queue.
        //
        if (x->is_final(e->i_state)) {
            float fw = x->final_weight(e->i_state);
            float ps = semiring->times(e->p_score, fw);
            q.push(new Entry(ps, fw, -1, e->s_state, e->u_node));
        }

        // Expand entry by following arcs.
        //
        for (FSTArcIter a = x->arcs_out(e->i_state); a.valid(); ++a) {
            float ps = semiring->times(e->p_score, a->weight);
            float ts = semiring->times(ps, x->potential(a->next));
            if (semiring->is_better(ts, threshold, 0)) {
                Node* node = e->u_node;
                switch (unique) {
                case FSTNone:
                    break;
                case FSTInput:
                    node = node->append(a->input);
                    break;
                case FSTOutput:
                    node = node->append(a->output);
                    break;
                }
                q.push(new Entry(ts, ps, a->next, e->s_state, node,
                                 a->input, a->output, a->weight));
            }
        }

        // We're finished with this entry.
        //
        delete e;
    }

    // Trim states not part of a complete path and return.
    //
    return FSTTrim(result);
}

FSTGeneric
FSTAstarGraph(const FSTGeneric& x, FSTAstarSide project,
              float threshold, int max_pops)
{
    typedef FSTAstarGraphQueueEntry Entry;

    float best = x->potential(x->initial());
    threshold += best;

    FSTSemiring semiring = x->semiring();

    FSTGeneric scratch = FSTBasic(semiring);
    scratch->add_states(x->n_states());
    scratch->set_initial(x->initial());

    bool reached_final = false;
    int n_pops = 0;

    std::set<int> visited_set;

    FSTAstarQueue<Entry> q(semiring);
    q.push(new Entry(best, semiring->one, x->initial()));

    // Continue until stopping criteria are met.
    //
    while ((!reached_final || !max_pops || n_pops < max_pops) && !q.empty()) {

        // Pop the best entry.
        //
        Entry* e = q.pop();
        ++n_pops;

        // Grow things to keep up with possibly dynamic input FST x.
        //
        int n_new_states = x->n_states() - scratch->n_states();
        if (n_new_states > 0) {
            scratch->add_states(n_new_states);
        }

        // Add pending arc if appropriate.
        //
        if (e->a_state >= 0) {
            scratch->add_arc(e->a_state, e->i_state,
                             e->a_input, e->a_output, e->a_weight);
        }

        // Mark i_state visited and note if it was unvisited before.
        //
        bool unvisited = visited_set.insert(e->i_state).second;

        // Set or update state potentials.
        //
        if (unvisited) {
            scratch->set_potential(e->i_state, e->p_score);
        } else {
            if (semiring->is_better(e->p_score,
                                    scratch->potential(e->i_state),
                                    0)) {
                scratch->set_potential(e->i_state, e->p_score);
            }
        }

        // If this i_state was visited before, do not expand it again.
        //
        if (!unvisited) {
            delete e;
            continue;
        }

        // Set final weight of i_state.
        //
        if (x->is_final(e->i_state)) {
            reached_final = true;
            float fw = x->final_weight(e->i_state);
            scratch->set_final(e->i_state, fw);
        }

        // Expand entry by following arcs.
        //
        for (FSTArcIter a = x->arcs_out(e->i_state); a.valid(); ++a) {
            float ps = semiring->times(e->p_score, a->weight);
            float ts = semiring->times(ps, x->potential(a->next));
            if (semiring->is_better(ts, threshold, 0)) {
                int input = 0;
                int output = 0;
                switch (project) {
                case FSTNone:
                    input = a->input;
                    output = a->output;
                    break;
                case FSTInput:
                    input = a->input;
                    break;
                case FSTOutput:
                    input = a->output;
                    break;
                }
                q.push(new Entry(ts, ps, a->next, e->i_state,
                                 input, output, a->weight));
            }
        }

        // We're finished with this entry.
        //
        delete e;
    }

    // Trim states not part of a complete path and return.
    //
    return FSTTrim(scratch);
}

// This is a modified A* that removes epsilons in the result.  It is commented
// out because I don't think it is faster than FSTAstarGraph -> FSTProject ->
// FSTRemoveEpsilonsA.
//
#if 0
FSTGeneric
FSTAstarGraph(const FSTGeneric& x, FSTAstarSide project, bool rmeps,
              float threshold, int max_pops)
{
    typedef FSTAstarGraphQueueEntry Entry;
    typedef pair<int, int> Pair;

    float best = x->potential(x->initial());
    threshold += best;

    FSTSemiring semiring = x->semiring();

    FSTGeneric scratch = FSTBasic(semiring);
    scratch->add_states(x->n_states());
    scratch->set_initial(x->initial());

    bool reached_final = false;
    int n_pops = 0;

    std::set<Pair> visited_set;

    FSTAstarQueue<Entry> q(semiring);
    q.push(new Entry(best, semiring->one, x->initial()));

    while ((!reached_final || !max_pops || n_pops < max_pops) && !q.empty()) {

        // Pop the best entry.
        //
        Entry* e = q.pop();
        ++n_pops;

        //cerr.form("popped (%d, %d, %s)\n", e->i_state, e->a_state, FSTstring_map[e->a_input]);

        // Grow things to keep up with possibly dynamic input FST x.
        //
        int n_new_states = x->n_states() - scratch->n_states();
        if (n_new_states > 0) {
            scratch->add_states(n_new_states);
        }

        // Add pending arc if appropriate.
        //
        if (e->a_state >= 0 && (!rmeps || e->a_input || e->a_output)) {
            scratch->add_arc(e->a_state, e->i_state,
                             e->a_input, e->a_output, e->a_weight);
        }
        if (e->a_state < 0) {
            e->a_state = e->i_state;
        }

        // Mark (i_state, a_state) visited and note if it was unvisited before.
        //
        bool unvisited = visited_set.insert(Pair(e->i_state, rmeps ? e->a_state : 0)).second;

        // Set or update state potentials.
        //
        if (unvisited) {
            scratch->set_potential(e->i_state, e->p_score);
        } else {
            if (semiring->is_better(e->p_score,
                                    scratch->potential(e->i_state),
                                    0)) {
                scratch->set_potential(e->i_state, e->p_score);
            }
        }

        // If this (i_state, a_state) was visited before, do not expand it again.
        //
        if (!unvisited) {
            delete e;
            continue;
        }

        // Set final weight of i_state, and if removing epsilons, possibly update
        // final weight of a_state.
        //
        if (x->is_final(e->i_state)) {
            reached_final = true;
            float fw = x->final_weight(e->i_state);
            scratch->set_final(e->i_state, fw);
            if (rmeps && !e->a_input && !e->a_output) {
                float w = semiring->plus(scratch->final_weight(e->a_state),
                                         semiring->times(e->a_weight, fw));
                scratch->set_final(e->a_state, w);
            }
        }

        // Expand entry by following arcs.
        //
        for (FSTArcIter a = x->arcs_out(e->i_state); a.valid(); ++a) {
            float ps = semiring->times(e->p_score, a->weight);
            float ts = semiring->times(ps, x->potential(a->next));
            if (semiring->is_better(ts, threshold, 0)) {
                int input = 0;
                int output = 0;
                switch (project) {
                case FSTNone:
                    input = a->input;
                    output = a->output;
                    break;
                case FSTInput:
                    input = a->input;
                    break;
                case FSTOutput:
                    input = a->output;
                    break;
                }
                if (rmeps && !e->a_input && !e->a_output) {
                    float w = semiring->times(e->a_weight, a->weight);
                    q.push(new Entry(ts, ps, a->next, e->a_state,
                                     input, output, w));
                } else {
                    q.push(new Entry(ts, ps, a->next, e->i_state,
                                     input, output, a->weight));
                }
            }
        }

        delete e;

    }

    return FSTTrim(scratch);
}
#endif

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 80
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
