////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

#include "fst/FST_i.h"

class FSTBackoffRep
    : public FSTGenericInt {

public: // methods
    FSTBackoffRep(const FSTGeneric& in, bool invert=false);
    virtual ~FSTBackoffRep();
    virtual const fst::set<int>& input_alphabet() const;
    virtual const fst::set<int>& output_alphabet() const;
    virtual bool is_final(int p) const;
    virtual float final_weight(int p) const;
    virtual float potential(int p) const;
    virtual FSTArcIter arcs_out(int p, FSTArcOrder order=FSTNone) const;
    virtual FSTArcIter arcs_out_with_input(int p, int i,
                                           FSTArcMode m=FSTEqual) const;
    virtual FSTArcIter arcs_out_with_output(int p, int o,
                                            FSTArcMode m=FSTEqual) const;
    virtual void force() const;
    virtual void reset() const;

protected: // types

protected: // members
    const FSTGenericInt*    _in;
    bool                    _invert;

protected: // methods
    FSTArcIter dynamic(const FSTArc* a, int p, float bo_weight) const;
};

FSTBackoffRep::FSTBackoffRep(const FSTGeneric& in, bool invert)
    : _in(in.pointer()),
      _invert(invert)
{
    _in->incref();
    _class_name = "FSTBackoff";
    _semiring = _in->semiring();
    _initial = _in->initial();
    _n_states = _in->n_states();
}

FSTBackoffRep::~FSTBackoffRep()
{
    _in->decref();
}

const fst::set<int>&
FSTBackoffRep::input_alphabet() const
{
    return _in->input_alphabet();
}

const fst::set<int>&
FSTBackoffRep::output_alphabet() const
{
    return _in->output_alphabet();
}

bool
FSTBackoffRep::is_final(int p) const
{
    return final_weight(p) != _semiring->zero;
}

float
FSTBackoffRep::final_weight(int p) const
{
    float w = _in->final_weight(p);

    if (w != _semiring->zero)
        return _invert ? _semiring->invert(w) : w;

    float bo = _semiring->one;
    FSTArcIter e;
    while (e = _in->arcs_out_with_input(p, 0), e.valid()) {
        assert(e.n() == 1);
        bo = _semiring->times(bo, e->weight);
        p = e->next;
        w = _in->final_weight(p);
        if (w != _semiring->zero) {
            w = _semiring->times(w, bo);
            return _invert ? _semiring->invert(w) : w;
        }
    }
    return _semiring->zero;
}

float
FSTBackoffRep::potential(int p) const
{
    return _in->potential(p);
}

FSTArcIter
FSTBackoffRep::arcs_out(int, FSTArcOrder) const
{
    FSTFatal("FSTBackoff::arcs_out not implemented");
    return FSTArcIter();
}

FSTArcIter
FSTBackoffRep::dynamic(const FSTArc* a, int p, float bo_weight) const
{
    float w = _semiring->times(a->weight, bo_weight);
    if (_invert)
        w = _semiring->invert(w);
    FSTArcIterData* data = new FSTArcIterData;
    data->arcs.push_back(new FSTArc(p, a->next, a->input, a->output, w));
    return FSTArcIter(data);
}

FSTArcIter
FSTBackoffRep::arcs_out_with_input(int p, int i, FSTArcMode mode) const
{
    assert(mode == FSTEqual);

    // We are essentially removing epsilons so there are none to output.
    //
    if (i == 0)
        return FSTArcIter();

    // See if we have input i directly at state p.  We do not need a dynamically
    // created arc if we are not inverting.
    //
    FSTArcIter a = _in->arcs_out_with_input(p, i);
    if (a.valid()) {
        assert(a.n() == 1);
        return _invert ? dynamic(a, p, _semiring->one) : a;
    }

    // Follow backoff epsilon arcs looking for state with input i.
    //
    float bo_weight = _semiring->one;
    for (FSTArcIter e = _in->arcs_out_with_input(p, 0);
         e.valid();
         e = _in->arcs_out_with_input(e->next, 0)) {
        assert(e.n() == 1);
        bo_weight = _semiring->times(bo_weight, e->weight);
        a = _in->arcs_out_with_input(e->next, i);
        if (a.valid()) {
            assert(a.n() == 1);
            return dynamic(a, p, bo_weight);
        }
    }

    // Didn't find a matching arc.
    //
    return FSTArcIter();
}

FSTArcIter
FSTBackoffRep::arcs_out_with_output(int, int, FSTArcMode) const
{
    FSTFatal("FSTBackoff::arcs_out_with_output not implemented");
    return FSTArcIter();
}

void
FSTBackoffRep::force() const
{
    _in->force();
}

void
FSTBackoffRep::reset() const
{
    _in->reset();
}

FSTGeneric
FSTBackoff(const FSTGeneric& x, bool invert)
{
    return FSTGeneric(new FSTBackoffRep(x, invert));
}

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 80
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
