////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

//#define USE_FORCE

#include <cstring>
#include "fst/FST_i.h"

#define FSTBasicDebug 0

class FSTBasicRep
    : public FSTGenericInt {

public: // methods

    FSTBasicRep(FSTSemiring semiring=FSTMinPlus);
    FSTBasicRep(const FSTGeneric& x, FSTArcOrder input, FSTArcOrder output,
                bool reverse, bool clear, bool convert_to_PlusTimes = false);
    FSTBasicRep(const FSTGeneric& x, FSTArcOrder input, FSTArcOrder output,
                bool reverse, bool clear, 
                FSTSemiring convert_to_semiring);
    FSTBasicRep(istream& is, const char* header);
    virtual ~FSTBasicRep();

    virtual const fst::set<int>& input_alphabet() const;
    virtual const fst::set<int>& output_alphabet() const;
    virtual int n_arcs_out(int p) const;
    virtual FSTArcIter arcs_out(int p, FSTArcOrder order=FSTNone) const;
    virtual FSTArc     arc_out(int p) const;
    virtual FSTArcIter arcs_out_with_input(int p, int i,
                                           FSTArcMode m=FSTEqual) const;
    virtual FSTArcIter arcs_out_with_output(int p, int o,
                                            FSTArcMode m=FSTEqual) const;
    virtual void       arcs_out(FSTArcVector& result,
                                int p, FSTArcOrder order=FSTNone) const;
    virtual bool       is_final(int p) const;
    virtual float      final_weight(int p) const;
    virtual float      potential(int p) const;
    virtual void       force() const;
    virtual void       reset() const;
    virtual void       dump_ascii(ostream& os, bool header=true) const;
    virtual void       dump_binary(ostream& os) const;
    virtual void       set_initial(int p);
    virtual void       set_final(int p, float weight);
    virtual void       set_potential(int p, float weight);
    virtual int        add_states(int n);
    virtual void       add_arc(int p, int q, int i, int o, float w);
    virtual void       add_arcs(int p, FSTArcVector& arcs,
                                FSTArcOrder order=FSTNone);
    virtual void       set_arcs(int p, FSTArcVector& arcs,
                                FSTArcOrder order=FSTNone);
    virtual void       delete_state(int p);
    virtual void       delete_states(const fst::vector<int>& states);

    static FSTGeneric  copy(const FSTGeneric& x);

    virtual void restore_input_output_index();

protected: // types

    struct _State;
    struct _TrainData;
    typedef fst::vector<_State*> _StatePtrVector;

protected: // members

    _StatePtrVector _states;
    mutable FSTArcOrder _arc_order;
    fst::set<int> _input_alphabet;
    fst::set<int> _output_alphabet;
    _TrainData* _train_data;

protected: // methods

    float _adjust(float weight, float scale, float offset) const {
        if (weight != _semiring->zero)
            weight = scale*weight + offset;
        return weight;
    }
    void _init();
    void _sort_arcs(FSTArcOrder order) const;
    void _maybe_grow(int p);
    virtual void _train_begin();
    virtual void _train_begin_left();
    virtual void _train_begin_right();
    virtual void _train_end();
    virtual void _train_normalize();
    virtual void _train_normalize_by_outputlabel();
    virtual void _train_update(fst::vector<double>& a_counts,
                               fst::vector<double>& f_counts,
                               double floor);
};

FSTGeneric
FSTBasic(FSTSemiring semiring)
{
    return FSTGeneric(new FSTBasicRep(semiring));
}

FSTGeneric
FSTBasic(const FSTGeneric& x)
{
    return FSTGeneric(new FSTBasicRep(x, FSTInput, FSTOutput, false, false));
}

FSTGeneric
FSTClearWeights(const FSTGeneric& x)
{
    return FSTGeneric(new FSTBasicRep(x, FSTInput, FSTOutput, false, true));
}

FSTGeneric
FSTProject(const FSTGeneric& x, FSTArcOrder input, FSTArcOrder output)
{
    return FSTGeneric(new FSTBasicRep(x, input, output, false, false));
}

FSTGeneric
FSTReverse(const FSTGeneric& x)
{
    return FSTGeneric(new FSTBasicRep(x, FSTInput, FSTOutput, true, false));
}

FSTGeneric
FSTConvert_to_PlusTimes(const FSTGeneric& x)
{
    return FSTGeneric(new FSTBasicRep(x, FSTInput, FSTOutput, false, false, true));
}

FSTGeneric
FSTConvert_to_MinPlus(const FSTGeneric& x)
{
    return FSTGeneric(new FSTBasicRep(x, FSTInput, FSTOutput, false, false, FSTMinPlus));
}

FSTGeneric
FSTConvert_to_MaxPlus(const FSTGeneric& x)
{
    return FSTGeneric(new FSTBasicRep(x, FSTInput, FSTOutput, false, false, FSTMaxPlus));
}

struct FSTBasicRep::_State {
    float weight;
    float potential;
    FSTArcVector arcs_out;
    _State(FSTSemiring s) : weight(s->zero), potential(s->one) {}
    ~_State() {
    for (FSTArcVector::iterator i = arcs_out.begin();
         i != arcs_out.end(); i++) {
        delete *i;
    }
    }
    FSTAllocator_new_delete(_State);
};

struct FSTBasicRep::_TrainData {
    FSTSemiring save_semiring;
    fst::vector<int> save_outputs;
    fst::vector<int> save_inputs;
    // NOTE: arc_map's 0th entry is NULL, and reserve for epislon arc.
    //       arc_map.size() = n_arcs + 1
    FSTPointerMap<FSTArc> arc_map;
    
    // arc_map's 0th entry is NULL to reserve for epislon arc,
    //     thus it contains n_arcs+1 entries
    _TrainData(int n_arcs) : save_outputs(n_arcs+1), save_inputs(n_arcs+1) {}
};

FSTBasicRep::FSTBasicRep(FSTSemiring semiring)
{
    _init();
    _semiring = semiring;
}

inline int
_transform(const FSTArcIter& a, FSTArcOrder which)
{
    int result;
    switch (which) {
    case FSTInput:
        result = a->input;
        break;
    case FSTOutput:
        result = a->output;
        break;
    default:
        result = 0;
    }
    return result;
}

// Make a copy of any FST, possibly reversing it and possibly projecting input
// and output labels.
//
FSTBasicRep::FSTBasicRep(const FSTGeneric& x,
                         FSTArcOrder input, FSTArcOrder output,
                         bool reverse, bool clear, bool convert_to_PlusTimes)
{
    _init();
    
    if (reverse && convert_to_PlusTimes) {
        cerr << "reverse and convert_to_PlusTimes both true not implemented" << endl;
        exit(0);
    }

    if (convert_to_PlusTimes) {
        _semiring = FSTPlusTimes;
    } else {
        _semiring = x->semiring();
    }

#ifdef USE_FORCE
    x->force();
#endif
    add_states(x->n_states());
    if (!reverse) {                             // not reversing
        set_initial(x->initial());
        for (FSTStateIter i = x->states(); i.valid(); i++) {
            int p = *i;
            float fw = x->final_weight(p);
            if (convert_to_PlusTimes) 
                fw = x->semiring()->convert(fw);
            if (fw != _semiring->zero)
                set_final(p, clear ? _semiring->one : fw);
            if (!clear){
                if (convert_to_PlusTimes)
                    set_potential(p, x->semiring()->convert(x->potential(p)));
                else 
                    set_potential(p, x->potential(p));
            }
            FSTArcIter j = x->arcs_out(p);
            _states[p]->arcs_out.reserve(j.n());
#ifndef USE_FORCE
            int new_states = x->n_states() - n_states();
            if (new_states > 0)
                add_states(new_states);
#endif
            for (; j.valid(); j++) {
                float aw = j->weight;
                if (convert_to_PlusTimes) 
                    aw = x->semiring()->convert(aw);
                add_arc(p, j->next,
                        _transform(j, input), _transform(j, output),
                        clear ? _semiring->one : aw);
            }
        }
    } else {                                    // reversing
        fst::set<int> final_states;
        float initial_potential = _semiring->zero;
        for (FSTStateIter i = x->states(); i.valid(); i++) {
            int p = *i;
            if (!clear)
                set_potential(p, x->potential(p));
            if (x->is_final(p)) {
                final_states.insert(p);
                initial_potential = _semiring->better(initial_potential,
                                                      x->potential(p));
            }
            FSTArcIter j = x->arcs_out(p);
#ifndef USE_FORCE
            int new_states = x->n_states() - n_states();
            if (new_states > 0)
                add_states(new_states);
#endif
            for (; j.valid(); j++)
                add_arc(j->next, p,
                        _transform(j, input), _transform(j, output),
                        clear ? _semiring->one : j->weight);
        }
        int new_initial = add_state();
        set_initial(new_initial);
        set_final(x->initial(), _semiring->one);
        for (fst::set<int>::const_iterator k = final_states.begin();
             k != final_states.end(); k++) {
            add_arc(new_initial, *k, 0, 0,
                    clear ? _semiring->one : x->final_weight(*k));
        }
        if (!clear)
            set_potential(new_initial, initial_potential);
    }
}


// Make a copy of any FST, possibly reversing it and possibly projecting input
// and output labels and possibly converting the semiring type.
//
FSTBasicRep::FSTBasicRep(const FSTGeneric& x,
                         FSTArcOrder input, FSTArcOrder output,
                         bool reverse, bool clear, 
                         FSTSemiring convert_to_semiring)
{
    _init();
    
    if (reverse && (convert_to_semiring != x->semiring())) {
        cerr << "reverse and convert_to_semiring both true not implemented" << endl;
        exit(0);
    }
        
    if ((x->semiring() == FSTMinPlus) && (convert_to_semiring != FSTPlusTimes) && (convert_to_semiring != FSTMaxPlus)) {
        cerr << "semiring of x " << x->semiring() << " and convert_to_semiring " << 
            convert_to_semiring << " pair not implemented" << endl;
        exit(0);
    }
    if ((x->semiring() == FSTMaxPlus) && (convert_to_semiring != FSTMinPlus)) {
        cerr << "semiring of x " << x->semiring() << " and convert_to_semiring " << 
            convert_to_semiring << " pair not implemented" << endl;
        exit(0);
    }
    if ((x->semiring() == FSTPlusTimes) && (convert_to_semiring != FSTMinPlus)) {
        cerr << "semiring of x " << x->semiring() << " and convert_to_semiring " << 
            convert_to_semiring << " pair not implemented" << endl;
        exit(0);
    }

    if (convert_to_semiring != x->semiring()) {
        _semiring = convert_to_semiring;
    } else {
        _semiring = x->semiring();
    }

#ifdef USE_FORCE
    x->force();
#endif
    add_states(x->n_states());
    if (!reverse) {                             // not reversing
        set_initial(x->initial());
        for (FSTStateIter i = x->states(); i.valid(); i++) {
            int p = *i;
            float fw = x->final_weight(p);
            if ((x->semiring() == FSTMinPlus) && (convert_to_semiring == FSTPlusTimes)) {
                // fw = x->semiring()->convert(fw);
                // return is_better(x, zero) ? exp(-x) : 0;
                fw = (fw > 1e30 - 1e-3) ? exp(-fw) : 0;
            } else if ((x->semiring() == FSTMinPlus) && (convert_to_semiring == FSTMaxPlus)) {
                fw = -fw;
            } else if ((x->semiring() == FSTMaxPlus) && (convert_to_semiring == FSTMinPlus)) {
                fw = -fw;
            } else if ((x->semiring() == FSTPlusTimes) && (convert_to_semiring == FSTMinPlus)) {
                // return (x > 0) ? -log(x) : zero;
                fw = (fw > 0) ? -log(fw) : 1e30;
            }
            if (fw != _semiring->zero)
                set_final(p, clear ? _semiring->one : fw);
            if (!clear){
                if ((x->semiring() == FSTMinPlus) && (convert_to_semiring == FSTPlusTimes)) {
                    // fw = x->semiring()->convert(fw);
                    // return is_better(x, zero) ? exp(-x) : 0;
                    float tmp_p = x->potential(p);
                    tmp_p = (tmp_p > 1e30 - 1e-3) ? exp(-tmp_p) : 0;
                    set_potential(p, tmp_p);
                } else if ((x->semiring() == FSTPlusTimes) && (convert_to_semiring == FSTMinPlus)) {
                    // return (x > 0) ? -log(x) : zero;
                    float tmp_p = x->potential(p);
                    tmp_p = (fw > 0) ? -log(tmp_p) : 1e30;
                    set_potential(p, tmp_p);
                } else 
                    set_potential(p, x->potential(p));
            }
            FSTArcIter j = x->arcs_out(p);
            _states[p]->arcs_out.reserve(j.n());
#ifndef USE_FORCE
            int new_states = x->n_states() - n_states();
            if (new_states > 0)
                add_states(new_states);
#endif
            for (; j.valid(); j++) {
                float aw = j->weight;

                if ((x->semiring() == FSTMinPlus) && (convert_to_semiring == FSTPlusTimes)) {
                    // aw = x->semiring()->convert(aw);
                    // return is_better(x, zero) ? exp(-x) : 0;
                    aw = (aw > 1e30 - 1e-3) ? exp(-aw) : 0;
                } else if ((x->semiring() == FSTPlusTimes) && (convert_to_semiring == FSTMinPlus)) {
                    // return (x > 0) ? -log(x) : zero;
                    aw = (aw > 0) ? -log(aw) : 1e30;
                }

                add_arc(p, j->next,
                        _transform(j, input), _transform(j, output),
                        clear ? _semiring->one : aw);
            }
        }
    } else {                                    // reversing
        fst::set<int> final_states;
        float initial_potential = _semiring->zero;
        for (FSTStateIter i = x->states(); i.valid(); i++) {
            int p = *i;
            if (!clear)
                set_potential(p, x->potential(p));
            if (x->is_final(p)) {
                final_states.insert(p);
                initial_potential = _semiring->better(initial_potential,
                                                      x->potential(p));
            }
            FSTArcIter j = x->arcs_out(p);
#ifndef USE_FORCE
            int new_states = x->n_states() - n_states();
            if (new_states > 0)
                add_states(new_states);
#endif
            for (; j.valid(); j++)
                add_arc(j->next, p,
                        _transform(j, input), _transform(j, output),
                        clear ? _semiring->one : j->weight);
        }
        int new_initial = add_state();
        set_initial(new_initial);
        set_final(x->initial(), _semiring->one);
        for (fst::set<int>::const_iterator k = final_states.begin();
             k != final_states.end(); k++) {
            add_arc(new_initial, *k, 0, 0,
                    clear ? _semiring->one : x->final_weight(*k));
        }
        if (!clear)
            set_potential(new_initial, initial_potential);
    }
}

// Load from basic file.
//
FSTBasicRep::FSTBasicRep(istream& is, const char* header)
{
    _init();

    const char* h;
    string line;
    char input[1024], output[1024];             // Yuck, hard-coded sizes.
    int p, q, n, new_states;
    float weight;
    float scale = 1;
    float offset = 0;

    if (strstr(header, "PlusTimesLog"))
        _semiring = FSTPlusTimesLog;
    else if (strstr(header, "PlusTimes"))
        _semiring = FSTPlusTimes;
    else if (strstr(header, "MaxPlus"))
        _semiring = FSTMaxPlus;
    else // default is MinPlusSemiring
        _semiring = FSTMinPlus;

    h = strstr(header, "WeightScale=");
    if (h) {
        sscanf(h, "WeightScale=%f", &scale);
    }

    h = strstr(header, "WeightOffset=");
    if (h) {
        sscanf(h, "WeightOffset=%f", &offset);
    }

    while (!getline(is, line).fail()) {
        const char *linec = line.c_str();
        if (strncmp(linec, "FSTBasic", 8) == 0)
            continue;
        switch (linec[0]) {
        case 'F':
            n = sscanf(linec, "F %d %f", &p, &weight);
            if (n < 1)
                FSTFatal("usage: F <state>");
            if (n < 2)
                weight = _semiring->one;
            weight = _adjust(weight, scale, offset);
        new_states = p - _n_states + 1;
        if (new_states > 0)
        add_states(new_states);
        set_final(p, weight);
            break;
        case 'I':
            n = sscanf(linec, "I %d", &p);
            if (n != 1)
                FSTFatal("file: usage: I <state>");
        new_states = p - _n_states + 1;
        if (new_states > 0)
        add_states(new_states);
        if (_initial >= 0 && p != _initial)
        FSTWarn("file: changing initial state from %d to %d "
             "(only one allowed)", _initial, p);
        set_initial(p);
            break;
        case 'P':
            n = sscanf(linec, "P %d %f", &p, &weight);
            weight = _adjust(weight, scale, offset);
            if (n < 2)
                FSTFatal("usage: P <state>");
        new_states = p - _n_states + 1;
        if (new_states > 0)
        add_states(new_states);
        set_potential(p, weight);
            break;
        case 'T':
            n = sscanf(linec, "T %d %d %s %s %f",
                       &p, &q, input, output, &weight);
            if (n < 2)
                FSTFatal("file: usage: "
                      "T <from> <to> [<input> [<output> [<weight>]]]");
            if (n < 5)
                weight = _semiring->one;
            weight = _adjust(weight, scale, offset);
            if (n < 4 || strcmp(output, ",") == 0)
                output[0] = 0;
            if (n < 3 || strcmp(input, ",") == 0)
                input[0] = 0;
        new_states = _max(p, q) - _n_states + 1;
        if (new_states > 0)
        add_states(new_states);
        add_arc(p, q, FSTstring_map[input], FSTstring_map[output], weight);
            break;
        }
    }
}

FSTBasicRep::~FSTBasicRep()
{
    for (_StatePtrVector::iterator i = _states.begin();
     i != _states.end(); i++) {
        delete *i;
    }
}

const fst::set<int>&
FSTBasicRep::input_alphabet() const
{
    return _input_alphabet;
}

const fst::set<int>&
FSTBasicRep::output_alphabet() const
{
    return _output_alphabet;
}

bool
FSTBasicRep::is_final(int p) const
{
    FST_ASSERT_VALID_STATE(p);
    return _states[p]->weight != _semiring->zero;
}

float
FSTBasicRep::final_weight(int p) const
{
    FST_ASSERT_VALID_STATE(p);
    return _states[p]->weight;
}

float
FSTBasicRep::potential(int p) const
{
    FST_ASSERT_VALID_STATE(p);
    return _states[p]->potential;
}

void
FSTBasicRep::force() const
{
    // Nothing to do because we're not lazy.
}

void
FSTBasicRep::reset() const
{
    // Nothing to do because we're not lazy.
}

void
FSTBasicRep::dump_ascii(ostream& os, bool header) const
{
    if (_train_data)
        FSTFatal("FSTBasicRep::dump_ascii: can't dump when in training mode");
    FSTGenericInt::dump_ascii(os, header);
}

void
FSTBasicRep::dump_binary(ostream& os) const
{
    if (_train_data)
        FSTFatal("FSTBasicRep::dump_binary: can't dump when in training mode");
    FSTGenericInt::dump_binary(os);
}

int
FSTBasicRep::n_arcs_out(int p) const
{
    FST_ASSERT_VALID_STATE(p);
    return _states[p]->arcs_out.size();
}

FSTArcIter
FSTBasicRep::arcs_out(int p, FSTArcOrder order) const
{
    FST_ASSERT_VALID_STATE(p);
    if (order != FSTNone && order != _arc_order)
        _sort_arcs(order);
    return FSTArcIter(_states[p]->arcs_out);
}

FSTArc
FSTBasicRep::arc_out(int p) const
{
    assert(p < (_train_data->arc_map).size());

    return _train_data->arc_map[p];
}

FSTArcIter
FSTBasicRep::arcs_out_with_input(int p, int i, FSTArcMode m) const
{
    FST_ASSERT_VALID_STATE(p);
    if (_arc_order != FSTInput)
        _sort_arcs(FSTInput);
    FSTArcCompareInput compare;
    const FSTArcVector& arcs = _states[p]->arcs_out;
    FSTArc a;
    a.input = i;
    typedef FSTArcVector::const_iterator Iter;
    Iter begin = arcs.end(), end = arcs.end();
    pair<Iter, Iter> range;
    switch (m) {
    case FSTEqual:
        range = equal_range(arcs.begin(), arcs.end(), &a, compare);
        begin = range.first;
        end = range.second;
        break;
    case FSTLessEqual:
        begin = arcs.begin();
        end = upper_bound(arcs.begin(), arcs.end(), &a, compare);
        break;
    case FSTGreaterEqual:
        begin = lower_bound(arcs.begin(), arcs.end(), &a, compare);
        end = arcs.end();
        break;
    }
    return FSTArcIter(arcs, begin, end);
}

FSTArcIter
FSTBasicRep::arcs_out_with_output(int p, int o, FSTArcMode m) const
{
    FST_ASSERT_VALID_STATE(p);
    if (_arc_order != FSTOutput)
        _sort_arcs(FSTOutput);
    FSTArcCompareOutput compare;
    const FSTArcVector& arcs = _states[p]->arcs_out;
    FSTArc a;
    a.output = o;
    typedef FSTArcVector::const_iterator Iter;
    Iter begin = arcs.end(), end = arcs.end();
    pair<Iter, Iter> range;
    switch (m) {
    case FSTEqual:
        range = equal_range(arcs.begin(), arcs.end(), &a, compare);
        begin = range.first;
        end = range.second;
        break;
    case FSTLessEqual:
        begin = arcs.begin();
        end = upper_bound(arcs.begin(), arcs.end(), &a, compare);
        break;
    case FSTGreaterEqual:
        begin = lower_bound(arcs.begin(), arcs.end(), &a, compare);
        end = arcs.end();
        break;
    }
    return FSTArcIter(arcs, begin, end);
}

void
FSTBasicRep::arcs_out(FSTArcVector& result, int p, FSTArcOrder order) const
{
    FST_ASSERT_VALID_STATE(p);
    if (order != FSTNone && order != _arc_order)
        _sort_arcs(order);
    FSTArcVector& arcs = _states[p]->arcs_out;
    for (FSTArcVector::iterator i = arcs.begin(); i != arcs.end(); i++) {
        result.push_back(new FSTArc(*i));
    }
}

void
FSTBasicRep::set_initial(int p)
{
    assert(!_train_data);
    assert(p >= 0);
    _maybe_grow(p);
    _initial = p;
}

void
FSTBasicRep::set_final(int p, float w)
{
    assert(!_train_data);
    assert(p >= 0);
    _maybe_grow(p);
    _states[p]->weight = w;
}

void
FSTBasicRep::set_potential(int p, float w)
{
    assert(!_train_data);
    assert(p >= 0);
    _maybe_grow(p);
    _states[p]->potential = w;
}

int
FSTBasicRep::add_states(int n)
{
    assert(!_train_data);
    assert(n >= 0);
    if (n > 0) {
        int first = _n_states;
        _n_states += n;
        while (n-- > 0)
            _states.push_back(new _State(_semiring));
        assert((size_t)_n_states == _states.size());
        return first;
    } else {
        return -1;
    }
}

void
FSTBasicRep::add_arc(int p, int q, int i, int o, float w)
{
    assert(!_train_data);
    assert(p >= 0);
    _maybe_grow(p);
    assert(q >= 0);
    _maybe_grow(q);
    _input_alphabet.insert(i);
    _output_alphabet.insert(o);
    FSTArcVector& arcs = _states[p]->arcs_out;
    FSTArc* a = new FSTArc(p, q, i, o, w);
    switch (_arc_order) {
    case FSTNone:
        arcs.push_back(a);
        break;
    case FSTInput:
        arcs.insert(lower_bound(arcs.begin(), arcs.end(), a,
                                FSTArcCompareInput()), a);
        break;
    case FSTOutput:
        arcs.insert(lower_bound(arcs.begin(), arcs.end(), a,
                                FSTArcCompareOutput()), a);
        break;
    }
    _n_arcs++;
}

void
FSTBasicRep::add_arcs(int p, FSTArcVector& a, FSTArcOrder order)
{
    // Note: this steals the arcs contained in a, leaving a empty.

#ifndef NDEBUG
    FST_ASSERT_VALID_STATE(p);
    assert(!_train_data);
    for (FSTArcVector::iterator i = a.begin(); i != a.end(); i++) {
    int q = (*i)->next;
    FST_ASSERT_VALID_STATE(q);
    }
#endif
    // Update alphabets.
    for (FSTArcVector::iterator i = a.begin(); i != a.end(); i++) {
        _input_alphabet.insert((*i)->input);
        _output_alphabet.insert((*i)->output);
    }

    // First, stick new arcs at end, remembering how many were originally there
    // (n).
    //
    FSTArcVector& arcs = _states[p]->arcs_out;
    int n = arcs.size();
    arcs.insert(arcs.end(), a.begin(), a.end());

    // Now, fix sorting if necessary.
    //
    switch (_arc_order) {
    case FSTInput:
        if (order != FSTInput)
        sort(arcs.begin() + n, arcs.end(), FSTArcCompareInput());
    inplace_merge(arcs.begin(), arcs.begin() + n, arcs.end(),
              FSTArcCompareInput());
    break;
    case FSTOutput:
    if (order != FSTOutput)
        sort(arcs.begin() + n, arcs.end(), FSTArcCompareOutput());
    inplace_merge(arcs.begin(), arcs.begin() + n, arcs.end(),
              FSTArcCompareOutput());
    break;
    case FSTNone:
    break;
    }
    _n_arcs += a.size();

    // Finally, clear out input a since we stole its arcs.
    //
    a.erase(a.begin(), a.end());
}

void
FSTBasicRep::set_arcs(int p, FSTArcVector& a, FSTArcOrder order)
{
    // Note: this steals the arcs contained in a, leaving a empty.

#ifndef NDEBUG
    FST_ASSERT_VALID_STATE(p);
    assert(!_train_data);
    for (FSTArcVector::iterator i = a.begin(); i != a.end(); i++) {
    int q = (*i)->next;
    FST_ASSERT_VALID_STATE(q);
    }
#endif

    // Update alphabets.
    //
    for (FSTArcVector::iterator i = a.begin(); i != a.end(); i++) {
        _input_alphabet.insert((*i)->input);
        _output_alphabet.insert((*i)->output);
    }

    // Remove old arcs.
    //
    FSTArcVector& arcs = _states[p]->arcs_out;
    for (FSTArcVector::iterator j = arcs.begin();
     j != arcs.end(); j++)
    delete *j;
    _n_arcs -= arcs.size();
    _n_arcs += a.size();
    arcs.erase(arcs.begin(), arcs.end());

    // Copy arcs.
    //
    arcs.insert(arcs.end(), a.begin(), a.end());

    // Fix sorting if necessary.
    //
    switch (_arc_order) {
    case FSTInput:
        if (order != FSTInput)
            sort(arcs.begin(), arcs.end(), FSTArcCompareInput());
        break;
    case FSTOutput:
        if (order != FSTOutput)
            sort(arcs.begin(), arcs.end(), FSTArcCompareOutput());
        break;
    case FSTNone:
        break;
    }

    // Finally, clear out input a since we stole its arcs.
    //
    a.erase(a.begin(), a.end());
}

void
FSTBasicRep::delete_state(int p)
{
    assert(!_train_data);
    FST_ASSERT_VALID_STATE(p);
    fst::vector<int> states;
    states.push_back(p);
    delete_states(states);
}

void
FSTBasicRep::delete_states(const fst::vector<int>& states)
{
    assert(!_train_data);
    if (_n_states == 0)
    return;
    assert(_n_states > 0);

    // Fill keep[] with boolean according to list of states.
    //
    fst::vector<bool> keep(_n_states, char(true));
    for (fst::vector<int>::const_iterator i = states.begin();
     i != states.end(); i++) {
    FST_ASSERT_VALID_STATE(*i);
    keep[*i] = false;
    }

    // Fill state_map[] so that it maps from old states to new states.
    //
    fst::vector<int> state_map(_n_states);
    state_map[0] = 0;
    int n_delete = !keep[0];
    for (int i = 1; i < _n_states; i++) {
    state_map[i] = state_map[i - 1] + keep[i - 1];
    if (!keep[i])
        n_delete++;
    }

    // Now we sweep through all states and arcs.  For each state i to be deleted
    // (i.e., keep[i] is false) we delete the state, which in turn deletes its
    // leaving arcs.  For other states, delete arcs headed to deleted states,
    // and relabel the states in remaining arcs according to state_map (computed
    // above).  Note that none of this disturbs any arc ordering in effect.
    //
    // Also, update the alphabets as we sweep through the arcs.
    //
    _input_alphabet.clear();
    _output_alphabet.clear();
    _n_arcs = 0;
    for (int i = 0; i < _n_states; i++) {
    FSTArcVector& arcs = _states[i]->arcs_out;
    FSTArcVector::iterator j, k;
    if (!keep[i]) {
        delete _states[i];
    } else {
        for (j = k = arcs.begin(); j != arcs.end(); j++) {
        if (!keep[(*j)->next]) {
            delete *j;
        } else {
                    _input_alphabet.insert((*j)->input);
                    _output_alphabet.insert((*j)->output);
            (*j)->next = state_map[(*j)->next];
                    (*j)->prev = state_map[(*j)->prev];
            *k++ = *j;
        }
        }
        arcs.erase(k, arcs.end());
            _n_arcs += arcs.size();
    }
    }

    // Move states to their new location and then delete unused portion of
    // _states vector at end.  We do not need to delete the associated state
    // pointers since those have already been deleted or moved above.
    //
    for (int i = 0; i < _n_states; i++)
    if (keep[i])
        _states[state_map[i]] = _states[i];

    _states.erase(_states.begin() + _n_states - n_delete, _states.end());
    _n_states = _states.size();

    if (_initial >= 0){
        if (keep[_initial])
            _initial = state_map[_initial];
        else
            _initial = -1;
    }
}

FSTGeneric
FSTBasicRep::copy(const FSTGeneric& x)
{
    return new FSTBasicRep(x, FSTInput, FSTOutput, false, false);
}

void
FSTBasicRep::_init()
{
    _class_name = "FSTBasic";
    _arc_order = FSTNone;
    _is_mutable = true;
    _train_data = 0;
}

void
FSTBasicRep::_sort_arcs(FSTArcOrder order) const
{
    if (order == _arc_order)
        return;
    if (order == FSTNone) {
        _arc_order = FSTNone;
        return;
    }
    if (_arc_order != FSTNone)
        FSTWarn("FSTBasic: resorting arcs from order %d to order %d is inefficient",
                int(_arc_order), int(order));
    _arc_order = order;
    FSTArcCompareInput input_compare;
    FSTArcCompareOutput output_compare;
    switch(_arc_order) {
    case FSTInput:
    for (_StatePtrVector::const_iterator i = _states.begin();
         i != _states.end(); i++) {
        FSTArcVector& arcs = (*i)->arcs_out;
        sort(arcs.begin(), arcs.end(), input_compare);
    }
    break;
    case FSTOutput:
    for (_StatePtrVector::const_iterator i = _states.begin();
         i != _states.end(); i++) {
        FSTArcVector& arcs = (*i)->arcs_out;
        sort(arcs.begin(), arcs.end(), output_compare);
    }
    break;
    case FSTNone:
    break;
    }
}

void
FSTBasicRep::_maybe_grow(int p)
{
    if (p >= _n_states)
        add_states(p - _n_states + 1);
}

void
FSTBasicRep::_train_begin()
{
    _StatePtrVector::iterator si;
    FSTArcVector::iterator ai;

    // Initialize training data.
    //
    _train_data = new _TrainData(_n_arcs);

    // Store arc index on output of every arc, and convert all weights to
    // PlusTimes semiring.
    //
    for (si = _states.begin(); si != _states.end(); si++) {
        _State* s = *si;
        s->weight = _semiring->convert(s->weight);
        for (ai = s->arcs_out.begin(); ai != s->arcs_out.end(); ai++) {
            FSTArc* a = *ai;
            int index = _train_data->arc_map[a];
            _train_data->save_outputs[index] = a->output;
            a->output = index;
            a->weight = _semiring->convert(a->weight);
        }
    }

    _train_data->save_semiring = _semiring;
    _semiring = FSTPlusTimes;
}

// mapping input:output of fst(this) to arcindex:output
void
FSTBasicRep::_train_begin_right()
{
    _StatePtrVector::iterator si;
    FSTArcVector::iterator ai;

    // Initialize training data.
    //
    _train_data = new _TrainData(_n_arcs);

    // Store arc index on output of every arc, and convert all weights to
    // PlusTimes semiring.
    for (si = _states.begin(); si != _states.end(); si++) {
        _State* s = *si;
        //        cerr << "weight: " << s->weight;
        s->weight = _semiring->convert(s->weight);
        //        cerr << " converted weight: " << s->weight << endl;
        for (ai = s->arcs_out.begin(); ai != s->arcs_out.end(); ai++) {
            FSTArc* a = *ai;
            int index = _train_data->arc_map[a];
            _train_data->save_inputs[index] = a->input;
            _train_data->save_outputs[index] = a->output;
            //cerr << "r: (" << _train_data->save_inputs[index] << ":" << index << ")" << endl;
            a->input = index;
            a->weight = _semiring->convert(a->weight);
        }
    }

    _train_data->save_semiring = _semiring;
    _semiring = FSTPlusTimes;
}

// ASSUMING _train_begin_right() has been called, so
// need to undo it first.
// mapping input:output of fst(this) to input:arcindex
void
FSTBasicRep::_train_begin_left()
{
    _StatePtrVector::iterator si;
    FSTArcVector::iterator ai;

    // Restore arc outputs and convert all weights to original semiring.
    //
    _semiring = _train_data->save_semiring;
    for (si = _states.begin(); si != _states.end(); si++) {
        _State* s = *si;
        s->weight = _semiring->unconvert(s->weight);
        for (ai = s->arcs_out.begin(); ai != s->arcs_out.end(); ai++) {
            FSTArc* a = *ai;
            int index = a->input;
            a->output = _train_data->save_outputs[index];
            a->input = _train_data->save_inputs[index];
            a->weight = _semiring->unconvert(a->weight);
        }
    }

    for (si = _states.begin(); si != _states.end(); si++) {
        _State* s = *si;
        s->weight = _semiring->convert(s->weight);
        for (ai = s->arcs_out.begin(); ai != s->arcs_out.end(); ai++) {
            FSTArc* a = *ai;
            int index = _train_data->arc_map[a];
            assert(_train_data->save_inputs[index] == a->input);
            assert(_train_data->save_outputs[index] == a->output);
            //cerr << "l: (" << _train_data->save_inputs[index] << ":" << index << ")" << endl;
            a->output = index;
            a->weight = _semiring->convert(a->weight);
        }
    }

    _train_data->save_semiring = _semiring;
    _semiring = FSTPlusTimes;
}



void
FSTBasicRep::_train_end()
{
    _StatePtrVector::iterator si;
    FSTArcVector::iterator ai;

    // Restore arc outputs and convert all weights to original semiring.
    //
    _semiring = _train_data->save_semiring;
    for (si = _states.begin(); si != _states.end(); si++) {
        _State* s = *si;
        s->weight = _semiring->unconvert(s->weight);
        for (ai = s->arcs_out.begin(); ai != s->arcs_out.end(); ai++) {
            FSTArc* a = *ai;
            int index = a->output;
            a->output = _train_data->save_outputs[index];
            a->weight = _semiring->unconvert(a->weight);
        }
    }
    delete _train_data;
    _train_data = 0;
}

void
FSTBasicRep::restore_input_output_index()
{
    _StatePtrVector::iterator si;
    FSTArcVector::iterator ai;

    // Restore arc input and outputs index with their label coding
    //
    for (si = _states.begin(); si != _states.end(); si++) {
        _State* s = *si;
        for (ai = s->arcs_out.begin(); ai != s->arcs_out.end(); ai++) {
            FSTArc* a = *ai;
            int index = atoi(FSTstring_map[a->input]);
            a->input = index;
            index = atoi(FSTstring_map[a->output]);
            a->output = index;
        }
    }
}

void
FSTBasicRep::_train_normalize()
{
    _StatePtrVector::iterator si;
    FSTArcVector::iterator ai;

    // Normalize weights so arcs and final weights for each state sum to one.
    //
    assert(_semiring == FSTPlusTimes);
    for (si = _states.begin(); si != _states.end(); si++) {
        _State* s = *si;
        float sum = s->weight;
        for (ai = s->arcs_out.begin(); ai != s->arcs_out.end(); ai++) {
            FSTArc* a = *ai;
            sum += a->weight;
        }
        if (sum == 0)
            continue;
        float factor = 1 / sum;
        s->weight *= factor;
        for (ai = s->arcs_out.begin(); ai != s->arcs_out.end(); ai++) {
            FSTArc* a = *ai;
            a->weight *= factor;
        }
    }
}

void
FSTBasicRep::_train_normalize_by_outputlabel()
{
    _StatePtrVector::iterator si;
    FSTArcVector::iterator ai;

    int eps_label = FSTstring_map[""];
        
    // Normalize weights so arcs and final weights for each state sum to one.
    //
    assert(_semiring == FSTPlusTimes);
    for (si = _states.begin(); si != _states.end(); si++) {
        _State* s = *si;

        fst::set<int> output_set;

        // this ensures eps_label thus s->weight is computed first 
        output_set.insert(eps_label);

        for (ai = s->arcs_out.begin(); ai != s->arcs_out.end(); ai++) {
            FSTArc* a = *ai;
            int index = _train_data->arc_map[a];
            output_set.insert(_train_data->save_outputs[index]);
        }
        
        for (fst::set<int>::iterator output_label_itr = output_set.begin(); 
             output_label_itr != output_set.end(); output_label_itr++) {
            
            double sum = 0.0;
            if (*output_label_itr == eps_label)
                sum += s->weight;
  
            for (ai = s->arcs_out.begin(); ai != s->arcs_out.end(); ai++) {
                FSTArc* a = *ai;
                int index = _train_data->arc_map[a];
                if (_train_data->save_outputs[index] == *output_label_itr) {
                    sum += a->weight;
                }
            }

            if (sum == 0.0)
                continue;
            double factor = 1.0 / sum;
            if (*output_label_itr == eps_label)
                s->weight *= factor;
            for (ai = s->arcs_out.begin(); ai != s->arcs_out.end(); ai++) {
                FSTArc* a = *ai;
                int index = _train_data->arc_map[a];
                if (_train_data->save_outputs[index] == *output_label_itr) {
                    a->weight *= factor;
                }
            }
        }
    }
}

void
FSTBasicRep::_train_update(fst::vector<double>& a_counts,
                           fst::vector<double>& f_counts,
                           double floor)
{
    int a, p;

    // Copy counts to weights.
    //
    if (int(a_counts.size()) <= _n_arcs) {
        a_counts.insert(a_counts.end(), _n_arcs - a_counts.size() + 1, floor);
    }
    if (int(f_counts.size()) < _n_states) {
        f_counts.insert(f_counts.end(), _n_states - f_counts.size(), floor);
    }
    assert(int(a_counts.size()) == _n_arcs + 1);
    assert(int(f_counts.size()) == _n_states);

    for (p = 0; p < _n_states; p++) {
        if (is_final(p)) {
            _states[p]->weight = floor + f_counts[p];
        }
    }
    for (a = 1; a <= _n_arcs; a++) {
        _train_data->arc_map[a]->weight = floor + a_counts[a];
    }

    // Normalize weights.
    //
    //_train_normalize_by_outputlabel();
    _train_normalize();

    // Return normalized weights in counts.
    //
    for (p = 0; p < _n_states; p++)
        f_counts[p] = _states[p]->weight;
    for (a = 1; a <= _n_arcs; a++)
        a_counts[a] = _train_data->arc_map[a]->weight;
}

FSTGeneric
FSTLoadBasic(istream& is, const char* header)
{
    return new FSTBasicRep(is, header);
}

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 80
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
