////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

#undef USE_HASH

#include <cstring>
#include "fst/FST_i.h"
#ifdef USE_HASH
#include "fst/FSTTimer.h"
#include <hash_map>
#endif

class FSTBinaryRep
    : public FSTGenericInt {

public: // methods

    FSTBinaryRep(istream& is, const char* header);
    virtual ~FSTBinaryRep();

    virtual const fst::set<int>& input_alphabet() const;
    virtual const fst::set<int>& output_alphabet() const;
    virtual int        n_arcs_out(int p) const;
    virtual FSTArcIter arcs_out(int p, FSTArcOrder order=FSTNone) const;
    virtual FSTArcIter arcs_out_with_input(int p, int i,
                                           FSTArcMode m=FSTEqual) const;
    virtual FSTArcIter arcs_out_with_output(int p, int o,
                                            FSTArcMode m=FSTEqual) const;
    virtual bool       is_final(int p) const;
    virtual float      final_weight(int p) const;
    virtual float      potential(int p) const;
    virtual void       force() const;
    virtual void       reset() const;
    virtual void       dump_ascii(ostream& os, bool header=true) const;
    virtual void       dump_binary(ostream& os) const;

protected: // types

    struct _State;
    struct _TrainData;
    typedef FSTArcVector::iterator _ArcIterator;
#ifdef USE_HASH
    typedef pair<int, int> _StateLabel;
    typedef pair<_ArcIterator, _ArcIterator> _ArcRange;
    struct  _StateLabelHash {
        size_t operator()(const _StateLabel& x) const {
            return 193*x.first + x.second;
        }
    };
    typedef fst::hash_map<_StateLabel, _ArcRange, _StateLabelHash> _HashMap;
#endif

protected: // members

    _State* _states;
    FSTArc* _arcs;
    fst::vector<FSTArc*> _arc_ptrs;
    mutable FSTArcOrder _arc_order;
#ifdef USE_HASH
    mutable _HashMap _hash_map;
#endif
    fst::set<int> _input_alphabet;
    fst::set<int> _output_alphabet;
    _TrainData* _train_data;

protected: // methods

    float _adjust(float weight, float scale, float offset) const {
        if (weight != _semiring->zero)
            weight = scale*weight + offset;
        return weight;
    }
    void _init();
    void _sort_arcs(FSTArcOrder order) const;
    virtual void _train_begin();
    virtual void _train_end();
    virtual void _train_normalize();
    virtual void _train_update(fst::vector<double>& a_counts,
                               fst::vector<double>& f_counts,
                               double floor);
};

struct FSTBinaryRep::_State {
    float weight;
    float potential;
    _ArcIterator arcs_out_begin;
    _ArcIterator arcs_out_end;
};

struct FSTBinaryRep::_TrainData {
    FSTSemiring save_semiring;
    fst::vector<int> save_outputs;

    _TrainData(int n_arcs) : save_outputs(n_arcs) {}
};

static void
_load_alphabet(fst::set<int>& alphabet,
              fst::vector<int>& mapping, istream& is)
{
    int size = FSTLoadInt(is);

    mapping.clear();
    mapping.reserve(size);

    string s;
    fst::set<int>::iterator pos = alphabet.begin();
    while (size-- > 0) {
        FSTLoadString(is, s);
        int l = FSTstring_map[s.c_str()];
        mapping.push_back(l);
        pos = alphabet.insert(pos, l);
    }
}

// Load from binary file.
//
FSTBinaryRep::FSTBinaryRep(istream& is, const char* header)
{
    const char* h;

#if 0
    is.setf(ios_base::binary);
    cerr << "binary? " << bool(is.flags() & ios_base::binary) << endl << flush;
#endif

    _init();

    if (strstr(header, "PlusTimesLog"))
        _semiring = FSTPlusTimesLog;
    else if (strstr(header, "PlusTimes"))
        _semiring = FSTPlusTimes;
    else if (strstr(header, "MaxPlus"))
        _semiring = FSTMaxPlus;
    else // default is MinPlusSemiring
        _semiring = FSTMinPlus;

    float scale = 1;
    h = strstr(header, "WeightScale=");
    if (h) {
        sscanf(h, "WeightScale=%f", &scale);
    }

    float offset = 0;
    h = strstr(header, "WeightOffset=");
    if (h) {
        sscanf(h, "WeightOffset=%f", &offset);
    }

    // arc order
    _arc_order = FSTArcOrder(FSTLoadInt(is));

    // initial state
    _initial = FSTLoadInt(is);

    // input labels
    fst::vector<int> input_mapping;
    _load_alphabet(_input_alphabet, input_mapping, is);

    // output labels
    fst::vector<int> output_mapping;
    _load_alphabet(_output_alphabet, output_mapping, is);

    // # arcs
    _n_arcs = FSTLoadInt(is);

    // arcs
    _arcs = new FSTArc[_n_arcs];
    _arc_ptrs.resize(_n_arcs);
    if (scale == 1 && offset == 0) {
        for (int a = 0; a < _n_arcs; a++) {
            FSTArc* ap = _arc_ptrs[a] = _arcs + a;
            ap->prev = FSTLoadInt(is);
            ap->next = FSTLoadInt(is);
            ap->input = input_mapping[FSTLoadInt(is)];
            ap->output = output_mapping[FSTLoadInt(is)];
            ap->weight = FSTLoadFloat(is);
        }
    } else {
        for (int a = 0; a < _n_arcs; a++) {
            FSTArc* ap = _arc_ptrs[a] = _arcs + a;
            ap->prev = FSTLoadInt(is);
            ap->next = FSTLoadInt(is);
            ap->input = input_mapping[FSTLoadInt(is)];
            ap->output = output_mapping[FSTLoadInt(is)];
            ap->weight = _adjust(FSTLoadFloat(is), scale, offset);
        }
    }

    // # states
    _n_states = FSTLoadInt(is);

    // states
    _states = new _State[_n_states];
    if (scale == 1 && offset == 0) {
        for (int s = 0; s < _n_states; s++) {
            _State* sp = _states + s;
            sp->weight = FSTLoadFloat(is);
            sp->potential = FSTLoadFloat(is);
            sp->arcs_out_begin = _arc_ptrs.begin() + FSTLoadInt(is);
            sp->arcs_out_end = _arc_ptrs.begin() + FSTLoadInt(is);
        }
    } else {
        for (int s = 0; s < _n_states; s++) {
            _State* sp = _states + s;
            sp->weight = _adjust(FSTLoadFloat(is), scale, offset);
            sp->potential = _adjust(FSTLoadFloat(is), scale, offset);
            sp->arcs_out_begin = _arc_ptrs.begin() + FSTLoadInt(is);
            sp->arcs_out_end = _arc_ptrs.begin() + FSTLoadInt(is);
        }
    }
}

FSTBinaryRep::~FSTBinaryRep()
{
    delete[] _states;
    delete[] _arcs;
}

const fst::set<int>&
FSTBinaryRep::input_alphabet() const
{
    return _input_alphabet;
}

const fst::set<int>&
FSTBinaryRep::output_alphabet() const
{
    return _output_alphabet;
}

bool
FSTBinaryRep::is_final(int p) const
{
    FST_ASSERT_VALID_STATE(p);
    return _states[p].weight != _semiring->zero;
}

float
FSTBinaryRep::final_weight(int p) const
{
    FST_ASSERT_VALID_STATE(p);
    return _states[p].weight;
}

float
FSTBinaryRep::potential(int p) const
{
    FST_ASSERT_VALID_STATE(p);
    return _states[p].potential;
}

void
FSTBinaryRep::force() const
{
    // Nothing to do because we're not lazy.
}

void
FSTBinaryRep::reset() const
{
    // Nothing to do because we're not lazy.
}

void
FSTBinaryRep::dump_ascii(ostream& os, bool header) const
{
    if (_train_data)
        FSTFatal("FSTBinaryRep::dump_ascii: can't dump when in training mode");
    FSTGenericInt::dump_ascii(os, header);
}

void
FSTBinaryRep::dump_binary(ostream& os) const
{
    if (_train_data)
        FSTFatal("FSTBinaryRep::dump_binary: can't dump when in training mode");
    FSTGenericInt::dump_binary(os);
}

int
FSTBinaryRep::n_arcs_out(int p) const
{
    FST_ASSERT_VALID_STATE(p);
    return _states[p].arcs_out_end - _states[p].arcs_out_begin;
}

FSTArcIter
FSTBinaryRep::arcs_out(int p, FSTArcOrder order) const
{
    FST_ASSERT_VALID_STATE(p);
    if (order != FSTNone && order != _arc_order)
        _sort_arcs(order);
    return FSTArcIter(_arc_ptrs, _states[p].arcs_out_begin, _states[p].arcs_out_end);
}

FSTArcIter
FSTBinaryRep::arcs_out_with_input(int p, int i, FSTArcMode m) const
{
    FST_ASSERT_VALID_STATE(p);
    if (_arc_order != FSTInput)
        _sort_arcs(FSTInput);
#ifdef USE_HASH
    _HashMap::const_iterator iter = _hash_map.find(_StateLabel(p, i));
    if (iter != _hash_map.end()) {
        _ArcRange range = (*iter).second;
        if (m == FSTLessEqual)
            range.first = _states[p].arcs_out_begin;
        else if (m == FSTGreaterEqual)
            range.second = _states[p].arcs_out_end;
        return FSTArcIter(range.first, range.second);
    } else if (m == FSTEqual) {
        return FSTArcIter();
    }
#endif
    typedef _ArcIterator Iter;
    Iter begin = _states[p].arcs_out_begin;
    Iter end = _states[p].arcs_out_end;
    FSTArcCompareInput compare;
    FSTArc a;
    a.input = i;
    pair<Iter, Iter> range;
    switch (m) {
    case FSTEqual:
        range = equal_range(begin, end, &a, compare);
        begin = range.first;
        end = range.second;
        break;
    case FSTLessEqual:
        end = upper_bound(begin, end, &a, compare);
        break;
    case FSTGreaterEqual:
        begin = lower_bound(begin, end, &a, compare);
        break;
    }
    return FSTArcIter(_arc_ptrs, begin, end);
}

FSTArcIter
FSTBinaryRep::arcs_out_with_output(int p, int o, FSTArcMode m) const
{
    FST_ASSERT_VALID_STATE(p);
    if (_arc_order != FSTOutput)
        _sort_arcs(FSTOutput);
#ifdef USE_HASH
    _HashMap::const_iterator iter = _hash_map.find(_StateLabel(p, o));
    if (iter != _hash_map.end()) {
        _ArcRange range = (*iter).second;
        if (m == FSTLessEqual)
            range.first = _states[p].arcs_out_begin;
        else if (m == FSTGreaterEqual)
            range.second = _states[p].arcs_out_end;
        return FSTArcIter(range.first, range.second);
    } else if (m == FSTEqual) {
        return FSTArcIter();
    }
#endif
    typedef fst::vector<FSTArc*>::const_iterator Iter;
    Iter begin = _states[p].arcs_out_begin;
    Iter end = _states[p].arcs_out_end;
    FSTArcCompareOutput compare;
    FSTArc a;
    a.output = o;
    pair<Iter, Iter> range;
    switch (m) {
    case FSTEqual:
        range = equal_range(begin, end, &a, compare);
        begin = range.first;
        end = range.second;
        break;
    case FSTLessEqual:
        end = upper_bound(begin, end, &a, compare);
        break;
    case FSTGreaterEqual:
        begin = lower_bound(begin, end, &a, compare);
        break;
    }
    return FSTArcIter(_arc_ptrs, begin, end);
}

void
FSTBinaryRep::_init()
{
    _class_name = "FSTBinary";
    _train_data = 0;
}

void
FSTBinaryRep::_sort_arcs(FSTArcOrder order) const
{
    if (order == _arc_order || order == FSTNone)
        return;
    if (_arc_order != FSTNone)
        FSTWarn("FSTBinary: resorting arcs from order %d to order %d is inefficient",
                int(_arc_order), int(order));
    _arc_order = order;
    FSTArcCompareInput input_compare;
    FSTArcCompareOutput output_compare;
    switch(_arc_order) {
    case FSTInput:
#ifdef USE_HASH
        _hash_map.clear();
        _hash_map.resize(_n_arcs);
#endif
        for (int s = 0; s < _n_states; s++) {
            _State* sp = _states + s;
            _ArcIterator end = sp->arcs_out_end;
            _ArcIterator i = sp->arcs_out_begin;
        sort(i, end, input_compare);
#ifdef USE_HASH
            while (i != end) {
                _ArcIterator j = i + 1;
                while (j != end && (*j)->input == (*i)->input)
                    ++j;
                _hash_map[_StateLabel(s, (*i)->input)] = _ArcRange(i, j);
                i = j;
            }
#endif
    }
    break;
    case FSTOutput:
#ifdef USE_HASH
        _hash_map.clear();
        _hash_map.resize(_n_arcs);
#endif
        for (int s = 0; s < _n_states; s++) {
            _State* sp = _states + s;
            _ArcIterator end = sp->arcs_out_end;
            _ArcIterator i = sp->arcs_out_begin;
        sort(i, end, output_compare);
#ifdef USE_HASH
            while (i != end) {
                _ArcIterator j = i + 1;
                while (j != end && (*j)->output == (*i)->output)
                    ++j;
                _hash_map[_StateLabel(s, (*i)->output)] = _ArcRange(i, j);
                i = j;
            }
#endif
    }
    break;
    case FSTNone:
    break;
    }
}

void
FSTBinaryRep::_train_begin()
{
    _ArcIterator ai;

    // Initialize training data.
    //
    _train_data = new _TrainData(_n_arcs);

    // Store arc index on output of every arc, and convert all weights to
    // PlusTimes semiring.
    //
    for (int p = 0; p < _n_states; p++) {
        _State* s = &_states[p];
        s->weight = _semiring->convert(s->weight);
        for (ai = s->arcs_out_begin; ai != s->arcs_out_end; ai++) {
            FSTArc* a = *ai;
            int index = ai - _arc_ptrs.begin();
            _train_data->save_outputs[index] = a->output;
            a->output = index;
            a->weight = _semiring->convert(a->weight);
        }
    }

    _train_data->save_semiring = _semiring;
    _semiring = FSTPlusTimes;
}

void
FSTBinaryRep::_train_end()
{
    _ArcIterator ai;

    // Restore arc outputs and convert all weights to original semiring.
    //
    _semiring = _train_data->save_semiring;
    for (int p = 0; p < _n_states; p++) {
        _State* s = &_states[p];
        s->weight = _semiring->unconvert(s->weight);
        for (ai = s->arcs_out_begin; ai != s->arcs_out_end; ai++) {
            FSTArc* a = *ai;
            a->output = _train_data->save_outputs[a->output];
            a->weight = _semiring->unconvert(a->weight);
        }
    }
    delete _train_data;
    _train_data = 0;
}

void
FSTBinaryRep::_train_normalize()
{
    _ArcIterator ai;

    // Normalize weights so arcs and final weights for each state sum to one.
    //
    assert(_semiring == FSTPlusTimes);
    for (int p = 0; p < _n_states; p++) {
        _State* s = &_states[p];
        float sum = s->weight;
        for (ai = s->arcs_out_begin; ai != s->arcs_out_end; ai++) {
            FSTArc* a = *ai;
            sum += a->weight;
        }
        if (sum == 0)
            continue;
        float factor = 1 / sum;
        s->weight *= factor;
        for (ai = s->arcs_out_begin; ai != s->arcs_out_end; ai++) {
            FSTArc* a = *ai;
            a->weight *= factor;
        }
    }
}

void
FSTBinaryRep::_train_update(fst::vector<double>& a_counts,
                            fst::vector<double>& f_counts,
                            double floor)
{
    int a, p;

    // Copy counts to weights.
    //
    assert(int(a_counts.size()) == _n_arcs);
    assert(int(f_counts.size()) == _n_states);
    for (p = 0; p < _n_states; p++)
        _states[p].weight = f_counts[p];
    for (a = 0; a < _n_arcs; a++)
        _arc_ptrs[a]->weight = a_counts[a];

    // Normalize weights.
    //
    _train_normalize();

    // Return normalized weights in counts.
    //
    for (p = 0; p < _n_states; p++)
        f_counts[p] = _states[p].weight;
    for (a = 0; a < _n_arcs; a++)
        a_counts[a] = _arc_ptrs[a]->weight;
}

FSTGeneric
FSTLoadBinary(istream& is, const char* header)
{
    return new FSTBinaryRep(is, header);
}

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 80
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
