////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

#include "fst/FST_i.h"

class FSTCacheRep
    : public FSTGenericInt {

public: // methods
    FSTCacheRep(const FSTGeneric& in);
    virtual ~FSTCacheRep();
    virtual const fst::set<int>& input_alphabet() const;
    virtual const fst::set<int>& output_alphabet() const;
    virtual bool is_final(int p) const;
    virtual float final_weight(int p) const;
    virtual float potential(int p) const;
    virtual int n_arcs_out(int p) const;
    virtual FSTArcIter arcs_out(int p, FSTArcOrder order=FSTNone) const;
    virtual FSTArcIter arcs_out_with_input(int p, int i,
                                           FSTArcMode m=FSTEqual) const;
    virtual FSTArcIter arcs_out_with_output(int p, int o,
                                            FSTArcMode m=FSTEqual) const;
    virtual void force() const;
    virtual void reset() const;

    virtual FSTGeneric reload(bool& flag);

    // Cache-specific methods:
    virtual float hit_rate() const;
    virtual float hit_rate(int& n_hits, int& n_total) const;

protected: // types
    struct _State;
    typedef fst::vector<_State*> _StatePtrVector;

protected: // members
    FSTGeneric _in;
    mutable _StatePtrVector _states;
    mutable FSTArcOrder _arc_order;
    mutable int _n_hits;
    mutable int _n_misses;
    mutable bool _forced;

protected: // methods
    void _sort_arcs(FSTArcOrder order) const;
    void _expand(int p) const;
    void _init_states(int p) const;
    void _init() const;
    void _clean() const;

};

FSTGeneric
FSTCache(const FSTGeneric& x)
{
    return FSTGeneric(new FSTCacheRep(x));
}

struct FSTCacheRep::_State {
    fst::vector<FSTArc*> arcs_out;
    _State() {}
    ~_State() {
        for (fst::vector<FSTArc*>::iterator i = arcs_out.begin();
             i != arcs_out.end(); i++) {
            delete *i;
        }
    }
    FSTAllocator_new_delete(_State);
};

FSTCacheRep::FSTCacheRep(const FSTGeneric& in)
    : _in(in)
{
    _class_name = "FSTCache";
    _semiring = _in->semiring();
    _init();
    _initial = _in->initial();
}

FSTCacheRep::~FSTCacheRep()
{
    _clean();
}

FSTGeneric
FSTCacheRep::reload(bool& flag)
{
    if (_in.reload()) flag = true;
    reset();
    return FSTGeneric(this);
}

void
FSTCacheRep::reset() const
{
    _in->reset();
    _clean();
    _init();
}

void
FSTCacheRep::_init() const
{
    _arc_order = FSTNone;
    _n_hits = 0;
    _n_misses = 0;
    _n_states = 0;
    _n_arcs = 0;
    _init_states(_in->n_states());
    _forced = false;
}

void
FSTCacheRep::_clean() const
{
    for (_StatePtrVector::iterator i = _states.begin();
     i != _states.end(); i++) {
        if (*i)
            delete *i;
    }
    _states.clear();
}

const fst::set<int>&
FSTCacheRep::input_alphabet() const
{
    return _in->input_alphabet();
}

const fst::set<int>&
FSTCacheRep::output_alphabet() const
{
    return _in->output_alphabet();
}

bool
FSTCacheRep::is_final(int p) const
{
    return _in->is_final(p);
}

float
FSTCacheRep::final_weight(int p) const
{
    return _in->final_weight(p);
}

float
FSTCacheRep::potential(int p) const
{
    return _in->potential(p);
}

int
FSTCacheRep::n_arcs_out(int p) const
{
    FST_ASSERT_VALID_STATE(p);
    if (_states[p]) {
        _n_hits++;
    } else {
        _n_misses++;
        _expand(p);
    }
    return _states[p]->arcs_out.size();
}

FSTArcIter
FSTCacheRep::arcs_out(int p, FSTArcOrder order) const
{
    FST_ASSERT_VALID_STATE(p);
    if (_states[p]) {
        _n_hits++;
    } else {
        _n_misses++;
        _expand(p);
    }
    if (order != FSTNone && order != _arc_order)
        _sort_arcs(order);
    const fst::vector<FSTArc*>& arcs = _states[p]->arcs_out;
    return FSTArcIter(arcs);
}

FSTArcIter
FSTCacheRep::arcs_out_with_input(int p, int i, FSTArcMode m) const
{
    FST_ASSERT_VALID_STATE(p);
    if (_arc_order != FSTInput)
        _sort_arcs(FSTInput);
    if (_states[p]) {
        _n_hits++;
    } else {
        _n_misses++;
        _expand(p);
    }
    const fst::vector<FSTArc*>& arcs = _states[p]->arcs_out;
    FSTArcCompareInput compare;
    FSTArc a;
    a.input = i;
    typedef fst::vector<FSTArc*>::const_iterator Iter;
    Iter begin = arcs.end(), end = arcs.end();
    pair<Iter, Iter> range;
    switch (m) {
    case FSTEqual:
        range = equal_range(arcs.begin(), arcs.end(), &a, compare);
        begin = range.first;
        end = range.second;
        break;
    case FSTLessEqual:
        begin = arcs.begin();
        end = upper_bound(arcs.begin(), arcs.end(), &a, compare);
        break;
    case FSTGreaterEqual:
        begin = lower_bound(arcs.begin(), arcs.end(), &a, compare);
        end = arcs.end();
        break;
    }
    return FSTArcIter(arcs, begin, end);
}

FSTArcIter
FSTCacheRep::arcs_out_with_output(int p, int o, FSTArcMode m) const
{
    FST_ASSERT_VALID_STATE(p);
    if (_arc_order != FSTOutput)
        _sort_arcs(FSTOutput);
    if (_states[p]) {
        _n_hits++;
    } else {
        _n_misses++;
        _expand(p);
    }
    const fst::vector<FSTArc*>& arcs = _states[p]->arcs_out;
    FSTArcCompareOutput compare;
    FSTArc a;
    a.output = o;
    typedef fst::vector<FSTArc*>::const_iterator Iter;
    Iter begin = arcs.end(), end = arcs.end();
    pair<Iter, Iter> range;
    switch (m) {
    case FSTEqual:
        range = equal_range(arcs.begin(), arcs.end(), &a, compare);
        begin = range.first;
        end = range.second;
        break;
    case FSTLessEqual:
        begin = arcs.begin();
        end = upper_bound(arcs.begin(), arcs.end(), &a, compare);
        break;
    case FSTGreaterEqual:
        begin = lower_bound(arcs.begin(), arcs.end(), &a, compare);
        end = arcs.end();
        break;
    }
    return FSTArcIter(arcs, begin, end);
}

void
FSTCacheRep::_expand(int p) const
{
    _states[p] = new _State;
    fst::vector<FSTArc*>& arcs = _states[p]->arcs_out;
    FSTArcIter i = _in->arcs_out(p, _arc_order);
    _init_states(_in->n_states());
    int n_new_arcs = i.n();
    if (n_new_arcs > 0) {
        arcs.reserve(n_new_arcs);
        for (; i.valid(); i++) {
            arcs.push_back(new FSTArc(*i));
        }
        _n_arcs += n_new_arcs;
    }
}

void
FSTCacheRep::_init_states(int p) const
{
    int n_new_states = p - _n_states;
    if (n_new_states > 0) {
        _states.resize(p + 1);
        _n_states += n_new_states;
    }
}

void
FSTCacheRep::force() const
{
    if (!_forced) {
        for (int p = 0; p < _n_states; p++)
            if (!_states[p])
                _expand(p);
        _forced = true;
    }
}

float
FSTCacheRep::hit_rate() const
{
    return float(_max(1, _n_hits)) / float(_max(1, _n_hits + _n_misses));
}

float
FSTCacheRep::hit_rate(int& n_hits, int& n_total) const
{
    n_hits = _n_hits;
    n_total = n_hits + _n_misses;
    return hit_rate();
}

void
FSTCacheRep::_sort_arcs(FSTArcOrder order) const
{
    if (order == _arc_order)
        return;
    if (order == FSTNone) {
        _arc_order = FSTNone;
        return;
    }
    if (_arc_order != FSTNone)
        FSTWarn("FSTCache: resorting arcs from order %d to order %d is inefficient",
             int(_arc_order), int(order));
    _arc_order = order;
    FSTArcCompareInput input_compare;
    FSTArcCompareOutput output_compare;
    switch(_arc_order) {
    case FSTInput:
    for (_StatePtrVector::iterator i = _states.begin();
         i != _states.end(); i++) {
            if (*i) {
                fst::vector<FSTArc*>& arcs = (*i)->arcs_out;
                sort(arcs.begin(), arcs.end(), input_compare);
            }
    }
    break;
    case FSTOutput:
    for (_StatePtrVector::iterator i = _states.begin();
         i != _states.end(); i++) {
            if (*i) {
                fst::vector<FSTArc*>& arcs = (*i)->arcs_out;
                sort(arcs.begin(), arcs.end(), output_compare);
            }
    }
    break;
    case FSTNone:
    break;
    }
}

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 80
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
