////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

#include "fst/FST_i.h"

class FSTCache2Rep
    : public FSTGenericInt {

public: // methods
    FSTCache2Rep(const FSTGeneric& in);
    virtual ~FSTCache2Rep();
    virtual const fst::set<int>& input_alphabet() const;
    virtual const fst::set<int>& output_alphabet() const;
    virtual bool is_final(int p) const;
    virtual float final_weight(int p) const;
    virtual float potential(int p) const;
    virtual FSTArcIter arcs_out(int p, FSTArcOrder order=FSTNone) const;
    virtual FSTArcIter arcs_out_with_input(int p, int i,
                                           FSTArcMode m=FSTEqual) const;
    virtual FSTArcIter arcs_out_with_output(int p, int o,
                                            FSTArcMode m=FSTEqual) const;
    virtual void force() const;
    virtual void reset() const;

    virtual FSTFinalStateIter final_states() const;
    virtual FSTArcIter arcs_in(int p, FSTArcOrder o=FSTNone) const;
    virtual FSTArcIter arcs_in_with_input(int p, int i,
                                          FSTArcMode m=FSTEqual) const;
    virtual FSTArcIter arcs_in_with_output(int p, int o,
                                           FSTArcMode m=FSTEqual) const;

    virtual FSTGeneric reload(bool& flag);

    // Cache2-specific methods:
    virtual float hit_rate() const;
    virtual float hit_rate(int& n_hits, int& n_total) const;

protected: // types
    struct _State;
    class _FinalStateIterRep;
    typedef fst::vector<_State*> _StatePtrVector;

protected: // members
    FSTGeneric _in;
    mutable _StatePtrVector _states;
    mutable FSTArcOrder _arc_out_order;
    mutable FSTArcOrder _arc_in_order;
    mutable fst::set<int> _final_states;
    mutable int _n_hits;
    mutable int _n_misses;
    mutable bool _forced;

protected: // methods
    void _sort_arcs_in(FSTArcOrder order) const;
    void _sort_arcs_out(FSTArcOrder order) const;
    void _expand(int p) const;
    void _init_states(int p) const;
    void _init() const;
    void _clean() const;

};

FSTGeneric
FSTCache2(const FSTGeneric& x)
{
    return FSTGeneric(new FSTCache2Rep(x));
}

struct FSTCache2Rep::_State {
    fst::vector<FSTArc*> arcs_out;
    fst::vector<FSTArc*> arcs_in;
    bool expanded;
    _State() : expanded(false) {}
    ~_State() {
        for (fst::vector<FSTArc*>::iterator i = arcs_out.begin();
             i != arcs_out.end(); i++) {
            delete *i;
        }
    }
    FSTAllocator_new_delete(_State);
};

class FSTCache2Rep::_FinalStateIterRep
    : public FSTFinalStateIterRep {

    fst::set<int>::const_iterator i, e;
    
public:
    _FinalStateIterRep(const fst::set<int>& x)
    : i(x.begin()), e(x.end()) {}
    virtual ~_FinalStateIterRep() {}
    virtual void operator++(int) { i++; }
    virtual bool valid() const   { return i != e; }
    virtual int state() const    { return *i; }
    FSTAllocator_new_delete(_FinalStateIterRep);
};

FSTCache2Rep::FSTCache2Rep(const FSTGeneric& in)
    : _in(in),
      _arc_out_order(FSTNone),
      _arc_in_order(FSTNone),
      _n_hits(0),
      _n_misses(0),
      _forced(false)
{
    _class_name = "FSTCache2";
    _is_bidirectional = true;
    _semiring = _in->semiring();
    _init();
    _initial = _in->initial();
}

FSTCache2Rep::~FSTCache2Rep()
{
    _clean();
}

FSTGeneric
FSTCache2Rep::reload(bool& flag)
{
    if (_in.reload()) flag = true;
    reset();
    return FSTGeneric(this);
}

void
FSTCache2Rep::reset() const
{
    _in->reset();
    _clean();
    _init();
}

void
FSTCache2Rep::_init() const
{
    _arc_out_order = FSTNone;
    _arc_in_order = FSTNone;
    _n_hits = 0;
    _n_misses = 0;
    _n_states = 0;
    _n_arcs = 0;
    _init_states(_in->n_states());
    _forced = false;
}

void
FSTCache2Rep::_clean() const
{
    for (_StatePtrVector::iterator i = _states.begin();
     i != _states.end(); i++) {
        if (*i)
            delete *i;
    }
    _states.clear();
}

const fst::set<int>&
FSTCache2Rep::input_alphabet() const
{
    return _in->input_alphabet();
}

const fst::set<int>&
FSTCache2Rep::output_alphabet() const
{
    return _in->output_alphabet();
}

bool
FSTCache2Rep::is_final(int p) const
{
    return _in->is_final(p);
}

float
FSTCache2Rep::final_weight(int p) const
{
    return _in->final_weight(p);
}

float
FSTCache2Rep::potential(int p) const
{
    return _in->potential(p);
}

FSTArcIter
FSTCache2Rep::arcs_out(int p, FSTArcOrder order) const
{
    FST_ASSERT_VALID_STATE(p);
    if (_states[p]->expanded) {
        _n_hits++;
    } else {
        _n_misses++;
        _expand(p);
    }
    if (order != FSTNone && order != _arc_out_order)
        _sort_arcs_out(order);
    const fst::vector<FSTArc*>& arcs = _states[p]->arcs_out;
    return FSTArcIter(arcs);
}

FSTArcIter
FSTCache2Rep::arcs_out_with_input(int p, int i, FSTArcMode m) const
{
    FST_ASSERT_VALID_STATE(p);
    if (_arc_out_order != FSTInput)
        _sort_arcs_out(FSTInput);
    if (_states[p]->expanded) {
        _n_hits++;
    } else {
        _n_misses++;
        _expand(p);
    }
    const fst::vector<FSTArc*>& arcs = _states[p]->arcs_out;
    FSTArcCompareInput compare;
    FSTArc a;
    a.input = i;
    typedef fst::vector<FSTArc*>::const_iterator Iter;
    Iter begin = arcs.end(), end = arcs.end();
    pair<Iter, Iter> range;
    switch (m) {
    case FSTEqual:
        range = equal_range(arcs.begin(), arcs.end(), &a, compare);
        begin = range.first;
        end = range.second;
        break;
    case FSTLessEqual:
        begin = arcs.begin();
        end = upper_bound(arcs.begin(), arcs.end(), &a, compare);
        break;
    case FSTGreaterEqual:
        begin = lower_bound(arcs.begin(), arcs.end(), &a, compare);
        end = arcs.end();
        break;
    }
    return FSTArcIter(arcs, begin, end);
}

FSTArcIter
FSTCache2Rep::arcs_out_with_output(int p, int o, FSTArcMode m) const
{
    FST_ASSERT_VALID_STATE(p);
    if (_arc_out_order != FSTOutput)
        _sort_arcs_out(FSTOutput);
    if (_states[p]->expanded) {
        _n_hits++;
    } else {
        _n_misses++;
        _expand(p);
    }
    const fst::vector<FSTArc*>& arcs = _states[p]->arcs_out;
    FSTArcCompareOutput compare;
    FSTArc a;
    a.output = o;
    typedef fst::vector<FSTArc*>::const_iterator Iter;
    Iter begin = arcs.end(), end = arcs.end();
    pair<Iter, Iter> range;
    switch (m) {
    case FSTEqual:
        range = equal_range(arcs.begin(), arcs.end(), &a, compare);
        begin = range.first;
        end = range.second;
        break;
    case FSTLessEqual:
        begin = arcs.begin();
        end = upper_bound(arcs.begin(), arcs.end(), &a, compare);
        break;
    case FSTGreaterEqual:
        begin = lower_bound(arcs.begin(), arcs.end(), &a, compare);
        end = arcs.end();
        break;
    }
    return FSTArcIter(arcs, begin, end);
}

FSTArcIter
FSTCache2Rep::arcs_in(int p, FSTArcOrder order) const
{
    FST_ASSERT_VALID_STATE(p);
    if (order != FSTNone && order != _arc_in_order)
        _sort_arcs_in(order);
    const fst::vector<FSTArc*>& arcs = _states[p]->arcs_in;
    return FSTArcIter(arcs);
}

FSTArcIter
FSTCache2Rep::arcs_in_with_input(int p, int i, FSTArcMode m) const
{
    FST_ASSERT_VALID_STATE(p);
    if (_arc_in_order != FSTInput)
        _sort_arcs_in(FSTInput);
    const fst::vector<FSTArc*>& arcs = _states[p]->arcs_in;
    FSTArcCompareInput compare;
    FSTArc a;
    a.input = i;
    typedef fst::vector<FSTArc*>::const_iterator Iter;
    Iter begin = arcs.end(), end = arcs.end();
    pair<Iter, Iter> range;
    switch (m) {
    case FSTEqual:
        range = equal_range(arcs.begin(), arcs.end(), &a, compare);
        begin = range.first;
        end = range.second;
        break;
    case FSTLessEqual:
        begin = arcs.begin();
        end = upper_bound(arcs.begin(), arcs.end(), &a, compare);
        break;
    case FSTGreaterEqual:
        begin = lower_bound(arcs.begin(), arcs.end(), &a, compare);
        end = arcs.end();
        break;
    }
    return FSTArcIter(arcs, begin, end);
}

FSTArcIter
FSTCache2Rep::arcs_in_with_output(int p, int o, FSTArcMode m) const
{
    FST_ASSERT_VALID_STATE(p);
    if (_arc_in_order != FSTOutput)
        _sort_arcs_in(FSTOutput);
    const fst::vector<FSTArc*>& arcs = _states[p]->arcs_in;
    FSTArcCompareOutput compare;
    FSTArc a;
    a.output = o;
    typedef fst::vector<FSTArc*>::const_iterator Iter;
    Iter begin = arcs.end(), end = arcs.end();
    pair<Iter, Iter> range;
    switch (m) {
    case FSTEqual:
        range = equal_range(arcs.begin(), arcs.end(), &a, compare);
        begin = range.first;
        end = range.second;
        break;
    case FSTLessEqual:
        begin = arcs.begin();
        end = upper_bound(arcs.begin(), arcs.end(), &a, compare);
        break;
    case FSTGreaterEqual:
        begin = lower_bound(arcs.begin(), arcs.end(), &a, compare);
        end = arcs.end();
        break;
    }
    return FSTArcIter(arcs, begin, end);
}

FSTFinalStateIter
FSTCache2Rep::final_states() const
{
    return FSTFinalStateIter(new _FinalStateIterRep(_final_states));
}

void
FSTCache2Rep::_expand(int p) const
{
    assert(!_states[p]->expanded);
    _states[p]->expanded = true;
    fst::vector<FSTArc*>& arcs_out = _states[p]->arcs_out;
    _in->arcs_out(arcs_out, p, _arc_out_order);
    _init_states(_in->n_states());
    _n_arcs += arcs_out.size();

    // Insert arcs into appropriate arcs_in in appropriate order.
    //
    for (fst::vector<FSTArc*>::iterator i = arcs_out.begin();
         i != arcs_out.end(); i++) {
        FSTArc* a = *i;
        fst::vector<FSTArc*>& arcs_in = _states[a->next]->arcs_in;
        switch (_arc_in_order) {
        case FSTInput:
            arcs_in.insert(lower_bound(arcs_in.begin(), arcs_in.end(), a,
                                       FSTArcCompareInput()), a);
            break;
        case FSTOutput:
            arcs_in.insert(lower_bound(arcs_in.begin(), arcs_in.end(), a,
                                       FSTArcCompareOutput()), a);
            break;
        case FSTNone:
            arcs_in.push_back(a);
            break;
        }
    }
}

void
FSTCache2Rep::_init_states(int n) const
{
    int n_new_states = n - _n_states;
    if (n_new_states > 0) {
        _states.resize(n);
        for (int i = _n_states; i < n; i++) {
            _states[i] = new _State;
            if (_in->is_final(i))
                _final_states.insert(i);
        }
        _n_states += n_new_states;
    }
}

void
FSTCache2Rep::force() const
{
    if (!_forced) {
        for (int p = 0; p < _n_states; p++)
            if (!_states[p]->expanded)
                _expand(p);
        _forced = true;
    }
}

float
FSTCache2Rep::hit_rate() const
{
    return float(_max(1, _n_hits)) / float(_max(1, _n_hits + _n_misses));
}

float
FSTCache2Rep::hit_rate(int& n_hits, int& n_total) const
{
    n_hits = _n_hits;
    n_total = n_hits + _n_misses;
    return hit_rate();
}

void
FSTCache2Rep::_sort_arcs_out(FSTArcOrder order) const
{
    if (order == _arc_out_order)
        return;
    if (order == FSTNone) {
        _arc_out_order = FSTNone;
        return;
    }
    if (_arc_out_order != FSTNone)
        FSTWarn("FSTCache2: resorting arcs from order %d to order %d is inefficient",
             int(_arc_out_order), int(order));
    _arc_out_order = order;
    FSTArcCompareInput input_compare;
    FSTArcCompareOutput output_compare;
    switch(_arc_out_order) {
    case FSTInput:
    for (_StatePtrVector::iterator i = _states.begin();
         i != _states.end(); i++) {
            if (*i) {
                fst::vector<FSTArc*>& arcs = (*i)->arcs_out;
                sort(arcs.begin(), arcs.end(), input_compare);
            }
    }
    break;
    case FSTOutput:
    for (_StatePtrVector::iterator i = _states.begin();
         i != _states.end(); i++) {
            if (*i) {
                fst::vector<FSTArc*>& arcs = (*i)->arcs_out;
                sort(arcs.begin(), arcs.end(), output_compare);
            }
    }
    break;
    case FSTNone:
    break;
    }
}

void
FSTCache2Rep::_sort_arcs_in(FSTArcOrder order) const
{
    if (order == _arc_in_order)
        return;
    if (order == FSTNone) {
        _arc_in_order = FSTNone;
        return;
    }
    if (_arc_in_order != FSTNone)
        FSTWarn("FSTCache2: resorting arcs from order %d to order %d is inefficient",
             int(_arc_in_order), int(order));
    _arc_in_order = order;
    FSTArcCompareInput input_compare;
    FSTArcCompareOutput output_compare;
    switch(_arc_in_order) {
    case FSTInput:
    for (_StatePtrVector::iterator i = _states.begin();
         i != _states.end(); i++) {
            if (*i) {
                fst::vector<FSTArc*>& arcs = (*i)->arcs_in;
                sort(arcs.begin(), arcs.end(), input_compare);
            }
    }
    break;
    case FSTOutput:
    for (_StatePtrVector::iterator i = _states.begin();
         i != _states.end(); i++) {
            if (*i) {
                fst::vector<FSTArc*>& arcs = (*i)->arcs_in;
                sort(arcs.begin(), arcs.end(), output_compare);
            }
    }
    break;
    case FSTNone:
    break;
    }
}

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 80
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
