////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

#include "fst/FST_i.h"

///////////////////////////////////////////////////////////////////////////////
//
// Implementation for Kleene-+ closure (1 or more copies).
//
// All final states have an eps:eps/final_weight arc added to the original
// initial state.

class FSTClosurePlusRep
    : public FSTGenericInt {

public: // methods

    FSTClosurePlusRep(const FSTGeneric& x, bool plus=false);
    virtual ~FSTClosurePlusRep();
    virtual const fst::set<int>& input_alphabet() const;
    virtual const fst::set<int>& output_alphabet() const;
    virtual bool is_final(int p) const;
    virtual float final_weight(int p) const;
    virtual float potential(int p) const;
    virtual FSTArcIter arcs_out(int p, FSTArcOrder order=FSTNone) const;
    virtual FSTArcIter arcs_out_with_input(int p, int i,
                                           FSTArcMode m=FSTEqual) const;
    virtual FSTArcIter arcs_out_with_output(int p, int o,
                                            FSTArcMode m=FSTEqual) const;
    virtual void force() const;
    virtual void reset() const;

    virtual FSTGeneric reload(bool& flag);

protected: // types

protected: // members
    FSTGeneric _in;
    fst::set<int> _alphabet_in, _alphabet_out;

protected: // methods
    inline void _add_arcs(fst::vector<FSTArc*>& arcs, FSTArcIter a, int p) const;
};

FSTClosurePlusRep::FSTClosurePlusRep(const FSTGeneric& x, bool plus)
    : _in(x)
{
    _class_name = "FSTClosurePlus";
    _initial = _in->initial();
    _n_states = _in->n_states();
    _semiring = _in->semiring();

    _alphabet_in = _in->input_alphabet();
    _alphabet_in.insert(0);
    _alphabet_out = _in->output_alphabet();
    _alphabet_out.insert(0);
}

FSTClosurePlusRep::~FSTClosurePlusRep()
{
}

FSTGeneric
FSTClosurePlusRep::reload(bool& flag)
{
    if (_in.reload()) {
        flag = true;
        _alphabet_in = _in->input_alphabet();
        _alphabet_in.insert(0);
        _alphabet_out = _in->output_alphabet();
        _alphabet_out.insert(0);
    }
    reset();
    return FSTGeneric(this);
}

const fst::set<int>&
FSTClosurePlusRep::input_alphabet() const
{
    return _alphabet_in;
}

const fst::set<int>&
FSTClosurePlusRep::output_alphabet() const
{
    return _alphabet_out;
}

bool
FSTClosurePlusRep::is_final(int p) const
{
    FST_ASSERT_VALID_STATE(p);
    return _in->is_final(p);
}

float
FSTClosurePlusRep::final_weight(int p) const
{
    FST_ASSERT_VALID_STATE(p);
    return _in->final_weight(p);
}

float
FSTClosurePlusRep::potential(int p) const
{
    FST_ASSERT_VALID_STATE(p);
    return _in->potential(p);
}

inline void
FSTClosurePlusRep::_add_arcs(fst::vector<FSTArc*>& arcs, FSTArcIter a, int p) const
{
    for (; a.valid(); a++)
        arcs.push_back(new FSTArc(p, a->next,
                                  a->input, a->output, a->weight));
}

FSTArcIter
FSTClosurePlusRep::arcs_out(int p, FSTArcOrder order) const
{
    FST_ASSERT_VALID_STATE(p);
    FSTArcIter result;
    float w = _in->final_weight(p);
    if (w == _semiring->zero || p == _initial) {
        result = _in->arcs_out(p, order);
    } else {
        FSTArcIterData* arc_data = new FSTArcIterData;
        fst::vector<FSTArc*>& arcs = arc_data->arcs;
        arcs.push_back(new FSTArc(p, _initial, 0, 0, w));
        _add_arcs(arcs, _in->arcs_out(p, order), p);
        result = FSTArcIter(arc_data);
    }
    _n_states = _in->n_states();
    return result;
}

FSTArcIter FSTClosurePlusRep::arcs_out_with_input(int p, int i, FSTArcMode m) const
{
    FST_ASSERT_VALID_STATE(p);
    FSTArcIter result;
    float w = _in->final_weight(p);
    if (w == _semiring->zero || p == _initial || (i > 0 && m != FSTLessEqual)) {
        result = _in->arcs_out_with_input(p, i, m);
    } else {
        FSTArcIterData* arc_data = new FSTArcIterData;
        fst::vector<FSTArc*>& arcs = arc_data->arcs;
        arcs.push_back(new FSTArc(p, _initial, 0, 0, w));
        _add_arcs(arcs, _in->arcs_out_with_input(p, i, m), p);
        result = FSTArcIter(arc_data);
    }
    _n_states = _in->n_states();
    return result;
}

FSTArcIter FSTClosurePlusRep::arcs_out_with_output(int p, int o, FSTArcMode m) const
{
    FST_ASSERT_VALID_STATE(p);
    FSTArcIter result;
    float w = _in->final_weight(p);
    if (w == _semiring->zero || p == _initial || (o > 0 && m != FSTLessEqual)) {
        result = _in->arcs_out_with_output(p, o, m);
    } else {
        FSTArcIterData* arc_data = new FSTArcIterData;
        fst::vector<FSTArc*>& arcs = arc_data->arcs;
        arcs.push_back(new FSTArc(p, _initial, 0, 0, w));
        _add_arcs(arcs, _in->arcs_out_with_output(p, o, m), p);
        result = FSTArcIter(arc_data);
    }
    _n_states = _in->n_states();
    return result;
}

void FSTClosurePlusRep::force() const
{
    _in->force();
    _n_states = _in->n_states();
}

void FSTClosurePlusRep::reset() const
{
    _in->reset();
    _n_states = _in->n_states();
}

///////////////////////////////////////////////////////////////////////////////
//
// Implementation for Kleene-* closure (0 or more copies).
//
// Basically, we create a new initial state (0), and connect it to the original
// initial state with eps:eps/1.  All state numbers in the original FST are
// increased by one.  All final states have an eps:eps/final_weight arc added to
// the original initial state.

class FSTClosureStarRep
    : public FSTGenericInt {

public: // methods

    FSTClosureStarRep(const FSTGeneric& x, bool plus=false);
    virtual ~FSTClosureStarRep();
    virtual const fst::set<int>& input_alphabet() const;
    virtual const fst::set<int>& output_alphabet() const;
    virtual bool is_final(int p) const;
    virtual float final_weight(int p) const;
    virtual float potential(int p) const;
    virtual FSTArcIter arcs_out(int p, FSTArcOrder order=FSTNone) const;
    virtual FSTArcIter arcs_out_with_input(int p, int i,
                                           FSTArcMode m=FSTEqual) const;
    virtual FSTArcIter arcs_out_with_output(int p, int o,
                                            FSTArcMode m=FSTEqual) const;
    virtual void force() const;
    virtual void reset() const;

    virtual FSTGeneric reload(bool& flag);

protected: // types

protected: // members
    FSTGeneric _in;
    fst::set<int> _alphabet_in, _alphabet_out;

protected: // methods
    inline void _add_arcs(fst::vector<FSTArc*>& arcs, FSTArcIter a, int p) const;
};

FSTClosureStarRep::FSTClosureStarRep(const FSTGeneric& x, bool plus)
    : _in(x)
{
    _class_name = "FSTClosureStar";
    _initial = 0;
    _n_states = _in->n_states() + 1;
    _semiring = _in->semiring();

    _alphabet_in = _in->input_alphabet();
    _alphabet_in.insert(0);
    _alphabet_out = _in->output_alphabet();
    _alphabet_out.insert(0);
}

FSTClosureStarRep::~FSTClosureStarRep()
{
}

FSTGeneric
FSTClosureStarRep::reload(bool& flag)
{
    if (_in.reload()) {
        flag = true;
        _alphabet_in = _in->input_alphabet();
        _alphabet_in.insert(0);
        _alphabet_out = _in->output_alphabet();
        _alphabet_out.insert(0);
    }
    reset();
    return FSTGeneric(this);
}

const fst::set<int>&
FSTClosureStarRep::input_alphabet() const
{
    return _alphabet_in;
}

const fst::set<int>&
FSTClosureStarRep::output_alphabet() const
{
    return _alphabet_out;
}

bool
FSTClosureStarRep::is_final(int p) const
{
    FST_ASSERT_VALID_STATE(p);
    return p == _initial || _in->is_final(p - 1);
}

float
FSTClosureStarRep::final_weight(int p) const
{
    FST_ASSERT_VALID_STATE(p);
    return (p == _initial) ? _semiring->one : _in->final_weight(p - 1);
}

float
FSTClosureStarRep::potential(int p) const
{
    FST_ASSERT_VALID_STATE(p);
    return (p == _initial) ? _semiring->one : _in->potential(p - 1);
}

inline void
FSTClosureStarRep::_add_arcs(fst::vector<FSTArc*>& arcs, FSTArcIter a, int p) const
{
    for (; a.valid(); a++)
        arcs.push_back(new FSTArc(p, a->next + 1,
                                  a->input, a->output, a->weight));
}

FSTArcIter
FSTClosureStarRep::arcs_out(int p, FSTArcOrder order) const
{
    FST_ASSERT_VALID_STATE(p);

    float w;
    FSTArcIterData* arc_data = new FSTArcIterData;
    fst::vector<FSTArc*>& arcs = arc_data->arcs;

    if (p == _initial) {
        arcs.push_back(new FSTArc(p, _in->initial() + 1, 0, 0, _semiring->one));
    } else {
        if ((w = _in->final_weight(p - 1)) != _semiring->zero) {
            arcs.push_back(new FSTArc(p, _in->initial() + 1, 0, 0, w));
        }
        _add_arcs(arcs, _in->arcs_out(p - 1, order), p);
        _n_states = _in->n_states() + 1;
    }
    return FSTArcIter(arc_data);
}

FSTArcIter FSTClosureStarRep::arcs_out_with_input(int p, int i, FSTArcMode m) const
{
    FST_ASSERT_VALID_STATE(p);

    float w;
    FSTArcIterData* arc_data = new FSTArcIterData;
    fst::vector<FSTArc*>& arcs = arc_data->arcs;

    if (p == _initial) {
        if (i == 0 || m == FSTLessEqual) {
            arcs.push_back(new FSTArc(p, _in->initial() + 1, 0, 0, _semiring->one));
        }
    } else {
        if ((i == 0 || m == FSTLessEqual) && (w = _in->final_weight(p - 1)) != _semiring->zero) {
            arcs.push_back(new FSTArc(p, _in->initial() + 1, 0, 0, w));
        }
        _add_arcs(arcs, _in->arcs_out_with_input(p - 1, i, m), p);
        _n_states = _in->n_states() + 1;
    }
    return FSTArcIter(arc_data);
}

FSTArcIter FSTClosureStarRep::arcs_out_with_output(int p, int o, FSTArcMode m) const
{
    FST_ASSERT_VALID_STATE(p);

    float w;
    FSTArcIterData* arc_data = new FSTArcIterData;
    fst::vector<FSTArc*>& arcs = arc_data->arcs;

    if (p == _initial) {
        if (o == 0 || m == FSTLessEqual) {
            arcs.push_back(new FSTArc(p, _in->initial() + 1, 0, 0, _semiring->one));
        }
    } else {
        if ((o == 0 || m == FSTLessEqual) && (w = _in->final_weight(p - 1)) != _semiring->zero) {
            arcs.push_back(new FSTArc(p, _in->initial() + 1, 0, 0, w));
        }
        _add_arcs(arcs, _in->arcs_out_with_output(p - 1, o, m), p);
        _n_states = _in->n_states() + 1;
    }
    return FSTArcIter(arc_data);
}

void FSTClosureStarRep::force() const
{
    _in->force();
    _n_states = _in->n_states() + 1;
}

void FSTClosureStarRep::reset() const
{
    _in->reset();
    _n_states = _in->n_states() + 1;
}

///////////////////////////////////////////////////////////////////////////////

FSTGeneric
FSTClosure(const FSTGeneric& x, bool plus)
{
    // Dispatch to Kleene-+ or Kleene-* implementation.  If we're doing Kleene-*
    // closure and the final weight of the initial state is 1, then it is
    // equivalent to Kleene-+ closure.
    return FSTGeneric((plus || x->final_weight(x->initial()) == x->semiring()->one)
                      ? static_cast<FSTGenericInt*>(new FSTClosurePlusRep(x))
                      : static_cast<FSTGenericInt*>(new FSTClosureStarRep(x)));
}

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 80
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
