////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

#include "fst/FST_i.h"

FSTComposeInt::~FSTComposeInt()
{
}

class FSTComposeRep
    : public FSTComposeInt {

public: // methods

    FSTComposeRep(const FSTGeneric& x, const FSTGeneric& y);
    virtual ~FSTComposeRep();
    virtual const fst::set<int>& input_alphabet() const;
    virtual const fst::set<int>& output_alphabet() const;
    virtual bool is_final(int p) const;
    virtual float final_weight(int p) const;
    virtual float potential(int p) const;
    virtual FSTArcIter arcs_out(int p, FSTArcOrder order=FSTNone) const;
    virtual FSTArcIter arcs_out_with_input(int p, int i,
                                           FSTArcMode m=FSTEqual) const;
    virtual FSTArcIter arcs_out_with_output(int p, int o,
                                            FSTArcMode m=FSTEqual) const;
    virtual void arcs_out(fst::vector<FSTArc*>& arcs,
                          int p, FSTArcOrder order=FSTNone) const;
    virtual void force() const;
    virtual void reset() const;

    virtual FSTGeneric reload(bool& flag);

public: // new methods

    virtual pair<int,int> state_pair(int state) const;

protected: // types

    struct _Triple {
        unsigned _rp;
        unsigned _q;
        
        _Triple() {}
        _Triple(unsigned p, unsigned q, unsigned r)
            : _rp(p | (r << 30)), _q(q) {}
        unsigned p() const {
            return _rp & 0x3fffffff;
        }
        unsigned q() const {
            return _q;
        }
        unsigned r() const {
            return _rp >> 30;
        }
        bool operator == (const _Triple& x) const {
            return _rp == x._rp && _q == x._q;
        }
        bool operator < (const _Triple& x) const {
            return FST_LESS3(_rp, x._rp,
                             FST_LESS2(_q, x._q));
        }

#ifdef NO_HASH_MAP
        struct Less {
            bool operator()(const _Triple& x, const _Triple& y) const {
                return FST_LESS3(x._rp, y._rp,
                                 FST_LESS2(x._q, y._q));
            }
        };
#else
        struct Hash {
            size_t operator()(const _Triple& x) const {
                return x._rp * 257 + x._q * 17;
            }
        };
        struct Equal {
            bool operator()(const _Triple& x, const _Triple& y) const {
                return x._rp == y._rp && x._q == y._q;
            }
        };
#endif

    };

#ifdef NO_HASH_MAP
    typedef fst::map<_Triple, int, _Triple::Less> _TripleIntMap;
#else
    typedef fst::hash_map<_Triple, int, _Triple::Hash, _Triple::Equal> _TripleIntMap;
#endif
    typedef fst::vector<const _Triple*> _TripleVector;

protected: // members

    FSTGeneric _x;
    FSTGeneric _y;
    mutable _TripleVector _triples;
    mutable _TripleIntMap _map;
    bool _epsilon_filter;
    fst::set<int> _input_alphabet;
    fst::set<int> _output_alphabet;

protected: // methods

    void _arcs_out(int p, int i, int o,
                   FSTArcMode m, FSTArcOrder order, fst::vector<FSTArc*>* arcs) const;
    inline int _find_or_add_state(const _Triple& t) const;
    inline void _add_arc(fst::vector<FSTArc*>* arcs, int from,
                         int p, int q, int r, int i, int o, float w) const;
    int _init() const;
    void _clean() const;
};

const FSTComposeRep* FSTComposeRepCast(const FSTGeneric& x);
FSTComposeRep* FSTComposeRepCast(FSTGeneric& x);

FSTComposeRep::FSTComposeRep(const FSTGeneric& x, const FSTGeneric& y)
    : _x(x), _y(y)
{
    _class_name = "FSTCompose";

    if (x->semiring() != y->semiring())
        FSTFatal("can't compose different semirings: %s and %s",
              x->semiring()->name, y->semiring()->name);
    _semiring = x->semiring();

    _input_alphabet = _x->input_alphabet();
    _input_alphabet.insert(0);
    _output_alphabet = _y->output_alphabet();
    _output_alphabet.insert(0);
    _initial = _init();

    // Filter epsilons iff x has output epsilons and y has input epsilons.
    //
    _epsilon_filter = 
        _x->output_alphabet().count(0) != 0 && _y->input_alphabet().count(0) != 0;
}

FSTComposeRep::~FSTComposeRep()
{
    _clean();
}

FSTGeneric
FSTComposeRep::reload(bool& flag)
{
    if (_x.reload()) flag = true;
    if (_y.reload()) flag = true;
    reset();
    return FSTGeneric(this);
}

void
FSTComposeRep::reset() const
{
    _x->reset();
    _y->reset();
    _clean();

    int new_initial = _init();
    assert(new_initial == _initial);
}

int
FSTComposeRep::_init() const
{
    int result = -1;
    if (_x->initial() >= 0 && _y->initial() >= 0)
        result = _find_or_add_state(_Triple(_x->initial(),
                                            _y->initial(), 0));
    return result;
}

void
FSTComposeRep::_clean() const
{
    // Swapping with an empty container is a good way to clear out any internal
    // allocation.

#ifdef _MSC_VER
    _triples = _TripleVector();
#else
    _TripleVector empty_triples;
    _triples.clear();
    _triples.swap(empty_triples);
#endif

    _TripleIntMap empty_map;
    _map.clear();
    _map.swap(empty_map);

    _n_states = 0;
    _n_arcs = 0;
}

const fst::set<int>&
FSTComposeRep::input_alphabet() const
{
    return _input_alphabet;
}

const fst::set<int>&
FSTComposeRep::output_alphabet() const
{
    return _output_alphabet;
}

bool
FSTComposeRep::is_final(int p) const
{
    FST_ASSERT_VALID_STATE(p);
    const _Triple* triple = _triples[p];
    return _x->is_final(triple->p()) && _y->is_final(triple->q());
}

float
FSTComposeRep::final_weight(int p) const
{
    FST_ASSERT_VALID_STATE(p);
    const _Triple* triple = _triples[p];
    float pw = _x->final_weight(triple->p());
    if (pw == _semiring->zero)
        return pw;
    float qw = _y->final_weight(triple->q());
    if (qw == _semiring->zero)
        return qw;
    return _semiring->times(pw, qw);
}

float
FSTComposeRep::potential(int p) const
{
    FST_ASSERT_VALID_STATE(p);
    const _Triple* t = _triples[p];
    return _semiring->times(_x->potential(t->p()), _y->potential(t->q()));
}

inline void
FSTComposeRep::_add_arc(fst::vector<FSTArc*>* arcs, int from,
                        int p, int q, int r, int i, int o, float w) const
{
    if (!_epsilon_filter) r = 0;
    int state = _find_or_add_state(_Triple(p, q, r));
    if (arcs)
        arcs->push_back(new FSTArc(from, state, i, o, w));
}

FSTArcIter
FSTComposeRep::arcs_out(int p, FSTArcOrder order) const
{
    FST_ASSERT_VALID_STATE(p);
    FSTArcIterData* arc_data = new FSTArcIterData;
    _arcs_out(p, -1, -1, FSTEqual, order, &arc_data->arcs);
    return FSTArcIter(arc_data);
}

FSTArcIter
FSTComposeRep::arcs_out_with_input(int p, int i, FSTArcMode m) const
{
    FST_ASSERT_VALID_STATE(p);
    FSTArcIterData* arc_data = new FSTArcIterData;
    _arcs_out(p, i, -1, m, FSTInput, &arc_data->arcs);
    return FSTArcIter(arc_data);
}

FSTArcIter
FSTComposeRep::arcs_out_with_output(int p, int o, FSTArcMode m) const
{
    FST_ASSERT_VALID_STATE(p);
    FSTArcIterData* arc_data = new FSTArcIterData;
    _arcs_out(p, -1, o, m, FSTOutput, &arc_data->arcs);
    return FSTArcIter(arc_data);
}

void
FSTComposeRep::arcs_out(fst::vector<FSTArc*>& result, int p, FSTArcOrder order) const
{
    FST_ASSERT_VALID_STATE(p);
    _arcs_out(p, -1, -1, FSTEqual, order, &result);
}

void
FSTComposeRep::_arcs_out(int state, int input, int output, FSTArcMode mode,
                         FSTArcOrder order, fst::vector<FSTArc*>* arcs) const
{
    const _Triple* triple = _triples[state];
    int p = triple->p();
    int q = triple->q();
    int r = triple->r();

    FSTArcIter pi, pt, qi, qt;
    int cached_value = -1;

    FSTArcOrder orig_order = order;

    // If we have not specified a desired order (very typical), we select the
    // order that allows us to walk down the smallest side and lookup in the
    // largest side.
    //
    if (order == FSTNone) {
        if (_x->n_arcs_out(p) <= _y->n_arcs_out(q))
            order = FSTInput;
        else
            order = FSTOutput;
    }

    switch (order) {

    case FSTNone: /////////////////////////////////////////////////////////////

        FSTFatal("internal error");
        break;

    case FSTInput: ////////////////////////////////////////////////////////////

        // X holding and Y making epsilon-input transitions.
        //
        if (r != 2 && (input <= 0 || mode == FSTLessEqual)) {
            for (qi = _y->arcs_out_with_input(q, 0, FSTEqual);
                 qi.valid(); qi++) {
                _add_arc(arcs, state, p, qi->next, 1,
                         0, qi->output, qi->weight);
            }
        }
        
        // Iterate through X's arcs (ordered by input if required).
        //
        if (input < 0)
            pi = _x->arcs_out(p, orig_order);
        else
            pi = _x->arcs_out_with_input(p, input, mode);
        for (; pi.valid(); pi++) {

            // X making epsilon-output transition and Y holding.
            //
            if (r != 1 && pi->output == 0) {
                _add_arc(arcs, state, pi->next, q, 2,
                         pi->input, 0, pi->weight);
            }

            // X and Y making transitions.
            //
            if (r == 0 || pi->output != 0) {
                if (pi->output != cached_value) {
                    cached_value = pi->output;
                    qt = _y->arcs_out_with_input(q, cached_value, FSTEqual);
                }
                for (qi = qt; qi.valid(); qi++) {
                    _add_arc(arcs, state, pi->next, qi->next, 0,
                             pi->input, qi->output,
                             _semiring->times(pi->weight, qi->weight));
                }
            }
        }

        break;

    case FSTOutput: ///////////////////////////////////////////////////////////

        // Y holding and X making epsilon-output transitions.
        //
        if (r != 1 && (output <= 0 || mode == FSTLessEqual)) {
            for (pi = _x->arcs_out_with_output(p, 0, FSTEqual);
                 pi.valid(); pi++) {
                _add_arc(arcs, state, pi->next, q, 2,
                         pi->input, 0, pi->weight);
            }
        }
        
        // Iterate through Y's arcs (ordered by output if required).
        //
        if (output < 0)
            qi = _y->arcs_out(q, orig_order);
        else
            qi = _y->arcs_out_with_output(q, output, mode);
        for (; qi.valid(); qi++) {

            // Y making epsilon-input transition and X holding.
            //
            if (r != 2 && qi->input == 0) {
                _add_arc(arcs, state, p, qi->next, 1,
                         0, qi->output, qi->weight);
            }

            // X and Y making transitions.
            //
            if (r == 0 || qi->input != 0) {
                if (qi->input != cached_value) {
                    cached_value = qi->input;
                    pt = _x->arcs_out_with_output(p, cached_value, FSTEqual);
                }
                for (pi = pt; pi.valid(); pi++) {
                    _add_arc(arcs, state, pi->next, qi->next, 0,
                             pi->input, qi->output,
                             _semiring->times(pi->weight, qi->weight));
                }
            }
        }

        break;

    }
}

void
FSTComposeRep::force() const
{
    // Expand all states.  We do this by asking for all arcs out of all states
    // (without actually creating all the arcs).
    //
    for (int p = 0; p < _n_states; p++)
        _arcs_out(p, -1, -1, FSTEqual, FSTNone, 0);
}

inline int
FSTComposeRep::_find_or_add_state(const _Triple& triple) const
{
    int p;
    _TripleIntMap::iterator i = _map.find(triple);

    if (i == _map.end()) {

        // State not found: build new state.
        //
        p = _n_states++;
        i = _map.insert(_TripleIntMap::value_type(triple, p)).first;
        _triples.push_back(&(*i).first);

    } else {

        // State found: look up state index and free triple (it isn't needed).
        //
        p = (*i).second;

    }

    return p;
}

pair<int,int>
FSTComposeRep::state_pair(int state) const
{
    FST_ASSERT_VALID_STATE(state);
    const _Triple* triple = _triples[state];
    return pair<int,int>(triple->p(), triple->q());
}

FSTGeneric
FSTCompose(const FSTGeneric& x, const FSTGeneric& y)
{
    return FSTGeneric(new FSTComposeRep(x, y));
}

const FSTComposeInt*
FSTComposeCast(const FSTGeneric& x)
{
    return dynamic_cast<const FSTComposeInt*>(x.pointer());
}

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 80
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
