////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

#include "fst/FST_i.h"

class FSTConcatenateRep
    : public FSTGenericInt {

public: // methods
    FSTConcatenateRep(const fst::vector<FSTGeneric>& inputs);
    virtual ~FSTConcatenateRep();
    virtual const fst::set<int>& input_alphabet() const;
    virtual const fst::set<int>& output_alphabet() const;
    virtual bool is_final(int p) const;
    virtual float final_weight(int p) const;
    virtual float potential(int p) const;
    virtual FSTArcIter arcs_out(int p, FSTArcOrder order=FSTNone) const;
    virtual FSTArcIter arcs_out_with_input(int p, int i, FSTArcMode m=FSTEqual) const;
    virtual FSTArcIter arcs_out_with_output(int p, int o, FSTArcMode m=FSTEqual) const;
    virtual void force() const;
    virtual void reset() const;
    virtual FSTGeneric reload(bool& flag);

protected: // types
    struct _State;
    typedef fst::vector<_State*> _StatePtrVector;
    typedef fst::map<_State*, int, ptr_less<_State*> > _StateIntMap;

protected: // members
    fst::vector<FSTGeneric> _inputs;
    int _n_inputs;
    mutable _StatePtrVector _states;
    mutable _StateIntMap _map;
    fst::set<int> _input_alphabet;
    fst::set<int> _output_alphabet;

protected: // methods
    inline int _find_or_add_state(_State* s) const;
    inline void _add_epsilon(fst::vector<FSTArc*>& arcs, int p,
                             int input, int state, float weight) const;
    inline void _add_arcs(fst::vector<FSTArc*>& arcs, int p,
                          int input, FSTArcIter a) const;
    void _setup();
    int  _init() const;
    void _clean() const;
};

FSTGeneric
FSTConcatenate(const FSTGeneric& x, const FSTGeneric& y)
{
    fst::vector<FSTGeneric> v;
    v.push_back(x);
    v.push_back(y);
    return FSTGeneric(new FSTConcatenateRep(v));
}

FSTGeneric
FSTConcatenate(const FSTGeneric& x, const FSTGeneric& y, const FSTGeneric& z)
{
    fst::vector<FSTGeneric> v;
    v.push_back(x);
    v.push_back(y);
    v.push_back(z);
    return FSTGeneric(new FSTConcatenateRep(v));
}

FSTGeneric
FSTConcatenate(const fst::vector<FSTGeneric>& inputs)
{
    return FSTGeneric(new FSTConcatenateRep(inputs));
}

struct FSTConcatenateRep::_State {
    int input;
    int state;
    _State(int i, int s) : input(i), state(s) {}
    bool operator < (const _State& x) {
        return FST_LESS3(input, x.input,
                         FST_LESS2(state, x.state));
    }
    FSTAllocator_new_delete(_State);
};

FSTConcatenateRep::FSTConcatenateRep(const fst::vector<FSTGeneric>& inputs)
    : _inputs(inputs), _n_inputs(_inputs.size())
{
    _class_name = "FSTConcatenate";

    _setup();
    _initial = _init();
}

void
FSTConcatenateRep::_setup()
{
    // Make sure input semirings are identical and set up alphabets.
    _input_alphabet.clear();
    _output_alphabet.clear();
    fst::vector<FSTGeneric>::const_iterator i = _inputs.begin();
    if (i != _inputs.end()) {
        _semiring = (*i)->semiring();
        for (; i != _inputs.end(); i++) {
            if (_semiring != (*i)->semiring())
                FSTFatal("can't concatenate different semirings: %s and %s",
                      _semiring->name, (*i)->semiring()->name);
            const fst::set<int>& inputs = (*i)->input_alphabet();
            const fst::set<int>& outputs = (*i)->output_alphabet();
            fst::set_insert(_input_alphabet, inputs.begin(), inputs.end());
            fst::set_insert(_output_alphabet, outputs.begin(), outputs.end());
        }
    }
    _input_alphabet.insert(0);
    _output_alphabet.insert(0);
}

FSTConcatenateRep::~FSTConcatenateRep()
{
    _clean();
}

FSTGeneric
FSTConcatenateRep::reload(bool& flag)
{
    for (fst::vector<FSTGeneric>::iterator i = _inputs.begin();
         i != _inputs.end(); i++) {
        if ((*i).reload()) flag = true;
    }
    if (flag)
        _setup();
    reset();
    return FSTGeneric(this);
}

void
FSTConcatenateRep::reset() const
{
    // Reset all input FSTs.
    //
    for (fst::vector<FSTGeneric>::const_iterator i = _inputs.begin();
         i != _inputs.end(); i++)
        (*i)->reset();

    _clean();

    int new_initial = _init();
    assert(new_initial = _initial);
}

int
FSTConcatenateRep::_init() const
{
    int result = -1;
    if (_n_inputs > 0) {
        int i = _inputs[0]->initial();
        if (i >= 0)
            result = _find_or_add_state(new _State(0, i));
    }
    return result;
}

void
FSTConcatenateRep::_clean() const
{
    for (_StatePtrVector::iterator i = _states.begin();
     i != _states.end(); i++) {
        delete *i;
    }
    _states.clear();
    _map.clear();
    _n_states = 0;
    _n_arcs = 0;
}



const fst::set<int>&
FSTConcatenateRep::input_alphabet() const
{
    return _input_alphabet;
}

const fst::set<int>&
FSTConcatenateRep::output_alphabet() const
{
    return _output_alphabet;
}

bool
FSTConcatenateRep::is_final(int p) const
{
    FST_ASSERT_VALID_STATE(p);
    _State* pp = _states[p];
    return pp->input == _n_inputs-1 && _inputs[pp->input]->is_final(pp->state);
}

float
FSTConcatenateRep::final_weight(int p) const
{
    FST_ASSERT_VALID_STATE(p);
    _State* pp = _states[p];
    if (pp->input == _n_inputs-1)
        return _inputs[pp->input]->final_weight(pp->state);
    else
        return _semiring->zero;
}

float
FSTConcatenateRep::potential(int p) const
{
    FST_ASSERT_VALID_STATE(p);
    _State* pp = _states[p];
    return _inputs[pp->input]->potential(pp->state);
}

inline void
FSTConcatenateRep::_add_epsilon(fst::vector<FSTArc*>& arcs, int p,
                                int input, int state, float weight) const
{
    int q = _find_or_add_state(new _State(input, state));
    arcs.push_back(new FSTArc(p, q, 0, 0, weight));
}

inline void
FSTConcatenateRep::_add_arcs(fst::vector<FSTArc*>& arcs, int p,
                             int input, FSTArcIter a) const
{
    for (; a.valid(); a++) {
        int q = _find_or_add_state(new _State(input, a->next));
        arcs.push_back(new FSTArc(p, q, a->input, a->output, a->weight));
    }
}

FSTArcIter
FSTConcatenateRep::arcs_out(int p, FSTArcOrder order) const
{
    FST_ASSERT_VALID_STATE(p);
    _State* pp = _states[p];
    int input = pp->input;
    int state = pp->state;
    const FSTGeneric& fst = _inputs[input];
    FSTArcIter a = fst->arcs_out(state, order);
    FSTArcIterData* arc_data = new FSTArcIterData;
    fst::vector<FSTArc*>& arcs = arc_data->arcs;

    if (input < _n_inputs-1 && fst->is_final(state)) {
        arcs.reserve(a.n() + 1);
        _add_epsilon(arcs, p, input + 1, _inputs[input + 1]->initial(),
                     fst->final_weight(state));
    } else {
        arcs.reserve(a.n());
    }
    _add_arcs(arcs, p, input, a);
    return FSTArcIter(arc_data);
}

FSTArcIter
FSTConcatenateRep::arcs_out_with_input(int p, int i, FSTArcMode m) const
{
    FST_ASSERT_VALID_STATE(p);
    _State* pp = _states[p];
    int input = pp->input;
    int state = pp->state;
    const FSTGeneric& fst = _inputs[input];
    FSTArcIter a = fst->arcs_out_with_input(state, i, m);
    FSTArcIterData* arc_data = new FSTArcIterData;
    fst::vector<FSTArc*>& arcs = arc_data->arcs;

    if ((i == 0 || m == FSTLessEqual) &&
        input < _n_inputs-1 && fst->is_final(state)) {
        arcs.reserve(a.n() + 1);
        _add_epsilon(arcs, p, input + 1, _inputs[input + 1]->initial(),
                     fst->final_weight(state));
    } else {
        arcs.reserve(a.n());
    }
    _add_arcs(arcs, p, input, a);
    return FSTArcIter(arc_data);
}

FSTArcIter
FSTConcatenateRep::arcs_out_with_output(int p, int o, FSTArcMode m) const
{
    FST_ASSERT_VALID_STATE(p);
    _State* pp = _states[p];
    int input = pp->input;
    int state = pp->state;
    const FSTGeneric& fst = _inputs[input];
    FSTArcIter a = fst->arcs_out_with_output(state, o, m);
    FSTArcIterData* arc_data = new FSTArcIterData;
    fst::vector<FSTArc*>& arcs = arc_data->arcs;

    if ((o == 0 || m == FSTLessEqual) &&
        input < _n_inputs-1 && fst->is_final(state)) {
        arcs.reserve(a.n() + 1);
        _add_epsilon(arcs, p, input + 1, _inputs[input + 1]->initial(),
                     fst->final_weight(p));
    } else {
        arcs.reserve(a.n());
    }
    _add_arcs(arcs, p, input, a);
    return FSTArcIter(arc_data);
}

void
FSTConcatenateRep::force() const
{
    // Create all reachable states.
    for (int p = 0; p < _n_states; p++) {
        int input = _states[p]->input;
        int state = _states[p]->state;
        for (FSTArcIter a = _inputs[input]->arcs_out(state); a.valid(); a++)
            _find_or_add_state(new _State(input, state));
        if (_inputs[input]->is_final(state) && input < _n_inputs - 1) {
            int initial = _inputs[input + 1]->initial();
            if (initial >= 0)
                _find_or_add_state(new _State(input + 1, initial));
        }
    }
}

inline int
FSTConcatenateRep::_find_or_add_state(_State* state) const
{
    int p;
    _StateIntMap::iterator i = _map.find(state);

    if (i == _map.end()) {

        // State not found: build new state.
        _states.push_back(state);
        p = _n_states++;
        _map[state] = p;

    } else {

        // State found: look up state index and free triple (it isn't needed).
        p = (*i).second;
        delete state;

    }

    return p;
}

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 80
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
