////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

//#define DEBUG
//#define DEBUG_SETS
//#define DEBUG_WEIGHT_AMBIGUITY

#include "fst/FST_i.h"
#include <climits>
#include "fst/FSTSymbolString.h"

class FSTDeterminizeRep
    : public FSTGenericInt {

public: // methods

    FSTDeterminizeRep(const FSTGeneric& in, int degree,
                      bool fsa, float eps, bool strict);
    virtual ~FSTDeterminizeRep();
    virtual const fst::set<int>& input_alphabet() const;
    virtual const fst::set<int>& output_alphabet() const;
    virtual bool is_final(int p) const;
    virtual float final_weight(int p) const;
    virtual float potential(int p) const;
    virtual FSTArcIter arcs_out(int p, FSTArcOrder order=FSTNone) const;
    virtual FSTArcIter arcs_out_with_input(int p, int i,
                                           FSTArcMode m=FSTEqual) const;
    virtual FSTArcIter arcs_out_with_output(int p, int o,
                                            FSTArcMode m=FSTEqual) const;
    virtual void force() const;
    virtual void reset() const;

    virtual FSTGeneric reload(bool& flag);

    void state_set(int p, fst::set<int>& result) const;

protected: // types

    struct _Pair {
        int state;
        int output;
        _Pair(int s, int o) : state(s), output(o) {}
        bool operator<(const _Pair& x) const {
            return FST_LESS3(state, x.state,
                             FST_LESS2(output, x.output));
        }
        bool operator==(const _Pair& x) const {
            return state == x.state && output == x.output;
        }
    };

    typedef fst::map<_Pair,int>              _PairMap;

    typedef fst::vector<const _Pair*>        _PairVector;

    struct _Triple {
        int state_output;
        float weight;
        _Triple(int so, float w) : state_output(so), weight(w) {}
    };

    struct _TripleCompare {
        float eps;
        _TripleCompare(float e) : eps(e) {}
        bool operator()(const _Triple& x, const _Triple& y) const {
            return FST_LESS3(x.state_output, y.state_output,
                             FST_LESS2(x.weight, y.weight - eps));
        }
    };


    class _TripleSet {
        
    protected: // types
        
        struct _Chunk {
            _Chunk* next;
            _Triple triples[0];
            
            static _Chunk* allocate(unsigned n) {
                unsigned n_bytes = sizeof(_Chunk) + n*sizeof(_Triple);
                void* bytes = FSTAllocator::allocate(n_bytes);
                return reinterpret_cast<_Chunk*>(bytes);
            }
            static void deallocate(_Chunk* chunk, unsigned n) {
                unsigned n_bytes = sizeof(_Chunk) + n*sizeof(_Triple);
                FSTAllocator::deallocate(chunk, n_bytes);
            }
        };
        
    public: // types
        
        class ConstIterator;
        
        class Iterator {
            friend class _TripleSet;
            
        private:
            _Chunk* chunk;
            _Triple* ptr;
            unsigned n;
            
            Iterator(_Chunk* chunk_, _Triple* ptr_, unsigned n_)
                : chunk(chunk_), ptr(ptr_), n(n_) {}
        public:
            Iterator(const Iterator& x)
                : chunk(x.chunk), ptr(x.ptr), n(x.n) {}
            bool valid() const {
                return n != 0;
            }
            _Triple& triple() const {
                return *ptr;
            }
            int& state_output() const {
                return ptr->state_output;
            }
            float& weight() const {
                return ptr->weight;
            }
            Iterator& operator++() {
                ++ptr;
                --n;
                if (ptr == &chunk->triples[_chunk_elements]) {
                    chunk = chunk->next;
                    ptr = &chunk->triples[0];
                }
                return *this;
            }
            Iterator operator++(int) {
                Iterator result(*this);
                ++(*this);
                return result;
            }
        };
        
        class ConstIterator {
            friend class _TripleSet;
            
        private:
            const _Chunk* chunk;
            const _Triple* ptr;
            unsigned n;
            
            ConstIterator(const _Chunk* chunk_, const _Triple* ptr_, unsigned n_)
                : chunk(chunk_), ptr(ptr_), n(n_) {}
        public:
            ConstIterator(const ConstIterator& x)
                : chunk(x.chunk), ptr(x.ptr), n(x.n) {}
            bool valid() const {
                return n != 0;
            }
            const _Triple& triple() const {
                return *ptr;
            }
            int state_output() const {
                return ptr->state_output;
            }
            float weight() const {
                return ptr->weight;
            }
            ConstIterator& operator++() {
                ++ptr;
                --n;
                if (ptr == &chunk->triples[_chunk_elements]) {
                    chunk = chunk->next;
                    ptr = &chunk->triples[0];
                }
                return *this;
            }
            ConstIterator operator++(int) {
                ConstIterator result(*this);
                ++(*this);
                return result;
            }
        };
        
        struct Compare {
            _TripleCompare compare;
            Compare(float eps) : compare(eps) {}
            bool operator()(const _TripleSet* x, const _TripleSet* y) const {
                if (x->size() < y->size()) return true;
                if (x->size() > y->size()) return false;
                _TripleSet::ConstIterator
                    xi(x->const_elements()), yi(y->const_elements());
                for (; xi.valid() && yi.valid(); ++xi, ++yi) {
                    if (compare(xi.triple(), yi.triple())) return true;
                    if (compare(yi.triple(), xi.triple())) return false;
                }
                return !xi.valid() && yi.valid();
            }
        };
        
    protected: // members
        
        unsigned _size;
        _Chunk* _chunk;

    public: // members

        static const unsigned _chunk_elements;
        
    public: // methods
        
        _TripleSet(int state_output, float weight)
            : _size(1), _chunk(_Chunk::allocate(1)) {
#if 0
            cerr << "set(1)" << endl;
#endif
            _Triple* t = &_chunk->triples[0];
            t->state_output = state_output;
            t->weight = weight;
        }
        
        _TripleSet(unsigned n)
            : _size(n), _chunk(0) {
#if 0
            cerr << "set(" << n << ')' << endl;
#endif
            if (_size) {
                _Chunk* c;
                _Chunk* tmp = 0;
                while (n != 0) {
                    unsigned s = (n >= _chunk_elements) ? _chunk_elements : n;
                    c = _Chunk::allocate(s);
                    c->next = tmp;
                    tmp = c;
                    n -= s;
                }
                while ((c = tmp) != 0) {
                    tmp = c->next;
                    c->next = _chunk;
                    _chunk = c;
                }
            }
        }
        
        ~_TripleSet() {
            unsigned n = _size;
            while (n != 0) {
                unsigned s = (n >= _chunk_elements) ? _chunk_elements : n;
                _Chunk* c = _chunk;
                _chunk = c->next;
                _Chunk::deallocate(c, s);
                n -= s;
            }
        }
        
        unsigned size() const {
            return _size;
        }
        
        Iterator elements() {
            return Iterator(_chunk, _chunk ? _chunk->triples : 0, _size);
        }
        
        ConstIterator const_elements() const {
            return ConstIterator(_chunk, _chunk ? _chunk->triples : 0, _size);
        }

        FSTAllocator_new_delete(_TripleSet);
        
    };

    typedef fst::map<_TripleSet*,int,_TripleSet::Compare>  _Map;

    struct _CollectionEntry {
        int input;
        int state;
        int output[2];
        float weight;

        _CollectionEntry(int i, int s, int o1, int o2, float w)
            : input(i), state(s), weight(w) {
            output[0] = o1 ? o1 : o2;
            output[1] = o1 ? o2 : 0;
        }

        // This comparison object is used when sorting entries in a collection.
        // Note that the order of comparision is input, output, state.
        //
        struct Compare {
            bool operator()(const _CollectionEntry* x,
                            const _CollectionEntry* y) const {
            return FST_LESS3(x->input, y->input,
                       FST_LESS3(x->output[0], y->output[0],
                           FST_LESS3(x->output[1], y->output[1],
                               FST_LESS2(x->state, y->state))));
            }
        };

        FSTAllocator_new_delete(_CollectionEntry);
    };

    typedef fst::vector<_CollectionEntry*> _Collection;

    struct _State {
        float weight;
        float potential;
        _TripleSet* set;
        _State(FSTSemiring r, _TripleSet* s)
            : weight(r->zero), potential(r->one), set(s) {}
        ~_State() {
            delete set;
        }
        FSTAllocator_new_delete(_State);
    };

    typedef fst::vector<_State*>             _StatePtrVector;

protected: // members

    FSTGeneric              _in;
    mutable _StatePtrVector _states;
    mutable _Map            _map;
    mutable _PairMap        _pair_map;
    mutable _PairVector     _pair_vector;
    mutable _Collection     _collection;
    int                     _degree;
    bool                    _fsa;
    bool                    _strict;
    fst::set<int>           _input_alphabet;
    fst::set<int>           _output_alphabet;
#ifdef NO_HASH_SET
    fst::set<int>           _pound_labels;
#else
    fst::hash_set<int>      _pound_labels;
#endif
    float                   _eps;

protected: // methods

    int _lookup_pair(int state, int output) const;
    const _Pair* _lookup_pair(int state_output) const {
        return _pair_vector[state_output];
    }
    FSTArc* _simple_arc(int p, const _CollectionEntry* e) const;
    int _find_or_add_state(_TripleSet* s) const;
    int _init() const;
    void _clean() const;
    void _print(ostream& os, const _TripleSet* s) const;
    void _print(ostream& os, const _CollectionEntry* e) const;
};

FSTGeneric
FSTDeterminize(const FSTGeneric& x, int degree, bool fsa, float eps, bool strict)
{
    return FSTGeneric(new FSTDeterminizeRep(x, degree, fsa, eps, strict));
}

FSTGeneric
FSTDeterminizeA(const FSTGeneric& x, int degree, float eps, bool strict)
{
    return FSTGeneric(new FSTDeterminizeRep(x, degree, true, eps, strict));
}

static inline float
_equal(float x, float y, float eps)
{
    float diff = x - y;
    return diff < eps && diff > -eps;
}

///////////////////////////////////////////////////////////////////////////////

const unsigned FSTDeterminizeRep::_TripleSet::_chunk_elements =
    (FSTAllocator_max_bytes - sizeof(FSTDeterminizeRep::_TripleSet::_Chunk)) / sizeof(FSTDeterminizeRep::_Triple);
        
FSTDeterminizeRep::FSTDeterminizeRep(
    const FSTGeneric& in, int degree, bool fsa, float eps, bool strict)
    : _in(in),
      _map(_TripleSet::Compare(eps)),
      _degree(degree),
      _fsa(fsa),
      _strict(strict),
      _eps(eps)
{
    _class_name = "FSTDeterminize";
    _semiring = _in->semiring();

    _input_alphabet = _in->input_alphabet();
    _input_alphabet.insert(0);
    _output_alphabet = _in->output_alphabet();
    _output_alphabet.insert(0);

    if (_degree <= 0)
        _degree = INT_MAX;

    if (_strict)
        _degree = 1;

    _initial = _init();

    _collection.reserve(1024);

    for (fst::set<int>::const_iterator i = _input_alphabet.begin();
         i != _input_alphabet.end(); ++i) {
        if (*i > 1 && FSTstring_map[*i][0] == '#')
            _pound_labels.insert(*i);
    }
}

FSTDeterminizeRep::~FSTDeterminizeRep()
{
    _clean();
}

FSTGeneric
FSTDeterminizeRep::reload(bool& flag)
{
    if (_in.reload()) flag = true;
    reset();
    return FSTGeneric(this);
}

void
FSTDeterminizeRep::reset() const
{
    _in->reset();
    _clean();

    int new_initial = _init();
    assert(new_initial == _initial);
}

int
FSTDeterminizeRep::_init() const
{
    int result = -1;
    if (_in->initial() >= 0) {
        _TripleSet* initial_set =
            new _TripleSet(_lookup_pair(_in->initial(), 0), _semiring->one);
        result = _find_or_add_state(initial_set);
    }
    return result;
}

void
FSTDeterminizeRep::_clean() const
{
    for (_StatePtrVector::iterator i = _states.begin();
         i != _states.end(); i++) {
        delete *i;
    }
    _states.clear();
    _map.clear();
    _n_states = 0;
}

const fst::set<int>&
FSTDeterminizeRep::input_alphabet() const
{
    return _input_alphabet;
}

const fst::set<int>&
FSTDeterminizeRep::output_alphabet() const
{
    return _output_alphabet;
}

bool
FSTDeterminizeRep::is_final(int p) const
{
    FST_ASSERT_VALID_STATE(p);
    return _states[p]->weight != _semiring->zero;
}

float
FSTDeterminizeRep::final_weight(int p) const
{
    FST_ASSERT_VALID_STATE(p);
    return _states[p]->weight;
}

float
FSTDeterminizeRep::potential(int p) const
{
    FST_ASSERT_VALID_STATE(p);
    return _states[p]->potential;
}

FSTArc*
FSTDeterminizeRep::_simple_arc(int p, const _CollectionEntry* e) const
{
    int input = e->input;
    int output = e->output[0];
    float weight = e->weight;

    int state = e->state;

    _TripleSet* new_set =
        new _TripleSet(_lookup_pair(state, e->output[1]), _semiring->one);

#ifdef DEBUG
    cerr << "simple arc: "
         << FSTstring_map[input] << ','
         << FSTstring_map[output] << ','
         << weight << " -> ";
    _print(cerr, new_set);
    cerr << endl;
#endif
    return new FSTArc(p, _find_or_add_state(new_set), input, output, weight);
}

FSTArcIter
FSTDeterminizeRep::arcs_out(int p, FSTArcOrder order) const
{
    FST_ASSERT_VALID_STATE(p);

    const _TripleSet* pset = _states[p]->set;
#ifdef DEBUG
    cerr << "working on: ";
    _print(cerr, pset);
    cerr << endl;
#endif

    //
    // Sweep through all states in pset, and store all arcs in collection.
    //
    _collection.clear();

    for (_TripleSet::ConstIterator pset_i = pset->const_elements();
         pset_i.valid(); ++pset_i) {

        const _Pair* so = _lookup_pair(pset_i.state_output());

        if (so->state >= 0) {

            // Real state: insert arcs into collection.
            //
            for (FSTArcIter a = _in->arcs_out(so->state); a.valid(); a++) {
                float w = _semiring->times(pset_i.weight(), a->weight);
                _collection.push_back
                    (new _CollectionEntry(a->input, a->next,
                                          so->output, a->output, w));
            }
            
        }
        
        if ((so->state < 0 || _in->is_final(so->state)) && so->output != 0) {
            
            // Epsilon arcs due to pending outputs at final states.
            //
            float w = (so->state < 0) ?
                pset_i.weight() :
                _semiring->times(pset_i.weight(), _in->final_weight(so->state));
            _collection.push_back
                (new _CollectionEntry(0, -1, so->output, 0, w));
            
        }
        
    }
    
    if (_collection.empty())
        return FSTArcIter();

    sort(_collection.begin(), _collection.end(), _CollectionEntry::Compare());

#ifdef DEBUG
    cerr << "collection (after sorting): " << _collection.size() << "\n";
    for (unsigned i = 0; i < _collection.size(); i++) {
        cerr << "\t";
        _print(cerr, _collection[i]);
        cerr << endl;
    }
#endif

    //
    // Merge consecutive (input,triple) that have identical input, state, and
    // output, combining weights (semiring plus).  Relies on collection already
    // being sorted.
    //

    _Collection::iterator i, j, k;
    i = j = _collection.begin();
    if (j != _collection.end())
        j++;
    for (; j != _collection.end(); j++) {
        _CollectionEntry* ii = *i;
        _CollectionEntry* jj = *j;
        if (ii->input == jj->input &&
            ii->state == jj->state &&
            ii->output[0] == jj->output[0] &&
            ii->output[1] == jj->output[1]) {
            // Combine i and j into i, then free j (it will get overwritten or
            // erased).
            ii->weight = _semiring->plus(ii->weight, jj->weight);
            delete jj;
        } else {
            // Advance i and move j into i.  This copies j into an unused slot.
            *++i = *j;
        }
    }
    assert(!_collection.empty());
    i++;
    _collection.erase(i, _collection.end());

#ifdef DEBUG
    cerr << "collection (after combining): " << _collection.size() << "\n";
    for (unsigned i = 0; i < _collection.size(); i++) {
        cerr << "\t";
        _print(cerr, _collection[i]);
        cerr << endl;
    }
#endif

    //
    // Sweep through collection, generating arcs and their destination sets.
    //
    FSTArcIterData* arc_data = new FSTArcIterData;
    fst::vector<FSTArc*>& arcs = arc_data->arcs;

    i = j = _collection.begin();
    while (j != _collection.end()) {
        j++;
        bool special_pound = _pound_labels.count((*i)->input);

        if (j == _collection.end() ||
            (*i)->input != (*j)->input ||
            ((_fsa || (*i)->input <= 1 || special_pound) &&
             ((*i)->output[0] != (*j)->output[0] ||
              (*i)->output[1] != (*j)->output[1]))) {

            // Optionally allow a degree of nondeterminism.
            //
            if (j - i <= _degree && !special_pound) {
#ifdef DEBUG
                cerr << "shortcut <= degree" << endl;
#endif
                for (k = i; k != j; k++) {
                    arcs.push_back(_simple_arc(p, *k));
                    delete *k;
                    *k = 0;
                }
                i = j;
                continue;
            }

            // If the inputs are different, then we are ready to create the arc
            // with (*i).input.  The destination set contains the triples [i,j).
            //
            int input = (*i)->input;

            // Extract output and weight from triples [i,j).
            //
            int output = (*i)->output[0];
            float weight = (*i)->weight;
            for (k = i + 1; k != j; k++) {
                if (output != (*k)->output[0])
                    output = 0;
                weight = _semiring->plus(weight, (*k)->weight);
            }
            if (output != 0) {
                // Extracted output, so shift entry outputs.
                for (k = i; k != j; k++) {
                    (*k)->output[0] = (*k)->output[1];
                    (*k)->output[1] = 0;
                }
            }
            int n_output_overflows = 0;
            if (!output) {
                // We failed to extract an output.  Look for any entries with
                // two pending outputs.  We cannot store two pending outputs in
                // a triple, so escape such output overflow by emitting
                // nondeterministic arcs.
                for (k = i; k != j; k++) {
                    if ((*k)->output[1]) {
                        arcs.push_back(_simple_arc(p, *k));
                        delete *k;
                        *k = 0;
                        n_output_overflows++;
                    }
                }
            }
            if (n_output_overflows == j - i) {
                i = j;
                continue;
            }
            if (weight != _semiring->one) {
                // Extracted weight, so remove from entry weights.
                for (k = i; k != j; k++) {
                    if (!*k) continue;
                    (*k)->weight = _semiring->divide((*k)->weight, weight);
                }
            }

            // Determine if we are faced with weight ambiguity caused by
            // self-loops with different weights.  We look for the set of
            // triples represented by [i,j) to have the same states and outputs
            // as pset, but different weights.  This is a recipe for infinite
            // states.  If detected, output nondeterministic arcs to avoid
            // problem.
            //
            if (!_strict &&
                n_output_overflows == 0 && j - i > 1 && j - i == int(pset->size())) {
                bool weight_ambiguity = false;
                _TripleSet::ConstIterator pset_i = pset->const_elements();
                for (k = i; k != j; ++k, ++pset_i) {
                    const _Pair* so = _lookup_pair(pset_i.state_output());
                    if (so->state != (*k)->state ||
                        so->output != (*k)->output[0]) {
                        weight_ambiguity = false;
                        break;
                    }
                    if (!_equal(pset_i.weight(), (*k)->weight, _eps)) {
                        weight_ambiguity = true;
                    }
                }
                if (weight_ambiguity) {
#if defined(DEBUG) || defined(DEBUG_WEIGHT_AMBIGUITY)
                    cerr << "resolving weight ambiguity for "
                         << FSTstring_map[(*i)->input] << endl;
                    cerr << "  from ";
                    _print(cerr, pset);
                    cerr << endl;
                    cerr << "    to {";
                    for (k = i; k != j; k++) {
                        if (k != i)
                            cerr << ",";
                        _print(cerr, *k);
                    }
                    cerr << "}" << endl;
#endif
                    for (k = i; k != j; k++) {
                        if (output) {
                            (*k)->output[1] = (*k)->output[0];
                            (*k)->output[0] = output;
                        }
                        (*k)->weight = _semiring->times((*k)->weight, weight);
                        arcs.push_back(_simple_arc(p, *k));
                        delete *k;
                        *k = 0;
                    }
                    i = j;
                    continue;
                }  
            }

            // Build new set from updated triples [i, j).
            //
            unsigned set_size = j - i - n_output_overflows;
            _TripleSet* new_set = new _TripleSet(set_size);
            _TripleSet::Iterator si = new_set->elements();
            for (k = i; k != j; k++) {
                if (!*k) continue;
                assert((*k)->output[1] == 0);
                assert(si.valid());
                si.state_output() = _lookup_pair((*k)->state, (*k)->output[0]);
                si.weight() = (*k)->weight;
                ++si;
                delete *k;
                *k = 0;
            }
            assert(!si.valid());

#ifdef DEBUG
            cerr << "arc: " << FSTstring_map[input] << "," << FSTstring_map[output] <<
                "," << weight << " -> ";
            _print(cerr, new_set);
            cerr << endl;
#endif
            arcs.push_back(new FSTArc(p, _find_or_add_state(new_set),
                                      input, output, weight));
            i = j;
        }
    }
    
    _collection.clear();

    return FSTArcIter(arc_data);
}

FSTArcIter
FSTDeterminizeRep::arcs_out_with_input(int p, int i, FSTArcMode m) const
{
    FSTFatal("FSTDeterminizeRep::arcs_out_with_input: not implemented");
    return FSTArcIter();
}

FSTArcIter
FSTDeterminizeRep::arcs_out_with_output(int p, int o, FSTArcMode m) const
{
    FSTFatal("FSTDeterminizeRep::arcs_out_with_output: not implemented");
    return FSTArcIter();
}

void
FSTDeterminizeRep::force() const
{
    // Expand all states.  We do this by asking for all arcs out of all states
    // (without actually creating all the arcs).
    FSTFatal("FSTDeterminizeRep::force not implemented");
}

int
FSTDeterminizeRep::_find_or_add_state(_TripleSet* ts) const
{
    int p;
    _Map::iterator i = _map.find(ts);

    if (i == _map.end()) {

        // State not found: build new state.
        //
        _State* pp = new _State(_semiring, ts);
        _states.push_back(pp);
        p = _n_states++;
        _map[ts] = p;

        // Calculate final weight for state.  Note that final states in the set
        // with a pending output do not contribute.  Not also that state == -1
        // with no pending outputs is also a final state.  Also compute potential.
        // It might be preferable to use lazy evaluation.
        //
        pp->potential = _semiring->zero;
        for (_TripleSet::ConstIterator j = ts->const_elements(); j.valid(); ++j) {
            const _Pair* so = _lookup_pair(j.state_output());
            if (so->state == -1 && !so->output) {
                pp->weight = j.weight();
            } else if (so->state >= 0 && _in->is_final(so->state) && !so->output) {
                float w = _semiring->times(_in->final_weight(so->state),
                                           j.weight());
                pp->weight = _semiring->plus(pp->weight, w);
            }
            if (so->state == -1) {
                pp->potential = _semiring->better(pp->potential, j.weight());
            } else {
                float w = _semiring->times(j.weight(), _in->potential(so->state));
                pp->potential = _semiring->better(pp->potential, w);
            }
        }

#ifdef DEBUG_SETS
        cerr << p << " :";
        for (_TripleSet::ConstIterator j = ts->const_elements(); j.valid(); ++j) {
            const _Pair* so = _lookup_pair(j.state_output());
            cerr.form(" (%d,%s,%.8g)",
                      so->state, FSTstring_map[so->output], j.weight());
        }
        cerr << "\n";
#endif

    } else {

        // State found: look up state index and free set (it isn't needed).
        //
        p = (*i).second;
        delete ts;

    }

    return p;
}

int
FSTDeterminizeRep::_lookup_pair(int state, int output) const
{
    _Pair p(state, output);
    _PairMap::iterator i = _pair_map.lower_bound(p);
    int result;
    if (i != _pair_map.end() && (*i).first == p) {
        result = (*i).second;
    } else {
        result = _pair_vector.size();
        i = _pair_map.insert(i, _PairMap::value_type(p, result));
        _pair_vector.push_back(&(*i).first);
        return result;
    }
    return result;
}

void
FSTDeterminizeRep::_print(ostream& os, const _TripleSet* s) const
{
    os << "{";
    bool first = true;
    for (_TripleSet::ConstIterator i = s->const_elements(); i.valid(); ++i) {
        const _Pair* so = _lookup_pair(i.state_output());
        if (!first) os << ",";
        first = false;
        os << "("
           << so->state << ","
           << FSTstring_map[so->output] << ","
           << i.weight() << ")";
    }
    os << "}";
}

void
FSTDeterminizeRep::_print(ostream& os, const _CollectionEntry* e) const
{
    os << "("
       << FSTstring_map[e->input] << ","
       << e->state << ",["
       << FSTstring_map[e->output[0]] << ","
       << FSTstring_map[e->output[1]] << "],"
       << e->weight << ")";
}

void
FSTDeterminizeRep::state_set(int p, fst::set<int>& result) const
{
    FST_ASSERT_VALID_STATE(p);

    result.clear();

    const _TripleSet* pset = _states[p]->set;
    for (_TripleSet::ConstIterator pset_i = pset->const_elements();
         pset_i.valid(); ++pset_i) {
        const _Pair* so = _lookup_pair(pset_i.state_output());
        result.insert(so->state);
    }
}

void
FSTDeterminizeStateSet(const FSTGeneric& x, int p, fst::set<int>& result)
{
    dynamic_cast<const FSTDeterminizeRep*>(x.pointer())->state_set(p, result);
}

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 80
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
