////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

//#define INSTRUMENT
#define HASH
//#define HASH_CONNECTIONS
//#define RE_IN

#include <cstring>
#include "fst/FST_i.h"
#include "fst/format.h"
#include <cctype>

#ifdef NO_HASH_SET
#  undef HASH
#endif

#define not_implemented   false
#define should_not_happen false

class FSTDynamicRep
    : public FSTGenericInt {

public: // methods

    FSTDynamicRep(const FSTGeneric& main);
    virtual ~FSTDynamicRep();
    virtual const fst::set<int>& input_alphabet() const;
    virtual const fst::set<int>& output_alphabet() const;
    virtual bool is_final(int p) const;
    virtual float final_weight(int p) const;
    virtual float potential(int p) const;
    virtual FSTArcIter arcs_out(int p, FSTArcOrder order) const;
    virtual FSTArcIter arcs_out_with_input(int p, int i, FSTArcMode m) const;
    virtual FSTArcIter arcs_out_with_output(int p, int o, FSTArcMode m) const;
    virtual void arcs_out(fst::vector<FSTArc*>& arcs, int p, FSTArcOrder order) const;
    virtual void force() const;
    virtual void reset() const;

    virtual FSTGeneric reload(bool& flag);

public: // new methods

    bool set_class(int name, const FSTGeneric& fst);
    FSTGeneric get_class(int name) const;

protected: // types

    class _State {
        int _main_state;
        int _constraint_or_state;
        const FSTGenericInt* _class_fst;

    public:

        _State(int ms)
            : _main_state(ms), _constraint_or_state(0), _class_fst(NULL) {
        }
        _State(int ms, int c, int cs, const FSTGenericInt* cf)
            : _main_state(ms), _constraint_or_state(c ? -c : cs), _class_fst(cf) {
            assert(c == 0 || _class_fst == NULL);
        }
                                                                 
        int main_state() const {
            return _main_state;
        }

        int constraint() const {
            return _constraint_or_state < -1 ? -_constraint_or_state : 0;
        }

        bool final() const {
            return _constraint_or_state == -1;
        }

        int class_state() const {
            return _constraint_or_state < 0 ? -1 : _constraint_or_state;
        }

        const FSTGenericInt* class_fst() const {
            return _class_fst;
        }
                                                                 
#ifdef HASH
#  if defined(__GNUC__) && !__GNUC_PREREQ(3,4)
#    define FRIEND friend
#  else
#    define FRIEND
#  endif
        FRIEND struct Hash {
            size_t operator()(const _State* x) const {
                return 17*x->_main_state - 5*x->_constraint_or_state +
                    size_t(x->_class_fst);
            }
        };
        FRIEND struct Equal {
            bool operator()(const _State* x, const _State* y) const {
                return x->_main_state == y->_main_state &&
                    x->_constraint_or_state == y->_constraint_or_state &&
                    x->_class_fst == y->_class_fst;
            }
        };
#else
        bool operator<(const _State& rhs) const {
            return FST_LESS3(_main_state, rhs._main_state,
                             FST_LESS3(_constraint_or_state, rhs._constraint_or_state,
                                       FST_LESS2(_class_fst, rhs._class_fst)));
        }
#endif
    };

#ifdef HASH
    typedef fst::hash_set<int> _LabelSet;
    typedef fst::hash_map<int, FSTGeneric> _ClassMap;
    typedef fst::hash_map<_State*, int, _State::Hash, _State::Equal> _StateMap;
#else
    typedef fst::set<int> _LabelSet;
    typedef fst::map<int, FSTGeneric> _ClassMap;
    typedef fst::map<_State*, int, ptr_less<_State*> > _StateMap;
#endif

#ifdef HASH_CONNECTIONS
    struct StateOutputWeight {
        int state;
        int output;
        float weight;
        StateOutputWeight(int s, int o, float w)
            : state(s), output(o), weight(w) {}
    };

    struct StateConstraint {
        int state_class;
        int constraint;
        StateConstraint(int sc, int c) : state_class(sc), constraint(c) {}

        friend struct Hash {
            size_t operator()(const StateConstraint& x) const {
                return 17*x.state_class + 5*x.constraint;
            }
        };

        friend struct Equal {
            size_t operator()(const StateConstraint& x, const StateConstraint& y) const {
                return x.state_class == y.state_class
                    && x.constraint  == y.constraint;
            }
        };
    };
    typedef pair<int,int> IntPair;
    typedef hash_multimap<StateConstraint, StateOutputWeight,
                          StateConstraint::Hash, StateConstraint::Equal> _Multimap;
#endif

protected: // members

    mutable fst::vector<_State*> _states;
    mutable _StateMap _state_map;
    FSTGeneric _main_fst;
    _LabelSet _special_set;
    _ClassMap _class_map;
    mutable _LabelSet _class_set;
    int _star;
    bool _epsilon_to_final;
#ifdef HASH_CONNECTIONS
    _Multimap _connection_map;
#endif
    mutable fst::set<int> _input_alphabet;
    mutable fst::set<int> _output_alphabet;
#ifdef INSTRUMENT
    mutable int _n_static_states;
    mutable int _n_static_arcs;
    mutable int _n_dynamic_arcs;
    mutable int _n_epsilon_arcs;
#endif

protected: // methods

    void _init_main();
    void _build_alphabets() const;
    void _merge_alphabets(const FSTGeneric& fst, bool is_main) const;

    const FSTGenericInt* _lookup_class(int name) const {
        _ClassMap::const_iterator i = _class_map.find(name);
        return (i == _class_map.end()) ? NULL : (*i).second.pointer();
    }

    void _arcs_out(fst::vector<FSTArc*>& arcs, int p, FSTArcOrder order) const;
    int _find_or_add_state(_State* s) const;

    void _add_arc(fst::vector<FSTArc*>& arcs, int from,
                  int main_state, int constraint, int class_state,
                  const FSTGenericInt* class_fst,
                  int i, int o, float w) const {
        int to = _find_or_add_state(new _State(main_state, constraint,
                                               class_state, class_fst));
        arcs.push_back(new FSTArc(from, to, i, o, w));
    }

    static void _add_arc(fst::vector<FSTArc*>& arcs, int from,
                         int to,
                         int i, int o, float w) {
        arcs.push_back(new FSTArc(from, to, i, o, w));
    }

};

FSTDynamicRep::FSTDynamicRep(const FSTGeneric& main_fst)
    : _main_fst(main_fst)
    , _epsilon_to_final(false)
#ifdef INSTRUMENT
    , _n_static_states(0)
    , _n_static_arcs(0)
    , _n_dynamic_arcs(0)
    , _n_epsilon_arcs(0)
#endif
{
    _class_name = "FSTDynamic";
    _semiring = _main_fst->semiring();
    _star = FSTstring_map["#star"];

    _init_main();
    reset();
}

FSTDynamicRep::~FSTDynamicRep()
{
#ifdef INSTRUMENT
    reset();
#endif
    for (fst::vector<_State*>::iterator i = _states.begin(); i != _states.end(); ++i) {
        delete *i;
    }
}

void
FSTDynamicRep::_init_main()
{
    _initial = _main_fst->initial();

    _n_arcs = 0;
    const int N = _n_states = _main_fst->n_states();
    
#ifdef INSTRUMENT
    _n_static_states = _n_static_arcs = _n_dynamic_arcs = _n_epsilon_arcs = 0;
#endif

    _special_set.clear();
    const fst::set<int>& alphabet = _main_fst->input_alphabet();
    for (fst::set<int>::const_iterator i = alphabet.begin(); i != alphabet.end(); ++i) {
        const char* s = FSTstring_map[*i];
        if (s[0] == '#' && s[1] != '\0') {
            _special_set.insert(*i);
        }
    }

    // Determine which main states have no constraint or #star arcs leaving,
    // meaning the arcs for these states can be passed through unchanged.  Such
    // states will have NULL in the _states vector.
    //
    assert(_states.empty());
    _states.insert(_states.end(), N, (_State*)NULL);
    _class_set.clear();
    _epsilon_to_final = false;
    for (int p = 0; p < N; p++) {
        bool special = false;
        for (FSTArcIter a = _main_fst->arcs_out(p); a.valid(); ++a) {
            if (a->input == _star) {
                _class_set.insert(a->output);
            }
            if (a->input == _star || _special_set.count(a->input)) {
                special = true;
            }
            if (a->input == 0 && _main_fst->is_final(a->next)) {
                _epsilon_to_final = true;
            }
            
        }
        if (special) {
            _states[p] = new _State(p);
#ifdef INSTRUMENT
        } else {
            _n_static_states++;
#endif
        }
    }
}

FSTGeneric
FSTDynamicRep::reload(bool& flag)
{
    reset();

    // First, do generic reload.  If it reloads, return because we're finished.
    //
    {
        bool generic_reload = false;
        FSTGeneric generic_reload_fst = FSTGenericInt::reload(generic_reload);
        if (generic_reload) {
            flag = true;
            return generic_reload_fst;
        }
    }

    // Now do FSTDynamic-specific reload.
    //
    if (_main_fst.reload()) {
        flag = true;
        _init_main();
    }
    _LabelSet reloaded;
    for (_ClassMap::iterator i = _class_map.begin(); i != _class_map.end(); ++i) {
        if ((*i).second.reload()) {
            flag = true;
            reloaded.insert((*i).first);
        }
    }
#ifdef HASH_CONNECTIONS
    for (_Multimap::iterator i = _connection_map.begin(); i != _connection_map.end();) {
        int state_class = (*i).first.state_class;
        if (state_class < 0 && reloaded.count(state_class)) {
            assert(not_implemented);
        }
    }
#endif
    if (flag)
        _build_alphabets();
    return FSTGeneric(this);
}

void
FSTDynamicRep::reset() const
{
#ifdef INSTRUMENT
    cerr << "FSTDynamic: states: "
         << _main_fst->n_states() << ' '
         << _states.size() << ' '
         << _n_static_states << ' '
         << "arcs: "
         << _n_dynamic_arcs << ' '
         << _n_static_arcs << ' '
         << _n_epsilon_arcs
         << endl;
    _n_static_arcs = _n_dynamic_arcs = _n_epsilon_arcs = 0;
#endif

    // Reset main FST.
    //
    _main_fst->reset();
    int N = _main_fst->n_states();

    // Reset class FSTs.
    //
    for (_ClassMap::const_iterator i = _class_map.begin();
         i != _class_map.end(); ++i) {
        (*i).second->reset();
    }

    // Free dynamically created states.
    //
    for (fst::vector<_State*>::iterator i = _states.begin() + N;
         i != _states.end(); ++i) {
        delete *i;
    }
    _states.erase(_states.begin() + N, _states.end());

    // Clear state map.
    //
    _state_map.clear();

    // Reset sizes.
    //
    _n_states = N;
    _n_arcs = 0;
}

bool
FSTDynamicRep::set_class(int name, const FSTGeneric& fst)
{
    if (_class_set.count(name)) {
        assert(fst->semiring() == _semiring);
        _class_map[name] = fst;
        _input_alphabet.clear();
        _output_alphabet.clear();
        const fst::set<int>& alphabet = fst->input_alphabet();
        for (fst::set<int>::const_iterator i = alphabet.begin(); i != alphabet.end(); ++i) {
            const char* s = FSTstring_map[*i];
            if (s[0] == '#' && s[1] != '\0') {
                _special_set.insert(*i);
            }
        }
        reset();
        return true;
    } else {
        return false;
    }
}

FSTGeneric
FSTDynamicRep::get_class(int name) const
{
    _ClassMap::const_iterator i = _class_map.find(name);
    if (i != _class_map.end())
        return FSTBasic((*i).second);
    else
        return FSTBasic();
}

#if 0
static void print(ostream& os, const char* prefix, const fst::set<int>& alphabet)
{
    for (fst::set<int>::const_iterator i = alphabet.begin(); i != alphabet.end(); ++i)
        os << prefix << FSTstring_map[*i] << endl;
}
#endif

void
FSTDynamicRep::_build_alphabets() const
{
    _input_alphabet.clear();
    _output_alphabet.clear();

    _merge_alphabets(_main_fst, true);
    for (_ClassMap::const_iterator i = _class_map.begin();
         i != _class_map.end(); ++i) {
        _merge_alphabets((*i).second, false);
    }

    _input_alphabet.insert(1); // pound
}

void
FSTDynamicRep::_merge_alphabets(const FSTGeneric& fst, bool is_main) const
{
    const fst::set<int>& ia = fst->input_alphabet();
    for (fst::set<int>::const_iterator i = ia.begin(); i != ia.end(); ++i)
        if (_special_set.count(*i) == 0)
            _input_alphabet.insert(*i);

    const fst::set<int>& oa = fst->output_alphabet();
    for (fst::set<int>::const_iterator i = oa.begin(); i != oa.end(); ++i)
        if (!is_main || _class_map.count(*i) == 0)
            _output_alphabet.insert(*i);
}

const fst::set<int>&
FSTDynamicRep::input_alphabet() const
{
    if (_input_alphabet.empty())
        _build_alphabets();
    return _input_alphabet;
}

const fst::set<int>&
FSTDynamicRep::output_alphabet() const
{
    if (_input_alphabet.empty())
        _build_alphabets();
    return _output_alphabet;
}

bool
FSTDynamicRep::is_final(int p) const
{
    FST_ASSERT_VALID_STATE(p);
    const _State* state = _states[p];
    if (!state)
        return _main_fst->is_final(p);
    return state->constraint() == 0
        && _main_fst->is_final(state->main_state())
        && (!state->class_fst() || state->class_fst()->is_final(state->class_state()));
}

float
FSTDynamicRep::final_weight(int p) const
{
    FST_ASSERT_VALID_STATE(p);
    const _State* state = _states[p];
    if (!state)
        return _main_fst->final_weight(p);
    if (state->constraint())
        return _semiring->zero;
    float main_w = _main_fst->final_weight(state->main_state());
    if (main_w == _semiring->zero)
        return _semiring->zero;
    if (state->class_fst()) {
        float class_w = state->class_fst()->final_weight(state->class_state());
        if (class_w == _semiring->zero)
            return _semiring->zero;
        else
            return _semiring->times(main_w, class_w);
    } else {
        return main_w;
    }
}

float
FSTDynamicRep::potential(int p) const
{
    // Not really implemented, nor likely necessary.

    FST_ASSERT_VALID_STATE(p);
    return _semiring->one;
}

void
FSTDynamicRep::_arcs_out(fst::vector<FSTArc*>& arcs, int p, FSTArcOrder order) const
{
    const int pound = 1;

    const _State* state = _states[p];
    assert(state != NULL);

    arcs.clear();

    if (state->class_fst() == NULL) {
        
        if (state->constraint()) {
            
            // in main FST with pending constraint
            
            for (FSTArcIter a = _main_fst->arcs_out_with_input(state->main_state(), _star);
                 a.valid(); ++a) {
                const FSTGenericInt* class_fst = _lookup_class(a->output);
                if (class_fst) {
                    for (FSTArcIter b = class_fst->arcs_out_with_input(class_fst->initial(), state->constraint());
                         b.valid(); ++b) {
#ifdef RE_IN
                        assert(b->output == 0);
                        assert(!class_fst->is_final(b->next));
                        float ab = _semiring->times(a->weight, b->weight);
                        for (FSTArcIter c = class_fst->arcs_out(b->next); c.valid(); ++c) {
                            if (true || _special_set.count(c->input) == 0) {
                                _add_arc(arcs, p,
                                         a->next, 0, c->next, class_fst,
                                         c->input, c->output, _semiring->times(ab, c->weight));
                            } else {
                                _add_arc(arcs, p,
                                         a->next, c->input, 0, NULL,
                                         pound, c->output, _semiring->times(ab, c->weight));
                            }
                        }
#else // !RE_IN
                        _add_arc(arcs, p,
                                 a->next, 0, b->next, class_fst,
                                 0, b->output, _semiring->times(a->weight, b->weight));
#ifdef INSTRUMENT
                        _n_epsilon_arcs++;
#endif // INSTRUMENT
#endif // RE_IN
                    }
                }
            }
            
            for (FSTArcIter a = _main_fst->arcs_out_with_input(state->main_state(), state->constraint());
                 a.valid(); ++a) {
                _add_arc(arcs, p,
                         a->next,
                         0, a->output, a->weight);
#ifdef INSTRUMENT
                _n_epsilon_arcs++;
#endif
            }

            for (FSTArcIter a = _main_fst->arcs_out_with_input(state->main_state(), 1, FSTLessEqual);
                 a.valid(); ++a) {
                _add_arc(arcs, p,
                         a->next, state->constraint(), -1, NULL,
                         0, a->output, a->weight); // intentionally map # (1) to eps (0)
#ifdef INSTRUMENT
                _n_epsilon_arcs++;
#endif
            }

        } else if (state->final()) {

            // in main FST with pending final

            for (FSTArcIter a = _main_fst->arcs_out_with_input(state->main_state(), 0);
                 a.valid(); ++a) {
                _add_arc(arcs, p,
                         a->next, pound, -1, NULL,
                         0, a->output, a->weight);
#ifdef INSTRUMENT
                _n_epsilon_arcs++;
#endif
            }

        } else {
            
            // in main FST with no pending constraint
            
            for (FSTArcIter a = _main_fst->arcs_out(state->main_state()); a.valid(); ++a) {
                if (_special_set.count(a->input) == 0) {
                    // normal arc
                    _add_arc(arcs, p,
                             a->next,
                             a->input, a->output, a->weight);
                } else if (a->input == _star) {
                    // star arc
                    const FSTGenericInt* class_fst = _lookup_class(a->output);
                    if (class_fst) {
                        int i = class_fst->initial();
                        float fw = class_fst->final_weight(i);
                        if (fw != _semiring->zero) {
                            _add_arc(arcs, p,
                                     a->next,
                                     0, 0, _semiring->times(a->weight, fw));
                        }
                        for (FSTArcIter b = class_fst->arcs_out(class_fst->initial());
                             b.valid(); ++b) {
                            if (_special_set.count(b->input) == 0) {
                                _add_arc(arcs, p,
                                         a->next, 0, b->next, class_fst,
                                         b->input, b->output, _semiring->times(a->weight, b->weight));
                            }
                        }
                    }
                } else {
                    // constraint arc
                    _add_arc(arcs, p,
                             a->next, a->input, 0, NULL,
                             pound, a->output, a->weight);
                }
            }
            
        }
        
    } else {
        
        // in class FST with no pending constraint
        
        const FSTGenericInt* class_fst = state->class_fst();

        // If the main FST has any epsilon-input arcs arriving at final states,
        // we will specially handle epsilon-input arcs leaving the next main
        // state when we are at a final class state by creating an epsilon arc
        // to a main state with pending finality.
        //
        if (_epsilon_to_final) {
            float fw = class_fst->final_weight(state->class_state());
            if (fw != _semiring->zero) {
                for (FSTArcIter a = _main_fst->arcs_out_with_input(state->main_state(), 0);
                     a.valid(); ++a) {
                    if (_main_fst->is_final(a->next)) {
                        float w = _semiring->times(fw, a->weight);
                        _add_arc(arcs, p,
                                 a->next, pound, -1, NULL,
                                 0, a->output, w);
                    }
                }
            }
        }

        for (FSTArcIter a = class_fst->arcs_out(state->class_state());
             a.valid(); ++a) {
            if (_special_set.count(a->input) == 0) {
                _add_arc(arcs, p,
                         state->main_state(), 0, a->next, class_fst,
                         a->input, a->output, a->weight);
            } else {
                float fw = class_fst->final_weight(a->next);
                if (fw == _semiring->zero) {
                    cerr << "ERROR: "
                         << a->prev << ','
                         << a->next << ','
                         << FSTstring_map[a->input] << ','
                         << FSTstring_map[a->output] << ','
                         << a->weight << endl;
                }
                assert(fw != _semiring->zero);
                _add_arc(arcs, p,
                         state->main_state(), a->input, 0, NULL,
                         pound, a->output, _semiring->times(a->weight, fw));
            }
        }
        
    }

    if (order != FSTNone)
        _sort_arcs(arcs, order);
}

void
FSTDynamicRep::arcs_out(fst::vector<FSTArc*>& arcs, int p, FSTArcOrder order) const
{
    const _State* state = _states[p];
    if (state == NULL) {
        _main_fst->arcs_out(arcs, p, order);
#ifdef INSTRUMENT
        _n_static_arcs += arcs.size();
#endif
    } else {
        _arcs_out(arcs, p, order);
#ifdef INSTRUMENT
        _n_dynamic_arcs += arcs.size();
#endif
    }
}

FSTArcIter
FSTDynamicRep::arcs_out(int p, FSTArcOrder order) const
{
    FST_ASSERT_VALID_STATE(p);
    const _State* state = _states[p];

    // Handle simple case of passing main FST arcs directly through.  We can do
    // this for states that have no #123 or #star arcs leaving.
    //
    if (state == NULL) {
#ifdef INSTRUMENT
        FSTArcIter arcs = _main_fst->arcs_out(p, order);
        _n_static_arcs += arcs.n();
        return arcs;
#else
        return _main_fst->arcs_out(p, order);
#endif
    }

    // Handle dynamically created states.
    //
    FSTArcIterData* arc_data = new FSTArcIterData;
    _arcs_out(arc_data->arcs, p, order);
#ifdef INSTRUMENT
    _n_dynamic_arcs += arc_data->arcs.size();
#endif
    return arc_data;
}

inline bool
_keep(int x, int key, FSTArcMode mode)
{
    if (mode == FSTEqual)
        return x == key;
    else if (mode == FSTGreaterEqual)
        return x <= key;
    else
        return x >= key;
}

FSTArcIter
FSTDynamicRep::arcs_out_with_input(int p, int i, FSTArcMode m) const
{
    FST_ASSERT_VALID_STATE(p);
    const _State* state = _states[p];

    if (state == NULL) {

        // Handle simple case of passing main FST arcs directly through.  We can
        // do this for states that have no #123 or #star arcs leaving.
        //
        return _main_fst->arcs_out_with_input(p, i, m);

    } else {

        // Handle dynamically created states.
        //
        FSTArcIterData* arc_data = new FSTArcIterData;
        _arcs_out(arc_data->arcs, p, FSTInput);
        
        // Select those arcs consistent with i and m.
        //
        FSTArcCompareInput cmp;
        FSTArc a;
        a.input = i;
        typedef fst::vector<FSTArc*>::iterator Iter;
        pair<Iter, Iter> range;
        switch (m) {
        case FSTEqual:
            range = equal_range(arc_data->arcs.begin(), arc_data->arcs.end(), &a, cmp);
            break;
        case FSTLessEqual:
            range.first = arc_data->arcs.begin();
            range.second = upper_bound(arc_data->arcs.begin(), arc_data->arcs.end(), &a, cmp);
            break;
        case FSTGreaterEqual:
            range.first = lower_bound(arc_data->arcs.begin(), arc_data->arcs.end(), &a, cmp);
            range.second = arc_data->arcs.end();
            break;
        }
        
        // Delete unused arcs.
        //
        for (Iter i = arc_data->arcs.begin(); i < range.first; i++)
            delete *i;
        for (Iter i = range.second; i < arc_data->arcs.end(); i++)
            delete *i;
        
        // Make new arc_data->arcs of the correct size containing range.
        //
        fst::vector<FSTArc*> arcs(range.first, range.second);
        arc_data->arcs.swap(arcs);
        
        return arc_data;

    }
}

FSTArcIter
FSTDynamicRep::arcs_out_with_output(int p, int o, FSTArcMode m) const
{
    FST_ASSERT_VALID_STATE(p);
    const _State* state = _states[p];

    // Because _arcs_out() makes use of _main_fst->arcs_out_with_input(), we
    // cannot efficiently just return _main_fst->arcs_out_with_input() for
    // non-dynamic states because _main_fst may repeatedly resort its arcs.
    // Thus, we'll handle all states dynamically.
    //
    FSTArcIterData* arc_data = new FSTArcIterData;
    
    if (state == NULL) {

        _main_fst->arcs_out(arc_data->arcs, p, FSTNone);
        _sort_arcs(arc_data->arcs, FSTOutput);

    } else {

        _arcs_out(arc_data->arcs, p, FSTOutput);

    }
        
    // Select those arcs consistent with o and m.
    //
    FSTArcCompareOutput cmp;
    FSTArc a;
    a.output = o;
    typedef fst::vector<FSTArc*>::iterator Iter;
    pair<Iter, Iter> range;
    switch (m) {
    case FSTEqual:
        range = equal_range(arc_data->arcs.begin(), arc_data->arcs.end(), &a, cmp);
        break;
    case FSTLessEqual:
        range.first = arc_data->arcs.begin();
        range.second = upper_bound(arc_data->arcs.begin(), arc_data->arcs.end(), &a, cmp);
        break;
    case FSTGreaterEqual:
        range.first = lower_bound(arc_data->arcs.begin(), arc_data->arcs.end(), &a, cmp);
        range.second = arc_data->arcs.end();
        break;
    }
    
    // Delete unused arcs.
    //
    for (Iter i = arc_data->arcs.begin(); i < range.first; i++)
        delete *i;
    for (Iter i = range.second; i < arc_data->arcs.end(); i++)
        delete *i;
    
    // Make new arc_data->arcs of the correct size containing range.
    //
    fst::vector<FSTArc*> arcs(range.first, range.second);
    arc_data->arcs.swap(arcs);
    
    return arc_data;
}

void
FSTDynamicRep::force() const
{
    // Expand all states.  This is a simplistic implementation that asks for
    // arcs out of all states.  It could be made more efficient if needed.
    //
    for (int p = 0; p < _n_states; p++)
        arcs_out(p, FSTNone);
}

int
FSTDynamicRep::_find_or_add_state(_State* state) const
{
    int p;
    _StateMap::iterator i = _state_map.find(state);

    if (i == _state_map.end()) {

        // State not found: build new state.
        //
        p = _n_states++;
        _state_map.insert(_StateMap::value_type(state, p));
        _states.push_back(state);
#if 0
        format(cerr, "add state %d = (%d,%d,%p) %s\n",
               p, state->main_state(), state->class_state(), state->class_fst(),
               state->class_state() < 0 ? FSTstring_map[-state->class_state()] : "none");
#endif
    } else {

        // State found: look up state index and free redundant _State*.
        //
        p = (*i).second;
        delete state;

    }

    return p;
}

FSTGeneric
FSTLoadDynamic(istream& is, const char* header)
{
    char *main_file_name = new char[strlen(header)];
    if (sscanf(header, "FSTDynamic %s", main_file_name) != 1) {
        FSTFatal("FSTLoadDynamic: cannot find main file name");
    }
    FSTGeneric result = FSTDynamic(FSTLoad(main_file_name));
    delete main_file_name;

    FSTGeneric empty = FSTBasic();
    empty->set_initial(empty->add_state());

    string class_name, fst_file_name;
    while (is >> class_name >> fst_file_name, is.good()) {
        if (fst_file_name == "<empty>") {
            FSTDynamicSet(result, class_name, empty);
        } else {
            FSTDynamicSet(result, class_name, FSTLoad(fst_file_name));
        }
    }

    return result;
}

FSTGeneric
FSTDynamic(const FSTGeneric& main_fst)
{
    return FSTGeneric(new FSTDynamicRep(main_fst));
}

bool
FSTIsDynamic(const FSTGeneric& fst)
{
    return dynamic_cast<const FSTDynamicRep*>(fst.pointer()) != NULL;
}

void
FSTDynamicSet(FSTGeneric& fst, const string& class_name, const FSTGeneric& class_fst)
{
    FSTDynamicSet(fst, FSTstring_map[class_name], class_fst);
}

void
FSTDynamicSet(FSTGeneric& fst, int class_name, const FSTGeneric& class_fst)
{
    FSTDynamicRep* dynamic_fst = dynamic_cast<FSTDynamicRep*>(fst.pointer());
    if (dynamic_fst) {
        bool success = dynamic_fst->set_class(class_name, class_fst);
        if (!success)
            FSTWarn("FSTDynamicSet failed: class %s not found", FSTstring_map[class_name]);
    } else {
        FSTWarn("FSTDynamicSet called on non-dynamic FST: setting ignored");
    }
}

FSTGeneric
FSTDynamicGet(const FSTGeneric& fst, const string& class_name)
{
    return FSTDynamicGet(fst, FSTstring_map[class_name]);
}

FSTGeneric
FSTDynamicGet(const FSTGeneric& fst, int class_name)
{
    const FSTDynamicRep* dynamic_fst = dynamic_cast<const FSTDynamicRep*>(fst.pointer());
    if (dynamic_fst) 
        return dynamic_fst->get_class(class_name);
    else
        return FSTBasic();
}

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 80
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
