/* -*- C++ -*-
 *
 * Copyright (c) 2004
 * Spoken Language Systems Group
 * MIT Computer Science and Artificial Intelligence Laboratory
 * Massachusetts Institute of Technology
 *
 * All Rights Reserved
 */


#include <cstring>
#include <string>
#include <cstdio>
#include "fst/set"
#include "fst/map"
#include "fst/vector"
#include "fst/FST.h"
#include "fst/FSTTimer.h"
#include "fst/format.h"

#include "fst/FSTDynamicClass.h"

using std::string;
using std::cerr;
using std::endl;
using std::pair;

static inline FSTGeneric
_empty()
{
    FSTGeneric empty = FSTBasic();
    empty->set_initial(empty->add_state());
    return empty;
}


FSTDynamicClass::FSTDynamicClass(const string& dict_file, const string& l2s_file,
                                 const string& cp_file, const string& nickname_file)
{
    _pound = FSTstring_map["#"];
    _iwt = FSTstring_map["_"];
    _dollar = FSTstring_map["$"];


    // words on input side and phonemes on output side
    if (dict_file == "" || dict_file == "None")
        _compute_L_star(_empty());
    else
        _compute_L_star(FSTLoad(dict_file.c_str()));

    // letters on input side and phonemes on output side
    if (l2s_file == "" || l2s_file == "None")
        _l2s_fst = _empty();
    else
        _l2s_fst = FSTLoad(l2s_file.c_str());

    // model labels on input side and phonemes on output side
    if (cp_file == "" || cp_file == "None")
        _CP = _empty();
    else
        _CP = FSTLoad(cp_file.c_str());

    if (nickname_file == "" || nickname_file == "None") {
        _nickname_fst = _empty();
        _nickname_spelling_fst = _empty();
    } else {
        _nickname_fst = FSTProject(FSTLoad(nickname_file.c_str()), FSTOutput, FSTInput);
        _nickname_spelling_fst = FSTBasic(_nickname_fst); // This isn't really
                                                          // correct, but will
                                                          // have to do for now.
    }

    // TODO: populate _class_map's canonical given the FST?
}

FSTDynamicClass::~FSTDynamicClass()
{
}

void
FSTDynamicClass::_compute_L_star(const FSTGeneric& L_in)
{
    // Add _* to end of all words, and closure arc eps:#.

    // Transpose to get word:phoneme, and also make a mutable copy.
    //
    _L = FSTProject(L_in, FSTOutput, FSTInput);
    int i = _L->initial();
    int N = _L->n_states();
    for (int p = 0; p < N; ++p) {
        if (_L->is_final(p)) {
            _L->add_arc(p, p, 0, _iwt);
            _L->add_arc(p, i, 0, _pound, _L->final_weight(p));
        }
    }
}

void
FSTDynamicClass::_merge(const FSTGeneric& L_add)
{
    fst::vector<int> state_map(L_add->n_states(), -1);
    state_map[L_add->initial()] = _L->initial();

    for (int p = 0; p < L_add->n_states(); ++p) {
        int& q = state_map[p];
        if (q < 0) q = _L->add_state();
        _L->set_final(q, L_add->final_weight(p));
        for (FSTArcIter a = L_add->arcs_out(p); a.valid(); ++a) {
            int& r = state_map[a->next];
            if (r < 0) r = _L->add_state();
            _L->add_arc(q, r, a->input, a->output, a->weight);
        }
        if (_L->is_final(q)) {
            _L->add_arc(q, q, 0, _iwt);
            _L->add_arc(q, _L->initial(), 0, _pound, _L->final_weight(q));
        }
    }
}

void
FSTDynamicClass::clear(FSTGeneric dynamic_fst, const string& class_name)
{
    FSTDynamicSet(dynamic_fst, class_name, clear(class_name));
}

FSTGeneric
FSTDynamicClass::clear(const string& class_name)
{
    int class_index = FSTstring_map[class_name];
    Class& c = _class_map[class_index];

    c.fst = FSTBasic();
    c.fst->set_initial(c.fst->add_state());

    c.canonical.clear();

    return c.fst;
}

void
FSTDynamicClass::null(FSTGeneric dynamic_fst, const string& class_name)
{
    FSTDynamicSet(dynamic_fst, class_name, null(class_name));
}

FSTGeneric
FSTDynamicClass::null(const string& class_name)
{
    int class_index = FSTstring_map[class_name];
    Class& c = _class_map[class_index];
    c.canonical.clear();

    FSTGeneric x = FSTBasic();
    x->add_states(3);
    x->set_initial(0);
    x->add_arc(0, 1, _dollar, 0);
    x->add_arc(1, 2, _dollar, 0);
    x->set_final(2);

    x = FSTTrim(FSTCompose(_CP, x));
    x = FSTOptimizeA(x);

    c.fst = x;

    return c.fst;
}

void
FSTDynamicClass::update(FSTGeneric dynamic_fst, bool replace,
                        const string& class_name, int n, const char** canonical_list,
                        bool spelling)
{
    FSTDynamicSet(dynamic_fst, class_name,
                  update(replace, class_name, n, canonical_list, spelling));
}

static int
_add_spelling(const char* word, FSTGeneric fst, int p)
{
    // Do nothing if we don't have [-A-Za-z']+
    //
    for (const char* l = word; *l; l++) {
        if (!isalpha(*l) && *l != '-' && *l != '\'')
            return p;
    }

    int q = p;
    char L_word[2];
    L_word[1] = 0;
    for (const char* l = word; *l; l++) {
        L_word[0] = toupper(*l);
        q = fst->add_state();
        fst->add_arc(p, q, 0, FSTstring_map[L_word]);
        p = q;
    }
    return q;
}

FSTGeneric
FSTDynamicClass::update(bool replace,
                        const std::string& class_name, const std::list<std::string>& canonical_list,
                        bool spelling)
{
    const char** c_canonical_list = new const char*[canonical_list.size()];
    int j = 0;
    for (std::list<std::string>::const_iterator i = canonical_list.begin(); i != canonical_list.end(); ++i) {
        c_canonical_list[j++] = (*i).c_str();
    }

    FSTGeneric result = update(replace, class_name, canonical_list.size(), c_canonical_list, spelling);

    delete[] c_canonical_list;

    return result;
}


FSTGeneric
FSTDynamicClass::update(bool replace,
                        const string& class_name, int n, const char** canonical_list,
                        bool spelling)
{
    FSTTimer timer;

    int class_index = FSTstring_map[class_name];

    Class& c = _class_map[class_index];

    if (n == 0) {
        if (replace) {
            return clear(class_name);
        }
        return c.fst;
    }

    if (replace && n == 1 && strcmp(canonical_list[0], "<null>") == 0) {
        return null(class_name);
    }

    // Reset the class FST if we're replacing.
    //
    if (replace) {
        c.fst = FSTBasic();
        c.fst->set_initial(c.fst->add_state());
        c.canonical.clear();
    }

    // Handle list containing a single FST file specification.  The
    // specification begins with '@' and can be a relative or fully-specified
    // filename.
    //
    // In the future we may support more than one file specification, possibly
    // in union with words and phrases, and possibly with weights.
    //
    if ((replace || c.canonical.empty()) && n == 1 && canonical_list[0][0] == '@') {
        return c.fst = FSTLoad(canonical_list[0] + 1);
    }

    // Construct a trivial FST that is the union of entries in the canonical
    // list that are not already in the class FST.
    //
    FSTGeneric& nickname_fst = spelling ? _nickname_spelling_fst : _nickname_fst;
    const IntSet& nickname_vocab = nickname_fst->input_alphabet();
    FSTGeneric canonical_fst = FSTBasic();
    canonical_fst->add_states(2);
    canonical_fst->set_initial(0);
    canonical_fst->set_final(1);
    bool already_finished = true;
    for (int i = 0; i < n; ++i) {
        // replace all ' ' with '_';
        char* c_copy = strdup(canonical_list[i]);
        for (char* cp = c_copy; *cp; cp++)
            if (*cp == ' ')
                *cp = '_';
        // Pull off optional weight "canonical entry|weight".
        float weight = 0;
        char* vbar = strchr(c_copy, '|');
        if (vbar) {
            *vbar = 0;
            sscanf(vbar + 1, "%f", &weight);
        }
        int w = FSTstring_map[c_copy];
        if (w == 0)
            goto next;
        if (c.canonical.insert(w).second) {
            // freshly inserted into canonical vocabulary
            canonical_fst->add_arc(0, 1, w, w, weight);
            already_finished = false;
            if (nickname_vocab.count(w) == 0) {
                if (_L->input_alphabet().count(w) != 0) {
                    // trivial nickname
                    int p = nickname_fst->initial();
                    int q = nickname_fst->add_state();
                    nickname_fst->set_final(q);
                    nickname_fst->add_arc(p, q, w, w);
                    if (spelling) {
                        int r = _add_spelling(c_copy, nickname_fst, q);
                        if (r != q)
                            nickname_fst->set_final(r);
                    }
                } else {
                    // don't have a nickname, generate one by splitting at ' ' or '_'.
                    int p = nickname_fst->initial();
                    for (char* cp = strtok(c_copy, "_ "); cp != NULL; cp = strtok(NULL, "_ ")) {
                        int q = nickname_fst->add_state();
                        nickname_fst->
                            add_arc(p, q, (p == nickname_fst->initial()) ? w : 0, FSTstring_map[cp], 0);
                        if (spelling) {
                            int r = _add_spelling(cp, nickname_fst, q);
                            if (r != q)
                                nickname_fst->add_arc(q, r);
                            p = r;
                        } else {
                            p = q;
                        }
                    }
                    nickname_fst->set_final(p);
                }
            }
        }
    next:
        free(c_copy);
    }
#if 0
    nickname_fst->dump_ascii("debug.nickname.fst");
#endif

    // If we're already fininshed (i.e., all the canonical entries are already
    // in the class FST), we're done.  There is nothing to do.
    //
    if (already_finished) {

        cerr << "all canonical entries already in class " << class_name << " -- no update needed"<< endl;
        return c.fst;

    }

    // Compose with nickname FST to expand.
    //
    FSTGeneric expanded_fst = FSTOptimizeA(FSTProject(FSTCompose(canonical_fst, nickname_fst), FSTOutput, FSTOutput));
#if 0
    expanded_fst->dump_ascii("debug.expanded.fst");
#endif

#if 1

    _expand_L(expanded_fst);
#else
    
    // Go through all unique words in expanded FST...
    //
    IntSet dict_words = _L->input_alphabet();
    const IntSet& expanded_words = expanded_fst->output_alphabet();
    fst::vector<FSTGeneric> new_dict_fsts;
    if (_l2s_fst->n_arcs() > 0) {
        for (IntSet::const_iterator si = expanded_words.begin(); si != expanded_words.end(); ++si) {
            int w = *si;
            if (!dict_words.count(w)) {
                // Word w is not known.
                
#if 0
                cerr << FSTstring_map[w] << " not in dictionary, trying ANGIE\n";
#endif
                
                // Split into letter FST.
                //
                string word = FSTstring_map[w];
                int n_letters = word.size();
                FSTGeneric letter_fst = FSTBasic();
                letter_fst->add_states(n_letters + 1);
                letter_fst->set_initial(0);
                letter_fst->set_final(n_letters);
                char letter[2]; letter[1] = '\0';
                for (unsigned i = 0; i < word.size(); ++i) {
                    letter[0] = word[i];
                    letter_fst->add_arc(i, i + 1, letter, letter);
                }
                
#if 0
                // Find 1-best pronunciation (letter:phoneme).
                //
                FSTGeneric phonemes = FSTSearch(letter_fst, _l2s_fst, 10, 0, false, NULL);
                if (phonemes->n_states() == 0) {
                    cerr << "warning: " << word << " failed letter-to-sound -- skipping" << endl;
                    continue;
                }
                phonemes = FSTClearWeights(phonemes);
                phonemes = FSTProject(phonemes, FSTOutput, FSTInput);
#else
                // Find pronunciations tied with the best (letter:phoneme).
                //
                FSTGeneric phonemes = FSTTrim(FSTCompose(letter_fst, _l2s_fst));
                if (phonemes->n_states() == 0) {
                    cerr << "warning: " << word << " failed letter-to-sound -- skipping" << endl;
                    continue;
                }
                phonemes = FSTClearWeights(FSTAstarGraph(FSTPotential(phonemes), FSTOutput, float(.1), 0));
#endif
                
                // Transform to word:phoneme.
                //
                FSTGeneric new_entry = FSTProject(phonemes, FSTNone, FSTInput);
                for (FSTArcIter a = new_entry->arcs_out(new_entry->initial()); a.valid(); ++a)
                    a.set_input_index(w);
#if 0
                cerr << "word " << word << " not known, using..." << endl;
                FSTTopologicalSort(new_entry)->dump_ascii(cerr);
#endif
                
                // Add to new FSTs to union into dictionary.
                //
                new_dict_fsts.push_back(new_entry);
            }
        }
    }
    
    // Update dictionary _L with the union of all words generated with
    // letter-to-sound (contained within vector new_dict_fsts).
    //
    if (!new_dict_fsts.empty()) {
        FSTTimer l_timer;
        FSTGeneric L_additions = FSTMinimize(FSTDeterminizeA(FSTCache(FSTRemoveEpsilonsA(FSTUnion(new_dict_fsts)))));
        _merge(L_additions);
        format(cerr, "time to update L: %.3fs\n", l_timer.value());
#if 0
        L_additions->dump_ascii("out.Ladd.fst");
        _L->dump_ascii("out.L.fst");
#endif
    }

#endif
    
    // Now that we have an L* that will cover our vocabulary, time to build the
    // dynamic class FST.
    
    FSTGeneric x = FSTTrim(FSTCompose(expanded_fst, _L));
    if (x->n_states() == 0) {
        cerr << "empty FST at " << __FILE__ << ':' << __LINE__ << endl;
#if 0
        expanded_fst->dump_ascii("debug.expanded_fst.fst");
        _L->dump_ascii("debug._L.fst");
#endif
        return clear(class_name);
    }
    x = FSTProject(x, FSTOutput, FSTInput);

    // Construct.
    //
    if (replace) {
        FSTGeneric empty;
        x = _construct(x, class_name, empty, false);
    } else {
        x = _construct(x, class_name, c.fst, false);
    }
    
    // Store internally for future non-replace update.
    //
    c.fst = x;

    //x->dump_binary(class_name + ".debug.fst");
    
    char buf[16];
    sprintf(buf, "%.3fs", timer.value());
    cerr << "FSTDynamicClass::update took " << buf << endl;

    return c.fst;
}

void
FSTDynamicClass::replace(FSTGeneric dynamic_fst, const string& class_name,
                         const FSTGeneric& contents_fst, bool no_optimize)
{
    FSTDynamicSet(dynamic_fst, class_name,
                  replace(class_name, contents_fst, no_optimize));
}

FSTGeneric
FSTDynamicClass::replace(const string& class_name, const FSTGeneric& contents_fst, bool no_optimize)
{
    FSTGeneric empty;
    return _construct(contents_fst, class_name, empty, no_optimize);
}

typedef pair<int,int> Pair;
typedef fst::map<Pair,int> PairMap;
typedef fst::vector<Pair> PairVector;

static int
_find_or_create_state(int state, int constraint, FSTGeneric fst, PairMap& pair_map, PairVector& pair_vector)
{
    int q = pair_map.size();
    pair<PairMap::iterator, bool> status =
        pair_map.insert(PairMap::value_type(Pair(state, constraint), q));
    if (status.second == true) {
        fst->add_state();
        pair_vector.push_back(Pair(state, constraint));
        //cerr << '(' << state << ',' << constraint << ") => " << q << endl;
    } else {
        q = (*status.first).second;
        //cerr << '(' << state << ',' << constraint << ") -> " << q << endl;
    }
    return q;
}

FSTGeneric
FSTDynamicClass::_construct(const FSTGeneric& in, const string& class_name,
                            const FSTGeneric& merge_fst, bool no_optimize)
{
    FSTGeneric x;
    if (in->is_mutable())
        x = in;
    else
        x = FSTBasic(in);

#if 0
    in->dump_ascii("debug.in.fst");
#endif

    // Prefix with $:<CLASS>.  The $ is a marker used by the augmented CoP.
    //
    int start_label = 0;
    if (class_name.length() > 0) {
        string label = "<";
        label += class_name;
        label += ">";
        start_label = FSTstring_map[label];
    }
    int new_initial = x->add_state();
    x->add_arc(new_initial, x->initial(), _dollar, start_label);
    x->set_initial(new_initial);
    
    // Suffix with $:</CLASS>.  The $ is a marker used by the augmented CoP.
    //
    int end_label = 0;
    if (class_name.length() > 0) {
        string label = "</";
        label += class_name;
        label += ">";
        end_label = FSTstring_map[label];
    }
    int new_final = x->add_state();
    for (int p = 0; p < x->n_states(); ++p) {
        if (x->is_final(p)) {
            x->add_arc(p, new_final, _dollar, end_label, x->final_weight(p));
            x->clear_final(p);
        }
    }
    x->set_final(new_final);
    
    // Compose with the augmented CoP.
    //
    // NOTE: It is possible for CoP, C in particular, to filter out paths and
    // potentially kill entire words.  The reason is that C is generally
    // constructed to rule out a list of "illegal" phone pairs.  At the moment,
    // I do not have a way to detect this, but perhaps this would be a useful
    // addition.
    //
#if 0
    x->dump_ascii("out.preCP.fst");
#endif
    {
        FSTTimer t;
        x = FSTTrim(FSTCompose(_CP, x));
        format(cerr, "time to apply CP: %.3fs\n", t.value());
    }

    if (x->n_states() == 0) {
        cerr << "empty FST at " << __FILE__ << ':' << __LINE__ << endl;
        return clear(class_name);
    }

    // If we're not replacing, we need to merge with the a pre-existing FST for
    // the class.
    //
    if (merge_fst && merge_fst->n_states() > 0) {
        x = FSTUnion(x, merge_fst);
    }

    // Optimize.
    //
    if (!no_optimize) {
        FSTTimer t;
        x = FSTBasic(FSTRemoveEpsilons(x));
        x = FSTBasic(FSTDeterminize(x));
        x = FSTMinimize(x);
        format(cerr, "time to optimize: %.3fs\n", t.value());
    }

    // Due to determinization, we could end up with a #i constraint arc followed
    // by additional arcs with pending outputs.  For example, with "lee" and
    // "leeds" in the class, we end up with "#i:lee" followed by "eps:</CLASS>".
    // Having the #i arc as non-final confuses FSTDynamic splicing.
    //
    // Delay such #i labels.
    //
    {
        //x->dump_binary("/tmp/debug.1.fst");

        PairMap pair_map;
        PairVector pair_vector;
        int p;

        pair_vector.reserve(x->n_states());

        FSTGeneric y = FSTBasic();
        pair_map[Pair(x->initial(), 0)] = p = y->add_state();
        y->set_initial(p);
        pair_vector.push_back(Pair(x->initial(), 0));

        for (unsigned i = 0; i < pair_vector.size(); i++) {
            int p = pair_vector[i].first;  // state
            int c = pair_vector[i].second; // constraint

            if (c == 0 && x->is_final(p)) {
                y->set_final(i, x->final_weight(p));
            }

            for (FSTArcIter a = x->arcs_out(p); a.valid(); a++) {
                const char* label = FSTstring_map[a->input];

                if (c == 0) {

                    if (p != x->initial() && !x->is_final(a->next) && label[0] == '#' && isdigit(label[1])) {
                        int q = _find_or_create_state(a->next, a->input, y, pair_map, pair_vector);
                        y->add_arc(i, q, 0, a->output, a->weight);
                    } else {
                        int q = _find_or_create_state(a->next, 0, y, pair_map, pair_vector);
                        y->add_arc(i, q, a->input, a->output, a->weight);
                    }

                } else {

                    assert(a->input == 0);

                    if (x->is_final(a->next)) {
                        int q = _find_or_create_state(a->next, 0, y, pair_map, pair_vector);
                        y->add_arc(i, q, c, a->output, a->weight);
                    } else {
                        int q = _find_or_create_state(a->next, c, y, pair_map, pair_vector);
                        y->add_arc(i, q, 0, a->output, a->weight);
                    }

                }

            }

        }

        //y->dump_binary("/tmp/debug.2.fst");

        x = y;
#if 0
        x = FSTMinimize(FSTBasic(FSTDeterminizeA(x)));
#endif
    }

#if 0
    x->dump_binary("debug-" + class_name + ".fst");
#endif

    return x;
}

const fst::set<int>&
FSTDynamicClass::input_alphabet() const
{
    return _CP->input_alphabet();
}

void
FSTDynamicClass::_expand_L(const FSTGeneric& expanded_fst) {
  string errorMsg;
  _expand_L(expanded_fst, errorMsg);
}

void
FSTDynamicClass::_expand_L(const FSTGeneric& expanded_fst, string& errorMsg)
{
    // Go through all unique words in expanded FST...
    //
    IntSet dict_words = _L->input_alphabet();
    const IntSet& expanded_words = expanded_fst->output_alphabet();
    fst::vector<FSTGeneric> new_dict_fsts;
    if (_l2s_fst->n_arcs() > 0) {
        for (IntSet::const_iterator si = expanded_words.begin(); si != expanded_words.end(); ++si) {
            int w = *si;
            if (!dict_words.count(w)) {
                // Word w is not known.
                
#if 0
                cerr << FSTstring_map[w] << " not in dictionary, trying ANGIE\n";
#endif
                
                // Split into letter FST.
                //
                string word = FSTstring_map[w];
                int n_letters = word.size();
                FSTGeneric letter_fst = FSTBasic();
                letter_fst->add_states(n_letters + 1);
                letter_fst->set_initial(0);
                letter_fst->set_final(n_letters);
                char letter[2]; letter[1] = '\0';
                for (unsigned i = 0; i < word.size(); ++i) {
                    letter[0] = word[i];
                    letter_fst->add_arc(i, i + 1, letter, letter);
                }
                
#if 0
                // Find 1-best pronunciation (letter:phoneme).
                //
                FSTGeneric phonemes = FSTSearch(letter_fst, _l2s_fst, 10, 0, false, NULL);
                if (phonemes->n_states() == 0) {
  		    errorMsg += "No pronunciation for: " + word + "\n";
                    cerr << "warning: " << word << " failed letter-to-sound -- skipping" << endl;
                    continue;
                }
                phonemes = FSTClearWeights(phonemes);
                phonemes = FSTProject(phonemes, FSTOutput, FSTInput);
#else
                // Find pronunciations tied with the best (letter:phoneme).
                //
                FSTGeneric phonemes = FSTTrim(FSTCompose(letter_fst, _l2s_fst));
                if (phonemes->n_states() == 0) {
		    errorMsg += "No pronunciation for: " + word + "\n";
                    cerr << "warning: " << word << " failed letter-to-sound -- skipping" << endl;
                    continue;
                }
                phonemes = FSTClearWeights(FSTAstarGraph(FSTPotential(phonemes), FSTOutput, float(.1), 0));
#endif
                
                // Transform to word:phoneme.
                //
                FSTGeneric new_entry = FSTProject(phonemes, FSTNone, FSTInput);
                for (FSTArcIter a = new_entry->arcs_out(new_entry->initial()); a.valid(); ++a)
                    a.set_input_index(w);
#if 0
                cerr << "word " << word << " not known, using..." << endl;
                FSTTopologicalSort(new_entry)->dump_ascii(cerr);
#endif
                
                // Add to new FSTs to union into dictionary.
                //
                new_dict_fsts.push_back(new_entry);
            }
        }
    }
    
    // Update dictionary _L with the union of all words generated with
    // letter-to-sound (contained within vector new_dict_fsts).
    //
    if (!new_dict_fsts.empty()) {
        FSTTimer l_timer;
        FSTGeneric L_additions = FSTMinimize(FSTDeterminizeA(FSTCache(FSTRemoveEpsilonsA(FSTUnion(new_dict_fsts)))));
        _merge(L_additions);
        format(cerr, "time to update L: %.3fs\n", l_timer.value());
#if 0
        L_additions->dump_ascii("out.Ladd.fst");
        _L->dump_ascii("out.L.fst");
#endif
    }
}
    
FSTGeneric  
FSTDynamicClass::build_static(const FSTGeneric& G)
{
    string errorMsg;
    return build_static(G, errorMsg);
}

FSTGeneric  
FSTDynamicClass::build_static(const FSTGeneric& G, string& errorMsg)
{
    // Transpose input FST because _L is words:phonemes.
    FSTGeneric x = FSTProject(G, FSTOutput, FSTInput);

    _expand_L(x, errorMsg);

    // Compute GoL.
    x = FSTTrim(FSTCompose(x, _L));
    if (x->n_states() == 0) {
        return x;
    }

    // Transpose back to LoG, phonemes:words.
    x = FSTProject(x, FSTOutput, FSTInput);

    // Add pauses at start and end.
    int hyphen = FSTstring_map["-"];

    FSTGeneric pause1 = FSTBasic();
    pause1->add_states(2);
    pause1->set_initial(0);
    pause1->set_final(1);
    pause1->add_arc(0, 0, hyphen);
    pause1->add_arc(0, 1, _pound);

    FSTGeneric pause2 = FSTBasic();
    pause2->add_states(1);
    pause2->set_initial(0);
    pause2->set_final(0);
    pause2->add_arc(0, 0, hyphen);


    x = FSTConcatenate(pause1, x, pause2);
#if 0
    x->dump_ascii("debug.C.fst");
#endif

    // Compute CoPoLoG.
    x = FSTTrim(FSTCompose(_CP, x));

    // Optimize.
    x = FSTBasic(FSTRemoveEpsilons(x));
    x = FSTBasic(FSTDeterminize(x));
    x = FSTMinimize(x);

#if 0
    x->dump_ascii("debug.static.fst");
#endif

    return x;
}
