#if 0
//
// (c) Copyright 1999
//     Massachusetts Institute of Technology
//     Spoken Language Systems Group
//     All Rights Reserved
//
#include "fst/FST_i.h"
#include <string>

class FSTExpressionParser {

    // A recursive descent parser for parsing FST expressions.  The
    // grammar is as follows and has had left recursion removed (a
    // necessity for recursive descent):
    //
    // expr
    //   : '<' filename '>' composeRest
    //   | '(' expr ')' composeRest
    //   | '{' expr '}' composeRest
    //   | cmd composeRest
    //
    // composeRest
    //   : 'o' expr
    //   | EMPTY
    //
    // cmd
    //   : command '(' expr ')'
    //
    // command
    //   : 'backoff'
    //   | 'cache'
    //   | 'det'
    //   | 'invert'
    //   | 'lru'
    //   | 'ngram_det'
    //   | 'ngram_invert'
    //   | 're'
    //   | 'reverse'
    //   | 'closure'
    //   | 'optA'
    //   | 'opt'
    //   | 'invert'
    //
    // Note that composition 'o' is right-associative due to the
    // nature of recursive descent.  That is, AoBoC = Ao(BoC).

public:
    FSTExpressionParser() {}
    ~FSTExpressionParser() {}
    bool parse(const string& input, FSTGeneric& result, string& error);

protected:
    bool _exprParse(FSTGeneric& result);
    bool _filenameParse(string& result);
    bool _cmdParse(FSTGeneric& result);
    bool _intParse(int& result);
    bool _match(char c);
    bool _match(const char* s);
    void _skip_ws();
    void _error();

protected:
    string _input_string;
    const char* _ptr;
    int _error_pos;

};

bool
FSTExpressionParser::parse(const string& input, FSTGeneric& result,
               string& error)
{
    _input_string = input;
    _ptr = _input_string.c_str();
    _error_pos = -1;

    error = "";

    _skip_ws();

    if (_exprParse(result) && *_ptr == 0) {

    // Successful parse that covered entire input.
    //
    return true;

    } else {

    // Parse failure.  Indicate position with error message.
    //
    error = "FSTExpression syntax error:\n" + input + "\n";
    while (_error_pos-- > 0)
        error += ' ';
    error += "^\n";
    return false;
    }
}

bool
FSTExpressionParser::_exprParse(FSTGeneric& result)
{
    const char* save = _ptr;
    string filename;

    if (_match('<') && _filenameParse(filename) && _match('>')) {
    result = FSTLoad(filename.c_str());
    } else if (_match('(') && _exprParse(result) && _match(')')) {
    ;
    } else if (_match('{') && _exprParse(result) && _match('}')) {
    ;
    } else if (_cmdParse(result)) {
    ;
    } else {
    _ptr = save;
    _error();
    return false;
    }

    if (_match('o')) {
    FSTGeneric rest;
    if (_exprParse(rest)) {
        result = FSTCompose(result, rest);
    } else {
        _error();
        return false;
    }
    }

    if (_match('|')) {
    FSTGeneric rest;
    if (_exprParse(rest)) {
        result = FSTUnion(result, rest);
    } else {
        _error();
        return false;
    }
    }

    if (_match('+')) {
    FSTGeneric rest;
    if (_exprParse(rest)) {
        result = FSTConcatenate(result, rest);
    } else {
        _error();
        return false;
    }
    }

    _skip_ws();
    return true;
}

bool
FSTExpressionParser::_filenameParse(string& result)
{
    result = "";
    while (*_ptr &&
       *_ptr != '<' && *_ptr != '>' &&
       *_ptr != '(' && *_ptr != ')' &&
       *_ptr != '{' && *_ptr != '}' &&
       !isspace(*_ptr)) {
    result += *_ptr++;
    }
    _skip_ws();
    return result.size() > 0;
}

bool
FSTExpressionParser::_intParse(int& result)
{
    bool success = false;
    result = 0;
    while (*_ptr && isdigit(*_ptr)) {
        success = true;
        result = 10*result + (*_ptr - '0');
        _ptr++;
    }
    _skip_ws();
    return success;
}

bool
FSTExpressionParser::_cmdParse(FSTGeneric& result)
{
    const char* save = _ptr;
    int value;

    if (_match("cache") &&
    _match('(') &&
    _exprParse(result) &&
    _match(')')) {
    result = FSTCache(result);
    } else if (_match("lru") &&
           _match('(') &&
           _exprParse(result) &&
               _match(',') &&
               _intParse(value) &&
           _match(')')) {
    result = FSTCacheLRU(result, value);
    } else if (_match("det") &&
           _match('(') &&
           _exprParse(result) &&
           _match(')')) {
    result = FSTDeterminize(result);
    } else if (_match("reverse") &&
           _match('(') &&
           _exprParse(result) &&
           _match(')')) {
    result = FSTReverse(result);
    } else if (_match("re") &&
           _match('(') &&
           _exprParse(result) &&
           _match(')')) {
    result = FSTRemoveEpsilons(result);
    } else if (_match("sort") &&
           _match('(') &&
           _exprParse(result) &&
           _match(')')) {
    result = FSTTopologicalSort(result);
    } else if (_match("closure") &&
           _match('(') &&
           _exprParse(result) &&
           _match(')')) {
    result = FSTClosure(result);
    } else if (_match("optA") &&
           _match('(') &&
           _exprParse(result) &&
           _match(')')) {
    result = FSTOptimizeA(result);
    } else if (_match("opt") &&
           _match('(') &&
           _exprParse(result) &&
           _match(')')) {
    result = FSTOptimize(result);
    } else if (_match("ngram_det") &&
           _match('(') &&
           _exprParse(result) &&
           _match(')')) {
    result = FSTNgramDeterminize(result);
    } else if (_match("ngram_invert") &&
           _match('(') &&
           _exprParse(result) &&
           _match(')')) {
    result = FSTNgramInvert(result);
    } else if (_match("invert") &&
           _match('(') &&
           _exprParse(result) &&
           _match(')')) {
    result = FSTInvert(result);
    } else {
    _ptr = save;
    _error();
    return false;
    }
    _skip_ws();
    return true;
}

bool
FSTExpressionParser::_match(char c)
{
    if (*_ptr == c) {
    _ptr++;
    _skip_ws();
    return true;
    } else {
    return false;
    }
}

bool
FSTExpressionParser::_match(const char* s)
{
    const char* save = _ptr;

    while (*_ptr && *s && *_ptr++ == *s++)
    ;
    if (*s == 0) {
    _skip_ws();
    return true;
    } else {
    _ptr = save;
    return false;
    }
}

void
FSTExpressionParser::_skip_ws()
{
    while (isspace(*_ptr))
    _ptr++;
}

void
FSTExpressionParser::_error()
{
    if (_error_pos < 0)
    _error_pos = _ptr - _input_string.c_str();
}

bool
FSTExpression(const string& input, FSTGeneric& result, string& error)
{
    FSTExpressionParser p;
    return p.parse(input, result, error);
}

FSTGeneric
FSTExpression(const string& input)
{
    FSTExpressionParser p;
    FSTGeneric result;
    string error;
    bool success = p.parse(input, result, error);

    if (!success)
    FSTFatal(error.c_str());

    return result;
}
#endif
