////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

#include <climits>
#include "fst/FST_i.h"
#include "fst/format.h"
#include <fstream>

extern time_t file_mtime(const char* filename);

#define DOT_EPSILON "@"

typedef fst::list<int> IntList;

#define NOT_SUPPORTED(method) \
    FSTFatal("%s::%s not supported", _class_name.c_str(), method);

//
// FSTGeneric implementation (rest is inline).
//

bool
FSTGeneric::reload()
{
    if (!_rep) return false;

#if 1
    bool flag = false;
    *this = _rep->reload(flag);
    return flag;
#else
    bool flag = false;
    FSTGeneric next = _rep->reload(flag);
    if (flag) {
        *this = next;
        return true;
    } else {
        return false;
    }
#endif
}

//
// FSTGenericInt implementation
//

FSTGenericInt::FSTGenericInt()
    : _class_name("FSTGeneric"),
      _semiring(0),
      _n_states(0),
      _n_arcs(0),
      _initial(-1),
      _is_mutable(false),
      _is_bidirectional(false),
      _filename(""),
      _mtime(0)
{
}

FSTGenericInt::~FSTGenericInt()
{
}

void
FSTGenericInt::enable_reload(const char* filename, time_t mtime)
{
    _filename = filename;
    _mtime = mtime;
}

FSTGeneric
FSTGenericInt::reload(bool& flag)
{
    if (_mtime != 0) {
        time_t mtime = file_mtime(_filename.c_str());
        if (mtime != _mtime) {
            cerr << "Reloading " << _filename << " because mtime "
                 << mtime << " != " << _mtime << "." << endl;
            flag = true;
            return FSTLoad(_filename);
        }
    }
    return FSTGeneric(this);
}

void
FSTGenericInt::dump_ascii(const string& filename, bool header) const
{
    dump_ascii(filename.c_str(), header);
}

void
FSTGenericInt::dump_ascii(const char* filename, bool header) const
{
    if (!filename || !filename[0] || (filename[0] == '-' && filename[1] == 0)) {
        dump_ascii(cout, header);
    } else {
        ofstream os(filename);
        if (!os)
            FSTFatal("can't write file %s", filename);
        dump_ascii(os, header);
    }
}

void
FSTGenericInt::dump_ascii(ostream& os, bool header) const
{
    int count = 0;

    if (header)
        os << "#FSTBasic " << _semiring->name << endl;

    if (initial() >= 0) {
        os << "I " << initial() << "\n";
        count++;
    }
    for (FSTStateIter s = states(); s.valid(); s++) {
        int p = *s;
        if (is_final(p)) {
            os << "F " << p;
            float weight = final_weight(p);
            if (weight != _semiring->one)
                os << " " << weight;
            os << "\n";
            count++;
        }
        float w = potential(p);
        if (w != _semiring->one) {
            os << "P " << p << " " << w << "\n";
            count++;
        }
        for (FSTArcIter a = arcs_out(p); a.valid(); a++) {
            int q = a->next;
            int input = a->input;
            int output = a->output;
            float weight = a->weight;
            // format: T <from> <to> [<input> [<output> [<weight>]]]
            os << "T " << p << " " << q;
            if (input || output || weight != _semiring->one)
                os << " " << ((input == 0) ? "," : FSTstring_map[input]);
            if (output || weight != _semiring->one)
                os << " " << ((output == 0) ? "," : FSTstring_map[output]);
            if (weight != _semiring->one)
                os << " " << weight;
            os << "\n";
            count++;
        }
    }
}

ostream&
operator<<(ostream& os, const FSTGeneric& x)
{
    // Is this correct?  Should << not dump the header?
    //
    x->dump_ascii(os, false);
    return os;
}

void
FSTGenericInt::dump_binary(const string& filename) const
{
    dump_binary(filename.c_str());
}

void
FSTGenericInt::dump_binary(const char* filename) const
{
    if (!filename || !filename[0] || (filename[0] == '-' && filename[1] == 0)) {
        dump_binary(FSTBinaryCout());
    } else {
#ifdef _MSC_VER
        ofstream os(filename, ios_base::binary);
#else
        ofstream os(filename);
#endif
        if (!os)
            FSTFatal("can't write file %s", filename);
        dump_binary(os);
    }
}

static void
dump_alphabet(const fst::set<int>& alphabet,
              fst::vector<int>& mapping, ostream& os)
{
    // Size mapping to maximum alphabet label, filling with -1.
    int max_label = (alphabet.size() == 0) ? -1 : *alphabet.rbegin();
    mapping.clear();
    mapping.resize(max_label + 1, -1);
    
    // dump alphabet size
    FSTDumpInt(alphabet.size(), os);

    // dump alphabet strings and build mapping
    int new_label = 0;
    for (fst::set<int>::const_iterator i = alphabet.begin();
         i != alphabet.end(); i++) {
        int label = *i;
        if (mapping[label] < 0)
            mapping[label] = new_label++;
        FSTDumpString(FSTstring_map[label], os);
    }
}

void
FSTGenericInt::dump_binary(ostream& os) const
{
    // header
    os << "#FSTBinary " << _semiring->name << endl;

    // arc order
    FSTDumpInt(int(FSTNone), os);

    // initial state
    FSTDumpInt(initial(), os);

    // input labels
    fst::vector<int> input_mapping;
    dump_alphabet(input_alphabet(), input_mapping, os);

    // output labels
    fst::vector<int> output_mapping;
    dump_alphabet(output_alphabet(), output_mapping, os);

    // number of arcs
    int n_arcs = 0;
    for (FSTStateIter s = states(); s.valid(); s++)
        n_arcs += arcs_out(*s).n();
    FSTDumpInt(n_arcs, os);

    // arcs
    int n_arcs_dumped = 0;
    fst::deque<int> arc_indices;
    for (FSTStateIter s = states(); s.valid(); s++) {
        int p = *s;
        arc_indices.push_back(n_arcs_dumped);
        for (FSTArcIter a = arcs_out(p); a.valid(); a++) {
            int input = input_mapping[a->input];
            int output = output_mapping[a->output];
            assert(input >= 0);
            assert(output >= 0);
            FSTDumpInt(p, os);
            FSTDumpInt(a->next, os);
            FSTDumpInt(input, os);
            FSTDumpInt(output, os);
            FSTDumpFloat(a->weight, os);
            n_arcs_dumped++;
        }
        arc_indices.push_back(n_arcs_dumped);
    }

    // # states
    int n_states_dumped = n_states();
    FSTDumpInt(n_states_dumped, os);

    // states
    for (int p = 0; p < n_states_dumped; p++) {
        FSTDumpFloat(final_weight(p), os);
        FSTDumpFloat(potential(p), os);
        FSTDumpInt(arc_indices.front(), os); arc_indices.pop_front();
        FSTDumpInt(arc_indices.front(), os); arc_indices.pop_front();
    }
}

void
FSTGenericInt::dump_mmap(const string& filename, FSTArcOrder order) const
{
    dump_mmap(filename.c_str(), order);
}

// FSTGenericInt::dump_mmap(const char*, FSTArcOrder) const is in
// FSTMmap.cpp.

static const char*
dot_label(int label)
{
    return label ? FSTstring_map[label] : DOT_EPSILON;
}

void
FSTGenericInt::dump_dot(const string& filename, const char* init) const
{
    dump_dot(filename.c_str(), init);
}

void
FSTGenericInt::dump_dot(const char* filename, const char* init) const
{
    if (!filename || !filename[0] || (filename[0] == '-' && filename[1] == 0)) {
        dump_dot(cout, init);
    } else {
        ofstream os(filename);
        if (!os)
            FSTFatal("can't write file %s", filename);
        dump_dot(os, init);
    }
}

void
FSTGenericInt::dump_dot(ostream& os, const char* init) const
{
    os << "digraph FST {\n";
    if (init)
        os << '\t' << init << ";\n";
    if (initial() >= 0)
        format(os, "\t-1 [label=\"\",peripheries=0];\n\t\t-1 -> %d;\n",
               initial());
    for (FSTStateIter s = states(); s.valid(); s++) {
        int p = *s;
        if (is_final(p)) {
            if (final_weight(p) != _semiring->one)
                format(os, "\t%d [label=\"%d/%g\",peripheries=2];\n",
                       p, p, final_weight(p));
            else
                format(os, "\t%d [peripheries=2];\n", p);
        }
        for (FSTArcIter a = arcs_out(p); a.valid(); a++) {
            format(os, "\t\t%d -> %d [label=\"%s:%s",
                   p, a->next,
                   dot_label(a->input), dot_label(a->output));
            if (a->weight != _semiring->one)
                format(os, "/%g", a->weight);
            os << "\"];\n";
        }
    }
    os << "}\n";
}

void
FSTGenericInt::trim()
{
    if (!is_mutable())
        FSTFatal("cannot trim immutable FST");

    fst::vector<int> dead;
    dead_states(dead);
    delete_states(dead);
}

// Overload this if the number of arcs out of state p can be trivially computed.
//
int
FSTGenericInt::n_arcs_out(int p) const
{
    return INT_MAX;
}

FSTArc 
FSTGenericInt::arc_out(int p) const
{
    NOT_SUPPORTED("arc_out()");

    return FSTArc();
}


// The following should be overloaded to make caching of dynamic arcs faster,
// but this does provide a backup.
//
void
FSTGenericInt::arcs_out(fst::vector<FSTArc*>& result, int p, FSTArcOrder order) const
{
    FSTArcIter i = arcs_out(p, order);
    result.clear();
    result.reserve(i.n());
    for (; i.valid(); i++)
        result.push_back(new FSTArc(*i));
}

//
// Determining dead states (states not on path from initial to final).
//

void
FSTGenericInt::dead_states(fst::vector<int>& dead) const
{
    FSTGeneric g(const_cast<FSTGenericInt*>(this));
    fst::vector<int> order;
    if (FSTTopologicalSortFromInitial(order, g)) {
        _dead_states_acyclic(dead, order);
    } else {
        _dead_states_cyclic(dead);
    }
}

void
FSTGenericInt::_dead_states_acyclic(fst::vector<int>& dead, const fst::vector<int>& order) const
{
    fst::vector<bool> keep(_n_states, false);
    for (fst::vector<int>::const_reverse_iterator i = order.rbegin(); i != order.rend(); ++i) {
        int p = *i;
        if (is_final(p)) {
            keep[p] = true;
        } else {
            for (FSTArcIter a = arcs_out(p); a.valid(); ++a) {
                int q = a->next;
                if (keep[q]) {
                    keep[p] = true;
                    break;
                }
            }
        }
    }

    for (int p = 0; p < _n_states; ++p) {
        if (!keep[p]) {
            dead.push_back(p);
        }
    }
}

void
FSTGenericInt::_dead_states_cyclic(fst::vector<int>& dead) const
{
    typedef fst::vector<char> CharVector;
    typedef fst::set<int> IntSet;
    typedef fst::vector<IntSet> IntSetVector;
    typedef IntSet::const_iterator IntSetIter;
    typedef pair<IntSetIter, IntSetIter> IntSetIterPair;

    static const char unvisited = 0;
    static const char forward   = 1;
    static const char keep      = 2;

    force();

    CharVector status(_n_states, unvisited);
    IntSetVector adjacency(_n_states);
    IntSet final;

    // Forward Pass: a depth-first traversal, marking states reachable from
    // initial state with status "forward", building up backwards adjacency
    // lists, and building up the set of final states.
    //
    if (_initial >= 0) {
        status[_initial] = forward;
        if (is_final(_initial))
            final.insert(_initial);
        stack<FSTArcIter> todo;
        todo.push(arcs_out(_initial));
        while (!todo.empty()) {
            FSTArcIter& i = todo.top();
            if (i.valid()) {
                status[i->prev] = forward;
                adjacency[i->next].insert(i->prev);
                if (status[i->next] == unvisited) {
                    status[i->next] = forward;
                    if (is_final(i->next))
                        final.insert(i->next);
                    todo.push(arcs_out(i->next));
                }
                ++i;
            } else {
                todo.pop();
            }
        }
    }

    // Backward Pass: a depth-first traversal in reverse, marking states
    // reachable from the initial state and any final state with status "keep".
    //
    if (_initial >= 0) {
        stack<IntSetIterPair> todo;
        for (IntSetIter fi = final.begin(); fi != final.end(); ++fi) {
            if (status[*fi] == forward) {
                status[*fi] = keep;
                todo.push(IntSetIterPair(adjacency[*fi].begin(), adjacency[*fi].end()));
            }
        }
        while (!todo.empty()) {
            IntSetIterPair& ip = todo.top();
            if (ip.first != ip.second) {
                int p = *ip.first;
                if (status[p] == forward) {
                    status[p] = keep;
                    todo.push(IntSetIterPair(adjacency[p].begin(), adjacency[p].end()));
                }
                ++ip.first;
            } else {
                todo.pop();
            }
        }
    }

    // Build up result vector of dead states.
    //
    dead.clear();
    for (int i = 0; i < _n_states; i++) {
        if (status[i] != keep)
            dead.push_back(i);
    }
}

//
// For the mutable and bidirectional methods, we provide stubs that cause
// run-time errors.
//

void
FSTGenericInt::set_initial(int)
{
    NOT_SUPPORTED("set_initial");
}

void
FSTGenericInt::set_final(int, float)
{
    NOT_SUPPORTED("set_final");
}

void
FSTGenericInt::set_potential(int, float)
{
    NOT_SUPPORTED("set_potential");
}

int
FSTGenericInt::add_states(int)
{
    NOT_SUPPORTED("add_states");
    return -1;
}

void
FSTGenericInt::add_arc(int, int, int, int, float)
{
    NOT_SUPPORTED("add_arc");
}

void
FSTGenericInt::add_arcs(int, fst::vector<FSTArc*>&, FSTArcOrder)
{
    NOT_SUPPORTED("add_arcs");
}

void
FSTGenericInt::set_arcs(int, fst::vector<FSTArc*>&, FSTArcOrder)
{
    NOT_SUPPORTED("set_arcs");
}

void
FSTGenericInt::delete_state(int)
{
    NOT_SUPPORTED("delete_state");
}

void
FSTGenericInt::delete_states(const fst::vector<int>&)
{
    NOT_SUPPORTED("delete_states");
}

FSTFinalStateIter
FSTGenericInt::final_states() const
{
    NOT_SUPPORTED("final_states");
    return FSTFinalStateIter();
}

FSTArcIter
FSTGenericInt::arcs_in(int, FSTArcOrder) const
{
    NOT_SUPPORTED("arcs_in");
    return FSTArcIter();
}

FSTArcIter
FSTGenericInt::arcs_in_with_input(int, int, FSTArcMode) const
{
    NOT_SUPPORTED("arcs_in_with_input");
    return FSTArcIter();
}

FSTArcIter
FSTGenericInt::arcs_in_with_output(int, int, FSTArcMode) const
{
    NOT_SUPPORTED("arcs_in_with_output");
    return FSTArcIter();
}

void
FSTGenericInt::_sort_arcs(fst::vector<FSTArc*>& arcs, FSTArcOrder order)
{
    FSTArcCompareInput input_compare;
    FSTArcCompareOutput output_compare;
    switch (order) {
    case FSTInput:
        sort(arcs.begin(), arcs.end(), input_compare);
        break;
    case FSTOutput:
        sort(arcs.begin(), arcs.end(), output_compare);
        break;
    default:
        break;
    }
}

void
FSTGenericInt::_train_begin()
{
    NOT_SUPPORTED("_train_begin()");
}

void
FSTGenericInt::_train_begin_left()
{
    NOT_SUPPORTED("_train_begin_left()");
}

void
FSTGenericInt::_train_begin_right()
{
    NOT_SUPPORTED("_train_begin_right()");
}

void
FSTGenericInt::_train_end()
{
    NOT_SUPPORTED("_train_end()");
}

void
FSTGenericInt::_train_normalize()
{
    NOT_SUPPORTED("_train_normalize()");
}

void
FSTGenericInt::_train_normalize_by_outputlabel()
{
    NOT_SUPPORTED("_train_normalize_by_outputlabel()");
}

void
FSTGenericInt::_train_update(fst::vector<double>&, fst::vector<double>&, double)
{
    NOT_SUPPORTED("_train_update()");
}

void
FSTGenericInt::restore_input_output_index()
{
    NOT_SUPPORTED("restore_input_output_index()");
}

void
FSTGenericInt::set_semiring(FSTSemiring semiring)
{
    _semiring = semiring;
}

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 80
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
