////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

// EM Training of arc and final weights.  The type-specific
// functionality is implemented by the _train_* virtual methods.
//

//#define FSTGenericTrain_DEBUG

#include <cstring>
#include "fst/FST_i.h"
#include "fst/format.h"
#include <cmath>
#include <cfloat>
#include <sstream>
#include <fstream>

using namespace std;

// For the time being we'll do all weight arithmetic through these macros.
//
#define ONE     double(1)
#define ZERO        double(0)
#define PLUS(x,y)   (double(x) + double(y))
#define TIMES(x,y)  (double(x) * double(y))
#define TIMES3(x,y,z)   (double(x) * double(y) * double(z))
#define DIVIDE(x,y) (double(x) / double(y))
#define IS_BETTER(x,y)  (double(x) > double(y))
#define IS_NOT_TOO_SMALL(x) (double(x) > (DBL_MIN * 1.0e10))

// Structures to build graph representing all parses of a sequence.  We build
// up a special graph instead of just using an FST because we need to store
// additional information for each state and arc for training purposes.

struct FSTTrainState;

struct FSTTrainArc {
    FSTTrainState* next;
    FSTTrainState* prev;
    double weight;
    int index;                    // fst arc index

    FSTTrainArc(FSTTrainState* n, FSTTrainState* p, double w, int i)
    : next(n), prev(p), weight(w), index(i) {}

    FSTAllocator_new_delete(FSTTrainArc);
};

typedef fst::vector<FSTTrainArc*> FSTTrainArcs;
typedef fst::vector<FSTGeneric> FSTGenerics;

struct FSTTrainState {
    FSTTrainArcs arcs;
    double alpha;
    double beta;
    bool is_final;
    double final;
    int index;                    // fst state index

    FSTTrainState(bool i_f, double f, int i)
    : is_final(i_f), final(f), index(i) {}

    FSTAllocator_new_delete(FSTTrainState);
};

typedef fst::vector<FSTTrainState*> FSTTrainParse;
typedef fst::vector<FSTTrainParse*> FSTTrainParses;

static void
Free(FSTTrainState* state)
{
    FSTTrainArcs::iterator i;
    for (i = state->arcs.begin(); i != state->arcs.end(); i++)
    delete *i;
    delete state;
}

static void
Free(FSTTrainParse* parse)
{
    FSTTrainParse::iterator i;
    for (i = parse->begin(); i != parse->end(); i++)
    Free(*i);
    delete parse;
}

static void
Free(FSTTrainParses& parses)
{
    FSTTrainParses::iterator i;
    for (i = parses.begin(); i != parses.end(); i++)
    Free(*i);
}

void 
describe_vector_double(const char* filename, const fst::vector<double>& vd) 
{
    int p, n;
    ofstream os(filename);

    n = vd.size();
    for (p = 0; p < n; p++) {
    os << p << ": " << vd[p] << endl;
    }
}

void 
describe_parses(const char* filename, FSTTrainParses& parses)
{
    FSTTrainParses::iterator pi;
    FSTTrainParse::iterator si;
    FSTTrainArcs::iterator ai;
    FSTTrainParse* parse;
    FSTTrainState* state;
    FSTTrainArc* arc;
    int p_index, s_index, a_index;

    ofstream os(filename);

    p_index = 0;
    for (pi = parses.begin(); pi != parses.end(); pi++) {
    parse = *pi;
    s_index = 0;
    os << "parse: " << p_index << endl;
    for (si = parse->begin(); si != parse->end(); si++) {
        state = *si;
        if (state->is_final) {
        os << "\tstate: " << s_index << " i: " << state->index; 
        os << " a: " << state->alpha << " b: " << state->beta;
        os << " F: " << state->final << endl; 
        } else {    
        os << "\tstate: " << s_index << " i: " << state->index; 
        os << " a: " << state->alpha << " b: " << state->beta << endl;
        }
        a_index = 0;
        for (ai = state->arcs.begin(); ai != state->arcs.end(); ai++) {
        arc = *ai;
        os << "\t\t\tarc: " << a_index << " i: " << arc->index; 
        os << " w: " << arc->weight;
        os << " pa: " << arc->prev->alpha << " pb: " << arc->prev->beta;
        os << " na: " << arc->next->alpha << " nb: " << arc->next->beta << endl;
        a_index++;
        }
        s_index++;
    }
    p_index++;
    }

}

void
describe_fst_with_arc_index(const char* filename, FSTGeneric arcindex_fst, 
                bool input_arcindex, bool output_arcindex, bool header = true) 
{
    int count = 0;

    ofstream os(filename);
    if (!os)
    FSTFatal("can't write file %s", filename);

    if (header)
        os << "#FSTBasic " << arcindex_fst->semiring()->name << endl;

    if (arcindex_fst->initial() >= 0) {
        os << "I " <<  arcindex_fst->initial() << endl;
        count++;
    }
    for (int p = 0; p < arcindex_fst->n_states(); p++) {
        if (arcindex_fst->is_final(p)) {
            os << "F " << p;
            float weight =  arcindex_fst->final_weight(p);
            if (weight != arcindex_fst->semiring()->one) 
                format(os, " %g", weight);
            os << "\n";
            count++;
        }
        float w = arcindex_fst->potential(p);
    // should this be semiring()->zero or semiring()->one ???
        if (w != arcindex_fst->semiring()->one) {
            os << "P " << p << " " << w << "\n";
            count++;
        }
        for (FSTArcIter a = arcindex_fst->arcs_out(p); a.valid(); a++) {
            int q = a->next;
            int input = a->input;
            int output = a->output;
            float weight = a->weight;
            // format: T <from> <to> [<input> [<output> [<weight>]]]
            os << "T " << p << " " << q;
        os << " ";
	if (input_arcindex || input || output ||
	    weight != arcindex_fst->semiring()->one) {
	    if (input_arcindex)
		os << input;
	    else if (input == 0) 
		os << ",";
	    else 
		os << FSTstring_map[input];
	}
	if (output_arcindex || output || 
	    weight != arcindex_fst->semiring()->one){
	    if (output_arcindex)
		os << " " << output;
	    else if (output == 0) 
		os << " ,";
	    else 
		os << " " << FSTstring_map[output];
            if (weight != arcindex_fst->semiring()->one)
                format(os, " %g", weight);
        }
	os << "\n";
	count++;
	}
    }
}

static bool
AlphaBeta(FSTTrainParse* parse)
{
    FSTTrainParse::iterator i;
    FSTTrainParse::reverse_iterator j;
    FSTTrainArcs::iterator a;

    for (i = parse->begin(); i != parse->end(); i++) {
    (*i)->alpha = ZERO;
    (*i)->beta = (*i)->final;
    }
    parse->front()->alpha = ONE;

    // Compute alphas walking in topological order.
    //
    for (i = parse->begin(); i != parse->end(); i++) {
    FSTTrainArcs& arcs = (*i)->arcs;
    for (a = arcs.begin(); a != arcs.end(); a++) {
        if ((*a)->next == *i)
        FSTFatal("self loop detected in parse");
        double w = TIMES((*i)->alpha, (*a)->weight);
        (*a)->next->alpha = PLUS((*a)->next->alpha, w);
    }
    }

    // Compute betas walking in reverse topological order.
    //
    for (j = parse->rbegin(); j != parse->rend(); j++) {
    FSTTrainArcs& arcs = (*j)->arcs;
    for (a = arcs.begin(); a != arcs.end(); a++) {
        double w = TIMES((*a)->weight, (*a)->next->beta);
        (*j)->beta = PLUS((*j)->beta, w);
    }
    }

    // Divide all alphas by beta(initial).  If beta(initial) is zero, meaning
    // that there is no non-zero parse, then set all alphas to zero instead so
    // this input will not contribute to any counts.
    //
    double beta_initial = parse->front()->beta;
    double factor = ZERO;
    if (IS_BETTER(beta_initial, ZERO))
    if (IS_NOT_TOO_SMALL(beta_initial))  // checking for underflow!!!!
        factor = DIVIDE(ONE, beta_initial);
    for (i = parse->begin(); i != parse->end(); i++) {
    (*i)->alpha = TIMES((*i)->alpha, factor);
    }

    return factor != ZERO;
}

static double
AlphaBeta(FSTTrainParses& parses)
{
    double sum = 0;
    int n = 0, np=0;
    

    FSTTrainParses::iterator i;
    for (i = parses.begin(); i != parses.end(); i++) {
    if (AlphaBeta(*i)) {
        sum += log((*i)->front()->beta);
        n++;
        np++;
    } else {
        cerr << "\tWARNING: ignoring " << np << "th parse" << endl;
        np++;
    }
    }
    if (n > 0)
    sum /= n;
    return sum;
}

static void
Reset(fst::vector<double>& a_counts, fst::vector<double>& f_counts, double value)
{
    fill(a_counts.begin(), a_counts.end(), value);
    fill(f_counts.begin(), f_counts.end(), value);
}

static void
Increment(fst::vector<double>& a_counts, fst::vector<double>& f_counts,
      FSTTrainParses& parses)
{
    FSTTrainParses::iterator pi;
    FSTTrainParse::iterator si;
    FSTTrainArcs::iterator ai;
    FSTTrainParse* parse;
    FSTTrainState* state;
    FSTTrainArc* arc;
    double w;
    
    for (pi = parses.begin(); pi != parses.end(); pi++) {
    parse = *pi;
    for (si = parse->begin(); si != parse->end(); si++) {
        state = *si;
        if (state->is_final) {
        w = TIMES(state->alpha, state->beta);
        f_counts[state->index] = PLUS(f_counts[state->index], w);
        }
        for (ai = state->arcs.begin(); ai != state->arcs.end(); ai++) {
        arc = *ai;
        w = TIMES3(arc->prev->alpha, arc->next->beta, arc->weight);
        a_counts[arc->index] = PLUS(a_counts[arc->index], w);
        }
    }
    }
}

static void
Update(const fst::vector<double>& a_counts,
       const fst::vector<double>& f_counts,
       FSTTrainParses parses)
{
    FSTTrainParses::iterator pi;
    FSTTrainParse::iterator si;
    FSTTrainArcs::iterator ai;
    FSTTrainParse* parse;
    FSTTrainState* state;
    FSTTrainArc* arc;
    
    for (pi = parses.begin(); pi != parses.end(); pi++) {
    parse = *pi;
    for (si = parse->begin(); si != parse->end(); si++) {
        state = *si;
        if (state->is_final) {
        state->final = f_counts[state->index];
        }
        for (ai = state->arcs.begin(); ai != state->arcs.end(); ai++) {
        arc = *ai;
        arc->weight = a_counts[arc->index];
        }
    }
    }
}

#ifdef FSTGenericTrain_DEBUG
static void
Debug(FSTGeneric x)
{
    cerr.form("I %d", x->initial()) << endl;
    for (int p = 0; p < x->n_states(); p++) {
    if (x->is_final(p))
        cerr.form("F %d %g", p, x->final_weight(p)) << endl;
    for (FSTArcIter a = x->arcs_out(p); a.valid(); a++)
        cerr.form("T %d %d %s %g", p, a->next, FSTstring_map[a->input],
              a->weight) << endl;
    }
}
#endif // FSTGenericTrain_DEBUG

FSTGeneric
Stringfst(istringstream& ls, FSTSemiring semiring_type, 
      bool add_pound = false, int pound_index = 0)
{
    FSTGeneric x = FSTBasic(semiring_type);

    string word;

    fst::vector<int> words;
    while (ls >> word) {
    words.push_back(FSTstring_map[word.c_str()]);
    }
    
#ifdef FSTGenericTrain_DEBUG
    if (0) {
    cerr << "words: "; 
    for (int i = 0; i < words.size(); i++) {
        cerr << words[i] << " ";
    }
    cerr << endl;
    }
#endif // FSTGenericTrain_DEBUG
    
    // Construct FST x from input sequence.
    //
    x->add_states(words.size() + 1);
    if (add_pound) {
    for (int i = 0; i < int(words.size()); i++) {
        x->add_arc(i, i + 1, words[i], words[i]);
        x->add_arc(i, i, pound_index, pound_index);
    }
    } else {
    for (int i = 0; i < int(words.size()); i++) {
        x->add_arc(i, i + 1, words[i], words[i]);
    }
    }
    x->set_initial(0);
    x->set_final(words.size());
    
    return x;

}

// compute Xl = x o A o Bl  
// compute Yr = Br o C o y
// then z = Xl o Yr
FSTGeneric
Parses_twosides_onesample_Binputoutput(FSTGeneric Bl, FSTGeneric Br, FSTGeneric Xl, FSTGeneric Yr)
{
    
    // compute Xl o Bl;
    Xl = FSTTrim(FSTCompose(Xl, Bl));
#ifdef FSTGenericTrain_DEBUG
    describe_fst_with_arc_index("Xl.fst", Xl, false, true);
#endif // FSTGenericTrain_DEBUG
    Xl = FSTProject(Xl, FSTOutput, FSTOutput);
    Xl = FSTBasic(FSTRemoveEpsilonsA(Xl));
#ifdef FSTGenericTrain_DEBUG
    describe_fst_with_arc_index("Xla.fst", Xl, true, true);
#endif // FSTGenericTrain_DEBUG

    // compute Br o Yr; 
    Yr = FSTTrim(FSTCompose(Br, Yr));
#ifdef FSTGenericTrain_DEBUG
    describe_fst_with_arc_index("Yr.fst", Yr, true, false);
#endif // FSTGenericTrain_DEBUG
    Yr = FSTProject(Yr, FSTInput, FSTInput);
    Yr = FSTBasic(FSTRemoveEpsilonsA(Yr));
#ifdef FSTGenericTrain_DEBUG
    describe_fst_with_arc_index("Yra.fst", Yr, true, true);
#endif // FSTGenericTrain_DEBUG

    return FSTTrim(FSTCompose(Xl, Yr)); 
}

// compute Xl = x o A o Bl  
// compute Yr = Br o C o y
// then z = Xl o Yr
FSTGeneric
Parses_twosides_onesample(FSTGeneric Bl, FSTGeneric Br, FSTGeneric A, FSTGeneric C, 
              istringstream& left_ls, istringstream& right_ls)
{
    int add_pound = true;
    int pound_index = FSTstring_map["#"];
    
    FSTGeneric Xl = Stringfst(left_ls, Bl->semiring(), add_pound, pound_index);
    FSTGeneric Yr = Stringfst(right_ls, Bl->semiring(), false);
#ifdef FSTGenericTrain_DEBUG
    Xl->dump_ascii("x.fst");
    Yr->dump_ascii("y.fst");
#endif // FSTGenericTrain_DEBUG

    if (A->n_states() > 0) {
    // compute Xl = x o A
    Xl = FSTTrim(FSTCompose(Xl, A));
#ifdef FSTGenericTrain_DEBUG
    describe_fst_with_arc_index("xA.fst", Xl, false, false);
#endif // FSTGenericTrain_DEBUG
    }
    
    if (C->n_states() > 0) {
    // compute Yr = C o y;
    Yr = FSTTrim(FSTCompose(C, Yr));
#ifdef FSTGenericTrain_DEBUG
    describe_fst_with_arc_index("Cy.fst", Yr, false, false);
#endif // FSTGenericTrain_DEBUG
    }

    return Parses_twosides_onesample_Binputoutput(Bl, Br, Xl, Yr);
}

void
Parses_twosides(FSTTrainParses& parses, 
        FSTGeneric Bl, FSTGeneric Br, FSTGeneric A, FSTGeneric C, 
        istream& left_is, istream& right_is,
        fst::vector<double>& a_counts, fst::vector<double>& f_counts, 
        const char* save_stats_dir, const char* load_stats_dir)
{
    int i, p, N;
    int max_a = 0, max_f = 0;
    
    string left_line, right_line;
    string left_tag, right_tag;
    int num_processed = 0;
    int next_threshold = 100;

    FSTGeneric z;

    if (strlen(load_stats_dir) > 0) 
    cerr << "loading from: " << load_stats_dir << endl;

    while (getline(left_is, left_line)) {
    left_tag = left_line.substr(left_line.rfind('(')+1,
                    left_line.rfind(')')-left_line.rfind('(')-1);
    left_line = left_line.substr(0, left_line.rfind('('));
    left_line = left_line.substr(0, left_line.find_last_not_of(' ')+1);
    istringstream left_ls(left_line.c_str());

    assert(getline(right_is, right_line));
    right_tag = right_line.substr(right_line.rfind('(')+1,
                    right_line.rfind(')')-right_line.rfind('(')-1);
    right_line = right_line.substr(0, right_line.rfind('('));
    right_line = right_line.substr(0, right_line.find_last_not_of(' ')+1);
    istringstream right_ls(right_line.c_str());

    // cerr << "tag: " << left_tag << "|" << right_tag << endl;
    // cerr << "line: " << left_line << "|" << right_line << endl;

    assert(left_tag == right_tag);

    if (strlen(load_stats_dir) > 0) {
        char zfile_name[200];
        sprintf(zfile_name, "%s/%s.fst", load_stats_dir, left_tag.c_str());
#ifdef FSTGenericTrain_DEBUG
        cerr << "loading stats from: " << zfile_name << endl;
#endif // FSTGenericTrain_DEBUG
        z = FSTBasic(FSTLoad(zfile_name));
        z->restore_input_output_index();
    } else {
        z = Parses_twosides_onesample(Bl, Br, A, C, left_ls, right_ls);
    }


#ifdef FSTGenericTrain_DEBUG
    describe_fst_with_arc_index("z.fst", z, true, true);
#endif // FSTGenericTrain_DEBUG
    if (strlen(save_stats_dir) > 0) {
        char zfile_name[200];
        sprintf(zfile_name, "%s/%s.fst", save_stats_dir, left_tag.c_str());
        describe_fst_with_arc_index(zfile_name, z, true, true);
    }

    N = z->n_states();
    
    num_processed++;
    if (N == 0) {
        cerr << "WARNING: parse empty for " << num_processed << "th input tag: " << left_tag << endl;
        continue;
    }

    // build the state_map 
    fst::vector<int> state_map(N, -1);
    
    for (int state_i = 0; state_i < N; state_i++) {
        FSTArcIter z_arcs = z->arcs_out(state_i);
        for (; z_arcs.valid(); z_arcs++) {  
        int z_arcindex = z_arcs.input();
        int z_prev_state = z_arcs.prev();
        int z_next_state = z_arcs.next();
        int Bl_prev_state = (Bl->arc_out(z_arcindex)).prev;
        int Bl_next_state = (Bl->arc_out(z_arcindex)).next;

        // assert that if every state in z would only map to one state in Bl
        if (state_map[z_prev_state] == -1) 
            state_map[z_prev_state] = Bl_prev_state;
        else 
            assert(state_map[z_prev_state] == Bl_prev_state);
        if (state_map[z_next_state] == -1) 
            state_map[z_next_state] = Bl_next_state;
        else 
            assert(state_map[z_next_state] == Bl_next_state);
        
        // optional: 
        // make sure all the final states in z are also final states
        if (z->is_final(z_prev_state) && !Bl->is_final(Bl_prev_state)) {
            cerr << "z(" << z_prev_state << ") is final ";
            cerr << "and Bl(" << Bl_prev_state << ") is not." << endl;
        }
        if (z->is_final(z_next_state) && !Bl->is_final(Bl_next_state)) {
            cerr << "z(" << z_prev_state << ") is final ";
            cerr << "and Bl(" << Bl_prev_state << ") is not." << endl;
        }
        }
    }
    for (p = 0; p < N; p++) {
        assert(state_map[p] != -1);
    }
    // Build training representation.
    //
    FSTTrainParse* states_ptr = new FSTTrainParse(N);
    FSTTrainParse& states = *states_ptr;
    for (p = 0; p < N; p++) {
        states[p] = new FSTTrainState(z->is_final(p), z->final_weight(p),
                      state_map[p]);
        if (state_map[p] > max_f)
        max_f = state_map[p];
    }
    for (p = 0; p < N; p++) {
        FSTArcIter a = z->arcs_out(p);
        FSTTrainState* s = states[p];
        s->arcs.reserve(a.n());
        for (; a.valid(); a++) {
        if (IS_BETTER(a->weight, ZERO)) {
            s->arcs.push_back(new FSTTrainArc(states[a->next],
                              states[p],
                              a->weight, a->output));
            if (a->output > max_a)
            max_a = a->output;
        }
        }
    }

    // Topologically sort, reordering states.
    //
    fst::vector<int> order;
    if (!FSTTopologicalSort(order, z))
        FSTFatal("cannot topologically sort (sequence o fst)");
    FSTTrainParse temp_states(N);
    for (i = 0; i < N; i++)
        temp_states[i] = states[order[i]];
#ifdef _MSC_VER
        states = temp_states;
#else
    states.swap(temp_states);
#endif
    
    if (states_ptr) 
        parses.push_back(states_ptr);

    if (num_processed >= next_threshold) {
        cerr << "\t" << num_processed << " done" << endl;
        //next_threshold += 5;
        next_threshold *= 2;
    }

    }
    
    // making sure 
    assert(!getline(right_is, right_line));

    cerr << "\t" << num_processed << " ALL DONE" << endl;
    
    // Size counts appropriately.  Until this point, we don't know the proper
    // size if fst is dynamic.
    //
    a_counts.clear();
    a_counts.insert(a_counts.end(), max_a + 1, ZERO);
    f_counts.clear();
    f_counts.insert(f_counts.end(), max_f + 1, ZERO);
}

void
FSTGenericInt::train_twosided(istream& left_is, istream& right_is, 
                  const char* Ain_file, const char* Cin_file, double floor, 
                  const char* save_stats_dir, const char* load_stats_dir)
{
    int iteration_stop = 1000;
    double avg_log_likelihood_stop = 1e-2;
    const string None = "None";
    FSTGeneric A, C;

    // load in A and C;
    if (Ain_file != None) {
    A = FSTLoad(Ain_file);
    if (A->n_states() == 0) {
        cerr << "A empty" << endl;
        exit(1);
    }
    } else {
    A = FSTBasic();
    }
        
    if (Cin_file != None) {
    C = FSTLoad(Cin_file);
    if (C->n_states() == 0) {
        cerr << "C empty" << endl;
        exit(1);
    }
    } else {
    C = FSTBasic();
    }
    
    // First, we must ready the FST for training.  This resets a possibly
    // dynamic FST, puts it into training mode, and temporarily converts its
    // weights to the PlusTimes semiring.
    //
    reset();

    _train_begin_right();
    //_train_normalize();
    // Alias for this FST.
    //
    FSTGeneric Br(this);
#ifdef FSTGenericTrain_DEBUG
    describe_fst_with_arc_index("Br.fst", Br, true, false);
#endif // FSTGenericTrain_DEBUG
    Br = FSTClearWeights(Br);

    _train_begin_left();
    FSTGeneric Bl(this);
#ifdef FSTGenericTrain_DEBUG
    describe_fst_with_arc_index("Bl.fst", Bl, false, true);
#endif // FSTGenericTrain_DEBUG

#ifdef FSTGenericTrain_DEBUG
    describe_fst_with_arc_index("A.PlusTimes.fst", FSTConvert_to_PlusTimes(A), false, false);
    describe_fst_with_arc_index("C.PlusTimes.fst", FSTConvert_to_PlusTimes(C), false, false);
#endif // FSTGenericTrain_DEBUG



    // NOW build up parses with XAl, YAr, and Bl
    cerr << "Building Parses from both sides..." << endl;
    cerr.flush();
    FSTTrainParses parses;
    fst::vector<double> a_counts, f_counts;
    Parses_twosides(parses, Bl, Br, FSTConvert_to_PlusTimes(A), FSTConvert_to_PlusTimes(C), 
            left_is, right_is, a_counts, f_counts, save_stats_dir, load_stats_dir);
    cerr << "\tDONE (Parses from both sides)" << endl;  // Parses DONE

    if (strlen(save_stats_dir) > 0) {
    cerr << endl << "Saved parses to " << save_stats_dir << endl 
         << "Skipping training... ALL DONE" << endl;  // Parses DONE
    return;
    }

#ifdef FSTGenericTrain_DEBUG
    describe_parses("itr.-1.parses", parses);
    char Bout_fn[200];
    sprintf(Bout_fn, "postincre.%d.acounts", -1);
    describe_vector_double(Bout_fn, a_counts); 
    sprintf(Bout_fn, "postincre.%d.fcounts", -1);
    describe_vector_double(Bout_fn, f_counts); 
#endif // FSTGenericTrain_DEBUG 

    // Iterate training.
    //
    cerr << "EM Training Weights..." << endl;
    cerr.flush();
    double last_avg_log_likelihood = -1e30;
    for (int iteration = 0; iteration < iteration_stop; iteration++) {

    double avg_log_likelihood = AlphaBeta(parses);

    format(cerr, "%4d: %.6f", iteration, avg_log_likelihood);
    cerr << endl;

    if (fabs(avg_log_likelihood - last_avg_log_likelihood) <
        avg_log_likelihood_stop) {
        break;
    }
    last_avg_log_likelihood = avg_log_likelihood;

    Reset(a_counts, f_counts, ZERO);

#ifdef FSTGenericTrain_DEBUG
    char Bout_fn[200];
    sprintf(Bout_fn, "prenorm.%d.parses", iteration);
    describe_parses(Bout_fn, parses);
    sprintf(Bout_fn, "B.prenorm.%d.fst", iteration);
    describe_fst_with_arc_index(Bout_fn, Bl, false, true);
#endif // FSTGenericTrain_DEBUG

    // compute a_counts and f_counts from the parses
    Increment(a_counts, f_counts, parses);
#ifdef FSTGenericTrain_DEBUG
    sprintf(Bout_fn, "postincre.%d.acounts", iteration);
    describe_vector_double(Bout_fn, a_counts); 
    sprintf(Bout_fn, "postincre.%d.fcounts", iteration);
    describe_vector_double(Bout_fn, f_counts); 
#endif // FSTGenericTrain_DEBUG 

    // Update fst weights from counts, normalize them, and store the
    // weights back in the counts.
    //
    _train_update(a_counts, f_counts, floor);

#ifdef FSTGenericTrain_DEBUG
    sprintf(Bout_fn, "B.postnorm.%d.fst", iteration);
    describe_fst_with_arc_index(Bout_fn, Bl, false, true);
#endif // FSTGenericTrain_DEBUG

    // put the a_counts and f_counts into parses
    Update(a_counts, f_counts, parses);

#ifdef FSTGenericTrain_DEBUG
    sprintf(Bout_fn, "postnorm.%d.parses", iteration);
    describe_parses(Bout_fn, parses);
#endif // FSTGenericTrain_DEBUG

    }

    // All finished.
    //
    Free(parses);
    reset();
    _train_end();
    cerr << " DONE" << endl;  // EM training

#ifdef FSTGenericTrain_DEBUG
    describe_fst_with_arc_index("Bl3.fst", Bl, false, false);
#endif // FSTGenericTrain_DEBUG
    
}

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 50
// fill-column: 79
// c-basic-offset: 4
// End:



