////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

#include "fst/FST_i.h"

class FSTInvertRep
    : public FSTGenericInt {

public: // methods
    FSTInvertRep(const FSTGeneric& in);
    virtual ~FSTInvertRep();
    virtual const fst::set<int>& input_alphabet() const;
    virtual const fst::set<int>& output_alphabet() const;
    virtual bool is_final(int p) const;
    virtual float final_weight(int p) const;
    virtual float potential(int p) const;
    virtual int n_arcs_out(int p) const;
    virtual FSTArcIter arcs_out(int p, FSTArcOrder order=FSTNone) const;
    virtual FSTArcIter arcs_out_with_input(int p, int i,
                                           FSTArcMode m=FSTEqual) const;
    virtual FSTArcIter arcs_out_with_output(int p, int o,
                                            FSTArcMode m=FSTEqual) const;
    virtual void force() const;
    virtual void reset() const;

protected: // members
    FSTGeneric _in;

protected: // members
    FSTArcIter _invert_arcs(int p, FSTArcIter a) const;
};

FSTGeneric
FSTInvert(const FSTGeneric& x)
{
    return FSTGeneric(new FSTInvertRep(x));
}

FSTInvertRep::FSTInvertRep(const FSTGeneric& in)
    : _in(in)
{
    _class_name = "FSTInvert";
    _semiring = _in->semiring();
    _initial = _in->initial();
    _n_states = _in->n_states();
}

FSTInvertRep::~FSTInvertRep()
{
}

const fst::set<int>&
FSTInvertRep::input_alphabet() const
{
    return _in->input_alphabet();
}

const fst::set<int>&
FSTInvertRep::output_alphabet() const
{
    return _in->output_alphabet();
}

bool
FSTInvertRep::is_final(int p) const
{
    return _in->is_final(p);
}

float
FSTInvertRep::final_weight(int p) const
{
    return _semiring->invert(_in->final_weight(p));
}

float
FSTInvertRep::potential(int p) const
{
    return _in->potential(p);
}

int
FSTInvertRep::n_arcs_out(int p) const
{
    return _in->n_arcs_out(p);
}

FSTArcIter
FSTInvertRep::_invert_arcs(int p, FSTArcIter a) const
{
    _n_states = _in->n_states();
    FSTArcIterData* out_arc_data = new FSTArcIterData();
    fst::vector<FSTArc*>& arcs = out_arc_data->arcs;
    arcs.reserve(a.n());
    for (; a.valid(); ++a) {
        arcs.push_back(new FSTArc(p,
                                  a->next,
                                  a->output,
                                  a->input,
                                  _semiring->invert(a->weight)));
    }
    return FSTArcIter(out_arc_data);
}

FSTArcIter
FSTInvertRep::arcs_out(int p, FSTArcOrder order) const
{
    return _invert_arcs(p, _in->arcs_out(p, order));
}

FSTArcIter
FSTInvertRep::arcs_out_with_input(int p, int i, FSTArcMode m) const
{
    return _invert_arcs(p, _in->arcs_out_with_input(p, i, m));
}

FSTArcIter
FSTInvertRep::arcs_out_with_output(int p, int o, FSTArcMode m) const
{
    return _invert_arcs(p, _in->arcs_out_with_output(p, o, m));
}

void
FSTInvertRep::force() const
{
    _in->force();
    _n_states = _in->n_states();
}

void
FSTInvertRep::reset() const
{
    _in->reset();
    _n_states = _in->n_states();
}

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 80
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
