////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2008, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

#include <cstring>
#include "fst/FST_i.h"

using namespace std;

FSTLetterToSound::FSTLetterToSound(const char* filename, int N, float S)
{
    this->N = N;
    this->S = S;

    FSTGeneric fst = FSTLoad(filename);

    if (fst->input_alphabet() == fst->output_alphabet()) {
        gp = fst;
        build_g2gp();
        build_gp2p();
    } else {
        g2p = fst;
    }
}

FSTLetterToSound::~FSTLetterToSound()
{
}

FSTGeneric
FSTLetterToSound::generateFst(const string& word)
{
    // Construct grapheme FST for word, which is just a linear sequence of
    // letters, ignoring whitespace.
    //
    int w = FSTstring_map[word];
    FSTGeneric fst = FSTBasic();
    int p = fst->add_state();
    fst->set_initial(p);
    char label[2];
    label[1] = 0;
    for (int i = 0; i < int(word.length()); i++) {
        char c = word[i];
        label[0] = c;
        if (c != ' ' && c != '\t' && c != '\n' && c != '_') {
            label[0] = c;
            int l = FSTstring_map[label];
            fst->add_state();
            fst->add_arc(p, p + 1, w, l);
            p++;
        }
        w = 0;
    }
    fst->set_final(p);
    if (fst->is_final(fst->initial())) {
        return fst; // empty result
    }

    // Compose with appropriate FST(s) to produce phonemes.
    //
    if (g2p) {
        //
        // We have a direct grapheme:phoneme FST.
        
        fst = FSTTrim(FSTCompose(fst, g2p));
        if (fst->initial() < 0) {
            return FSTGeneric(); // error result
        }

    } else {
        //
        // We have a grapheme:graphone o graphone:graphone o graphone:phoneme cascade of FSTs.

        fst = FSTTrim(FSTCompose(fst, g2gp));
        if (fst->initial() < 0) {
            return FSTGeneric(); // error result
        }
        fst = FSTTrim(FSTCompose(fst, gp));
        fst = FSTTrim(FSTCompose(fst, gp2p));
    }

    // Search for N best.
    //
    fst = FSTPotential(fst);
    fst = FSTAstarNbest(fst, FSTOutput, N, S);
    if (fst->initial() < 0) {
        return FSTGeneric(); // error result
    }

    return fst;
}

string
FSTLetterToSound::generateString(const string& word)
{
    FSTGeneric nbest = generateFst(word);
    if (!nbest) {
        return "";
    }
    string result;
    for (FSTArcIter i = nbest->arcs_out(nbest->initial()); i.valid(); i++) {
        bool word_first = true;
        if (result.size() > 0)
            result += ',';
        for (FSTArcIter a = i; a.valid(); a = nbest->arcs_out(a->next)) {
            if (a->output) {
                if (!word_first)
                    result += ' ';
                word_first = false;
                result += FSTstring_map[a->output];
            }
        }
    }

    return result;
}

void
FSTLetterToSound::build_g2gp()
{
    // Construct FST mapping letter sequences to graphone sequences.
    //
    g2gp = FSTBasic();
    g2gp->add_state();
    g2gp->set_initial(0);
    g2gp->set_final(0);

    const fst::set<int>& alphabet = gp->input_alphabet();
    char input[2];
    input[1] = 0;
    for (fst::set<int>::const_iterator i = alphabet.begin(); i != alphabet.end(); ++i) {
        int output_index = *i;
        char* graphone = strdup(FSTstring_map[output_index]);
        char* slash = strchr(graphone, '/');
        if (slash) {
            if (graphone == slash) {
                // empty grapheme sequence
                g2gp->add_arc(0, 0, 0, output_index);
            } else {
                *slash = 0;
                if (strchr(graphone, '_')) {
                    int p = 0;
                    char *end = slash;
                    for (char *save, *cp = strtok_r(graphone, "_", &save);
                         cp != NULL;
                         cp = strtok_r(NULL, "_", &save)) {
                        int q = (cp + strlen(cp) < end) ? g2gp->add_state() : 0;
                        int input_index = FSTstring_map[cp];
                        g2gp->add_arc(p, q, input_index, output_index);
                        output_index = 0;
                        p = q;
                    }
                } else {
                    int p = 0;
                    for (const char* cp = graphone; cp[0]; ++cp) {
                        int q = cp[1] ? g2gp->add_state() : 0;
                        input[0] = *cp;
                        int input_index = FSTstring_map[input];
                        g2gp->add_arc(p, q, input_index, output_index);
                        output_index = 0;
                        p = q;
                    }
                }
            }
        }
        free(graphone);
    }
}

void
FSTLetterToSound::build_gp2p()
{
    // Construct FST mapping graphone sequences to phoneme sequences.
    //

    gp2p = FSTBasic();
    gp2p->add_state();
    gp2p->set_initial(0);
    gp2p->set_final(0);

    const fst::set<int>& alphabet = gp->output_alphabet();
    for (fst::set<int>::const_iterator i = alphabet.begin(); i != alphabet.end(); ++i) {
        int input_index = *i;
        char* graphone = strdup(FSTstring_map[input_index]);
        char* end = graphone + strlen(graphone);
        char* slash = strchr(graphone, '/');
        if (slash) {
            if (slash[1] == 0) {
                // empty phoneme sequence
                gp2p->add_arc(0, 0, input_index, 0);
            } else {
                int p = 0;
                for (char *save, *cp = strtok_r(slash + 1, "_", &save);
                     cp != NULL;
                     cp = strtok_r(NULL, "_", &save)) {
                    int q = (cp + strlen(cp) < end) ? gp2p->add_state() : 0;
                    int output_index = FSTstring_map[cp];
                    gp2p->add_arc(p, q, input_index, output_index);
                    input_index = 0;
                    p = q;
                }
            }
        }
        free(graphone);
    }

    gp2p = FSTMinimize(gp2p);
}

#ifdef UNIT_TEST

#include <iostream>

int main(int argc, char** argv)
{
    FSTLetterToSound l2s(argv[1]);
    cout << l2s.generate(argv[2]) << endl;
    return 0;
}

#endif

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 80
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
