////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

#include "fst/FST_i.h"
#include "fst/FSTTimer.h"
using namespace std;

//#define TIMING
//#define DEBUG
//#define FORCE_CYCLIC

static inline int
_max(int x, int y)
{
    return (x > y) ? x : y;
}

///////////////////////////////////////////////////////////////////////////////
//
// Types
//

typedef fst::vector<const FSTArc*> FSTConstArcPtrVector;

struct FSTMinimizeStateInfo {
    FSTConstArcPtrVector arcs;
    FSTArcIter iterator;
    fst::vector<int> states_in;
    float final_weight;
};

typedef fst::vector<FSTMinimizeStateInfo> FSTMinimizeStateInfoVector;

struct FSTArcCompareInputOutputWeightNext
    : binary_function<const FSTArc*, const FSTArc*, bool> {
    
    bool operator () (const FSTArc* x, const FSTArc* y) const
    {
        if ( x->input < y->input ) return true;
        if ( y->input < x->input ) return false;
        if (x->output < y->output) return true;
        if (y->output < x->output) return false;
        if (x->weight < y->weight) return true;
        if (y->weight < x->weight) return false;
        return x->next < y->next;
    }
};

//
// The following are a "comparison objects" for use in STL maps.  We store some
// state in a comparision object, and then call it like a function.  These
// return true iff one state is lexicographically "less" than another state.
//

// For differentiating states with respect to (final_weight, (input, output,
// weight, block[next])*).
//
struct FSTMinimizeCompare
    : binary_function<int, int, bool> {
    
    const FSTMinimizeStateInfoVector& state_info;
    const fst::vector<int>& block_map;
    float eps;
    
    // The constructor requires precomputed state info and a block map.
    //
    FSTMinimizeCompare(const FSTMinimizeStateInfoVector& state_info_,
                       const fst::vector<int>& block_map_, float eps_)
        : state_info(state_info_), block_map(block_map_), eps(eps_)
    {
    }
    
    // Compare states p and q.
    //
    bool operator()(int p, int q) const
    {
        const FSTMinimizeStateInfo& pi = state_info[p];
        const FSTMinimizeStateInfo& qi = state_info[q];
        if (pi.final_weight < qi.final_weight - eps) return true;
        if (qi.final_weight < pi.final_weight - eps) return false;
        if (pi.arcs.size() < qi.arcs.size()) return true;
        if (qi.arcs.size() < pi.arcs.size()) return false;
        FSTConstArcPtrVector::const_iterator pai, qai;
        for (pai = pi.arcs.begin(), qai = qi.arcs.begin();
             pai != pi.arcs.end();
             pai++, qai++) {
            const FSTArc* pa = *pai;
            const FSTArc* qa = *qai;
            if (pa->input < qa->input) return true;
            if (qa->input < pa->input) return false;
            if (pa->output < qa->output) return true;
            if (qa->output < pa->output) return false;
            if (pa->weight < qa->weight - eps) return true;
            if (qa->weight < pa->weight - eps) return false;
            int pb = block_map[pa->next];
            int qb = block_map[qa->next];
            if (pb < qb) return true;
            if (qb < pb) return false;
        }
        return false;
    }
};

// For differentiating states with respect to (block[next]*).  This is like
// FSTMinimizeCompare above, except that we have already differentiated states
// with respect to final_weight, ((input, output, weight)*).  This is used in the
// iterative splitting of cyclic_minimize_partition.
//
struct FSTMinimizeCompare2
    : binary_function<int, int, bool> {

    const FSTMinimizeStateInfoVector& state_info;
    const fst::vector<int>& block_map;

    // The constructor requires precomputed state info and a block map.
    //
    FSTMinimizeCompare2(const FSTMinimizeStateInfoVector& state_info_,
                        const fst::vector<int>& block_map_)
        : state_info(state_info_), block_map(block_map_)
    {
    }

    // Compare states p and q.
    //
    bool operator()(int p, int q) const
    {
        const FSTMinimizeStateInfo& pi = state_info[p];
        const FSTMinimizeStateInfo& qi = state_info[q];
        FSTConstArcPtrVector::const_iterator pai, qai;
        for (pai = pi.arcs.begin(), qai = qi.arcs.begin();
             pai != pi.arcs.end();
             pai++, qai++) {
            const FSTArc* pa = *pai;
            const FSTArc* qa = *qai;
            int pb = block_map[pa->next];
            int qb = block_map[qa->next];
            if (pb < qb) return true;
            if (qb < pb) return false;
        }
        return false;
    }
};

// FSTMinimizePartition holds all information associated with a partition of
// states.  block_for_state[p] holds the block index associated with a given
// state.  blocks[b] holds information about a particular block, including the
// first (arbitrary) state in the block and the number of states in a block.
// links[p] holds the forward and backward links for the doubly linked block
// lists, terminated by -1.  There are methods for moving a state to a new block
// and moving a state to an existing block.
//
struct FSTMinimizePartition {

    struct Block {
        int start;
        int count;
    };
    struct Link {
        int prev;
        int next;
    };
    int S;
    fst::vector<Link> links;
    fst::vector<Block> blocks;
    fst::vector<int> block_for_state;
    float zero;

    FSTMinimizePartition(int S_, FSTSemiring semiring)
        : S(S_), links(S), blocks(), block_for_state(size_t(S), -1),
          zero(semiring->zero)
    {
    }

    // Reset partition to no blocks, with all states marked as in block -1.
    //
    void reset()
    {
        blocks.clear();
        fill(block_for_state.begin(), block_for_state.end(), -1);
    }

    // Return number of blocks.
    //
    int n_blocks() const
    {
        return blocks.size();
    }

    // Return next state in block.
    //
    int next(int p) const
    {
        return links[p].next;
    }

    // Move p to a new block.  Does not update block_for_states[].
    //
    int new_block(int p)
    {
        remove(p);
        blocks.push_back(Block());
        Block& block = blocks.back();
        block.start = p;
        block.count = 1;
        Link& link = links[p];
        link.next = link.prev = -1;
        return blocks.size() - 1;
    }

    // Move p to start of existing block b.  Does not update block_for_states[].
    //
    void move(int p, int b)
    {
        remove(p);
        Block& block = blocks[b];
        block.count++;
        Link& link = links[p];
        link.prev = -1;
        link.next = block.start;
        links[block.start].prev = p;
        block.start = p;
    }

    // Remove p from current block.
    //
    void remove(int p)
    {
        int b = block_for_state[p];
        if (b >= 0) {
            Block& block = blocks[b];
            assert(block.count > 1);
            block.count--;
            if (block.start == p)
                block.start = links[p].next;
            Link& link = links[p];
            if (link.next >= 0) links[link.next].prev = link.prev;
            if (link.prev >= 0) links[link.prev].next = link.next;
        }
    }

    // Update block_for_state[] for blocks b..blocks.size().  We do not update
    // immediately when we move a state to a new block because that could
    // adversely affect the construction of an ordered map (i.e., in effect the
    // keys would be changing).
    //
    void update(int b)
    {
        for (; b < int(blocks.size()); b++)
            for (int p = blocks[b].start; p >= 0; p = links[p].next)
                block_for_state[p] = b;
    }

    // Print block b.
    //
    void print(int b, ostream& os=cerr) const
    {
        os << '{';
        int start = blocks[b].start;
        for (int p = start; p >= 0; p = links[p].next) {
            if (p != start)
                os << ',';
            os << p;
        }
        os << '}' << endl;
    }

    // Print all blocks.
    //
    void print(ostream& os=cerr) const
    {
        os << "blocks " << 0 << ".." << blocks.size() - 1 << endl;
        for (int b = 0; b < int(blocks.size()); b++) {
            os << "  block " << b << " = ";
            print(b, os);
        }
    }
};

///////////////////////////////////////////////////////////////////////////////
//
// Forward static function declarations.
//
static void minimize_partition(
    fst::vector<int>& result,
    const FSTMinimizeStateInfoVector& info,
    int i,
    float eps,
    FSTSemiring semiring);
static bool acyclic_minimize_partition(
    FSTMinimizePartition& P,
    const FSTMinimizeStateInfoVector& info,
    int i,
    float eps);
static void cyclic_minimize_partition(
    FSTMinimizePartition& P,
    const FSTMinimizeStateInfoVector& info,
    float eps);
static bool compute_initial_partition_dfs(
    FSTMinimizePartition& P,
    const FSTMinimizeStateInfoVector& info,
    int p);
static bool compute_initial_partition(
    FSTMinimizePartition& P,
    const FSTMinimizeStateInfoVector& info,
    int i);

///////////////////////////////////////////////////////////////////////////////
//
// Functions.
//

// FSTMinimize performs FSA minimization.  If fst is mutable, the minimization
// will be in place, otherwise a copy of fst will be made.  Note that
// minimization does not include epsilon removal, determinization, output
// pushing, or weight pushing.  All of those are separate operations.
// Minimization merges states that are equivalent.
//
FSTGeneric
FSTMinimize(FSTGeneric fst, float eps)
{
    int p;
    FSTTimer timer;

    // If fst doesn't have an initial state, we do nothing.
    //
    if (fst->initial() < 0)
        return fst;

    // Make sure that fst is mutable.
    //
    if (!fst->is_mutable())
        fst = FSTMutableCopy(fst);

    // Build up a FSTMinimizeStateInfo for each state.  This structure will hold
    // arc pointers fully sorted (so we can correctly compare state arc
    // signatures), states that have arcs arriving at the state, and the
    // finality of the state.  The purpose for doing this is to avoid any FST
    // method calls during minimization which results in a big speedup.  Note
    // that we also store the arc iterator so that if the FST produces
    // dynamically allocated arcs they won't be deleted before we're finished
    // with them.
    //
    timer.reset();
    int S = fst->n_states();
    FSTMinimizeStateInfoVector info(S);
    FSTArcCompareInputOutputWeightNext comparator;
    for (p = 0; p < S; p++) {
        FSTMinimizeStateInfo& i = info[p];
        FSTArcIter a = fst->arcs_out(p);
        i.iterator = a;
        i.arcs.reserve(a.n());
        for (; a.valid(); a++) {
            i.arcs.push_back(*a);
            info[a->next].states_in.push_back(p);
        }
        sort(i.arcs.begin(), i.arcs.end(), comparator);
        i.final_weight = fst->final_weight(p);
    }
    for (p = 0; p < S; p++) {
        fst::vector<int>& states_in = info[p].states_in;
        sort(states_in.begin(), states_in.end());
        fst::vector<int>::iterator i = unique(states_in.begin(), states_in.end());
        states_in.erase(i, states_in.end());
    }
#ifdef TIMING
    cerr.form("preprocess time: %.3fs\n", timer.value());
#endif

    // Compute partition of states.
    //
    fst::vector<int> mapping;
    timer.reset();
    minimize_partition(mapping, info, fst->initial(), eps, fst->semiring());
#ifdef TIMING
    cerr.form("partition time: %.3fs\n", timer.value());
#endif

    // Merge states according to partition.  Note that when merging state p with
    // state q, we only need to move the transitions arriving at p to q before
    // deleting p.  Transitions leaving p already leave q also.  Otherwise, p
    // and q would not be equivalent during minimization.
    //
    timer.reset();
    fst::vector<int> to_delete;
    for (int p = 0; p < int(mapping.size()); p++) {
        int q = mapping[p];
        if (q != p) {
            
            // Redirect arcs into p to q instead.
            //
            const fst::vector<int>& states_in = info[p].states_in;
            for (fst::vector<int>::const_iterator i = states_in.begin();
                 i != states_in.end(); i++) {
                int r = *i;
                const FSTConstArcPtrVector& arcs = info[r].arcs;
                for (FSTConstArcPtrVector::const_iterator j = arcs.begin();
                     j != arcs.end(); j++) {
                    const FSTArc* a = *j;
                    if (a->next == p)
                        fst->add_arc(r, q, a->input, a->output, a->weight);
                }
            }
            to_delete.push_back(p);
        }
    }
    fst->set_initial(mapping[fst->initial()]);
    fst->delete_states(to_delete);
#ifdef TIMING
    cerr.form("delete time: %.3fs\n", timer.value());
#endif

    return fst;
}


#ifdef DEBUG
ostream& operator<<(ostream& os, const fst::vector<int>& x)
{
    for (fst::vector<int>::const_iterator i = x.begin(); i != x.end(); i++)
        os << " " << *i;
    return os;
}
#endif



// General FST minimization: O(N log N)
//
// Algorithm (adapted from Hopcroft's algorithm):
//
//  1. partition all states on (final_weight, (input, output, weight,
//     next->final_weight)...) --  O(E)
//  2. put all resulting blocks into to-do list
//  3. while to-do list not empty:
//  4.   select and delete any block i from to-do list
//  5.   inverse <- all blocks containg at least one states with arcs into
//       block  i
//  6.   for each block j in inverse:
//  7.     block q <- block j intersect inverse
//  8.     block j <- block j - block q
//  9.     partition block q on the set of arcs with next states mapped to
//         blocks
//
// The primary difference of this algorithm with that of Hopcroft is that we do
// not iterate through individual labels for 5-9 above.  Instead, we split based
// on all labels simultaneously, although this means we can't really do 6
// properly.  Instead, we split all blocks in inverse, including those where the
// entire block is in the inverse.  We must do this since the states in the
// block could be in the inverse due to different labels.  Hopcroft only
// considers a single label at a time.
//
static void
cyclic_minimize_partition(FSTMinimizePartition& P,
                          const FSTMinimizeStateInfoVector& info,
                          float eps)
{
    int i, j, k, p, q;
    int S = info.size();
    FSTTimer timer;

    // Stack of blocks waiting to be partitioned against.  in_waiting is used to
    // quickly determine if a block has already been put in waiting.  Note that
    // order of access from waiting is unimportant; we could also use a queue.
    //
    stack<int> waiting;
    fst::vector<bool> in_waiting;

    // We use state_block_map when splitting states apart.  It maps a state to a
    // block.  The comparator object differentiates states based on their
    // signature (final_weight, (input, output, weight, block[next])*).
    //
    typedef fst::map<int, int, FSTMinimizeCompare> StateBlockMap;
    FSTMinimizeCompare comparator(info, P.block_for_state, eps);
    StateBlockMap state_block_map(comparator);

    // Reset partition P which may have been altered by a failed acyclic
    // minimize partition attempt.
    //
    P.reset();

#ifdef DEBUG
    cerr << "at start: " << P.n_blocks() << " blocks" << endl;
#endif

    // Now, distinguish states as fully as possible initially.  This will be the
    // seed partition that we will further refine in the main loop below.
    //
    timer.reset();
    for (p = 0; p < S; p++) {
        StateBlockMap::iterator sbmi = state_block_map.find(p);
        if (sbmi == state_block_map.end()) {
            // not found -- new block
            int b = P.new_block(p);
#ifdef DEBUG
            cerr << "start " << p << " into block " << b << endl;
#endif
            state_block_map[p] = b;
            waiting.push(b);
            in_waiting.push_back(true);
        } else {
#ifdef DEBUG
            cerr << "state " << p << " into block " << (*sbmi).second << endl;
#endif
            P.move(p, (*sbmi).second);
        }
    }
#ifdef TIMING
    cerr << form << "initial partition time: " << timer.value() << endl;
#endif
#ifdef DEBUG
    cerr << "initial partition: ";
    P.print();
#endif
    P.update(0);
    state_block_map.clear();                    // no longer needed

    // Get second comparator and state block map ready.  This one differentiates
    // solely on (block[next]*) since we have already differentiated on
    // everything else.
    //
#if 0
    cerr << "DEBUG\n";
    typedef fst::map<int, int, FSTMinimizeCompare2> StateBlockMap2;
    FSTMinimizeCompare2 comparator2(info, P.block_for_state);
    StateBlockMap2 state_block_map2(comparator2);
#else
#   define StateBlockMap2 StateBlockMap
    StateBlockMap state_block_map2(comparator);
#endif

    //
    //
    fst::vector<int> inverse_states;
    fst::vector<bool> state_in_inverse(S, false);
    fst::vector<int> block_inverse_state_counts(size_t(P.n_blocks()), 0);
    fst::vector<int> inverse_blocks;

    // Main Loop
    //
    timer.reset();
    while (!waiting.empty()) {

#ifdef DEBUG
        {
            const int* wp = &waiting.top();
            cerr << "waiting:";
            for (int z = 0; z < waiting.size(); z++)
                cerr << " " << wp[-z];
            cerr << endl;
        }
#endif

    i = waiting.top();
    waiting.pop();
    in_waiting[i] = false;

    // Compute states that can reach block[i], the "inverse" across all
        // arcs.
    //
    for (k = 0; k < int(inverse_states.size()); k++)
        state_in_inverse[inverse_states[k]] = false;
    inverse_states.clear();
    inverse_blocks.clear();
    for (p = P.blocks[i].start; p >= 0; p = P.links[p].next) {
        const fst::vector<int>& states_in = info[p].states_in;
        for (fst::vector<int>::const_iterator si = states_in.begin();
             si != states_in.end();
             si++) {
            q = *si;
            if (!state_in_inverse[q]) {
                state_in_inverse[q] = true;
                inverse_states.push_back(q);
                int b = P.block_for_state[q];
                if (block_inverse_state_counts[b] == 0)
                    inverse_blocks.push_back(b);
                block_inverse_state_counts[b]++;
            }
        }
    }
#ifdef DEBUG
        cerr << "inverse states:" << inverse_states << "\n";
        cerr << "inverse blocks:" << inverse_blocks << "\n";
#endif

        //
    for (k = 0; k < int(inverse_blocks.size()); k++) {
        j = inverse_blocks[k];

            if (P.blocks[j].count < 2) {

#ifdef DEBUG
                cerr << "no need to split block " << j << endl;
#endif

            } else {

#ifdef DEBUG
                cerr << "splitting block " << j << " with respect to " << i << endl;
#endif
                
                // Split block j.  The first state in j remains in j.
                // Subsequent unequal states go in new blocks.
                //
                state_block_map2.clear();
                int s = P.blocks[j].start;
                assert(s >= 0);
                state_block_map2[s] = j;
#ifdef DEBUG
                cerr << "  state " << s << " stays in block " << j << endl;
#endif
                int first_new_b = P.n_blocks();
                for (s = P.next(s); s >= 0;) {
                    int next_state = P.next(s);
                    int next_block = P.n_blocks();
                    StateBlockMap2::value_type entry(s, next_block);
                    pair<StateBlockMap2::iterator, bool> status =
                        state_block_map2.insert(entry);
                    if (status.second) {
                        // A new entry was created in the map, so start new
                        // block.
                        P.new_block(s);
                        in_waiting.push_back(false);
                        block_inverse_state_counts.push_back(0);
                    } else {
                        // There was already a matching entry in the map.
                        P.move(s, (*status.first).second);
                    }
#ifdef DEBUG
                    cerr << "  state " << s << " into block " << (*status.first).second << endl;
#endif
                    s = next_state;
                    
                }
                P.update(first_new_b);
                
                // Push appropriate blocks onto waiting stack.
                //
                if (in_waiting[j]) {
                    for (int b = first_new_b; b < P.n_blocks(); b++) {
                        if (!in_waiting[b]) {
                            waiting.push(b);
                            in_waiting[b] = true;
                        }
                    }
                } else {
                    
                    // Determine largest block, and push all others.
                    //
                    int biggest_block = j;
                    int biggest_block_size = P.blocks[j].count;
                    int b;
                    for (b = first_new_b; b < P.n_blocks(); b++) {
                        int size = P.blocks[b].count;
                        if (size > biggest_block_size) {
                            biggest_block_size = size;
                            biggest_block = b;
                        }
                    }
                    if (j != biggest_block) {
                        waiting.push(j);
                        in_waiting[j] = true;
                    }
                    for (b = first_new_b; b < P.n_blocks(); b++) {
                        if (b != biggest_block) {
                            waiting.push(b);
                            in_waiting[b] = true;
                        }
                    }
                }
            }

            // Reset for next iteration.
            //
        block_inverse_state_counts[j] = 0;
    
    }

    }
#ifdef TIMING
    cerr.form("iteration partition time: %.3fs\n", timer.value());
#endif
}

// Depth-first search (DFS) for computing initial partition used by
// acyclic_minimize_partition.  The result is put into partition P, with states
// of height h put into block h.  (Height for a state is the maximal
// distance to a final state.)  We fail by returning false if we detect a cycle
// or if there are any states that cannot reach a final state.
//

typedef FSTConstArcPtrVector::const_iterator ArcIter;

struct FSTMinimizeStackEntry {
    int state;
    int height;
    ArcIter iter, end;
    FSTMinimizeStackEntry(int s, ArcIter i, ArcIter e)
        : state(s), height(0), iter(i), end(e) {
    }
};

typedef FSTMinimizeStackEntry StackEntry;

static bool
compute_initial_partition_dfs(FSTMinimizePartition& P,
                              const FSTMinimizeStateInfoVector& info, int p)
{
    P.block_for_state[p] = -2;
    if (info[p].final_weight == P.zero && info[p].arcs.size() == 0)
        return false;                           // non-final with no arcs

    stack<StackEntry> todo;
    todo.push(StackEntry(p, info[p].arcs.begin(), info[p].arcs.end()));

    while (!todo.empty()) {
        StackEntry& top = todo.top();
        if (top.iter != top.end) {
            
            const FSTArc* a = *top.iter;
            int q = a->next;
            if (P.block_for_state[q] == -2) {   // cycle detected: fail
                
                return false;
                
            } else if (P.block_for_state[q] == -1) { // q's height unknown
                
                P.block_for_state[q] = -2;
                if (info[q].final_weight == P.zero && info[q].arcs.size() == 0)
                    return false;               // non-final with no arcs
                todo.push(StackEntry(q, info[q].arcs.begin(), info[q].arcs.end()));
                
            } else {                            // q's height known

                top.height = _max(top.height, P.block_for_state[q] + 1);
            }
            ++top.iter;
            
        } else {
            // finished children
            int h = top.height;
            if (h < int(P.blocks.size())) {
                P.move(top.state, h);
            } else {
                assert(h == int(P.blocks.size()));
                P.new_block(top.state);
            }
            assert(P.block_for_state[top.state] == -2);
            P.block_for_state[top.state] = h;
            todo.pop();

            // propagate up to parent
            if (!todo.empty())
                todo.top().height = _max(todo.top().height, h + 1);
        }
        
    }

    return true;
}

static bool
compute_initial_partition(FSTMinimizePartition& P,
                          const FSTMinimizeStateInfoVector& info,
                          int initial)
{
    // P.block_for_state[i] will hold height of state i.  Initialize them all to
    // -1 to indicate that they have not been computed yet.  -2 will be used in
    // detecting cycles.
    //
    fill(P.block_for_state.begin(), P.block_for_state.end(), -1);

    // DFS computing heights and looking for cycles.
    //
    return compute_initial_partition_dfs(P, info, initial);
}

// Acyclic FST minimization: O(N)
//
// Algorithm (adapted from Revuz's algorithm):
//
//  1. partition all states on their maximum height from a final state using
//     a DFS
//  2. foreach block i in 1..H, where H is maximum h from step 1:
//  3.   split block i on (final_weight, (input, output, weight, block[to_state])*)
//
static bool
acyclic_minimize_partition(FSTMinimizePartition& P,
                           const FSTMinimizeStateInfoVector& info,
                           int initial,
                           float eps)
{
    int b;

    // Initial partition, where block B[i] represents all states of height i.
    // Height of a state is its maximum distance from a final state.  We detect
    // a cyclic FST while computing height, and bail out, letting the general
    // algorithm handle it.
    //
    if (!compute_initial_partition(P, info, initial)) {
        fill(P.block_for_state.begin(), P.block_for_state.end(), -1);
        return false;
    }
    int H = P.blocks.size();

    // Main loop: split B[0], B[1], ..., B[H-1] on (final_weight, (input, output,
    // weight, block[to_state])*).  We do not have to split new blocks B[H],
    // B[H+1], ... ever again.
    //
    typedef fst::map<int, int, FSTMinimizeCompare> StateBlockMap;
    FSTMinimizeCompare comparator(info, P.block_for_state, eps);
    StateBlockMap state_block_map(comparator);
    int first_new_block = H;
    
    for (b = 0; b < H; b++) {
        
        state_block_map.clear();
        
        int p = P.blocks[b].start;
        assert(p >= 0);
        state_block_map[p] = b;                 // First state stays put.
        p = P.links[p].next;
        while (p >= 0) {
            int next_p = P.links[p].next;
            StateBlockMap::iterator sbmi = state_block_map.find(p);
            if (sbmi == state_block_map.end()) {
                // not found -- new block
                state_block_map[p] = P.new_block(p);
            } else {
                P.move(p, (*sbmi).second);
            }
            p = next_p;
        }
        P.update(first_new_block);
        first_new_block = P.blocks.size();
        
    }
    return true;
}

static void
minimize_partition(fst::vector<int>& result,
                   const FSTMinimizeStateInfoVector& info,
                   int initial,
                   float eps,
                   FSTSemiring semiring)
{
    int S = info.size();
    
    // Try O(N) acyclic minimization first, else if cycles are detected fall
    // back to O(N log N) general minimization.
    //
    FSTMinimizePartition P(S, semiring);
#ifdef FORCE_CYCLIC
    cyclic_minimize_partition(P, info, eps);
#else
    if (!acyclic_minimize_partition(P, info, initial, eps))
        cyclic_minimize_partition(P, info, eps);
#endif
    
    // Convert FSTMinimizePartition structure to vector mapping current states
    // to new states.
    //
    result.resize(S);
#ifdef DEBUG
    cerr << "minimize partition:" << endl;
#endif
    for (int p = 0; p < S; p++) {
        int b = P.block_for_state[p];
        result[p] = (b < 0) ? -1 : P.blocks[b].start;
#ifdef DEBUG
        cerr << "    " << p << " ~ " <<  result[p] << endl;
#endif
    }
#ifdef DEBUG
    P.print();
#endif
}

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 80
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:


