////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

#include <cstring>
#include "fst/FST_i.h"

#ifndef _MSC_VER

#include <fstream>
#include <unistd.h>
#include <fcntl.h>
#include <cerrno>
#include <sys/mman.h>

#define MMAP_ALIGNMENT 4096

static void
_dump_int(ostream& os, int x)
{
    os.write((const char*)&x, 4);
}

static void
_dump_float(ostream& os, float x)
{
    os.write((const char*)&x, 4);
}

static int
_load_int(istream& is)
{
    int x;
    is.read((char*)&x, 4);
    return x;
}

static void
_dump_label_map(ostream& os, const fst::set<int>& labels)
{

    // size of map
    _dump_int(os, labels.size());

    // place holder for length of longest label
    streampos longest_length_pos = os.tellp();
    _dump_int(os, -1);

    int longest_length = 0;
    for (fst::set<int>::const_iterator i(labels.begin()); i != labels.end(); ++i) {
        const char* label = FSTstring_map[*i];
        int length = strlen(label) + 1;
        if (length > longest_length)
            longest_length = length;
        os.write(label, length);
        _dump_int(os, *i);
    }
    
    // fix longest length
    os.seekp(longest_length_pos, ios::beg);
    _dump_int(os, longest_length);
    os.seekp(0, ios::end);
}

static void
_load_label_map(istream& is, fst::set<int>& labels)
{
    labels.clear();

    int n_labels = _load_int(is);

    int longest_label = _load_int(is);
    char* buf = new char[longest_label];
    while (n_labels--) {
        for (char* p = buf; true; p++) {
            is.read(p, 1);
            if (*p == 0)
                break;
        }
        int index = _load_int(is);
        if (index != FSTstring_map[buf])
            FSTFatal("FSTMmap: label `%s' != %d", buf, index);
        labels.insert(index);
    }

    delete[] buf;
}

//
// Dump out mmap-able file.
//

void
FSTGenericInt::dump_mmap(const char* filename, FSTArcOrder order) const
{
    ofstream os(filename);
    if (!os)
        FSTFatal("can't write file %s", filename);

    // header
    os << "#FSTMmap " << semiring()->name << endl;

    // byte-order indicator
    _dump_int(os, 0x12345678);

    // file format version
    _dump_int(os, 1);

    // arc order
    _dump_int(os, int(order));

    // initial state
    _dump_int(os, initial());

    // number of arcs
    int n_arcs = 0;
    for (int p = 0; p < n_states(); p++)
        n_arcs += arcs_out(p).n();
    _dump_int(os, n_arcs);

    // number of states
    _dump_int(os, _n_states);

    // input/output label maps
    _dump_label_map(os, input_alphabet());
    _dump_label_map(os, output_alphabet());

    // mmap offset (start of arcs)
    size_t mmap_pos = ((size_t)os.tellp() + MMAP_ALIGNMENT-1) & ~(MMAP_ALIGNMENT-1);
    _dump_int(os, mmap_pos);
    os.seekp(mmap_pos, ios::beg);

    //
    // Start of mmap.
    //

    // Dump arcs, sorted for each state.
    //
    assert(sizeof(FSTArc) == 20);
    fst::vector<int> n_arcs_for_state(_n_states);
    for (int p = 0; p < _n_states; p++) {
        FSTArcIter a = arcs_out(p, order);
        n_arcs_for_state[p] = a.n();
        for (; a.valid(); a++)
            os.write((const char*)*a, sizeof(FSTArc));
    }

    // Insert filler so states are aligned on 16-byte boundary of mmap region.
    //
    size_t n_filler = (16 - (((size_t)os.tellp() - mmap_pos) & 15)) & 15;
    os.seekp(n_filler, ios::cur);

    assert((size_t)os.tellp() - mmap_pos == ((n_arcs*sizeof(FSTArc) + 15) & ~15));

    // Dump states.
    //
    int arc_count = 0;
    for (int p = 0; p < _n_states; p++) {
        _dump_float(os, final_weight(p));
        _dump_float(os, potential(p));
        _dump_int(os, arc_count);
        _dump_int(os, arc_count + n_arcs_for_state[p]);
        arc_count += n_arcs_for_state[p];
    }    

}

class FSTMmapRep
    : public FSTGenericInt {

public: // methods

    FSTMmapRep(istream& is, const char* header, const char* filename);
    virtual ~FSTMmapRep();

    virtual const fst::set<int>& input_alphabet() const;
    virtual const fst::set<int>& output_alphabet() const;
    virtual FSTArcIter arcs_out(int p, FSTArcOrder order=FSTNone) const;
    virtual FSTArcIter arcs_out_with_input(int p, int i,
                                           FSTArcMode m=FSTEqual) const;
    virtual FSTArcIter arcs_out_with_output(int p, int o,
                                            FSTArcMode m=FSTEqual) const;
    virtual bool       is_final(int p) const;
    virtual float      final_weight(int p) const;
    virtual float      potential(int p) const;
    virtual void       force() const;
    virtual void       reset() const;

protected: // types

    struct _State;
    struct _TrainData;

protected: // members

    _State* _states;
    fst::vector<FSTArc*> _arcs;
    FSTArcOrder _arc_order;
    fst::set<int> _input_alphabet;
    fst::set<int> _output_alphabet;

protected: // methods

};

struct FSTMmapRep::_State {
    float  weight;
    float  potential;
    size_t arcs_begin;
    size_t arcs_end;
};

// Load from Mmap file.
//
FSTMmapRep::FSTMmapRep(istream& is, const char* header, const char* filename)
{
    const char* h;

    _class_name = "FSTMmap";

    if (strstr(header, "PlusTimesLog"))
        _semiring = FSTPlusTimesLog;
    else if (strstr(header, "PlusTimes"))
        _semiring = FSTPlusTimes;
    else if (strstr(header, "MaxPlus"))
        _semiring = FSTMaxPlus;
    else // default is MinPlusSemiring
        _semiring = FSTMinPlus;

    h = strstr(header, "WeightScale=");
    if (h)
        FSTWarn("WeightScale not supported for FSTMmap");

    h = strstr(header, "WeightOffset=");
    if (h)
        FSTWarn("WeightOffset not supported for FSTMmap");

    int byte_order = _load_int(is);
    if (byte_order != 0x12345678)
        FSTFatal("Incorrect byte order for FSTMmap (%x)", byte_order);

    int file_version = _load_int(is);
    if (file_version != 1)
        FSTFatal("Bad file version for FSTMmap (%d)", file_version);

    _arc_order = FSTArcOrder(_load_int(is));

    _initial = _load_int(is);

    _n_arcs = _load_int(is);

    _n_states = _load_int(is);

    _load_label_map(is, _input_alphabet);
    _load_label_map(is, _output_alphabet);

    size_t mmap_beg = _load_int(is);
    is.seekg(mmap_beg, ios::beg);

    is.seekg(0, ios::end);
    size_t mmap_end = is.tellg();

    int fd = open(filename, O_RDONLY);
    if (fd < 0)
        FSTFatal("can't read file %s", filename);

    void* mmap_ptr = mmap(NULL, mmap_end - mmap_beg,
                          PROT_READ, MAP_PRIVATE,
                          fd, mmap_beg);

    close(fd);

    if (mmap_ptr == MAP_FAILED)
        FSTFatal("can't mmap file %s (%s)", filename, strerror(errno));

    FSTArc* arcs_begin = (FSTArc*)mmap_ptr;

    _arcs.resize(_n_arcs);
    for (int i = 0; i < _n_arcs; i++)
        _arcs[i] = arcs_begin + i;
    
    _states = (_State*)(((size_t)(arcs_begin + _n_arcs) + 15) & ~15);
}

FSTMmapRep::~FSTMmapRep()
{
}

const fst::set<int>&
FSTMmapRep::input_alphabet() const
{
    return _input_alphabet;
}

const fst::set<int>&
FSTMmapRep::output_alphabet() const
{
    return _output_alphabet;
}

bool
FSTMmapRep::is_final(int p) const
{
    FST_ASSERT_VALID_STATE(p);
    return _states[p].weight != _semiring->zero;
}

float
FSTMmapRep::final_weight(int p) const
{
    FST_ASSERT_VALID_STATE(p);
    return _states[p].weight;
}

float
FSTMmapRep::potential(int p) const
{
    FST_ASSERT_VALID_STATE(p);
    return _states[p].potential;
}

void
FSTMmapRep::force() const
{
    // Nothing to do because we're not lazy.
}

void
FSTMmapRep::reset() const
{
    // Nothing to do because we're not lazy.
}

FSTArcIter
FSTMmapRep::arcs_out(int p, FSTArcOrder order) const
{
    FST_ASSERT_VALID_STATE(p);
    if (order != FSTNone && order != _arc_order)
        FSTFatal("FSTMmap: asking for order %d, but compiled with order %d",
                 order, _arc_order);
    return FSTArcIter(_arcs,
                      _arcs.begin() + _states[p].arcs_begin,
                      _arcs.begin() + _states[p].arcs_end);
}

FSTArcIter
FSTMmapRep::arcs_out_with_input(int p, int i, FSTArcMode m) const
{
    FST_ASSERT_VALID_STATE(p);
    if (_arc_order != FSTInput)
        FSTFatal("FSTMmap: need order %d, but compiled with order %d",
                 FSTInput, _arc_order);
    typedef fst::vector<FSTArc*>::const_iterator Iter;
    Iter begin = _arcs.begin() + _states[p].arcs_begin;
    Iter end = _arcs.begin() + _states[p].arcs_end;
    FSTArcCompareInput compare;
    FSTArc a;
    a.input = i;
    pair<Iter, Iter> range;
    switch (m) {
    case FSTEqual:
    range = equal_range(begin, end, &a, compare);
    begin = range.first;
    end = range.second;
    break;
    case FSTLessEqual:
    end = upper_bound(begin, end, &a, compare);
    break;
    case FSTGreaterEqual:
    begin = lower_bound(begin, end, &a, compare);
    break;
    }
    return FSTArcIter(_arcs, begin, end);
}

FSTArcIter
FSTMmapRep::arcs_out_with_output(int p, int o, FSTArcMode m) const
{
    FST_ASSERT_VALID_STATE(p);
    if (_arc_order != FSTOutput)
        FSTFatal("FSTMmap: need order %d, but compiled with order %d",
                 FSTOutput, _arc_order);
    typedef fst::vector<FSTArc*>::const_iterator Iter;
    Iter begin = _arcs.begin() + _states[p].arcs_begin;
    Iter end = _arcs.begin() + _states[p].arcs_end;
    FSTArcCompareOutput compare;
    FSTArc a;
    a.output = o;
    pair<Iter, Iter> range;
    switch (m) {
    case FSTEqual:
    range = equal_range(begin, end, &a, compare);
    begin = range.first;
    end = range.second;
    break;
    case FSTLessEqual:
    end = upper_bound(begin, end, &a, compare);
    break;
    case FSTGreaterEqual:
    begin = lower_bound(begin, end, &a, compare);
    break;
    }
    return FSTArcIter(_arcs, begin, end);
}

FSTGeneric
FSTLoadMmap(istream& is, const char* header, const char* filename)
{
    return new FSTMmapRep(is, header, filename);
}

#else // _MSC_VER

FSTGeneric
FSTLoadMmap(istream& is, const char* header, const char* filename)
{
    FSTFatal("FSTMmap not supported in Windows");
    return FSTBasic();
}

void
FSTGenericInt::dump_mmap(const char* filename, FSTArcOrder order) const
{
    FSTFatal("FSTMmap not supported in Windows");
}

#endif // _MSC_VER

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 80
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
