////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

//
// FSTNgram2: 12-bytes per n-gram FST implementation
//

//#define CACHE_STATS
//#define HASH_STATS

#include "fst/FST_i.h"
#include <cstring>
#include <fstream>
#include "fst/hash_set"
#include "sq.h"
#include "BitStreamWriter32.h"
#include "MD5.h"
#include "fst/format.h"

#define STATE(STATE_PTR) (STATE_PTR - &_states[0])

#define HASH_C1 unsigned(0x11)
#define HASH_C2 unsigned(0x1001)

#define PNGRAM1 0x5751d867 // retired due to child_end and arc_end bug
#define PNGRAM2 0xabc3a62b
#define PNGRAM3 0xefb19065 // adds support for <$CLASS> ... </$CLASS>

#define F2Q(X,Q) int(rintf(X * (1 << Q)))

class FSTNgram2Rep
    : public FSTNgramInt {

public: // methods

    FSTNgram2Rep(istream& is, const char* header);
    virtual ~FSTNgram2Rep();
    virtual const fst::set<int>& input_alphabet() const;
    virtual const fst::set<int>& output_alphabet() const;
    virtual bool is_final(int p) const;
    virtual float final_weight(int p) const;
    virtual float potential(int p) const;
    virtual FSTArcIter arcs_out(int p, FSTArcOrder order) const;
    virtual FSTArcIter arcs_out_with_input(int p, int i, FSTArcMode m) const;
    virtual FSTArcIter arcs_out_with_output(int p, int o, FSTArcMode m) const;
    virtual void force() const;
    virtual void reset() const;
    virtual void dump_binary(ostream& os) const;

    virtual void determinize();
    virtual void invert();
    virtual void subtract(const FSTNgramInt* x);
    virtual void quantize(int bits);

    virtual bool ngram_weight(int p, int input, float& weight) const;
    virtual int  ngram_next_state() const;

    virtual void add_classed_word(int word, int word_class, float weight);

    virtual void dump_pngram(const char* vocab_file, const char* out_file, int weight_width) const;
    virtual void dump_pngram(const fst::list<std::string>& vocab_list, FILE* fp, int weight_width) const;

protected: // types (forward declarations)

    struct _Arc;
    struct _State;
    struct _ClassedWord;
    struct _SingleArcIterData;

    typedef fst::vector<_State>        _StateVector;
    typedef fst::vector<_ClassedWord*> _ClassedWordVector;

    struct _CacheLinks {

        _CacheLinks* next;
        _CacheLinks* prev;

        _CacheLinks() {
            next = this;
            prev = this;
        }
    };

    struct _CacheEntry : _CacheLinks {

        const int state;
        const int input;
        float weight;
        int next_state;

        _CacheEntry(int state_, int input_)
            : state(state_), input(input_) {}

#ifdef NO_HASH_SET
        bool operator < (const _CacheEntry& e) const {
            return FST_LESS3(state, e.state,
                             FST_LESS2(input, e.input));
        }
#else
        struct Hash {
            size_t operator()(const _CacheEntry& e) const {
                return 0x1001 * e.state + 0x11 * e.input;
            }
        };

        struct Equal {
            bool operator()(const _CacheEntry& a, const _CacheEntry& b) const {
                return a.state == b.state && a.input == b.input;
            }
        };
#endif

    };

#ifdef NO_HASH_SET
    typedef fst::set<_CacheEntry> _CacheSet;
    typedef fst::set<int> _IntSet;
#else
    typedef fst::hash_set<_CacheEntry, _CacheEntry::Hash, _CacheEntry::Equal> _CacheSet;
    typedef fst::hash_set<int> _IntSet;
#endif

#ifdef NO_HASH_MAP
    typedef fst::map<int, int> _IntIntMap;
#else
    typedef fst::hash_map<int, int> _IntIntMap;
#endif

protected: // members

    _StateVector         _states;
    const _State**       _hash_table;
    unsigned             _hash_mask;
    unsigned             _n;
    _ClassedWordVector   _classed_words;
    bool                 _deterministic;
    bool                 _inverted;
    fst::set<int>        _alphabet;
    int                  _initial_unit;
    int                  _final_unit;
    int                  _n_static_states;
    mutable _IntIntMap   _class_start_map;
    mutable _IntSet      _class_end_set;
    int                  _cache_capacity;
    mutable _CacheSet    _cache_set;
    mutable _CacheLinks  _cache_list;
    mutable _CacheEntry* _cache_entry;
#ifdef HASH_STATS
    mutable unsigned     _store_probes;
    mutable unsigned     _store_count;
    mutable unsigned     _hit_probes;
    mutable unsigned     _hit_count;
    mutable unsigned     _miss_probes;
    mutable unsigned     _miss_count;
#endif
#ifdef CACHE_STATS
    mutable unsigned     _cache_weight_try;
    mutable unsigned     _cache_weight_hit;
    mutable unsigned     _cache_next_try;
    mutable unsigned     _cache_next_hit;
#endif

protected: // methods

    void load_class_info(istream& is, const fst::map<int,int>& label_map, float scale);

    void hash_table_init();

    void hash_table_store(const _State* state);

    const _Arc* find_arc(const _State* state, unsigned unit,
             float& bo_weight) const;

    const _Arc* find_arc_deterministic(const _State* state, unsigned unit,
                                       float& bo_weight) const;

    int  next_state(const _State* state, unsigned unit) const;

    const _State* corresponding_state(const _State* state) const;

    inline void append_arc(FSTArcVector& arcs,
                           int p, int q, int io, float w) const;

    inline FSTArcIter single_arc(int p, int q, int io, float w) const;
    inline FSTArcIter single_arc(int p, int q, int i, int o, float w) const;

    void cache_insert_at_head(_CacheEntry* e) const;
    void cache_move_to_head(_CacheEntry* e) const;

    float adjust(float weight, float scale) const {
        if (weight != _semiring->zero)
            weight *= scale;
        return weight;
    }

};

FSTGeneric
FSTLoadNgram2(istream& is, const char* header)
{
    return new FSTNgram2Rep(is, header);
}

// There is one _Arc for every n-gram probability that exists.  This is
// currently 6 bytes, as small as we can get without quantizing the probability
// (weight) down to a 16-bit value.  We play games with unaligned_weight rather
// than use a float directly to avoid a 2-byte gap between unit and the weight
// due to the alignment requirements of a float.
//
struct FSTNgram2Rep::_Arc {
    unsigned short unit;
    char unaligned_weight[sizeof(float)];
    
    _Arc() {}

    _Arc(unsigned u) : unit(u) {}

    float weight() const {
        float weight;
        memcpy(&weight, unaligned_weight, sizeof(float));
        return weight;
    }
    
    void set_weight(float weight) {
        memcpy(unaligned_weight, &weight, sizeof(float));
    }

    bool operator < (const _Arc& a) const {
        return unit < a.unit;
    }

#if 0    
    struct Compare {
        bool operator()(const _Arc& a, const _Arc& b) const {
            return a.unit < b.unit;
        }
    };
#endif
};

// There is one _State for every n-gram context (history).  The range
// [arcs_begin, arcs_end) specify the available arcs, sorted by _Arc::Compare so
// that we can look up a particular unit with a binary search.  The unit and
// parent slots represent a linked list containing the state's context; this is
// used when computing the next state for an arc.
//
struct FSTNgram2Rep::_State {
    _Arc* arcs_begin;
    _Arc* arcs_end;
    unsigned short unit;
    unsigned short depth;
    const _State* parent;
    const _State* backoff;
    float backoff_weight;
    float final_weight;

    // Return the arc with the particular unit label, or 0 for none.  This uses
    // a binary search.
    //
    const _Arc* find(unsigned unit) const {
        _Arc key(unit);
        const _Arc* arc =
            lower_bound(arcs_begin, arcs_end, key);
        return (arc != arcs_end && arc->unit == unit) ? arc : 0;
    }

    // Compute the hash value of the state.
    //
    unsigned compute_hash() const {
        const _State* state;
        unsigned result = 0;
        for (state = this; state->unit; state = state->parent) {
            result = HASH_C2*result + HASH_C1*state->unit;
        }
        return result;
        
    }

    // Compute the hash value of the state, limiting the length of the context
    // to length.
    //
    unsigned compute_hash(unsigned length) const {
        const _State* state;
        unsigned result = 0;
        for (state = this; length-- && state->unit; state = state->parent) {
            result = HASH_C2*result + HASH_C1*state->unit;
        }
        return result;
        
    }

    // Compute the hash value of the state with unit appended to the context,
    // limiting the length of the context to length.
    //
    unsigned compute_hash(unsigned unit, unsigned length) const {
        const _State* state;
        unsigned result = HASH_C1*unit;
        for (state = this; --length && state->unit; state = state->parent) {
            result = HASH_C2*result + HASH_C1*state->unit;
        }
        return result;
    }

    // Return whether this state's context is equal to b's context limited to
    // length.
    //
    bool equal(const _State* b, unsigned length) const {
        const _State* a = this;
        while (length > 0 && a->unit == b->unit && a->unit) {
            length--;
            a = a->parent;
            b = b->parent;
        }
        return a->unit == 0 && length == 0;     // end of a && end of b
    }
    // Return whether this state's context is equal to b's context with unit
    // appended and limited to length.
    //
    bool equal(unsigned unit, const _State* b, unsigned length) const {
        const _State* a = this;
        if (a->unit == unit) {
            a = a->parent;
            while (--length && a->unit == b->unit && a->unit) {
                a = a->parent;
                b = b->parent;
            }
        }
        return a->unit == 0 && length == 0;     // end of a && end of b
    }
};

// Structure to hold information about classed words.  The link is to form a
// linked list enumerating all the words in a given class.
//
struct FSTNgram2Rep::_ClassedWord {
    unsigned word;
    unsigned unit;
    float weight;
    _ClassedWord* link;
    
    _ClassedWord(unsigned word_, unsigned unit_, float weight_)
        : word(word_), unit(unit_), weight(weight_), link(0) {}

    FSTAllocator_new_delete(_ClassedWord);
};

// Structure to hold a single dynamic arc for creating an FSTArcIter when in
// deterministic mode.
//
struct FSTNgram2Rep::_SingleArcIterData : FSTReferenceCounted {
    FSTArc arc;
    FSTArc* arc_ptr;
    
    _SingleArcIterData(int p, int q, int i, int o, float w)
        : arc(p, q, i, o, w), arc_ptr(&arc) {}
    virtual ~_SingleArcIterData() {}
    FSTAllocator_new_delete(_SingleArcIterData);
};

// Append a new arc to a vector of arcs.
//
inline void
FSTNgram2Rep::append_arc(FSTArcVector& arcs,
                         int p, int q, int io, float w) const {
    if (_inverted) w = -w;
    arcs.push_back(new FSTArc(p, q, io, io, w));
}

// Create an FSTArcIter containing a single arc.
//

inline FSTArcIter
FSTNgram2Rep::single_arc(int p, int q, int io, float w) const {
    if (_inverted) w = -w;
    _SingleArcIterData* arc_data = new _SingleArcIterData(p, q, io, io, w);
    return FSTArcIter(&arc_data->arc_ptr, &arc_data->arc_ptr + 1, arc_data);
}

inline FSTArcIter
FSTNgram2Rep::single_arc(int p, int q, int i, int o, float w) const {
    if (_inverted) w = -w;
    _SingleArcIterData* arc_data = new _SingleArcIterData(p, q, i, o, w);
    return FSTArcIter(&arc_data->arc_ptr, &arc_data->arc_ptr + 1, arc_data);
}

// Create an FSTArcIter containing no arcs.
//
inline static FSTArcIter no_arcs()
{
    return FSTArcIter();
}

// The constructor.
//
FSTNgram2Rep::FSTNgram2Rep(istream& is, const char* header)
    : _deterministic(false)
    , _inverted(false)
#ifdef HASH_STATS
    , _store_probes(0), _store_count(0)
    , _hit_probes(0), _hit_count(0)
    , _miss_probes(0), _miss_count(0)
#endif
#ifdef CACHE_STATS
    , _cache_weight_try(0), _cache_weight_hit(0)
    , _cache_next_try(0), _cache_next_hit(0)
#endif
{
    const char* h;

    _class_name = "FSTNgram2";
    _semiring = FSTMinPlus;

    if (strstr(header, "Version=2") == 0) {
        FSTFatal("FSTNgram2: only Version=2 supported");
    }

    float scale = 1;
    if ((h = strstr(header, "WeightScale=")) != 0) {
        sscanf(h, "WeightScale=%f", &scale);
    }

    if (strstr(header, "WeightOffset=") != 0) {
        FSTWarn("FSTNgram2: WeightOffset not supported");
    }

    if (strstr(header, "Deterministic") != 0) {
        _deterministic = true;
    }

    // Load n of n-gram.
    //
    _n = FSTLoadUnsignedShort(is);

    // Load unit (string, index) list, and build up label_map.
    //
    const int mit_initial_final = FSTstring_map["<>"];
    const int cmu_initial       = FSTstring_map["<s>"];
    const int cmu_final         = FSTstring_map["</s>"];
    _initial_unit = 0;
    _final_unit = 0;
    string s;
    fst::map<int,int> label_map;
    int n_units = FSTLoadUnsignedShort(is);
    while (n_units-- > 0) {
        FSTLoadString(is, s);
        int unit = FSTstring_map[s];
        label_map[FSTLoadUnsignedShort(is)] = unit;
        _alphabet.insert(unit);
        if (unit == mit_initial_final) {
            _initial_unit = _final_unit = unit;
        } else if (unit == cmu_initial && !_initial_unit) {
            _initial_unit = unit;
        } else if (unit == cmu_final && !_final_unit) {
            _final_unit = unit;
        }
    }
    _alphabet.erase(_initial_unit);
    _alphabet.erase(_final_unit);
    if (!_deterministic) {
        _alphabet.insert(0);
    }

    // Load word class information.
    //
    load_class_info(is, label_map, scale);

    // Load initial state.
    //
    _initial = FSTLoadInt(is);

    // Load states and their transitions.
    //
    _n_states = _n_static_states = FSTLoadInt(is);

    hash_table_init();

    _n_arcs = 0;
    _states.resize(_n_states);
    for (int p = 0; p < _n_states; ++p) {
        
        _State* state = &_states[p];
        
        state->final_weight = adjust(FSTLoadFloat(is), scale);
        
        int backoff_state = FSTLoadInt(is);
        if (backoff_state >= 0) {
            state->backoff = &_states[backoff_state];
            state->backoff_weight = adjust(FSTLoadFloat(is), scale);
            _n_arcs++;
        } else {
            state->backoff = 0;
        }
        
        int unit = label_map[FSTLoadUnsignedShort(is)];
        assert(unit < (1<<16));
        state->unit = unit;
        
        int parent_state = FSTLoadInt(is);
        if (parent_state >= 0) {
            state->parent = &_states[parent_state];
        } else {
            state->parent = 0;
        }
        
        int n_arcs = FSTLoadUnsignedShort(is);
        _n_arcs += n_arcs;
        
        if (n_arcs > 0) {
            void* where = FSTAllocator::allocate(n_arcs * sizeof(_Arc));
            state->arcs_begin = new(where) _Arc[n_arcs];
        } else {
            state->arcs_begin = 0;
        }
        state->arcs_end = state->arcs_begin + n_arcs;
        for (int a = 0; a < n_arcs; ++a) {
            _Arc* arc = state->arcs_begin + a;
            int unit = label_map[FSTLoadUnsignedShort(is)];
            assert(unit < (1<<16));
            arc->unit = unit;
            arc->set_weight(adjust(FSTLoadFloat(is), scale));
            assert(arc->unit != 0);
        }
        sort(state->arcs_begin, state->arcs_end);

    }

    // Set cache capacity.
    //
    _cache_capacity = _max(_n_arcs / 100, 100000);
#ifndef NO_HASH_SET
    _cache_set.resize(_cache_capacity);
#endif

    // Compute and store depth [0,n-1] for each state, and store state in hash
    // table.
    //
#ifdef HASH_STATS
    cerr << "hashing...";
#endif
    for (int p = 0; p < _n_states; p++) {
        
#ifdef HASH_STATS
        if ((p % 100000) == 0) cerr << p << "...";
#endif
        
        _State* state = &_states[p];
        
        hash_table_store(state);
        
        unsigned depth = 0;
        for (const _State* node = state; (node = node->parent) != 0;) {
            depth++;
        }
        state->depth = depth;
        
        assert(depth > 0 || state->unit == 0);
    }
#ifdef HASH_STATS
    cerr << "done" << endl;
#endif
}

void
FSTNgram2Rep::load_class_info(istream& is, const fst::map<int,int>& label_map, float scale)
{
    int n_classes = FSTLoadUnsignedShort(is);

    if (n_classes == 0) return;

    // Load word class info, storing in tmp.
    //
    string s;
    int max_label = 0, n_classed_words = 0;
    _ClassedWordVector tmp;
#ifndef NO_HASH_MAP
    _class_start_map.resize(n_classes);
#endif
#ifndef NO_HASH_SET
    _class_end_set.resize(n_classes);
#endif
    while (n_classes-- > 0) {

        fst::map<int,int>::const_iterator i =
            label_map.find(FSTLoadUnsignedShort(is));
        assert(i != label_map.end());
        int unit = (*i).second;
        max_label = _max(max_label, unit);

        _alphabet.erase(unit);

        int n_words = FSTLoadUnsignedShort(is);
        n_classed_words += n_words;
        while (n_words-- > 0) {
            FSTLoadString(is, s);
            int word = FSTstring_map[s];
            max_label = _max(max_label, word);
            _alphabet.insert(word);
            tmp.push_back(new _ClassedWord(word, unit, adjust(FSTLoadFloat(is), scale)));
        }

        // Construct sets containing <CLASS> and </CLASS> labels.
        //
        string label;
        int label_index;

        label = "<";
        label += FSTstring_map[unit];
        label += ">";
        _class_start_map[label_index = FSTstring_map[label]] = unit;
        _alphabet.insert(label_index);

        label = "</";
        label += FSTstring_map[unit];
        label += ">";
        _class_end_set.insert(label_index = FSTstring_map[label]);
        _alphabet.insert(label_index);

    }

    // Move word class info from tmp to _classed_words, where it can be rapidly
    // looked up via word or unit.
    //
    _classed_words.clear();
    _classed_words.resize(max_label + 1);
    for (_ClassedWordVector::iterator i = tmp.begin();
         i != tmp.end(); ++i) {
        _ClassedWord* classed_word = *i;
        _classed_words[classed_word->word] = classed_word;
        classed_word->link = _classed_words[classed_word->unit];
        _classed_words[classed_word->unit] = classed_word;
    }
}

void
FSTNgram2Rep::add_classed_word(int word, int unit, float weight)
{
    // Grow _classed_words vector if necessary.
    //
    while (int(_classed_words.size()) <= word)
        _classed_words.push_back(NULL);

    // Insert into _classed_words
    //
    assert(_classed_words[word] == NULL);
    _classed_words[word] = new _ClassedWord(word, unit, weight);
    _classed_words[word]->link = _classed_words[unit];
    _classed_words[unit] = _classed_words[word];
    _alphabet.insert(word);
}

// The destructor.
//
FSTNgram2Rep::~FSTNgram2Rep()
{
    //FSTAllocator::status(cerr);
    for (int p = 0; p < _n_static_states; ++p) {
        _State* state = &_states[p];
        if (state->arcs_begin) {
            int n_arcs = state->arcs_end - state->arcs_begin;
            FSTAllocator::deallocate(state->arcs_begin, n_arcs * sizeof(_Arc));
            //delete[] _states[p].arcs_begin;
        }
    }
    
    _ClassedWordVector tmp;
    for (int i = 0; i < int(_classed_words.size()); ++i) {
        _ClassedWord* cw = _classed_words[i];
        if (cw && int(cw->word) == i)
            tmp.push_back(cw);
    }
    for (int i = 0; i < int(tmp.size()); ++i)
        delete tmp[i];

    delete _hash_table;

#ifdef HASH_STATS
    cerr.form("hash statistics (%d buckets)\n", _hash_mask + 1);
    cerr.form("C1 = %d (%x), C2 = %d (%x)\n",
              HASH_C1, HASH_C1, HASH_C2, HASH_C2);
    cerr.form("store: %.2f (%d/%d)\n",
          float(_store_probes) / float(_max(_store_count, 1U)),
          _store_probes, _store_count);
    cerr.form("hit:   %.2f (%d/%d)\n",
          float(_hit_probes) / float(_max(_hit_count, 1U)),
          _hit_probes, _hit_count);
    cerr.form("miss:  %.2f (%d/%d)\n",
          float(_miss_probes) / float(_max(_miss_count, 1U)),
          _miss_probes, _miss_count);
    cerr.form("arc size   = %d\n", sizeof(_Arc));
    cerr.form("state size = %d\n", sizeof(_State));
#endif
}

void
FSTNgram2Rep::dump_binary(ostream& os) const
{
    // Dump header.
    //
    os << "#FSTNgram2 Version=2";
    if (_deterministic)
        os << " Deterministic";
    os << endl;

    // Dump order (n) of n-gram.
    //
    FSTDumpUnsignedShort(_n, os);

    // Dump list of (label, index) for all units.
    //
    fst::set<int> units;
    for (int p = 0; p < _n_static_states; p++) {
        const _State* sp = &_states[p];
        if (sp->unit) {
            units.insert(sp->unit);
        }
        for (_Arc* ap = sp->arcs_begin; ap != sp->arcs_end; ap++) {
            units.insert(ap->unit);
        }
    }
    int n_units = units.size();
    assert(n_units < (1<<16));
    FSTDumpUnsignedShort(n_units, os);
    for (fst::set<int>::const_iterator i = units.begin(); i != units.end(); ++i) {
        FSTDumpString(FSTstring_map[*i], os);
        FSTDumpUnsignedShort(*i, os);
    }

    // Dump class information.
    //
    int n_classes = 0, n_classed_words = 0;
    for (unsigned i = 0; i < _classed_words.size(); i++) {
        const _ClassedWord* cw = _classed_words[i];
        if (cw && cw->word != i) {
            n_classes++;
        }
    }
    FSTDumpUnsignedShort(n_classes, os);
    if (n_classes > 0) {
        for (unsigned i = 0; i < _classed_words.size(); i++) {
            const _ClassedWord* cw = _classed_words[i];
            if (cw && cw->word != i) {
                int n_words = 0;
                for (const _ClassedWord* cw2 = cw; cw2 != 0; cw2 = cw2->link) {
                    n_words++;
                }
                FSTDumpUnsignedShort(cw->unit, os);
                FSTDumpUnsignedShort(n_words, os);
                for (const _ClassedWord* cw2 = cw; cw2 != 0; cw2 = cw2->link) {
                    FSTDumpString(FSTstring_map[cw2->word], os);
                    FSTDumpFloat(cw2->weight, os);
                }
                n_classed_words += n_words;
            }
        }
    }

    // Dump initial state.
    //
    FSTDumpInt(_initial, os);

    // Dump states.
    //
    FSTDumpInt(_n_static_states, os);
    for (int p = 0; p < _n_static_states; p++) {
        const _State* sp = &_states[p];

        // Dump final weight.
        //
        FSTDumpFloat(sp->final_weight, os);

        // Dump backoff state (-1 for none) and weight if backoff state exists.
        //
        if (sp->backoff) {
            FSTDumpInt(STATE(sp->backoff), os);
            FSTDumpFloat(sp->backoff_weight, os);
        } else {
            FSTDumpInt(-1, os);
        }

        // Dump immediate context.
        //
        FSTDumpUnsignedShort(sp->unit, os);

        // Dump parent.
        //
        FSTDumpInt(sp->parent ? STATE(sp->parent) : -1, os);

        // Dump n-grams.
        //
        unsigned n_arcs = sp->arcs_end - sp->arcs_begin;
        assert(n_arcs < (1<<16));
        FSTDumpUnsignedShort(n_arcs, os);
        for (const _Arc* ap = sp->arcs_begin; ap != sp->arcs_end; ap++) {
            FSTDumpUnsignedShort(ap->unit, os);
            FSTDumpFloat(ap->weight(), os);
        }
    }
}

void
FSTNgram2Rep::hash_table_init()
{
    // Set size to be power of 2 between 1.33 and 2.67 times larger
    // than number of states.
    //
    unsigned x = (_n_static_states * 4) / 3;
    unsigned size = 1;
    while (x) {
        size <<= 1;
        x >>= 1;
    }

#ifdef HASH_STATS
    cerr << "hash table size " << size << endl;
#endif

    _hash_mask = size - 1;
    _hash_table = new const _State*[size];
    for (unsigned h = 0; h < size; h++) {
        _hash_table[h] = 0;
    }
}

const fst::set<int>&
FSTNgram2Rep::input_alphabet() const
{
    return _alphabet;
}

const fst::set<int>&
FSTNgram2Rep::output_alphabet() const
{
    return _alphabet;
}

bool
FSTNgram2Rep::is_final(int p) const
{
    FST_ASSERT_VALID_STATE(p);
    if (p >= _n_static_states) {
        return true; // this is a hack to allow ending in the middle of a dynamic class!!!
    } else {
        return _states[p].final_weight != _semiring->zero;
    }
}

float
FSTNgram2Rep::final_weight(int p) const
{
    FST_ASSERT_VALID_STATE(p);
    if (p >= _n_static_states) {
        float w = _states[p - _n_static_states].final_weight;
        return (w == _semiring->zero || !_inverted) ? w : -w;
    } else {
        float w = _states[p].final_weight;
        return (w == _semiring->zero || !_inverted) ? w : -w;
    }
}

float
FSTNgram2Rep::potential(int p) const
{
    return _semiring->one;
}

void
FSTNgram2Rep::force() const
{
}

void
FSTNgram2Rep::reset() const
{
#ifdef CACHE_STATS
    cerr.form("ngram_cache: %d/%d, %.3f %d, %.3f %d\n",
              _cache_set.size(),
              _cache_capacity,
              float(_cache_weight_hit) / float(_max(1U, _cache_weight_try)),
              _cache_weight_try,
              float(_cache_next_hit) / float(_max(1U, _cache_next_try)),
              _cache_next_try);
    _cache_weight_hit = _cache_weight_try = 0;
    _cache_next_hit = _cache_next_try = 0;
#endif
    _cache_set.clear();
    _cache_list.next = _cache_list.prev = &_cache_list;
}

void
FSTNgram2Rep::determinize()
{
    _deterministic = true;
    _alphabet.erase(0);
}

void
FSTNgram2Rep::invert()
{
    _deterministic = true;
    _inverted = true;
    _alphabet.erase(0);
}

FSTArcIter
FSTNgram2Rep::arcs_out(int p, FSTArcOrder order) const
{
    assert(p >= 0);
    assert(p < int(_states.size()));

    const _State* state = &_states[p];

    if (_deterministic) {
        FSTFatal("FSTNgram2Rep: arcs_out not implemented for deterministic");
    }

    FSTArcIterData* arc_data = new FSTArcIterData;
    FSTArcVector& arcs = arc_data->arcs;

    // Backoff epsilon arc.
    //
    if (state->backoff) {
        append_arc(arcs, p, STATE(state->backoff), 0, state->backoff_weight);
    }

    // Non-epsilon arcs.  Expand classes to words if classes in use.
    //
    for (_Arc* arc = state->arcs_begin; arc != state->arcs_end; ++arc) {
    int next = next_state(state, arc->unit);
        _ClassedWord* cw = 
            _classed_words.empty() ? 0 : _classed_words[arc->unit];
        if (cw) {
            while (cw) {
                append_arc(arcs, p, next, cw->word, arc->weight() + cw->weight);
                cw = cw->link;
            }
        } else {
            append_arc(arcs, p, next, arc->unit, arc->weight());
        }
    }

    // Sort arcs if requested.
    //
    if (order != FSTNone) {
        sort(arcs.begin(), arcs.end(), FSTArcCompareInput());
    }

    return FSTArcIter(arc_data);
}

FSTArcIter
FSTNgram2Rep::arcs_out_with_input(int p, int i, FSTArcMode m) const
{
    assert(p >= 0);
    assert(p < _n_states);

    // Have not implemented FSTLessEqual or FSTGreaterEqual yet.
    //
    assert(m == FSTEqual);

    const _State* state = &_states[p];

    // Check to see if we are in a special state for <CLASS>:eps *:* ... </CLASS>.
    //
    if (p >= _n_static_states) {
        if (_class_end_set.count(i)) {
            return single_arc(p, p - _n_static_states, i, i, 0);
        } else if (i > 0) {
            return single_arc(p, p, i, i, 0);
        } else {
            return no_arcs();
        }
    }

    // Check for <CLASS> input label and transition to special state.
    //
    _IntIntMap::const_iterator csmi = _class_start_map.find(i);
    if (csmi != _class_start_map.end()) {
        _n_states = 2*_n_static_states;
        float extra_weight = 0;
        int unit = (*csmi).second;
        const _Arc* arc = find_arc(state, unit, extra_weight);
        if (arc) {
            return single_arc(p, next_state(state, arc->unit) + _n_static_states,
                              i, i, arc->weight() + extra_weight);
        } else {
            return no_arcs();
        }
    }

    // Normal, non <CLASS> </CLASS> processing...
    
    if (i == 0) {

        if (!_deterministic && state->backoff) {
            return single_arc(p, STATE(state->backoff), 0,
                              state->backoff_weight);
        } else {
            return no_arcs();
        }

    } else {

        float extra_weight = 0;
        const _Arc* arc = 0;

        const _ClassedWord* cw = 0;
        if (!_classed_words.empty() && i < int(_classed_words.size())) {
            cw = _classed_words[i];
        }

        if (cw) {
            if (i == int(cw->word)) {
                extra_weight = cw->weight;
                arc = find_arc(state, cw->unit, extra_weight);
            }
        } else {
            arc = find_arc(state, i, extra_weight);
        }
        if (arc) {
            return single_arc(p, next_state(state, arc->unit), i,
                  arc->weight() + extra_weight);
        } else {
            return no_arcs();
        }
    }

}

bool
FSTNgram2Rep::ngram_weight(int p, int input, float& weight) const
{
    pair<_CacheSet::iterator, bool> status =
        _cache_set.insert(_CacheEntry(p, input));

    _CacheEntry* entry = _cache_entry =
        const_cast<_CacheEntry*>(&*status.first);

#ifdef CACHE_STATS
    _cache_weight_try++;
#endif

    if (status.second) {

        // Cache miss.

        cache_insert_at_head(entry);

        const _State* state = &_states[p];
        
        const _ClassedWord* cw = 0;
        if (!_classed_words.empty() && input < int(_classed_words.size())) {
            cw = _classed_words[input];
        }
        
        const _Arc* arc = 0;
        if (cw) {
            if (input == int(cw->word)) {
                entry->weight = cw->weight;
                arc = find_arc_deterministic(state, cw->unit, entry->weight);
            }
        } else {
            entry->weight = 0;
            arc = find_arc_deterministic(state, input, entry->weight);
        }
        
        if (arc) {
            entry->weight += arc->weight();
            weight = entry->weight;
            entry->next_state = -1;             // lazy value
            return true;
        } else {
            entry->next_state = -2;             // no transition value
            return false;
        }

    } else {

        // Cache hit.

#ifdef CACHE_STATS
        _cache_weight_hit++;
#endif

        cache_move_to_head(entry);

        if (entry->next_state != -2) {
            weight = entry->weight;
            return true;
        } else {
            return false;
        }

    }
}

int
FSTNgram2Rep::ngram_next_state() const
{
    assert(_cache_entry->next_state != -2);

#ifdef CACHE_STATS
    _cache_next_try++;
#endif


    if (_cache_entry->next_state >= 0) {
#ifdef CACHE_STATS
        _cache_next_hit++;
#endif
        return _cache_entry->next_state;
    }

    int input = _cache_entry->input;
    int unit = 0;

    const _ClassedWord* cw = 0;
    if (!_classed_words.empty() && input < int(_classed_words.size())) {
        cw = _classed_words[input];
    }

    if (cw) {
        assert(input == int(cw->word));
        unit = cw->unit;
    } else {
        unit = input;
    }

    assert(unit != 0);

    _cache_entry->next_state =
        next_state(&_states[_cache_entry->state], unit);

    return _cache_entry->next_state;
}

void
FSTNgram2Rep::cache_insert_at_head(_CacheEntry* e) const
{
    if (int(_cache_set.size()) > _cache_capacity) {
        // remove tail
        _CacheEntry* t = static_cast<_CacheEntry*>(_cache_list.prev);
        t->prev->next = t->next;
        t->next->prev = t->prev;
        _cache_set.erase(*t);
    }
    // insert at head
    e->next = _cache_list.next;
    e->prev = &_cache_list;
    _cache_list.next = e;
    e->next->prev = e;
}

void
FSTNgram2Rep::cache_move_to_head(_CacheEntry* e) const
{
    if (_cache_list.next != e) {
        // Remove from previous location.
        e->prev->next = e->next;
        e->next->prev = e->prev;
        // Insert at head.
        e->next = _cache_list.next;
        e->prev = &_cache_list;
        _cache_list.next = e;
        e->next->prev = e;
    }
}

FSTArcIter
FSTNgram2Rep::arcs_out_with_output(int p, int o, FSTArcMode m) const
{
    return arcs_out_with_input(p, o, m);
}

// Find unit at state, or if deterministic follow backoff if necessary.
//
const FSTNgram2Rep::_Arc*
FSTNgram2Rep::find_arc(const _State* state, unsigned unit,
               float& bo_weight) const
{
    const _Arc* arc = state->find(unit);

    if (arc || !_deterministic) return arc;

    for (;;) {
        if (!state->backoff) return 0;
        bo_weight += state->backoff_weight;
        state = state->backoff;
        arc = state->find(unit);
        if (arc) return arc;
    }
}

// Find unit at state, following backoff if necessary.
//
const FSTNgram2Rep::_Arc*
FSTNgram2Rep::find_arc_deterministic(const _State* state, unsigned unit,
                                     float& bo_weight) const
{
    const _Arc* arc = state->find(unit);

    if (arc) return arc;

    for (;;) {
        if (!state->backoff) return 0;
        bo_weight += state->backoff_weight;
        state = state->backoff;
        arc = state->find(unit);
        if (arc) return arc;
    }
}

int
FSTNgram2Rep::next_state(const _State* state, unsigned unit) const
{
    // Construct context of next state.
    //
    _State context;
    context.unit = unit;
    context.parent = state;

    unsigned length = state->depth + 1;
    if (length > _n - 1) length = _n - 1;

    // Lookup in linear hash table, reducing length of context until
    // found or empty context reached.
    //
    while (length > 0) {
        unsigned h = state->compute_hash(unit, length) & _hash_mask;
#ifdef HASH_STATS
        unsigned probes = 1;
#endif
        while (_hash_table[h]) {
            if (_hash_table[h]->equal(unit, state, length)) {
#ifdef HASH_STATS
                _hit_count++;
                _hit_probes += probes;
#endif
                return STATE(_hash_table[h]);
            }
            h = (h + 1) & _hash_mask;
#ifdef HASH_STATS
            probes++;
#endif
        }
        length--;
#ifdef HASH_STATS
        _miss_count++;
        _miss_probes += probes;
#endif
    }

    // Reached empty context.
    //
    return 0;
}

void
FSTNgram2Rep::hash_table_store(const _State* state)
{
    assert((state->parent && state->unit != 0) || state->unit == 0);

    unsigned h = state->compute_hash() & _hash_mask;

#ifdef HASH_STATS
    _store_count++;
    _store_probes++;
#endif
    while (_hash_table[h]) {
        h = (h + 1) & _hash_mask;
#ifdef HASH_STATS
        _store_probes++;
#endif
    }
    _hash_table[h] = state;
}

// Find state in this corresponding to state from another n-gram.  By
// corresponding state we mean the state with the longest matching context
// (history).
//
const FSTNgram2Rep::_State*
FSTNgram2Rep::corresponding_state(const _State* state) const
{
    unsigned length = state->depth;
    if (length > _n - 1) length = _n - 1;

    // Lookup in linear hash table, reducing length of context until
    // found or empty context reached.
    //
    while (length > 0) {
        unsigned h = state->compute_hash(length) & _hash_mask;
        while (_hash_table[h]) {
            if (_hash_table[h]->equal(state, length)) {
                return _hash_table[h];
            }
            h = (h + 1) & _hash_mask;
        }
        length--;
    }

    // Reached empty context (root node, state 0).
    //
    return &_states[0];
}

void
FSTNgram2Rep::subtract(const FSTNgramInt* xx)
{
    const FSTNgram2Rep* x = dynamic_cast<const FSTNgram2Rep*>(xx);
    assert(x != 0);

    // We now must be used deterministically.
    //
    _deterministic = true;

    // Make sure vocabularies are the same.
    //
    if (_alphabet != x->_alphabet) {
        goto alphabet_mismatch;
    }

    // Make sure word classes are the same, but the within-class weights can be
    // different.  Update within-class weights.
    //
    assert(_classed_words.size() == x->_classed_words.size());
    for (int i = 0; i < int(_classed_words.size()); i++) {
        if (bool(_classed_words[i]) != bool(x->_classed_words[i])) {
            goto word_class_mismatch;
        }
        if (_classed_words[i] && int(_classed_words[i]->word) == i) {
            if (_classed_words[i]->unit != x->_classed_words[i]->unit) {
                goto word_class_mismatch;
            }
            _classed_words[i]->weight -= x->_classed_words[i]->weight;
        }
    }

    // Iterate through states.
    //
    for (int s = 0; s < _n_static_states; ++s) {
        _State* sp = &_states[s];

        // Finding corresponding state in x.
        //
        const _State* xsp;
        if (s == _initial) {
            xsp = &x->_states[x->_initial];
        } else {
            xsp = x->corresponding_state(sp);
        }

        // Update final weight.
        //
        sp->final_weight -= xsp->final_weight;

        // Update backoff weight, but only if the states are of the same depth.
        //
        if (sp->depth == xsp->depth) {
            sp->backoff_weight -= xsp->backoff_weight;
        }

        // Iterate through arcs (existing n-grams).
        //
        for (_Arc* ap = sp->arcs_begin; ap != sp->arcs_end; ap++) {

            // Find corresponding weight in x, following backoff as necessary.
            //
            float xw = 0;
            const _Arc* xap = x->find_arc_deterministic(xsp, ap->unit, xw);
            xw += xap->weight();

            // Update weight.
            //
            ap->set_weight(ap->weight() - xw);

        }

    }

    return;

 alphabet_mismatch:
    FSTFatal("FSTNgram2Rep::subtract: two n-grams must have same alphabet");

 word_class_mismatch:
    FSTFatal("FSTNgram2Rep::subtract: two n-grams must have same word classes");
}

// I have not found any real benefit to using separate quantizers for arc,
// backoff, and final weights.
//
#define SEPARATE_QUANTIZERS 0

void
FSTNgram2Rep::quantize(int bits)
{
    int n = 1 << bits;
    void* b_trainer = sq_trainer_lloyd_max_new(bits);
#if SEPARATE_QUANTIZERS
    void* f_trainer = sq_trainer_lloyd_max_new(bits);
    void* a_trainer = sq_trainer_lloyd_max_new(bits);
#else
    void* f_trainer = b_trainer;
    void* a_trainer = b_trainer;
#endif

    // Quantizer training is not probablistically weighted, which might help,
    // but would require computing state occupancy probabilities.  That would be
    // much easier if we still had the n-gram count tree.

    // Gather training data for quantizer(s).
    //
    for (unsigned p = 0; p < _states.size(); ++p) {
        const _State* state = &_states[p];
        assert(state->backoff_weight != _semiring->zero);
        assert(state->final_weight != _semiring->zero);
        sq_trainer_incorporate(b_trainer, state->backoff_weight, 1);
        sq_trainer_incorporate(f_trainer, state->final_weight, 1);
        for (const _Arc* a = state->arcs_begin; a != state->arcs_end; ++a) {
            float w = a->weight();
            assert(w != _semiring->zero);
            sq_trainer_incorporate(a_trainer, w, 1);
        }
    }

    // Train quantizer(s).
    //
    std::cerr << "training weight quantizer";
    float *b_d = NULL, *b_r = NULL;
    sq_trainer_train(b_trainer, &b_d, &b_r);
#if SEPARATE_QUANTIZERS
    float *f_d = NULL, *f_r = NULL;
    sq_trainer_train(f_trainer, &f_d, &f_r);
    float *a_d = NULL, *a_r = NULL;
    sq_trainer_train(a_trainer, &a_d, &a_r);
#else
    float *f_d = b_d, *f_r = b_r;
    float *a_d = b_d, *a_r = b_r;
#endif
    std::cerr << "done\n";

    // Quantize weights.
    //
    for (unsigned p = 0; p < _states.size(); ++p) {
        _State* state = &_states[p];
        state->backoff_weight = b_r[sq_quantize(state->backoff_weight, b_d, n)];
        state->final_weight   = f_r[sq_quantize(state->final_weight,   f_d, n)];
        for (_Arc* a = state->arcs_begin; a != state->arcs_end; ++a) {
            a->set_weight(a_r[sq_quantize(a->weight(), a_d, n)]);
            //float w = a_r[sq_quantize(a->weight(), a_d, n)];
            //memcpy(a->unaligned_weight, &w, sizeof(float));
        }
    }

    // Clean up.
    //
    sq_trainer_delete(b_trainer);
#if SEPARATE_QUANTIZERS
    sq_trainer_delete(f_trainer);
    sq_trainer_delete(a_trainer);
#endif
}

////////////////////////////////////////////////////////////////////////////////

static unsigned bitsNeeded(unsigned n)
{
    unsigned c;
    n |= n >> 1;
    n |= n >> 2;
    n |= n >> 4;
    n |= n >> 8;
    n |= n >> 16;
    for (c = 0; n; c++) {
        n &= n - 1;
    }
    return c;
}

static u32 swap32(u32 x)
{
    x = ((x >> 8) & 0x00ff00ff) | ((x & 0x00ff00ff) << 8);
    x = ( x >> 16             ) | ( x               << 16);
    return x;
}

static u16 swap16(u16 x)
{
    x = (x >> 8) | (x << 8);
    return x;
}

static void dump_u32(FILE* fp, int x, bool swap)
{
    if (swap) x = swap32(x);
    fwrite(&x, sizeof(x), 1, fp);
}

static void dump_u16(FILE* fp, int x, bool swap)
{
    assert((x >> 16) == 0);
    u16 y(x);
    if (swap) y = swap16(y);
    fwrite(&y, sizeof(y), 1, fp);
}

static void dump_s16(FILE* fp, int x, bool swap)
{
    assert((x >> 16) == 0 || (x >> 16) == -1);
    s16 y(x);
    if (swap) y = swap16(y);
    fwrite(&y, sizeof(y), 1, fp);
}

static void dump_u8(FILE* fp, int x)
{
    assert((x & 0xffffff00) == 0);
    u8 y(x);
    fwrite(&y, sizeof(y), 1, fp);
}

static void dump_pad(FILE* fp, int n)
{
    while (ftell(fp) & (n - 1)) {
        dump_u8(fp, 0);
    }
}

typedef std::pair<int,int> Pair;
typedef fst::set<Pair> PairSet;

static ostream&
operator<<(ostream& os, const PairSet& ps)
{
    os << '{';
    for (PairSet::const_iterator i = ps.begin(); i != ps.end(); ++i) {
        os << " (" << (*i).first << ',' << (*i).second << ')';
    }
    return os << " }";
}

void
FSTNgram2Rep::dump_pngram(const char* vocab_file, const char* out_file, int weight_width) const
{
    fst::list<string> vocab_list;
    ifstream is(vocab_file);
    string word;
    while (getline(is, word)) {
        vocab_list.push_back(word);
    }
    is.close();
    cerr << "loaded " << vocab_list.size() << " words from file " << vocab_file << ".\n";

    FILE* fp = fopen(out_file, "wb");
    dump_pngram(vocab_list, fp, weight_width);
    fclose(fp);
}

void
FSTNgram2Rep::dump_pngram(const fst::list<string>& vocab_list_in, FILE* fp, int weight_width) const
{
    bool swap = false;

    typedef fst::map<int,int> IntIntMap;
    typedef fst::map<string,int> StringIntMap;
    typedef fst::vector<int> IntVector;

    // Make sure initial state's unit is in vocab_list.
    fst::list<string> vocab_list(vocab_list_in);
    string initial_unit = FSTstring_map[_states[_initial].unit];
    if (find(vocab_list.begin(), vocab_list.end(), initial_unit) == vocab_list.end()) {
        cerr << "adding initial unit: " << initial_unit << '\n';
        vocab_list.push_back(initial_unit);
    }

    // Build word_index_map: FST word index -> vocab word index
    //
    IntIntMap word_map; // fst -> vocab
    StringIntMap word_string_map; // string -> vocab
    IntIntMap unit_map; // fst unit -> packed unit
    IntIntMap vocab_unit_map; // vocab -> packed unit
    u8 vocab_digest[16];
    {
        MD5 md5;
        unit_map[0] = 0;
        for (fst::list<string>::const_iterator i = vocab_list.begin();
             i != vocab_list.end(); ++i) {
            string word = *i;
            if (word.length() == 0) {
                continue;
            }
            md5.hash((u8*) word.c_str(), word.length());
            int vocab_index = word_map.size();
            int fst_index = FSTstring_map[word];
            word_map[fst_index] = vocab_index;
            word_string_map[word] = vocab_index;
            const _ClassedWord* cw =
                fst_index < (int)_classed_words.size() ? _classed_words[fst_index] : NULL;
            int fst_unit;
            if (!cw) {
                if (_class_start_map.count(fst_index)) {
                    fst_unit = _class_start_map[fst_index];
                } else if (!_class_end_set.count(fst_index)) {
                    fst_unit = fst_index;
                } else {
                    fst_unit = 0; // No unit for </$CLASS> words.
                }
            } else {
                fst_unit = cw->unit;
            }
            int packed_unit;
            if (!unit_map.count(fst_unit)) {
                packed_unit = unit_map.size();
                unit_map[fst_unit] = packed_unit;
            } else {
                packed_unit = unit_map[fst_unit];
            }
            vocab_unit_map[vocab_index] = packed_unit;
        }
        md5.finish(vocab_digest);
        cerr << "digest: ";
        for (int i = 0; i < 16; ++i) {
            format(cerr, "%02x", vocab_digest[i]);
        }
        cerr << '\n';
        int vocab_index = word_map.size();
        int fst_index = FSTstring_map["<>"];
        word_map[fst_index] = vocab_index;
        word_string_map["<>"] = vocab_index;
        int fst_unit = fst_index;
        int packed_unit = unit_map.size();
        unit_map[fst_unit] = packed_unit;

        cerr << "units: " << unit_map.size() << '\n';
    }

    typedef fst::set<int> IntSet;
    typedef fst::list<int> IntList;
    typedef fst::vector<IntIntMap> StateChildrenMap;
    StateChildrenMap state_children(_n_static_states);

    // Build tree represented by state_children.
    //
    IntSet keep_states;
    keep_states.insert(0);
    IntIntMap unit_for_state;
    for (int p = 0; p < _n_static_states; ++p) {
        const _State* ps = &_states[p];
        if (unit_map.count(ps->unit) == 0) {
            continue;
        }
        const _State* qs = ps->backoff;
        if (qs) {
            int q = qs - &_states[0];
            int u = ps->unit;
            for (const _State* rs = ps->parent; rs && rs->parent; rs = rs->parent) {
                if (unit_map.count(rs->unit) == 0) {
                    u = 0;
                    break;
                }
                u = rs->unit;
            }
            if (u == 0) {
                continue;
            }
            //cerr << q << " -> " << p << ' ' << FSTstring_map[u] << ' ' << unit_map[u] << '\n';
            assert(unit_map.count(qs->unit));
            state_children[q][unit_map[u]] = p;
            unit_for_state[p] = unit_map[u];
            keep_states.insert(p);
        }
    }

#if 0
    for (int p = 0; p < _n_static_states; ++p) {
        cout << "children[" << p << ']';
        const IntIntMap& children = state_children[p];
        for (IntIntMap::const_iterator i = children.begin(); i != children.end(); ++i) {
            cout << ' ' << (*i).first << ':' << (*i).second;
        }
        cout << '\n';
    }
#endif

    // Sanity check: backoff should point at parent.
    //
    assert(keep_states.count(_initial));
    for (int p = 0; p < _n_static_states; ++p) {
        if (keep_states.count(p)) {
            const _State* ps = &_states[p];
            const IntIntMap& children = state_children[p];
            for (IntIntMap::const_iterator i = children.begin(); i != children.end(); ++i) {
                int u = (*i).first;
                int q = (*i).second;
                assert(unit_for_state[q] == u);
                // wrong: assert(unit_map[_states[q].unit] == u);
                //assert(unit_map[_states[(*i).second].unit] == (*i).first);
                assert(_states[q].backoff == ps);
            }
        }
    }

    // Assign new state numbers in breadth-first traversal order.
    //
    int n_non_terminal_states = 0;
    IntIntMap state_map; // fst -> packed
    IntIntMap reverse_state_map; // packed -> fst
    state_map[0] = 0;
    reverse_state_map[0] = 0;
    for (int d = 1; d <= (int)_n; ++d) {
        for (int p = 0; p < _n_static_states; ++p) {
            const _State* ps = &_states[p];
            if (ps->depth != d) {
                continue;
            }
            if (keep_states.count(p) == 0) {
                continue;
            }
            int new_state = state_map.size();
            state_map[p] = new_state;
            reverse_state_map[new_state] = p;
            //cerr << p << " : " << new_state << '\n';
            if (d < (int)_n - 1) {
                n_non_terminal_states = new_state + 1;
            }
        }
    }
    cerr << "states: " << keep_states.size() << ' ' << state_map.size() << '\n';
    cerr << "non-terminal states: " << n_non_terminal_states << '\n';

    // Dynamic class tags.
    //
    PairSet open_close_set;
    PairSet close_state_set;
    for (_IntIntMap::const_iterator i = _class_start_map.begin(); i != _class_start_map.end(); ++i) {
        int fst_open_index = (*i).first;
        int fst_unit = (*i).second;
        string open_word = FSTstring_map[fst_open_index];
        string close_word = string("</") + string(FSTstring_map[fst_unit]) + string(">");
        if (word_string_map.count(open_word) == 0 || word_string_map.count(close_word) == 0) {
            continue;
        }
        int packed_open_index = word_string_map[open_word];
        int packed_close_index = word_string_map[close_word];
        open_close_set.insert(Pair(packed_open_index, packed_close_index));
        for (int p = 0; p < _n_static_states; ++p) {
            int fst_state = next_state(&_states[p], fst_unit);
            assert(state_map.count(fst_state));
            int packed_state = state_map[fst_state];
            close_state_set.insert(Pair(packed_close_index, packed_state));
        }
    }
    cerr << "open/close set size: " << open_close_set.size() << '\n';
    cerr << open_close_set << '\n';
    cerr << "close/state set size: " << close_state_set.size() << '\n';
    cerr << close_state_set << '\n';

    // Train weight quantizer.
    //
#define ZERO_BIAS 10
    void* sq = sq_trainer_lloyd_max_new(weight_width);
    for (int p = 0; p < _n_static_states; ++p) {
        if (keep_states.count(p)) {
            const _State* ps = &_states[p];
            sq_trainer_incorporate(sq, ps->final_weight,
                                   ps->final_weight == 0 ? ZERO_BIAS : 1);
            sq_trainer_incorporate(sq, ps->backoff_weight,
                                   ps->backoff_weight == 0 ? ZERO_BIAS : 1);
            for (const _Arc* a = ps->arcs_begin; a != ps->arcs_end; ++a) {
                if (unit_map.count(a->unit)) {
                    sq_trainer_incorporate(sq, a->weight(),
                                           a->weight() == 0 ? ZERO_BIAS : 1);
                }
            }
        }
    }
    float* sq_d;
    float* sq_r;
    cerr << "training weight quantizer";
    sq_trainer_train(sq, &sq_d, &sq_r);
    cerr << "done\n";

    // Pack arcs.
    //
    int unit_width = bitsNeeded(unit_map.size());
    BitStreamWriter32 arc_writer(swap);
    IntVector arc_table;
    int arc_index = 0;
    for (IntIntMap::iterator i = reverse_state_map.begin(); i != reverse_state_map.end(); ++i) {
        int p = (*i).second;
        assert(keep_states.count(p));
        const _State* ps = &_states[p];
        arc_table.push_back(arc_index);
        IntIntMap sortedArcs;
        for (const _Arc* a = ps->arcs_begin; a != ps->arcs_end; ++a) {
            if (unit_map.count(a->unit)) {
                sortedArcs[unit_map[a->unit]] =
                    sq_quantize(a->weight(), sq_d, 1 << weight_width);
            }
        }
        for (IntIntMap::iterator j = sortedArcs.begin(); j != sortedArcs.end(); ++j) {
            arc_writer.write((*j).first, unit_width);
            arc_writer.write((*j).second, weight_width);
            arc_index++;
        }
    }
    arc_writer.flush();
    arc_table.push_back(arc_index);
    int arc_index_width = bitsNeeded(arc_index + 1);

    // Compute field bit widths.
    //
    int bo_state_width = bitsNeeded(n_non_terminal_states);
    int child_state_width = bitsNeeded(state_map.size());

    // Pack child pointers.
    //
    BitStreamWriter32 child_writer(swap);
    IntVector child_table;
    int child_index = 0;
    for (IntIntMap::iterator i = reverse_state_map.begin(); i != reverse_state_map.end(); ++i) {
        int new_state = (*i).first;
        int p = (*i).second;
        if (new_state < n_non_terminal_states) {
            child_table.push_back(child_index);
            const IntIntMap& children = state_children[p];
            for (IntIntMap::const_iterator j = children.begin(); j != children.end(); ++j) {
                int u = (*j).first;
                int q = (*j).second;
                assert(u == unit_for_state[q]);
                assert(state_map.count(q) == 1);
                child_writer.write(state_map[q], child_state_width);
                child_index++;
            }
        }
    }
    child_writer.flush();
    child_table.push_back(child_index);
    assert((int)child_table.size() == n_non_terminal_states + 1);
    int child_index_width = bitsNeeded(child_index + 1);

    // Pack states.
    //
    BitStreamWriter32 state_writer(swap);
    for (IntIntMap::iterator i = reverse_state_map.begin(); i != reverse_state_map.end(); ++i) {
        int new_state = (*i).first;
        int p = (*i).second;
        const _State* ps = &_states[p];
        state_writer.write(unit_for_state[p], unit_width);
        if (!ps->backoff) {
            state_writer.write(new_state, bo_state_width);
        } else {
            int b = ps->backoff - &_states[0];
            assert(state_map.count(b) == 1);
            state_writer.write(state_map[b], bo_state_width);
        }
        state_writer.write(sq_quantize(ps->backoff_weight, sq_d, 1 << weight_width), weight_width);
        state_writer.write(sq_quantize(ps->final_weight, sq_d, 1 << weight_width), weight_width);
        //int pos = state_writer.position();
        state_writer.write(arc_table[new_state], arc_index_width);
        //cerr << "arc " << new_state << ' ' << arc_table[new_state] << ' ' << pos << '\n';
        if (new_state < n_non_terminal_states) {
            state_writer.write(child_table[new_state], child_index_width);
        }
    }
    state_writer.flush();
    assert(state_map.size() == reverse_state_map.size());

    cerr << "arc bytes: " << arc_writer.length() << '\n';
    cerr << "state bytes: " << state_writer.length() << '\n';
    cerr << "child bytes: " << child_writer.length() << '\n';
    
    cerr << "weight width: " << weight_width << '\n';
    cerr << "unit width: " << unit_width << '\n';
    cerr << "backoff state width: " << bo_state_width << '\n';
    cerr << "child state width: " << child_state_width << '\n';
    cerr << "child index width: " << child_index_width << '\n';
    cerr << "arc index width: " << arc_index_width << '\n';

    //
    // Output to file
    //

    u32 startPos = ftell(fp);

    // version
    dump_u32(fp, PNGRAM3, swap);

    // file size (set later)
    u32 sizePos = ftell(fp);
    dump_u32(fp, 0, swap);

    // state table pointer (set later)
    dump_u32(fp, 0, swap);

    // child table pointer (set later)
    dump_u32(fp, 0, swap);

    // arc table pointer (set later)
    dump_u32(fp, 0, swap);

    // initial state
    dump_u32(fp, state_map[_initial], swap);

    // number of normal (tree) states
    dump_u32(fp, state_map.size(), swap);

    // number of non-terminal (tree branch) states
    dump_u32(fp, n_non_terminal_states, swap);

    // n
    dump_u8(fp, _n);

    // unit width
    dump_u8(fp, unit_width);

    // backoff pointer width
    dump_u8(fp, bo_state_width);

    // weight width
    dump_u8(fp, weight_width);

    // arc index width
    dump_u8(fp, arc_index_width);

    // child index width
    dump_u8(fp, child_index_width);

    // child pointer width
    dump_u8(fp, child_state_width);

    // vocabulary digest
    for (int i = 0; i < 16; ++i) {
        dump_u8(fp, vocab_digest[i]);
    }

    // unit map
    dump_pad(fp, 2);
    dump_u16(fp, vocab_unit_map.size(), swap);
    for (int i = 0; i < (int)vocab_unit_map.size(); ++i) {
        assert((vocab_unit_map[i] >> 16) == 0);
        dump_u16(fp, vocab_unit_map[i], swap);
    }

    // open/close map
    dump_pad(fp, 2);
    dump_u16(fp, open_close_set.size(), swap);
    for (PairSet::const_iterator i = open_close_set.begin(); i != open_close_set.end(); ++i) {
        dump_u16(fp, (*i).first,  swap);
        dump_u16(fp, (*i).second, swap);
    }

    // close/state set
    dump_pad(fp, 2);
    dump_u16(fp, close_state_set.size(), swap);
    dump_pad(fp, 4);
    for (PairSet::const_iterator i = close_state_set.begin(); i != close_state_set.end(); ++i) {
        dump_u32(fp, (*i).first,  swap); // could be u16, but would disturb 32-bit alignment
        dump_u32(fp, (*i).second, swap);
    }

    // child end position
    dump_pad(fp, 4);
    dump_u32(fp, child_table[child_table.size() - 1], swap);

    // arc end position
    dump_pad(fp, 4);
    dump_u32(fp, arc_table[arc_table.size() - 1], swap);
    cerr << "arc end " << arc_table[arc_table.size() - 1] << '\n';

    // weight reconstruction values
    dump_pad(fp, 2);
    for (int i = 0; i < (1 << weight_width); ++i) {
        dump_s16(fp, s16(F2Q(sq_r[i], 6)), swap);
    }

    // state table
    dump_pad(fp, 4);
    u32 stateTablePos = ftell(fp) - startPos;
    fwrite(state_writer.data(), state_writer.length(), 1, fp);

    // child pointer table
    dump_pad(fp, 4);
    u32 childTablePos = ftell(fp) - startPos;
    fwrite(child_writer.data(), child_writer.length(), 1, fp);

    // arc table
    dump_pad(fp, 4);
    u32 arcTablePos = ftell(fp) - startPos;
    fwrite(arc_writer.data(), arc_writer.length(), 1, fp);

    // fix up postponed pointers
    u32 size = ftell(fp) - startPos;
    fseek(fp, sizePos, SEEK_SET);
    dump_u32(fp, size, swap);
    dump_u32(fp, stateTablePos, swap);
    dump_u32(fp, childTablePos, swap);
    dump_u32(fp, arcTablePos, swap);

    // done
    cout << "Wrote " << size << " bytes.\n";

#if 0
    {
        static char* strings[] = {
            "show",
            "songs",
            "by",
            "<$ARTISTNAME>",
            "dire",
            "straits",
            "</$ARTISTNAME>",
            NULL
        };

        int p = initial(), q;
        float w;

        for (char** word = strings; *word; ++word) {
            int fw = FSTstring_map[*word];
            int pw = word_map[fw];
            FSTArcIter a = arcs_out_with_input(p, fw, FSTEqual);
            assert(a.n() == 1);
            q = a->next;
            w = a->weight;
            cout << "T " << p << ' ' << q << ' ' << *word << ' ' << w
                 << " | " << pw << ' ' << state_map[p] << ' ' << state_map[q] << '\n';
            p = q;
        }
        cout << "F " << p << ' ' << final_weight(p) << " | " << state_map[p] << '\n';
    }
#endif
    
}

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 80
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
