/* -*- C++ -*-
 *
 * Copyright (c) 2007
 * Spoken Language Systems Group
 * MIT Computer Science and Artificial Intelligence Laboratory
 * Massachusetts Institute of Technology
 *
 * All Rights Reserved
 */

#include <iostream>
#include <sstream>
#include <fstream>
#include <cstdlib>
#include <cassert>
#include "fst/FST.h"
#include "fst/map"
#include "BitStreamWriter32.h"
#include "sq.h"
#include "MD5.h"

// echo -n PFST3 | md5sum | cut -c1-8
#define PFST3         0x4278b59f

// echo -n PFST2 | md5sum | cut -c1-8
#define PFST2         0x2d74266e

// echo -n PFST1 | md5sum | cut -c1-8
#define PFST1         0xc420ef26

// echo -n QFST1 | md5sum | cut -c1-8
#define QFST1         0xff8e81bf

#define F2Q(X,Q) u32((X*(1 << Q)) + 0.5F)

using namespace std;

typedef std::pair<int,int> Pair;
typedef fst::map<Pair,int> PairIntMap;
typedef fst::vector<Pair>  IntPairMap;
typedef fst::map<int,int> IntIntMap;
typedef fst::vector<int> IntVector;
typedef fst::set<int> IntSet;

// Compute number of bits needed to encode unsigned integer n.
//
// Derived from:
//   http://graphics.stanford.edu/~seander/bithacks.html#VariableSignExtend
//   http://graphics.stanford.edu/~seander/bithacks.html#CountBitsSetKernighan
//
static unsigned
bitsNeeded(unsigned n)
{
    unsigned c;
    n |= n >> 1;
    n |= n >> 2;
    n |= n >> 4;
    n |= n >> 8;
    n |= n >> 16;
    for (c = 0; n; c++)
        n &= n - 1;
    return c;
}

#if 0
static unsigned
nextHighestPowerOf2(unsigned v)
{
    v--;
    v |= v >> 1;
    v |= v >> 2;
    v |= v >> 4;
    v |= v >> 8;
    v |= v >> 16;
    v++;
    return v;
}
#endif

static u32
swap32(u32 x)
{
    x = ((x >> 8) & 0x00ff00ff) | ((x & 0x00ff00ff) << 8);
    x = ( x >> 16             ) | ( x               << 16);
    return x;
}

static u16
swap16(u16 x)
{
    x = (x >> 8) | (x << 8);
    return x;
}

///////////////////////////////////////////////////////////////////////////////

class ByteWriter {
public:
    virtual ~ByteWriter() = 0;
    virtual void write(const void* bytes, u32 nBytes) = 0;
            void write_u32(int x, bool swap);
            void write_u16(int x, bool swap);
            void write_s16(int x, bool swap);
            void write_u8(int x);
            void pad(int n);
    virtual void seek(u32 position) = 0;
    virtual u32  tell() const = 0;
};

ByteWriter::~ByteWriter()
{
}

void
ByteWriter::write_u32(int x, bool swap)
{
    if (swap) x = swap32(x);
    write(&x, sizeof(x));
}

void
ByteWriter::write_u16(int x, bool swap)
{
    assert((x >> 16) == 0);
    u16 y(x);
    if (swap) y = swap16(y);
    write(&y, sizeof(y));
}

void
ByteWriter::write_s16(int x, bool swap)
{
    assert((x >> 16) == 0 || (x >> 16) == -1);
    s16 y(x);
    if (swap) y = swap16(y);
    write(&y, sizeof(y));
}

void
ByteWriter::write_u8(int x)
{
    assert((x & 0xffffff00) == 0);
    u8 y(x);
    write(&y, sizeof(y));
}

void 
ByteWriter::pad(int n)
{
    while (tell() & (n - 1)) {
        write_u8(0);
    }
}

///////////////////////////////////////////////////////////////////////////////

class FileByteWriter : public ByteWriter {
    FILE* _fp;
public:
    FileByteWriter(const char* filename);
    virtual ~FileByteWriter();
    virtual void write(const void* bytes, u32 nBytes);
    virtual void seek(u32 position);
    virtual u32  tell() const;
};

FileByteWriter::FileByteWriter(const char* filename)
{
    _fp = fopen(filename, "wb");
}

FileByteWriter::~FileByteWriter()
{
    fclose(_fp);
}

void
FileByteWriter::write(const void* bytes, u32 nBytes)
{
    fwrite(bytes, 1, nBytes, _fp);
}

void
FileByteWriter::seek(u32 position)
{
    fseek(_fp, position, SEEK_SET);
}

u32
FileByteWriter::tell() const
{
    return ftell(_fp);
}

///////////////////////////////////////////////////////////////////////////////

class MemoryByteWriter : public ByteWriter {
    fst::vector<u8> _data;
    u32 _position;
public:
    MemoryByteWriter(u32 initialCapacity);
    virtual ~MemoryByteWriter();
    virtual void write(const void* bytes, u32 nBytes);
    virtual void seek(u32 position);
    virtual u32  tell() const;
    const u8* data() const;
    u32 size() const;
};

MemoryByteWriter::MemoryByteWriter(u32 initialCapacity)
{
    _data.reserve(initialCapacity);
    _position = 0;
}

MemoryByteWriter::~MemoryByteWriter()
{
}

void
MemoryByteWriter::write(const void* bytes, u32 nBytes)
{
    const u8* in = reinterpret_cast<const u8*>(bytes);
    s32 nReplace = std::min(s32(nBytes), s32(_data.size() - _position));
    if (nReplace > 0) {
        std::copy(in, in + nReplace, _data.begin() + _position);
        in += nReplace;
        nBytes -= nReplace;
        _position += nReplace;
    }
    if (nBytes > 0) {
        _data.insert(_data.end(), in, in + nBytes);
        _position += nBytes;
    }
}

void
MemoryByteWriter::seek(u32 position)
{
    if (position > _data.size()) {
        _data.insert(_data.end(), u8(0), position - _data.size());
    }
    _position = position;
}

u32
MemoryByteWriter::tell() const
{
    return _position;
}

const u8*
MemoryByteWriter::data() const
{
    return &_data[0];
}

u32
MemoryByteWriter::size() const
{
    return _data.size();
}

///////////////////////////////////////////////////////////////////////////////

class FSTPack_ArcCompare {
    const IntIntMap& _mapping;
    int _lookup(int index) const {
        IntIntMap::const_iterator i = _mapping.find(index);
        return (i != _mapping.end()) ? (*i).second : 0;
    }
public:
    FSTPack_ArcCompare(const IntIntMap& mapping)
        : _mapping(mapping) {
    }
    bool operator()(const FSTArc* x, const FSTArc* y) const {
        return _lookup(x->input) < _lookup(y->input);
    }
};

static int
constraintFromLabel(const char* label)
{
    if (label[0] == '#' && isdigit(label[1])) {
        return atoi(label + 1);
    } else {
        return 0;
    }
}

static bool
isClassLabel(const char* label)
{
    return label[0] == '$' && label[1] != 0;
}

class FSTPackArcPredicate : public FSTArcPredicate {
public:
    IntSet inputs;
    IntSet outputs;

    virtual ~FSTPackArcPredicate() {
    }
    virtual bool operator()(const FSTArc* a) const {
        return inputs.count(a->input) != 0 && outputs.count(a->output) == 0;
    }
};

static bool
needTopoSort(const FSTGeneric& fst, const FSTPackArcPredicate& predicate)
{
    for (int p = 0; p < fst->n_states(); ++p) {
        for (FSTArcIter a = fst->arcs_out(p); a.valid(); ++a) {
            if (predicate(a) && a->next <= p) {
                return true;
            }
        }
    }
    return false;
}

static FSTGeneric
uniquifyReentryLabels(const FSTGeneric& fst)
{
    IntIntMap constraintMap;

    const IntSet& inputs = fst->input_alphabet();
    for (IntSet::const_iterator i = inputs.begin(); i != inputs.end(); ++i) {
        int constraint = constraintFromLabel(FSTstring_map[*i]);
        if (constraint != 0) {
            constraintMap[*i] = constraint;
        }
    }

    IntSet classSet;
    
    const IntSet& outputs = fst->output_alphabet();
    for (IntSet::const_iterator i = outputs.begin(); i != outputs.end(); ++i) {
        if (isClassLabel(FSTstring_map[*i])) {
            classSet.insert(*i);
        }
    }

    bool anyEntry = false;
    bool anyNonUnique = false;
    PairIntMap stateConstraintCount;
    for (int p = 0; p < fst->n_states(); ++p) {
        for (FSTArcIter a = fst->arcs_out_with_input(p, 2, FSTGreaterEqual); a.valid(); ++a) {
            IntIntMap::const_iterator i = constraintMap.find(a->input);
            if (i != constraintMap.end()) {
                int c = (*i).second;
                if (classSet.count(a->output)) {
                    // class entry
                    anyEntry = true;
                } else {
                    // main reentry
                    Pair pc(p, c);
                    if (++stateConstraintCount[pc] > 1) {
                        anyNonUnique = true;
                    }
                }
            }
        }
    }
    if (!anyEntry || !anyNonUnique) {
        return fst;
    }

    // Uniquify.

    // First, assign new states for (state,constraint) for those that have count > 1.
    //
    PairIntMap stateConstraintMap;
    for (PairIntMap::const_iterator i = stateConstraintCount.begin(); i != stateConstraintCount.end(); ++i) {
        const Pair& pc = (*i).first;
        int count = (*i).second;
        if (count > 1) {
            int r = stateConstraintMap.size() + fst->n_states();
            stateConstraintMap[pc] = r;
        }
    }

    // Then, construct FST.
    //
    FSTGeneric out = FSTBasic();
    out->add_states(fst->n_states() + stateConstraintMap.size());
    out->set_initial(fst->initial());
    for (int p = 0; p < fst->n_states(); ++p) {
        out->set_final(p, fst->final_weight(p));
        for (FSTArcIter a = fst->arcs_out(p); a.valid(); ++a) {
            int q = a->next;
            IntIntMap::const_iterator i = constraintMap.find(a->input);
            int c = (i != constraintMap.end()) ? (*i).second : 0;
            Pair pc(p, c);
            PairIntMap::const_iterator j = stateConstraintMap.find(pc);
            if (j != stateConstraintMap.end()) {
                out->add_arc((*j).second, q, 0, a->output, a->weight);
                if (--stateConstraintCount[pc] == 0) {
                    out->add_arc(p, (*j).second, a->input, 0, 0);
                }
            } else {
                out->add_arc(p, q, a->input, a->output, a->weight);
            }
        }
    }

    return out;
}

///////////////////////////////////////////////////////////////////////////////

static FSTGeneric
FSTPack2(ByteWriter& byteWriter, const char* out_label_file,
         const FSTGeneric& in, int weightBits, bool swap, int verbosity)
{
    if (verbosity >= 1) {
        cerr << "version: 2\n";
    }

    // This is hard-coded in PFST2 loading/walking code inside PocketSUMMIT.
    // PFST3 adds settable weightBits support.
    //
    if (weightBits != 5) {
        FSTWarn("PFST2 requires 5 weight bits -- using 5\n");
        weightBits = 5;
    }

    // Topologically sort with respect to null (epsilon and #) transitions.  We
    // set up the arc predicate to accept epsilon, #, and #i inputs and reject
    // $Class outputs.
    //
    FSTPackArcPredicate predicate;
    predicate.inputs.insert(0);
    predicate.inputs.insert(1);
    FSTGeneric fst = in;
    if (needTopoSort(in, predicate)) {
        if (verbosity >= 2) {
            cerr << "topological sort\n";
        }
        fst =FSTTopologicalSort(in, &predicate);
    } else {
        if (verbosity >= 2) {
            cerr << "skipping topological sort\n";
        }
    }
    if (fst->n_states() == 0) {
        FSTWarn("WARNING: topological sort with respect to null arcs failed -- aborting FSTPack\n");
        return FSTBasic();
    }

    // Train Lloyd-Max (non-linear) weight quantizer.
    //
    void* trainer = sq_trainer_lloyd_max_new(weightBits);
    bool anyNonZero = false;
    for (int p = 0; p < fst->n_states(); ++p) {
        if (fst->is_final(p) && fst->final_weight(p) != 0) {
            sq_trainer_incorporate(trainer, fst->final_weight(p), 1);
        }
        for (FSTArcIter a = fst->arcs_out(p); a.valid(); ++a) {
            if (a->weight != 0) {
                sq_trainer_incorporate(trainer, a->weight, 1);
                anyNonZero = true;
            }
        }
    }
    float* weightD = 0;
    float* weightR = 0;
    if (anyNonZero) {
        cerr << "Training weight quantizer";
        sq_trainer_train(trainer, &weightD, &weightR);
        cerr << ".done.\n";
    }
    if (verbosity >= 3) {
        cerr << "weights:";
        for (int i = 0; i < (1 << weightBits); ++i) {
            cerr << ' ' << weightR[i];
        }
        cerr << '\n';
    }
    sq_trainer_delete(trainer);

    // Compute inputMap...
    //
    const IntSet* inputs = &in->input_alphabet();
    int iMax = 0;
    int tMax = 0;
    int inputMax = 0;
    IntIntMap inputMap;
    IntIntMap inputSortMap; // used to later sort arcs
    for (IntSet::const_iterator i = inputs->begin(); i != inputs->end(); ++i) {
        const char* label = FSTstring_map[*i];
        if (*i == 0 || *i == 1) {
            // do nothing
        } else if (((label[0] == 't' || label[0] == 'i') && label[1] == '.')) {
            int index;
            index = atoi(label + 2) + 2;
            inputSortMap[*i] = index;
            inputMap[*i] = index;
            if (label[0] == 'i' && index > iMax) {
                iMax = index;
            }
            if (label[0] == 't' && index > tMax) {
                tMax = index;
            }
            if (index > inputMax) {
                inputMax = index;
            }
        } else if (!constraintFromLabel(label)) {
            cerr << "unhandled input label: '" << label << "'\n";
        }
    }
    assert(tMax > iMax);
    int cMax = 0;
    for (IntSet::const_iterator i = inputs->begin(); i != inputs->end(); ++i) {
        const char* label = FSTstring_map[*i];
        int constraint = constraintFromLabel(label);
        if (constraint) {
            int index = tMax + constraint;
            inputMap[*i] = index;
            inputSortMap[*i] = -1; // put constraint arcs at beginning
            if (index > inputMax) {
                inputMax = index;
            }
            if (index > cMax) {
                cMax = index;
            }
        }
    }

    inputMap[0] = 0;
    inputMap[1] = 1;
    inputSortMap[0] = 0;
    inputSortMap[1] = 1;
    if (verbosity >= 2) {
        cerr << "iMax = " << iMax << '\n';
        cerr << "tMax = " << tMax << '\n';
        cerr << "cMax = " << cMax << '\n';
    }

    u32 nStates = fst->n_states();
    u32 nInputs = inputMax + 1;

    IntSet outputs = fst->output_alphabet();
    outputs.insert(0);

    typedef fst::map<string,int> StringIntMap;
    StringIntMap outputStringMap;
    for (IntSet::const_iterator i = outputs.begin(); i != outputs.end(); ++i) {
        outputStringMap[FSTstring_map[*i]] = 0;
    }
    int outputIndex = 0;
    for (StringIntMap::iterator i = outputStringMap.begin();
         i != outputStringMap.end();
         ++i) {
        (*i).second = outputIndex++;
    }
    IntIntMap outputIntMap;
    for (IntSet::const_iterator i = outputs.begin(); i != outputs.end(); ++i) {
        outputIntMap[*i] = outputStringMap[FSTstring_map[*i]];
    }

    typedef fst::vector<string> StringVector;
    StringVector inputStrings;
    inputStrings.resize(inputMax + 1);
    for (IntIntMap::const_iterator i = inputMap.begin(); i != inputMap.end(); ++i) {
        inputStrings[(*i).second] = FSTstring_map[(*i).first];
    }
    for (int i = 2; i <= tMax; i++) {
        char buf[16];
        sprintf(buf, "%c.%d", (i <= iMax) ? 'i' : 't', i - 2);
        inputStrings[i] = buf;
    }

    if (verbosity >= 4) {
        cerr << "inputs:\n";
        for (StringVector::iterator i = inputStrings.begin(); i != inputStrings.end(); ++i) {
            cerr << '\t' << *i << '\n';
        }
    }

    assert(nInputs == inputStrings.size());

    u32 inputBits = bitsNeeded(nInputs);
    u32 stateBits = bitsNeeded(nStates);
    u32 outputBits = bitsNeeded(outputs.size());

    if (verbosity >= 2) {
        cerr << "bit widths:"
             << " input=" << inputBits
             << " output=" << outputBits
             << " next=" << stateBits
             << '\n';
    }

    // Pack all arcs, keeping track of bit offsets for the first arc leaving
    // each states.
    //
    BitStreamWriter32 arcWriter(swap);
    fst::vector<u32> positionForState;
    typedef fst::vector<const FSTArc*> ArcVector;
    ArcVector arcVector;
    FSTPack_ArcCompare arcCompare(inputSortMap);
    for (u32 p = 0; p < nStates; ++p) {
        positionForState.push_back(arcWriter.position());

        // state finality
        arcWriter.write(fst->is_final(p), 1);
        if (fst->is_final(p)) {
            if (fst->final_weight(p) != 0) {
                int weight = sq_quantize(fst->final_weight(p), weightD, 1 << weightBits);
                arcWriter.write(1, 1);
                assert((weight >> weightBits) == 0);
                arcWriter.write(weight, weightBits);
            } else {
                arcWriter.write(0, 1);
            }
        }

        // Sort arcs using arcCompare (so that #i constraints come first).
        //
        arcVector.clear();
        FSTArcIter a = fst->arcs_out(p); // outside loop in case arcs are dynamic
        for (; a.valid(); ++a) {
            arcVector.push_back(a);
        }
        stable_sort(arcVector.begin(), arcVector.end(), arcCompare);

        // Walk sorted arcs.
        //
        for (ArcVector::const_iterator i = arcVector.begin(); i != arcVector.end(); ++i) {
            const FSTArc* a = *i;

            if (inputMap.count(a->input) == 0) {
                cerr << "unhandled input label: '" << FSTstring_map[a->input] << "' -- ignoring\n";
                continue;
            }

            // flag not finished with arcs
            arcWriter.write(1, 1);

            // input
            assert(inputMap.count(a->input) != 0);
            int mappedInput = inputMap[a->input];
            assert((mappedInput >> inputBits) == 0);
            arcWriter.write(mappedInput, inputBits);

            assert(a->input > 1 || a->next > int(p));

            if (mappedInput < 2 || mappedInput > iMax) {

                // next state
                assert((a->next >> stateBits) == 0);
                arcWriter.write(a->next, stateBits);

                // output (word)
                int output = outputIntMap[a->output];
                assert(!a->output || output);
                assert(output < int(outputStringMap.size()));
                assert((output >> outputBits) == 0);
                if (output) {
                    arcWriter.write(1, 1);
                    arcWriter.write(output, outputBits);
                } else {
                    arcWriter.write(0, 1);
                }
                    
                // weight
                if (a->weight == 0) {
                    arcWriter.write(0, 1);
                } else {
                    int weight = sq_quantize(a->weight, weightD, 1 << weightBits);
                    arcWriter.write(1, 1);
                    assert((weight >> weightBits) == 0);
                    arcWriter.write(weight, weightBits);
                }

            } else { // internal

                if (a->next == int(p)) {
                    arcWriter.write(0, 1); // no next state needed
                } else {
                    arcWriter.write(1, 1);
                    assert((a->next >> stateBits) == 0);
                    arcWriter.write(a->next, stateBits);
                    int output = outputIntMap[a->output];
                    assert(!a->output || output);
                    assert(output < int(outputStringMap.size()));
                    assert((output >> outputBits) == 0);
                    if (output) {
                        arcWriter.write(1, 1);
                        arcWriter.write(output, outputBits);
                    } else {
                        arcWriter.write(0, 1);
                    }
                }

            }
        }
        
        // flag finished with arcs
        arcWriter.write(0, 1);
    }
    arcWriter.flush();
    positionForState.push_back(arcWriter.position());

    if (verbosity >= 2) {
        u32 nArcs = fst->n_arcs();
        u32 nBits = arcWriter.position();
        cerr << "arcs: " << nBits << " bits, " << nArcs << " arcs, "
             << double(nBits)/double(nArcs) << " bits/arc\n";
    }

    // Determine shift to use with state table 1.
    //
    u32 tableBits = 16;
    u32 tableShift = 0;
    for (u32 shift = 16; shift > 0; --shift) {
        u32 cumArcBits = 0;
        bool ok = true;
        for (u32 p = 0; p < nStates; ++p) {
            if ((p & ((1 << shift) - 1)) == 0) {
                cumArcBits = 0;
            }
            u32 arcBits = positionForState[p + 1] - positionForState[p];
            u32 pos = cumArcBits + arcBits;
            if (pos >> tableBits) {
                ok = false;
                break;
            }
            cumArcBits += arcBits;
        }
        if (ok) {
            tableShift = shift;
            break;
        }
    }
    
    // Construct the state tables.
    //
    fst::vector<u32> stateTable1;
    fst::vector<u16> stateTable2;
    for (u32 p = 0, cumArcBits = 0; p < nStates; ++p) {
        if ((p & ((1 << tableShift) - 1)) == 0) {
            stateTable1.push_back(positionForState[p]);
            cumArcBits = 0;
        }
        assert((cumArcBits >> 16) == 0);
        stateTable2.push_back(cumArcBits);
        u32 arcBits = positionForState[p + 1] - positionForState[p];
        cumArcBits += arcBits;
    }
    if (verbosity >= 2) {
        cerr << "state table shift: " << tableShift << '\n';
    }

    // Construct the word table.
    //
    fst::vector<u32> wordTable;
    u32 wordOffset = 0;
    for (StringIntMap::iterator i = outputStringMap.begin();
         i != outputStringMap.end();
         ++i) {
        wordTable.push_back(wordOffset);
        wordOffset += (*i).first.length() + 1;
    }
    if (verbosity >= 2) {
        cerr << "word table size = " << wordTable.size() << '\n';
    }

    // Construct the input table.
    //
    fst::vector<u32> inputTable;
    u32 inputOffset = 0;
    for (StringVector::iterator i = inputStrings.begin(); i != inputStrings.end(); ++i) {
        inputTable.push_back(inputOffset);
        inputOffset += (*i).length() + 1;
    }
    if (verbosity >= 2) {
        cerr << "input table size = " << inputTable.size() << '\n';
    }

    ////////////////////////////////////////////////////////////////////////////
    // Output to file
    ////////////////////////////////////////////////////////////////////////////

    // version
    byteWriter.write_u32(PFST2, swap);

    // filesize (set later)
    u32 sizePos = byteWriter.tell();
    byteWriter.write_u32(0, swap);

    // state table 1 pointer (set later)
    byteWriter.write_u32(0, swap);

    // state table 2 pointer (set later)
    byteWriter.write_u32(0, swap);

    // arc data pointer (set later)
    byteWriter.write_u32(0, swap);

    // word table pointer (set later)
    byteWriter.write_u32(0, swap);

    // word string pointer (set later)
    byteWriter.write_u32(0, swap);

    // number of states
    byteWriter.write_u32(nStates, swap);
    
    // initial state
    byteWriter.write_u32(fst->initial(), swap);

    // max internal label
    byteWriter.write_u16(iMax, swap);

    // state table shift
    byteWriter.write_u8(tableShift);

    // input bit width
    byteWriter.write_u8(inputBits);

    // next state bit width
    byteWriter.write_u8(stateBits);

    // output bit width
    byteWriter.write_u8(outputBits);

    // weight reconstruction values
    byteWriter.pad(2);
    for (u32 i = 0; i < u32(1 << weightBits); ++i) {
        byteWriter.write_s16(weightR ? (s16)(F2Q(weightR[i], 6)) : 0, swap);
    }

    // state table 1
    byteWriter.pad(32);
    u32 stateTable1Pos = byteWriter.tell();
    for (fst::vector<u32>::iterator i = stateTable1.begin(); i != stateTable1.end(); ++i) {
        byteWriter.write_u32(*i, swap);
    }

    // state table 2
    byteWriter.pad(32);
    u32 stateTable2Pos = byteWriter.tell();
    for (fst::vector<u16>::iterator i = stateTable2.begin(); i != stateTable2.end(); ++i) {
        byteWriter.write_u16(*i, swap);
    }

    // arc data
    byteWriter.pad(32);
    u32 arcPos = byteWriter.tell();
    byteWriter.write(arcWriter.data(), arcWriter.length());

    // word table
    byteWriter.pad(32);
    u32 wordTablePos = byteWriter.tell();
    for (fst::vector<u32>::iterator i = wordTable.begin(); i != wordTable.end(); ++i) {
        byteWriter.write_u32(*i, swap);
    }

    // word strings
    byteWriter.pad(32);
    u32 wordStringPos = byteWriter.tell();
    for (StringIntMap::iterator i = outputStringMap.begin();
         i != outputStringMap.end();
         ++i) {
        byteWriter.write((*i).first.c_str(), (*i).first.length() + 1);
    }

    // fix up postponed pointers
    u32 size = byteWriter.tell();
    byteWriter.seek(sizePos);
    byteWriter.write_u32(size, swap);
    byteWriter.write_u32(stateTable1Pos, swap);
    byteWriter.write_u32(stateTable2Pos, swap);
    byteWriter.write_u32(arcPos, swap);
    byteWriter.write_u32(wordTablePos, swap);
    byteWriter.write_u32(wordStringPos, swap);
    if (verbosity >= 3) {
        cerr << "stateTable1Pos = " << stateTable1Pos << '\n';
        cerr << "stateTable2Pos = " << stateTable2Pos << '\n';
    }

    // done

    // optionally dump vocabulary (alphabet) file
    if (out_label_file && out_label_file[0]) {
        FILE* fp = fopen(out_label_file, "w");
        for (StringIntMap::iterator i = outputStringMap.begin();
             i != outputStringMap.end();
             ++i) {
            fputs((*i).first.c_str(), fp);
            fputc('\n', fp);
        }
        fclose(fp);
        if (verbosity >= 1) {
            cerr << "Wrote " << outputStringMap.size()
                 << " words to vocabulary file " << out_label_file << ".\n";
        }
    }

    if (verbosity >= 1) {
        cerr << "Wrote " << size << " bytes";
    }

    return fst;
}

////////////////////////////////////////////////////////////////////////////////

static string
digestString(const u8 digest[16])
{
    string result;
    for (int i = 0; i < 16; ++i) {
        char buf[3];
        sprintf(buf, "%02x", digest[i]);
        result += buf;
    }
    return result;
}

static FSTGeneric
FSTPack3(ByteWriter& byteWriter, const char* out_label_file,
         const FSTGeneric& in, int weightBits, const IntSet& externalInputs, bool swap, int verbosity)
{
    if (verbosity >= 1) {
        cerr << "version: 3\n";
    }

    FSTGeneric fst = uniquifyReentryLabels(in);

    // Topologically sort with respect to null (epsilon and #) transitions.  We
    // set up the arc predicate to accept epsilon, #, and #i inputs and reject
    // $Class outputs.
    //
    FSTPackArcPredicate predicate;
    predicate.inputs.insert(0);
    predicate.inputs.insert(1);
    const IntSet& inInputs = fst->input_alphabet();
    for (IntSet::const_iterator i = inInputs.begin(); i != inInputs.end(); ++i) {
        const char* label = FSTstring_map[*i];
        if (label[0] == '#' && label[1] >= '0' && label[1] <= '9') {
            predicate.inputs.insert(*i);
        }
    }
    const IntSet& inOutputs = fst->output_alphabet();
    for (IntSet::const_iterator i = inOutputs.begin(); i != inOutputs.end(); ++i) {
        const char* label = FSTstring_map[*i];
        if (label[0] == '$' && label[1] != 0) {
            predicate.outputs.insert(*i);
        }
    }
    if (needTopoSort(fst, predicate)) {
        if (verbosity >= 2) {
            cerr << "topological sort\n";
        }
        fst = FSTTopologicalSort(fst, &predicate);
    } else {
        if (verbosity >= 2) {
            cerr << "skipping topological sort\n";
        }
    }
    if (fst->n_states() == 0) {
        cerr << "WARNING: topological sort with respect to null arcs failed -- aborting FSTPack\n";
        return FSTBasic();
    }

    // Train Lloyd-Max (non-linear) weight quantizer.
    //
    float* weightD = 0;
    float* weightR = 0;
    if (weightBits > 0) {
        void* trainer = sq_trainer_lloyd_max_new(weightBits);
        bool anyNonZero = false;
        for (int p = 0; p < fst->n_states(); ++p) {
            if (fst->is_final(p) && fst->final_weight(p) != 0) {
                sq_trainer_incorporate(trainer, fst->final_weight(p), 1);
            }
            for (FSTArcIter a = fst->arcs_out(p); a.valid(); ++a) {
                if (a->weight != 0) {
                    sq_trainer_incorporate(trainer, a->weight, 1);
                    anyNonZero = true;
                }
            }
        }
        if (anyNonZero) {
            cerr << "Training weight quantizer";
            sq_trainer_train(trainer, &weightD, &weightR);
            cerr << ".done.\n";
        }
        if (verbosity >= 3) {
            cerr << "weights:";
            for (int i = 0; i < (1 << weightBits); ++i) {
                cerr << ' ' << weightR[i];
            }
            cerr << '\n';
        }
        sq_trainer_delete(trainer);
    }

    // Compute inputMap...
    //
    const IntSet* inputs = (externalInputs.size() > 0) ? &externalInputs : &inInputs;
    int iMax = 0;
    int tMax = 0;
    int inputMax = 0;
    IntIntMap inputMap;
    IntIntMap inputSortMap; // used to later sort arcs
    for (IntSet::const_iterator i = inputs->begin(); i != inputs->end(); ++i) {
        const char* label = FSTstring_map[*i];
        if (*i == 0 || *i == 1) {
            // do nothing
        } else if (((label[0] == 't' || label[0] == 'i') && label[1] == '.')) {
            int index;
            index = atoi(label + 2) + 2;
            inputSortMap[*i] = index;
            inputMap[*i] = index;
            if (label[0] == 'i' && index > iMax) {
                iMax = index;
            }
            if (label[0] == 't' && index > tMax) {
                tMax = index;
            }
            if (index > inputMax) {
                inputMax = index;
            }
        } else if (!constraintFromLabel(label)) {
            cerr << "unhandled input label: '" << label << "'\n";
        }
    }
    assert(tMax > iMax);
    int cMax = 0;
    for (IntSet::const_iterator i = inputs->begin(); i != inputs->end(); ++i) {
        const char* label = FSTstring_map[*i];
        int constraint = constraintFromLabel(label);
        if (constraint) {
            int index = tMax + constraint;
            inputMap[*i] = index;
            inputSortMap[*i] = -1; // put constraint arcs at beginning
            if (index > inputMax) {
                inputMax = index;
            }
            if (index > cMax) {
                cMax = index;
            }
        }
    }

    inputMap[0] = 0;
    inputMap[1] = 1;
    inputSortMap[0] = 0;
    inputSortMap[1] = 1;
    if (verbosity >= 2) {
        cerr << "iMax = " << iMax << '\n';
        cerr << "tMax = " << tMax << '\n';
        cerr << "cMax = " << cMax << '\n';
    }

    u32 nStates = fst->n_states();
    u32 nInputs = inputMax + 1;

    IntSet outputs = fst->output_alphabet();
    outputs.insert(0);

    typedef fst::map<string,int> StringIntMap;
    StringIntMap outputStringMap;
    for (IntSet::const_iterator i = outputs.begin(); i != outputs.end(); ++i) {
        string label = FSTstring_map[*i];
        outputStringMap[label] = 0;
        if (isClassLabel(label.c_str())) {
            outputStringMap["<" + label + ">"] = 0;
            outputStringMap["</" + label + ">"] = 0;
        }
    }
    int outputIndex = 0;
    for (StringIntMap::iterator i = outputStringMap.begin();
         i != outputStringMap.end();
         ++i) {
        (*i).second = outputIndex++;
    }
    IntIntMap outputIntMap;
    for (IntSet::const_iterator i = outputs.begin(); i != outputs.end(); ++i) {
        outputIntMap[*i] = outputStringMap[FSTstring_map[*i]];
    }

    typedef fst::vector<string> StringVector;
    StringVector inputStrings;
    inputStrings.resize(inputMax + 1);
    for (IntIntMap::const_iterator i = inputMap.begin(); i != inputMap.end(); ++i) {
        inputStrings[(*i).second] = FSTstring_map[(*i).first];
    }
    for (int i = 2; i <= tMax; i++) {
        char buf[16];
        sprintf(buf, "%c.%d", (i <= iMax) ? 'i' : 't', i - 2);
        inputStrings[i] = buf;
    }

    if (verbosity >= 4) {
        cerr << "inputs:\n";
        for (StringVector::iterator i = inputStrings.begin(); i != inputStrings.end(); ++i) {
            cerr << '\t' << *i << '\n';
        }
    }

    assert(nInputs == inputStrings.size());

    u32 inputBits = bitsNeeded(nInputs);
    u32 stateBits = bitsNeeded(nStates);
    u32 outputBits = bitsNeeded(outputStringMap.size() - 1);

    if (verbosity >= 2) {
        cerr << "bit widths:"
             << " next=" << stateBits
             << " input=" << inputBits
             << " output=" << outputBits
             << " weight=" << weightBits
             << '\n';
    }

    // Pack all arcs, keeping track of bit offsets for the first arc leaving
    // each states.
    //
    BitStreamWriter32 arcWriter(swap);
    fst::vector<u32> positionForState;
    typedef fst::vector<const FSTArc*> ArcVector;
    ArcVector arcVector;
    FSTPack_ArcCompare arcCompare(inputSortMap);
    for (u32 p = 0; p < nStates; ++p) {
        positionForState.push_back(arcWriter.position());

        // state finality
        arcWriter.write(fst->is_final(p), 1);
        if (weightBits > 0 && fst->is_final(p)) {
            if (fst->final_weight(p) != 0) {
                arcWriter.write(1, 1);
                int weight = sq_quantize(fst->final_weight(p), weightD, 1 << weightBits);
                assert((weight >> weightBits) == 0);
                arcWriter.write(weight, weightBits);
            } else {
                arcWriter.write(0, 1);
            }
        }

        // Sort arcs using arcCompare (so that #i constraints come first).
        //
        arcVector.clear();
        FSTArcIter a = fst->arcs_out(p); // outside loop in case arcs are dynamic
        for (; a.valid(); ++a) {
            arcVector.push_back(a);
        }
        stable_sort(arcVector.begin(), arcVector.end(), arcCompare);

        // Walk sorted arcs.
        //
        for (ArcVector::const_iterator i = arcVector.begin(); i != arcVector.end(); ++i) {
            const FSTArc* a = *i;

            if (inputMap.count(a->input) == 0) {
                cerr << "unhandled input label: '" << FSTstring_map[a->input] << "' -- ignoring\n";
                continue;
            }

            // flag not finished with arcs
            arcWriter.write(1, 1);

            // input
            assert(inputMap.count(a->input) != 0);
            int mappedInput = inputMap[a->input];
            assert((mappedInput >> inputBits) == 0);
            arcWriter.write(mappedInput, inputBits);

            assert(a->input > 1 || a->next > int(p));

            if (mappedInput < 2 || mappedInput > iMax) {

                // next state
                assert((a->next >> stateBits) == 0);
                arcWriter.write(a->next, stateBits);

                // output (word)
                int output = outputIntMap[a->output];
                assert(!a->output || output);
                assert(output < int(outputStringMap.size()));
                assert((output >> outputBits) == 0);
                if (output) {
                    arcWriter.write(1, 1);
                    arcWriter.write(output, outputBits);
                } else {
                    arcWriter.write(0, 1);
                }
                    
                // weight
                if (weightBits > 0) {
                    if (a->weight == 0) {
                        arcWriter.write(0, 1);
                    } else {
                        arcWriter.write(1, 1);
                        int weight = sq_quantize(a->weight, weightD, 1 << weightBits);
                        assert((weight >> weightBits) == 0);
                        arcWriter.write(weight, weightBits);
                    }
                }

            } else { // internal

                if (a->next == int(p)) {
                    arcWriter.write(0, 1); // no next state needed
                } else {
                    arcWriter.write(1, 1);
                    assert((a->next >> stateBits) == 0);
                    arcWriter.write(a->next, stateBits);
                    int output = outputIntMap[a->output];
                    assert(!a->output || output);
                    assert(output < int(outputStringMap.size()));
                    assert((output >> outputBits) == 0);
                    if (output) {
                        arcWriter.write(1, 1);
                        arcWriter.write(output, outputBits);
                    } else {
                        arcWriter.write(0, 1);
                    }
                }

            }
        }
        
        // flag finished with arcs
        arcWriter.write(0, 1);
    }
    arcWriter.flush();
    positionForState.push_back(arcWriter.position());

    if (verbosity >= 2) {
        u32 nArcs = fst->n_arcs();
        u32 nBits = arcWriter.position();
        cerr << "arcs: " << nBits << " bits, " << nArcs << " arcs, "
             << double(nBits)/double(nArcs) << " bits/arc\n";
    }

    // Determine shift to use with state table 1.
    //
    u32 tableBits = 16;
    u32 tableShift = 0;
    for (u32 shift = 16; shift > 0; --shift) {
        u32 cumArcBits = 0;
        bool ok = true;
        for (u32 p = 0; p < nStates; ++p) {
            if ((p & ((1 << shift) - 1)) == 0) {
                cumArcBits = 0;
            }
            u32 arcBits = positionForState[p + 1] - positionForState[p];
            u32 pos = cumArcBits + arcBits;
            if (pos >> tableBits) {
                ok = false;
                break;
            }
            cumArcBits += arcBits;
        }
        if (ok) {
            tableShift = shift;
            break;
        }
    }
    
    // Construct the state tables.
    //
    fst::vector<u32> stateTable1;
    fst::vector<u16> stateTable2;
    for (u32 p = 0, cumArcBits = 0; p < nStates; ++p) {
        if ((p & ((1 << tableShift) - 1)) == 0) {
            stateTable1.push_back(positionForState[p]);
            cumArcBits = 0;
        }
        assert((cumArcBits >> 16) == 0);
        stateTable2.push_back(cumArcBits);
        u32 arcBits = positionForState[p + 1] - positionForState[p];
        cumArcBits += arcBits;
    }
    if (verbosity >= 2) {
        cerr << "state table shift: " << tableShift << '\n';
    }

    // Construct the word table.
    //
    MD5 md5;
    fst::vector<u32> wordTable;
    u32 wordOffset = 0;
    for (StringIntMap::iterator i = outputStringMap.begin();
         i != outputStringMap.end();
         ++i) {
        const string& output = (*i).first;
        md5.hash((u8*) output.c_str(), output.length());
        wordTable.push_back(wordOffset);
        wordOffset += output.length() + 1;
    }
    u8 outputDigest[16];
    md5.finish(outputDigest);
    if (verbosity >= 2) {
        cerr << "word table size = " << wordTable.size() << '\n';
        cerr << "word digest = " << digestString(outputDigest) << '\n';
    }

    // Construct the input table.
    //
    md5.reset();
    fst::vector<u32> inputTable;
    u32 inputOffset = 0;
    for (StringVector::iterator i = inputStrings.begin(); i != inputStrings.end(); ++i) {
        const string& input = *i;
        md5.hash((u8*) input.c_str(), input.length());
        inputTable.push_back(inputOffset);
        inputOffset += input.length() + 1;
    }
    u8 inputDigest[16];
    md5.finish(inputDigest);
    if (verbosity >= 2) {
        cerr << "input table size = " << inputTable.size() << '\n';
        cerr << "input digest = " << digestString(inputDigest) << '\n';
    }

    ////////////////////////////////////////////////////////////////////////////
    // Output to file
    ////////////////////////////////////////////////////////////////////////////

    // version
    byteWriter.write_u32(PFST3, swap);

    // filesize (set later)
    u32 sizePos = byteWriter.tell();
    byteWriter.write_u32(0, swap);

    // state table 1 pointer (set later)
    byteWriter.write_u32(0, swap);

    // state table 2 pointer (set later)
    byteWriter.write_u32(0, swap);

    // arc data pointer (set later)
    byteWriter.write_u32(0, swap);

    // input table pointer (set later)
    byteWriter.write_u32(0, swap);
    
    // input string pointer (set later)
    byteWriter.write_u32(0, swap);

    // word table pointer (set later)
    byteWriter.write_u32(0, swap);

    // word string pointer (set later)
    byteWriter.write_u32(0, swap);

    // number of states
    byteWriter.write_u32(nStates, swap);
    
    // initial state
    byteWriter.write_u32(fst->initial(), swap);

    // state table shift
    byteWriter.write_u8(tableShift);

    // input bit width
    byteWriter.write_u8(inputBits);

    // next state bit width
    byteWriter.write_u8(stateBits);

    // output bit width
    byteWriter.write_u8(outputBits);

    // weight bit width
    byteWriter.write_u8(weightBits);

    // max internal label
    byteWriter.write_u16(iMax, swap);

    // number of inputs
    byteWriter.write_u32(inputTable.size(), swap);
    
    // number of words
    byteWriter.write_u32(wordTable.size(), swap);

    // input digest
    for (int i = 0; i < 16; ++i) {
        byteWriter.write_u8(inputDigest[i]);
    }

    // output digest
    for (int i = 0; i < 16; ++i) {
        byteWriter.write_u8(outputDigest[i]);
    }

    // weight reconstruction values
    if (weightBits > 0) {
        byteWriter.pad(2);
        for (u32 i = 0; i < u32(1 << weightBits); ++i) {
            byteWriter.write_s16(weightR ? (s16)(F2Q(weightR[i], 6)) : 0, swap);
        }
    }

    // state table 1
    byteWriter.pad(32);
    u32 stateTable1Pos = byteWriter.tell();
    for (fst::vector<u32>::iterator i = stateTable1.begin(); i != stateTable1.end(); ++i) {
        byteWriter.write_u32(*i, swap);
    }

    // state table 2
    byteWriter.pad(32);
    u32 stateTable2Pos = byteWriter.tell();
    for (fst::vector<u16>::iterator i = stateTable2.begin(); i != stateTable2.end(); ++i) {
        byteWriter.write_u16(*i, swap);
    }

    // arc data
    byteWriter.pad(32);
    u32 arcPos = byteWriter.tell();
    byteWriter.write(arcWriter.data(), arcWriter.length());

    // input table
    byteWriter.pad(32);
    u32 inputTablePos = byteWriter.tell();
    for (fst::vector<u32>::iterator i = inputTable.begin(); i != inputTable.end(); ++i) {
        byteWriter.write_u32(*i, swap);
    }
    
    // input strings
    byteWriter.pad(32);
    u32 inputStringPos = byteWriter.tell();
    for (StringVector::iterator i = inputStrings.begin(); i != inputStrings.end(); ++i) {
        byteWriter.write((*i).c_str(), (*i).length() + 1);
    }

    // word table
    byteWriter.pad(32);
    u32 wordTablePos = byteWriter.tell();
    for (fst::vector<u32>::iterator i = wordTable.begin(); i != wordTable.end(); ++i) {
        byteWriter.write_u32(*i, swap);
    }

    // word strings
    byteWriter.pad(32);
    u32 wordStringPos = byteWriter.tell();
    for (StringIntMap::iterator i = outputStringMap.begin();
         i != outputStringMap.end();
         ++i) {
        byteWriter.write((*i).first.c_str(), (*i).first.length() + 1);
    }

    // fix up postponed pointers
    u32 size = byteWriter.tell();
    byteWriter.seek(sizePos);
    byteWriter.write_u32(size, swap);
    byteWriter.write_u32(stateTable1Pos, swap);
    byteWriter.write_u32(stateTable2Pos, swap);
    byteWriter.write_u32(arcPos, swap);
    byteWriter.write_u32(inputTablePos, swap);
    byteWriter.write_u32(inputStringPos, swap);
    byteWriter.write_u32(wordTablePos, swap);
    byteWriter.write_u32(wordStringPos, swap);
    if (verbosity >= 3) {
        cerr << "stateTable1Pos = " << stateTable1Pos << '\n';
        cerr << "stateTable2Pos = " << stateTable2Pos << '\n';
    }

    // done

    // optionally dump vocabulary (alphabet) file
    if (out_label_file && out_label_file[0]) {
        FILE* fp = fopen(out_label_file, "w");
        for (StringIntMap::iterator i = outputStringMap.begin();
             i != outputStringMap.end();
             ++i) {
            fputs((*i).first.c_str(), fp);
            fputc('\n', fp);
        }
        fclose(fp);
        if (verbosity >= 1) {
            cerr << "Wrote " << outputStringMap.size()
                 << " words to vocabulary file " << out_label_file << ".\n";
        }
    }

    if (verbosity >= 1) {
        cerr << "Wrote " << size << " bytes.\n";
    }

    return fst;
}

// Main entry point.
//
static FSTGeneric
FSTPack(ByteWriter& byteWriter, const char* out_label_file,
        const FSTGeneric& in, int weightBits,
        const IntSet& externalInputs, bool swap, int verbosity, int version)
{
    switch (version) {
    default:
        FSTWarn("Unsupported PackedFST version %d -- using 3");
    case 0:
    case 3:
        return FSTPack3(byteWriter, out_label_file, in, weightBits, externalInputs, swap, verbosity);
    case 2:
        return FSTPack2(byteWriter, out_label_file, in, weightBits, swap, verbosity);
    }

}

FSTGeneric
FSTPackToFile(const char* pfst_file, const char* out_label_file,
              const FSTGeneric& in, int weight_bits, const IntSet& inputs,
              bool swap, int verbosity, int version)
{
    MemoryByteWriter byteWriter(0);
    FSTGeneric out = FSTPack(byteWriter, out_label_file, in, weight_bits, inputs, swap, verbosity, version);
    FILE* fp = fopen(pfst_file, "wb");
    fwrite(byteWriter.data(), byteWriter.size(), 1, fp);
    fclose(fp);
    if (verbosity >= 1) {
        cerr << " to file " << pfst_file << ".\n";
    }
    return out;
}

FSTMemory*
FSTPackToMemory(const FSTGeneric& in, int weight_bits, const IntSet& inputs, int verbosity, int version)
{
    MemoryByteWriter byteWriter(0);
    FSTPack(byteWriter, NULL, in, weight_bits, inputs, false, verbosity, version);
    return new FSTMemory(byteWriter.data(), byteWriter.size(), 32);
}

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 80
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
