////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2008, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

#include <cstring>
#include <iostream>
#include <fstream>
#include "fst/FST.h"
#include "fst/FSTPronRules.h"

// plurals (not yet implemented)
//
//   +s      remove s, check, append /'s/
//   +es     remove es, check, append /'s/
//   -y+ies  remove ies, add y, check, append /'s/

// possessive (implemented)
//
//   -'s     remove s, check, append /'s/
//   -'      remove ', check, append /'s/

using namespace std;

typedef fst::set<int> Set;
typedef fst::map<int,int> Map;
typedef pair<int,int> Pair;
typedef fst::map<Pair,int> PairMap;
typedef fst::vector<Pair> PairVector;

FSTDictionary::FSTDictionary(const char* lFile, const char* letterToSoundFiles,
                             int N, float S)
{
    _iwt = FSTstring_map["_"];
    _pound = FSTstring_map["#"];
    _poundStar = FSTstring_map["#star"];
    _apostrophe_s = FSTstring_map["'s"];

    if (lFile && lFile[0] && strcmp(lFile, "None") != 0) {
        _L = FSTLoad(lFile);
    }

    if (letterToSoundFiles && letterToSoundFiles[0] && strcmp(letterToSoundFiles, "None") != 0) {
        char* letterToSoundFilesCopy = strdup(letterToSoundFiles);
        const char* sep = ",";
        for (char *save, *file = strtok_r(letterToSoundFilesCopy, sep, &save);
             file != NULL;
             file = strtok_r(NULL, sep, &save)) {
            _letterToSoundList.push_back(new FSTLetterToSound(file, N, S));
        }
        free(letterToSoundFilesCopy);
    }

    if (_L) {
        if (!_L->is_mutable()) {
            _L = FSTBasic(_L);
        }
        _poundState = _L->initial();
        _iwtState = _L->add_state();
        _L->set_final(_iwtState);
        _L->add_arc(_iwtState, _iwtState, 0, _iwt);
        _L->add_arc(_iwtState, _poundState, 0, _pound);
        _L->clear_final(_poundState);
    }
}

FSTDictionary::~FSTDictionary()
{
    for (fst::list<FSTLetterToSound*>::iterator i = _letterToSoundList.begin();
         i != _letterToSoundList.end(); ++i) {
        delete *i;
    }
}

FSTGeneric
FSTDictionary::compose(const FSTGeneric& G)
{
    _addWords(G->input_alphabet());
    return FSTProject(FSTTrim(FSTCompose(FSTProject(G, FSTOutput, FSTInput), _L)), FSTOutput, FSTInput);
}

void
FSTDictionary::_addWords(const fst::set<int>& words)
{
    for (fst::set<int>::const_iterator i = words.begin(); i != words.end(); ++i) {
        int w = *i;
        if (_L->input_alphabet().count(w) == 0) {
            _addWord(w);
        }
    }
}

void
FSTDictionary::_addWord(int w)
{
    const char* word = FSTstring_map[w];
    if (word[0] != '$' && strchr(word, '_') != NULL) {
        _addUnderbarredWord(w, word);
    } else {
        _addSimpleWord(w, word);
    }
}

static bool
extractSuffix(string& word, const string& suffix)
{
    string::size_type i = word.rfind(suffix);
    if (i != string::npos && i == word.length() - suffix.length()) {
        word = word.substr(0, i);
        return true;
    } else {
        return false;
    }
}

static FSTGeneric
singleArcFst(int i, int o)
{
    FSTGeneric fst = FSTBasic();
    fst->add_states(2);
    fst->set_initial(0);
    fst->set_final(1);
    fst->add_arc(0, 1, i, o);
    return fst;
}

FSTGeneric
FSTDictionary::_appendPossessive(FSTGeneric fst)
{
    FSTGeneric s = singleArcFst(0, _apostrophe_s);
    fst = FSTBasic(FSTConcatenate(fst, s));
    fst = FSTBasic(FSTRemoveEpsilonsA(fst));
    return fst;
}

FSTGeneric
FSTDictionary::_stripLabel(const FSTGeneric& fst, int o)
{
    FSTGeneric filter = FSTBasic();
    filter->add_state();
    filter->set_initial(0);
    filter->set_final(0);
    const fst::set<int>& phonemes = fst->output_alphabet();
    for (fst::set<int>::const_iterator i = phonemes.begin(); i != phonemes.end(); ++i) {
        if (*i != o) {
            filter->add_arc(0, 0, *i, *i);
        }
    }

    return FSTTrim(FSTCompose(fst, filter));
}

void
FSTDictionary::_addSimpleWord(int w, const char* wordIn)
{
    if (wordIn[0] == '$') {
        //
        // Dynamic class: pronunciation is "#star".

        _merge(singleArcFst(w, _poundStar), false);
        return;
    }

    // TO DO: support stripping of regular plural suffixes.

    string word = wordIn;
    bool possessive = false;
    if (extractSuffix(word, "'s") || extractSuffix(word, "'")) {
        possessive = true;
        cerr << "possessive: " << wordIn << " -> " << word << '\n';
    }

    int wordIndex = FSTstring_map[word];
    if (possessive && _L->input_alphabet().count(wordIndex) != 0) {
        //
        // After stripping possessive suffix, we found the word in the dictionary.

        FSTGeneric fst = singleArcFst(w, wordIndex);
        fst = FSTTrim(FSTCompose(fst, _L));
        fst = _stripLabel(fst, _iwt);
        if (possessive) {
            fst = _appendPossessive(fst);
        }
        _merge(fst, true);

    } else {

        // Use first letterToSound that produces output.
        //
        FSTGeneric fst;
        for (fst::list<FSTLetterToSound*>::iterator i = _letterToSoundList.begin();
             i != _letterToSoundList.end(); ++i) {
            fst = (*i)->generateFst(word);
            fst = FSTClearWeights(fst);
            if (fst) {
                if (possessive) {
                    fst = _appendPossessive(fst);
                }
                _merge(fst, true);
                return;
            }
        }

        FSTWarn("no pronunciation for '%s' -- ignoring", wordIn);
    }

}

void
FSTDictionary::_addUnderbarredWord(int w, const char* word)
{
    // Split word at _.
    //
    char* wordCopy = strdup(word);
    FSTGeneric fst = FSTBasic();
    int p = fst->add_state();
    fst->set_initial(p);
    for (char *save, *subword = strtok_r(wordCopy, "_", &save);
         subword != NULL;
         subword = strtok_r(NULL, "_", &save)) {
        int q = fst->add_state();
        int s = FSTstring_map[subword];
        if (_L->input_alphabet().count(s) == 0) {
            _addSimpleWord(s, subword);
        }
        fst->add_arc(p, q, (p == fst->initial()) ? w : 0, s);
        p = q;
    }
    fst->set_final(p);
    free(wordCopy);
    //
    // Currently, fst of the form abc_def_ghi:abc, eps:def, eps:ghi.

    fst = FSTTrim(FSTCompose(fst, _L));
    //
    // Currently, fst of the form abc_def_ghi:/a/, eps:/b/, eps:/c/, ... eps:_*

    fst = _stripLabel(fst, _pound);

    // Merge into _L.
    //
    _merge(fst, true);
}

void
FSTDictionary::_merge(const FSTGeneric& fst, bool addIwt)
{
    if (fst->initial() >= 0) {
        fst::vector<int> stateMap(fst->n_states(), -1);
        stateMap[fst->initial()] = _L->initial();
        
        for (int p = 0; p < fst->n_states(); ++p) {
            int& pMapped = stateMap[p];
            if (pMapped < 0) {
                pMapped = _L->add_state();
            }
            _L->set_final(pMapped, fst->final_weight(p));
            if (_L->is_final(pMapped)) {
                _L->add_arc(pMapped, _poundState, 0, _pound, _L->final_weight(pMapped));
                if (addIwt) {
                    _L->add_arc(pMapped, _iwtState, 0, _iwt, _L->final_weight(pMapped));
                }
            }
            for (FSTArcIter a = fst->arcs_out(p); a.valid(); ++a) {
                int q = a->next;
                int& qMapped = stateMap[q];
                if (qMapped < 0) {
                    qMapped = _L->add_state();
                }
                _L->add_arc(pMapped, qMapped, a->input, a->output, a->weight);
            }
        }
#if DICT_DUMP
        fst->dump_ascii("debug.merge.in.fst");
        _L->dump_ascii("debug.merge.out.fst");
#endif
    }
}

FSTPocketCompiler::FSTPocketCompiler(const char* mcpFile, const char* lFile,
                                     const char* letterToSoundFiles, int weightBits)
    : _mcp(FSTLoad(mcpFile))
    , _dict(lFile, letterToSoundFiles)
    , _weightBits(weightBits)
{
    _verbosity = 2;
}

FSTPocketCompiler::~FSTPocketCompiler()
{
}

FSTMemory*
FSTPocketCompiler::jsgfCompile(const char* jsgf)
{
    return _jsgfCompile(NULL, jsgf);
}

FSTMemory*
FSTPocketCompiler::jsgfCompileClass(const char* className, const char* jsgf)
{
    return _jsgfCompile(className, jsgf);
}

FSTMemory*
FSTPocketCompiler::_jsgfCompile(const char* className, const char* jsgf)
{
    bool gIgnoreTags  = false;
    bool gIgnoreWords = false;
    bool gRtnParse    = false;
    bool gPause12     = false;

    // Attach JSGF header if not supplied.
    //
    string buffer;
    if (jsgf[0] != '#') {
        if (_verbosity >= 1) cerr << "attaching JSGF header\n";
        buffer = "#JSGF V1.0;\ngrammar jsgf;\n";
        buffer += jsgf;
        jsgf = buffer.c_str();
    }

    // Compute G from JSGF.
    //
    FSTGeneric g;
    string errorMsg;
    if (!FSTJsgfParseString(g, errorMsg, jsgf, gIgnoreTags, gIgnoreWords, gRtnParse, gPause12, false)) {
        if (_verbosity >= 1) {
            cerr << "JSGF parse error: " << errorMsg;
        }
        return NULL;
    }
    buffer.clear();
    if (_verbosity >= 1) cerr << "optimizing..." << flush;
    g = FSTOptimizeA(g);
    if (_verbosity >= 1) cerr << "done\n";

    return _compile(className, g);
}

FSTMemory*
FSTPocketCompiler::fstCompile(const FSTGeneric& g)
{
    return _compile(NULL, g);
}

static void
_addPhrase(FSTGeneric& g, char* phrase)
{
    int p = 0;
    const char* sep = " ";
    for (char *save, *word = strtok_r(phrase, sep, &save);
         word != NULL;
         word = strtok_r(NULL, sep, &save)) {
        if (word[0]) {
            int w = FSTstring_map[word];
            int q = g->add_state();
            g->add_arc(p, q, w, w);
            p = q;
        }
    }
    g->set_final(p);
}

FSTMemory*
FSTPocketCompiler::listCompileClass(const char* className, const char* phrases)
{
    FSTGeneric g = FSTBasic();
    g->add_state();
    g->set_initial(0);

    char* copy = strdup(phrases);
    const char* sep = ",;\r\n";
    for (char *save, *phrase = strtok_r(copy, sep, &save);
         phrase != NULL;
         phrase = strtok_r(NULL, sep, &save)) {
        _addPhrase(g, phrase);
    }
    free(copy);

    g = FSTBasic(FSTRemoveEpsilonsA(g));
    g = FSTBasic(FSTDeterminizeA(g));
    g = FSTMinimize(g);

    return _compile(className, g);
}

FSTMemory*
FSTPocketCompiler::fstCompileClass(const char* className, const FSTGeneric& lg)
{
    return _compile(className, lg, true);
}

FSTMemory*
FSTPocketCompiler::_compile(const char* className, const FSTGeneric& g, bool isLG)
{
    FSTGeneric lg;

    if (!isLG) {

        // Compute LoG.  The FSTDictionary will automatically generate pronunciations as needed.
        //
        if (_verbosity >= 1) cerr << "computing LoG..." << flush;
        lg = _dict.compose(g);
        if (_verbosity >= 1) cerr << "done\n";

    } else {

        lg = g;

    }

    // Apply appropriate prefix and suffix to LoG.
    //
    fst::vector<FSTGeneric> sequence;

    if (className == NULL) {
        //
        // main: Prefix with -* # and suffix with # -*.

        int dash = FSTstring_map["-"];
        FSTGeneric pause = singleArcFst(dash, 0);
        pause->add_arc(1, 1, dash, 0);
        FSTGeneric pound = singleArcFst(1, 0);
        sequence.push_back(pause);
        sequence.push_back(pound);
        sequence.push_back(lg);
        sequence.push_back(pound);
        sequence.push_back(pause);

    } else {
        //
        // class: Prefix with $:<class> and suffix with $:</class>.

        string prefixString = "<";
        prefixString += className;
        prefixString += ">";

        string suffixString = "</";
        suffixString += className;
        suffixString += ">";

        int prefix = FSTstring_map[prefixString];
        int suffix = FSTstring_map[suffixString];
        int dollar = FSTstring_map["$"];

        sequence.push_back(singleArcFst(dollar, prefix));
        sequence.push_back(lg);
        sequence.push_back(singleArcFst(dollar, suffix));

    }
    lg = FSTConcatenate(sequence);
    sequence.clear();

    // Optimize.
    //
    if (!isLG) {
        if (_verbosity >= 1) cerr << "optimizing..." << flush;
        lg = FSTOptimize(lg);
        if (_verbosity >= 1) cerr << "done\n";
        if (_verbosity >= 2) cerr << "size: " << lg->n_states() << ' ' << lg->n_arcs() << '\n';
    }

    // Determine if this grammar supports dynamic vocabulary.
    //
    bool dynamic = className != NULL
        || lg->input_alphabet().count(FSTstring_map["#star"]) != 0;
    FSTGeneric mcp;
    if (dynamic) {
        if (!_mcpMain) {
            _mcpMain = FSTDynamicAugmentMain(_mcp);
        }
        if (!_mcpClass) {
            _mcpClass = FSTDynamicAugmentClass(_mcp);
        }
        if (className == NULL) {
            if (_verbosity >= 1) cerr << "using dynamic main MCP\n";
            mcp = _mcpMain;
        } else {
            if (_verbosity >= 1) cerr << "using dynamic class MCP\n";
            mcp = _mcpClass;
        }
    } else {
        if (_verbosity >= 1) cerr << "using non-dynamic MCP\n";
        mcp = _mcp;
    }

    // Compute MoCoPo(LoG).
    //
    if (_verbosity >= 1) cerr << "computing MoCoPoLoG...";
    FSTGeneric mcplg = FSTTrim(FSTCompose(mcp, lg));
    lg.destroy();
    if (_verbosity >= 1) cerr << "done\n";
    if (!isLG) {
        if (_verbosity >= 1) cerr << "optimizing..." << flush;
        mcplg = FSTOptimizeA(mcplg); // need A version???
        if (_verbosity >= 1) cerr << "done\n";
    }

    if (className != NULL) {
        mcplg = _delaySuffix(mcplg);
    }
    if (_verbosity >= 2) cerr << "size: " << mcplg->n_states() << ' ' << mcplg->n_arcs() << '\n';

    // Pack to PocketSUMMIT format.
    //
    if (_verbosity >= 1) cerr << "packing to PocketSUMMIT format...\n";
    FSTMemory* result = FSTPackToMemory(mcplg, _weightBits,
                                        dynamic ? _mcpMain->input_alphabet() : fst::set<int>(),
                                        _verbosity);

    return result;
}

static int
_find_or_create_state(int state, int constraint, FSTGeneric fst, PairMap& pair_map, PairVector& pair_vector)
{
    int q = pair_map.size();
    pair<PairMap::iterator, bool> status =
        pair_map.insert(PairMap::value_type(Pair(state, constraint), q));
    if (status.second == true) {
        fst->add_state();
        pair_vector.push_back(Pair(state, constraint));
    } else {
        q = (*status.first).second;
    }
    return q;
}

FSTGeneric
FSTPocketCompiler::_delaySuffix(const FSTGeneric& x)
{
    // Due to determinization, we could end up with a #i constraint arc followed
    // by additional arcs with pending outputs.  For example, with "lee" and
    // "leeds" in the class, we end up with "#i:lee" followed by "eps:</CLASS>".
    // Having the #i arc as non-final confuses FSTDynamic splicing.
    //
    // Delay such #i labels.
    //

    Set constraints;
    const Set& inputs = x->input_alphabet();
    for (Set::const_iterator i = inputs.begin(); i != inputs.end(); ++i) {
        const char* label = FSTstring_map[*i];
        if (label[0] == '#' && isdigit(label[1])) {
            constraints.insert(*i);
        }
    }

    PairMap pair_map;
    PairVector pair_vector;
    int p;
    
    pair_vector.reserve(x->n_states());
    
    FSTGeneric y = FSTBasic();
    pair_map[Pair(x->initial(), 0)] = p = y->add_state();
    y->set_initial(p);
    pair_vector.push_back(Pair(x->initial(), 0));
    
    for (unsigned i = 0; i < pair_vector.size(); i++) {
        int p = pair_vector[i].first;  // state
        int c = pair_vector[i].second; // constraint
        
        if (c == 0 && x->is_final(p)) {
            y->set_final(i, x->final_weight(p));
        }
        
        for (FSTArcIter a = x->arcs_out(p); a.valid(); a++) {
            
            if (c == 0) {
                
                if (p != x->initial() && !x->is_final(a->next) && constraints.count(a->input)) {
                    int q = _find_or_create_state(a->next, a->input, y, pair_map, pair_vector);
                    y->add_arc(i, q, 0, a->output, a->weight);
                } else {
                    int q = _find_or_create_state(a->next, 0, y, pair_map, pair_vector);
                    y->add_arc(i, q, a->input, a->output, a->weight);
                }
                
            } else {
                
                assert(a->input == 0);
                
                if (x->is_final(a->next)) {
                    int q = _find_or_create_state(a->next, 0, y, pair_map, pair_vector);
                    y->add_arc(i, q, c, a->output, a->weight);
                } else {
                    int q = _find_or_create_state(a->next, c, y, pair_map, pair_vector);
                    y->add_arc(i, q, 0, a->output, a->weight);
                }
                
            }
            
        }
        
    }

    return y;
}

static int
constraint_label(int p)
{
    char buf[16];
    sprintf(buf, "#%d", p);
    return FSTstring_map[buf];
}

static bool
is_constraint_label(int i)
{
    const char* label = FSTstring_map[i];
    return label[0] == '#' && label[1] >= '0' && label[1] <= '9';
}

static void
compute_constraints(Map& constraints, const FSTGeneric& in)
{
    // Determine set of constraints by examining destinations of # arcs.
    //
    const int pound = FSTstring_map["#"];
    for (int p = 0; p < in->n_states(); ++p) {
        for (FSTArcIter a = in->arcs_out_with_output(p, pound); a.valid(); ++a) {
            constraints.insert(Map::value_type(a->next, constraints.size() + 1));
        }
    }
}

FSTGeneric
FSTDynamicAugmentMain(const FSTGeneric& in)
{
    const int pound = FSTstring_map["#"];
    const int star = FSTstring_map["#star"];

    in->force();
    const int N = in->n_states();

    FSTGeneric out = FSTBasic(in);

    Map constraints;
    compute_constraints(constraints, in);

    // New final state.
    //
    int new_final = out->add_state();
    out->set_final(new_final);

    // New #star end state.
    //
    int star_end = out->add_state();

    // New constraint-specific #star start states.
    //
    int star_start = out->add_states(constraints.size());
    for (Map::const_iterator i = constraints.begin(); i != constraints.end(); ++i) {
        int c = (*i).second;
        out->add_arc(star_start + c - 1, star_end, constraint_label(c), star);
    }

    // New initial state and its transitions.
    //
    int old_initial = out->initial();
    int new_initial = out->add_state();
    out->set_initial(new_initial);
    out->set_final(new_initial);
    for (FSTArcIter a = out->arcs_out(old_initial); a.valid(); ++a) {
        out->add_arc(new_initial, a->next, a->input, a->output, a->weight);
    }
    out->add_arc(new_initial, star_start + constraints[old_initial] - 1, 0, 0);

    // Transitions after #star.
    //
    for (Map::const_iterator i = constraints.begin(); i != constraints.end(); ++i) {
        int q = (*i).first;
        int c = (*i).second;
        int c_label = constraint_label(c);
        out->add_arc(star_end, q, c_label, pound);
        out->add_arc(star_end, star_start + c - 1, c_label, pound);
        if (in->is_final(q)) {
            out->add_arc(star_end, new_final, c_label, 0, in->final_weight(q));
        }
    }

    // Process all original states.
    //
    for (int p = 0; p < N; ++p) {
        for (FSTArcIter a = in->arcs_out_with_output(p, pound); a.valid(); ++a) {
            int q = a->next;
            int c = constraints[q];
            out->add_arc(p, star_start + c - 1, pound, pound, a->weight);
        }
    }

    return out;
}

FSTGeneric
FSTDynamicAugmentClass(const FSTGeneric& in)
{
    const int pound = FSTstring_map["#"];
    const int dollar = FSTstring_map["$"];

    in->force();
    const int N = in->n_states();

    FSTGeneric out = FSTBasic(in);

    Map constraints;
    compute_constraints(constraints, in);

    // New final state.
    //
    int new_final = out->add_state();
    out->set_final(new_final);

    // New initial state.
    //
    int new_initial = out->add_state();
    out->set_initial(new_initial);

    // New transitions.
    //
    for (Map::const_iterator i = constraints.begin(); i != constraints.end(); ++i) {
        int q = (*i).first;
        int q_prime = (*i).second;
        int pound_q = constraint_label(q_prime);
        out->add_arc(new_initial, q, pound_q, dollar);
    }
    for (int p = 0; p < N; ++p) {
        for (FSTArcIter a = in->arcs_out_with_output(p, pound); a.valid(); ++a) {
            if (in->is_final(p)) {
                int q = a->next;
                int q_prime = constraints[q];
                int pound_q = constraint_label(q_prime);
                out->add_arc(p, new_final, pound_q, dollar, in->final_weight(p));
                out->clear_final(p);
            }
        }
    }

    // Non-null constraint.  Must start with at least one non-null (epsilon/#) arc.
    //
    FSTGeneric constraint_fst = FSTBasic();
    constraint_fst->add_states(4);
    constraint_fst->set_initial(0);
    const fst::set<int>& alphabet = out->input_alphabet();
    for (fst::set<int>::const_iterator i = alphabet.begin(); i != alphabet.end(); ++i) {
        if (is_constraint_label(*i)) {
            constraint_fst->add_arc(0, 1, *i, *i);
            constraint_fst->add_arc(2, 3, *i, *i);
        } else {
            if (*i > 1) {
                constraint_fst->add_arc(1, 2, *i, *i);
            }
            if (*i > 0) {
                constraint_fst->add_arc(2, 2, *i, *i);
            }
        }
    }
    constraint_fst->set_final(3);

    out = FSTTrim(FSTCompose(constraint_fst, out));

    return out;
}

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 80
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
