////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

#include "fst/FST_i.h"
#include <cstdio>
#include <queue>
#include "Heap.h"

static void
_FSTPotentialAcyclic(fst::vector<float>& potential, FSTGeneric m,
                     const fst::vector<int>& order)
{
    FSTSemiring semiring = m->semiring();

    // We rely on the fact that m has been forced by the sort in sizing the
    // potential[p] vector.
    //
    potential.clear();
    potential.insert(potential.begin(), m->n_states(), semiring->zero);

    // Walk through states in reverse order.  When we reach a state p, the
    // potentials for all states reachable from p will be finished.
    //
    for (fst::vector<int>::const_reverse_iterator i = order.rbegin();
         i != order.rend(); i++) {
        int p = *i;
        potential[p] = semiring->better(potential[p], m->final_weight(p));
        for (FSTArcIter a = m->arcs_out(p); a.valid(); a++) {
            int q = a->next;
            if (q == p) {
                if (semiring->is_better(a->weight, semiring->one))
                    FSTFatal("FSTPotential: self loop weight better than %g",
                             semiring->one);
            } else {
                float w = semiring->times(potential[q], a->weight);
                potential[p] = semiring->better(potential[p], w);
            }
        }
    }
}

static void
_FSTPotentialCyclicBellmanFord(fst::vector<float>& potential, FSTGeneric m)
{
    int N = m->n_states();
    FSTSemiring semiring = m->semiring();

    // Initialize.
    //
    potential.clear();
    potential.reserve(N);
    for (int p = 0; p < N; p++) {
        potential.push_back(m->final_weight(p));
    }

    // Relaxation (N-1 passes) and negative cost loop check (1 pass), if needed.
    //
    for (int i = N; i > 0; i--) {
        bool relaxed = false;
        for (int p = 0; p < N; p++) {
            float& pw = potential[p];
            for (FSTArcIter a = m->arcs_out(p); a.valid(); a++) {
                int q = a->next;
                float qw = potential[q];
                if (semiring->is_better(qw, semiring->zero)) {
                    float w = semiring->times(qw, a->weight);
                    if (i > 1) { // relaxation pass
                        if (semiring->is_better(w, pw)) {
                            pw = w;
                            relaxed = true;
                        }
                    } else { // negative cost loop check, final pass
                        if (semiring->is_better(w, pw)) {
                            FSTWarn("FSTPotential: "
                                    "cycle with weight better than semiring->one detected");
                            return;
                        }
                    }
                }
            }
        }
        if (!relaxed) {
            break;
        }
    }
}

namespace _FSTPotential {

    struct Arc {
        int state;
        float weight;
        Arc(int s, float w) : state(s), weight(w) {}
    };

    typedef fst::list<Arc>  ArcList;

    struct DijkstraEntry {
        float weight;
        int index;
        ArcList arcs;
        DijkstraEntry(float w) : weight(w), index(0) {
        }
        void set_heap_index(int i) {
            index = i;
        }
        int get_heap_index() const {
            return index;
        }
        struct Better {
            FSTSemiring semiring;
            Better() : semiring(0) {
            }
            Better(FSTSemiring s) : semiring(s) {
            }
            bool operator()(const DijkstraEntry* x, const DijkstraEntry* y) const {
                return semiring->is_better(x->weight, y->weight);
            }
        };
    };

}

using namespace _FSTPotential;

static void
_FSTPotentialCyclicDijkstra(fst::vector<float>& potential, FSTGeneric m)
{
    typedef Heap<DijkstraEntry, DijkstraEntry::Better> DijkstraEntryHeap;
    typedef fst::vector<DijkstraEntry> DijkstraEntryVector;

    int N = m->n_states();
    FSTSemiring semiring = m->semiring();

    DijkstraEntry initial_entry(semiring->zero);
    DijkstraEntryVector entry_vector(N, initial_entry);

    DijkstraEntry::Better better(semiring);
    DijkstraEntryHeap entry_heap(better);

    // Build up lists of reverse arcs for each state, and insert all
    // final states into priority queue.
    //
    for (int p = 0; p < N; p++) {
        if (m->is_final(p)) {
            entry_vector[p].weight = m->final_weight(p);
            entry_heap.insert(&entry_vector[p]);
        }
        for (FSTArcIter a = m->arcs_out(p); a.valid(); a++) {
            entry_vector[a->next].arcs.push_front(Arc(p, a->weight));
        }
    }

    // Run best-first search.
    //
    while (entry_heap.size() != 0) {
        DijkstraEntry* top = entry_heap.top();
        entry_heap.pop();
        for (ArcList::const_iterator i = top->arcs.begin(); i != top->arcs.end(); i++) {
            int q = (*i).state;
            float qw = semiring->times(top->weight, (*i).weight);
            DijkstraEntry* next = &entry_vector[q];
            if (semiring->is_better(qw, next->weight)) {
                next->weight = qw;
                if (next->index > 0) {
                    entry_heap.decrease(next);
                } else {
                    entry_heap.insert(next);
                }
            }
        }
    }

    // Output resulting potentials.
    //
    potential.clear();
    for (int p = 0; p < N; p++) {
        potential.push_back(entry_vector[p].weight);
    }
}

// Use the Bellman-Ford algorithm if we have any "negative" costs.  Otherwise,
// use Dijkstra's algorithm.
//
static void
_FSTPotentialCyclic(fst::vector<float>& potential, FSTGeneric m)
{
    int N = m->n_states();
    FSTSemiring semiring = m->semiring();
    for (int p = 0; p < N; p++) {
        for (FSTArcIter a = m->arcs_out(p); a.valid(); a++) {
            if (semiring->is_better(a->weight, semiring->one)) {
                _FSTPotentialCyclicBellmanFord(potential, m);
                return;
            }
        }
    }
    _FSTPotentialCyclicDijkstra(potential, m);
}

// Use either acyclic or cyclic algorithm.
//
void
FSTPotential(fst::vector<float>& potential, const FSTGeneric& x)
{
    fst::vector<int> order;
    if (FSTTopologicalSort(order, x))
        _FSTPotentialAcyclic(potential, x, order);
    else
        _FSTPotentialCyclic(potential, x);
}

FSTGeneric
FSTPotential(FSTGeneric x)
{
    FSTGeneric m = x->is_mutable() ? x : FSTMutableCopy(x);

    fst::vector<float> potential;
    FSTPotential(potential, x);
    for (int p = 0; p < int(potential.size()); p++)
        m->set_potential(p, potential[p]);

    return m;
}

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 80
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
