////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

// Phonological Rule Transducer Construction:
//
// P  = C o E o L o R
//
// Pl = C o E o L
// Pr = R
//
// C: enforce link consistency (e.g., foo$ must be followed by $foo)
// E: rule expansion
// L: left context
// R: right context (applied right-to-left)

#include "fst/FST_i.h"
#include "fst/FSTPronRules.h"
#include "fst/FSTTimer.h"
#include "fst/format.h"
#include <string>
#include <sstream>

using namespace std;

typedef pair<int, int> IntPair;

struct FSTPronRules::_Rule {
    int name;                                   // e.g., ax.2 (ax rule 2)
    int index;
    int middle;
    fst::vector<int> left;                           // sorted left labels
    fst::vector<int> right;                          // sorted right labels
    FSTGeneric fst;
};

FSTPronRules::FSTPronRules(const string& filename, int verbosity)
    : _verbosity(verbosity)
{
    if (_parse(filename)) {
        _compile();
    } else {
        _rule_map.clear();
    }
}

FSTPronRules::~FSTPronRules()
{
    _RuleMap::iterator i;
    for (i = _rule_map.begin(); i != _rule_map.end(); ++i) {
        _RuleVector& rv = (*i).second;
        _RuleVector::iterator j;
        for (j = rv.begin(); j != rv.end(); ++j) {
            delete *j;
        }
    }
}

bool
FSTPronRules::bad() const
{
    return _rule_map.empty();
}

static inline void
_assign(fst::vector<int>& dst, const fst::set<int>& src)
{
    dst.clear();
    dst.reserve(src.size());
    for (fst::set<int>::const_iterator i = src.begin(); i != src.end(); ++i)
        dst.push_back(*i);
}

void
FSTPronRules::set_alphabet(const fst::set<int>& labels)
{
    _assign(_alphabet, labels);
}

void
FSTPronRules::set_initial(const fst::set<int>& labels)
{
    _assign(_initial, labels);
}

void
FSTPronRules::set_final(const fst::set<int>& labels)
{
    _assign(_final, labels);
}

void
FSTPronRules::set_ignore(const fst::set<int>& labels)
{
    _assign(_ignore, labels);
}

void
FSTPronRules::add_link(int final, int initial)
{
    _links_f[final] = initial;
}

int
FSTPronRules::expr_context(int label, bool left)
{
    fst::map<int, int>& m = left ? _links_l : _links_r;
    int& result = m[label];
    if (result == 0) {                          // New label mapping.
        string new_label;
        if (left) {                             // label -> ___label
            new_label = "___";
            new_label += FSTstring_map[label];
        } else {                                // label -> label___
            new_label = FSTstring_map[label];
            new_label += "___";
        }
        result = FSTstring_map[new_label];
    }
    return result;
}

void
FSTPronRules::add_rule(const fst::set<int>& left, int label,
                       const fst::set<int>& right, FSTGeneric fst)
{
    _RuleVector& rv = _rule_map[label];
    _Rule* r = new _Rule;

    r->index = rv.size();
    r->middle = label;
    fst::vector_insert(r->left, left.begin(), left.end());
    fst::vector_insert(r->right, right.begin(), right.end());

    ostringstream os;
    format(os, "%s.%d", FSTstring_map[r->middle], r->index);
    r->name = FSTstring_map[os.str()];

    r->fst = FSTOptimizeA(fst);

    rv.push_back(r);
}

static ostream& operator << (ostream& os, const fst::vector<int>& x)
{
    os << '{';
    fst::vector<int>::const_iterator i = x.begin();
    if (i != x.end()) {
        os << FSTstring_map[*i];
        while (++i != x.end())
            os << ' ' << FSTstring_map[*i];
    }
    os << '}';
    return os;
}

inline static ostream&
operator << (ostream& os, const FSTPronRules& rs)
{
    return rs.print(os);
}

inline static ostream&
operator << (ostream& os, const FSTPronRules::_Rule& r)
{
    os << r.left << ' ' << FSTstring_map[r.name] << ' ' << r.right << endl;
    return os;
}

ostream&
FSTPronRules::print(ostream& os) const
{
    os << "alphabet:" << endl << "  " << _alphabet << endl;
    os << "rules:" << endl;
    _RuleMap::const_iterator i;
    for (i = _rule_map.begin(); i != _rule_map.end(); ++i) {
        const _RuleVector& rv = (*i).second;
        _RuleVector::const_iterator j;
        for (j = rv.begin(); j != rv.end(); ++j) {
            os << "  " << *(*j);
        }
    }
    return os;
}

inline bool
FSTPronRules::_contains(const fst::vector<int>& s, int label)
{
    return s.empty() || binary_search(s.begin(), s.end(), label);
}

void
FSTPronRules::_find_left_matches(_RuleVector& result, const _RuleVector& rules, int l)
{
    result.clear();
    for (_RuleVector::const_iterator i = rules.begin(); i != rules.end(); i++) {
        if (_contains((*i)->left, l))
            result.push_back(const_cast<_Rule*>(*i));
    }
}

void
FSTPronRules::_find_right_matches(_RuleVector& result, const _RuleVector& rules, int r)
{
    result.clear();
    for (_RuleVector::const_iterator i = rules.begin(); i != rules.end(); i++) {
        if (_contains((*i)->right, r))
            result.push_back(const_cast<_Rule*>(*i));
    }
}

bool
FSTPronRules::_any_right_matches(const _RuleVector& rules, int r)
{
    for (_RuleVector::const_iterator i = rules.begin(); i != rules.end(); i++) {
        if (_contains((*i)->right, r))
            return true;
    }
    return false;
}

bool
FSTPronRules::_any_matches(const _RuleVector& rules, int l, int r)
{
    for (_RuleVector::const_iterator i = rules.begin(); i != rules.end(); i++) {
        const _Rule* rule = *i;
        if (_contains(rule->left, l) && _contains(rule->right, r))
            return true;
    }
    return false;
}

const FSTPronRules::_Rule*
FSTPronRules::_find_first_left_match(const _RuleVector& rules, int r)
{
    // Find first rules matching l on left.
    //
    for (_RuleVector::const_iterator i = rules.begin();
         i != rules.end(); i++) {
        if (_contains((*i)->left, r))
            return *i;
    }
    return NULL;
}

// Add label to the output side of the input FST.  This is functionally
// equivalent to prepending epsilon:label to the input FST, but produces a
// better result.  Basically, label is added to all initial arcs.  However, if
// an arc is an initial link arc, the label is placed on subsequent arcs.  If
// the initial state (or the state following an initial link arc) is final, an
// epsilon:label arc to a new final state is created.
//
FSTGeneric
FSTPronRules::_label_fst(const FSTGeneric& in, int label)
{
    FSTGeneric out = FSTBasic();
    out->add_states(in->n_states());

    int initial = in->initial();
    out->set_initial(initial);

    int final = out->add_state();
    out->set_final(final);

    if (in->is_final(initial))
        out->add_arc(initial, final, 0, label);

    for (int p = 0; p < in->n_states(); ++p) {
        if (in->is_final(p))
            out->set_final(p);
        for (FSTArcIter a = in->arcs_out(p); a.valid(); a++) {
            if (p == initial && _links_s.count(a->input)) {
                // An initial link arc.
                int q = out->add_state();
                out->add_arc(p, q, a->input, 0);
                if (in->is_final(a->next))
                    out->add_arc(q, final, 0, label);
                for (FSTArcIter b = in->arcs_out(a->next); b.valid(); b++) {
                    out->add_arc(q, b->next, b->input, label);
                }
            } else {
                out->add_arc(p, a->next, a->input, (p == initial) ? label : 0);
            }
        }
    }
    return FSTMinimize(FSTBasic(FSTDeterminizeA(out)));
}

void
FSTPronRules::_compile()
{
    // Construct _rules from _rule_map.  This contains all the rules in a flat
    // list.
    //
    _rules.clear();
    for (_RuleMap::iterator i = _rule_map.begin(); i != _rule_map.end(); ++i) {
        _RuleVector& rv = (*i).second;
        for (_RuleVector::iterator j = rv.begin(); j != rv.end(); ++j) {
            _rules.push_back(*j);
        }
    }

    // Construct _links_s, containing all initial and left links.
    //
    _links_s.clear();
    for (fst::map<int, int>::const_iterator i = _links_f.begin();
         i != _links_f.end(); ++i) {
        _links_s.insert((*i).second);
        if (_links_f.count((*i).second))
            format(cerr, "warning: `%s' both an initial and final link",
                      FSTstring_map[(*i).second]) << endl;
    }
    for (fst::map<int, int>::const_iterator i = _links_l.begin();
         i != _links_l.end(); ++i) {
        _links_s.insert((*i).second);
    }

    // Construct _links_a, containing all (initial, final, left, right) links.
    //
    _links_a = _links_s;
    for (fst::map<int, int>::const_iterator i = _links_f.begin();
         i != _links_f.end(); ++i) {
        _links_a.insert((*i).first);
    }
    for (fst::map<int, int>::const_iterator i = _links_r.begin();
         i != _links_r.end(); ++i) {
        _links_a.insert((*i).second);
    }

    // Construct transducers C, E, L, and R.
    //
    C = _construct_C();
    E = _construct_E();
    _construct_LR(L, R);

    // Compute final Pl and Pr and we're done.
    //
    _Pr = R;

    FSTGeneric CE = FSTOptimizeA(FSTBasic(FSTCompose(C, E)));
    
    _Pl = FSTOptimizeA(FSTCache(FSTCompose(CE, L)));

    // Warn if all possible input sequences are not supported by the rules.
    //
    _check_coverage(CE, L, R);

}

// Construct L and R.
//
void
FSTPronRules::_construct_LR(FSTGeneric& L, FSTGeneric& R)
{
    _RuleVector matches;

    // Populate map from rule sets matching (middle, right) to FST labels.
    //
    fst::map<_RuleVector, int> rules_label_map;
    _alphabet.push_back(0);
    for (fst::vector<int>::const_iterator mi = _alphabet.begin();
         mi != _alphabet.end(); ++mi) {
        int m = *mi;
        if (m == 0)
            continue;
        const _RuleVector& m_rules = _rule_map[m];
        if (m_rules.empty())
            continue;
        for (fst::vector<int>::const_iterator ri = _alphabet.begin();
             ri != _alphabet.end(); ++ri) {
            int r = *ri;
            _find_right_matches(matches, m_rules, r);
            if (matches.empty())
                continue;
            int& label = rules_label_map[matches];
            if (label == 0) {
                ostringstream scratch;
                scratch << FSTstring_map[m];
                for (_RuleVector::const_iterator i = matches.begin();
                     i != matches.end(); ++i) {
                    format(scratch, ".%d", (*i)->index);
                }
                label = FSTstring_map[scratch.str()];
            }
        }
    }
    _alphabet.pop_back();

    // Populate map from label to state.
    //
    fst::map<int, int> label_state_map;
    label_state_map[0] = 0;
    int p = 1;
    for (fst::vector<int>::const_iterator i = _alphabet.begin();
         i != _alphabet.end(); ++i) {
        label_state_map[*i] = p++;
    }

    // Build R.  R is applied right-to-left to an FST, converting input
    // symbols to rule sets that are compatible with the right context.
    //
    R = FSTBasic();
    R->add_states(label_state_map.size());
    R->set_initial(0);
    if (_final.empty())
        R->set_final(0);
    for (fst::map<int, int>::const_iterator i = label_state_map.begin();
         i != label_state_map.end(); ++i) {
        int x = (*i).first;
        int p = (*i).second;
        if (p == 0)
            continue;
        if (_contains(_initial, x))
            R->set_final(p);
        if (_contains(_final, x)) {
            _find_right_matches(matches, _rule_map[x], 0);
            if (matches.empty())
                continue;
            int label = rules_label_map[matches];
            assert(label != 0);
            R->add_arc(0, p, label, x);
        }
        for (fst::vector<int>::const_iterator j = _alphabet.begin();
             j != _alphabet.end(); ++j) {
            int y = *j;
            _find_right_matches(matches, _rule_map[y], x);
            if (matches.empty())
                continue;                       // already warned
            int label = rules_label_map[matches];
            R->add_arc(p, label_state_map[y], label, y);
        }
    }
    R = FSTMinimize(R);
    _add_ignore_loops(R);

    // Build L.  L is applied left-to-right to the result of (R o input),
    // and resolves the rule sets down to the first rule that is compatible
    // with the left context.
    //
    L = FSTBasic();
    L->add_states(label_state_map.size());
    L->set_initial(0);
    if (_initial.empty())
        L->set_final(0);
    for (fst::map<_RuleVector, int>::const_iterator i = rules_label_map.begin();
         i != rules_label_map.end(); ++i) {
        const _RuleVector& rules = (*i).first;
        const _Rule* rule = _find_first_left_match(rules, 0);
        if (rule && _contains(_initial, rule->middle)) {
            L->add_arc(0, label_state_map[rule->middle],
                        rule->name, (*i).second);
        }
    }
    for (fst::map<int, int>::const_iterator i = label_state_map.begin();
         i != label_state_map.end(); ++i) {
        int x = (*i).first;
        int p = (*i).second;
        if (p == 0)
            continue;
        if (_contains(_final, x))
            L->set_final(p);
        for (fst::map<_RuleVector, int>::const_iterator j = rules_label_map.begin();
             j != rules_label_map.end(); ++j) {
            const _RuleVector& rules = (*j).first;
            const _Rule* rule = _find_first_left_match(rules, x);
            if (rule) {
                L->add_arc(p, label_state_map[rule->middle],
                            rule->name, (*j).second);
            }
        }
    }                
    L = FSTMinimize(L);
    _add_ignore_loops(L);
}

// Construct E to expand rules.
//
FSTGeneric
FSTPronRules::_construct_E()
{
    fst::vector<FSTGeneric> rule_fsts;
    for (_RuleMap::iterator i = _rule_map.begin();
         i != _rule_map.end(); ++i) {
        _RuleVector& rv = (*i).second;
        for (_RuleVector::iterator j = rv.begin(); j != rv.end(); ++j) {
            rule_fsts.push_back(_label_fst((*j)->fst, (*j)->name));
        }
    }

    FSTGeneric E = FSTBasic(FSTRemoveEpsilonsA(FSTUnion(rule_fsts)));
    assert(E->initial() == 0);
    for (fst::vector<int>::const_iterator i = _ignore.begin();
         i != _ignore.end(); ++i) {
        E->add_arc(0, 0, *i, *i);
    }
    E = FSTBasic(FSTClosure(E));

    return E;
}

static inline int
_find(const fst::map<int,int>& m, int label)
{
    fst::map<int,int>::const_iterator mi = m.find(label);
    return (mi != m.end()) ? (*mi).second : 0;
}

// Construct C to enforce link consistency (e.g., foo$ must be followed by $foo
// if "connect foo$ $foo;" was specified).
//
FSTGeneric
FSTPronRules::_construct_C()
{
    // Construct alphabet of all non-link symbols in all rules.
    //
    fst::vector<int> alphabet;
    fst::set<int> tmp;
    for (_RuleMap::const_iterator i = _rule_map.begin();
         i != _rule_map.end(); ++i) {
        const _RuleVector& rv = (*i).second;
        for (_RuleVector::const_iterator j = rv.begin();
             j != rv.end(); ++j) {
            const fst::set<int>& a = (*j)->fst->input_alphabet();
            fst::set_insert(tmp, a.begin(), a.end());
        }
    }
    set_difference(tmp.lower_bound(1), tmp.end(),
                   _links_a.begin(), _links_a.end(),
                   back_insert_iterator< fst::vector<int> >(alphabet));

    FSTGeneric C = FSTBasic();
    C->add_state();
    C->set_initial(0);
    C->set_final(0);

    fst::vector<int>::const_iterator vi, vj;
    fst::map<int, int>::const_iterator mi;

    // (eps:raux(sigma))* sigma:sigma (eps:laux(sigma))*
    //
    for (vi = alphabet.begin(); vi != alphabet.end(); ++vi) {
        int sigma = *vi;
        int p = C->add_states(2);
        C->add_arc(0, p, 0, 0);
        C->add_arc(p, p + 1, sigma, sigma);
        C->add_arc(p + 1, 0, 0, 0);
        int raux = _find(_links_r, sigma);
        if (raux) C->add_arc(p, p, 0, raux);
        int laux = _find(_links_l, sigma);
        if (laux) C->add_arc(p + 1, p + 1, 0, laux);
    }

    // sigma1:sigma1 (eps:laux(sigma1) | eps:raux(sigma2))* sigma2:sigma2
    //
    for (vi = alphabet.begin(); vi != alphabet.end(); ++vi) {
        int sigma1 = *vi;
        int laux = _find(_links_l, sigma1);
        if (laux) {
            for (vj = alphabet.begin(); vj != alphabet.end(); ++vj) {
                int sigma2 = *vj;
                int raux = _find(_links_r, sigma2);
                if (raux) {
                    int p = C->add_state();
                    C->add_arc(0, p, sigma1, sigma1);
                    C->add_arc(p, p, 0, laux);
                    C->add_arc(p, p, 0, raux);
                    C->add_arc(p, 0, sigma2, sigma2);
                }
            }
        }
    }

    // eps:rconn(alpha) eps:lconn(alpha)
    //
    // Final to initial links.
    //
    for (mi = _links_f.begin(); mi != _links_f.end(); ++mi) {
        int rconn = (*mi).first;
        int lconn = (*mi).second;
        int p = C->add_state();
        C->add_arc(0, p, 0, rconn);
        C->add_arc(p, 0, 0, lconn);
    }

    C = FSTOptimizeA(C);
    _add_ignore_loops(C);

    return C;
}

bool
FSTPronRules::_compatible(const _Rule* a, const _Rule* b)
{
    return _contains(a->right, b->middle) && _contains(b->left, a->middle);
}

void
FSTPronRules::_add_ignore_loops(FSTGeneric x)
{
    if (!_ignore.empty()) {
        int N = x->n_states();
        for (int p = 0; p < N; ++p) {
            for (fst::vector<int>::const_iterator i = _ignore.begin();
                 i != _ignore.end(); ++i) {
                x->add_arc(p, p, *i, *i);
            }
        }
    }
}

void
FSTPronRules::_compute_transitions(fst::set<IntPair>& transitions, FSTGeneric fsa)
{
    fst::set<IntPair> explored;                      // set of (left label, state)
    fst::list<IntPair> todo;

    todo.push_back(IntPair(0, fsa->initial()));
    while (!todo.empty()) {
        if (explored.count(todo.front())) {
            todo.pop_front();
            continue;
        }
        explored.insert(todo.front());
        int left = todo.front().first;
        int p = todo.front().second;
        todo.pop_front();
        for (FSTArcIter a = fsa->arcs_out(p); a.valid(); a++) {
            assert(a->input != 0);
            if (!_ignore.empty() && _contains(_ignore, a->input)) {
                todo.push_back(IntPair(left, a->next));
            } else {
                todo.push_back(IntPair(a->input, a->next));
                if (left != 0)
                    transitions.insert(IntPair(left, a->input));
            }
        }
    }
}

void
FSTPronRules::_check_coverage(FSTGeneric CE, FSTGeneric L, FSTGeneric R)
{
    // Check that coverage of left/middle/right input labels.  This check does
    // not look at the rule realizations (surface form) at all, and thus does
    // not detect input sequences blocked by left/right connections or
    // left/right surface constraints.
    //
    fst::set<int> right_warned;
    _RuleVector lm_rules;
    for (fst::vector<int>::const_iterator m = _alphabet.begin();
         m != _alphabet.end(); ++m) {
        const _RuleVector& m_rules = _rule_map[*m];
        if (m_rules.empty()) {
            format(cerr, "warning: `%s' has no rules",
                   FSTstring_map[*m]) << endl;
            continue;
        }
        right_warned.clear();
        for (fst::vector<int>::const_iterator r = _alphabet.begin();
             r != _alphabet.end(); ++r) {
            if (!_any_right_matches(m_rules, *r)) {
                right_warned.insert(*r);
                format(cerr, "warning: `%s' has no match for right `%s'",
                       FSTstring_map[*m], FSTstring_map[*r]) << endl;
                continue;
            }
        }
        for (fst::vector<int>::const_iterator l = _alphabet.begin();
             l != _alphabet.end(); ++l) {
            _find_left_matches(lm_rules, m_rules, *l);
            if (lm_rules.empty()) {
                format(cerr, "warning: `%s' has no match for left `%s'",
                       FSTstring_map[*m], FSTstring_map[*l]) << endl;
                continue;
            }
            for (fst::vector<int>::const_iterator r = _alphabet.begin();
                 r != _alphabet.end(); ++r) {
                if (!right_warned.count(*r) && !_any_matches(lm_rules, *l, *r)) {
                    format(cerr, "warning: `%s' has no match "
                           "for left `%s' and right `%s'",
                           FSTstring_map[*m], FSTstring_map[*l],
                           FSTstring_map[*r]) << endl;
                    continue;
                }
            }
        }
    }

    //
    // Check realizations.
    //

    // Compute FSA representing allowable rule sequences according to LoR.
    //
    FSTGeneric LRvalid =
        FSTOptimizeA(FSTProject(FSTCompose(FSTBasic(L),
                                           FSTOptimizeA(FSTReverse(R))),
                                FSTInput, FSTInput));

    // Intersect the LoR sequences with those imposed by CE (i.e., realization
    // constraints).
    //
    FSTGeneric CEvalid =
        FSTOptimizeA(FSTBasic(FSTCompose(FSTProject(CE, FSTOutput, FSTOutput),
                                         LRvalid)));

    // Compute true and actual rule transitions.
    //
    fst::set<IntPair> true_transitions, actual_transitions;
    _compute_transitions(true_transitions, LRvalid);
    _compute_transitions(actual_transitions, CEvalid);

    // Compute true transitions that are missing in actual, and issue warning.
    //
    fst::vector<IntPair> missing;
    set_difference(true_transitions.begin(), true_transitions.end(),
                   actual_transitions.begin(), actual_transitions.end(),
                   back_insert_iterator< fst::vector<IntPair> >(missing));
    if (_verbosity >= 2 && !missing.empty()) {
        cerr << "warning: "
            "the following rule transitions fail due to surface constraints"
             << endl;
        for (fst::vector<IntPair>::const_iterator i = missing.begin();
             i != missing.end(); ++i) {
            cerr << '\t'
                 << FSTstring_map[(*i).first] << ' '
                 << FSTstring_map[(*i).second] << endl;
        }
    }
}

FSTGeneric
FSTPronRules::apply(FSTGeneric in) const
{
    if (bad())
        return FSTBasic();

    // Check that rules support all input FST labels and warn if not.
    //
    const fst::set<int>& rule_labels = _Pr->output_alphabet();
    const fst::set<int>& in_labels = in->input_alphabet();
    fst::vector<int> missing_labels;
    set_difference(in_labels.lower_bound(1), in_labels.end(),
                   rule_labels.begin(), rule_labels.end(),
                   back_insert_iterator< fst::vector<int> >(missing_labels));
    if (!missing_labels.empty()) {
        cerr << "warning: the following labels not supported by rules:"
             << endl;
        for (fst::vector<int>::const_iterator i = missing_labels.begin();
             i != missing_labels.end(); ++i) {
            cerr << '\t' << FSTstring_map[*i] << endl;
        }
    }

    // Apply rules: Pr right-to-left and Pl left-to-right.
    //
    FSTGeneric x;
    x = FSTReverse(FSTCompose(_Pr, FSTReverse(in)));
    x = FSTBasic(FSTCompose(_Pl, x));

    return x;
}

void
FSTPronRules::dump(const char* a_file, const char* b_file, const char* c_file) const
{
    FSTGeneric r = FSTOptimizeA(FSTReverse(R));
    FSTGeneric l = FSTBasic(L);
    FSTGeneric LR = FSTOptimizeA(FSTTrim(FSTCompose(l, r)));

    C->dump_ascii(a_file);
    E->dump_ascii(b_file);
    LR ->dump_ascii(c_file);
}

FSTGeneric
FSTPronRules::single_fst() const
{
    return FSTBasic(FSTCompose(_Pl, FSTOptimizeA(FSTReverse(_Pr))));
}

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 79
// indent-tabs-mode: nil
// c-basic-offset: 4
// End:
