////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

#include "fst/FST_i.h"

FSTGeneric
FSTPushWeights(const FSTGeneric& x)
{
    // Compute potential for each state.
    //
    fst::vector<float> potential;
    FSTPotential(potential, x);
    assert(int(potential.size()) == x->n_states());

    // Build new FST with pushed weights.
    //
    FSTSemiring semiring = x->semiring();
    int N = x->n_states();
    int initial = x->initial();
    FSTGeneric y = FSTBasic(semiring);
    y->add_states(N);
    y->set_initial(initial);

    // Perform an iterative DFS to determine which states are part of an initial cycle.
    // cycle[p] will be true iff state p is part of such a cycle.
    //
    typedef std::pair<int, FSTArcIter> StateArcs;
    typedef fst::vector<StateArcs> StateArcsVector;
    fst::vector<bool> visited(N, false), cycle(N, false);
    StateArcsVector todo;
    visited[initial] = true;
    todo.push_back(StateArcs(initial, x->arcs_out(initial)));
    while (!todo.empty()) {
        StateArcs& top = todo.back();
        FSTArcIter& a = top.second;
        if (a.valid()) {
            int q = a->next;
            ++a;
            if (q == initial || cycle[q]) {
                for (int i = todo.size() - 1; i >= 0 && !cycle[todo[i].first]; i--) {
                    cycle[todo[i].first] = true;
                }
            } else if (!visited[q]) {
                visited[q] = true;
                todo.push_back(StateArcs(q, x->arcs_out(q)));
            }
        } else {
            todo.pop_back();
        }
    }

    // Produce output FST, incorporating potential[initial] when leaving the initial cycle
    // (if there is an initial cycle).
    //
    if (cycle[initial]) {
        for (int p = 0; p < N; p++) {
            if (x->is_final(p)) {
                float fw = semiring->divide(x->final_weight(p), potential[p]);
                if (cycle[p]) {
                    fw = semiring->times(fw, potential[initial]);
                }
                y->set_final(p, fw);
            }
            for (FSTArcIter a = x->arcs_out(p); a.valid(); a++) {
                int q = a->next;
                float aw =
                    semiring->divide(semiring->times(a->weight, potential[q]),
                                     potential[p]);
                if (cycle[p] && !cycle[q]) {
                    aw = semiring->times(aw, potential[initial]);
                }
                y->add_arc(p, q, a->input, a->output, aw);
            }
        }
    } else {
        potential[initial] = semiring->one;
        for (int p = 0; p < N; p++) {
            if (x->is_final(p)) {
                float fw = semiring->divide(x->final_weight(p), potential[p]);
                y->set_final(p, fw);
            }
            for (FSTArcIter a = x->arcs_out(p); a.valid(); a++) {
                int q = a->next;
                float aw =
                    semiring->divide(semiring->times(a->weight, potential[q]),
                                     potential[p]);
                y->add_arc(p, q, a->input, a->output, aw);
            }
        }
    }

    return y;
}

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 80
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
