////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

#define FINAL_TAG     -1
#define POTENTIAL_TAG -2

//#define DEBUG

#ifdef DEBUG
#  define protected public
#endif

#include "fst/FST_i.h"
#include "Heap.h"

class FSTRemoveEpsilonsRep
    : public FSTGenericInt {

public: // methods
    FSTRemoveEpsilonsRep(const FSTGeneric& in, bool fsa, float delta);
    virtual ~FSTRemoveEpsilonsRep();
    virtual const fst::set<int>& input_alphabet() const;
    virtual const fst::set<int>& output_alphabet() const;
    virtual bool is_final(int p) const;
    virtual float final_weight(int p) const;
    virtual float potential(int p) const;
    virtual FSTArcIter arcs_out(int p, FSTArcOrder order=FSTNone) const;
    virtual FSTArcIter arcs_out_with_input(int p, int i,
                                           FSTArcMode m=FSTEqual) const;
    virtual FSTArcIter arcs_out_with_output(int p, int o,
                                            FSTArcMode m=FSTEqual) const;
    virtual void force() const;
    virtual void reset() const;

    virtual FSTGeneric reload(bool& flag);

    void state_set(int p, fst::set<int>& result) const;

protected: // types
    struct _State;
    struct _ColorWeight;
    typedef fst::vector<_State*> _StatePtrVector;
    typedef pair<int, int> _StateOutput;
    typedef fst::map<_StateOutput, float> _Set;
    typedef _Set _Map;
    struct _SetLess : binary_function<_Set*, _Set*, bool> {
        float delta;
        _SetLess(float delta_) : delta(delta_) {}
        bool operator()(const _Set*, const _Set*) const;
    };
    typedef fst::map<_Set*, int, _SetLess > _SetMap;

#ifdef HASH
    struct _StateOutputHash {
        size_t operator()(const _StateOutput& so) const {
            return 0x1001*so.first + 0x11*so.second;
        }
    };
    struct _StateOutputEqual {
        bool operator()(const _StateOutput& a, const _StateOutput& b) const {
            return a == b;
        }
    };
    typedef fst::hash_map<_StateOutput, int, _StateOutputHash, _StateOutputEqual> _CacheMap;
#else
    typedef fst::map<_StateOutput, int> _CacheMap;
#endif
public:
    typedef pair<_StateOutput, float> _Entry;

    typedef _Set::iterator _SetIter;

    struct _StackEntry {
        _SetIter iter;
        FSTArcIter arcs;
        _StackEntry(_SetIter si, const FSTArcIter& a)
            : iter(si), arcs(a) {
        }
    };

    struct _QueueEntry {
        int state;
        int output;
        mutable float weight;
        mutable int index;
        _QueueEntry(int s, int o, float w)
            : state(s), output(o), weight(w), index(-1) {
        }
        void set_heap_index(int i) const {
            index = i;
        }
        int get_heap_index() const {
            return index;
        }
        struct Index {
            bool operator()(const _QueueEntry& x, const _QueueEntry& y) const {
                if (x.state < y.state) return true;
                if (x.state > y.state) return false;
                return x.output < y.output;
            }
        };
        struct Better {
            FSTSemiring semiring;
            Better() : semiring(0) {
            }
            Better(FSTSemiring s) : semiring(s) {
            }
            bool operator()(const _QueueEntry* x, const _QueueEntry* y) const {
                return semiring->is_better(x->weight, y->weight);
            }
        };
    };

protected: // members
    FSTGeneric              _in;
    mutable _StatePtrVector _states;
    mutable _SetMap         _set_map;
    mutable _CacheMap       _cache_map;
    bool                    _fsa;

protected: // methods
    int _init() const;
    void _clean() const;
    void _add_arcs(fst::vector<FSTArc*>& arcs, int p, FSTArcIter& a, int o, float weight) const;
    void _add_final_arc(fst::vector<FSTArc*>& arcs, int p, int q, int o, float w) const;
    int _find_or_add_state(int in_state, int output) const;
    _Set* _closure(const _StateOutput& state_output) const;
    bool _closure_idempotent_acyclic(_Set* set_, const _StateOutput& state_output, bool& negative) const;
    void _closure_idempotent_cyclic_bellman_ford(_Set* set_, const _StateOutput& state_output) const;
    void _closure_idempotent_cyclic_dijkstra(_Set* set_, const _StateOutput& state_output) const;
    void _closure_idempotent(_Set* set_, const _StateOutput& state_output) const;
    void _closure_non_idempotent(_Set* set_, const _StateOutput& state_output) const;
    void _closure_non_idempotent_dfs(fst::vector<int>& order, fst::map<int, _ColorWeight>& color_weight,
                                     int p, int q, float w) const;
    void _prune(_Set* set_) const;

};

typedef FSTRemoveEpsilonsRep::_Entry _Entry;

struct FSTRemoveEpsilonsRep::_State {
    _Set* set_;
    _State(_Set* s) : set_(s) {}
    ~_State() { delete set_; }
    FSTAllocator_new_delete(_State);
};

struct FSTRemoveEpsilonsRep::_ColorWeight {

    enum {White, Gray, Black} color;
    float self_loop_weight;

    _ColorWeight() : color(White) {}
};

inline static int
_state(const _Entry& e)
{
    return e.first.first;
}

inline static int
_output(const _Entry& e)
{
    return e.first.second;
}

inline static float
_weight(const _Entry& e)
{
    return e.second;
}

inline static float&
_weight(_Entry& e)
{
    return e.second;
}

bool
FSTRemoveEpsilonsRep::_SetLess::operator()(const _Set* x, const _Set* y) const
{
    _Set::const_iterator xi, yi;
    for (xi = x->begin(), yi = y->begin();
         xi != x->end() && yi != y->end();
         ++xi, ++yi) {
        if (_state(*xi)  < _state(*yi) ) return true;
        if (_state(*yi)  < _state(*xi) ) return false;
        if (_output(*xi) < _output(*yi)) return true;
        if (_output(*yi) < _output(*xi)) return false;
        if (_weight(*xi) + delta < _weight(*yi)) return true;
        if (_weight(*yi) + delta < _weight(*xi)) return false;
    }
    return xi == x->end() &&  yi != y->end();
}

#if 0
inline static bool
operator < (const _Entry& x, const _Entry& y) {
    return FST_LESS3(_weight(x), _weight(y),
                     FST_LESS3(_state(x), _state(y),
                               FST_LESS2(_output(x), _output(y))));
}
#endif

#ifdef DEBUG
static ostream&
operator << (ostream& os, const FSTRemoveEpsilonsRep::_Set::value_type& x)
{
    return os << '('
              << _state(x) << ','
              << FSTstring_map[_output(x)] << ','
              << _weight(x) << ')';
}

static ostream&
operator << (ostream& os, const FSTRemoveEpsilonsRep::_Set* x)
{
    os << '{';
    for (FSTRemoveEpsilonsRep::_Set::const_iterator i = x->begin();
         i != x->end(); ++i) {
        if (i != x->begin())
            os << ',';
        os << *i;
    }
    os << '}';
    return os;
}
#endif

FSTRemoveEpsilonsRep::FSTRemoveEpsilonsRep(const FSTGeneric& in,
                                           bool fsa, float delta)
    : _in(in),
      _set_map(_SetLess(delta)),
      _fsa(fsa)
{
    _class_name = "FSTRemoveEpsilons";
    _semiring = _in->semiring();
    assert(_semiring->idempotent() || _fsa);

    _initial = _init();
}


FSTRemoveEpsilonsRep::~FSTRemoveEpsilonsRep()
{
    _clean();
}

FSTGeneric
FSTRemoveEpsilonsRep::reload(bool& flag)
{
    if (_in.reload()) flag = true;
    reset();
    return FSTGeneric(this);
}

void
FSTRemoveEpsilonsRep::reset() const
{
    _in->reset();
    _clean();

    int new_initial = _init();
    assert(new_initial == _initial);
}

int
FSTRemoveEpsilonsRep::_init() const
{
    return (_in->initial() >= 0) ?
        _find_or_add_state(_in->initial(), 0)
        : -1;
}

void
FSTRemoveEpsilonsRep::_clean() const
{
    for (_StatePtrVector::iterator i = _states.begin();
         i != _states.end(); i++) {
        delete *i;
    }
    _states.clear();
    _cache_map.clear();
    _set_map.clear();
    _n_states = 0;
}

const fst::set<int>&
FSTRemoveEpsilonsRep::input_alphabet() const
{
    return _in->input_alphabet();
}

const fst::set<int>&
FSTRemoveEpsilonsRep::output_alphabet() const
{
    return _in->output_alphabet();
}

bool
FSTRemoveEpsilonsRep::is_final(int p) const
{
    FST_ASSERT_VALID_STATE(p);
    _Set::const_iterator i = _states[p]->set_->find(_StateOutput(FINAL_TAG, 0));
    return (*i).second != _semiring->zero;
}

float
FSTRemoveEpsilonsRep::final_weight(int p) const
{
    FST_ASSERT_VALID_STATE(p);
    _Set::const_iterator i = _states[p]->set_->find(_StateOutput(FINAL_TAG, 0));
    return (*i).second;
}

float
FSTRemoveEpsilonsRep::potential(int p) const
{
    FST_ASSERT_VALID_STATE(p);
    _Set::const_iterator i = _states[p]->set_->find(_StateOutput(POTENTIAL_TAG, 0));
    return (*i).second;
}

FSTArcIter
FSTRemoveEpsilonsRep::arcs_out(int p, FSTArcOrder order) const
{
    FST_ASSERT_VALID_STATE(p);
    _Set& set_ = *_states[p]->set_;
    FSTArcIterData* arc_data = new FSTArcIterData;
    fst::vector<FSTArc*>& arcs = arc_data->arcs;
    for (_Set::iterator si = set_.begin(); si != set_.end(); si++) {
        int q = _state(*si);
        if (q >= 0) {
            FSTArcIter a = _in->arcs_out(q);
            _add_arcs(arcs, p, a, _output(*si), _weight(*si));
            if (_output(*si) && _in->is_final(q))
                _add_final_arc(arcs, p, q, _output(*si), _weight(*si));
        }
    }
    if (order != FSTNone)
        _sort_arcs(arcs, order);
    return FSTArcIter(arc_data);
}

FSTArcIter
FSTRemoveEpsilonsRep::arcs_out_with_input(int p, int i, FSTArcMode m) const
{
    FST_ASSERT_VALID_STATE(p);
    _Set& set_ = *_states[p]->set_;
    FSTArcIterData* arc_data = new FSTArcIterData;
    fst::vector<FSTArc*>& arcs = arc_data->arcs;
    for (_Set::iterator si = set_.begin(); si != set_.end(); si++) {
        int q = _state(*si);
        if (q >= 0) {
            bool epsilon_included =
                (i == 0) ||
                (m == FSTLessEqual && i > 0) ||
                (m == FSTGreaterEqual && i < 0);
            if (epsilon_included && _output(*si) && _in->is_final(q))
                _add_final_arc(arcs, p, q, _output(*si), _weight(*si));
            FSTArcIter a = _in->arcs_out_with_input(q, i, m);
            _add_arcs(arcs, p, a, _output(*si), _weight(*si));
        }
    }
    // An optimization might be to use inplace_merge(...).
    _sort_arcs(arcs, FSTInput);
    return FSTArcIter(arc_data);
}

FSTArcIter
FSTRemoveEpsilonsRep::arcs_out_with_output(int p, int o, FSTArcMode m) const
{
    FST_ASSERT_VALID_STATE(p);
    _Set& set_ = *_states[p]->set_;
    FSTArcIterData* arc_data = new FSTArcIterData;
    fst::vector<FSTArc*>& arcs = arc_data->arcs;
    for (_Set::iterator si = set_.begin(); si != set_.end(); si++) {
        int q = _state(*si);
        if (q >= 0) {
            bool output_included =
                (o == _output(*si)) ||
                (m == FSTLessEqual && o > _output(*si)) ||
                (m == FSTGreaterEqual && o < _output(*si));
            if (output_included && _output(*si) && _in->is_final(q))
                _add_final_arc(arcs, p, q, _output(*si), _weight(*si));
            if (output_included && _output(*si)) {
                FSTArcIter a = _in->arcs_out(q);
                _add_arcs(arcs, p, a, _output(*si), _weight(*si));
            } else if (!_output(*si)) {
                FSTArcIter a = _in->arcs_out_with_output(q, o, m);
                _add_arcs(arcs, p, a, 0, _weight(*si));
            }
        }
    }
    // An optimization might be to use inplace_merge(...).
    _sort_arcs(arcs, FSTInput);
    return FSTArcIter(arc_data);
}

void
FSTRemoveEpsilonsRep::force() const
{
    // This could be improved by using something that doesn't actually store all
    // the arcs that we are just going to toss away.
    //
    for (int p = 0; p < _n_states; p++)
        arcs_out(p);
}


void
FSTRemoveEpsilonsRep::_add_arcs(
    fst::vector<FSTArc*>& arcs, int p, FSTArcIter& a, int o, float w) const
{
    // This was fatal to performance!
    // arcs.reserve(arcs.size() + a.n());
    if (_fsa) {
        for (; a.valid(); a++) {
            if (a->input || a->output) {
                int q = _find_or_add_state(a->next, 0);
                arcs.push_back(new FSTArc(p, q, a->input, a->output,
                                          _semiring->times(a->weight, w)));
            }
        }
    } else {
        for (; a.valid(); a++) {
            if (a->input || a->output) {
                int pending_output = (o && a->output) ? a->output : 0;
                if (!pending_output && !a->input)
                    continue;                       // a covered by closure
                int arc_output = o ? o : a->output;
                int q = _find_or_add_state(a->next, pending_output);
                assert(a->input || arc_output);
                arcs.push_back(new FSTArc(p, q, a->input, arc_output,
                                          _semiring->times(a->weight, w)));
            }
        }
    }
}

void
FSTRemoveEpsilonsRep::_add_final_arc(
    fst::vector<FSTArc*>& arcs, int p, int q, int o, float w) const
{
    int r = _find_or_add_state(-1, 0);
    assert(o != 0);
    assert(_fsa == false);
    arcs.push_back(new FSTArc(p, r, 0, o,
                              _semiring->times(_in->final_weight(q), w)));
}

int
FSTRemoveEpsilonsRep::_find_or_add_state(int in_state, int output) const
{
    _StateOutput state_output(in_state, output);

    // Check to see if we have already mapped this input state to a state.
    //
    _CacheMap::iterator ci = _cache_map.find(state_output);
    if (ci != _cache_map.end())
        return (*ci).second;

    // Compute closure of (input state, output), producing a _Set* containing
    // (state, output, weight) triples.
    //
    _Set* set_ = _closure(state_output);

    // P will hold the resulting state number.
    //
    int p;

    _SetMap::iterator i = _set_map.find(set_);
    if (i == _set_map.end()) {

        // State not found: build new state.
        //
        _State* pp = new _State(set_);
        _states.push_back(pp);
        p = _n_states++;
        _set_map[set_] = p;
        _cache_map[state_output] = p;
#ifdef DEBUG
        cerr << "new state: " << p << ' ' << set_ << endl;
#endif

    } else {

        // State found: look up state index and free set (it isn't needed).
        p = (*i).second;
        delete set_;

    }

    // Store in cache.
    //
    _cache_map[state_output] = p;

    return p;
}

FSTRemoveEpsilonsRep::_Set*
FSTRemoveEpsilonsRep::_closure(const _StateOutput& state_output) const
{
    // THIS COMMENT IS OUT OF DATE!
    //
    // Compute weighted epsilon closure: all (state, weight) pairs reachable
    // from _in state p.  We only keep at most one pair for each state, hence
    // the use of a state -> weight map (called _Set).  We compute the closure
    // using a best-first search along epsilon arcs.
    //
    // For the MinPlus semiring, this will not terminate if there is an epsilon
    // cycle with negative weight.  For the PlusTimes and PlusTimesLog semirings
    // this will not terminate if there is an epsilon cycle with non-zero
    // weight.  This cases could be detected with a depth-first search (and the
    // states might be topologically sortable), but this would cause complexity
    // and likely slow things down unnecessarily.
    //

#ifdef DEBUG
    cerr << "computing closure on (" << state_output.first << ", " << state_output.second << ")" << endl;
#endif
    _Set* set_ = new _Set;

    // Special final state.
    //
    if (state_output.first < 0) {
        set_->insert(_Entry(_StateOutput(FINAL_TAG, 0), _semiring->one));
        set_->insert(_Entry(_StateOutput(POTENTIAL_TAG, 0), _semiring->one));
        return set_;
    }

    // Branch to appropriate closure routine.
    //
    if (_semiring->idempotent())
        _closure_idempotent(set_, state_output);
    else
        _closure_non_idempotent(set_, state_output);

#ifdef DEBUGX
    cerr << "after search:" << endl << set_ << endl;
#endif

    // Prune entries that cannot produce arcs.
    //
    _prune(set_);

    return set_;
}

// TODO:
//
// USE THIS FOR NON-IDEMPOTENT SEMIRINGS AS WELL.  WE WILL HAVE TO HANDLE
// SELF-LOOP WEIGHTS USING SEMIRING->CLOSURE, WHICH WE ARE NOT CURRENTLY
// DOING IN THIS CODE.

bool
FSTRemoveEpsilonsRep::_closure_idempotent_acyclic(
    _Set* set_, const _StateOutput& state_output, bool& negative) const
{
    // We use these weight values as flags during the DFS to detect cycles.
    //
    const float BUSY = 1;
    const float DONE = 2;

    assert(set_->empty());

    // Topologically sort epsilon-reachable states using an interative DFS.  The
    // result is a list of set iterators.  We use the set to keep track of which
    // (state,output) have been visited.
    //
    typedef fst::list<_SetIter> List;
    typedef fst::list<_StackEntry> Stack;
    List order;
    Stack todo;
    bool cyclic = false;
    negative = false;

    todo.push_back(_StackEntry(set_->insert(_Set::value_type(state_output, BUSY)).first,
                               _in->arcs_out_with_input(state_output.first, 0)));
    while (!todo.empty()) {
        _StackEntry& top = todo.back();
        FSTArcIter& a = top.arcs;
        if (a.valid()) {
            _SetIter psi = top.iter;
            const _StateOutput& pso = (*psi).first;
            //int p = pso.first;
            int po = pso.second;
            if (!(a->output && (po || _fsa))) {
                if (!negative && _semiring->is_better(a->weight, _semiring->one)) {
                    negative = true;
                }
                int q = a->next;
                int qo = po ? po : a->output;
                pair<_Set::iterator,bool> status =
                    set_->insert(_Set::value_type(_StateOutput(q, qo), BUSY));
                _SetIter qsi = status.first;
                if (status.second) {
                    todo.push_back(_StackEntry(status.first, _in->arcs_out_with_input(q, 0)));
                } else if ((*qsi).second == BUSY && qsi != psi) {
                    cyclic = true;
                }
            }
            ++a; // OK here because list-based stack does not invalidate references on growth.
        } else {
            (*top.iter).second = DONE;
            order.push_front(top.iter);
            todo.pop_back();
        }
    }

    // Abort if we have detected a cycle.  However, we are returning the set of
    // (state,output) that are reachable and whether or not there were any
    // negative cost arcs.
    //

    if (cyclic) {
        return false;
    }

    // Reset weights in set (should be DONE): zero for all (state,output) except
    // initial, which gets one.
    //
    for (List::iterator i = order.begin(); i != order.end(); i++) {
        (*(*i)).second = _semiring->zero;
    }
    (*set_)[state_output] = _semiring->one;

    // Traverse (state,output) in topological order, pushing weights forward.
    //
    for (List::iterator i = order.begin(); i != order.end(); i++) {
        _Entry e = *(*i);
        int p = _state(e);
        int po = _output(e);
        float pw = _weight(e);
        for (FSTArcIter a = _in->arcs_out_with_input(p, 0); a.valid(); a++) {
            if (a->output && (po || _fsa)) {
                continue;
            }
            int q = a->next;
            int qo = po ? po : a->output;
            float qw = _semiring->times(pw, a->weight);
            float& w = (*set_)[_StateOutput(q, qo)]; // the set value is already initialized
            w = _semiring->plus(w, qw);
        }
    }

    return true;
}

void
FSTRemoveEpsilonsRep::_closure_idempotent_cyclic_bellman_ford(
    _Set* set_, const _StateOutput& state_output) const
{
    int N = set_->size();

    // Initialization.
    //
    for (_SetIter j = set_->begin(); j != set_->end(); j++) {
        (*j).second = _semiring->zero;
    }
    (*set_)[state_output] = _semiring->one;

    // Relaxation (N-1 passes) and negative cost loop check (1 pass), if needed.
    //
    for (int i = N; i > 0; i--) {
        bool relaxed = false;
        for (_SetIter j = set_->begin(); j != set_->end(); j++) {
            float pw = (*j).second;
            if (!_semiring->is_better(pw, _semiring->zero)) {
                continue;
            }
            const _StateOutput& pso = (*j).first;
            int p = pso.first;
            int po = pso.second;
            for (FSTArcIter a = _in->arcs_out_with_input(p, 0); a.valid(); a++) {
                if (a->output && (po || _fsa)) {
                    continue;
                }
                int q = a->next;
                int qo = po ? po : a->output;
                float qw = _semiring->times(pw, a->weight);
                float& stored = (*set_)[_StateOutput(q, qo)];
                float w = _semiring->plus(stored, qw);
                if (i > 1) { // relaxation pass
                    if (stored != w) {
                        stored = w;
                        relaxed = true;
                    }
                } else { // negative cost loop check, final pass
                    if (stored != w) {
                        FSTWarn("FSTRemoveEpsilons: "
                                "epsilon cycle with weight better than semiring->one detected");
                        return;
                    }
                }
            }
        }
        if (!relaxed) { // early exit as pass updated nothing
            break;
        }
    }
}

void
FSTRemoveEpsilonsRep::_closure_idempotent_cyclic_dijkstra(
    _Set* set_, const _StateOutput& state_output) const
{
    set_->clear();

    typedef Heap<const _QueueEntry, _QueueEntry::Better> QueueHeap;
    typedef fst::set<_QueueEntry, _QueueEntry::Index> QueueSet;

    _QueueEntry::Better better(_semiring);
    QueueHeap queue_heap(better);
    QueueSet queue_set;

    std::pair<QueueSet::iterator,bool> status =
        queue_set.insert(_QueueEntry(state_output.first, state_output.second, _semiring->one));
    queue_heap.insert(&*status.first); // insert pointer to entry contained in set

    while (queue_heap.size()) {
        const _QueueEntry& top = *queue_heap.top();
        queue_heap.pop();
        int p = top.state;
        int po = top.output;
        float pw = top.weight;
        for (FSTArcIter a = _in->arcs_out_with_input(p, 0); a.valid(); a++) {
            if (a->output && (po || _fsa)) {
                continue;
            }
            int q = a->next;
            int qo = po ? po : a->output;
            float qw = _semiring->times(pw, a->weight);
            status = queue_set.insert(_QueueEntry(q, qo, qw));
            const _QueueEntry& stored = *status.first;
            if (status.second) { // new
                queue_heap.insert(&stored);
            } else { // preexisting
                qw = _semiring->plus(qw, stored.weight);
                if (qw != stored.weight) {
                    stored.weight = qw;
                    if (stored.index <= 0) { // no longer in heap
                        // I don't think this should ever happen.  It would
                        // require a negative cost weight, but those are handled
                        // by the Bellman-Ford algorithm instead.
                        queue_heap.insert(&stored);
                    } else { // still in heap
                        queue_heap.decrease(&stored);
                    }
                }
            }
        }
    }

    // Copy results from queue_set to set_.
    //
    for (QueueSet::iterator i = queue_set.begin(); i != queue_set.end(); i++) {
        (*set_)[_StateOutput((*i).state, (*i).output)] = (*i).weight;
    }
}

void
FSTRemoveEpsilonsRep::_closure_idempotent(
    _Set* set_, const _StateOutput& state_output) const
{
    bool negative = false;
    if (!_closure_idempotent_acyclic(set_, state_output, negative)) {
        if (negative) {
            _closure_idempotent_cyclic_bellman_ford(set_, state_output);
        } else {
            _closure_idempotent_cyclic_dijkstra(set_, state_output);
        }
    }
}

void
FSTRemoveEpsilonsRep::_closure_non_idempotent(
    _Set* set_, const _StateOutput& state_output) const
{
    assert(_fsa);
    assert(state_output.second == 0);

    // 1. Topological sort (forward DFS) to produce state order and check for
    // more than self-loops.  Sum up any self-loop weights.  Zero weights for
    // states visited.

    fst::map<int, _ColorWeight> color_weight;
    fst::vector<int> order;

    _closure_non_idempotent_dfs(order, color_weight, -1, state_output.first, _semiring->zero);
    reverse(order.begin(), order.end());

    //
    // 2. Walk through states in topological order, pushing weights forward.

    fst::map<int, float> weight;

    for (fst::vector<int>::const_iterator i = order.begin(); i != order.end(); ++i) {
        weight[*i] = _semiring->zero;
    }
    weight[order[0]] = _semiring->one;

    for (fst::vector<int>::const_iterator i = order.begin(); i != order.end(); ++i) {
        int p = *i;
        float pw = weight[p] =
            _semiring->times(weight[p],
                             _semiring->closure(color_weight[p].self_loop_weight));
        set_->insert(_Entry(_StateOutput(p, 0), pw));
#ifdef DEBUG2
        cerr << ' ' << p << ' ' << color_weight[*i].self_loop_weight << ' ' << pw << endl;
#endif
        for (FSTArcIter a = _in->arcs_out_with_input(p, 0); a.valid(); ++a) {
            if (a->output) continue;
            int q = a->next;
            float& qw = weight[q];
            qw = _semiring->plus(qw, _semiring->times(pw, a->weight));
        }
    }
}

void
FSTRemoveEpsilonsRep::_closure_non_idempotent_dfs(
    fst::vector<int>& order, fst::map<int, _ColorWeight>& color_weight, int p, int q, float w) const
{
    _ColorWeight& qcw = color_weight[q];

    switch (qcw.color) {
    case _ColorWeight::White: // unvisited: mark active
        qcw.color = _ColorWeight::Gray;
        qcw.self_loop_weight = _semiring->zero;
        break;
    case _ColorWeight::Gray: // active: cycle detected, handle self-loops
        if (q == p) {
            qcw.self_loop_weight = _semiring->plus(qcw.self_loop_weight, w);
            return;
        } else {
            FSTFatal("non-self-loop epsilon cycle detected: not yet supported");
        }
        break;
    case _ColorWeight::Black: // already finished
        return;
        break;
    }
                                                   
    for (FSTArcIter a = _in->arcs_out_with_input(q, 0); a.valid(); ++a) {
        if (a->output) continue;
        _closure_non_idempotent_dfs(order, color_weight, q, a->next, a->weight);
    }

    qcw.color = _ColorWeight::Black;
    order.push_back(q);
}

void
FSTRemoveEpsilonsRep::_prune(_Set* set_) const
{
    // Now, prune set by removing all entries that will not produce any arcs.
    // We have to be careful advancing the iterator since we may be removing
    // entries during the traversal.  We must compute final_weight and potential
    // here before states are pruned.
    //
    float final_weight = _semiring->zero;
    float potential    = _semiring->zero;
    for (_Set::iterator i = set_->begin(); i != set_->end(); ) {
        _Set::iterator j = i;
        j++;

        int q = _state(*i);
        int o = _output(*i);
        float w = _weight(*i);

        if (!o) {
            final_weight = _semiring->plus(final_weight,
                                           _semiring->times(w,
                                                            _in->final_weight(q)));
        }
        potential = _semiring->better(potential,
                                      _semiring->times(w,
                                                       _in->potential(q)));

        bool prune = true;
        if (o && _in->is_final(q)) {
            prune = false;
        } else if (_in->arcs_out_with_input(q, 1, FSTGreaterEqual).n() > 0) {
            prune = false;
        } else if (o || _fsa) {
            // Look for any epsilon:output arcs.
            for (FSTArcIter a = _in->arcs_out_with_input(q, 0); a.valid(); a++) {
                if (a->output) {
                    prune = false;
                    break;
                }
            }
        }
        if (prune) set_->erase(i);
        i = j;
    }
#ifdef DEBUGX
    cerr << "after prune:" << endl << set_ << endl;
#endif

    if (!_semiring->is_better(final_weight, _semiring->zero) &&
        !_semiring->is_better(_semiring->zero, final_weight))
        final_weight = _semiring->zero;

    // Add two dummy entries, (FINAL_TAG, final_weight) and (POTENTIAL_TAG,
    // potential), to keep state from being merged with incompatible sets that
    // are otherwise the same.  This is also where we store the final weight and
    // potential for a state (i.e., in its set).
    //
    set_->insert(_Entry(_StateOutput(FINAL_TAG, 0), final_weight));
    set_->insert(_Entry(_StateOutput(POTENTIAL_TAG, 0), potential));
}

#if 0
// This is work in progress and completely untested...

// Return reference to the value t[so], setting the default value if it has
// never before been accessed.
//
float&
FSTRemoveEpsilonsRep::_value(_Map& t, const _StateOutput& so, float default_value)
{
    pair<_Map::iterator, bool> status = t->insert(_Entry(so, default_value));
    return (*status.first).second;
}

FSTRemoveEpsilonsRep::_Set*
FSTRemoveEpsilonsRep::_closure(const _StateOutput& state_output) const
{

    // Compute weighted epsilon closure: all (state, weight) pairs reachable
    // from _in state p.  We only keep at most one pair for each state, hence
    // the use of a state -> weight map (called _Set).  We compute the closure
    // using a best-first search along epsilon arcs.
    //
    // For the MinPlus semiring, this will not terminate if there is an epsilon
    // cycle with negative weight.  For the PlusTimes and PlusTimesLog semirings
    // this will not terminate if there is an epsilon cycle with non-zero
    // weight.  This cases could be detected with a depth-first search (and the
    // states might be topologically sortable), but this would cause complexity
    // and likely slow things down unnecessarily.
    //

    _Set* d_set = new _Set;
    _Map& d_table = *d_set;
    _Map r_table;

    // Special final state.
    //
    if (state_output.first < 0) {
        d_table[_StateOutput(FINAL_TAG, 0)] = _semiring->one;
        d_table[_StateOutput(POTENTIAL_TAG, 0)] = _semiring->one;
        return &d_table;
    }

    typedef priority_queue<_Entry> Queue;
    Queue queue;

    fst::set<_StateOutput> in_queue;

    // Put initial entry in set and queue.
    //
    _Entry e(state_output, _semiring->one);
    d_table.insert(e);
    queue.push(e);
    in_queue.insert(state_output);

    // Main loop of the best-first search.
    //
    while (!queue.empty()) {

#ifdef DEBUGX
        cerr << "pop: " << queue.top() << endl;
#endif

        // Pop top entry.
        //
        const _Entry& qe = queue.top();
        _StateOutput& q_so = qe.first;
        int qs   = _state(qe);
        int qo   = _output(qe);
        float qw = _weight(qe);
        float& r_q  = _value(r_table, q_so, _semiring->zero);
        float r = r_qe;
        r_qe = _semiring->zero;
        in_queue.erase(q_so);
        queue.pop();

        // REWRITE THIS: Push all epsilon-input extensions that are either not
        // already in the set or have a better score.  However, an extension is
        // not possible if there is both a pending output (qo) and an arc
        // output, or if we are in FSA mode and there is an arc output.
        //
        for (FSTArcIter a = _in->arcs_out_with_input(qs, 0); a.valid(); a++) {
            if (a->output && (qo || _fsa))
                continue;
            int ns = a->next;
            int no = qo ? qo : a->output;

            _StateOutput n_so(ns, no);
            float& d_n = _value(d_table, n_so, _semiring->zero);
            float r_times_aw = _semiring->times(r, a->weight);
            float d_n_new = _semiring->plus(d_n, r_times_aw);

            if (!_semiring->equal(d_n, d_n_new)) {
                d_n = d_n_new;
                float& r_n = _value(r_table, n_so, _semiring->zero);
                r_n = _semiring->plus(r_n, r_times_aw);
                if (!in_queue.count(n_so)) {
                    queue.push(_Entry(n_so, d_n));
                    in_queue.insert(n_so);
                }
            }

        }

    }

#ifdef DEBUGX
    cerr << "after search:" << endl << d_set << endl;
#endif

    // Now, prune set by removing all entries that will not produce any arcs.
    // We have to be careful advancing the iterator since we may be removing
    // entries during the traversal.  We must compute final_weight and potential
    // here before states are pruned.
    //
    float final_weight = _semiring->zero;
    float potential    = _semiring->zero;
    for (_Map::iterator i = d_table.begin(); i != d_table.end(); ) {
        _Map::iterator j = i;
        j++;

        int q = _state(*i);
        int o = _output(*i);
        float w = _weight(*i);

        if (!o) {
            final_weight = _semiring->plus(final_weight,
                                           _semiring->times(w,
                                                            _in->final_weight(q)));
        }
        potential = _semiring->better(potential,
                                      _semiring->times(w,
                                                       _in->potential(q)));

        bool prune = true;
        if (o && _in->is_final(q)) {
            prune = false;
        } else if (_in->arcs_out_with_input(q, 1, FSTGreaterEqual).n() > 0) {
            prune = false;
        } else if (o || _fsa) {
            // Look for any epsilon:output arcs.
            for (FSTArcIter a = _in->arcs_out_with_input(q, 0); a.valid(); a++) {
                if (a->output) {
                    prune = false;
                    break;
                }
            }
        }
        if (prune) d_table.erase(i);
        i = j;
    }
#ifdef DEBUGX
    cerr << "after prune:" << endl << d_set << endl;
#endif

    if (!_semiring->is_better(final_weight, _semiring->zero) &&
        !_semiring->is_better(_semiring->zero, final_weight))
        final_weight = _semiring->zero;

    // Add two dummy entries, (FINAL_TAG, final_weight) and (POTENTIAL_TAG,
    // potential), to keep state from being merged with incompatible sets that
    // are otherwise the same.  This is also where we store the final weight and
    // potential for a state (i.e., in its set).
    //
    d_table[_StateOutput(FINAL_TAG, 0)] = final_weight;
    d_table[_StateOutput(POTENTIAL_TAG, 0)] = potential;

    return &d_table;
}
#endif

void
FSTRemoveEpsilonsRep::state_set(int p, fst::set<int>& result) const
{
    FST_ASSERT_VALID_STATE(p);

    result.clear();
    const _Set& set_ = *_states[p]->set_;
    for (_Set::const_iterator si = set_.begin(); si != set_.end(); si++) {
        int q = _state(*si);
        if (q >= 0)
            result.insert(q);
    }
}

FSTGeneric
FSTRemoveEpsilons(const FSTGeneric& x, bool fsa, float delta)
{
    return FSTGeneric(new FSTRemoveEpsilonsRep(x, fsa, delta));
}

FSTGeneric
FSTRemoveEpsilonsA(const FSTGeneric& x, float delta)
{
    return FSTGeneric(new FSTRemoveEpsilonsRep(x, true, delta));
}

void
FSTRemoveEpsilonsStateSet(const FSTGeneric& x, int p, fst::set<int>& result)
{
    dynamic_cast<const FSTRemoveEpsilonsRep*>(x.pointer())->state_set(p, result);
}

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 80
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
