////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

// To Do:
//
// * Figure out if there are simple changes that would allow left recursive
// and/or right recursive grammars.  How is what we are doing different from
// what an Earley parser does?
//
// * Figure out if stack loop collapsing, as described in Pereira and Wright
// paper, helps with above left/right recursive grammars.  I expect it will do
// exactly that.

//#define FORCE_TRAIN
//#define DEBUG

#include <cstring>
#include "fst/FST_i.h"
#include "fst/format.h"

class FSTRtnRep
    : public FSTGenericInt {

public: // methods

    FSTRtnRep(istream& is, const char* header);
    FSTRtnRep(const FSTGeneric& root_fst,
              const fst::vector<string>& nonterminals,
              const fst::vector<FSTGeneric>& nonterminal_fsts,
              bool tagged);
    virtual ~FSTRtnRep();
    virtual const fst::set<int>& input_alphabet() const;
    virtual const fst::set<int>& output_alphabet() const;
    virtual bool is_final(int p) const;
    virtual float final_weight(int p) const;
    virtual float potential(int p) const;
    virtual FSTArcIter arcs_out(int p, FSTArcOrder order=FSTNone) const;
    virtual FSTArcIter arcs_out_with_input(int p, int i,
                                           FSTArcMode m=FSTEqual) const;
    virtual FSTArcIter arcs_out_with_output(int p, int o,
                                            FSTArcMode m=FSTEqual) const;
    virtual void arcs_out(fst::vector<FSTArc*>& arcs,
                          int p, FSTArcOrder order=FSTNone) const;
    virtual void force() const;
    virtual void reset() const;
    virtual void dump_ascii(ostream& os, bool header=true) const;
    virtual void dump_binary(ostream& os) const;

    void setFst(const string& name, const FSTGeneric& fst);

protected: // types

    struct _Stack;
    struct _StackItem;
    struct _State;
    struct _TrainData;

    struct _StackItem
        : FSTReferenceCounted {
        
        const FSTGeneric& net;
        int state;
        const _StackItem* next;
        
        _StackItem(const FSTGeneric& net_, int state_, const _StackItem* next_)
            : net(net_), state(state_), next(next_) {
            if (next) next->incref();
        }
        ~_StackItem() {
            if (next) next->decref();
        }
        bool operator<(const _StackItem& x) const {
            return &x != 0 &&
                (this == 0 || FST_LESS3(&net, &x.net,
                                        FST_LESS3(state, x.state,
                                                  FST_LESS2(*next, *x.next))));
        }
        FSTAllocator_new_delete(_StackItem);
    };
    
    struct _Stack
        : FSTReferenceCounter<const _StackItem> {
        
        typedef FSTReferenceCounter<const _StackItem> _RC;

        _Stack() {}
        _Stack(const _Stack& s) : _RC(s._rep) {}
        _Stack(const _StackItem* i) : _RC(i) {}
        ~_Stack() {}
        bool operator<(const _Stack& x) const {
            return *_rep < *x._rep;
        }                 
        const FSTGeneric& top_net() const {
            return _rep->net;
        }
        int top_state() const {
            return _rep->state;
        }
        bool empty() const {
            return _rep == 0;
        }
        _Stack pop() const {
            return _Stack(_rep->next);
        }
        _Stack push(const FSTGeneric& n, int s) const {
            return _Stack(new _StackItem(n, s, _rep));
        }
    };
    
    typedef fst::map<_Stack, float>
        _ContextSet;

    struct _State {
        float weight;
        _ContextSet* set_;
        _State(FSTSemiring s, _ContextSet* c) : weight(s->zero), set_(c) {}
        ~_State() {
            delete set_;
        }
        FSTAllocator_new_delete(_State);
    };

    struct _TrainData {
        FSTSemiring save_semiring;
        FSTPointerMap<void> ptr_map;
        fst::vector<bool> is_arc;
    };
    
    typedef fst::map<_ContextSet*, int, ptr_less<_ContextSet*> >
        _ContextSetIntMap;
    typedef fst::map<_Stack, int>
        _StackIntMap;
    typedef fst::vector<_State*>
        _StatePtrVector;
    typedef fst::map<string, int>
        _StringIntMap;
    typedef pair<int, int>
        _NetworkStatePair;
    typedef fst::map<_NetworkStatePair, int>
        _ColorMap;
    typedef pair<int, int>
        _TagPair;
    typedef fst::vector<_TagPair>
        _TagPairVector;

protected: // members

    fst::vector<FSTGeneric> _networks;
    fst::vector<string> _network_names;
    fst::vector<int> _network_labels;
    _TagPairVector _network_tags;
    mutable _StatePtrVector _states;
    mutable _ContextSetIntMap _set_state_map;
    mutable _StackIntMap _stack_state_map;
    fst::set<int> _input_alphabet;
    fst::set<int> _output_alphabet;
    bool _is_transducer;
    bool _is_tagged;
    _TrainData* _train_data;

protected: // methods

    void _arcs_out(int p, int input,
                   FSTArcMode m, FSTArcOrder order, fst::vector<FSTArc*>* arcs) const;
    int _find_or_add_state(const _Stack& stack_) const;
    inline void _add_arc(fst::vector<FSTArc*>* arcs, int from,
                         const _Stack& stack_, int i, int o, float w) const;
    _ContextSet* _closure(const _Stack& stack_) const;
    void _closure(_ContextSet& out, const _Stack& stack_, float weight) const;
    bool _insert(_ContextSet& set_, const _Stack& stack_, float weight) const;
    FSTGeneric _relabel(const FSTGeneric& in, const fst::vector<int>& map_);
    void _compute_tags();
    void _compute_alphabets();
    int  _init() const;
    void _clean() const;
    void _describe(ostream& os, const _Stack& x) const;
    void _describe(ostream& os, const _ContextSet& x) const;
    void _warn_left_recursion() const;
    void _warn_left_recursion_dfs(_ColorMap& color, _NetworkStatePair& here) const;

    static inline int _string_to_index(const _StringIntMap& map_,
                                       const char* s);

    int _train_index(const FSTArc* a) const;
    int _train_index(const _State* s) const;

    virtual void _train_begin();
    virtual void _train_end();
    virtual void _train_normalize();
    virtual void _train_update(fst::vector<double>& a_counts,
                               fst::vector<double>& f_counts,
                               double floor);
};

//static const float eps = 0.001;

FSTGeneric
FSTRtn(const FSTGeneric& root_fst,
       const fst::vector<string>& nonterminals,
       const fst::vector<FSTGeneric>& nonterminal_fsts,
       bool tagged)
{
    return FSTGeneric(new FSTRtnRep(root_fst, nonterminals,
                                    nonterminal_fsts, tagged));
}

FSTRtnRep::FSTRtnRep(const FSTGeneric& root_fst,
                     const fst::vector<string>& nonterminals,
                     const fst::vector<FSTGeneric>& nonterminal_fsts,
                     bool tagged)
    : _is_tagged(tagged),
      _train_data(0)
{
    unsigned i;

    _class_name = "FSTRtn";
    _semiring = root_fst->semiring();
    _network_names.push_back("");
    _network_names.insert(_network_names.end(),
                          nonterminals.begin(), nonterminals.end());

    assert(nonterminals.size() == nonterminal_fsts.size());

    _network_labels.reserve(_network_names.size());
    _network_labels.push_back(0);
    for (i = 0; i < nonterminals.size(); i++)
        _network_labels.push_back(FSTstring_map[nonterminals[i]]);
    fst::vector<int> label_map(FSTstring_map.size());
    for (i = 0; i < nonterminals.size(); i++)
        label_map[FSTstring_map[nonterminals[i]]] = -int(i) - 1;
    for (i = 0; i < label_map.size(); i++)
        if (label_map[i] == 0)
            label_map[i] = i;

    // Copy FSTs, relabeling nonterminals to negative indices using label_map.
    //
    _networks.resize(nonterminals.size() + 1);
    _networks[0] = _relabel(root_fst, label_map);
    for (i = 0; i < nonterminals.size(); i++) {
        _networks[i + 1] = _relabel(nonterminal_fsts[i], label_map);
    }

    _compute_tags();
    _compute_alphabets();
    _initial = _init();
    _warn_left_recursion();
}

FSTRtnRep::FSTRtnRep(istream& is, const char* header)
    : _is_tagged(false),
      _train_data(0)
{
    _class_name = "FSTRtn";

    _StringIntMap nonterminal_map;

    const char *h;
    char input[1024], output[1024];             // Yuck, hard-coded sizes.
    int p, q, n, new_states;
    int n_nonterminals = 0;
    float weight;
    float scale = 1;
    float offset = 0;
    int fst_n = -1;
    FSTGeneric fst;

    if (strstr(header, "PlusTimesLog"))
        _semiring = FSTPlusTimesLog;
    else if (strstr(header, "PlusTimes"))
        _semiring = FSTPlusTimes;
    else if (strstr(header, "MaxPlus"))
        _semiring = FSTMaxPlus;
    else // default is MinPlusSemiring
        _semiring = FSTMinPlus;

    h = strstr(header, "WeightScale=");
    if (h) {
        sscanf(h, "WeightScale=%f", &scale);
    }

    h = strstr(header, "WeightOffset=");
    if (h) {
        sscanf(h, "WeightOffset=%f", &offset);
    }

    _is_transducer = (strstr(header, "Transducer") != NULL);

    _is_tagged = (strstr(header, "Tagged") != NULL);

    _networks.push_back(FSTBasic(_semiring));
    _network_names.push_back("");
    _network_labels.push_back(0);

    string line;
    while (!getline(is, line).fail()) {
        const char* linec = line.c_str();
        switch (linec[0]) {
        case 'N':                               // nonterminal declaration
            n = sscanf(linec, "N %s", input);
            if (n < 1)
                FSTFatal("usage: N <nonterminal>");
            if (!nonterminal_map[input]) {
                nonterminal_map[input] = ++n_nonterminals;
                _networks.push_back(FSTBasic(_semiring));
                _network_names.push_back(input);
                _network_labels.push_back(FSTstring_map[input]);
            }
            break;
        case 'S':                               // start of subnetwork
            if (fst_n >= 0) {
                _networks[fst_n] = fst;
            }
            fst = FSTBasic(_semiring);
            n = sscanf(linec, "S %s", input);
            if (n < 1 || input[0] == 0) {
                fst_n = 0;
            } else {
                n = nonterminal_map[input];
                if (!n)
                    FSTFatal("S %s before N %s", input, input);
                fst_n = n;
            }
            break;
        case 'F':
            n = sscanf(linec, "F %d %f", &p, &weight);
            if (n < 1)
                FSTFatal("usage: F <state>");
            if (n < 2)
                weight = _semiring->one;
            else if (weight != _semiring->zero)
                weight = scale*weight + offset;
        new_states = p - fst->n_states() + 1;
        if (new_states > 0)
        fst->add_states(new_states);
        fst->set_final(p, weight);
            break;
        case 'I':
            n = sscanf(linec, "I %d", &p);
            if (n != 1)
                FSTFatal("file: usage: I <state>");
        new_states = p - fst->n_states() + 1;
        if (new_states > 0)
        fst->add_states(new_states);
        if (fst->initial() >= 0 && p != fst->initial())
        FSTWarn("file: changing initial state from %d to %d "
             "(only one allowed)", fst->initial(), p);
        fst->set_initial(p);
            break;
        case 'P':
            n = sscanf(linec, "P %d %f", &p, &weight);
            if (weight != _semiring->zero)
                weight = scale*weight + offset;
            if (n < 2)
                FSTFatal("usage: P <state>");
        new_states = p - fst->n_states() + 1;
        if (new_states > 0)
        fst->add_states(new_states);
        fst->set_potential(p, weight);
            break;
        case 'T':
            n = sscanf(linec, "T %d %d %s %s %f",
                       &p, &q, input, output, &weight);
            if (n < 2)
                FSTFatal("file: usage: "
                      "T <from> <to> [<input> [<output> [<weight>]]]");
            if (n < 5)
                weight = _semiring->one;
            if (weight != _semiring->zero)
                weight = scale*weight + offset;
            if (n < 4 || strcmp(output, ",") == 0)
                output[0] = 0;
            if (n < 3 || strcmp(input, ",") == 0)
                input[0] = 0;
            if (output[0]) {
                if (nonterminal_map.count(input)) {
                    FSTWarn("ignoring output `%s' on nonterminal transition "
                            "`%s' (not transducer)",
                            output, input);
                    output[0] = 0;
                } else if (!_is_transducer && strcmp(input, output) != 0) {
                    FSTWarn("setting output `%s' to `%s' (not transducer)",
                            output, input);
                    strcpy(output, input);
                }
            } else if (!_is_transducer && !nonterminal_map.count(input)) {
                strcpy(output, input);
            }
        new_states = _max(p, q) - fst->n_states() + 1;
        if (new_states > 0)
        fst->add_states(new_states);
        fst->add_arc(p, q,
                        _string_to_index(nonterminal_map, input),
                        FSTstring_map[output], weight);
            break;
        }
        
    }
    if (fst_n >= 0) {
        _networks[fst_n] = fst;
    }

    _warn_left_recursion();
    _compute_tags();
    _compute_alphabets();
    _initial = _init();
#ifdef FORCE_TRAIN
    _train_begin();
#endif
}

FSTRtnRep::~FSTRtnRep()
{
    _clean();
}

void
FSTRtnRep::_compute_tags()
{
    _network_tags.reserve(_networks.size());
    _network_tags.push_back(_TagPair(0, 0));
    for (unsigned i = 1; i < _networks.size(); i++) {
        string name = _network_names[i];
        string begin_tag, end_tag;
        if (name[0] == '<' && name[name.size() - 1] == '>') {
            begin_tag = name;
            end_tag = name;
            end_tag.insert(1, "/");
        } else {
            begin_tag = '<' + name + '>';
            end_tag = "</" + name + '>';
        }
        _network_tags.push_back(_TagPair(FSTstring_map[begin_tag],
                                         FSTstring_map[end_tag]));
    }
}

void
FSTRtnRep::_compute_alphabets()
{
    // Compute input and output alphabets, which are the union of terminals in
    // all networks.  Note that nonterminals are negative, and the lower_bound
    // skips over them.
    //
    for (unsigned j = 0; j < _networks.size(); j++) {
        const fst::set<int>& i = _networks[j]->input_alphabet();
        fst::set_insert(_input_alphabet,  i.lower_bound(0), i.end());
        const fst::set<int>& o = _networks[j]->output_alphabet();
        fst::set_insert(_output_alphabet, o.lower_bound(0), o.end());
    }
}

int
FSTRtnRep::_init() const
{
    // Create initial state with initial context set.  Note that closure is
    // computed by _find_or_add_state.
    //
    _Stack stack_ = _Stack().push(_networks[0], _networks[0]->initial());
    return _find_or_add_state(stack_);
}

void
FSTRtnRep::_clean() const
{
    for (_StatePtrVector::iterator i = _states.begin();
         i != _states.end(); i++) {
        delete *i;
    }
    _states.clear();
    _set_state_map.clear();
    _stack_state_map.clear();
    _n_states = 0;
    _n_arcs = 0;
}

void
FSTRtnRep::reset() const
{
    _clean();

    int new_initial = _init();
    assert(new_initial == _initial);
}

const fst::set<int>&
FSTRtnRep::input_alphabet() const
{
    return _input_alphabet;
}

const fst::set<int>&
FSTRtnRep::output_alphabet() const
{
    return _output_alphabet;
}

bool
FSTRtnRep::is_final(int p) const
{
    FST_ASSERT_VALID_STATE(p);
    return _states[p]->weight != _semiring->zero;
}

float
FSTRtnRep::final_weight(int p) const
{
    FST_ASSERT_VALID_STATE(p);
    return _states[p]->weight;
}

float
FSTRtnRep::potential(int p) const
{
    FST_ASSERT_VALID_STATE(p);
    return _semiring->one;
}

inline void
FSTRtnRep::_add_arc(fst::vector<FSTArc*>* arcs, int from, const _Stack& stack_,
                    int i, int o, float w) const
{
    int state = _find_or_add_state(stack_);
    if (arcs)
        arcs->push_back(new FSTArc(from, state, i, o, w));
}

FSTArcIter
FSTRtnRep::arcs_out(int p, FSTArcOrder order) const
{
    FST_ASSERT_VALID_STATE(p);
    FSTArcIterData* arc_data = new FSTArcIterData;
    _arcs_out(p, 0, FSTGreaterEqual, order, &arc_data->arcs);
    return FSTArcIter(arc_data);
}

FSTArcIter
FSTRtnRep::arcs_out_with_input(int p, int i, FSTArcMode m) const
{
    FST_ASSERT_VALID_STATE(p);
    FSTArcIterData* arc_data = new FSTArcIterData;
    _arcs_out(p, i, m, FSTInput, &arc_data->arcs);
    return FSTArcIter(arc_data);
}

FSTArcIter
FSTRtnRep::arcs_out_with_output(int p, int o, FSTArcMode m) const
{
    FSTFatal("FSTRtnRep::arcs_out_with_output not implemented");
    return arcs_out_with_input(p, o, m);
}

void
FSTRtnRep::arcs_out(fst::vector<FSTArc*>& result, int p, FSTArcOrder order) const
{
    FST_ASSERT_VALID_STATE(p);
    _arcs_out(p, 0, FSTGreaterEqual, order, &result);
}

int
FSTRtnRep::_train_index(const FSTArc* a) const
{
    int i = _train_data->ptr_map[const_cast<FSTArc*>(a)];
    fst::vector<bool>& is_arc = _train_data->is_arc;
    assert(i <= int(is_arc.size()));
    if (i == int(is_arc.size()))
        is_arc.push_back(true);
    return i;
}

int
FSTRtnRep::_train_index(const _State* s) const
{
    int i = _train_data->ptr_map[const_cast<_State*>(s)];
    fst::vector<bool>& is_arc = _train_data->is_arc;
    assert(i <= int(is_arc.size()));
    if (i == int(is_arc.size()))
        is_arc.push_back(false);
    return i;
}

void
FSTRtnRep::_arcs_out(int state, int input, FSTArcMode mode,
                     FSTArcOrder order, fst::vector<FSTArc*>* arcs) const
{
    assert(order != FSTOutput);

    // Make sure that if we see FSTLessEqual that we do not return
    // nonterminals, which are negative.

    _State* sp = _states[state];
    const _ContextSet& set_ = *sp->set_;

    bool add_tags = _is_tagged || _train_data;

    // Iterate through contexts in set, generating requested arcs.
    //
    for (_ContextSet::const_iterator i = set_.begin(); i != set_.end(); i++) {
        const _Stack& stack_ = (*i).first;
        float weight = (*i).second;

        if (stack_.empty())
            continue;

        const FSTGeneric& top_net = stack_.top_net();
        int top_state = stack_.top_state();
        float final_weight = top_net->final_weight(top_state);

        // If appropriate, output an epsilon-input arc to close a tagged
        // nonterminal expansion.
        //
        if (add_tags
            && (input == 0 || mode == FSTLessEqual)
            && top_net != _networks[0]
            && final_weight != _semiring->zero) {
            int output;
            if (_train_data) {
#ifdef FORCE_TRAIN
                output = 0;
#else
                output = _train_index(sp);
#endif
            } else {
                int net_index = &top_net - &_networks[0];
                output = _network_tags[net_index].second;
            }
            float new_weight = _semiring->times(weight, final_weight);
            _Stack new_stack = stack_.pop();
            _add_arc(arcs, state, new_stack, 0, output, new_weight);
        }

        // Generate arcs.
        //
        if (!add_tags) {

            // Normal case when not tagging or training.
            //
            FSTArcIter a = top_net->arcs_out_with_input(top_state, input, mode);
            for (; a.valid(); a++) {
                if (a->input >= 0) {
                    _Stack new_stack =
                        stack_.pop().
                        push(stack_.top_net(), a->next);
                    float new_weight = _semiring->times(weight, a->weight);
                    _add_arc(arcs, state, new_stack,
                             a->input, a->output, new_weight);
                }
            }

        } else {

            // Tagging or training.  We will generate arcs for nonterminal
            // transitions too, but we have to be careful setting the arc
            // iterator because nonterminal arcs have negative inputs.
            //
            FSTArcIter a;
            if (input != 0 || mode == FSTLessEqual)
                a = top_net->arcs_out_with_input(top_state, input, mode);
            else if (mode == FSTGreaterEqual) // && input == 0
                a = top_net->arcs_out(top_state, order);
            else // mode == FSTEqual && input == 0
                a = top_net->arcs_out_with_input(top_state, 0, FSTLessEqual);
            for (; a.valid(); a++) {
                float new_weight = _semiring->times(weight, a->weight);
                if (a->input >= 0) {            // normal arc
                    _Stack new_stack =
                        stack_.pop().
                        push(stack_.top_net(), a->next);
                    int output;
                    if (_train_data) {
#ifdef FORCE_TRAIN
                        output = 0;
#else
                        output = _train_index(*a);
#endif
                    } else {
                        output = a->output;
                    }
                    _add_arc(arcs, state, new_stack,
                             a->input, output, new_weight);
                } else {                        // nonterminal arc
                    const FSTGeneric& nt = _networks[-a->input];
                    _Stack new_stack =
                        stack_.pop().
                        push(stack_.top_net(), a->next).
                        push(nt, nt->initial());
                    int output;
                    if (_train_data) {
#ifdef FORCE_TRAIN
                        output = 0;
#else
                        output = _train_index(*a);
#endif
                    } else {
                        int net_index = &new_stack.top_net() - &_networks[0];
                        output = _network_tags[net_index].first;
                    }
                    _add_arc(arcs, state, new_stack, 0, output, new_weight);
                }
            }
        }

    }

    // Sort arcs if necessary.
    //
    if (order != FSTNone && set_.size() > 1 && arcs->size() > 1)
        sort(arcs->begin(), arcs->end(), FSTArcCompareInput());
}

void
FSTRtnRep::force() const
{
    // Expand all states.  We do this by asking for all arcs out of all states
    // (without actually creating all the arcs).
    //
    for (int p = 0; p < _n_states; p++)
        _arcs_out(p, 0, FSTGreaterEqual, FSTNone, 0);
}

void
FSTRtnRep::dump_ascii(ostream& os, bool header) const
{
    unsigned i;

    if (_train_data)
        FSTFatal("FSTRtnRep::dump_ascii: can't dump when in training mode");

    if (header) {
        os << "#FSTRtn " << _semiring->name;
        if (_is_transducer)
            os << " Transducer";
        if (_is_tagged)
            os << " Tagged";
        os << '\n';
    }

    for (i = 1; i < _network_names.size(); i++)
        os << "N " << _network_names[i] << '\n';

    for (i = 0; i < _networks.size(); i++) {

        const FSTGeneric& network = _networks[i];

        if (i == 0)
            os << "S\n";
        else
            os << "S " << _network_names[i] << '\n';
        os << "I " << network->initial() << '\n';
        for (FSTStateIter s = network->states(); s.valid(); s++) {
            int p = *s;
            if (network->is_final(p)) {
                float weight = network->final_weight(p);
                os << "F " << p;
                if (weight != _semiring->one)
                    format(os, " %g", weight);
                os << '\n';
            }
            for (FSTArcIter a = network->arcs_out(p); a.valid(); a++) {
                int q = a->next;
                int input = a->input;
                int output = a->output;
                float weight = a->weight;
                // format: T <from> <to> [<input> [<output> [<weight>]]]
                os << "T " << p << ' ' << q;
                if (input || output || weight != _semiring->one){
                    if (input < 0)
                        os << ' ' << _network_names[-input];
                    else
                        os << ' ' << ((input == 0) ? "," : FSTstring_map[input]);
                }
                if (output || weight != _semiring->one)
                    os << ' ' << ((output == 0) ? "," : FSTstring_map[output]);
                if (weight != _semiring->one)
                    format(os, " %g", weight);
                os << '\n';
            }
        }
    }
}

void
FSTRtnRep::dump_binary(ostream& os) const
{
    if (_train_data)
        FSTFatal("FSTRtnRep::dump_binary: can't dump when in training mode");
    FSTGenericInt::dump_binary(os);
}

int
FSTRtnRep::_find_or_add_state(const _Stack& stack_) const
{
    // Check to see if we have already mapped this stack to a state.
    //
    _StackIntMap::iterator ci = _stack_state_map.find(stack_);
    if (ci != _stack_state_map.end())
        return (*ci).second;

    // Compute closure of stack, creating a _ContextSet*.
    //
    _ContextSet* set_ = _closure(stack_);

    int p;
    _ContextSetIntMap::iterator i = _set_state_map.find(set_);

    if (i == _set_state_map.end()) {

        // State not found: build new state.
        //
        _State* pp = new _State(_semiring, set_);
        _states.push_back(pp);
        p = _n_states++;
        _set_state_map[set_] = p;
        _stack_state_map[stack_] = p;

        // Determine finality.  Final states will have a set element with an
        // empty stack, and this element will be first.
        //
        if (set_->size() > 0) {
            _ContextSet::const_iterator j = set_->begin();
            if ((*j).first.empty()) {
                pp->weight = (*j).second;
            }
#ifdef DEBUG
            format(cerr, "new state %d%s: ", p, (*j).first.empty() ? "*" : "");
            _describe(cerr, *set_);
            cerr << '\n';
#endif        
        } else {
            pp->weight = _semiring->zero;
#ifdef DEBUG
            format(cerr, "new state %d: {}\n", p);
#endif
        }

    } else {

        // State found: look up state index and free set (it isn't needed).
        p = (*i).second;
        delete set_;

    }

    return p;
}

bool 
FSTRtnRep::_insert(_ContextSet& set_, const _Stack& stack_, float weight) const
{
    // Possibly insert (stack, weight) into set, returning true if the
    // insertion takes place.  It will take place if stack wasn't in set or if
    // weight is better than its previous value.
    //
    pair<_ContextSet::iterator, bool> result =
        set_.insert(_ContextSet::value_type(stack_, weight));
    if (!result.second &&
        _semiring->is_better(weight, (*result.first).second)) {
        result.second = true;
        (*result.first).second = weight;
    }
    return result.second;
}

void
FSTRtnRep::_closure(_ContextSet& out, const _Stack& stack_, float weight) const
{
    const FSTGeneric& top_net = stack_.top_net();
    int top_state = stack_.top_state();

    // Recursively explore context, possibly adding context to set as we go.
    //

    if (!_is_tagged && !_train_data) {

        // If top element has nonterminals leaving, add new context by pushing.
        // This is a parser "shift".  Note that all nonterminals are negative.
        //
        FSTArcIter a = top_net->arcs_out_with_input(top_state, -1, FSTLessEqual);
        for (; a.valid(); a++) {
            
            // Create new context, changing top element to destination of
            // nonterminal and pushing a new element representing the initial
            // state of the nonterminal network.
            //
            const FSTGeneric& nt_net = _networks[-a->input];
            float new_weight = _semiring->times(weight, a->weight);
            _Stack new_stack =
                stack_.pop().
                push(top_net, a->next).
                push(nt_net, nt_net->initial());
            if (_insert(out, new_stack, new_weight)) {
#ifdef DEBUG
                cerr << "shifting [" << new_weight;
                _describe(cerr, new_stack);
                cerr << "]\n";
#endif
                _closure(out, new_stack, new_weight);
            }
        }
    }

    if ((!_is_tagged && !_train_data) || top_net == _networks[0]) {

        // If top element is a final state, add new context by popping.  This
        // is a parser "reduce".
        //
        float top_final_weight =
            top_net->final_weight(top_state);
        if (top_final_weight != _semiring->zero) {
            float new_weight = _semiring->times(weight, top_final_weight);
            _Stack new_stack = stack_.pop();
            if (_insert(out, new_stack, new_weight)) {
#ifdef DEBUG
                cerr << "reducing [" << new_weight;
                _describe(cerr, new_stack);
                cerr << "]\n";
#endif
                if (!new_stack.empty())
                    _closure(out, new_stack, new_weight);
            }
        }
    }

}

FSTRtnRep::_ContextSet*
FSTRtnRep::_closure(const _Stack& stack_) const
{
    float weight = _semiring->one;

    // Create result context set, with (stack, weight) as initial member.
    //
    _ContextSet* pout = new _ContextSet;
    _ContextSet& out = *pout;
    out[stack_] = weight;

#ifdef XDEBUG
    cerr << "closure input: ";
    _describe(cerr, out);
    cerr << "\n";
#endif
    
    // Recursively explore contexts in set, starting with initial member
    // (stack, weight).
    //
    _closure(out, stack_, weight);

    // As an optimization, remove all contexts (stacks) in result set that
    // cannot generate arcs.  Note that we are careful about advancing iterator
    // j before erasing an element, otherwise we couldn't advance.
    //
    // We do not perform this optimization when training because we need the
    // full set of states associated with a final state.
    //
    if (!_train_data) {
        for (_ContextSet::iterator j = out.begin(); j != out.end();) {
            const _Stack& s = (*j).first;
            if (!s.empty()) {
                
                const FSTGeneric& top_net = s.top_net();
                int top_state = s.top_state();
                
                bool any_arcs;
                if (_is_tagged)
                    any_arcs = top_net->arcs_out(top_state).valid()
                        || (top_net != _networks[0]
                            && top_net->is_final(top_state));
                else
                    any_arcs =
                        top_net->arcs_out_with_input(top_state, 0,
                                                    FSTGreaterEqual).valid();
                if (!any_arcs) {
                    _ContextSet::iterator k = j;
#ifdef DEBUG
                    cerr << "removing [" << (*j).second;
                    _describe(cerr, stack_);
                    cerr << "]\n";
#endif
                    j++;
                    out.erase(k);
                } else {
                    j++;
                }
            } else {
                j++;
            }
        }
    }

#ifdef XDEBUG
    cerr << "closure output: ";
    _describe(cerr, out);
    cerr << "\n";
#endif

    return pout;
}

FSTGeneric
FSTRtnRep::_relabel(const FSTGeneric& in, const fst::vector<int>& map_)
{
    FSTGeneric out = FSTBasic(in->semiring());
    out->add_states(in->n_states());
    for (FSTStateIter i = in->states(); i.valid(); i++) {
        int p = *i;
        out->set_final(p, in->final_weight(p));
        FSTArcIter j = in->arcs_out(p);
        int new_states = in->n_states() - out->n_states();
        if (new_states > 0)
            out->add_states(new_states);
        for (; j.valid(); j++) {
            if (j->input != j->output)
                _is_transducer = true;
            int input = map_[j->input];
            out->add_arc(p, j->next, input,
                        input < 0 ? 0 : j->output, j->weight);
        }
    }
    out->set_initial(in->initial());
    return out;
}

inline int
FSTRtnRep::_string_to_index(const _StringIntMap& map_, const char* s)
{
    _StringIntMap::const_iterator i = map_.find(s);
    return (i != map_.end()) ? -(*i).second : FSTstring_map[s];
}

void
FSTRtnRep::_describe(ostream& os, const _Stack& x) const
{
    _Stack y = x;
    for (; !y.empty(); y = y.pop()) 
        format(os, ",(%s,%d)",
                _network_names[&y.top_net() - &_networks[0]].c_str(),
                y.top_state());
}

void
FSTRtnRep::_describe(ostream& os, const _ContextSet& x) const
{
    os << "{";
    for (_ContextSet::const_iterator i = x.begin(); i != x.end(); i++) {
        if (i != x.begin())
            os << ", ";
        format(os, "[%g", (*i).second);
        _describe(os, (*i).first);
        os << "]";
    }
    os << "}";
}

#define COLOR_WHITE 0
#define COLOR_GRAY  1
#define COLOR_BLACK 2

void
FSTRtnRep::_warn_left_recursion() const
{
    int N = _networks.size();
    _ColorMap color;

    for (int n = 1; n < N; n++) {
        _NetworkStatePair where(n, _networks[n]->initial());
        if (color[where] == COLOR_WHITE)
            _warn_left_recursion_dfs(color, where);
    }
}

void
FSTRtnRep::_warn_left_recursion_dfs(_ColorMap& color, _NetworkStatePair& here) const
{
    color[here] = COLOR_GRAY;
    for (FSTArcIter a = _networks[here.first]->arcs_out(here.second);
         a.valid(); a++) {
        if (a->input < 0) {
            _NetworkStatePair there(-a->input, _networks[-a->input]->initial());
            if (color[there] == COLOR_GRAY)
                FSTWarn("left recursion involving %s and %s",
                        _network_names[here.first].c_str(),
                        _network_names[there.first].c_str());
            else if (color[there] == COLOR_WHITE)
                _warn_left_recursion_dfs(color, there);
#if 0
        } else if (a->input == 0) {
            _NetworkStatePair there(here.first, a->next);
            if (color[there] == COLOR_WHITE)
                _warn_left_recursion_dfs(color, there);
#endif
        }
    }
    color[here] = COLOR_BLACK;
}

void
FSTRtnRep::_train_begin()
{
    _train_data = new _TrainData;

    // Convert all weights to PlusTimes semiring.
    //
    for (int i = 0; i < int(_networks.size()); i++) {
        FSTGeneric& fst = _networks[i];
        for (int p = 0; p < fst->n_states(); p++) {
            fst->set_final(p, _semiring->convert(fst->final_weight(p)));
            for (FSTArcIter a = fst->arcs_out(p); a.valid(); a++) {
                FSTArc* ap = const_cast<FSTArc*>(*a);
                ap->weight = _semiring->convert(ap->weight);
            }
        }
    }
    _train_data->save_semiring = _semiring;
    _semiring = FSTPlusTimes;

    // We reset after changing semirings so that we don't have stale state
    // final weights.
    //
    reset();
}

void
FSTRtnRep::_train_end()
{
    // Convert all weights to original semiring.
    //
    _semiring = _train_data->save_semiring;
    for (int i = 0; i < int(_networks.size()); i++) {
        FSTGeneric& fst = _networks[i];
        for (int p = 0; p < fst->n_states(); p++) {
            fst->set_final(p, _semiring->unconvert(fst->final_weight(p)));
            for (FSTArcIter a = fst->arcs_out(p); a.valid(); a++) {
                FSTArc* ap = const_cast<FSTArc*>(*a);
                ap->weight = _semiring->unconvert(ap->weight);
            }
        }
    }

    delete _train_data;
    _train_data = 0;

    // We reset after changing semirings so that we don't have stale state
    // final weights.
    //
    reset();
}

void
FSTRtnRep::_train_normalize()
{
    for (unsigned i = 0; i < _networks.size(); i++) {
        FSTGeneric& fst = _networks[i];
        for (int p = 0; p < fst->n_states(); p++) {
            double final_weight = fst->final_weight(p);
            double sum = final_weight;
            for (FSTArcIter a = fst->arcs_out(p); a.valid(); a++) {
                sum += a->weight;
            }
            if (sum == 0)
                continue;
            double factor = 1 / sum;
            fst->set_final(p, final_weight * factor);
            for (FSTArcIter a = fst->arcs_out(p); a.valid(); a++) {
                FSTArc* ap = const_cast<FSTArc*>(*a);
                ap->weight *= factor;
            }
        }
    }
}

inline void
_increment_final_weight(FSTGeneric& fst, int p, double w)
{
    fst->set_final(p, fst->final_weight(p) + w);
}

void
FSTRtnRep::_train_update(fst::vector<double>& a_counts,
                         fst::vector<double>& f_counts,
                         double floor)
{
    // Set all arc weights and all final state weights to floor.  Note that we
    // do this here rather than setting the a_counts and f_counts in case the
    // counts do not cover the whole FST.
    //
    for (int i = 0; i < int(_networks.size()); i++) {
        FSTGeneric& fst = _networks[i];
        for (int p = 0; p < fst->n_states(); p++) {
            fst->set_final(p, (fst->final_weight(p) != 0) ? floor : 0);
            for (FSTArcIter a = fst->arcs_out(p); a.valid(); a++) {
                FSTArc* ap = const_cast<FSTArc*>(*a);
                ap->weight = floor;
            }
        }
    }

    // Copy final counts to state final weights.
    //
    for (unsigned p = 0; p < f_counts.size(); p++) {
        if (f_counts[p] == 0)
            continue;
        _ContextSet* cs = _states[p]->set_;
        assert(cs->size() == 2);
        const _Stack& s = (*cs->rbegin()).first;
        _increment_final_weight(const_cast<FSTGeneric&>(s.top_net()),
                                s.top_state(), f_counts[p]);
    }

    // Copy arc counts to arc and final weights.  Note that the weight of
    // nonterminal closing arcs is due to a nonterminal final weight.
    //
    for (unsigned a = 0; a < a_counts.size(); a++) {
        if (_train_data->is_arc[a]) {
            FSTArc* ap = reinterpret_cast<FSTArc*>(_train_data->ptr_map[a]);
            ap->weight += a_counts[a];
        } else {
            _State* sp = reinterpret_cast<_State*>(_train_data->ptr_map[a]);
            _ContextSet* cs = sp->set_;
            assert(cs->size() == 1);
            const _Stack& s = (*cs->begin()).first;
            _increment_final_weight(const_cast<FSTGeneric&>(s.top_net()),
                                    s.top_state(), a_counts[a]);
        }
    }

    // Normalize.
    //
    _train_normalize();

    // Copy state final weights to f_counts.
    //
    fill(f_counts.begin(), f_counts.end(), 0);
    for (unsigned p = 0; p < f_counts.size(); p++) {
        _ContextSet* cs = _states[p]->set_;
        if (cs->size() == 2) {
            const _Stack& s = (*cs->rbegin()).first;
            f_counts[p] += s.top_net()->final_weight(s.top_state());
        }
    }

    // Copy arc and final weights to a_counts.
    //
    fill(a_counts.begin(), a_counts.end(), 0);
    for (unsigned a = 0; a < a_counts.size(); a++) {
        if (_train_data->is_arc[a]) {
            FSTArc* ap = reinterpret_cast<FSTArc*>(_train_data->ptr_map[a]);
            a_counts[a] += ap->weight;
        } else {
            _State* sp = reinterpret_cast<_State*>(_train_data->ptr_map[a]);
            _ContextSet* cs = sp->set_;
            assert(cs->size() == 1);
            const _Stack& s = (*cs->begin()).first;
            a_counts[a] += s.top_net()->final_weight(s.top_state());
        }
    }
}

void
FSTRtnRep::setFst(const string& name, const FSTGeneric& fst)
{
    for (int i = 0; i < int(_network_names.size()); i++) {
        if (_network_names[i] == name) {
            _networks[i] = fst;
            reset();
        }
    }
}

FSTGeneric
FSTLoadRtn(istream& is, const char* header)
{
    return new FSTRtnRep(is, header);
}

void
FSTRtnSet(FSTGeneric& rtn_fst, const string& name, const FSTGeneric& fst)
{
    FSTRtnRep* rtn = dynamic_cast<FSTRtnRep*>(rtn_fst.pointer());
    if (rtn) {
        rtn->setFst(name, fst);
    }
}

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 79
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
