////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

// Code supporting context-dependent rules of the form
//
//   phi REPLACE psi / lambda ___ rho
//
// where phi is the input, psi the output, lambda the left context, and rho the
// right context, with all being regular expressions.  The following REPLACE
// operators have been implemented:
//
//   @->  mandatory, left-to-right, directed (maximal match)
//   ->@  mandatory, right-to-left, directed (maximal match)
//   *->  mandatory, left-to-right
//   ->*  mandatory, right-to-left
//   ->   mandatory, simultaneous
//   ?->  optional, left-to-right
//   ->?  optional, right-to-left
//   ?->? optional, simultaneous
//
// NOTE: Parallel rules are not correctly implemented yet.  I believe that most
// rules in parallel become optional.  We should either fix this behavior or
// remove the parallel capability entirely.  Also, ignore isn't implemented for
// parallel rules.
//
// The algorithm for mandatory left-to-right is from Mohri and Sproat.  The
// algorithms for the rest are in the same spirit, but the specifics are by Lee
// Hetherington.  The following were very useful references during development:
//
// M. Mohri and R. Sproat, "An efficient compiler for weighted rewrite rules,"
// Proc. ACL-96?, 1996.
//
// R. Kaplan and M. Kay, "Regular models of phonological rule systems,"
// Computational Linguistics, 20(3), pp. 331-378, 1994.
//
// L. Karttunen, "The replace operator," in Finite State Language Processing,
// ed. by E. Roche and Y. Schabes, MIT Press (Cambridge, CA), pp. 117-147,
// 1997.
//
// L. Karttunen, "Directed replacement," Proc. 34th Annual Meeting of the
// Association for Computational Linguistics (Santa Cruz, CA), pp. 108-115,
// 1996.
//

//#define SIGMA_DEBUG

#include "fst/FST_i.h"
#include "fst/FSTRules.h"
#include <iostream>

//
// Auxiliary symbols.
//
#define AUXILIARY_LEFT1   "AUX<1"
#define AUXILIARY_LEFT2   "AUX<2"
#define AUXILIARY_RIGHT1  "AUX>1"
#define AUXILIARY_RIGHT2  "AUX>2"
#define AUXILIARY_EPSILON "AUX<>"

//
// Convenient typedefs.
//
typedef FSTArcIter ArcIter;
typedef Set::const_iterator SetIter;

//
// Forward declarations.
//
static bool _lookup(Fst& result, FSTRules* context, int l);
static void _warn_missing(FSTRules* context, SetRef set, int line);
static void _warn_missing(FSTRules* context, SetRef sigma2, SetRef set,
                          int line);
static Fst _automaton(SetRef sigma);
static Fst _automaton(int input);
static Fst _complement(SetRef sigma, FstRef x);
static Fst _compose(FstRef x, FstRef y);
static Fst _determinize(FstRef x);
static Fst _empty();
static Fst _i_loops_constrained(FstRef x, SetRef sigma, SetRef ignore);
static Fst _minimize(FstRef x);
static Fst _optional(FstRef x);
static Fst _star(SetRef sigma, int label1=0, int label2=0, int label3=0);
static Fst _transducer(int input, int output);
static Fst _trim(FstRef x);
static Fst _union(FstRef x, FstRef y);


#ifdef UNUSED
static Fst _all_final(FstRef x);
static Fst _build_f(SetRef sigma, FstRef phi, int left1, int left2, int right1);
static Fst _build_l1(FstRef alpha, int left1, bool keep);
static Fst _build_l2(FstRef alpha, int left2);
static Fst _build_r1(SetRef sigma, FstRef rho, int right1);
static Fst _build_r2(SetRef sigma, FstRef rho, int right2);
static Fst _build_m(SetRef sigma, FstRef phi,
                    int left1, int left2, int right1, int right2);
static Fst _concat(FstRef x, FstRef y);
static Fst _concat(FstRef x, FstRef y, FstRef z);
static Fst _cross(FstRef x, FstRef y);
static Fst _i_loops(FstRef x, int l1, int l2=0, int l3=0);
static Fst _introduce(SetRef sigma, int label, int extra_sigma=0);
static Fst _mark_after_first(SetRef sigma, FstRef x, int mark);
static Fst _marker1(FstRef sigma, int insert1, int insert2=0);
static Fst _marker2(FstRef sigma, int input, int output=0);
static Fst _marker3(FstRef sigma, int remove);
static Fst _partial(FstRef x);
static Fst _r_loops(FstRef x, int l1, int l2=0, int l3=0);
static Fst _reverse(FstRef x);
static Fst _shorten(FstRef x);
static Fst _star(FstRef x);
#endif

#ifdef SIGMA_DEBUG
static ostream&
operator<<(ostream& os, SetRef s)
{
    for (SetIter i = s.begin(); i != s.end(); i++)
        os << " " << FSTstring_map[*i];
    return os;
}
#endif

// For convenience we'll define multiplication of automata to be composition.
//
inline Fst operator*(FstRef x, FstRef y)
{
    return _compose(x, y);
}

///////////////////////////////////////////////////////////////////////////////
// FSTRules

FSTRules::FSTRules(bool warn_)
    : root(0),
      warn(warn_)
{}

FSTRules::~FSTRules()
{
    delete root;
}

void
FSTRules::eval()
{
    assignments.clear();
    fst = _minimize(root->eval(this));
}

///////////////////////////////////////////////////////////////////////////////

Fst
FSTRules::SymbolNode::eval(FSTRules* context)
{
    Fst result;
    if (!_lookup(result, context, _label))
        result = _automaton(_label);
    return result;
}

///////////////////////////////////////////////////////////////////////////////

Fst
FSTRules::SetNode::eval(FSTRules* context)
{
    // Expand variables in set by inserting their input alphabets (ignoring
    // their topology).
    //
    Set orig_set = _set;
    _set.clear();
    for (SetIter i = orig_set.begin(); i != orig_set.end(); i++) {
        Fst value;
        if (_lookup(value, context, *i)) {
            SetRef s = value->input_alphabet();
            fst::set_insert(_set, s.begin(), s.end());
        } else {
            _set.insert(*i);
        }
    }

    // Now build automaton for set.
    //
    return _automaton(_set);
}

///////////////////////////////////////////////////////////////////////////////

Fst
FSTRules::DotNode::eval(FSTRules* context)
{
    Fst x = FSTBasic();
    x->add_states(2);
    x->set_initial(0);
    x->set_final(1);
    for (SetIter i = context->sigma.begin(); i != context->sigma.end(); i++)
        x->add_arc(0, 1, *i, *i);
    return x;
}

///////////////////////////////////////////////////////////////////////////////

Fst
FSTRules::OptionalNode::eval(FSTRules* context)
{
    return _optional(_node->eval(context));
}

///////////////////////////////////////////////////////////////////////////////

Fst
FSTRules::StarNode::eval(FSTRules* context)
{
    return _trim(FSTClosure(_node->eval(context)));
}

///////////////////////////////////////////////////////////////////////////////

Fst
FSTRules::PlusNode::eval(FSTRules* context)
{
    return _trim(FSTClosure(_node->eval(context), true));
}

///////////////////////////////////////////////////////////////////////////////

Fst
FSTRules::ComplementNode::eval(FSTRules* context)
{
    return _complement(context->sigma, _node->eval(context));
}

///////////////////////////////////////////////////////////////////////////////

FSTRules::ConcatenateNode::~ConcatenateNode()
{
    for (NodeList::iterator i = _nodes.begin(); i != _nodes.end(); i++)
        delete *i;
}

Fst
FSTRules::ConcatenateNode::eval(FSTRules* context)
{
    if (_nodes.empty()) {
        return FSTBasic();
    } else {
        fst::vector<FSTGeneric> v;
        for (NodeList::iterator i = _nodes.begin(); i != _nodes.end(); i++)
            v.push_back((*i)->eval(context));
        return _trim(FSTConcatenate(v));
    }
}

///////////////////////////////////////////////////////////////////////////////

Fst
FSTRules::UnionNode::eval(FSTRules* context)
{
    return _union(_node_a->eval(context), _node_b->eval(context));
}

///////////////////////////////////////////////////////////////////////////////

Fst
FSTRules::IntersectNode::eval(FSTRules* context)
{
    return _compose(_node_a->eval(context), _node_b->eval(context));
}

///////////////////////////////////////////////////////////////////////////////

Fst
FSTRules::AssignNode::eval(FSTRules* context)
{
    context->assignments[_label] = _node->eval(context);
    context->sigma.erase(_label);
    return FSTBasic();
}

///////////////////////////////////////////////////////////////////////////////

Fst
FSTRules::IgnoreNode::eval(FSTRules* context)
{
    Set ignore_save = context->ignore;
    context->ignore = _ignore_node->eval(context)->input_alphabet();
    context->ignore.erase(0);
    Fst result = _series_node->eval(context);
    context->ignore = ignore_save;
    return result;
}

///////////////////////////////////////////////////////////////////////////////

Fst
FSTRules::SeriesNode::eval(FSTRules* context)
{
    if (_nodes.size() == 0) {
        return _star(context->sigma);
    } else {
        NodeList::iterator i = _nodes.begin();
        Fst x = (*i)->eval(context);
        for (i++; i != _nodes.end(); i++) {
            x = _minimize(_compose(x, (*i)->eval(context)));
            if (x->n_states() == 1 && x->n_arcs() == 0)
                x = FSTBasic();
        }
        return x;
    }
}

///////////////////////////////////////////////////////////////////////////////

Fst
FSTRules::ExprNode::eval(FSTRules* context)
{
    Fst x = _node->eval(context);

    if (context->warn) {
        context->warned.clear();
        _warn_missing(context, x->input_alphabet(), _line);
    }

    // Ensure that all labels are in sigma, even if they were not listed in
    // the alphabet specification.
    //
    SetRef alphabet = x->input_alphabet();
    for (SetIter i = alphabet.begin(); i != alphabet.end(); i++)
        context->sigma.insert(*i);

    return _i_loops_constrained(x, context->sigma, context->ignore);
}

///////////////////////////////////////////////////////////////////////////////

FSTRules::RuleNode::RuleNode(int line, int type,
                             Node* phi, Node* psi, Node* lambda, Node* rho)
    : _line(line),
      _type(type),
      _node_phi(phi),
      _node_psi(psi),
      _node_lambda(lambda),
      _node_rho(rho)
{
}

FSTRules::RuleNode::~RuleNode()
{
    delete _node_phi;
    delete _node_psi;
    delete _node_lambda;
    delete _node_rho;
}

Fst
FSTRules::RuleNode::eval(FSTRules* context)
{
    return _empty();

#ifdef UNUSED

    // Decode type.
    //
    bool mandatory = _type & MANDATORY;
    bool l_to_r    = _type & DIRECTION_RL;
    bool r_to_l    = _type & DIRECTION_RL;
    bool directed  = _type & DIRECTED;

    // Evaluate component expressions.
    //
    _phi    = _node_phi->eval(context);
    _psi    = _node_psi->eval(context);
    _lambda = _node_lambda ? _node_lambda->eval(context) : _automaton(0);
    _rho    = _node_rho    ? _node_rho->eval(context)    : _automaton(0);

    // Warn if expressions symbols are not in relevant alphabet.
    //
    if (context->warn) {
        context->warned.clear();
        _warn_missing(context, _phi->input_alphabet(), _line);
        if (l_to_r)
            _warn_missing(context, _rho->input_alphabet(), _line);
        else
            _warn_missing(context, _psi->input_alphabet(),
                          _rho->input_alphabet(), _line);
        if (r_to_l)
            _warn_missing(context, _lambda->input_alphabet(), _line);
        else
            _warn_missing(context, _psi->input_alphabet(),
                          _lambda->input_alphabet(), _line);
    }

    // Build local copy of sigma that does not contain ignored symbols.
    //
    Set sigma = context->sigma;
    for (SetIter i = context->ignore.begin(); i != context->ignore.end(); i++)
        sigma.erase(*i);

    // Quick result: if phi, lambda, or rho accept nothing, then the rule will
    // never fire so return sigma star.
    //
    if (_phi->initial() < 0 || _lambda->initial() < 0 || _rho->initial() < 0)
        return _star(sigma);

    // Build the rule's transducer.
    //
    Fst rule;
    if (r_to_l && l_to_r)
        rule = _build_s(sigma, mandatory);
    else
        rule = _build_d(sigma, mandatory, directed, r_to_l);

    // Add ignore loops.
    //
    if (context->ignore.size() != 0)
        rule = _i_loops_constrained(rule, sigma, context->ignore);

#ifdef SIGMA_DEBUG
    cerr << "line: " << _line << "\n";
    cerr << "sigma:   " << context->sigma << "\n";
    cerr << "rule in: " << rule->input_alphabet() << "\n";
    cerr << "rule out:" << rule->output_alphabet() << "\n";
#endif

    // Set current alphabet to rule's output alphabet.
    //
    context->sigma = rule->output_alphabet();

    // Clean up (not necessary).
    //
    _phi.destroy();
    _psi.destroy();
    _lambda.destroy();
    _rho.destroy();

    return rule;

#endif // UNUSED
}

#ifdef UNUSED

Fst
FSTRules::RuleNode::_build_s(SetRef sigma, bool mandatory)
{
    // Minimize all automata.  (The minimum we could do is determinize phi.)
    //
    _phi    = _minimize(_phi);
    _psi    = _minimize(_psi);
    _lambda = _minimize(_lambda);
    _rho    = _minimize(_rho);

    // Lookup auxiliary symbols.
    //
    int left1  = FSTstring_map[AUXILIARY_LEFT1];
    int left2  = FSTstring_map[AUXILIARY_LEFT2];
    int right1 = FSTstring_map[AUXILIARY_RIGHT1];

    // Compute r1 that inserts right1 before all instances of rho.
    //
    Fst r1 = _build_r1(sigma, _rho, right1);

    // Compute f that inserts {left1, left2} before phi right1.
    //
    Fst f = _build_f(sigma, _phi, left1, left2, right1);

    // Compute l1 that accepts and deletes left1 following lambda.
    //
    Fst alpha = _determinize(_concat(_star(sigma), _lambda));
    Fst l1 = _i_loops(_build_l1(alpha, left1, true), left2, right1);

    // Compute l2 that rejects left2 following lambda if mandatory, otherwise
    // does nothing.
    //
    Fst l2;
    if (mandatory) {
        l2 = _i_loops(_build_l2(alpha, left2), left1, right1);
    } else {
        l2 = FSTBasic();
    }

    // Compute replace.
    //
    Fst a = _automaton(sigma);
    a->add_arc(0, 1, right1, 0);
    Fst b = _transducer(left1, 0);
    if (mandatory) {
        b = _concat(b, _optional(_concat(_r_loops(_partial(_phi),
                                                  left1, right1),
                                         _transducer(left1, 0))));
    } else {
        a->add_arc(0, 1, left1, 0);
    }
    Fst replace = 
        _star(_union(a,
                     _concat(b,
                             _r_loops(_cross(_phi, _psi), left1, right1),
                             _transducer(right1, 0))));
#if 0
    r1->dump_ascii("r1");
    f->dump_ascii("f");
    l1->dump_ascii("l1");
    l2->dump_ascii("l2");
    replace->dump_ascii("rep");
#endif

    // Combine components and return.
    //
    return _minimize(r1 * f * l1 * l2 * replace);
}

Fst
FSTRules::RuleNode::_build_d(SetRef sigma, bool mandatory, bool directed,
                             bool right_to_left)
{
    // If we're building a right-to-left rule, reverse all inputs and swap
    // lambda and rho.  We'll reverse the result later to complete the
    // right-to-left behavior.
    //
    if (right_to_left) {
        _phi    = _reverse(_phi);
        _psi    = _reverse(_psi);
        _lambda = _reverse(_lambda);
        _rho    = _reverse(_rho);
        swap(_lambda, _rho);
    }

    // Minimize all input automata.  (At minimum we must determinize phi.)
    //
    _phi    = _minimize(_phi);
    _psi    = _minimize(_psi);
    _lambda = _minimize(_lambda);
    _rho    = _minimize(_rho);

    // Lookup auxiliary symbols.
    //
    int left1  = FSTstring_map[AUXILIARY_LEFT1];
    int left2  = FSTstring_map[AUXILIARY_LEFT2];
    int right1 = FSTstring_map[AUXILIARY_RIGHT1];
    int right2 = FSTstring_map[AUXILIARY_RIGHT2];

    // We will need to use r2 if we're doing directed replacement and rho is
    // non-empty.
    //
    bool use_r2 = directed && !_rho->is_final(_rho->initial());

    // Optionally compute r2 that insert right2 before the shortest possible
    // rho.  This is only used for directed replacement and indicates the last
    // possible place that we can transition from replacement to context.
    //
    Fst r2;
    if (use_r2) {
        r2 = _build_r2(sigma, _rho, right2);
    } else {
        r2 = FSTBasic();
    }

    // Compute r1 that inserts right1 before all instances of rho.
    //
    Fst r1 = _build_r1(sigma, _rho, right1);
    if (use_r2)
        r1 = _i_loops(r1, right2);

    // Compute f that inserts {left1, left2} before phi right1.
    //
    Fst f = _build_f(sigma, _phi, left1, left2, right1);
    if (use_r2)
        f = _i_loops(f, right2);

    // Optionally compute m that enforces directed (maximal) replacements by
    // allowing an exit from phi only if phi could not have been extended or if
    // there is a right2 (which forces an exit from phi).
    //
    Fst m;
    if (directed) {
        m = _build_m(sigma, _phi, left1, left2, right1, right2);
    } else {
        m = FSTBasic();
    }

    // Compute replace.
    //
    Fst a = _automaton(sigma);
    a->add_arc(0, 1, right1, 0);
    a->add_arc(0, 1, left2, mandatory ? left2 : 0);
    Fst replace = 
        _star(_union(a,
                     _concat(_transducer(left1, left1),
                             _r_loops(_cross(_phi, _psi), right1, left1),
                             _transducer(right1, 0))));

    // Compute l1 that accepts and deletes left1 following lambda.
    //
    Fst alpha = _determinize(_concat(_star(sigma), _lambda));
    Fst l1 = _i_loops(_build_l1(alpha, left1, false), left2);

    // Compute l2 that rejects left2 following lambda if mandatory, otherwise
    // does nothing.
    //
    Fst l2;
    if (mandatory) {
        l2 = _build_l2(alpha, left2);
    } else {
        l2 = FSTBasic();
    }

    // Combine components.
    //
    Fst result = r2 * r1 * f * m * replace * l1 * l2;
    
    // Reverse result if building a right-to-left rule.
    //
    if (right_to_left)
        result = _reverse(result);

    // Finally, return minimized result.
    //
    return _minimize(result);
}

#endif // UNUSED

///////////////////////////////////////////////////////////////////////////////
//
// Static support functions.
//

// Look up label l in context->assignments, returning true and setting result
// if found, else returning false.
//
static bool
_lookup(Fst& result, FSTRules* context, int l)
{
    FstMap::iterator i = context->assignments.find(l);
    if (i != context->assignments.end()) {
        result = (*i).second;
        return true;
    } else {
        return false;
    }
}

// Warn if symbols in set are not in sigma.
//
static void
_warn_missing(FSTRules* context, SetRef set, int line)
{
    _warn_missing(context, Set(), set, line);
}

// Warn if symbols in set are not in sigma or sigma2.
//
static void
_warn_missing(FSTRules* context, SetRef sigma2, SetRef set, int line)
{
    for (SetIter i = set.begin(); i != set.end(); i++) {
        if (*i && !context->sigma.count(*i) && !sigma2.count(*i) &&
            !context->warned.count(*i)) {
            context->warned.insert(*i);
            FSTWarn("`%s' not in alphabet at line %d",
                    FSTstring_map[*i], line);
        }
    }
}

// Remove epsilons and dead states.  Make sure we are left with at least one
// state.
//
static Fst
_trim(FstRef x)
{
    Fst y = FSTTrim(FSTRemoveEpsilons(x));
    if (y->initial() < 0)
        y = _empty();
    return y;
}

// Automata determinization.
//
static Fst
_determinize(FstRef x)
{
    // The FSTBasic is needed because FSTDeterminize doesn't implement
    // arcs_out_with_input yet.  The true forced automata determinization
    // instead of transducer determinization.
    //
    return FSTBasic(FSTDeterminize(_trim(x), 1, true));
}

// Automata minimization.
//
static Fst
_minimize(FstRef x)
{
    return FSTMinimize(_determinize(x));
}

// Automata intersection or transducer composition, treating empty transducers
// as identity.
//
static Fst
_compose(FstRef x, FstRef y)
{
    if (x->n_states() == 0)
        return y;
    else if (y->n_states() == 0)
        return x;
    else
        return(_trim(FSTCompose(x, y)));
}

#if UNUSED
// Concatenation of automata or transducers.
//
static Fst
_concat(FstRef x, FstRef y)
{
    return _trim(FSTConcatenate(x, y));
}
static Fst
_concat(FstRef x, FstRef y, FstRef z)
{
    return _trim(FSTConcatenate(x, y, z));
}
#endif // UNUSED

// Union of automata or transducers.
//
static Fst
_union(FstRef x, FstRef y)
{
    return _trim(FSTUnion(x, y));
}

// Compute the complement of automaton x.
//
static Fst
_complement(SetRef sigma, FstRef x)
{
    Fst y = _determinize(x);

    int N = y->n_states();
    Fst result = FSTBasic();
    result->add_states(N + 1);
    int dead_state = N;

    result->set_initial(y->initial());
    result->set_final(dead_state);
    for (SetIter i = sigma.begin();
         i != sigma.end(); i++) {
        for (int p = 0; p < N; p++) {
            ArcIter a = y->arcs_out_with_input(p, *i);
            if (a.valid())
                result->add_arc(p, a->next, *i, *i);
            else
                result->add_arc(p, dead_state, *i, *i);
        }
        result->add_arc(N, N, *i, *i);
    }
    for (int p = 0; p < N; p++)
        if (!y->is_final(p))
            result->set_final(p);

    return result;
}

// Compute reverse of x.
//
#if UNUSED
static Fst
_reverse(FstRef x)
{
    return _trim(FSTReverse(x));
}
#endif // UNUSED

// Compute optional x.
//
static Fst
_optional(FstRef x)
{
    Fst y = FSTBasic(x);
    y->set_final(y->initial());
    return y;
}

// Compute Kleene-star closure of regular expression x.
//
#if UNUSED
static Fst
_star(FstRef x)
{
    return _trim(FSTClosure(x));
}
#endif

// Compute Kleene-star closure of alphabet sigma, treating label1, label2, and
// label3 as part of sigma if they are not epsilon (0).
//
static Fst
_star(SetRef sigma, int label1, int label2, int label3)
{
    Fst x = FSTBasic();
    x->add_state();
    x->set_initial(0);
    x->set_final(0);
    for (SetIter i = sigma.begin(); i != sigma.end(); i++)
        x->add_arc(0, 0, *i, *i);
    if (label1)
        x->add_arc(0, 0, label1, label1);
    if (label2)
        x->add_arc(0, 0, label2, label2);
    if (label3)
        x->add_arc(0, 0, label3, label3);
    return x;
}

// Compute cross product of two automata, producing a transducer.  To simplify
// some other code, if the cross product is empty, produce a single initial
// state that is non-final.
//
#if UNUSED
static Fst
_cross(FstRef x, FstRef y)
{
    Fst z =
        FSTProject(x, FSTInput, FSTNone) *
        FSTProject(y, FSTNone, FSTOutput);
    if (z->n_states() == 0) {
        z = FSTBasic();
        z->add_state();
        z->set_initial(0);
    }
    return z;
}
#endif // UNUSED

// Introduce a single instance label in all positions.  The language is the
// union of sigma and extra_sigma (if specified).
//
#if UNUSED
static Fst
_introduce(SetRef sigma, int label, int extra_sigma)
{
    Fst r = FSTBasic();
    r->add_states(2);
    r->set_initial(0);
    r->set_final(1);
    r->add_arc(0, 1, 0, label);
    for (SetIter i = sigma.begin();
         i != sigma.end(); i++) {
        r->add_arc(1, 0, *i, *i);
    }
    if (extra_sigma)
        r->add_arc(1, 0, extra_sigma, extra_sigma);
    return r;
}
#endif // UNUSED

// Construct automaton by making all non-final states final and vice-a-versa.
// This requires x to be deterministic.  This should accept "partial" strings.
// It is similar to complementation except that we don't add arcs for
// non-matching labels.
//
#if UNUSED
static Fst
_partial(FstRef x)
{
    Fst y = FSTBasic(x);
    for (int p = 0; p < y->n_states(); p++)
        if (y->is_final(p))
            y->clear_final(p);
        else
            y->set_final(p);
    return _trim(y);
}
#endif // UNUSED

// Construct automaton that accepts only shortest prefixes of x.  We construct
// this from x by removing all arcs from final states.  x is required to be
// deterministic.
//
#if UNUSED
static Fst
_shorten(FstRef x)
{
    fst::vector<FSTArc*> no_arcs;
    Fst y = _determinize(x);
    for (int p = 0; p < y->n_states(); p++)
        if (y->is_final(p))
            y->set_arcs(p, no_arcs);
    return _trim(y);
}
#endif // UNUSED

// Construct empty automaton.  We make it have a single state that is initial
// but not final.  This simplifies some of the other code so that it doesn't
// need to deal with zero states and initial = -1.
//
static Fst
_empty()
{
    Fst x = FSTBasic();
    x->set_initial(x->add_state());
    return x;
}

// Construct automaton accepting single symbol from sigma.
//
static Fst
_automaton(SetRef sigma)
{
    Fst x = FSTBasic();
    x->add_states(2);
    x->set_initial(0);
    x->set_final(1);
    for (SetIter i = sigma.begin(); i != sigma.end(); i++)
        x->add_arc(0, 1, *i, *i);
    return x;
}

// Construct one-arc automaton.
//
static Fst
_automaton(int input)
{
    return _transducer(input, input);
}


// Construct one-arc transducer.
//
static Fst
_transducer(int input, int output)
{
    Fst x = FSTBasic();
    if (!input && !output) {
        x->add_state();
        x->set_initial(0);
        x->set_final(0);
    } else {
        x->add_states(2);
        x->set_initial(0);
        x->set_final(1);
        x->add_arc(0, 1, input, output);
    }
    return x;
}

// Make all states final.
//
#if UNUSED
static Fst
_all_final(FstRef x)
{
    Fst y = FSTBasic(x);
    for (int p = 0; p < y->n_states(); p++)
        y->set_final(p);
    return y;
}
#endif // UNUSED

// Construct transducer that inserts mark after first occurence of regular
// expression x.  If there is a sequence of x's, only the first (shortest) is
// marked.  x is required to be deterministic.
//
#if UNUSED
static Fst
_mark_after_first(SetRef sigma, FstRef x, int mark)
{
    Fst a = _shorten(x);
    Fst b = _transducer(0, mark);
    Fst c = _all_final(FSTClosure(x));

    Fst y = _minimize(FSTConcatenate(a, b, c));

    // For every state p in y and every label l in sigma such that there is no
    // arc leaving p with label l: add arc from p to the initial state q with
    // label l:l.  However, do not add any arcs to states with epsilon arcs
    // leaving (i.e., those whose only arc leaving is eps:mark).
    //
    int q = y->initial();
    y->set_final(q);
    for (int p = 0; p < y->n_states(); p++) {
        if (y->arcs_out_with_input(p, 0).n() == 0) {
            for (SetIter i = sigma.begin();
                 i != sigma.end(); i++) {
                int l = *i;
                if (y->arcs_out_with_input(p, l).n() == 0) {
                    y->add_arc(p, q, l, l);
                }
            }
        }
    }

    return y;
}
#endif // UNUSED

#if UNUSED
static Fst
_marker1(FstRef alpha, int insert1, int insert2)
{
    int p;

    alpha->force();
    int N = alpha->n_states();

    Fst result = FSTBasic();
    result->add_states(N);
    result->set_initial(alpha->initial());

    // For each alpha final state p create a new state and store its index in
    // map[p].  We also create an arc epsilon:insert1 and optionally an arc
    // epsilon:insert2 from the new state to the original final state p.
    //
    fst::vector<int> map(alpha->n_states());
    for (p = 0; p < N; p++) {
        if (alpha->is_final(p)) {
            map[p] = result->add_state();
            result->add_arc(map[p], p, 0, insert1);
            if (insert2) {
                result->add_arc(map[p], p, 0, insert2);
            }
        } else {
            map[p] = -1;
        }
    }

    // Go through all alpha states, changing non-final states to final and
    // changing all arcs to final states to their new states according to map.
    //
    for (p = 0; p < N; p++) {
        result->set_final(p);
        for (ArcIter a = alpha->arcs_out(p); a.valid(); a++) {
            int q = a->next;
            if (map[q] >= 0)
                q = map[q];
            result->add_arc(p, q, a->input, a->input, a->weight);
        }
    }

    if (alpha->n_states() == 1)
        for (p = 0; p < N; p++)
            if (map[p] >= 0)
                result->set_final(map[p]);

    return result;
}

static Fst
_marker2(FstRef alpha, int input, int output)
{
    Fst result = FSTBasic(alpha);
    for (int p = 0; p < result->n_states(); p++) {
        if (result->is_final(p)) {
            result->add_arc(p, p, input, output);
        } else {
            result->set_final(p);
        }
    }

    return result;
}

static Fst
_marker3(FstRef alpha, int remove)
{
    Fst result = FSTBasic(alpha);
    for (int p = 0; p < result->n_states(); p++) {
        if (result->is_final(p)) {
        } else {
            if (remove)
                result->add_arc(p, p, remove, 0);
            result->set_final(p);
        }
    }

    return result;
}
#endif // UNUSED

// Construct machine with that removes labels by adding l1:eps, l2:eps, and
// l3:eps loops at every state.
//
#if UNUSED
static Fst
_r_loops(FstRef x, int l1, int l2, int l3)
{
    Fst y = FSTBasic(x);
    for (int p = 0; p < y->n_states(); p++) {
        y->add_arc(p, p, l1, 0);
        if (l2)
            y->add_arc(p, p, l2, 0);
        if (l3)
            y->add_arc(p, p, l3, 0);
    }
    return y;
}
#endif // UNUSED

// Construct machine with that ignores labels by adding l1:l1, l2:l2, and l3:l3
// loops at every state.
//
#if UNUSED
static Fst
_i_loops(FstRef x, int l1, int l2, int l3)
{
    Fst y = FSTBasic(x);
    for (int p = 0; p < y->n_states(); p++) {
        y->add_arc(p, p, l1, l1);
        if (l2)
            y->add_arc(p, p, l2, l2);
        if (l3)
            y->add_arc(p, p, l3, l3);
    }
    return y;
}
#endif // UNUSED

// Add ignore loops, but constained so that they do not come after input
// epsilon arcs.  Without this contraint, rule insertions can come both before
// and after ignored symbols (e.g., "#"), causing unwanted extra paths.
//

static Fst
_i_loops_constrained(FstRef x, SetRef sigma, SetRef ignore)
{
    if (ignore.empty())
        return x;

    // First, relabel all input epsilons with auxiliary symbol eps and add
    // ignore loops x:x to every state.
    //
    int eps = FSTstring_map["<<eps>>"];
    Fst y = FSTBasic();
    y->add_states(x->n_states());
    y->set_initial(x->initial());
    for (int p = 0; p < x->n_states(); p++) {
        y->set_final(p, x->final_weight(p));
        for (ArcIter a = x->arcs_out(p); a.valid(); a++)
            y->add_arc(p, a->next,
                      a->input ? a->input : eps,
                      a->output, a->weight);
        for (SetIter i = ignore.begin(); i != ignore.end(); i++)
            y->add_arc(p, p, *i, *i);
    }

    // Next, build a filter transducer that will insert ignore arcs except
    // before eps, and relabel eps to epsilon.
    //
    Fst z = _star(sigma);
    int p = z->initial();
    int q = z->add_state();
    z->set_final(q);
    z->add_arc(p, p, 0, eps);
    for (SetIter i = ignore.begin(); i != ignore.end(); i++) {
        z->add_arc(p, q, *i, *i);
        z->add_arc(q, q, *i, *i);
    }
    for (SetIter i = sigma.begin(); i != sigma.end(); i++)
        z->add_arc(q, p, *i, *i);

    // Finally, compose the filter transducer z and the relabeled input x.
    //
    return z * y;
}

// Build transducer r1 that inserts right1 before every rho.
// 
#if UNUSED
static Fst
_build_r1(SetRef sigma, FstRef rho, int right1)
{
    // Rho allowing empty is equivalent to no right context, so simply
    // introduce right1 everywhere in this case.
    //
    if (rho->is_final(rho->initial()))
        return _introduce(sigma, right1);

    Fst alpha = _determinize(_concat(_star(sigma), _reverse(rho)));
    return _reverse(_marker1(alpha, right1));
}
#endif // UNUSED

// Build transducer f that inserts {left1, left2} before every phi right1.
//
#if UNUSED
static Fst
_build_f(SetRef sigma, FstRef phi, int left1, int left2, int right1)
{
    Fst alpha =
        _determinize(_concat(_star(sigma, right1),
                             _reverse(_concat(_i_loops(phi, right1),
                                              _automaton(right1)))));
    return _reverse(_marker1(alpha, left1, left2));
}
#endif // UNUSED

// Build transducer l1 that accepts and optionally deletes left1 following
// deterministic alpha.
//
#if UNUSED
static Fst
_build_l1(FstRef alpha, int left1, bool keep)
{
    return _trim(_marker2(alpha, left1, keep ? left1 : 0));
}
#endif // UNUSED

// Build transducer l1 that rejects left1 following deterministic alpha.
//
#if UNUSED
static Fst
_build_l2(FstRef alpha, int left2)
{
    return _trim(_marker3(alpha, left2));
}
#endif // UNUSED

// Build transducer r2 that inserts right2 before rho such that only one rho
// will be inserted immediately before the shortest possible rho.  This is used
// for directed replacement.
//
#if UNUSED
static Fst _build_r2(SetRef sigma, FstRef rho, int right2)
{
    return _minimize(_reverse(_mark_after_first(sigma, _reverse(rho), right2)));
}
#endif // UNUSED

// Build transducer m that enforces maximal replacements by allowing an exit
// from phi only if phi could not have been extended or if there is a right2.
// The "could not have been extended" has a feel of complementation.
//
// Unfortunately, I have not found a clean way to express m using regular
// expression operations.  It probably could be, but I doubt it would be as
// efficient as the following.
//

#if UNUSED
static Fst _build_m(SetRef sigma, FstRef phi,
                           int left1, int left2, int right1, int right2)
{
    Fst m = FSTBasic(phi);
    int old_initial = m->initial();
    int initial = m->add_state();

    m->set_initial(initial);
    m->set_final(initial);

    for (SetIter i = sigma.begin(); i != sigma.end(); i++)
        m->add_arc(initial, initial, *i, *i);
    m->add_arc(initial, initial, right1, right1);
    m->add_arc(initial, initial, left2, left2);
    m->add_arc(initial, old_initial, left1, left1);

    for (int p = 0; p < phi->n_states(); p++) {
        m->add_arc(p, p, right1, 0);
        m->add_arc(p, p, left1, 0);
        m->add_arc(p, p, left2, left2);
        if (m->is_final(p)) {
            m->clear_final(p);
            m->add_arc(p, initial, right2, 0);
            if (m->arcs_out(p).n() == 0) {
                m->add_arc(p, initial, right1, right1);
            } else {
                int q = m->add_state();
                m->set_final(q);
                m->add_arc(p, q, right1, right1);
                m->add_arc(q, q, right1, right1);
                m->add_arc(q, q, left1, left1);
                m->add_arc(q, q, left2, left2);
                for (SetIter i = sigma.begin(); i != sigma.end(); i++)
                    if (m->arcs_out_with_input(p, *i).n() == 0)
                        m->add_arc(q, initial, *i, *i);
            }
        }
    }
    m = _minimize(_r_loops(m, right2));
    return m;
}
#endif // UNUSED

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 79
// indent-tabs-mode: nil
// c-basic-offset: 4
// End:
