////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

// Issues:
//
// * # special handling disabled
// * epsilon composition filtering?

#include "fst/FST.h"
#include "fst/FSTTimer.h"
#include "fst/format.h"

using std::cerr;
using std::endl;
using std::flush;

#define BAD_SCORE 1e38

#ifdef DEBUG_SEARCH
#include <fstream>
static std::ofstream debug("search.log");
#endif

class FSTSearchRep {

protected:
    
    // A Node is a dynamic programming node and holds a link to the best arriving
    // path.
    //
    struct Node {
        float   score;
        int     state_a;
        int     state_b;
        Node*   prev;
        int     input;
        int     output;
        float   weight;

        Node(int a, int b) : score(BAD_SCORE), state_a(a), state_b(b) {}
        FSTAllocator_new_delete(Node);
    };
    typedef fst::vector<Node*>      NodeVector;
    typedef fst::vector<NodeVector> NodeVectorVector;

    // Node comparison objects (for sorting).
    //
    struct NodeScoreGreater {
        bool operator() (const Node* a, const Node* b) const {
            return a->score > b->score;
        }
    };
    struct NodeStateLess {
        bool operator() (const Node* a, const Node* b) const {
            return a->state_b < b->state_b;
        }
    };

protected:

    // FST inputs.
    //
    FSTGeneric fst_a;
    FSTGeneric fst_b;

    // Pruning parameters.
    //
    float vprune;
    int   vnodes;
    bool  sort_nodes;

    // Sparse matrix of active nodes at each boundary.
    //
    NodeVectorVector matrix;

    // Information about nodes at the frontier boundary.  Frontier maps state
    // to node.  Frontier_active is used as a list active states.
    //
    NodeVector frontier;
    NodeVector frontier_active;
    float      frontier_best_score;

    // Miscellaneous.
    //
    bool verbose;
    int n_arcs;
    int n_nodes;
    typedef fst::vector<int> IntVector;
    IntVector *nodes;
    FSTTimer     timer;

public:

    FSTSearchRep(const FSTGeneric& fst_a, const FSTGeneric& fst_b,
                 float vprune, int vnodes, bool sort_nodes, bool verbose,
                 IntVector *nodes=0);
    ~FSTSearchRep();

    FSTGeneric viterbi();
    FSTGeneric viterbi_graph();

protected:
    void viterbi_initialize();
    void viterbi_update(int state_a);
    void viterbi_prune(bool& count_reached);
    void viterbi_epsilons();
    void viterbi_compress(int state_a);
    void viterbi_forward();
    void viterbi_sort();
    int viterbi_number();
    inline const int find_node(int a, int b);
    FSTGeneric viterbi_backtrace();
    void sort_matrix();
    void propagate_epsilons(Node* from_node, float max_score);
    inline Node*       find_or_create_node(int a, int b);
};

///////////////////////////////////////////////////////////////////////////////

FSTSearchRep::FSTSearchRep(const FSTGeneric& fst_a_, const FSTGeneric& fst_b_,
                           float vprune_, int vnodes_, bool sort_nodes_,
                           bool verbose_, IntVector *nodes_)
    : fst_a(FSTCache2(fst_a_)), fst_b(FSTCache2(fst_b_)),
      vprune(vprune_), vnodes(vnodes_), sort_nodes(sort_nodes_),
      verbose(verbose_), nodes(nodes_)
{
    assert(FSTstring_map["#"] == 1);
    assert(fst_a->initial() == 0);
    assert(fst_a->semiring() == FSTMinPlus);
    assert(fst_b->semiring() == FSTMinPlus);

    fst_a->force();
}

FSTSearchRep::~FSTSearchRep()
{
    for (NodeVectorVector::iterator i = matrix.begin();
         i != matrix.end(); i++) {
        NodeVector& v = *i;
        for (NodeVector::iterator j = v.begin(); j != v.end(); j++) {
            Node* n = *j;
            delete n;
        }
    }
}

FSTGeneric
FSTSearchRep::viterbi()
{
    viterbi_forward();
    return viterbi_backtrace();
}

void
FSTSearchRep::viterbi_sort()
{
    NodeStateLess comparator;
    for (NodeVectorVector::iterator i = matrix.begin(); i != matrix.end(); i++)
        sort((*i).begin(), (*i).end(), comparator);
}

int
FSTSearchRep::viterbi_number()
{
    int s = 0;
    for (NodeVectorVector::iterator i = matrix.begin(); i != matrix.end(); i++)
        for (NodeVector::iterator j = (*i).begin(); j != (*i).end(); j++)
            (*j)->output = s++;
    return s;
}

inline const int
FSTSearchRep::find_node(int a, int b)
{
    Node key(a, b);
    const NodeVector& v = matrix[a];
    NodeStateLess comparator;
    NodeVector::const_iterator i = lower_bound(v.begin(), v.end(),
                                               &key, comparator);
    return (i == v.end() || (*i)->state_b != b) ? -1 : (*i)->output;
}

typedef std::pair<int, int> Pair;
typedef fst::map<Pair, int> PairMap;

#if 0
static int
_find_or_create_result_node(int a, int b, PairMap& table)
{
    std::pair<PairMap::iterator, bool> status = table.insert(PairMap::value_type(Pair(a, b), table.size()));
    return (*i).second;
}
#endif

FSTGeneric
FSTSearchRep::viterbi_graph()
{
    // Perform forward pass, sort, and number nodes
    //
    viterbi_forward();
    if(!sort_nodes)
        viterbi_sort();
    int num_states = viterbi_number();

    // Initialize result
    //
    FSTGeneric result = FSTBasic();
    result->add_states(num_states);
    assert(fst_a->initial() == 0);
    int initial = find_node(0, fst_b->initial());
    assert(initial >= 0);
    result->set_initial(initial);

    // Walk horizontally through trellis
    //
    for (int a = 0; a < fst_a->n_states(); a++) {
        NodeVector& nodes = matrix[a];
        bool a_final = fst_a->is_final(a);
        float a_final_weight = fst_a->final_weight(a);

        // Walk vertically up slice of trellis
        //
        for (NodeVector::iterator ni = nodes.begin(); ni != nodes.end(); ni++) {
            Node *node = *ni;
            assert(node->state_a == a);
            int b = node->state_b;
            int p = node->output;
            assert(p >= 0);
            if (a_final && fst_b->is_final(b))
                result->set_final(p, a_final_weight + fst_b->final_weight(b));

            // Handle B input epsilons, in which A will not make a transition.
            //
            for (FSTArcIter bi = fst_b->arcs_out_with_input(b, 0); bi.valid(); bi++) {
                int q = find_node(a, bi->next);
                if (q >= 0) {
                    result->add_arc(p, q, 0, bi->output, bi->weight);
#if 0
                    if(fst_a->is_final(a) && fst_b->is_final(bi->next))
                        result->set_final(q,
                                          fst_a->final_weight(a) +
                                          fst_b->final_weight(bi->next));
#endif
                }
            }

            // Iterate through all A arcs departing state a.
            //
            for (FSTArcIter ai = fst_a->arcs_out(a); ai.valid(); ai++) {

                // Handle A output epsilons, in which B will not make a transition.
                //
                if (ai->output == 0) {
                    int q = find_node(ai->next, b);
                    if (q >= 0) {
                        result->add_arc(p, q, ai->input, 0, ai->weight);
#if 0
                        if(fst_a->is_final(ai->next) && fst_b->is_final(b))
                            result->set_final(q,
                                              fst_a->final_weight(ai->next) +
                                              fst_b->final_weight(b));
#endif
                    }
                }

                // Find B arcs compatible with A arc ai.
                //
                for (FSTArcIter bi = fst_b->arcs_out_with_input(b, ai->output);
                     bi.valid(); bi++) {
                    int q = find_node(ai->next, bi->next);
                    if (q >= 0) {
                        result->add_arc(p, q, ai->input, bi->output,
                                        ai->weight + bi->weight);
#if 0
                        if(fst_a->is_final(ai->next) && fst_b->is_final(bi->next))
                            result->set_final(q,
                                              fst_a->final_weight(ai->next) +
                                              fst_b->final_weight(bi->next));
#endif
                    }
                }
            }
        }
    }

    return result;
}

void
FSTSearchRep::viterbi_forward()
{
    viterbi_initialize();

    for (int a = 0; a < fst_a->n_states(); a++) {

        viterbi_update(a);

        bool count_reached = false;
        viterbi_prune(count_reached);

        viterbi_epsilons();

        viterbi_compress(a);

        if(nodes) {
            nodes->push_back(matrix[a].size());
            if(count_reached)
                *nodes->rbegin() *= -1;
        }

        if (verbose) {
            cerr << " "
                 << matrix[a].size()
                 << (count_reached ? "*" : "")
                 << flush;
        }

    }
    if (verbose)
        cerr << endl;
    if (sort_nodes)
        viterbi_sort();
}

void
FSTSearchRep::viterbi_initialize()
{
    int a = fst_a->initial();
    int b = fst_b->initial();
    assert(b >= 0);
    Node* n = find_or_create_node(a, b);
    n->score = 0;
    n->prev = 0;
    n->input = 0;
    n->output = 0;
    n->weight = 0;
}

void
FSTSearchRep::viterbi_update(int a)
{
    frontier_best_score = BAD_SCORE;
    float max_score = BAD_SCORE;

    // Iterate through all A arcs arriving at state a.
    //
    for (FSTArcIter ai = fst_a->arcs_in(a); ai.valid(); ai++) {
        int a_from = ai->prev;

        if (ai->next == ai->prev) {
            FSTWarn("A self loop (%d, %d, %s, %s, %g) -- skipping",
                       ai->prev, ai->next,
                       FSTstring_map[ai->input], FSTstring_map[ai->output],
                       ai->weight);
            continue;
        }

        if (ai->next < ai->prev) {
            FSTFatal("A not topologically sorted (%d, %d, %s, %s, %g) -- aborting",
                     ai->prev, ai->next,
                     FSTstring_map[ai->input], FSTstring_map[ai->output],
                     ai->weight);
        }

        // Handle A output epsilons, in which B will not make a transition.
        //
        if (ai->output == 0) {

            // Iterate through all nodes at A state a_from.
            //
            NodeVector& from_nodes = matrix[a_from];
            for (NodeVector::iterator ni = from_nodes.begin();
                 ni != from_nodes.end(); ni++) {
                
                Node* from_n = *ni;

                float weight = ai->weight;
                float to_score = from_n->score + weight;

                if (to_score < frontier_best_score) {
                    frontier_best_score = to_score;
                    max_score = frontier_best_score + vprune;
                }
                
                n_arcs++;
                Node* n_to = find_or_create_node(ai->next, from_n->state_b);
                if (to_score < n_to->score) {
                    n_to->score = to_score;
                    n_to->prev = from_n;
                    n_to->input = ai->input;
                    n_to->output = 0;
                    n_to->weight = weight;
                }
            }
            continue;
        }

        // Iterate through all nodes at A state a_from.
        //
        NodeVector& from_nodes = matrix[a_from];
        for (NodeVector::iterator ni = from_nodes.begin();
             ni != from_nodes.end(); ni++) {

            Node* from_n = *ni;

            // Find B arcs compatible with A arc ai.
            //
            for (FSTArcIter bi = fst_b->arcs_out_with_input(from_n->state_b, ai->output);
                 bi.valid(); bi++) {
                float weight = ai->weight + bi->weight;
                float to_score = from_n->score + weight;
#ifdef DEBUG_SEARCH
                format(debug, "%8d%8d%8d%8d%16s%16s%16g\n",
                       from_n->state_a,
                       from_n->state_b,
                       ai->next,
                       bi->next,
                       FSTstring_map[ai->input],
                       FSTstring_map[bi->output],
                       weight);
#endif
                if (to_score < frontier_best_score) {
                    frontier_best_score = to_score;
                    max_score = frontier_best_score + vprune;
                }
                
                n_arcs++;
                Node* n_to = find_or_create_node(ai->next, bi->next);
                if (to_score < n_to->score) {
                    n_to->score = to_score;
                    n_to->prev = from_n;
                    n_to->input = ai->input;
                    n_to->output = bi->output;
                    n_to->weight = weight;
                }
            }
        }
    }

}

void
FSTSearchRep::viterbi_prune(bool& count_reached)
{
    // Beam pruning based on relative score.  Note that we perform the pruning
    // before propagating the epsilons.  Otherwise, if there are epsilon
    // arcs with positive weights we might prune a predecessor node,
    // breaking the backtrace.  We do actively prune during epsilon
    // propagation.

    // Sweep through frontier deleting pruned nodes and setting their entry in
    // frontier_active to 0.
    //
    float max_score = frontier_best_score + vprune;
    for (NodeVector::iterator i = frontier_active.begin();
         i != frontier_active.end(); i++) {
        if ((*i)->score > max_score) {
            frontier[(*i)->state_b] = 0;
            delete *i;
            *i = 0;
        }
    }

    // Remove all 0 entries left in frontier_active.
    //
    frontier_active.
        erase(remove(frontier_active.begin(), frontier_active.end(), (Node*)0),
              frontier_active.end());

    int n_active = frontier_active.size();

    // Optionally prune based on maximum number of active nodes.
    //
    if (vnodes && n_active > vnodes) {

        count_reached = true;

        // Build heap containing all nodes.
        //
        NodeScoreGreater comparator;
        NodeVector heap;
        heap.reserve(n_active);
        heap.insert(heap.begin(),
                    frontier_active.begin(), frontier_active.end());
        make_heap(heap.begin(), heap.end(), comparator);

        // Pop vnodes best nodes, rebuilding frontier_active.
        //
        frontier_active.clear();
        for (int i = 0; i < vnodes; i++) {
            pop_heap(heap.begin(), heap.end() - i, comparator);
            Node* n = *(heap.end() - i - 1);
            frontier_active.push_back(n);
        }

        // Go through rest of nodes, removing them from frontier.
        //
        int n_delete = n_active - vnodes;
        for (int i = 0; i < n_delete; i++) {
            Node* n = heap[i];
            frontier[n->state_b] = 0;
            delete n;
        }

    }
}        

void
FSTSearchRep::viterbi_epsilons()
{
    // Propagate epsilon (null) arcs along frontier.  Note that the
    // word_boundary `#' is treated as epsilon.
    //
    // We make a single sweep through all the states, firing off a depth-first
    // exploration along epsilon arcs.  This may not be optimal, but it is
    // probably not worth improving it (e.g., topologically sorting states
    // involved in epsilon arcs).
    //
    // Note that since frontier_active can grow during this sweep, we cannot
    // use an iterator because that can be invalidated by a reallocation.
    // Also, even if frontier_active grows, we do not propagate epsilons for
    // new active states since those would already have been propagated by the
    // depth-first exploration.
    //
    float max_score = frontier_best_score + vprune;
    int n_active = frontier_active.size();
    for (int i = 0; i < n_active; i++) {
        Node* n = frontier_active[i];
        propagate_epsilons(n, max_score);
    }
}

void
FSTSearchRep::propagate_epsilons(Node* from_node, float max_score)
{
    float from_score = from_node->score;
#if 0
    FSTArcIter a =
        fst_b->arcs_out_with_input(from_node->state_b, 1, FSTLessEqual);
#else
    FSTArcIter a =
        fst_b->arcs_out_with_input(from_node->state_b, 0);
#endif
    for (; a.valid(); a++) {
#ifdef DEBUG_SEARCH
        format(debug, "%8d%8d%8d%8d%16s%16s%16g\n",
               from_node->state_a,
               from_node->state_b,
               from_node->state_a,
               a->next,
               "",
               FSTstring_map[a->output],
               a->weight);
#endif
        float score = from_score + a->weight;
        n_arcs++;
        if (score <= max_score) {
            Node* n =
                find_or_create_node(from_node->state_a, a->next);
            if (score < n->score) {
                n->score = score;
                n->prev = from_node;
                n->input = a->input;
                n->output = a->output;
                n->weight = a->weight;
                propagate_epsilons(n, max_score); // recursive call
            }                
        }
    }
}

void
FSTSearchRep::viterbi_compress(int state_a)
{
    // Compress active nodes on the frontier into vector-based list.  We might
    // want to sort the active nodes in descending order of score so as to
    // improve active pruning.
    //
    assert(matrix.size() == (unsigned)state_a);
    matrix.push_back(NodeVector());
    NodeVector& compressed = matrix[state_a];
    for (NodeVector::iterator j = frontier_active.begin();
         j != frontier_active.end(); j++) {
        compressed.push_back(*j);
        frontier[(*j)->state_b] = 0;
    }
    n_nodes += compressed.size();
    frontier_active.clear();
}

FSTGeneric
FSTSearchRep::viterbi_backtrace()
{
    Node* n;

    // Find best score among all final nodes.  A node is final if state_a and
    // state_b are both final.
    //
    Node* best_node = 0;
    for (FSTFinalStateIter fi = fst_a->final_states(); fi.valid(); fi++) {
        int a = fi.state();
        NodeVector& nodes = matrix[a];
        for (NodeVector::iterator i = nodes.begin(); i != nodes.end(); i++) {
            n = *i;
            if (fst_b->is_final(n->state_b)) {
                float score = n->score
                    + fst_a->final_weight(n->state_a)
                    + fst_b->final_weight(n->state_b);
                if (!best_node || score < best_node->score)
                    best_node = n;
            }
        }
    }

    if (!best_node) {
        if (verbose)
            cerr << "*** no complete path ***\n";
        return FSTBasic();
    }

    if (verbose)
        format(cerr, "best score = %f\n", best_node->score);

    // Backtrace.
    //
    FSTGeneric result = FSTBasic();
    result->add_state();
    result->set_final(0,
                     fst_a->final_weight(best_node->state_a) +
                     fst_b->final_weight(best_node->state_b));
    int p = 0;
    IntVector::iterator ivi;
    if(nodes) {
        reverse(nodes->begin(), nodes->end());
        ivi = nodes->begin();
    }
    for (n = best_node; n != NULL && n->prev != NULL; n = n->prev) {
        result->add_state();
        result->add_arc(p + 1, p, n->input, n->output, n->weight);
        if(nodes) {
            if(n->prev) {
                int n_delete = n->state_a - n->prev->state_a - 1;
                if(n_delete > 0) // Could be epsilons transitions in B
                    nodes->erase(ivi + 1, ivi + n_delete + 1);
            }
            ivi++;
        }
        p++;
    }
    if(nodes)
        reverse(nodes->begin(), nodes->end());
    result->set_initial(p);

    return result;
}

inline FSTSearchRep::Node*
FSTSearchRep::find_or_create_node(int a, int b)
{
    if (frontier.size() <= (unsigned)b) {
        frontier.insert(frontier.end(), size_t(b - frontier.size() + 1),
                        (Node*)0);
    }
    if (!frontier[b]) {
        frontier[b] = new Node(a, b);
        frontier_active.push_back(frontier[b]);
    }
    return frontier[b];
}

FSTGeneric
FSTSearch(const FSTGeneric& a, const FSTGeneric& b, float vprune, int vnodes,
          bool verbose, fst::vector<int> *nodes)
{
    FSTSearchRep* search = new FSTSearchRep(a, b, vprune, vnodes, false,
                                            verbose, nodes);
    FSTGeneric result = search->viterbi();
    delete search;
    return result;
}

FSTGeneric
FSTSearchGraph(const FSTGeneric& a, const FSTGeneric& b,
               float vprune, int vnodes, bool verbose)
{
    FSTSearchRep* search = new FSTSearchRep(a, b, vprune, vnodes, true,
                                            verbose);
    FSTGeneric result = search->viterbi_graph();
    delete search;
    return result;
}


FSTGeneric
FSTSearch(const FSTGeneric& x)
{
    fst::vector<int> order;
    if (!FSTTopologicalSort(order, x)) {
        FSTFatal("FSTSearch not implemented for non-topologically sortable FST -- aborting");
    }
    int N = order.size();

    FSTSemiring semiring = x->semiring();

    fst::vector<float> score(N, semiring->zero);
    fst::vector<FSTArc*> arc(N, (FSTArc*)0);

    for (fst::vector<int>::reverse_iterator i = order.rbegin(); i != order.rend(); i++) {
        int p = *i;
        score[p] = x->final_weight(p);
        for (FSTArcIter a = x->arcs_out(p); a.valid(); a++) {
            int q = a->next;
            if (q == p && semiring->is_better(a->weight, semiring->one)) {
                FSTFatal("FSTSearch: self loop weight better than %g -- aborting",
                         semiring->one);
            }
            float w = semiring->times(score[q], a->weight);
            if (semiring->is_better(w, score[p])) {
                score[p] = w;
                FSTArc*& arc_p = arc[p];
                if (!arc_p) {
                    arc_p = new FSTArc(p, q, a->input, a->output, a->weight);
                } else {
                    arc_p->next = q;
                    arc_p->input = a->input;
                    arc_p->output = a->output;
                    arc_p->weight = a->weight;
                }
            }
        }
    }

    // backtrace

    FSTGeneric result = FSTBasic();
    int r = result->add_state();
    result->set_initial(r);

    int p = x->initial();
    while (arc[p]) {
        result->add_state();
        FSTArc* arc_p = arc[p];
        result->add_arc(r, r + 1, arc_p->input, arc_p->output, arc_p->weight);
        r++;
        p = arc_p->next;
    }
    result->set_final(r, score[p]);

    for (p = 0; p < N; p++)
        delete arc[p];

    return result;
}

// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 79
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
