////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

#include "fst/FST_i.h"
#include "fst/FSTSemirings.h"

static inline float
log_plus(float x, float y)
{
    if (x > y)
        return y - log(1 + exp(y - x));
    else
        return x - log(1 + exp(x - y));
}

static inline float
absval(float x)
{
    return (x < 0) ? -x : x;
}

///////////////////////////////////////////////////////////////////////////////

FSTSemiringBase::FSTSemiringBase(float z, float o, float d, const char* n)
  : zero(z), one(o), delta(d), name(n)
{
}

FSTSemiringBase::~FSTSemiringBase()
{
}

///////////////////////////////////////////////////////////////////////////////

FSTMinPlusSemiring::FSTMinPlusSemiring()
    : FSTSemiringBase(1e30, 0, 1e-3, "MinPlus")
{
}

FSTMinPlusSemiring::~FSTMinPlusSemiring()
{
}

float
FSTMinPlusSemiring::times(float x, float y) const
{
    return x + y;
}

float
FSTMinPlusSemiring::times(float x, float y, float z) const
{
    return x + y + z;
}

float
FSTMinPlusSemiring::plus(float x, float y) const
{
    return _min(x, y);
}

float
FSTMinPlusSemiring::plus(float x, float y, float z) const
{
    return _min(_min(x, y), z);
}

float
FSTMinPlusSemiring::closure(float x) const
{
    return one;
}

float
FSTMinPlusSemiring::divide(float x, float y) const
{
    return x - y;
}

float
FSTMinPlusSemiring::invert(float x) const
{
    return (x != zero) ? -x : x;
}

float
FSTMinPlusSemiring::better(float x, float y) const
{
    return _min(x, y);
}

bool
FSTMinPlusSemiring::is_better(float x, float y) const
{
    return x < y - delta;
}

bool
FSTMinPlusSemiring::is_better(float x, float y, float delta) const
{
    return x < y - delta;
}

bool
FSTMinPlusSemiring::is_equal(float x, float y) const
{
    return absval(x - y) < delta;
}

bool
FSTMinPlusSemiring::is_equal(float x, float y, float delta) const
{
    return absval(x - y) < delta;
}

float
FSTMinPlusSemiring::convert(float x) const
{
    return is_better(x, zero) ? exp(-x) : 0;
}

float
FSTMinPlusSemiring::unconvert(float x) const
{
    return (x > 0) ? -log(x) : zero;
}

bool
FSTMinPlusSemiring::idempotent() const
{
    return true;
}

///////////////////////////////////////////////////////////////////////////////

FSTMaxPlusSemiring::FSTMaxPlusSemiring()
    : FSTSemiringBase(-1e30, 0, 1e-3, "MaxPlus")
{
}

FSTMaxPlusSemiring::~FSTMaxPlusSemiring()
{
}

float
FSTMaxPlusSemiring::times(float x, float y) const
{
    return x + y;
}

float
FSTMaxPlusSemiring::times(float x, float y, float z) const
{
    return x + y + z;
}

float
FSTMaxPlusSemiring::plus(float x, float y) const
{
    return _max(x, y);
}

float
FSTMaxPlusSemiring::plus(float x, float y, float z) const
{
    return _max(_max(x, y), z);
}

float
FSTMaxPlusSemiring::closure(float x) const
{
    return one;
}

float
FSTMaxPlusSemiring::divide(float x, float y) const
{
    return x - y;
}

float
FSTMaxPlusSemiring::invert(float x) const
{
    return (x != zero) ? -x : x;
}

float
FSTMaxPlusSemiring::better(float x, float y) const
{
    return _max(x, y);
}

bool
FSTMaxPlusSemiring::is_better(float x, float y) const
{
    return x > y + delta;
}

bool
FSTMaxPlusSemiring::is_better(float x, float y, float delta) const
{
    return x > y + delta;
}

bool
FSTMaxPlusSemiring::is_equal(float x, float y) const
{
    return absval(x - y) < delta;
}

bool
FSTMaxPlusSemiring::is_equal(float x, float y, float delta) const
{
    return absval(x - y) < delta;
}

float
FSTMaxPlusSemiring::convert(float x) const
{
    return is_better(x, zero) ? exp(x) : 0;
}

float
FSTMaxPlusSemiring::unconvert(float x) const
{
    return (x > 0) ? log(x) : zero;
}

bool
FSTMaxPlusSemiring::idempotent() const
{
    return true;
}

///////////////////////////////////////////////////////////////////////////////

FSTPlusTimesSemiring::FSTPlusTimesSemiring()
    : FSTSemiringBase(0, 1, 1e-6, "PlusTimes")
{
}

FSTPlusTimesSemiring::~FSTPlusTimesSemiring()
{
}

float
FSTPlusTimesSemiring::times(float x, float y) const
{
    return x * y;
}

float
FSTPlusTimesSemiring::times(float x, float y, float z) const
{
    return x * y * z;
}

float
FSTPlusTimesSemiring::plus(float x, float y) const
{
    return x + y;
}

float
FSTPlusTimesSemiring::plus(float x, float y, float z) const
{
    return x + y + z;
}

float
FSTPlusTimesSemiring::closure(float x) const
{
    return 1 / (1 - x);
}

float
FSTPlusTimesSemiring::divide(float x, float y) const
{
    return x / y;
}

float
FSTPlusTimesSemiring::invert(float x) const
{
    return (x != zero) ? 1 / x : x;
}

float
FSTPlusTimesSemiring::better(float x, float y) const
{
    return _max(x, y);
}

bool
FSTPlusTimesSemiring::is_better(float x, float y) const
{
    return x > y + delta;
}

bool
FSTPlusTimesSemiring::is_better(float x, float y, float delta) const
{
    return x > y + delta;
}

bool
FSTPlusTimesSemiring::is_equal(float x, float y) const
{
    return absval(x - y) < delta;
}

bool
FSTPlusTimesSemiring::is_equal(float x, float y, float delta) const
{
    return absval(x - y) < delta;
}

float
FSTPlusTimesSemiring::convert(float x) const
{
    return x;
}

float
FSTPlusTimesSemiring::unconvert(float x) const
{
    return x;
}

bool
FSTPlusTimesSemiring::idempotent() const
{
    return false;
}

///////////////////////////////////////////////////////////////////////////////

FSTPlusTimesLogSemiring::FSTPlusTimesLogSemiring()
    : FSTSemiringBase(1e30, 0, 1e-3, "PlusTimesLog")
{
}

FSTPlusTimesLogSemiring::~FSTPlusTimesLogSemiring()
{
}

float
FSTPlusTimesLogSemiring::times(float x, float y) const
{
    return x + y;
}

float
FSTPlusTimesLogSemiring::times(float x, float y, float z) const
{
    return x + y + z;
}

float
FSTPlusTimesLogSemiring::plus(float x, float y) const
{
    return log_plus(x, y);
}

float
FSTPlusTimesLogSemiring::plus(float x, float y, float z) const
{
    return log_plus(log_plus(x, y), z);
}

float
FSTPlusTimesLogSemiring::closure(float x) const
{
    return -log(1 - exp(-x));
}

float
FSTPlusTimesLogSemiring::divide(float x, float y) const
{
    return x - y;
}

float
FSTPlusTimesLogSemiring::invert(float x) const
{
    return (x != zero) ? -x : x;
}

float
FSTPlusTimesLogSemiring::better(float x, float y) const
{
    return _min(x, y);
}

bool
FSTPlusTimesLogSemiring::is_better(float x, float y) const
{
    return x < y - delta;
}

bool
FSTPlusTimesLogSemiring::is_better(float x, float y, float delta) const
{
    return x < y - delta;
}

bool
FSTPlusTimesLogSemiring::is_equal(float x, float y) const
{
    return absval(x - y) < delta;
}

bool
FSTPlusTimesLogSemiring::is_equal(float x, float y, float delta) const
{
    return absval(x - y) < delta;
}

float
FSTPlusTimesLogSemiring::convert(float x) const
{
    return is_better(x, zero) ? exp(-x) : 0;
}

float
FSTPlusTimesLogSemiring::unconvert(float x) const
{
    return (x > 0) ? -log(x) : zero;
}

bool
FSTPlusTimesLogSemiring::idempotent() const
{
    return false;
}

///////////////////////////////////////////////////////////////////////////////

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 80
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
