////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

#include "fst/FST_i.h"

using namespace std;

typedef fst::vector<int>        IntVector;
typedef fst::set<int>           IntSet;

#ifndef NO_HASH_MAP
typedef fst::hash_map<int, int> IntIntMap;
#else
typedef fst::map<int, int> IntIntMap;
#endif

class FSTInputSetArcPredicate : public FSTArcPredicate {
    const IntSet& inputs;
public:
    FSTInputSetArcPredicate(const IntSet& inputs_)
        : inputs(inputs_) {
    }
    virtual ~FSTInputSetArcPredicate() {
    }
    virtual bool operator()(const FSTArc* arc) const {
        return inputs.count(arc->input) != 0;
    }
    
};

static bool _topological_sort_dfs(IntVector& result, IntIntMap& color,
                                  const FSTGeneric& x, int p, const FSTArcPredicate* predicate);
static FSTGeneric _relabel_states(const FSTGeneric& x, const IntVector& map_);

FSTGeneric
FSTTopologicalSort(const FSTGeneric& x)
{
    return FSTTopologicalSort(x, (FSTArcPredicate*)0);
}

FSTGeneric
FSTTopologicalSort(const FSTGeneric& x, const IntSet& inputs)
{
    FSTInputSetArcPredicate predicate(inputs);
    return FSTTopologicalSort(x, &predicate);
}

FSTGeneric
FSTTopologicalSort(const FSTGeneric& x, const FSTArcPredicate* predicate)
{
    // Build new FST with states in topological order.  However, if x is cyclic
    // (even ignoring self loops) we return an empty result.
    //
    IntVector new2old;
    if (!FSTTopologicalSort(new2old, x, predicate))
        return FSTBasic();

    // Build mapping from old states to new states.
    //
    int N = new2old.size();
    IntVector old2new(N);
    for (int p = 0; p < N; p++)
    old2new[new2old[p]] = p;

    // Relabel states to form result.
    //
    return _relabel_states(x, old2new);

}

bool
FSTTopologicalSortFromInitial(fst::vector<int>& result, const FSTGeneric& x)
{
    // Perform topological sort, returning state order in result.  However, if
    // x is cyclic (even ignoring self loops) we return false and result is
    // empty.
    //
    result.clear();
    x->force();
#ifndef NO_HASH_MAP
    IntIntMap color(x->n_states());
#else
    IntIntMap color;
#endif
    if (x->initial() < 0
        || !_topological_sort_dfs(result, color, x, x->initial(), (FSTArcPredicate*)0)) {
        result.clear();
        return false;
    }
    reverse(result.begin(), result.end());
    //cout << "debug: " << result.size() << '\n';
    return true;
}

bool
FSTTopologicalSort(fst::vector<int>& result, const FSTGeneric& x)
{
    return FSTTopologicalSort(result, x, (FSTArcPredicate*)0);
}

bool
FSTTopologicalSort(fst::vector<int>& result, const FSTGeneric& x, const IntSet& inputs)
{
    FSTInputSetArcPredicate predicate(inputs);
    return FSTTopologicalSort(result, x, &predicate);
}

bool
FSTTopologicalSort(fst::vector<int>& result, const FSTGeneric& x,
                   const FSTArcPredicate* predicate)
{
    // Perform topological sort, returning state order in result.  However, if
    // x is cyclic (even ignoring self loops) we return false and result is
    // empty.
    //
    result.clear();
    x->force();
#ifndef NO_HASH_MAP
    IntIntMap color(x->n_states());
#else
    IntIntMap color;
#endif
    for (int p = x->n_states() - 1; p >= 0; --p) {
	if (!_topological_sort_dfs(result, color, x, p, predicate)) {
            result.clear();
            return false;
        }
    }
    reverse(result.begin(), result.end());
    return true;
}

static bool
_topological_sort_dfs(IntVector& result, IntIntMap& color,
                      const FSTGeneric& x, int p, const FSTArcPredicate* predicate)
{
    static const int white = 0, gray = 1, black = 2;

    if (color[p] == black)
        return true;

    typedef pair<int, FSTArcIter> StateArcs;
    stack<StateArcs> todo;

    color[p] = gray;
    todo.push(StateArcs(p, x->arcs_out(p)));
    while (!todo.empty()) {
        StateArcs& top = todo.top();
        FSTArcIter& a = top.second;
        if (a.valid()) {
            if (a->next != a->prev && (!predicate || (*predicate)(a))) {
                p = a->next;
                int& p_color = color[p];
                if (p_color == gray) {
                    // cycle detected
                    return false;
                } else if (p_color == white) {
                    // unvisited
                    p_color = gray;
                    todo.push(StateArcs(p, x->arcs_out(p)));
                }
            }
            ++a;
        } else {
            color[top.first] = black;
            result.push_back(top.first);
            todo.pop();
        }
    }
    return true;
}

static FSTGeneric
_relabel_states(const FSTGeneric& x, const IntVector& map_)
{
    // Assumes x has already been forced.
    FSTGeneric y = FSTBasic(x->semiring());
    y->add_states(x->n_states());
    y->set_initial(map_[x->initial()]);
    for (FSTStateIter si = x->states(); si.valid(); si++) {
    int o = *si;
    int n = map_[o];
    y->set_final(n, x->final_weight(o));
    y->set_potential(n, x->potential(o));
    for (FSTArcIter ai = x->arcs_out(o); ai.valid(); ai++) {
        y->add_arc(n, map_[ai->next],
              ai->input, ai->output, ai->weight);
    }
    }
    return y;
}

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 79
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
