////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

#include "fst/FST_i.h"

class FSTUnionRep
    : public FSTGenericInt {

public: // methods

    FSTUnionRep(const fst::vector<FSTGeneric>& fsts,
                const fst::vector<float>& weights,
                const fst::vector<int>& input_tags,
                const fst::vector<int>& output_tags);
    virtual ~FSTUnionRep();
    virtual const fst::set<int>& input_alphabet() const;
    virtual const fst::set<int>& output_alphabet() const;
    virtual bool is_final(int p) const;
    virtual float final_weight(int p) const;
    virtual float potential(int p) const;
    virtual FSTArcIter arcs_out(int p, FSTArcOrder order=FSTNone) const;
    virtual FSTArcIter arcs_out_with_input(int p, int i,
                                           FSTArcMode m=FSTEqual) const;
    virtual FSTArcIter arcs_out_with_output(int p, int o,
                                            FSTArcMode m=FSTEqual) const;
    virtual void force() const;
    virtual void reset() const;
    virtual FSTGeneric reload(bool& flag);

protected: // types

    struct _State;
    typedef fst::vector<_State*> _StatePtrVector;
    typedef fst::map<_State*, int, ptr_less<_State*> > _StateIntMap;

    struct _Tags {
        int ii, io, fi, fo;
    };

protected: // members

    fst::vector<FSTGeneric> _fsts;
    fst::vector<_Tags> _tags;
    fst::vector<float> _weights;
    int _n_fsts;
    mutable _StatePtrVector _states;
    mutable _StateIntMap _map;
    fst::set<int> _input_alphabet;
    fst::set<int> _output_alphabet;
    fst::set<int> _tags_alphabet;
    int _final;

protected: // methods

    int  _find_or_add_state(int fst, int state) const;
    inline void _add_arcs(fst::vector<FSTArc*>& arcs, int p,
                          int fst, FSTArcIter a) const;
    inline void _add_arc(fst::vector<FSTArc*>& arcs, int p,
                         int fst, int state, int i, int o, float w) const;
    void _setup_alphabets();
    void _clean() const;
};

FSTGeneric
FSTUnion(const FSTGeneric& x, const FSTGeneric& y)
{
    fst::vector<FSTGeneric> v;
    v.push_back(x);
    v.push_back(y);
    fst::vector<int> no_tags;
    fst::vector<float> no_weights;
    return FSTGeneric(new FSTUnionRep(v, no_weights, no_tags, no_tags));
}

FSTGeneric
FSTUnion(const FSTGeneric& x, const FSTGeneric& y, const FSTGeneric& z)
{
    fst::vector<FSTGeneric> v;
    v.push_back(x);
    v.push_back(y);
    v.push_back(z);
    fst::vector<int> no_tags;
    fst::vector<float> no_weights;
    return FSTGeneric(new FSTUnionRep(v, no_weights, no_tags, no_tags));
}

FSTGeneric
FSTUnion(const fst::vector<FSTGeneric>& fsts)
{
    fst::vector<int> no_tags;
    fst::vector<float> no_weights;
    return FSTGeneric(new FSTUnionRep(fsts, no_weights, no_tags, no_tags));
}

FSTGeneric
FSTUnion(const fst::vector<FSTGeneric>& fsts,
         const fst::vector<float>& weights)
{
    fst::vector<int> no_tags;
    return FSTGeneric(new FSTUnionRep(fsts, weights, no_tags, no_tags));
}

FSTGeneric
FSTUnion(const fst::vector<FSTGeneric>& fsts,
         const fst::vector<float>& weights,
         const fst::vector<int>& input_tags,
         const fst::vector<int>& output_tags)
{
    return FSTGeneric(new FSTUnionRep(fsts, weights, input_tags, output_tags));
}

FSTGeneric
FSTUnion(const fst::vector<FSTGeneric>& fsts,
         const fst::vector<int>& input_tags,
         const fst::vector<int>& output_tags)
{
    fst::vector<float> no_weights;
    return FSTGeneric(new FSTUnionRep(fsts, no_weights, input_tags, output_tags));
}

struct FSTUnionRep::_State {
    int fst;
    int state;
    _State(int i, int s) : fst(i), state(s) {}
    bool operator < (const _State& x) {
        return FST_LESS3(fst, x.fst,
                         FST_LESS2(state, x.state));
    }
    FSTAllocator_new_delete(_State);
};

static void
_build_tags(int tag, int& initial, int& final)
{
    if (tag) {

        string is = "<";
        is += FSTstring_map[tag];
        is += '>';
        initial = FSTstring_map[is];

        string fs = "</";
        fs += FSTstring_map[tag];
        fs += '>';
        final = FSTstring_map[fs];

    } else {
        initial = final = 0;
    }
}

FSTUnionRep::FSTUnionRep(const fst::vector<FSTGeneric>& fsts,
                         const fst::vector<float>& weights,
                         const fst::vector<int>& input_tags,
                         const fst::vector<int>& output_tags)
    : _fsts(fsts), _n_fsts(_fsts.size())
{
    _class_name = "FSTUnion";

    // Sanity check on vector dimensions.
    //
    assert(input_tags.size() == output_tags.size());
    assert(input_tags.size() == 0 || input_tags.size() == fsts.size());

    // Set up tags.
    //
    for (int i = 0; i < int(input_tags.size()); i++) {
        _Tags tags;
        _build_tags(input_tags[i], tags.ii, tags.fi);
        _build_tags(output_tags[i], tags.io, tags.fo);
        _tags.push_back(tags);
        _tags_alphabet.insert(tags.ii);
        _tags_alphabet.insert(tags.fi);
        _tags_alphabet.insert(tags.io);
        _tags_alphabet.insert(tags.fo);
    }
    assert(_tags.size() == input_tags.size());

    _setup_alphabets();

    // Set up weights.
    //
    _weights = weights;

    _initial = _find_or_add_state(-1, -1);
    _final = _tags.empty() ? -1 : _find_or_add_state(-2, -2);
}

FSTUnionRep::~FSTUnionRep()
{
    _clean();
}

void
FSTUnionRep::_setup_alphabets()
{
    // Make sure input semirings are identical and set up alphabets.
    //
    _input_alphabet.clear();
    _output_alphabet.clear();
    fst::vector<FSTGeneric>::const_iterator i = _fsts.begin();
    if (i != _fsts.end()) {
        _semiring = (*i)->semiring();
        for (; i != _fsts.end(); i++) {
            if (_semiring != (*i)->semiring())
                FSTFatal("can't take union of different semirings: %s and %s",
                      _semiring->name, (*i)->semiring()->name);
            const fst::set<int>& fsts = (*i)->input_alphabet();
            _input_alphabet.insert(fsts.begin(), fsts.end());
            const fst::set<int>& outputs = (*i)->output_alphabet();
            _input_alphabet.insert(outputs.begin(), outputs.end());
        }
    }
    _input_alphabet.insert(0);
    _output_alphabet.insert(0);
    _input_alphabet.insert(_tags_alphabet.begin(), _tags_alphabet.end());
}

FSTGeneric
FSTUnionRep::reload(bool& flag)
{
    for (fst::vector<FSTGeneric>::iterator i = _fsts.begin();
         i != _fsts.end(); i++) {
        if ((*i).reload()) flag = true;
    }
    if (flag)
        _setup_alphabets();
    reset();
    return FSTGeneric(this);
}

void
FSTUnionRep::reset() const
{

    // Reset all input FSTs.
    //
    for (fst::vector<FSTGeneric>::const_iterator i = _fsts.begin();
         i != _fsts.end(); i++) {
        (*i)->reset();
    }

    _clean();
    _find_or_add_state(-1, -1);
    if (!_tags.empty())
        _find_or_add_state(-2, -2);
}

void
FSTUnionRep::_clean() const
{
    for (_StatePtrVector::iterator j = _states.begin();
         j != _states.end(); j++) {
        delete *j;
    }
    _states.clear();
    _map.clear();
    _n_states = 0;
}

const fst::set<int>&
FSTUnionRep::input_alphabet() const
{
    return _input_alphabet;
}

const fst::set<int>&
FSTUnionRep::output_alphabet() const
{
    return _output_alphabet;
}

bool
FSTUnionRep::is_final(int p) const
{
    FST_ASSERT_VALID_STATE(p);
    _State* state = _states[p];
    if (p == _final) {
        return true;
    } else if (p == _initial || !_fsts[state->fst]->is_final(state->state)) {
        return false;
    } else {
        return _tags.empty();
    }
}

float
FSTUnionRep::final_weight(int p) const
{
    FST_ASSERT_VALID_STATE(p);
    _State* state = _states[p];
    if (p == _final) {
        return _semiring->one;
    } else if (p == _initial || !_fsts[state->fst]->is_final(state->state)) {
        return _semiring->zero;
    } else {
        return _tags.empty()
            ? _fsts[state->fst]->final_weight(state->state)
            : _semiring->zero;
    }
}

float
FSTUnionRep::potential(int p) const
{
    FST_ASSERT_VALID_STATE(p);
    if (p == _initial) {
        float result = _semiring->zero;
        for (int i = 0; i < int(_fsts.size()); i++) {
            const FSTGeneric& fst = _fsts[i];
            if (fst->initial() >= 0) {
                float w = i < int(_weights.size())
                    ? _semiring->times(_weights[i], fst->potential(fst->initial()))
                    : fst->potential(fst->initial());
                result = _semiring->better(result, w);
            }
        }
        return result;
    } else if (p == _final) {
        return _semiring->one;
    } else {
        _State* pp = _states[p];
        return _fsts[pp->fst]->potential(pp->state);
    }
}

inline void
FSTUnionRep::_add_arcs(fst::vector<FSTArc*>& arcs, int p,
                       int fst, FSTArcIter a) const
{
    //arcs.reserve(arcs.size() + a.n());
    for (; a.valid(); a++)
        arcs.push_back(new FSTArc(p, _find_or_add_state(fst, a->next),
                               a->input, a->output, a->weight));
}

inline void
FSTUnionRep::_add_arc(fst::vector<FSTArc*>& arcs, int p, int fst, int state,
                      int i, int o, float w) const
{
    arcs.push_back(new FSTArc(p, _find_or_add_state(fst, state),
                              i, o, w));
}

FSTArcIter
FSTUnionRep::arcs_out(int p, FSTArcOrder order) const
{
    FST_ASSERT_VALID_STATE(p);
    FSTArcIterData* arc_data = new FSTArcIterData;
    fst::vector<FSTArc*>& arcs = arc_data->arcs;
    if (p == _initial) {
        arcs.reserve(_n_fsts);
        for (int fst = 0; fst < _n_fsts; fst++) {
            int q = _fsts[fst]->initial();
            float w = fst < int(_weights.size()) ? _weights[fst] : _semiring->one;
            if (q >= 0) {
                if (_tags.empty()) {
                    _add_arc(arcs, p, fst, q, 0, 0, w);
                } else {
                    const _Tags& tags = _tags[fst];
                    _add_arc(arcs, p, fst, q, tags.ii, tags.io, w);
                }
            }
        }
        if (!_tags.empty() && order != FSTNone)
            _sort_arcs(arcs, order);
    } else if (p != _final) {
        int fst = _states[p]->fst;
        int state = _states[p]->state;
        _add_arcs(arcs, p, fst, _fsts[fst]->arcs_out(state, order));
        if (!_tags.empty() && _fsts[fst]->is_final(state)) {
            const _Tags& tags = _tags[fst];
            _add_arc(arcs, p, -2, -2, tags.fi, tags.fo,
                     _fsts[fst]->final_weight(state));
            if (order != FSTNone)
                _sort_arcs(arcs, order);
        }
    }
    return FSTArcIter(arc_data);
}

inline bool
_keep(int x, int key, FSTArcMode mode)
{
    if (mode == FSTEqual)
        return x == key;
    else if (mode == FSTGreaterEqual)
        return x <= key;
    else
        return x >= key;
}

FSTArcIter FSTUnionRep::arcs_out_with_input(int p, int i, FSTArcMode m) const
{
    FST_ASSERT_VALID_STATE(p);
    FSTArcIterData* arc_data = new FSTArcIterData;
    fst::vector<FSTArc*>& arcs = arc_data->arcs;
    if (p == _initial) {
        for (int fst = 0; fst < _n_fsts; fst++) {
            int q = _fsts[fst]->initial();
            float w = fst < int(_weights.size()) ? _weights[fst] : _semiring->one;
            if (q >= 0) {
                if (_tags.empty()) {
                    if (_keep(0, i, m))
                        _add_arc(arcs, p, fst, q, 0, 0, w);
                } else {
                    const _Tags& tags = _tags[fst];
                    if (_keep(tags.ii, i, m))
                        _add_arc(arcs, p, fst, q, tags.ii, tags.io, w);
                }
            }
        }
    } else if (p != _final) {
        int fst = _states[p]->fst;
        int state = _states[p]->state;
        _add_arcs(arcs, p, fst, _fsts[fst]->arcs_out_with_input(state, i, m));
        if (!_tags.empty() && _fsts[fst]->is_final(state)) {
            const _Tags& tags = _tags[fst];
            if (_keep(tags.fi, i, m))
                _add_arc(arcs, p, -2, -2, tags.fi, tags.fo,
                         _fsts[fst]->final_weight(state));
        }
    }
    return FSTArcIter(arc_data);
}

FSTArcIter FSTUnionRep::arcs_out_with_output(int p, int o, FSTArcMode m) const
{
    FST_ASSERT_VALID_STATE(p);
    FSTArcIterData* arc_data = new FSTArcIterData;
    fst::vector<FSTArc*>& arcs = arc_data->arcs;
    if (p == _initial) {
        for (int fst = 0; fst < _n_fsts; fst++) {
            int q = _fsts[fst]->initial();
            float w = fst < int(_weights.size()) ? _weights[fst] : _semiring->one;
            if (q >= 0) {
                if (_tags.empty()) {
                    if (_keep(0, o, m))
                        _add_arc(arcs, p, fst, q, 0, 0, w);
                } else {
                    const _Tags& tags = _tags[fst];
                    if (_keep(tags.io, o, m))
                        _add_arc(arcs, p, fst, q, tags.ii, tags.io, w);
                }
            }
        }
    } else if (p != _final) {
        int fst = _states[p]->fst;
        int state = _states[p]->state;
        _add_arcs(arcs, p, fst, _fsts[fst]->arcs_out_with_output(state, o, m));
        if (!_tags.empty() && _fsts[fst]->is_final(state)) {
            const _Tags& tags = _tags[fst];
            if (_keep(tags.fo, o, m))
                _add_arc(arcs, p, -2, -2, tags.fi, tags.fo,
                         _fsts[fst]->final_weight(state));
        }
    }
    return FSTArcIter(arc_data);
}

void FSTUnionRep::force() const
{
    // Create all reachable states.
    for (int f = 0; f < _n_fsts; f++) {
        const FSTGeneric& fst = _fsts[f];
        fst->force();
        int N = fst->n_states();
        for (int p = 0; p < N; p++) {
            _find_or_add_state(f, p);
        }
    }
}

int FSTUnionRep::_find_or_add_state(int fst, int state) const
{
    int p;
    _State* pp = new _State(fst, state);
    _StateIntMap::iterator i = _map.find(pp);
    if (i == _map.end()) {
        p = _n_states++;
        _states.push_back(pp);
        _map[pp] = p;
    } else {
        p = (*i).second;
        delete pp;
    }
    return p;
}

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 80
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
