////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2007, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

#include "fst/FST_i.h"
#include <fstream>

extern "C" {
#include "fst/sysdep.h"
}

#ifdef _MSC_VER
#include <fcntl.h>
#endif

using namespace std;

extern time_t file_mtime(const char* filename);

FSTStringMap FSTstring_map;

static FSTMinPlusSemiring _min_plus;
static FSTMaxPlusSemiring _max_plus;
static FSTPlusTimesSemiring _plus_times;
static FSTPlusTimesLogSemiring _plus_times_log;

FSTSemiring FSTMinPlus      = &_min_plus;
FSTSemiring FSTMaxPlus      = &_max_plus;
FSTSemiring FSTPlusTimes    = &_plus_times;
FSTSemiring FSTPlusTimesLog = &_plus_times_log;

#ifdef _MSC_VER
#include <fcntl.h>
istream&
FSTBinaryCin()
{
    _setmode(_fileno(stdin), _O_BINARY);
    return cin;
}
ostream&
FSTBinaryCout()
{
    _setmode(_fileno(stdout), _O_BINARY);
    return cout;
}
#else
istream&
FSTBinaryCin()
{
    return cin;
}
ostream&
FSTBinaryCout()
{
    return cout;
}
#endif

void
FSTLabels(const char* filename)
{
    string label;
    ifstream in(filename);
    while (!in.eof() && in.good()) {
        in >> label;
        FSTstring_map[label];
    }
}

static FSTGeneric
_FSTLoad(istream& is, const char* filename)
{
    FSTGeneric result;

    time_t mtime = 0;
    if (filename && filename[0])
        mtime = file_mtime(filename);

    char c;
    is >> c;
    if (c != '#') {
        is.putback(c);
        result = FSTLoadBasic(is, "");
    } else {
        string line;
        getline(is, line);
        if (is.fail())
            FSTFatal("error reading from file %s (no header?)", filename);
        const char* linec = line.c_str();
        char type[256];
        int type_length;
        sscanf(linec, "%s%n", type, &type_length);

        if (strcmp(type, "FSTBasic") == 0) {
            result = FSTLoadBasic(is, linec);
        } else if (strcmp(type, "FSTBinary") == 0) {
            result = FSTLoadBinary(is, linec);
        } else if (strcmp(type, "FSTDynamic") == 0) {
            result = FSTLoadDynamic(is, linec);
        } else if (strcmp(type, "FSTMmap") == 0) {
            result = FSTLoadMmap(is, linec, filename);
#if 0
        } else if (strcmp(type, "FSTNgram") == 0) {
            result = FSTLoadNgram(is, linec);
#endif
        } else if (strcmp(type, "FSTNgram2") == 0) {
            result = FSTLoadNgram2(is, linec);
        } else if (strcmp(type, "FSTNgram3") == 0) {
            result = FSTLoadNgram3(is, linec);
        } else if (strcmp(type, "FSTRtn") == 0) {
            result = FSTLoadRtn(is, linec);
        } else if (strcmp(type, "FSTExpression") == 0) {
            result = FSTExpression(linec + type_length);
        } else {
            FSTFatal("don't understand header type %s", type);
            abort();
        }
    }

    if (mtime)
        result->enable_reload(filename, mtime);

    return result;
}

FSTGeneric
FSTLoad(const string& file_or_exp)
{
    return FSTLoad(file_or_exp.c_str());
}

FSTGeneric
FSTLoad(const char* file_or_exp)
{
    FSTGeneric result;

    // First, try to interpret file_or_exp as an expression.
    //
    string parse_error;
    if (FSTExpression(file_or_exp, result, parse_error))
        return result;

    // Second, try to load the file.

    if (file_or_exp && file_or_exp[0] &&
        (file_or_exp[0] != '-' || file_or_exp[1])) {
#ifdef _MSC_VER
        ifstream is(file_or_exp, ios_base::binary);
#else
        ifstream is(file_or_exp);
#endif
        if (!is) {
            FSTFatal("can't read file %s or parse as expression\n%s",
                     file_or_exp, parse_error.c_str());
        }
        return _FSTLoad(is, file_or_exp);
    } else {
        return _FSTLoad(FSTBinaryCin(), file_or_exp);
    }
}

FSTGeneric
FSTLoadFile(const string& filename)
{
    return FSTLoad(filename.c_str());
}

FSTGeneric
FSTLoadFile(const char* filename)
{
#ifdef _MSC_VER
    ifstream is(filename, ios_base::binary);
#else
    ifstream is(filename);
#endif
    if (!is) {
        FSTFatal("can't read file %s", filename);
    }
    return _FSTLoad(is, filename);
}

FSTGeneric
FSTCopy(const FSTGeneric& x)
{
    return FSTBasic(x);
}

FSTGeneric
FSTMutableCopy(const FSTGeneric& x)
{
    return FSTBasic(x);
}

FSTGeneric
FSTOptimize(const FSTGeneric& x, int degree, float eps)
{
    FSTGeneric y = FSTBasic(FSTRemoveEpsilons(x));
    y = FSTPushWeights(y);
    y = FSTBasic(FSTDeterminize(y, degree, false, eps));
    y = FSTMinimize(y, eps);
    return y;
}

FSTGeneric
FSTOptimize2(const FSTGeneric& x, int degree, float eps)
{
    FSTGeneric y = FSTPushWeights(x);
    y = FSTBasic(FSTRemoveEpsilons(x));
    y = FSTBasic(FSTDeterminize(y, degree, false, eps));
    y = FSTMinimize(y, eps);
    return y;
}

FSTGeneric
FSTOptimizeA(const FSTGeneric& x, int degree, float eps)
{
    FSTGeneric y = FSTBasic(FSTRemoveEpsilonsA(x));
    y = FSTPushWeights(y);
    y = FSTBasic(FSTDeterminize(y, degree, true, eps));
    y = FSTMinimize(y, eps);
    return y;
}

FSTGeneric
FSTOptimize2A(const FSTGeneric& x, int degree, float eps)
{
    FSTGeneric y = FSTPushWeights(x);
    y = FSTBasic(FSTRemoveEpsilonsA(x));
    y = FSTBasic(FSTDeterminize(y, degree, true, eps));
    y = FSTMinimize(y, eps);
    return y;
}

// Note that if x is mutable, x will be trimmed in-place.
//
FSTGeneric
FSTTrim(FSTGeneric x)
{
    if (!x->is_mutable())
        x = FSTMutableCopy(x);
    x->trim();
    return x;
}

FSTNgramInt::~FSTNgramInt()
{
}

const FSTNgramInt*
FSTNgramCast(const FSTGeneric& x)
{
    return dynamic_cast<const FSTNgramInt*>(x.pointer());
}

FSTGeneric
FSTNgramDeterminize(FSTGeneric x)
{
    FSTNgramInt* xx = dynamic_cast<FSTNgramInt*>(x.pointer());
    if (xx) {
        xx->determinize();
        return x;
    } else {
        return FSTBackoff(x, false);
    }
}

FSTGeneric
FSTNgramInvert(FSTGeneric x)
{
    FSTNgramInt* xx = dynamic_cast<FSTNgramInt*>(x.pointer());
    if (xx) {
        xx->invert();
        return x;
    } else {
        return FSTBackoff(x, true);
    }
}

void
FSTNgramSubtract(FSTGeneric x, const FSTGeneric& y)
{
    FSTNgramInt* xx = dynamic_cast<FSTNgramInt*>(x.pointer());
    if (!xx)
        FSTFatal("FSTNgram2Subtract called with non-FSTNgram first argument");
    const FSTNgramInt* yy = dynamic_cast<const FSTNgramInt*>(y.pointer());
    if (!yy)
        FSTFatal("FSTNgram2Subtract called with non-FSTNgram second argument");
    xx->subtract(yy);
}

void
FSTNgramDumpPacked(const FSTGeneric& x, const char* vocab_file, const char* out_file,
                   int weight_width)
{
    dynamic_cast<const FSTNgramInt*>(x.pointer())->dump_pngram(vocab_file, out_file, weight_width);
}

void
FSTNgramDumpPacked(const FSTGeneric& x, const fst::list<string>& vocab_list, FILE* fp,
                   int weight_width)
{
    dynamic_cast<const FSTNgramInt*>(x.pointer())->dump_pngram(vocab_list, fp, weight_width);
}

void
FSTNgramAddClassedWord(FSTGeneric& ngram, int word, int word_class, float weight)
{
    FSTNgramInt* fst = dynamic_cast<FSTNgramInt*>(ngram.pointer());
    if (fst) {
        fst->add_classed_word(word, word_class, weight);
    }
}

//
// Fatal and warn functions.  Eventually, these will be modified so that their
// output can be redirected.
//

void
FSTFatal(const char* fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);
    FSTFatal(fmt, ap);
    va_end(ap);
}

void
FSTFatal(const char* fmt, va_list ap)
{
    fprintf(stderr, "FATAL: ");
    vfprintf(stderr, fmt, ap);
    fprintf(stderr, "\n");
    fflush(stderr);
#ifdef _WIN32
    exit(1);
#else
    abort();
#endif
}

void
FSTWarn(const char* fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);
    FSTWarn(fmt, ap);
    va_end(ap);
}

void
FSTWarn(const char* fmt, va_list ap)
{
    fprintf(stderr, "WARNING: ");
    vfprintf(stderr, fmt, ap);
    fprintf(stderr, "\n");
}

//
// Binary I/O routines, providing byte-swapping as necessary.
//

// We have byte-swapping assembly code for x86, but for some reason gcc 2.8.1
// seems to miscompile it with -O3.  Thus, we only use it for gcc newer than
// 2.8.1.  This has been tested with egcs 1.1.2 and gcc 2.95.1.

#define USE_X86_ASM (defined(__x86__) && defined(__GNUC__) && \
             (__GNUC__ >= 3 || (__GNUC__ == 2 && __GNUC_MINOR__ > 8)))

inline void
_swap2(void *p)
{
#if __LITTLE_ENDIAN__
# if USE_X86_ASM
    __asm__ __volatile__
    ("movw  (%0),%%ax \n\t"
     "xchgb %%al,%%ah \n\t"
     "movw  %%ax,(%0) \n\t"
     : : "r" (p) : "eax", "memory");
# else
    char t, *c = (char *) p;
    t = c[0]; c[0] = c[1]; c[1] = t;
# endif
#endif
}

inline void
_swap4(void *p)
{
#if __LITTLE_ENDIAN__
# if USE_X86_ASM
    __asm__ __volatile__
    ("movl  (%0),%%eax \n\t"
     "bswap %%eax      \n\t"
     "movl  %%eax,(%0) \n\t"
     : : "r" (p) : "eax", "memory");
# else
    char t, *c = (char *) p;
    t = c[0]; c[0] = c[3]; c[3] = t;
    t = c[1]; c[1] = c[2]; c[2] = t;
# endif
#endif
}

inline void
_swap8(void *p)
{
#if __LITTLE_ENDIAN__
# if USE_X86_ASM
    __asm__ __volatile__
    ("movl  (%0),%%eax  \n\t"
     "bswap %%eax       \n\t"
     "xchgl %%eax,4(%0) \n\t"
     "bswap %%eax       \n\t"
     "movl  %%eax,(%0)  \n\t"
     : : "r" (p) : "eax", "memory");
# else
    char t, *c = (char *) p;
    t = c[0]; c[0] = c[7]; c[7] = t;
    t = c[1]; c[1] = c[6]; c[6] = t;
    t = c[2]; c[2] = c[5]; c[5] = t;
    t = c[3]; c[3] = c[4]; c[4] = t;
# endif
#endif
}

void
FSTDumpShort(short x, ostream& os)
{
    _swap2(&x);
    os.write((const char*)&x, 2);
}

void
FSTDumpUnsignedShort(unsigned short x, ostream& os)
{
    _swap2(&x);
    os.write((const char*)&x, 2);
}

short
FSTLoadShort(istream& is)
{
    short x;
    is.read((char*)&x, 2);
    _swap2(&x);
    return x;
}

unsigned short
FSTLoadUnsignedShort(istream& is)
{
    unsigned short x;
    is.read((char*)&x, 2);
    _swap2(&x);
    return x;
}

void
FSTDumpInt(int x, ostream& os)
{
    _swap4(&x);
    os.write((const char*)&x, 4);
}

void
FSTDumpUnsignedInt(unsigned int x, ostream& os)
{
    _swap4(&x);
    os.write((const char*)&x, 4);
}

int
FSTLoadInt(istream& is)
{
    int x;
    is.read((char*)&x, 4);
    _swap4(&x);
    return x;
}

unsigned int
FSTLoadUnsignedInt(istream& is)
{
    unsigned int x;
    is.read((char*)&x, 4);
    _swap4(&x);
    return x;
}

void
FSTDumpCompressedUnsigned(unsigned int x, ostream& os)
{
    union {
        unsigned int i;
        unsigned char c[4];
    } u;
    u.i = x;

    // put in MSB order
    _swap4(&u.i);

    if ((x & 0xffffffc0) == 0) {
        os.write((const char*)&u.c[3], 1);
    } else if ((x & 0xffffc000) == 0) {
        u.c[2] |= 0x40;
        os.write((const char*)&u.c[2], 2);
    } else if ((x & 0xffc00000) == 0) {
        u.c[1] |= 0x80;
        os.write((const char*)&u.c[1], 3);
    } else if ((x & 0xc0000000) == 0) {
        u.c[0] |= 0xc0;
        os.write((const char*)&u.c[0], 4);
    } else {
        FSTFatal("FSTDumpCompressedUnsigned: %u too large", x);
    }
    assert(!os.fail());
}

unsigned int
FSTLoadCompressedUnsigned(istream& is)
{
    union {
        unsigned int i;
        unsigned char c[4];
    } u;
    u.i = 0;

    unsigned char b;
    is.read((char *)&b, 1);
    switch (b & 0xc0) {
    case 0x00:
        u.c[3] = b;
        break;
    case 0x40:
        u.c[2] = b & 0x3f;
        is.read((char *)&u.c[3], 1);
        break;
    case 0x80:
        u.c[1] = b & 0x3f;
        is.read((char *)&u.c[2], 2);
        break;
    case 0xc0:
        u.c[0] = b & 0x3f;
        is.read((char *)&u.c[1], 3);
        break;
    }

    assert(!is.fail());

    _swap4(&u.i);

    return u.i;
}

void
FSTDumpCompressedInt(int x, ostream& os)
{
    union {
        int i;
        unsigned char c[4];
    } u;
    u.i = x;

    // put in MSB order
    _swap4(&u.i);

    if (x & 0x80000000) {
        // negative
        if ((x & 0xffffffe0) == 0xffffffe0) {
            u.c[3] &= 0x3f;
            os.write((const char*)&u.c[3], 1);
        } else if ((x & 0xffffe000) == 0xffffe000) {
            u.c[2] &= 0x7f;
            os.write((const char*)&u.c[2], 2);
        } else if ((x & 0xffe00000) == 0xffe00000) {
            u.c[1] &= 0xbf;
            os.write((const char*)&u.c[1], 3);
        } else if ((x & 0xe0000000) == 0xe0000000) {
            os.write((const char*)&u.c[0], 4);
        } else {
            FSTFatal("FSTDumpCompressedInt: %d too large", x);
        }
    } else {
        // positive
        if ((x & 0xffffffe0) == 0) {
            os.write((const char*)&u.c[3], 1);
        } else if ((x & 0xffffe000) == 0) {
            u.c[2] |= 0x40;
            os.write((const char*)&u.c[2], 2);
        } else if ((x & 0xffe00000) == 0) {
            u.c[1] |= 0x80;
            os.write((const char*)&u.c[1], 3);
        } else if ((x & 0xe0000000) == 0) {
            u.c[0] |= 0xc0;
            os.write((const char*)&u.c[0], 4);
        } else {
            FSTFatal("FSTDumpCompressedInt: %u too large", x);
        }
    }
}

int
FSTLoadCompressedInt(istream& is)
{
    union {
        int i;
        unsigned char c[4];
    } u;

    unsigned char b;
    is.read((char *)&b, 1);

    if (b & 0x20) {
        // negative
        u.i = -1;
        switch (b & 0xc0) {
        case 0x00:
            u.c[3] = b | 0xc0;
            break;
        case 0x40:
            u.c[2] = b | 0xc0;
            is.read((char *)&u.c[3], 1);
            break;
        case 0x80:
            u.c[1] = b | 0xc0;
            is.read((char *)&u.c[2], 2);
            break;
        case 0xc0:
            u.c[0] = b;
            is.read((char *)&u.c[1], 3);
            break;
        }        
    } else {
        // positive
        u.i = 0;
        switch (b & 0xc0) {
        case 0x00:
            u.c[3] = b;
            break;
        case 0x40:
            u.c[2] = b & 0x3f;
            is.read((char *)&u.c[3], 1);
            break;
        case 0x80:
            u.c[1] = b & 0x3f;
            is.read((char *)&u.c[2], 2);
            break;
        case 0xc0:
            u.c[0] = b & 0x3f;
            is.read((char *)&u.c[1], 3);
            break;
        }
    }

    _swap4(&u.i);

    return u.i;
}

void
FSTDumpFloat(float x, ostream& os)
{
    _swap4(&x);
    os.write((const char*)&x, 4);
}

float
FSTLoadFloat(istream& is)
{
    float x;
    is.read((char*)&x, 4);
    _swap4(&x);
    return x;
}

void
FSTDumpDouble(double x, ostream& os)
{
    _swap8(&x);
    os.write((const char*)&x, 8);
}

double
FSTLoadDouble(istream& is)
{
    double x;
    is.read((char*)&x, 8);
    _swap8(&x);
    return x;
}

void
FSTDumpString(const string& s, ostream& os)
{
    int l = s.size();
    FSTDumpInt(l, os);
    os.write(s.c_str(), l + 1);
}

void
FSTLoadString(istream& is, string& x)
{
    int l = FSTLoadInt(is);
    if (l >= 0) {
        x.resize(l + 1);
        is.read((char*)&x[0], l + 1);
        x.resize(l);                            // remove '\0'
    } else {
        x.resize(0);
    }
}

string
FSTLoadString(istream& is)
{
    string x;
    int l = FSTLoadInt(is);
    if (l >= 0) {
        x.resize(l + 1);
        is.read((char*)&x[0], l + 1);
        x.resize(l);                            // remove '\0'
    }
    return x;
}

void
FSTDumpChars(const char* s, ostream& os)
{
    if (s) {
    int l = strlen(s);
    FSTDumpInt(l, os);
    os.write(s, l + 1);
    } else {
    FSTDumpInt(-1, os);
    }
}

char*
FSTLoadChars(istream& is)
{
    int l = FSTLoadInt(is);
    if (l < 0)
        return 0;
    char *x = new char[l + 1];
    is.read(x, l + 1);
    return x;
}

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 80
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
