#include <vector>
#include <deque>
#include <algorithm>
#include <cmath>
#include <vector>
#include <deque>
#include "sq.h"

#include <iostream>

using namespace std;

typedef pair<float,float> FloatFloat;

class LloydMaxTrainer : public ScalarQuantizerTrainer {
 public:
    LloydMaxTrainer(unsigned nBits);
    virtual ~LloydMaxTrainer();
    virtual void train(vector<float>& d, vector<float>& r);
protected:
    unsigned nBits;
};

LloydMaxTrainer::LloydMaxTrainer(unsigned nBits)
{
    this->nBits = nBits;
}

LloydMaxTrainer::~LloydMaxTrainer()
{
}

static inline double
square(double x)
{
    return x * x;
}

static double
updateR(vector<float>& r, const vector<float>& d, const deque<FloatFloat>& data)
{
    double distortion_sum = 0;
    float distortion_count = 0;
    float bin_sum = 0;
    float bin_count = 0;
    int i = 0;
    for (deque<FloatFloat>::const_iterator j = data.begin(); j != data.end(); ++j) {
        float e = (*j).first;
        float w = (*j).second;
        if (e <= d[i]) {
            bin_sum += w*e;
            bin_count += w;
        } else {
            r[i] = bin_sum / bin_count;
            bin_sum = w*e;
            bin_count = w;
            ++i;
        }
        distortion_sum += w*square(e - r[i]);
        distortion_count += w;
    }
    r[i] = bin_sum / bin_count;
    return distortion_sum / distortion_count;
}

static void
updateD(vector<float>& d, const vector<float>& r)
{
    int n = r.size();
    for (int i = 0; i < n - 1; i++) {
        d[i] = (r[i] + r[i + 1])/2;
    }
}

void
LloydMaxTrainer::train(vector<float>& d, vector<float>& r)
{
    unsigned n = 1 << nBits;

    sort(data.begin(), data.end());

    vector<float> unique;
    for (unsigned i = 0; i < data.size() && unique.size() <= n; ++i) {
        float value = data[i].first;
        if (unique.empty() || value != unique.back()) {
            unique.push_back(value);
        }
    }
    if (unique.empty()) {
        unique.push_back(0);
    }

    if (unique.size() <= n) {
        //
        // Perfect quantization.

        d.resize(n);
        r.resize(n);
        while (unique.size() <= n) {
            unique.push_back(1e+38);
        }
        for (unsigned i = 0; i < n; ++i) {
            d[i] = (unique[i] + unique[i + 1])/2;
            r[i] = unique[i];
        }

    } else {
        //
        // Run Lloyd-Max.

        d.clear();
        d.resize(n + 1);
        
        r.clear();
        r.resize(n);
        
        // Initial d[i] are currently set without regard to weights.  This could
        // probably be improved.
        //
        unsigned chunk = data.size() / n;
        for (unsigned i = 0, j = chunk; i < n - 1; i++, j += chunk) {
            d[i] = data[j].first;
        }
        d[n - 1] = float(1.001)*data.back().first;
        
        double last_distortion = 1e38;
        double distortion;
        double ratio;
        double ratio_target = 1 - 1e-3;
        do {
            cerr << '.' << flush;
            distortion = updateR(r, d, data);
            updateD(d, r);
            ratio = distortion / last_distortion;
            last_distortion = distortion;
        } while (ratio < ratio_target);
        
        d.erase(d.end() - 1);

    }

    deque<FloatFloat>().swap(data); // clear and deallocate data
}

void* sq_trainer_lloyd_max_new(unsigned bits)
{
    return new LloydMaxTrainer(bits);
}
