/* M. D. Atkinson, J.-R. Sack, N. Santaro, and T. Strothotte, "Min-max heaps and
   generalized priority queues," Communications of the ACM, vol. 29, num. 10,
   Oct. 1986. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "min_max_heap.h"

struct min_max_heap;

typedef int (*MinMaxHeapVirtualCompare)(struct min_max_heap* h, int i, int j);

typedef struct min_max_heap {
    void *data;
    unsigned int element_size;
    unsigned int n;
    unsigned int capacity;
    unsigned short lazy;
    unsigned short level;
    unsigned int level_remaining;
    MinMaxHeapVirtualCompare min_max_compare;
    MinMaxHeapCompareFunc compare_func;
    MinMaxHeapIndexFunc index_func;
    void *func_data;
} MinMaxHeapRep;

/*
   lazy=0     -> Not lazy, push to end and bubble up.

   lazy=1,2,3 -> data[lazy] has been popped and is available for push
                 followed by trickle down.  On a pop+push, this saves
                 a bubble up for the push.
*/

#define _DATA(I)        ((char *)h->data + (I)*h->element_size)
#define _INDEX(I)       (h->index_func && (h->index_func(_DATA(I), I, h->func_data), 0))
#define _MOVE(D,S)      memcpy(D, S, h->element_size)
#define _MOVEI(DI,SI)   memcpy(_DATA(DI), _DATA(SI), h->element_size)
#define _SWAP(I,J)      _MOVEI(0, I), _MOVEI(I, J), _MOVEI(J, 0), _INDEX(I), _INDEX(J)
#define _COMPARE(I,J)   h->compare_func(_DATA(I), _DATA(J), h->func_data)
#define _LESS(I,J)      (_COMPARE(I, J) < 0)
#define _VLESS(I,J)     (h->min_max_compare(h, I, J) < 0)
#define _VARGLESS(I,J)  (_VLESS(I, J) ? (I) : (J))
#define _SET_MODE(LEVEL) (h->min_max_compare = ((LEVEL) & 1) ? _max_compare : _min_compare)
#define _PARENT(I)      ((I) >> 1)
#define _GRANDPARENT(I) ((I) >> 2)
#define _CHILD(I)       ((I) << 1)

/* virtual comparison for "min" mode */
static int
_min_compare(MinMaxHeapRep* h, int i, int j)
{
    return _COMPARE(i, j);
}

/* virtual comparison for "max" mode */
static int
_max_compare(MinMaxHeapRep* h, int i, int j)
{
    return _COMPARE(j, i);
}

/* Bubble up in min/max mode.  Note use of _VLESS which calls the appropriate
   comparison function depending on prior use of _MIN_MODE or _MAX_MODE. */

static int
_bubble_up(MinMaxHeapRep* h, int i)
{
    int i_start = i;
    int g;
    while ((g = _GRANDPARENT(i)) > 0 && _VLESS(i, g)) {
        _SWAP(i, g);
        i = g;
    }
    return i != i_start;
}

/* Trickle in min/max mode.  Note use of _VLESS and _VARGLESS which call the
   appropriate comparison function depending on prior use of _MIN_MODE or
   _MAX_MODE. */

static int
_extreme2(MinMaxHeapRep* h, int i)
{
    unsigned int g, g2, m;
    unsigned int c = _CHILD(i);
    if (c > h->n) {
        /* i has no children */
        return 0;
    } else if (c == h->n) {
        /* i has one child, c */
        return c;
    } else {
        /* i has two children, c and c+1 */
        g = _CHILD(c);
        if (g > h->n) {
            /* c and c+1 have no children (i.e., i has no grandchildren) */
            m = _VARGLESS(c, c + 1);
        } else if (g == h->n) {
            /* c has one child */
            m = _VARGLESS(g, c + 1);
        } else {
            /* c has two children */
            m = _VARGLESS(g, g + 1);
            g2 = _CHILD(c + 1);
            if (g2 > h->n) {
                /* c+1 has no children */
                m = _VARGLESS(m, c + 1);
            } else if (g2 <= h->n) {
                /* c+1 has at least one child */
                m = _VARGLESS(m, g2);
                if (g2 + 1 <= h->n) {
                    /* c+1 has two children */
                    m = _VARGLESS(m, g2 + 1);
                }
            }
        }
        return m;
    }
}

static void
_trickle_down(MinMaxHeapRep* h, int i)
{
    int p;
    int m = _extreme2(h, i);
    if (m != 0) {
        /* any children */
        if (m > _CHILD(i) + 1) {
            /* m is grandchild */
            if (_VLESS(m, i)) {
                _SWAP(m, i);
                p = _PARENT(m);
                if (_VLESS(p, m)) {
                    _SWAP(p, m);
                }
                _trickle_down(h, m); /* tail recursive */
            }
        } else {
            /* m is child */
            if (_VLESS(m, i)) {
                _SWAP(m, i);
            }
        }
    }
}

static void
_do_lazy(MinMaxHeapRep* h)
{
    if (h->n == 1) {
        h->n = 0;
    } else {
        _MOVEI(h->lazy, h->n);
        _INDEX(h->lazy);
        --h->n;
        _SET_MODE(h->lazy > 1);
        _trickle_down(h, h->lazy);
    }
    if (++h->level_remaining > (unsigned int)1 << h->level) {
        --h->level;
        h->level_remaining = 1;
    }
    h->lazy = 0;
}

MinMaxHeap*
min_max_heap_new(unsigned element_size,
                 unsigned initial_capacity,
                 MinMaxHeapCompareFunc compare_func,
                 MinMaxHeapIndexFunc index_func,
                 void* func_data)
{
    MinMaxHeapRep *h;

    if (initial_capacity < 1) {
        initial_capacity = 1;
    }

    h = (MinMaxHeapRep*) malloc(sizeof(MinMaxHeapRep));
    h->data = (void *)malloc((initial_capacity + 1)*element_size);
    h->element_size = element_size;
    h->n = 0;
    h->capacity = initial_capacity;
    h->compare_func = compare_func;
    h->index_func = index_func;
    h->func_data = func_data;
    h->min_max_compare = NULL;
    h->lazy = 0;
    h->level = 0;
    h->level_remaining = 1;

    return (MinMaxHeap*) h;
}

void
min_max_heap_delete(MinMaxHeap* h_)
{
    MinMaxHeapRep* h = (MinMaxHeapRep*) h_;

    free(h->data);
    free(h);
}

void
min_max_heap_clear(MinMaxHeap* h_)
{
    MinMaxHeapRep* h = (MinMaxHeapRep*) h_;

    h->n = 0;
    h->lazy = 0;
    h->level = 0;
    h->level_remaining = 1;
}

unsigned
min_max_heap_size(MinMaxHeap* h_)
{
    MinMaxHeapRep* h = (MinMaxHeapRep*) h_;

    return (h->lazy != 0) ? h->n - 1 : h->n;
}

const void*
min_max_heap_data(MinMaxHeap* h_)
{
    MinMaxHeapRep* h = (MinMaxHeapRep*) h_;

    if (h->lazy != 0) {
        _do_lazy(h);
    }
    return (char *)h->data + h->element_size;
}

void
min_max_heap_push(MinMaxHeap* h_, const void* source)
{
    MinMaxHeapRep* h = (MinMaxHeapRep*) h_;
    int i, p;

    if (h->lazy != 0) {

        _MOVE(_DATA(h->lazy), source);
        _INDEX(h->lazy);
        if (h->lazy == 1) {
            _SET_MODE(0);
        } else {
            _SET_MODE(1);
        }
        _trickle_down(h, h->lazy);
        h->lazy = 0;

    } else {

        if (h->n == h->capacity) {
            int capacity = 2*h->capacity;
            h->data = (char *)realloc(h->data, (capacity + 1)*h->element_size);
            h->capacity = capacity;
        }
        h->n++;
        _MOVE(_DATA(h->n), source);
        _INDEX(h->n);
        
        i = h->n;
        _SET_MODE(h->level);
        p = _PARENT(i);
        if (p > 0 && _VLESS(p, i)) {
            _SWAP(p, i);
            _SET_MODE(h->level - 1);
            _bubble_up(h, p);
        } else {
            _bubble_up(h, i);
        }
        if (--h->level_remaining == 0) {
            h->level_remaining = 1 << ++h->level;
        }

    }
}

int
min_max_heap_peek_min(MinMaxHeap* h_, void* destination)
{
    MinMaxHeapRep* h = (MinMaxHeapRep*) h_;

    if (h->lazy != 0) {
        _do_lazy(h);
    }
    if (h->n == 0) {
        return 0;
    }
    if (destination) {
        _MOVE(destination, _DATA(1));
    }
    return 1;
}

int
min_max_heap_pop_min(MinMaxHeap* h_, void* destination)
{
    MinMaxHeapRep* h = (MinMaxHeapRep*) h_;

    h->lazy = min_max_heap_peek_min(h_, destination);
    if (h->index_func && h->lazy != 0) {
        h->index_func(_DATA(h->lazy), 0, h->func_data);
    }

    return h->lazy;
}

int
min_max_heap_peek_max(MinMaxHeap* h_, void* destination)
{
    MinMaxHeapRep* h = (MinMaxHeapRep*) h_;
    int m;

    if (h->lazy != 0) {
        _do_lazy(h);
    }
    if (h->n == 0) {
        return 0;
    }
    if (h->n == 1) {
        m = 1;
    } else {
        m = (h->n == 2) ? 2 : (_LESS(3, 2) ? 2 : 3);
    }
    if (destination) {
        _MOVE(destination, _DATA(m));
    }
    return m;
}

int
min_max_heap_pop_max(MinMaxHeap* h_, void* destination)
{
    MinMaxHeapRep* h = (MinMaxHeapRep*) h_;

    h->lazy = min_max_heap_peek_max(h_, destination);
    if (h->index_func && h->lazy != 0) {
        h->index_func(_DATA(h->lazy), 0, h->func_data);
    }

    return h->lazy;
}

static unsigned
_log2(unsigned x)
{
    unsigned log = 0;
    while (x >>= 1) {
        ++log;
    }
    return log;
}

void
min_max_heap_preadjust(MinMaxHeap* h_)
{
    MinMaxHeapRep* h = (MinMaxHeapRep*) h_;

    if (h->lazy != 0) {
        _do_lazy(h);
    }
}

void
min_max_heap_adjust(MinMaxHeap* h_, unsigned i)
{
    MinMaxHeapRep* h = (MinMaxHeapRep*) h_;
    unsigned level;
    unsigned p;

    assert(h->lazy == 0);

    level = _log2(i);
    _SET_MODE(level);
    p = _PARENT(i);
    if (p > 0 && _VLESS(p, i)) {
        _SWAP(p, i);
        _trickle_down(h, i);
        _SET_MODE(level + 1);
        _bubble_up(h, p);
    } else {
        if (!_bubble_up(h, i)) {
            _trickle_down(h, i);
        }
    }
}

unsigned
min_max_heap_verify(MinMaxHeap* h_)
{
    MinMaxHeapRep* h = (MinMaxHeapRep*) h_;
    int i = h->n;
    int level = _log2(i);
    int level_remaining = (1 << (level + 1)) - i;
    int p, g;
    
    assert(h->lazy == 0);

    _SET_MODE(level);
    while (i >= 1) {
        p = _PARENT(i);
        if (p > 0 && _VLESS(p, i)) {
            return i;
        }
        g = _GRANDPARENT(i);
        if (g > 0 && _VLESS(i, g)) {
            return i;
        }
        if (++level_remaining > 1 << level) {
            --level;
            level_remaining = 1;
            _SET_MODE(level);
        }
        --i;
    }
    return 0;
}
