/* -*- C++ -*-
 *
 * Copyright (c) 2009
 * Spoken Language Systems Group
 * MIT Computer Science and Artificial Intelligence Laboratory
 * Massachusetts Institute of Technology
 *
 * All Rights Reserved
 */

#include <iostream>
#include <fstream>
#include "fst/FST.h"
#include "../lib/BitStreamWriter32.h"
#include "../lib/sq.h"

// echo -n Address1 | md5sum | cut -c1-1
#define ADDRESS1 0x93d03fe3

using namespace std;

typedef fst::map<string,int> StringIntMap;
typedef fst::list<int> Sequence;
typedef fst::map<int,Sequence> IntSequenceMap;

typedef fst::map<string, string> StringStringMap;

typedef fst::list<string> StringList;
typedef fst::vector<string> StringVector;

typedef fst::list<int> IntList;

typedef fst::set<int> IntSet;

typedef fst::vector<int> IntVector;

// (pad:8, "city name\0", pad:32, word index+ 0, rtree)+    <- city data
// (pad:32, "code", #cities, city position+)                <- state data
//   :
// (pad:8, "city name\0", word index+ 0, rtree)+    <- city data
// (pad:32, "code" #cities, city position+)                <- state data

// (pad:8, "al, "alabama", "ak", "alaska", ... "wy", "wyoming")
// (pad:32, #states, state name position+)
// (pad:32, state position+)                        <- #states

static const char* stateCodes[] = {
    "alaska", "ak",
    "alabama", "al",
    "arkansas", "ar",
    "arizona", "az",
    "california", "ca",
    "colorado", "co",
    "connecticut", "ct",
    "d c", "dc",
    "district of columbia", "dc",
    "delaware", "de",
    "florida", "fl",
    "georgia", "ga",
    "hawaii", "hi",
    "iowa", "ia",
    "idaho", "id",
    "illinois", "il",
    "indiana", "in",
    "kansas", "ks",
    "kentucky", "ky",
    "louisiana", "la",
    "mass", "ma",
    "massachusetts", "ma",
    "maryland", "md",
    "maine", "me",
    "michigan", "mi",
    "minnesota", "mn",
    "missouri", "mo",
    "mississippi", "ms",
    "montana", "mt",
    "north carolina", "nc",
    "north dakota", "nd",
    "nebraska", "ne",
    "new hampshire", "nh",
    "new jersey", "nj",
    "new mexico", "nm",
    "nevada", "nv",
    "new york", "ny",
    "ohio", "oh",
    "oklahoma", "ok",
    "oregon", "or",
    "pennsylvania", "pa",
    "rhode island", "ri",
    "south carolina", "sc",
    "south dakota", "sd",
    "tennessee", "tn",
    "texas", "tx",
    "utah", "ut",
    "virginia", "va",
    "vermont", "vt",
    "washington", "wa",
    "wisconsin", "wi",
    "west virginia", "wv",
    "wyoming", "wy",
    NULL
};

struct City {
    //int index;     // index 0... within State
    int position;  // position withing bit stream
};

typedef fst::map<string, City> StringCityMap;

struct State {
    //int index;     // index 0... within Country
    int position;  // position within bitstream
    StringCityMap cities;
};

typedef fst::map<string, State> StringStateMap;

struct Country {
    StringStringMap codes; // map state name (or code) to code
    StringStateMap states; // by code only!
};

void
usage()
{
    cerr << "usage: words.txt MCP.fst l2s.baseforms\n";
    exit(1);
}

static void
load_index_file(const char* filename, StringIntMap& out)
{
    out.clear();
    out[""] = 0;
    ifstream is(filename);
    for (string line; getline(is,line);) {
        out[line] = out.size();
    }
}

static void
splitString(const string& s, StringList& out)
{
    out.clear();

    string::size_type i = 0, j;
    for (;;) {
        j = s.find_first_of(" \t", i);
        if (j == string::npos) {
            out.push_back(s.substr(i));
            break;
        }
        out.push_back(s.substr(i, j - i));
        i = s.find_first_not_of(" \t", j);
        if (i == string::npos) {
            break;
        }
    }
}

static bool
map_sequence(StringList& strings, StringIntMap& stringMap, Sequence& out)
{
    out.clear();
    for (StringList::iterator i = strings.begin(); i != strings.end(); ++i) {
        StringIntMap::iterator j = stringMap.find(*i);
        if (j == stringMap.end()) {
            out.clear();
            return false;
        }
        out.push_back((*j).second);
    }
    return true;
}

static void
load_baseforms(const char* filename,
               StringIntMap& wordMap, StringIntMap& phonemeMap,
               IntSequenceMap& out)
{
    out.clear();
    ifstream is(filename);
    for (string line; getline(is,line);) {
        StringList phonemes;
        splitString(line, phonemes);
        string word = phonemes.front();
        phonemes.pop_front();
        phonemes.pop_front();
        StringIntMap::iterator i = wordMap.find(word);
        if (i != wordMap.end()) {
            int word = (*i).second;
            Sequence sequence;
            if (map_sequence(phonemes, phonemeMap, sequence)) {
                out[word] = sequence;
            }
        }
    }
}

static unsigned
bitsNeeded(unsigned n)
{
    unsigned c;
    n |= n >> 1;
    n |= n >> 2;
    n |= n >> 4;
    n |= n >> 8;
    n |= n >> 16;
    for (c = 0; n; c++)
        n &= n - 1;
    return c;
}

static void writeWordIndex(BitStreamWriter32& writer, int w)
{
    const int w0 = 6;
    const int w1 = 10;
    const int w2 = 13;
    const int w3 = 18;

    const int e0 = (1 << w0);
    const int e1 = (1 << w1) + e0;
    const int e2 = (1 << w2) + e1;
    const int e3 = (1 << w3) + e2;

    if (w < e0) {
        writer.write(0, 1);
        writer.write(0, 1);
        writer.write(w, w0);
    } else if (w < e1) {
        writer.write(0, 1);
        writer.write(1, 1);
        writer.write(w - e0, w1);
    } else if (w < e2) {
        writer.write(1, 1);
        writer.write(0, 1);
        writer.write(w - e1, w2);
    } else if (w < e3) {
        writer.write(1, 1);
        writer.write(1, 1);
        writer.write(w - e2, w3);
    } else {
        abort();
    }
}

static void pad(FILE* fp, int n)
{
    int p = ftell(fp);
    int m = p % n;
    if (m != 0) {
        for (int i = 0; i < n - m; ++i) {
            fputc(0, fp);
        }
    }
}

static void pad32(FILE* fp)
{
    pad(fp, 4);
}

static void write32(FILE* fp, int x)
{
    fwrite(&x, 4, 1, fp);
}

static void write16(FILE* fp, u16 x)
{
    fwrite(&x, 2, 1, fp);
}

static void writeString(FILE* fp, const string& city)
{
    fputs(city.c_str(), fp);
    fputc(0, fp);
}

static void processCity(Country& us,
                        const string& stateCode, const string& cityName, FSTGeneric fst,
                        StringIntMap& wordMap, FILE* fp)
{
    assert(us.states.count(stateCode) == 1);
    State& state = us.states[stateCode];
    City& city = state.cities[cityName];

    city.position = ftell(fp);
    writeString(fp, cityName);

    StringList cityWords;
    splitString(cityName, cityWords);
    pad32(fp);
    for (StringList::iterator i = cityWords.begin(); i != cityWords.end(); ++i) {
        assert(wordMap.count(*i) != 0);
        write32(fp, wordMap[*i]);
    }
    write32(fp, 0);

    BitStreamWriter32 writer;

    fst = FSTDeterminizeA(fst);

    IntList fifo;
    fifo.push_back(fst->initial());
    while (!fifo.empty()) {
        int p = fifo.front();
        fifo.pop_front();
        bool anyArcs = false;
        for (FSTArcIter a = fst->arcs_out(p); a.valid(); ++a) {
            string word = FSTstring_map[a->input];
            if (wordMap.count(word)) {
                anyArcs = true;
                writer.write(1, 1);
                writeWordIndex(writer, wordMap[word]);
                fifo.push_back(a->next);
            } else {
                // DO SOMETHING HERE!
            }
        }
        writer.write(0, 1);
        if (anyArcs) {
            writer.write(fst->is_final(p), 1);
        }
    }
    writer.flush();
    fwrite(writer.data(), writer.length(), 1, fp);
}

static void processState(Country& us, const string& stateCode, StringIntMap& wordMap, FILE* fp)
{
    cout << stateCode << '\n';

    assert(us.states.count(stateCode) == 1);
    State& state = us.states[stateCode];

    state.position = ftell(fp);
    writeString(fp, stateCode);

    // word indices (any used, no particular order)
    IntSet wordIndices;
    for (int i = 0; stateCodes[i] != NULL; i += 2) {
        if (stateCode == stateCodes[i + 1]) {
            StringList words;
            splitString(stateCodes[i], words);
            for (StringList::const_iterator j = words.begin(); j != words.end(); ++j) {
                assert(wordMap.count(*j) != 0);
                wordIndices.insert(wordMap[*j]);
            }
        }
    }
    pad32(fp);
    for (IntSet::const_iterator i = wordIndices.begin(); i != wordIndices.end(); ++i) {
        write32(fp, *i);
    }
    write32(fp, 0);

    write32(fp, state.cities.size());
    for (StringCityMap::iterator i = state.cities.begin(); i != state.cities.end(); ++i) {
        City& city = (*i).second;
        write32(fp, city.position);
    }

    // precompiled PFST for $Filler <city> <state>
    string pfst_file = "cities/";
    pfst_file += stateCode;
    pfst_file += ".pfst";
    FILE *pfst_fp = fopen(pfst_file.c_str(), "rb");
    assert(pfst_fp != NULL);
    std::vector<char> data;
    char buf;
    for (size_t n; (n = fread(&buf, 1, 1, pfst_fp)) > 0;) {
        data.push_back(buf);
    }
    write32(fp, data.size());
    pad(fp, 32);
    fwrite(&data[0], data.size(), 1, fp);
    cout << '\t' << data.size() << '\n';
}

// dirName should be "db/streets"
int processDb(const char* file, StringIntMap& wordMap, FILE* fp)
{
    // ./ak/adak
    // amitchka circle
    // beverly cove
    // camel cove
    // caribou court
    // dolly varden
    // expedition harbor road
    // finger bay

    Country us;
    for (int i = 0; stateCodes[i] != NULL; i += 2) {
        string code = stateCodes[i + 1];
        string state = stateCodes[i];
        us.codes[state] = code;
        us.codes[code] = code;
        us.states[code]; // just creating, filled in later by processState()
    }

    FSTGeneric fst;
    string state, city;

    ifstream is(file);
    for (string line; getline(is, line);) {
        if (line.find("./") == 0) {
            if (state.length() > 0) {
                processCity(us, state, city, fst, wordMap, fp);
            }
            string prevState = state;
            state = line.substr(2, 2);
            city = line.substr(5);
            fst = FSTBasic();
            fst->set_initial(fst->add_state());
            if (prevState.length() > 0 && prevState != state) {
                processState(us, prevState, wordMap, fp);
            }
        } else {
            StringList words;
            splitString(line, words);
            int p = 0;
            for (StringList::reverse_iterator i = words.rbegin();
                 i != words.rend(); ++i) {
                int q = fst->add_state();
                fst->add_arc(p, q, FSTstring_map[*i]);
                p = q;
            }
            fst->set_final(p);
        }
    }
    if (state.length() > 0) {
        processCity(us, state, city, fst, wordMap, fp);
        processState(us, state, wordMap, fp);
    }

    // state names and codes as strings
    IntList positions;
    for (StringStringMap::iterator i = us.codes.begin(); i != us.codes.end(); ++i) {
        positions.push_back(ftell(fp));
        writeString(fp, (*i).first);
    }

    pad32(fp);
    int countryPosition = ftell(fp);

    // pointers to strings
    write32(fp, positions.size());
    for (IntList::iterator i = positions.begin(); i != positions.end(); ++i) {
        write32(fp, *i);
    }

    // pointers to states, full names and codes too
    for (StringStringMap::iterator i = us.codes.begin(); i != us.codes.end(); ++i) {
        const string& code = (*i).second;
        assert(us.states.count(code) != 0);
        State& state = us.states[code];
        write32(fp, state.position);
    }

    return countryPosition;
}

static int processProns(StringIntMap& wordMap, StringIntMap& phonemeMap,
                        IntSequenceMap& l2sBaseforms, FILE* fp)
{
    int W = wordMap.size();

    BitStreamWriter32 wordPronWriter(false);
    BitStreamWriter32 wordPronPosWriter(false);

    int pBits = bitsNeeded(phonemeMap.size());
    
    Sequence positions;
    
    const int pad = 16;
    int maxPosition = 0;
    
    for (int w = 0; w < W; ++w) {
        Sequence& pSequence = l2sBaseforms[w];
        wordPronWriter.pad(pad);
        int position = wordPronWriter.position()/pad;
        maxPosition = std::max(maxPosition, position);
        positions.push_back(position);
        if (pSequence.size() > 0) {
            wordPronWriter.write(1, 1);
            wordPronWriter.write(0, 1);
            wordPronWriter.write(0, 1);
            for (Sequence::const_iterator i = pSequence.begin();
                 i != pSequence.end(); ++i) {
                wordPronWriter.write(*i, pBits);
            }
            wordPronWriter.write(0, pBits);
        } else {
            wordPronWriter.write(0, 1);
        }
    }

    int posBits = bitsNeeded(maxPosition);
    for (Sequence::iterator i = positions.begin(); i != positions.end(); ++i) {
        wordPronPosWriter.write(*i, posBits);
    }

#if 0
    StringVector reversePhonemeMap(phonemeMap.size());
    for (StringIntMap::iterator i = phonemeMap.begin(); i != phonemeMap.end(); ++i) {
        reversePhonemeMap[(*i).second] = (*i).first;
    }
    pad32(fp);
    int phonemeMapOffset = ftell(fp);
    write32(fp, reversePhonemeMap.size());
    for (StringVector::const_iterator i = reversePhonemeMap.begin();
         i != reversePhonemeMap.end(); ++i) {
        writeString(fp, *i);
    }
#endif

    pad32(fp);
    int pronOffset = ftell(fp);
    fwrite(wordPronWriter.data(), wordPronWriter.length(), 1, fp);

    pad32(fp);
    int posOffset = ftell(fp);
    write32(fp, W);
    write32(fp, pronOffset);
#if 0
    write32(fp, phonemeMapOffset);
#endif
    write32(fp, posBits);
    write32(fp, pBits);
    fwrite(wordPronPosWriter.data(), wordPronWriter.length(), 1, fp);

    return posOffset;
}

static int processWordTable(StringIntMap& wordMap, FILE* fp)
{
    int W = wordMap.size();
    int base = ftell(fp);
    IntVector positions(W);
    int maxPosition = 0;
    for (StringIntMap::iterator i = wordMap.begin(); i != wordMap.end(); ++i) {
        const string& word = (*i).first;
        int position = ftell(fp) - base;
        positions[(*i).second] = position;
        maxPosition = std::max(maxPosition, position);
        writeString(fp, word.c_str());
    }
    int wordStringPosBits = bitsNeeded(maxPosition);

    pad32(fp);

    int result = ftell(fp);
    write32(fp, base);
    write32(fp, W);
    write32(fp, wordStringPosBits);

    BitStreamWriter32 writer;
    for (int i = 0; i < W; ++i) {
        writer.write(positions[i], wordStringPosBits);
    }
    writer.flush();
    fwrite(writer.data(), writer.length(), 1, fp);

    return result;
}

static void alphabetToMap(const fst::set<int>& a, StringIntMap& m)
{
    int index = 0;
    m[""] = index++;
    m["#"] = index++;
    for (fst::set<int>::const_iterator i = a.begin(); i != a.end(); ++i) {
        m.insert(StringIntMap::value_type(FSTstring_map[*i], m.size()));
    }
}

static void writeMap(FILE* fp, const StringIntMap& m)
{
    StringVector reverse(m.size());
    for (StringIntMap::const_iterator i = m.begin(); i != m.end(); ++i) {
        reverse[(*i).second] = (*i).first;
    }

    pad32(fp);
    write32(fp, m.size());
    for (StringVector::iterator i = reverse.begin(); i != reverse.end(); ++i) {
        writeString(fp, *i);
    }
}

#define F2Q8(f) s16((f)*(1<<8)+0.5f)

static int processMCP(StringIntMap& pMap, const FSTGeneric& fst, FILE* fp)
{
    StringIntMap mMap;
    alphabetToMap(fst->input_alphabet(), mMap);

    pad32(fp);
    int position = ftell(fp);

    writeMap(fp, mMap);

    writeMap(fp, pMap);

    pad32(fp);
    write32(fp, fst->n_states());
    write32(fp, fst->initial());

    const int weightBits = 5;
    const int weightN = 1 << weightBits;
    void* trainer = sq_trainer_lloyd_max_new(weightBits);
    for (int p = 0; p < fst->n_states(); ++p) {
        FSTArcIter a = fst->arcs_out(p);
        int value = a.n();
        if (fst->is_final(p)) {
            value |= (1 << 31);
            sq_trainer_incorporate(trainer, fst->final_weight(p), 1);
        }
        write32(fp, value);
        for (; a.valid(); ++a) {
            write32(fp, a->next);
            write16(fp, mMap[FSTstring_map[a->input]]);
            write16(fp, pMap[FSTstring_map[a->output]]);
            write32(fp, F2Q8(a->weight));
            if (a->weight != 0) {
                sq_trainer_incorporate(trainer, a->weight, 1);
            }
        }
    }

    float *weightD = 0;
    float *weightR = 0;
    cout << "Training weight quantizer";
    sq_trainer_train(trainer, &weightD, &weightR);
    sq_trainer_delete(trainer);
    cout << ".done.\n";

    pad32(fp);
    write16(fp, weightBits);
    for (int i = 0; i < weightN; ++i) {
        write16(fp, F2Q8(weightD[i]));
    }
    for (int i = 0; i < weightN; ++i) {
        write16(fp, F2Q8(weightR[i]));
    }

    return position;
}

int
main(int argc, char** argv)
{
    bool error = false;

    char c;
    while ((c = getopt(argc, argv, "")) != EOF) {
        switch (c) {
        default:
            error = true;
            break;
        }
    }
    argc -= optind;
    argv += optind;
    if (error || argc != 3)
        usage();

    const char* words_txt_file     = *argv++;
    const char* mcp_file           = *argv++;
    //const char* phonemes_txt_file  = *argv++;
    const char* l2s_baseforms_file = *argv++;
    //const char* words_split_file   = *argv++;
    //const char* dict_fst_file      = *argv++;

    StringIntMap wordMap;
    load_index_file(words_txt_file, wordMap);
    cout << wordMap.size() << " words.\n";

    StringIntMap phonemeMap;
#if 0
    load_index_file(phonemes_txt_file, phonemeMap);
#endif
    FSTGeneric mcp = FSTLoad(mcp_file);
    alphabetToMap(mcp->output_alphabet(), phonemeMap);
    cout << phonemeMap.size() << " phonemes.\n";

    IntSequenceMap l2sBaseforms;
    load_baseforms(l2s_baseforms_file, wordMap, phonemeMap, l2sBaseforms);
    cout << l2sBaseforms.size() << " l2s baseforms.\n";

#if 0
    IntSequenceMap splits;
    load_baseforms(words_split_file, wordMap, wordMap, splits);
    cout << splits.size() << " splits.\n";
#endif

    FILE* fp = fopen("address.bin", "w");

    write32(fp, ADDRESS1);

    int offsets = ftell(fp);
    write32(fp, 0); // db start offset
    write32(fp, 0); // pron start offset
    write32(fp, 0); // words start offset

    int dbPos = processDb("streets", wordMap, fp);
    int pronPos = processProns(wordMap, phonemeMap, l2sBaseforms, fp);
    int wordPos = processWordTable(wordMap, fp);
    int mcpPos = processMCP(phonemeMap, mcp, fp);

    cout << "total bytes: " << ftell(fp) << '\n';

    fseek(fp, offsets, SEEK_SET);
    write32(fp, dbPos);
    write32(fp, pronPos);
    write32(fp, wordPos);
    write32(fp, mcpPos);

    fclose(fp);

    return 0;
}

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 80
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
