/* -*- C++ -*-
 *
 * Copyright (c) 2009
 * Spoken Language Systems Group
 * MIT Computer Science and Artificial Intelligence Laboratory
 * Massachusetts Institute of Technology
 *
 * All Rights Reserved
 */

#include <iostream>
#include <fstream>
#include <cstdio>
#include "fst/FST.h"
#include "../lib/BitStreamWriter32.h"
#include "../lib/sq.h"

// echo -n Dynamic1 | md5sum | cut -c1-8
#define DYNAMIC1 0x81a7f4e5

// echo -n Dynamic2 | md5sum | cut -c1-8
#define DYNAMIC2 0x3bd377a4

// echo -n Dynamic3 | md5sum | cut -c1-8
#define DYNAMIC3 0x8d405d80

using namespace std;

typedef fst::set<int> IntSet;
typedef fst::map<string,int> StringIntMap;
typedef fst::vector<string> StringVector;
typedef fst::vector<int> IntVector;
typedef fst::vector<u8> ByteVector;
typedef fst::map<int,int> IntIntMap;

void
usage()
{
    cerr <<
        "usage: L.fst MCP.fst L2S.fst out.bin\n"
        "\n"
        "MCP.fst should be general, non-main, non-class version,\n"
        "as the main- and class-specific versions will be derived.\n"
        "\n";
    exit(1);
}


// Phong's linear congruential hash
static u32
hash2(const char* key)
{
    u32 h = 0;
    while (*key) {
        h =  0x63c63cd9*h + 0x9c39c33d + *key;
        ++key;
    }
    return h;
}

static u32
hash2(const string& key)
{
    return hash2(key.c_str());
}

static int
computeTableSize(int n)
{
    u32 size = 1;
    for (u32 x = (n*4)/3; x; x >>= 1) {
        size <<= 1;
    }
    return size;
}

static bool
isLinear(const FSTGeneric& fsa)
{
    for (int p = 0, n = fsa->n_states(); p < n; ++p) {
        FSTArcIter a = fsa->arcs_out(p);
        if (a.n() > 1) return false;
        if (a.n() > 0 && fsa->is_final(p)) return false;
    }
    return true;
}

static u32
bitsNeeded(u32 n)
{
    u32 c;
    n |= n >> 1;
    n |= n >> 2;
    n |= n >> 4;
    n |= n >> 8;
    n |= n >> 16;
    for (c = 0; n; c++)
        n &= n - 1;
    return c;
}

static inline int
get(const IntIntMap& mapping, int key)
{
    return (*mapping.find(key)).second;
}

static void
packPronunciationLinear(BitStreamWriter32& writer, const FSTGeneric& fst,
                        const IntIntMap& mapping)
{
    int width = bitsNeeded(mapping.size());

    writer.write(0, 1); // flag as linear
    for (int p = fst->initial();;) {
        FSTArcIter a = fst->arcs_out(p);
        assert(a.n() < 2);
        if (a.n() == 1) {
            writer.write(1, 1);
            writer.write(get(mapping, a->input), width);
            p = a->next;
        } else {
            break;
        }
    }
    writer.write(0, 1);
}

static void
packPronunciationGraph(BitStreamWriter32& writer, const FSTGeneric& fst,
                       const IntIntMap& mapping)
{
    int n = fst->n_states();
    assert(n < 32);

    int stateWidth = bitsNeeded(n);
    int labelWidth = bitsNeeded(mapping.size());

    writer.write(1, 1); // flag as graph
    //cout << "\tgraph\n";

    writer.write(stateWidth, 5);
    writer.write(n, stateWidth);

    assert(fst->initial() == 0);

    for (int p = 0; p < n; ++p) {
        for (FSTArcIter a = fst->arcs_out(p); a.valid(); ++a) {
            int q = a->next;
            writer.write(1, 1);
            writer.write(get(mapping, a->input), labelWidth);
            if (q == p + 1) {
                writer.write(0, 1);
            } else {
                writer.write(1, 1);
                writer.write(q, stateWidth);
            }
        }
        writer.write(0, 1);
        writer.write(fst->is_final(p), 1);
    }
}

static void
packPronunciation(BitStreamWriter32& writer, const FSTGeneric& L, int wi,
                  const IntIntMap& phoneMap)
{
    assert(wi > 0);

    FSTGeneric fst = FSTBasic();
    fst->add_states(2);
    fst->set_initial(0);
    fst->set_final(1);
    fst->add_arc(0, 1, wi, wi);

    fst = FSTCompose(fst, L);
    fst = FSTTrim(fst);

    assert(fst->n_states() > 0);

    fst = FSTProject(fst, FSTOutput, FSTNone);
    fst = FSTOptimizeA(fst);

    if (isLinear(fst)) {
        packPronunciationLinear(writer, fst, phoneMap);
    } else {
        packPronunciationGraph(writer, fst, phoneMap);
    }
}

static FSTGeneric
filterIwt(const FSTGeneric& L)
{
    FSTGeneric filter = FSTBasic();
    filter->add_state();
    filter->set_initial(0);
    filter->set_final(0);

    int underbar = FSTstring_map["_"];
    const IntSet& phonemes = L->output_alphabet();
    for (IntSet::const_iterator i = phonemes.begin(); i != phonemes.end(); ++i) {
        if (*i > 1 && *i != underbar) {
            filter->add_arc(0, 0, *i, *i);
        }
    }

    return FSTTrim(FSTCompose(L, filter));
}

static void pad(FILE* fp, int n)
{
    int p = ftell(fp);
    int m = p % n;
    if (m != 0) {
        for (int i = 0; i < n - m; ++i) {
            fputc(0, fp);
        }
    }
}

static void pad32(FILE* fp)
{
    pad(fp, 4);
}

static void write32(FILE* fp, int x)
{
    fwrite(&x, 4, 1, fp);
}

static void write16(FILE* fp, u16 x)
{
    fwrite(&x, 2, 1, fp);
}

static void writeString(FILE* fp, const string& city)
{
    fputs(city.c_str(), fp);
    fputc(0, fp);
}

static void alphabetToMap(const fst::set<int>& a, StringIntMap& m)
{
    m[""] = 0;
    m["#"] = 1;
    for (fst::set<int>::const_iterator i = a.begin(); i != a.end(); ++i) {
        m.insert(StringIntMap::value_type(FSTstring_map[*i], m.size()));
    }
}

static void writeMap(FILE* fp, const StringIntMap& m)
{
    StringVector reverse(m.size());
    for (StringIntMap::const_iterator i = m.begin(); i != m.end(); ++i) {
        reverse[(*i).second] = (*i).first;
    }

    pad32(fp);
    write32(fp, m.size());
    for (StringVector::iterator i = reverse.begin(); i != reverse.end(); ++i) {
        writeString(fp, *i);
    }
}

#define F2Q8(f) s16((f)*(1<<8)+0.5f)

static u32
processMCP(StringIntMap& pMap, const FSTGeneric& fst, FILE* fp)
{
    pMap.clear();
    alphabetToMap(fst->output_alphabet(), pMap);

    StringIntMap mMap;
    alphabetToMap(fst->input_alphabet(), mMap);

    pad32(fp);
    u32 position = ftell(fp);

    writeMap(fp, mMap);

    writeMap(fp, pMap);

    pad32(fp);
    write32(fp, bitsNeeded(pMap.size()));
    write32(fp, fst->n_states());
    write32(fp, fst->initial());
    for (int p = 0; p < fst->n_states(); ++p) {
        FSTArcIter a = fst->arcs_out(p);
        int value = a.n();
        if (fst->is_final(p)) {
            value |= (1 << 31);
        }
        write32(fp, value);
        for (; a.valid(); ++a) {
            write32(fp, a->next);
            write16(fp, mMap[FSTstring_map[a->input]]);
            write16(fp, pMap[FSTstring_map[a->output]]);
        }
    }

    return position;
}

static u32
processL(FILE* fp, const StringIntMap& pMap, FSTGeneric L)
{
    L = filterIwt(L);

    IntIntMap phoneMap;
    for (StringIntMap::const_iterator i = pMap.begin(); i != pMap.end(); ++i) {
        phoneMap[FSTstring_map[(*i).first]] = (*i).second;
    }

    const IntSet& wordIndices = L->input_alphabet();

    StringIntMap wordBucketMap;
    for (IntSet::const_iterator i = wordIndices.begin(); i != wordIndices.end(); ++i) {
        if (*i != 0) {
            wordBucketMap[FSTstring_map[*i]] = -1;
        }
    }

    int n = computeTableSize(wordBucketMap.size());
    int mask = n - 1;

    // Assign keys to buckets.  We know we have plenty of buckets so don't have
    // to worry about not terminating while looping through buckets due to hash
    // collisions.
    StringVector bucketStrings(n);
    for (StringIntMap::iterator i = wordBucketMap.begin(); i != wordBucketMap.end(); ++i) {
        const string& key = (*i).first;
        int b = (int) hash2(key) & mask;
        while (bucketStrings[b] != "") {
            b = (b + 1) & mask;
        }
        bucketStrings[b] = key;
        wordBucketMap[key] = b;
    }

    // Pack pronunciations in bucket order.
    BitStreamWriter32 pronWriter;
    IntVector pronPosition(n, -1);
    int pronPositionMax = 0;
    for (int b = 0; b < n; ++b) {
        string word = bucketStrings[b];
        if (word.length() > 0) {
            //cout << "word=" << word << " pos=" << pronWriter.position() << '\n';
            pronWriter.pad(32);
            pronPosition[b] = pronPositionMax = pronWriter.position()/8;
            packPronunciation(pronWriter, L, FSTstring_map[word], phoneMap);
        }
    }
    assert(pronPositionMax < (1 << 24));
    cout << "pronunciations: " << pronWriter.length() << " bytes\n";

    // Pack (string, pronunciation offset) tuples in bucket order.
    ByteVector tupleBytes;
    tupleBytes.push_back(0); // filler so that first position is not 0 (NULL).
    IntVector tuplePosition(n, 0);
    int tuplePositionMax = 0;
    for (int b = 0; b < n; ++b) {
        string word = bucketStrings[b];
        if (word.length() > 0) {
            tuplePosition[b] = tuplePositionMax = tupleBytes.size();
            for (u32 i = 0; i < word.length(); ++i) {
                tupleBytes.push_back(word[i]);
            }
            tupleBytes.push_back(0);
            tupleBytes.push_back((pronPosition[b] >>  0) & 0xff);
            tupleBytes.push_back((pronPosition[b] >>  8) & 0xff);
            tupleBytes.push_back((pronPosition[b] >> 16) & 0xff);
        }
    }
    cout << "tuples: " << tupleBytes.size() << " bytes\n";

    // Pack buckets.
    BitStreamWriter32 bucketWriter;
    int tuplePositionBits = bitsNeeded(tuplePositionMax);
    for (int b = 0; b < n; ++b) {
        bucketWriter.write(tuplePosition[b], tuplePositionBits);
    }
    cout << "buckets: " << bucketWriter.length() << " bytes\n";

    // Header...
    // tuple start position <u32>
    // pron start position <u32>
    // bucket start position <u32>

    u32 tupleOffset = 0;
    u32 pronOffset = 0;

    pad32(fp);
    u32 position = ftell(fp);
    write32(fp, n);
    write32(fp, tuplePositionBits);
    u32 save = ftell(fp);
    write32(fp, tupleOffset); // fix later
    write32(fp, pronOffset); // fix later

    fwrite(bucketWriter.data(), bucketWriter.length(), 1, fp);

    pad32(fp);
    tupleOffset = ftell(fp);
    fwrite(&tupleBytes[0], tupleBytes.size(), 1, fp);

    pad32(fp);
    pronOffset = ftell(fp);
    fwrite(pronWriter.data(), pronWriter.length(), 1, fp);

    fseek(fp, save, SEEK_SET);
    write32(fp, tupleOffset); // fixing
    write32(fp, pronOffset); // fixing

    fseek(fp, 0, SEEK_END);

    return position;
}

static string
removeUnderbars(string s)
{
    if (s != "_") {
        for (string::size_type i; (i = s.find('_')) != string::npos;) {
            s.erase(i, 1);
        }
    }
    return s;
}

u32 processL2S(FILE* fp, const FSTGeneric& ngram, const StringIntMap& pMap)
{
    typedef fst::set<int> IntSet;
    typedef fst::list<int> IntList;
    typedef fst::list<string> StringList;
    typedef fst::map<string, int> StringIntMap;
    typedef fst::map<string, IntList> StringIntListMap;
    typedef fst::list<IntList> IntListList;

    StringIntMap graphoneMap;
    const IntSet& alphabet = ngram->input_alphabet();
    for (IntSet::const_iterator i = alphabet.begin(); i != alphabet.end(); ++i) {
        if (*i) {
            graphoneMap[FSTstring_map[*i]] = 0;
        }
    }
    StringList graphoneList;
    StringIntListMap lettersUnitsMap;
    IntListList unitPhonemes;
    int unit = 0;
    int maxLetters = 0;
#if 0
    graphoneList.push_back("");
    unitPhonemes.push_back(IntList());
    unit = 1;
#endif
    for (StringIntMap::iterator i = graphoneMap.begin(); i != graphoneMap.end(); ++i) {
        string label = (*i).first;
#if 0
        if (label.length() == 0) {
            (*i).second = 0;
            unitPhonemes.push_back(IntList());
            graphoneList.push_back(label);
            continue;
        }
#endif
        string::size_type slash = label.find('/');
        assert(slash != string::npos);
        string letters = removeUnderbars(label.substr(0, slash));
        maxLetters = std::max(maxLetters, int(letters.length()));
        IntList l;
        bool phonemeError = false;
        char* phonemes = strdup(label.substr(slash + 1).c_str());
        for (char *save, *phoneme = strtok_r(phonemes, "_", &save);
             phoneme != NULL;
             phoneme = strtok_r(NULL, "_", &save)) {
            StringIntMap::const_iterator j = pMap.find(phoneme);
            if (j == pMap.end()) {
                phonemeError = true;
                l.clear();
                break;
            } else {
                l.push_back((*j).second);
            }
        }
        if (phonemeError) {
            cerr << "phoneme error for unit: " << label << '\n';
            (*i).second = -1;
        } else {
            (*i).second = unit;
            lettersUnitsMap[letters].push_back(unit);
            unitPhonemes.push_back(l);
            graphoneList.push_back(label);
            ++unit;
        }
    }

    IntVector lettersUnitsData;
    IntVector lettersUnitsPositions;
    for (StringIntListMap::const_iterator i = lettersUnitsMap.begin();
         i != lettersUnitsMap.end(); ++i) {
        lettersUnitsPositions.push_back(lettersUnitsData.size());
        const IntList& l = (*i).second;
        for (IntList::const_iterator j = l.begin(); j != l.end(); ++j) {
            lettersUnitsData.push_back(*j);
        }
        lettersUnitsData.push_back(-1);
    }

    IntVector unitPhonemesData;
    IntVector unitPhonemesPositions;
    for (IntListList::const_iterator i = unitPhonemes.begin();
         i != unitPhonemes.end(); ++i) {
        unitPhonemesPositions.push_back(unitPhonemesData.size());
        const IntList& l = *i;
        for (IntList::const_iterator j = l.begin(); j != l.end(); ++j) {
            unitPhonemesData.push_back(*j);
        }
        unitPhonemesData.push_back(-1);
    }

    IntVector unitLabelPositions;
    int unitLabelPos = 0;
#if 0
    unitLabelPositions.push_back(unitLabelPos);
    unitLabelPos += 1;
#endif
    for (StringList::const_iterator i = graphoneList.begin();
         i != graphoneList.end(); ++i) {
        unitLabelPositions.push_back(unitLabelPos);
        unitLabelPos += (*i).length() + 1;
    }

    pad32(fp);
    u32 position = ftell(fp);

    write32(fp, maxLetters);
    write32(fp, lettersUnitsMap.size());
    for (StringIntListMap::const_iterator i = lettersUnitsMap.begin();
         i != lettersUnitsMap.end(); ++i) {
        string letters = (*i).first;
        while ((int)letters.length() < maxLetters) {
            letters += '\0';
        }
        fwrite(letters.c_str(), 1, letters.size(), fp);
    }

    pad32(fp);
    write32(fp, lettersUnitsPositions.size());
    write32(fp, lettersUnitsData.size());
    fwrite(&lettersUnitsPositions[0], 4, lettersUnitsPositions.size(), fp);
    fwrite(&lettersUnitsData[0], 4, lettersUnitsData.size(), fp);

    write32(fp, unitPhonemesPositions.size());
    write32(fp, unitPhonemesData.size());
    fwrite(&unitPhonemesPositions[0], 4, unitPhonemesPositions.size(), fp);
    fwrite(&unitPhonemesData[0], 4, unitPhonemesData.size(), fp);

    cerr << "graphone list size: " << graphoneList.size() << '\n';
    cerr << "graphone data size: " << unitLabelPos << '\n';
    cerr << "graphone positions: " << *graphoneList.begin() << ' ' << *++graphoneList.begin() << " ... " << *graphoneList.rbegin() << '\n';

    write32(fp, graphoneList.size());
    write32(fp, unitLabelPos);
    fwrite(&unitLabelPositions[0], 4, graphoneList.size(), fp);
    for (StringList::const_iterator i = graphoneList.begin();
         i != graphoneList.end(); ++i) {
        string label = *i;
        fwrite(label.c_str(), 1, label.size() + 1, fp);
    }

    pad32(fp);
    u32 ngramPos = ftell(fp);
    cerr << "ngram @ " << ngramPos << '\n';
    int weightBits = 5;
    FSTNgramDumpPacked(ngram, graphoneList, fp, weightBits);

    return position;
}

int
main(int argc, char** argv)
{
    bool error = false;

    char c;
    while ((c = getopt(argc, argv, "")) != EOF) {
        switch (c) {
        default:
            error = true;
            break;
        }
    }
    argc -= optind;
    argv += optind;
    if (error || argc != 4)
        usage();

    const char* lFile = *argv++;
    const char* mcpFile = *argv++;
    const char* l2sFile = *argv++;
    const char* binFile = *argv++;

    FSTGeneric L2S = FSTLoad(l2sFile); // first intentionally
    FSTGeneric L = FSTLoad(lFile);
    FSTGeneric MCP = FSTLoad(mcpFile);
    FSTGeneric MCPm = FSTDynamicAugmentMain(MCP);
    FSTGeneric MCPc = FSTDynamicAugmentClass(MCP);

    FILE* fp = fopen(binFile, "wb");
    write32(fp, DYNAMIC3);
    write32(fp, 0); // size
    u32 mcpPos = 0;
    write32(fp, mcpPos); // placeholder, fix later
    u32 lPos = 0;
    write32(fp, lPos); // placeholder, fix later
    u32 l2sPos = 0;
    write32(fp, l2sPos); // placeholder, fix later

    StringIntMap pMap;
    mcpPos = processMCP(pMap, MCPm, fp);
    processMCP(pMap, MCPc, fp);

    lPos = processL(fp, pMap, L);

    l2sPos = processL2S(fp, L2S, pMap);

    cerr << ftell(fp) << '\n';

    fseek(fp, 0, SEEK_END);
    u32 size = ftell(fp);

    cerr << size << ' ' << mcpPos << ' ' << lPos << ' ' << l2sPos << '\n';

    fseek(fp, 4, SEEK_SET);
    write32(fp, size);
    write32(fp, mcpPos);
    write32(fp, lPos);
    write32(fp, l2sPos);

    fclose(fp);

    return 0;
}



// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 80
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
