////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

/******************************************************************************
=head1 NAME

fst_alphabet - output input or output alphabet

=head1 SYNOPSIS

B<fst_alphabet> (B<-i> | B<-o>) I<fst-in>

=head1 DESCRIPTION

B<fst_alphabet> prints the input or output alphabet of finite-state transducer (FST) I<fst-in> to standard output, one symbol per line.  If epsilon is part of the alphabet, it is not output (i.e., there will be no empty lines).

The specification of the input FST I<fst-in> can be a filename, an expression
(see fst_expression(1)), or "-" indicating standard input.

=head1 OPTIONS

=over 12

=item B<-i>

select the input alphabet

=item B<-o>

select the output alphabet

=back

=head1 SEE ALSO

fst_expression(1)

=head1 AUTHOR

I. Lee Hetherington <ilh@mit.edu>

=cut
******************************************************************************/

#include <cstring>
#include <iostream>
#include "fst.h"

using namespace std;

struct Compare {
    bool operator()(const char* a, const char* b) const {
        return strcmp(a, b) < 0;
    }
};

static void
usage()
{
    cerr <<
        "usage: fst alphabet (-i | -o) in.fst\n"
        "\n"
        "-i     select the input alphabet\n"
        "-o     select the output alphabet\n"
        "\n";
}

static int
fst_alphabet(int argc, char** argv)
{
    if (argc < 0) {
        usage();
        return 1;
    }

    FSTArcOrder which = FSTNone;
    bool error = false;

    char c;
    while ((c = getopt(argc, argv, "io")) != EOF) {
        switch (c) {
        case 'i':
            if (which != FSTNone)
                error = true;
            else
                which = FSTInput;
            break;
        case 'o':
            if (which != FSTNone)
                error = true;
            else
                which = FSTOutput;
            break;
        default:
            error = true;
            break;
        }
    }
    argc -= optind;
    argv += optind;
    if (error || which == FSTNone || argc != 1) {
        usage();
        return 1;
    }

    const char* in_file  = argv[0];
    FSTGeneric x = FSTLoad(in_file);

    const fst::set<int>& alphabet =
        (which == FSTInput) ? x->input_alphabet() : x->output_alphabet();

    // Build up copy that we can sort alphabetically.
    //
    typedef fst::vector<const char*> CharPtrVector;
    CharPtrVector sorted;
    sorted.reserve(alphabet.size());
    for (fst::set<int>::const_iterator i = alphabet.begin();
         i != alphabet.end(); i++) {
        if (*i)
            sorted.push_back(FSTstring_map[*i]);
    }
    
    sort(sorted.begin(), sorted.end(), Compare());

    // Output.
    //
    for (CharPtrVector::iterator j = sorted.begin();
         j != sorted.end(); j++) {
        cout << *j << endl;
    }

    return 0;
}

// Register as subcommand.
//
static FstTool alphabet_tool("alphabet", fst_alphabet);

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 80
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
