////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

/******************************************************************************
=head1 NAME

fst_build_cplg - construct speech recognition FST CoPoLoG

=head1 SYNOPSIS

B<fst_build_cplg> [I<OPTIONS>] I<C-fst> I<P-rules> I<L-fst> I<G-fst> I<result-fst>

=head1 DESCRIPTION

B<fst_build_cplg> computes speech recognition FST CoPoLoG by composing the
components together and optimizing the result.  The optimization includes
epsilon removal, determinization, weight pushing, and minimization.

The specification of the input FSTs I<C-fst>, I<L-fst>, and I<G-fst> can be a
filename, an expression (see fst_expression(1)), or "-" indicating standard
input.  The specification of the output FST I<result-fst> can be a filename or
"-" indicating standard output.  The output I<result-fst> is in binary format.

=head1 INPUTS

=over 12

=item I<C-fst>

FST I<C-fst> represents context-dependent relabeling.  It converts from
context-independent labels on its right side to context-dependent (e.g.,
diphones or triphones) on its left side.  If I<C-fst> equals "None", it is not
used.

=item I<P-rules>

Phonological rules I<P-rules> specify context-dependent phonological rules
mapping from phonemic sequences on its right side to phonetic sequences on its
left side.  See fst_pron_rules(1) for a description of such rules.  If
I<P-rules> equals "None", no rules are applied.

=item I<L-fst>

FST I<L-fst> represents the lexicon mapping word sequences on its right side to
phoneme sequences on its left side.  Note that the lexicon FST must have
closure, meaning that it should allow sequences of words and not just individual
words.  See fst_from_baseforms(1) for a tool to construct such a lexicon FST
given textual phonemic baseforms.  The options -c and -p are highly recommended.

=item I<G-fst>

FST I<G-fst> represents the grammar or language model and places constraints on
allowable word sequences.  In general, it will be an FSA input and output
labels, but that is not a requirement.

=back

=head1 OPTIONS

=over 12

=item B<-b> I<label>

Use I<label> for n-gram backoff.  More documentation on this in the future.

=item B<-d> I<degree>

Allow up to I<degree> of non-determinism at each state for each label.
I<degree> 1 is maximally non-deterministic.  Higher values may result in a
smaller I<result-fst>.  The default is 1.

=item B<-e> I<eps>

Consider two weights whose difference is less than I<eps> to be equal.  Larger
values may allow for a smaller I<result-fst> by collapsing weights that are only
approximately equal.  The default is 0.001.

=back

=head1 SEE ALSO

fst_pron_rules(1), fst_from_baseforms(1), fst_expression(1)

=head1 AUTHOR

I. Lee Hetherington <ilh@mit.edu>

=cut
******************************************************************************/

#include <iostream>
#include "fst.h"
#include "fst/FST.h"
#include "fst/FSTPronRules.h"

using namespace std;

static void
usage()
{
    cerr <<
        "usage: fst build_cplg [options] C.fst P.rules L.fst G.fst out.fst\n"
        "\n"
        "       specify None for C.fst to skip context dependency\n"
        "       specify None for P.rules to skip rule application\n"
        "       specify None for G.fst to skip language model\n"
        "-b lab use label for backoff\n"
        "-d n   let degree n non-determinism remain (default 4)\n"
        "-e x   weight difference considered equivalent (default 0.001)\n"
        "\n";
}

static int
fst_build_cplg(int argc, char** argv)
{
    if (argc < 0) {
        usage();
        return 1;
    }

    const string None = "None";

    int backoff = 0;
    int degree = 4;
    float eps = 0.001F;
    bool error = false;
    char c;
    while ((c = getopt(argc, argv, "b:d:e:")) != EOF) {
        switch (c) {
        case 'b':
            backoff = FSTstring_map[optarg];
            break;
        case 'd':
            degree = atoi(optarg);
            break;
        case 'e':
            eps = atof(optarg);
            break;
        default:
            error = true;
            break;
        }
    }
    argc -= optind;
    argv += optind;
    if (error || argc != 5) {
        usage();
        return 1;
    }

    const char* C_file = *argv++;
    const char* P_file = *argv++;
    const char* L_file = *argv++;
    const char* G_file = *argv++;
    const char* CPLG_file = *argv++;

    // Load lexicon L.
    //
    FSTGeneric x = FSTLoad(L_file);

    // Apply phonological rules P.
    //
    if (P_file != None) {
        cerr << "applying phonological rules: PoL" << endl;
        FSTPronRules rules(P_file);
        if (rules.bad()) {
            cout << "loading " << P_file << " as a fst file" << endl;
            x = FSTBasic(FSTCompose(FSTLoad(P_file), x));
        } else {
            x = rules.apply(x);
        }
    }

    // Initial FSA optimization of PoL (no word label delaying).
    //
    cerr << "removing epsilons" << endl;
    x = FSTBasic(FSTRemoveEpsilons(x));
    cerr << "determinizing (FSA)" << endl;
    x = FSTBasic(FSTDeterminizeA(x));
#if 0
    cerr << "minimizing" << endl;
    x = FSTMinimize(x);
#endif
        
    if (G_file != None) {

        // Compute PoLoG.
        //
        cerr << "composing with G: PoLoG" << endl;
        x = FSTBasic(FSTCompose(x, FSTLoad(G_file)));

    }

    // Final FST optimization of PoLoG.
    //
    cerr << "trimming" << endl;
    x = FSTTrim(x);
    cerr << "removing epsilons (FSA)" << endl;
    x = FSTBasic(FSTRemoveEpsilonsA(x));
    cerr << "determinizing" << endl;
    x = FSTBasic(FSTDeterminize(x, degree, false, eps));
    if (backoff) {
        cerr << "mapping backoff label " << FSTstring_map[backoff] << " to epsilon" << endl;
        FSTGeneric y = FSTBasic();
        y->add_states(x->n_states());
        y->set_initial(x->initial());
        for (int p = 0; p < y->n_states(); p++) {
            if (x->is_final(p)) {
                y->set_final(p, x->final_weight(p));
            }
            for (FSTArcIter a = x->arcs_out(p); a.valid(); a++) {
                int i = a->input;
                if (backoff && i == backoff) {
                    i = 0;
                }
                y->add_arc(p, a->next, i, a->output, a->weight);
            }
        }
        x = y;
    }
    cerr << "pushing weights" << endl;
    x = FSTPushWeights(x);
    cerr << "minimizing" << endl;
    x = FSTMinimize(x, eps);

    if (C_file != None) {

        // Compute CoPoLoG.
        //
        cerr << "composing with C: CoPoLoG" << endl;
        x = FSTMinimize(FSTBasic(FSTCompose(FSTLoad(C_file), x)));

    }
        
    // Dump output.
    //
    x->dump_binary(CPLG_file);

    return 0;
}

// Register as subcommand.
//
static FstTool build_cplg_tool("build_cplg", fst_build_cplg);

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 80
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
