////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

/******************************************************************************
=head1 NAME

fst_build_dynamic_class - construct FST that can be used as a dynamic class

=head1 SYNOPSIS

B<fst_build_dynamic_class> [B<-s>] I<class-name> I<L-fst> I<L2S-fst> I<CP-fst>
I<nickname-fst> I<out-fst>

B<fst_build_dynamic_class> I<class-name> I<CP-fst> I<in-fst> I<out-fst>

=head1 DESCRIPTION

B<fst_build_dynamic_class> constructs an FST that can be loaded into an
FSTDynamic, which allows for associating a class name with a particular FST to
be spliced in on-demand while respecting context constraints.

The first form reads a list of words or phrases from the standard input,
applies "nickname" expansion, looks words up in the provided dictionary,
optionally runs letter-to-sound generation on those words not found in the
dictionary, composes with CoP, and brackets the output with class name tags.

The second form composes CoP with the input FST and brackets the output with
class name tags.

This needs more documentation!

=head1 INPUTS

=over 12

=item I<class-name>

String I<class-name> will be used to bracket the contents of the dynamic
class FST with <I<class-name>> ... </I<class-name>>.

=item I<L-fst>

FST I<L-fst> represents a dictionary mapping words or phrases (phrases can have
underbars, e.g., "lets_go") on the right side to phonemes on the left side.  It
must have closure.  If I<L-fst> is "None", no dictionary lookup will occur, and
you must use I<L2S-fst> to provide word pronunciations.

=item I<L2S-fst>

FST I<L2S-fst> maps letter sequences to phoneme sequences.  If I<L2S-fst> is
"None", no letter-to-sound generation will occur, and any words not found in the
dictionary will be dropped.

=item I<CP-fst>

FST I<CP-fst> represents CoP, the composition of context-dependent relabeling
and phonological rules.

=item I<nickname-fst>

FST I<nickname-fst> can contain expansions of "words" into other words or
phrases representing aliases or nicknames.

=item I<in-fst>

FST I<in-fst> represents the contents of the dynamic class down to the phonemic
level (i.e., without CoP applied).

=back

=head1 OPTIONS

=over 12

=item B<-s>

Enable optional spelling mode.  This is very experimental.

=back

=head1 SEE ALSO

fst_dynamic_augment(1)

=head1 AUTHOR

I. Lee Hetherington <ilh@mit.edu>

=cut
******************************************************************************/

#include <cstring>
#include <iostream>
#include "fst.h"
#include "fst/FST.h"
#include "fst/FSTDynamicClass.h"

using namespace std;

static void
usage()
{
    cerr <<
        "usage: fst build_dynamic_class [-s] class-name dict.fst l2s.fst cp.fst nickname.fst out.fst\n"
        "       fst build_dynamic_class [-o] class-name cp.fst in.fst out.fst\n"
        "\n"
        "-o     do not optimize\n"
        "-s     enable optional word speak and spell\n"
        "\n";
}

static int
fst_build_dynamic_class(int argc, char** argv)
{
    if (argc < 1) {
        usage();
        return 1;
    }

    bool no_optimize_flag = false;
    bool spelling_flag = false;
    bool error = false;

    char c;
    while ((c = getopt(argc, argv, "os")) != EOF) {
        switch (c) {
        case 'o':
            no_optimize_flag = true;
            break;
        case 's':
            spelling_flag = true;
            break;
        default:
            error = true;
        }
    }
    argc -= optind;
    argv += optind;

    if (argc == 6) {

        const char* class_name = *argv++;
        const char* dict_file = *argv++;
        const char* l2s_file = *argv++;
        const char* cp_file = *argv++;
        const char* nickname_file = *argv++;
        const char* out_file = *argv++;
        
        FSTDynamicClass dynamic(dict_file, l2s_file, cp_file, nickname_file);
        
        fst::vector<char*> contents;
        string line;
        while (getline(cin, line)) {
            contents.push_back(strdup(line.c_str()));
        }
        
        dynamic.update(true, class_name, contents.size(), (const char**) &contents[0], spelling_flag)
            ->dump_binary(out_file);
        
        for (fst::vector<char*>::iterator i = contents.begin(); i != contents.end(); i++)
            free(*i);

    } else if (argc == 4) {

        const char* class_name = *argv++;
        const char* cp_file = *argv++;
        const char* in_file = *argv++;
        const char* out_file = *argv++;

        FSTDynamicClass dynamic("None", "None", cp_file, "None");
        dynamic.replace(class_name, FSTLoad(in_file), no_optimize_flag)->dump_binary(out_file);

    } else {

        usage();
        return 1;

    }

    return 0;
}

// Register as subcommand.
//
static FstTool build_dynamic_class_tool("build_dynamic_class", fst_build_dynamic_class);

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 80
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
