////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

/******************************************************************************
=head1 NAME

fst_collect - collapse arcs with identical origin, destination, input, and output

=head1 SYNOPSIS

B<fst_collect> [B<-B>] I<fst-in> I<fst-out>

=head1 DESCRIPTION

B<fst_collect> collapses arcs with identical origin, destination, input, and
output into a single arc with the weights summed according to the semiring.

The specification of the input FST I<fst-in> can be a filename, an expression
(see fst_expression(1)), or "-" indicating standard input.  The specification of
the output FST I<fst-out> can be a filename or "-" indicating standard output.

=head1 OPTIONS

=over 12

=item B<-B>

Output I<fst-out> in binary format.

=head1 AUTHOR

I. Lee Hetherington <ilh@mit.edu>

=cut
******************************************************************************/

#include <iostream>
#include "fst.h"

using namespace std;

static void
usage()
{
    cerr <<
        "usage: fst_collect [-B] in.fst out.fst\n"
        "\n"
        "-B     binary output format\n"
        "\n";
}

struct Entry {
    int next, input, output;
    Entry(int n, int i, int o) : next(n), input(i), output(o) {}
};

#define FST_LESS2(a, b)     ((a) < (b))
#define FST_LESS3(a, b, c)  ((a) < (b) || ((a) == (b) && (c)))

namespace std {
    template<> struct less<Entry> {
        bool operator()(const Entry& x, const Entry& y) const {
            return FST_LESS3(x.next, y.next,
                             FST_LESS3(x.input, y.input,
                                       FST_LESS2(x.output, y.output)));
        }
    };
}

typedef fst::map<Entry,float> Map;

static FSTGeneric
FSTCollect(const FSTGeneric& in)
{
    FSTSemiring semiring = in->semiring();
    in->force();
    int N = in->n_states();

    FSTGeneric out = FSTBasic(in->semiring());
    out->add_states(N);
    out->set_initial(in->initial());

    Map m;
    for (int p = 0; p < N; p++) {
        out->set_final(p, in->final_weight(p));
        for (FSTArcIter a = in->arcs_out(p); a.valid(); ++a) {
            Entry e(a->next, a->input, a->output);
            pair<Map::iterator,bool> status = m.insert(Map::value_type(e, a->weight));
            if (!status.second)
                (*status.first).second = semiring->plus((*status.first).second, a->weight);
        }
        for (Map::const_iterator i = m.begin(); i != m.end(); ++i)
            out->add_arc(p, (*i).first.next, (*i).first.input, (*i).first.output, (*i).second);
        m.clear();
    }

    return out;
}

static int
fst_collect(int argc, char** argv)
{
    if (argc < 0) {
        usage();
        return 1;
    }

    bool binary = false;

    bool error = false;
    char c;
    while ((c = getopt(argc, argv, "B")) != EOF) {
        switch (c) {
        case 'B':
            binary = true;
            break;
        default:
            error = true;
            break;
        }
    }
    argc -= optind;
    argv += optind;

    if (error || argc != 2) {
    usage();
        return 1;
    }

    const char* in_file  = argv[0];
    const char* out_file = argv[1];

    FSTGeneric x = FSTLoad(in_file);
    FSTGeneric y = FSTCollect(x);

    if (binary)
        y->dump_binary(out_file);
    else
        y->dump_ascii(out_file);

    return 0;
}

// Register as subcommand.
//
static FstTool collect_tool("collect", fst_collect);

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 80
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
