////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

#include <iostream>
#include "fst.h"
#include "fst/format.h"

using namespace std;

typedef fst::vector<bool>  BoolVector;
typedef fst::set<int>      IntSet;
typedef fst::vector<IntSet> IntSetVector;

static void compare(const FSTGeneric& a, const FSTGeneric& b);
static void compare_dfs(const FSTGeneric& a, BoolVector& a_visited,
                        IntSetVector& a_matches, int a_state,
                        const FSTGeneric& b, BoolVector& b_visited,
                        IntSetVector& b_matches, int b_state);

static void
usage()
{
    cerr << "usage: fst compare a.fst b.fst\n\n";
}

static int
fst_compare(int argc, char** argv)
{
    if (argc < 0) {
        usage();
        return 1;
    }

    argc--;
    argv++;

    if (argc != 2) {
    usage();
        return 1;
    }

    const char* a_file = argv[0];
    const char* b_file = argv[1];

    FSTGeneric a = FSTLoad(a_file);
    FSTGeneric b = FSTLoad(b_file);

    compare(a, b);

    return 0;
}

static ostream&
operator << (ostream& os, const IntSet& s)
{
    for (IntSet::const_iterator i = s.begin(); i != s.end(); i++)
        os << " " << *i;
    return os;
}

static void
compare(const FSTGeneric& a, const FSTGeneric& b)
{
    a->force();
    b->force();

    int n_a = a->n_states();
    int n_b = b->n_states();

    BoolVector a_visited(n_a);
    BoolVector b_visited(n_b);

    IntSetVector a_matches(n_a);
    IntSetVector b_matches(n_b);

    compare_dfs(a, a_visited, a_matches, a->initial(),
                b, b_visited, b_matches, b->initial());

    int i;
    for (i = 0; i < n_a; i++) {
        if (a_matches[i].size() > 1)
            cout << "*** ";
        cout << "a " << i << " :" << a_matches[i] << "\n";
    }
    for (i = 0; i < n_b; i++) {
        if (b_matches[i].size() > 1)
            cout << "*** ";
        cout << "b " << i << " :" << b_matches[i] << "\n";
    }
}

static void
compare_dfs(const FSTGeneric& a, BoolVector& a_visited,
            IntSetVector& a_matches, int a_state,
            const FSTGeneric& b, BoolVector& b_visited,
            IntSetVector& b_matches, int b_state)
{
    a_matches[a_state].insert(b_state);
    b_matches[b_state].insert(a_state);
    if (!a_visited[a_state]) {
        a_visited[a_state] = true;
        for (FSTArcIter a_arcs = a->arcs_out(a_state, FSTInput);
             a_arcs.valid(); a_arcs++) {
            FSTArcIter b_arcs = b->arcs_out_with_input(b_state,
                                                       a_arcs->input);
            if (b_arcs.n() == 0) {
                format(cout, "a(%d, %d, %s, %s) missing for b(%d)\n",
                       a_state, a_arcs->next,
                       FSTstring_map[a_arcs->input],
                       FSTstring_map[a_arcs->output],
                       b_state);
            }
            for (; b_arcs.valid(); b_arcs++) {
                if (a_arcs->output != b_arcs->output)
                    format(cout, "a(%d, %d, %s, %s) != b(%d, %d, %s, %s)\n",
                           a_state, a_arcs->next,
                           FSTstring_map[a_arcs->input],
                           FSTstring_map[a_arcs->output],
                           b_state, b_arcs->next,
                           FSTstring_map[b_arcs->input],
                           FSTstring_map[b_arcs->output]);
                compare_dfs(a, a_visited, a_matches, a_arcs->next,
                            b, b_visited, b_matches, b_arcs->next);
            }
        }
    }
}

// Register as subcommand.
//
static FstTool compare_tool("compare", fst_compare);

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 80
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:

