////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

/******************************************************************************
=head1 NAME

fsa_complement - compute the complement of a finite-state acceptor

=head1 SYNOPSIS

B<fsa_complement> [I<OPTIONS>] I<fst-in> I<fsa-out>

=head1 DESCRIPTION

B<fsa_complement> extracts a finite-state acceptor (FSA) from either the input
or output side of the input finite-state transducer (FST) I<fst-in> and then
computes its complement.  The complement FSA accepts the set of strings that is
the complement of the set of strings accepted by the input FSA.

The specification of the input FST I<fst-in> can be a filename, an expression
(see fst_expression(1)), or "-" indicating standard input.  The specification of
the output FSA I<fsa-out> can be a filename or "-" indicating standard output.

The output FSA is an FST with equal input and output labels on each arc.

=head1 OPTIONS

=over 12

=item B<-a> I<file>

Supplement the alphabet contained in I<fst-in> with that contained in I<file>,
where alphabet symbols are listed one per line.

=item B<-o>

Extract the FSA from the output side of I<fst-in>.  The default is to extract
from the input side.

=back

=head1 SEE ALSO

fst_expression(1)

=head1 AUTHOR

I. Lee Hetherington <ilh@mit.edu>

=cut
******************************************************************************/

#include <iostream>
#include <fstream>
#include "fst.h"

using namespace std;

static void
usage()
{
    cerr <<
        "usage: fst complement [options] in.fst out.fst\n"
        "\n"
        "-a F   supplement alphabet with list of symbols in file F\n"
        "-o     extract the FSA from the output side (input default)\n"
        "\n"
        "Note:  only works for FSA and not FST, which is why only the\n"
        "       input or output side is processed.\n"
        "\n";
}

static void
load_alphabet(fst::set<int>& alphabet, const char* filename)
{
    ifstream is(filename);
    string line;
    while (getline(is, line)) {
        if (line.length() > 0) {
            alphabet.insert(FSTstring_map[line]);
        }
    }
}

static FSTGeneric
complement(const FSTGeneric& x, const fst::set<int>& sigma)
{
    int N = x->n_states();

    fst::set<int>::const_iterator si;

    FSTGeneric y = FSTBasic(x->semiring());
    y->add_states(N + 1);

    y->set_initial(x->initial());
    y->set_final(N);

    for (si = sigma.begin(); si != sigma.end(); ++si) {
        for (int p = 0; p < N; p++) {
            FSTArcIter a = x->arcs_out_with_input(p, *si);
            int q = a.valid() ? a->next : N;
            y->add_arc(p, q, *si, *si);
        }
        y->add_arc(N, N, *si, *si);
    }
    for (int p = 0; p < N; p++)
        if (!x->is_final(p))
            y->set_final(p);

    return y;
}

static int
fsa_complement(int argc, char** argv)
{
    if (argc < 0) {
        usage();
        return 1;
    }

    FSTArcOrder side = FSTInput;
    bool error = false;
    char* aFile = NULL;

    char c;
    while ((c = getopt(argc, argv, "a:o")) != EOF) {
        switch (c) {
        case 'a':
            aFile = optarg;
            break;
        case 'o':
            side = FSTOutput;
            break;
        default:
            error = true;
            break;
        }
    }
    argc -= optind;
    argv += optind;
    if (error || argc != 2) {
        usage();
        return 1;
    }

    const char* in_file  = *argv++;
    const char* out_file = *argv++;


    FSTGeneric x = FSTLoad(in_file);
    x = FSTProject(x, side, side);

    fst::set<int> alphabet(x->input_alphabet());
    if (aFile && aFile[0]) {
        load_alphabet(alphabet, aFile);
    }
    x = FSTClearWeights(x);
    x = FSTOptimizeA(x);
    x = complement(x, alphabet);
    x = FSTMinimize(x);
    x->dump_ascii(out_file);

    return 0;
}

// Register as subcommand.
//
static FstTool complement_tool("complement", fsa_complement);

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 80
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
