////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

/******************************************************************************
=head1 NAME

fst_compose - compute FST composition

=head1 SYNOPSIS

B<fst_compose> [I<OPTIONS>] I<A-fst> I<B-fst> I<out-fst>

=head1 DESCRIPTION

B<fst_compose> computes the composition AoB from I<A-fst> and I<B-fst> and
outputs to I<out-fst>.

The specification of the input FSTs I<A-fst> and I<B-fst> can be a filename, an
expression (see fst_expression(1)), or "-" indicating standard input.  The
specification of the output FST I<out-fst> can be a filename or "-" indicating
standard output.

=head1 OPTIONS

=over 12

=item B<-B>

Save I<out-fst> in binary format.

=item B<-i>

Order arcs by input (internal index) order, which is the default.

=item B<-l>

Set the input side of an output arc from the output side of the left FST
(I<A-fst>) arc.  This does not compute the standard composition.  For
example, computing a:b o b:c will output b:c instead of a:c.

=item B<-o>

Order arcs by output (internal index) order.  This is ignored if B<-B>
is specified, and is generally of dubious usefulness.

=item B<-r>

Set the output side of an output arc from the input side of the right FST
(I<B-fst>) arc.  This does not compute the standard composition.  For
example, computing a:b o b:c will output b:c instead of a:c.

=item B<-t>

Trim dead states from result, equivalent to running fst_trim(1) on I<out-fst>.

=back

=head1 SEE ALSO

fst_trim(1)

=head1 AUTHOR

I. Lee Hetherington <ilh@mit.edu>

=cut
******************************************************************************/

#include <iostream>
#include "fst.h"

using namespace std;

static void
usage()
{
    cerr <<
        "usage: fst_compose [options] x.fst y.fst out.fst\n"
        "\n"
        "-B     save in FSTBinary format\n"
        "-i     order arcs by input [default]\n"
        "-l     project output of left FST to input\n"
        "-o     order arcs by output (ignored with -B)\n"
        "-r     project input of right FST to output\n"
        "-t     trim dead states from result (can't stream)\n"
        "\n";
}

static int
fst_compose(int argc, char** argv)
{
    if (argc < 0) {
        usage();
        return 1;
    }

    FSTArcOrder order = FSTInput;
    bool trim = false;
    bool binary = false;
    bool error = false;
    bool lp = false;
    bool rp = false;
    char c;
    while ((c = getopt(argc, argv, "Bilrot")) != EOF) {
        switch (c) {
        case 'B':
            binary = true;
            break;
        case 'i':
            order = FSTInput;
            break;
        case 'l':
            lp = true;
            break;
        case 'o':
            order = FSTOutput;
            break;
        case 'r':
            rp = true;
            break;
        case 't':
            trim = true;
            break;
        case '?':
        default:
            error = true;
            break;
        }
    }
    argc -= optind;
    argv += optind;
    if (error || argc != 3) {
        usage();
        return 1;
    }

    const char* x_file = argv[0];
    const char* y_file = argv[1];
    const char* z_file = argv[2];

    FSTGeneric x = FSTLoad(x_file);
    if (lp)
        x = FSTProject(x, FSTOutput, FSTOutput);
    FSTGeneric y = FSTLoad(y_file);
    if (rp)
        y = FSTProject(y, FSTInput, FSTInput);
    FSTGeneric z = FSTCompose(x, y);

    if (trim)
        z = FSTTrim(z);

    if (binary) {
        z->dump_binary(z_file);
    } else {
        z->dump_ascii(z_file, order);
    }

    return 0;
}

// Register as subcommand.
//
static FstTool compose_tool("compose", fst_compose);

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 79
// indent-tabs-mode: nil
// c-basic-offset: 4
// End:
