////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

#include <iostream>
#include "fst.h"

using namespace std;

static void
usage()
{
    cerr <<
        "usage: fst_compose2 [options] x.fst a.fst b.fst y.fst out.fst\n"
        "\n"
        "-B     save in FSTBinary format\n"
        "-l     project output of left FST to input\n"
        "-r     project input of right FST to output\n"
        "-t     trim dead states from result (can't stream)\n"
        "-p     to optimize (default no optimizing)\n"
        "-a     FSA-style optimization for output\n"
        "\n";
}

static int
fst_compose2(int argc, char** argv)
{
    if (argc < 0) {
        usage();
        return 1;
    }

    bool trim = false;
    bool binary = false;
    bool error = false;
    bool lp = false;
    bool rp = false;
    bool fsa = false;
    bool optimize = false;
    char c;
    while ((c = getopt(argc, argv, "Blrtap")) != EOF) {
    switch (c) {
        case 'B':
            binary = true;
            break;
        case 'l':
            lp = true;
            break;
        case 'r':
            rp = true;
            break;
        case 't':
            trim = true;
            break;
        case 'a':
            fsa = true;
            break;
        case 'p':
            optimize = true;
            break;
    case '?':
    default:
        error = true;
        break;
    }
    }
    argc -= optind;
    argv += optind;
    if (error || argc != 5) {
        usage();
        return 1;
    }

    const string None = "None";
    const char* x_file = argv[0];
    const char* a_file = argv[1];
    const char* b_file = argv[2];
    const char* y_file = argv[3];
    const char* z_file = argv[4];

    if ((x_file == None && a_file == None) ||
        (b_file == None && y_file == None))
        usage();

    FSTGeneric l;
    if(x_file != None && a_file != None) {
        FSTGeneric x = FSTLoad(x_file);
        FSTGeneric a = FSTLoad(a_file);
        l = FSTCompose(x, a);
        if (trim)
            l = FSTTrim(l);
        l = FSTOptimizeA(l);
    } else if(x_file != None)
        l = FSTLoad(x_file);
    else if(a_file != None)
        l = FSTLoad(a_file);
    if (lp)
        l = FSTProject(l, FSTOutput, FSTOutput);

    FSTGeneric r;
    if(b_file != None && y_file != None) {
        FSTGeneric b = FSTLoad(b_file);
        FSTGeneric y = FSTLoad(y_file);
        r = FSTCompose(b, y);
        if (trim)
            r = FSTTrim(r);
        r = FSTOptimizeA(r);
    } else if(b_file != None)
        r = FSTLoad(b_file);
    else if(y_file != None)
        r = FSTLoad(y_file);
    if (rp)
        r = FSTProject(r, FSTInput, FSTInput);

    FSTGeneric z = FSTCompose(l, r);
    if (trim)
        z = FSTTrim(z);
    if (optimize)
        if (fsa)
            z = FSTOptimizeA(z);
        else
            z = FSTOptimize(z);

    if (binary)
        z->dump_binary(z_file);
    else
        z->dump_ascii(z_file);

    return 0;
}

// Register as subcommand.
//
static FstTool compose2_tool("compose2", fst_compose2);

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 79
// indent-tabs-mode: nil
// c-basic-offset: 4
// End:
