////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

/******************************************************************************
=head1 NAME

fst_determinize - determinize FST

=head1 SYNOPSIS

B<fst_determinize> [I<OPTIONS>] I<fst-in> I<fst-out>

=head1 DESCRIPTION

B<fst_determinize> determinizes the input FST I<fst-in> to produce the output
FST I<fst-out>.  Options are provided to treat the FST as an FSA, allow for some
degree of non-determinism to be left non-determinisitic, and threshold for
determining weight equality.

In general, due to weight and output ambiguity, it is not possible to fully
determinize an FST.  To combat output ambiguity, the special input symbol "#"
can be used to force an output symbol rather than allowing it to be delayed
indefinitely.  For example, in a dictionary FST with phonemes on the input side
and words on the output side, with closure, determinization in the face of
homophones can cause the standard determinization algorithm to not terminate.
However, in this implementation, if the closure uses an arc labelled #:epsilon
between words, homophone word labels will be pushed back as far as the # arc and
no further.

In general determinization, each state is associated with a set of (state,
output sequence, weight) tuples.  A limitation of this implementation, for
efficiency, is that the pending output sequence can contain only 0 or 1
symbols.  If a longer pending output sequence would be required, epsilon-input
arcs will be admitted to force out pending output symbols.  In addition, various
heuristics are used to attempt to detect weight ambiguity and again output
epsilon-input arcs to avoid nontermination.

The specification of the input FST I<fst-in> can be a filename, an expression
(see fst_expression(1)), or "-" indicating standard input.  The specification of
the output FST I<fst-out> can be a filename or "-" indicating standard output.

=head1 OPTIONS

=over 12

=item B<-B>

Output I<fst-out> in binary format.

=item B<-a>

Treat I<fst-in> as an FSA with compound input:output labels.  No output pushing
will take place.

=item B<-d> I<N>

Allow nondeterminism of degree I<N> through.  For example, if there would be I<N>
or fewer arcs with identical input labels leaving a particular output state, they
would not be subject to determinization.  This can be used to make size/determinism
tradeoffs.  The default value of <N> is 1.

=item B<-e> I<thresh>

Treat two weights as equal if their difference is less than threshold I<thresh>.  The
default value of I<thresh> is 0.001.

=item B<-s>

Force strict determinism.  This currently turns off a weight ambiguity heuristic.
This is experimental.

=back

=head1 ALGORITHM

The algorithm is based on that published by Mehyar Mohri while he was at AT&T
[insert reference], and extended by the author.

=head1 AUTHOR

I. Lee Hetherington <ilh@mit.edu>

=cut
******************************************************************************/

#include <iostream>
#include "fst.h"

using namespace std;

static void
usage()
{
    cerr <<
        "usage: fst determinize [options] in.fst out.fst\n"
        "\n"
        "-B      binary output format\n"
        "-a      treat as FSA with compound input:output labels\n"
        "-d N    degree of non-determinism allowed [1]\n"
        "-e x    difference allowed when considering weights to be equal [0.001]\n"
        "-s      strict determinism (currently turns off weight ambiguity heuristic)\n"
        "\n";
}

static int
fst_determinize(int argc, char** argv)
{
    if (argc < 0) {
        usage();
        return 1;
    }

    bool binary = false;
    float eps = float(0.001);
    bool fsa = false;
    int degree = 1;
    bool strict = false;

    bool error = false;
    char c;
    while ((c = getopt(argc, argv, "Bad:e:s")) != EOF) {
        switch (c) {
        case 'B':
            binary = true;
            break;
        case 'a':
            fsa = true;
            break;
        case 'd':
            degree = atoi(optarg);
            break;
        case 'e':
            eps = atof(optarg);
            break;
        case 's':
            strict = true;
            break;
        default:
            error = true;
            break;
        }
    }
    argc -= optind;
    argv += optind;

    if (error || argc != 2) {
    usage();
        return 1;
    }

    const char* in_file  = argv[0];
    const char* out_file = argv[1];

    FSTGeneric x = FSTLoad(in_file);
    FSTGeneric y = FSTDeterminize(x, degree, fsa, eps, strict);

    if (binary)
        FSTBasic(y)->dump_binary(out_file);
    else
        y->dump_ascii(out_file);

    return 0;
}

// Register as subcommand.
//
static FstTool determinize_tool("determinize", fst_determinize);

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 80
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
