////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2008, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

// Augment an FST to make it compatible with FSTDynamic for dynamic splicing of
// FSTs.  For a dynamic FST to be spliced into the top-level FST, it must start
// and end with $, which can be accomplished by concatenating $ dynamic-FST $.
// For the top-level FST, the $s are optional; they are not needed, but they
// don't really hurt either.  The net result is the insertion of #i arcs on #
// transitions at the edges of top-level and dynamic splices.

#include <iostream>
#include <cstdio>
#include "fst.h"

using namespace std;

typedef fst::map<int,int> Map;

static void
usage()
{
    cerr <<
        "usage: [-c | -m] in.fst out.fst\n"
        "\n"
        "-c     class FST mode\n"
        "-m     main FST mode\n"
        "\n";
}

static int
constraint_label(int p)
{
    char buf[16];
    sprintf(buf, "#%d", p);
    return FSTstring_map[buf];
}

static FSTGeneric
augment_original(const FSTGeneric& in)
{
    const int dollar = FSTstring_map["$"];
    const int pound = FSTstring_map["#"];
    const int star = FSTstring_map["#star"];

    in->force();
    const int N = in->n_states();

    FSTGeneric out = FSTBasic(in);

    // New final state.
    //
    int new_final = out->add_state();
    out->set_final(new_final);

    // Two new states with #star:#star transition between them.  Later, we'll
    // add #i:# arcs into star1, and #i:# arcs out of star2.
    //
    int star1 = out->add_state();
    int star2 = out->add_state();
    out->add_arc(star1, star2, star, star);
    out->add_arc(star2, new_final, 0, dollar);
    out->add_arc(star2, star1, pound, pound);

    // New initial state.  There are additional #i:$ arcs added after sweeping
    // through all the states.
    //
    int old_initial = out->initial();
    int new_initial = out->add_state();
    out->set_initial(new_initial);
    for (FSTArcIter a = out->arcs_out(old_initial); a.valid(); ++a) {
        out->add_arc(new_initial, a->next, a->input, a->output, a->weight);
    }
    out->add_arc(new_initial, old_initial, 0, dollar);
    out->add_arc(new_initial, star1, 0, dollar);
    out->add_arc(new_initial, star2, star, star);

    // Process all original states.
    //
    fst::vector<bool> to_state_finished(N, false);
    for (int p = 0; p < N; ++p) {
        if (out->is_final(p)) {
            out->add_arc(p, new_final, 0, dollar, out->final_weight(p));
        }
        for (FSTArcIter a = in->arcs_out_with_output(p, pound);
             a.valid(); ++a) {
            int q = a->next;
            int pound_q = constraint_label(q);
            out->add_arc(p, star1, pound_q, pound, a->weight);
            if (out->is_final(p)) {
                out->add_arc(p, new_final, pound_q, dollar, out->final_weight(p));
            }
            if (!to_state_finished[q]) {
                out->add_arc(star2, q, pound_q, pound);
                out->add_arc(new_initial, q, pound_q, dollar);
                to_state_finished[q] = true;
            }
        }
    }

    // For every arc star2 -> i #i:#, add an arc new_initial -> i #i:$.
    //
    for (FSTArcIter a = out->arcs_out(star2); a.valid(); ++a) {
        if (a->output == pound) {
            out->add_arc(new_initial, a->next, a->input, dollar);
        }
    }

    return out;
}

static int
fst_dynamic_augment(int argc, char** argv)
{
    bool error = false;
    bool m_flag = false;
    bool c_flag = false;

    char c;
    while ((c = getopt(argc, argv, "mc")) != EOF) {
        switch (c) {
        case 'c':
            c_flag = true;
            break;
        case 'm':
            m_flag = true;
            break;
        default:
            error = true;
            break;
        }
    }
    argc -= optind;
    argv += optind;
    if (error || argc != 2 || m_flag && c_flag) {
        usage();
        return 1;
    }

    const char* in_file  = argv[0];
    const char* out_file = argv[1];

    FSTGeneric in = FSTLoad(in_file);
    FSTGeneric out;

    if (m_flag) {
        out = FSTDynamicAugmentMain(in);
    } else if (c_flag) {
        out = FSTDynamicAugmentClass(in);
    } else {
        out = augment_original(in);
    }

    out->dump_binary(out_file);

    return 0;
}

// Register as subcommand.
//
static FstTool dynamic_augment_tool("dynamic_augment", fst_dynamic_augment);

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 80
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:



