////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

/******************************************************************************
=head1 NAME

fst_epsilon_cycles - detect if an FST has any epsilon cycles

=head1 SYNOPSIS

B<fst_epsilon_cycles> I<fst-in>

=head1 DESCRIPTION

B<fst_epsilon_cycles> detects any epsilon cycles in FST I<fst-in>.  If none are
found, the return value is 0 and there is nothing is printed.  If at least one
is found, the return value is 1, and the list of states of the first cycle found
are printed to standard output.

The specification of the input FST I<fst-in> can be a filename, an expression
(see fst_expression(1)), or "-" indicating standard input.

=head1 OPTIONS

None.

=head1 AUTHOR

I. Lee Hetherington <ilh@mit.edu>

=cut
******************************************************************************/

#include <iostream>
#include "fst.h"

using namespace std;

static const int white = 0;
static const int gray = 1;
static const int black = 2;

static const int acyclic = -1;
static const int cyclic = -2;

typedef fst::map<int, int> IntIntMap;
typedef fst::list<int>     IntList;

static void
usage()
{
    cerr << "usage: fst epsilon_cycles in.fst\n\n";
}

static int
dfs(IntList& result, IntIntMap& color, const FSTGeneric& x, int p)
{
    int& p_color = color[p];
    if (p_color == white) {
    // unvisited
    p_color = gray;
    } else if (p_color == gray) {
    // cycle
    return p;
    } else {
    // finished
    return acyclic;
    }
    for (FSTArcIter a = x->arcs_out_with_input(p, 0); a.valid(); a++) {
    if (a->next == p) {
        // self-loop cycle detected
        result.push_front(p);
        return cyclic;
    }

    int q = dfs(result, color, x, a->next);
    if (q == cyclic)
        return cyclic;
    if (q >= 0) {
        // cycle detected
        result.push_front(p);
        if (q == p)
        return cyclic;
        else
        return q;
    }
    }
    p_color = black;
    return acyclic;
}

static bool
detect_epsilon_cycles(IntList& result, FSTGeneric x)
{
    IntIntMap color;
    for (int p = 0; p < x->n_states(); p++) {
    int status = dfs(result, color, x, p);
    assert(status < 0);
    if (status == cyclic)
        return true;
    assert(status == acyclic);
    }
    return false;
}

static int
fst_epsilon_cycles(int argc, char** argv)
{
    if (argc < 0) {
        usage();
        return 1;
    }

    argc--;
    argv++;

    if (argc != 1) {
    usage();
        return 1;
    }

    const char* in_file  = argv[0];

    FSTGeneric x = FSTLoad(in_file);

    IntList result;
    detect_epsilon_cycles(result, x);

    if (!result.empty()) {
    cout << "cycle:";
    for (IntList::const_iterator i = result.begin(); i != result.end(); i++)
        cout << ' ' << *i;
    cout << endl;
    return 1;
    } else {
    return 0;
    }
}

// Register as subcommand.
//
static FstTool epsilon_cycles_tool("epsilon_cycles", fst_epsilon_cycles);

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 80
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
