/* -*- C++ -*-
 *
 * Copyright (c) 2007
 * Spoken Language Systems Group
 * MIT Computer Science and Artificial Intelligence Laboratory
 * Massachusetts Institute of Technology
 *
 * All Rights Reserved
 */
#include <iostream>
#include <fstream>
#include "fst.h"

using namespace std;

typedef set<int> Set;
typedef vector<int> Sequence;
typedef list<Sequence> SequenceList;

static void
usage()
{
    cerr <<
        "usage: [options] in.txt out.fst\n"
        "\n"
        "-B               binary output\n"
        "-s <file>        stop words\n"
        "-t               tag indices\n"
        "\n";
}

static void
build_stop_word_set(Set& out, const char* filename)
{
    out.clear();

    ifstream is(filename);
    string line;
    string::size_type i, j;
    while (!getline(is, line).fail()) {
        i = line.find_first_not_of(" \t\r\n");
        j = line.find_last_not_of(" \t\r\n");
        if (i != string::npos && j != string::npos) {
            out.insert(FSTstring_map[line.substr(i, j - i + 1)]);
        }
    }
}

static void
build_sequence(Sequence& out, const string& text)
{
    out.clear();

    const char* ws = " \t\r\n";
    string::size_type i, j;

    for (j = 0; j != string::npos && (i = text.find_first_not_of(ws, j)) != string::npos;) {
        j = text.find_first_of(ws, i);
        string word;
        if (j == string::npos) {
            word = text.substr(i);
        } else {
            word = text.substr(i, j - i);
        }
        out.push_back(FSTstring_map[word]);
    }
}

static void
build_sequences(SequenceList& out, const char* filename, Set& stop_word_set)
{
    out.clear();

    ifstream is(filename);
    string line;
    while (!getline(is, line).fail()) {
        Sequence sequence;
        build_sequence(sequence, line);
        out.push_back(sequence);
    }
}

static FSTGeneric
build_untagged(const SequenceList& sequences, const Set& stop_word_set)
{
    FSTGeneric fst = FSTBasic();
    fst->add_state();
    fst->set_initial(0);
    
    for (SequenceList::const_iterator i = sequences.begin(); i != sequences.end(); ++i) {
        const Sequence& sequence = *i;
        int p = fst->add_state();
        for (unsigned j = 0; j < sequence.size(); ++j) {
            int word = sequence[j];
            int q = fst->add_state();
            if (j < sequence.size() - 1 || stop_word_set.count(word) == 0) {
                fst->add_arc(0, p);
            }
            fst->add_arc(p, q, word);
            if (stop_word_set.count(word) == 0) {
                // Allow finality only if word is not a stop word.
                fst->set_final(q);
            }
            p = q;
        }
        // Allow finality at end of sequence, even if stop word.
        fst->set_final(p);
    }

    fst = FSTOptimizeA(fst);

    // Eliminate zero-length sequence.
    //
    fst->clear_final(0);

    return fst;
}

static FSTGeneric
build_tagged(const SequenceList& sequences, const Set& stop_word_set)
{
    FSTGeneric fst = FSTBasic();
    fst->add_states(2);
    fst->set_initial(0);
    fst->set_final(1);
    
    int index = 0;
    for (SequenceList::const_iterator i = sequences.begin(); i != sequences.end(); ++i, ++index) {
        const Sequence& sequence = *i;
        int p = fst->add_state();
        for (Sequence::const_iterator j = sequence.begin(); j != sequence.end(); ++j) {
            int word = *j;
            int q = fst->add_state();
            char buf[12];
            sprintf(buf, "$%d", index);
            int index_label = FSTstring_map[buf];
            fst->add_arc(0, p, 0, 0, float(index));
            fst->add_arc(p, q, word, 0);
            fst->add_arc(q, 1, index_label, 0);
            p = q;
        }
    }

    fst = FSTOptimizeA(fst);

    // Generate final weight sets.

    assert("not yet implemented" == 0);


#if 0
    // Eliminate zero-length sequence.
    //
    fst->clear_final(0);

    // Eliminate stop words.
    //
    for (FSTArcIter a = fst->arcs_out(0); a.valid(); ++a) {
        if (stop_word_set.count(a->input)) {
            fst->clear_final(a->next);
        }
    }
#endif

    return fst;
}

static FSTGeneric
build(const SequenceList& sequences, const Set& stop_word_set, bool tag)
{
    FSTGeneric fst;
    if (tag) {
        fst = build_tagged(sequences, stop_word_set);
    } else {
        fst = build_untagged(sequences, stop_word_set);
    }
    return fst;
}

static int
fst_factors(int argc, char** argv)
{
    if (argc < 0) {
        usage();
        return 1;
    }

    const char* stop_word_file = NULL;
    bool binary_flag = false;
    bool tag_flag = false;
    bool error = false;

    char c;
    while ((c = getopt(argc, argv, "Bs:t")) != EOF) {
        switch (c) {
        case 'B':
            binary_flag = true;
            break;
        case 's':
            stop_word_file = optarg;
            break;
        case 't':
            tag_flag = true;
            break;
        default:
            error = true;
            break;
        }
    }
    argc -= optind;
    argv += optind;
    if (error || argc != 2) {
        usage();
        return 1;
    }

    const char* in_file  = argv[0];
    const char* out_file = argv[1];

    Set stop_word_set;
    if (stop_word_file) {
        build_stop_word_set(stop_word_set, stop_word_file);
    }

    SequenceList sequences;
    build_sequences(sequences, in_file, stop_word_set);

    FSTGeneric fst = build(sequences, stop_word_set, tag_flag);

    if (binary_flag) {
        fst->dump_binary(out_file);
    } else {
        fst->dump_ascii(out_file);
    }

    return 0;
}

// Register as subcommand.
//
static FstTool factors_tool("factors", fst_factors);

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 80
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
