////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

/******************************************************************************
=head1 NAME

fst_from_baseforms - convert text baseform file to FST

=head1 SYNOPSIS

B<fst_from_baseforms> [I<OPTIONS>] I<baseforms-in> I<fst-out>

=head1 DESCRIPTION

B<fst_from_baseforms> converts a textual baseform file to an FST.  The output
side will contain the word labels and the input side the pronunciation
symbols.  The word labels will be on word-initial arcs.

Sample baseform entries:

    boston       : b ao s t ax n
    boston       : b ( ao | aa ) s t ax n
    silence1plus : - +
    silence0plus : - *
    silence01    : - ?

The basic syntax is word separated from its pronunciation by " : " (i.e.,
one or more whitespace characters, then colon, then one or more whitespace
characters).  Within the pronunciation, you can use alternation "|",
parentheses "(" and ")" for grouping, Kleene "+" closure for one or more,
Kleene "*" closure for zero or more, or "?" for optional.

Note that pronunciation symbols are "sticky".  By this we mean that if any of
"|()+*?" are not separated by whitespace from a symbol then they are part of
the symbol.  For example, "ax+" is not one or more "ax", but rather a symbol
"ax+".

The specification of the output FST I<fst-out> can be a filename or "-"
indicating standard output.

=head1 OPTIONS

=over 12

=item B<-B>

Output in binary format.

=item B<-b> I<label>

Support I<label> for n-gram back-off.

=item B<-c>

Perform *-closure on lexicon.

=item B<-d>

Special dictionary treatment for use with PocketSUMMIT dynamic vocabulary.

=item B<-i> I<label>

Add I<label> self loop to the end of words.  A typical use might be to add an
optional pause at the end of every word.  This is equivalent to putting
"I<label> *" at the end of every baseform entry.

=item B<-o>

Perform FSA optimization, equivalent to that of fst_optimize(1) with the B<-a>
flag.  Output (word) labels will not be pushed back.

=item B<-P> I<label>

Put optional pauses I<label> at the start and end of word strings.

=item B<-p>

Put pound "#" symbol on closure arcs between words.  This can be very useful
when determinizing in the face of output ambiguity (e.g., homophones).

=item B<-w> I<wtw>

Put word transition weight I<wtw> at beginning of every word.

=item B<-W>

Put word-specific pseudo epsilon B<#>I<word> at end of every word.

=back

=head1 CAVEATS

B<fst_from_baseforms> uses a rather slow regular expression parser and FST
generator.  It really needs to be rewritten, perhaps along the lines of the one
used by B<fst_from_jsgf>.

=head1 AUTHOR

I. Lee Hetherington <ilh@mit.edu>

=cut
******************************************************************************/

#include "fst.h"
#include "fst/FSTRules.h"
#include <iostream>
#include <fstream>
#include <string>

static void
usage()
{
    cerr <<
        "usage: fst from_baseforms [options] in.baseforms out.fst\n"
        "\n"
        "-B         output binary file\n"
        "-b label   support label for backoff\n"
        "-c         perform *-closure on lexicon\n"
        "-d         special dictionary format for PocketSUMMIT\n"
        "-i label   add label self-loop at end of words\n"
        "-o         to optimize (default no optimizing)\n"
//      "-P         put word-specific # on closure arcs\n"
        "-P label   put optional pauses at start and end of word strings\n"
        "-p         put pound # on closure arcs\n"
        "-w wtw     put word transition weight at beginning of every word\n"
        "-W         put word-specific epsilon #<word> at end of every word\n"
        "\n";
}

// Parse a line to build an FST representing the baseform and merge it into the
// growing fst.  It will start from fst state 0 and end in fst state 1.
//
static bool
build_baseform(FSTGeneric& fst, string line, int iwt, float wtw, bool word_pound)
{
    int s_state = 0;                            // start state
    int e_state = 1;                            // end state

    //
    // A line looks as follows:  "word : regular-expression"
    //

    string::size_type index, ws_index;

    ws_index = line.find_first_of(" \t");
    if (ws_index == string::npos) {
        cerr << "no whitespace found" << endl;
        return false;
    }
    index = line.find(':', ws_index);
    if (index == string::npos) {
        cerr << "no : found" << endl;
        return false;
    }

    // Get rid of DOS carriage returns.
    //
    for (string::size_type i = line.find('\r'); i != string::npos; i = line.find('\r', i)) {
        line.erase(i);
    }

    string re = line.substr(index + 1);
    string word = line.substr(0, ws_index);

    // Change ` , ' to ` | '.
    //
    while ((index = re.find(" , ")) != string::npos)
        re[index + 1] = '|';

    // If there are any remaining commas, change them to ` | '.
    //
    while ((index = re.find(',')) != string::npos)
        re.replace(index, 1, " | ");

    // Change `&' to iwt * if iwt specified.
    //
    if (iwt) {
        string iwt_star = FSTstring_map[iwt];
        iwt_star += " *";
        while ((index = re.find('&')) != string::npos) {
            re.replace(index, 1, iwt_star);
        }
    } else {
        while ((index = re.find('&')) != string::npos)
            re.erase(index, 1);
    }

    // Append ` ;'.
    //
    re += " ;";

    // Call parser in single expression mode.
    //
    FSTGeneric baseform;
    if (!FSTRulesParseString(baseform, re, false, false))
        return false;

    // Merge word's FST into fst.
    //
    int epsilon = FSTstring_map["<epsilon>"];
    int offset = fst->add_states(baseform->n_states());
    fst->add_arc(s_state, baseform->initial() + offset, 0, 0, wtw);
    for (int p = 0; p < baseform->n_states(); p++) {
        int output = (p == baseform->initial()) ? FSTstring_map[word] : 0;
        if (baseform->is_final(p)) {
            //fst->set_final(p + offset);
            fst->add_arc(p + offset, e_state,
                         word_pound ? FSTstring_map["#" + word] : 0,
                         output);
        }
        for (FSTArcIter a = baseform->arcs_out(p); a.valid(); a++) {
            int input = a->input == epsilon ? 0 : a->input;
            fst->add_arc(p + offset, a->next + offset, input, output);
        }
    }

    return true;
}

static void
build(FSTGeneric& fst, istream& is, int iwt, float wtw, bool word_pound)
{
    int line_no = 1;
    string line;
    while (!getline(is, line).fail()) {
        if (!build_baseform(fst, line, iwt, wtw, word_pound))
            FSTWarn("problem parsing line %d\n", line_no);
        line_no++;
    }
}

static FSTGeneric
dictionary_format(const FSTGeneric& in)
{
    const int N = in->n_states();
    const int pound = FSTstring_map["#"];
    const int iwt = FSTstring_map["_"];
    const float zero = in->semiring()->zero;

    FSTGeneric out = FSTProject(in, FSTOutput, FSTInput);
    int cycleState = out->initial();
    int iwtState = out->add_state();

    out->set_final(iwtState);
    out->add_arc(iwtState, iwtState, 0, iwt);
    out->add_arc(iwtState, cycleState, 0, pound);

    for (int p = 0; p < N; ++p) {
        float w = in->final_weight(p);
        if (w != zero) {
            out->add_arc(p, iwtState, 0, iwt, w);
            out->add_arc(p, cycleState, 0, pound, w);
        }
    }

    return out;
}

static int
fst_from_baseforms(int argc, char** argv)
{
    if (argc < 0) {
        usage();
        return 1;
    }

    bool binary = false;
    int backoff = 0;
    bool closure = false;
    int iwt = 0;
    bool optimize = false;
    bool word_pound = false;
    bool pound = false;
    bool add_pauses = false;
    int pause[2];
    float wtw = 0;
    bool error = false;
    bool dictionary = false;
    char c;

    pause[0] = pause[1] = 0;

    while ((c = getopt(argc, argv, "Bb:cdi:oP:pw:W")) != EOF) {
        switch (c) {
        case 'B':
            binary = true;
            break;
        case 'b':
            backoff = FSTstring_map[optarg];
            break;
        case 'c':
            closure = true;
            break;
        case 'd':
            dictionary = true;
            break;
        case 'i':
            iwt = FSTstring_map[optarg];
            break;
        case 'o':
            optimize = true;
            break;
        case 'P':
            if (pause[0]) {
                pause[1] = FSTstring_map[optarg];
            } else {
                pause[0] = FSTstring_map[optarg];
            }
            add_pauses = true;
           break;
        case 'p':
            pound = true;
            break;
        case 'w':
            wtw = atof(optarg);
            break;
        case 'W':
            word_pound = true;
            break;
        case '?':
        default:
            error = true;
            break;
        }
    }
    argc -= optind;
    argv += optind;
    if (error || argc != 2) {
        usage();
        return 1;
    }

    if (dictionary) {
        backoff = 0;
        closure = false;
        iwt = 0;
        pause[0] = pause[1] = 0;
        add_pauses = false;
        pound = false;
        word_pound = false;
        optimize = true;
        binary = true;
    }

    if (!pause[1])
        pause[1] = pause[0];

    const string baseforms_file = *argv++;
    const string fst_file = *argv++;

    // Initialize lexicon FST.  State 0 will be the initial state and 1 the
    // final state.  Optionally, We connect 0 and 1 with a "#" label.
    //
    FSTGeneric fst = FSTBasic();
    fst->add_states(2);
    fst->set_initial(0);
    fst->set_final(1);
    if (closure) {
        fst->add_arc(1, 0, (pound && !word_pound) ? FSTstring_map["#"] : 0);
        if (!add_pauses) {
            int p = fst->add_state();
            fst->set_final(p);
            fst->add_arc(p, fst->initial());
            fst->set_initial(p);
        }
    }
    if (iwt) {
        fst->add_arc(1, 1, iwt, 0);
    }
    if (backoff) {
        fst->add_arc(0, 0, backoff, backoff);
    }
        
    // Process either standard input or file.
    //
    if (baseforms_file.empty() || baseforms_file == "-") {
        build(fst, cin, iwt, wtw, word_pound);
    } else {
        ifstream is(baseforms_file.c_str());
        if (is.bad()) {
            cerr << "error reading file `" << baseforms_file << "'" << endl;
            return 1;
        }
        build(fst, is, iwt, wtw, word_pound);
    }

    if ( add_pauses ) {
        // Build beginning pause model
        FSTGeneric pause1 = FSTBasic(); 
        pause1->add_states(3);
        pause1->set_initial(0);
        pause1->set_final(0);
        pause1->add_arc(0, 1, pause[0], 0);
        pause1->add_arc(1, 1, pause[0], 0);
        pause1->add_arc(1, 2, pound ? FSTstring_map["#"] : 0, 0);
        pause1->set_final(2);

        // Build ending pause model
        FSTGeneric pause2 = FSTBasic(); 
        pause2->add_states(3);
        pause2->set_initial(0);
        pause2->set_final(0);
        pause2->add_arc(0, 1, pound ? FSTstring_map["#"] : 0, 0);
        pause2->add_arc(1, 2, pause[1], 0);
        pause2->add_arc(2, 2, pause[1], 0);
        pause2->set_final(2);

        // Build pause model for utts with no output words
        FSTGeneric silence = FSTBasic(); 
        silence->add_states(1);
        silence->set_initial(0);
        silence->set_final(0);
        silence->add_arc(0, 0, pause[0], 0);

        // Concatenate pauses with lexicon
        fst = FSTBasic(FSTConcatenate(pause1, fst, pause2));
        fst = FSTBasic(FSTUnion(fst, silence));
    }

    // Optimize.
    //
    if (optimize)
        fst = FSTOptimizeA(fst);

    // Optional dictionary mode.
    //
    if (dictionary)
        fst = dictionary_format(fst);

    // Dump.
    //
    if (binary)
        fst->dump_binary(fst_file);
    else
        fst->dump_ascii(fst_file);

    return 0;
}

// Register as subcommand.
//
static FstTool from_baseforms_tool("from_baseforms", fst_from_baseforms);

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 79
// indent-tabs-mode: nil
// c-basic-offset: 4
// End:
