////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

/******************************************************************************
=head1 NAME

fst_from_jsgf - compile JSGF into recursive-transition network FST

=head1 SYNOPSIS

B<fst_from_jsgf> [I<OPTIONS>] I<jsgf-in> I<fst-out>

=head1 DESCRIPTION

B<fst_from_jsgf> compiles a context-free grammar (CFG) specified in the Java
Speech Grammar Format (JSGF) into a recursive-transition network (RTN) FST.

Left-recursive grammars are not currently supported, even if the left recursion
itself could be implemented by a finite-state cycle.  Other recursion is
supported.

The importing of subgrammars through JSGF's C<import> statement is not
supported.

Program code in the C<{...}> tags is not supported.  Instead, tags can be used
to specify additional output symbols when particular rules or subrules fire.

=head1 OPTIONS

=over 12

=item B<-p>

Add non-terminal parse tags to output.  For example, if the grammar has a
non-terminal C<<< <foo> >>>, then a realization of it will be bracketed by C<<<
<foo> >>> and C<<< </foo> >>>.  Note that this may have performance implications
when composing the resulting FST with input.

=item B<-t>

Ignore C<{...}> tags in JSGF file.

=item B<-w>

Do not output accepted input sequence (e.g., words), but instead only output
C<{...}> tags.

=item B<-s> I<semiring>

Specify the I<semiring> for the resulting FST.  Supported semirings are
B<minplus> for (min,+) and B<plustimes> for (+,*).  The default is B<minplus>
(min,+).

=back

=head1 SEE ALSO

FSTRtn(5)

http://java.sun.com/products/java-media/speech/forDevelopers/JSGF/

=head1 AUTHOR

I. Lee Hetherington <ilh@mit.edu>

=cut
******************************************************************************/

#include <iostream>
#include "fst.h"

using namespace std;

static void
usage()
{
    cerr <<
        "usage: fst from_jsgf [options] in out.fst\n"
        "\n"
        "-P     surround with <pause1> and <pause2>\n"
        "-p     add non-terminal parse tags (e.g., ... <foo> ... </foo> ...)\n"
        "-t     ignore {} tags in JSGF file\n"
        "-w     ignore words (don't output); tags will be only output\n"
        "\n";
}

static int
fst_from_jsgf(int argc, char** argv)
{
    if (argc < 0) {
        usage();
        return 1;
    }

    bool pause12 = false;
    bool rtn_parse = false;
    bool ignore_tags = false;
    bool ignore_words = false;
    bool error = false;

    char c;
    while ((c = getopt(argc, argv, "Pptw")) != EOF) {
        switch (c) {
        case 'P':
            pause12 = true;
            break;
        case 'p':
            rtn_parse = true;
            break;
        case 't':
            ignore_tags = true;
            break;
        case 'w':
            ignore_words = true;
            break;
        default:
            error = true;
        }
    }
    argc -= optind;
    argv += optind;

    if (error || argc != 2) {
    usage();
        return 1;
    }

    const char* in_file  = argv[0];
    const char* out_file = argv[1];

    FSTGeneric result;
    if (!FSTJsgfParseFile(result, in_file, ignore_tags, ignore_words, rtn_parse, pause12))
        return 1;

    result->dump_ascii(out_file);
    return 0;
}

// Register as subcommand.
//
static FstTool from_jsgf_tool("from_jsgf", fst_from_jsgf);

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 80
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
