////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

/******************************************************************************
=head1 NAME

fst_from_list - build FST from list

=head1 SYNOPSIS

B<fst_list> [I<OPTIONS>] I<fst-in>

=head1 DESCRIPTION

B<fst_info> constructs an FST from a list of phrases in a text file.  Each
line will be an optional path, and each path can have one or more words.

=head1 OPTIONS

=over 12

=item B<-B>

Output in binary format.

=back

=head1 SEE ALSO

fst_from_string(1)

=head1 AUTHOR

I. Lee Hetherington <ilh@mit.edu>

=cut
******************************************************************************/

#include <iostream>
#include <fstream>
#include <cmath>
#include "fst.h"

using namespace std;

struct Node;

typedef fst::map<int, Node*> Map;

struct Node {
    bool isFinal;
    int count;
    Map children;

    Node() : isFinal(false), count(0) {
    }
    ~Node() {
        for (Map::iterator i = children.begin(); i != children.end(); ++i) {
            delete (*i).second;
        }
    }
};

typedef fst::set<int> Set;
typedef fst::vector<int> Sequence;
typedef fst::list<Sequence> SequenceList;

static void
usage()
{
    cerr <<
        "usage: [options] in.txt out.fst\n"
        "\n"
        "-B               binary output\n"
        "\n";
}

static void
build_sequence(Node* root, const string& text)
{
    const char* ws = " \t\r\n";
    string::size_type i, j;

    root->count++;
    Node* node = root;

    for (j = 0; j != string::npos && (i = text.find_first_not_of(ws, j)) != string::npos;) {
        j = text.find_first_of(ws, i);
        string word;
        if (j == string::npos) {
            word = text.substr(i);
        } else {
            word = text.substr(i, j - i);
        }
        int label = FSTstring_map[word];
        Node*& nextNode = node->children[label];
        if (!nextNode) {
            nextNode = new Node();
        }
        nextNode->count++;
        node = nextNode;
    }

    if (node != root) {
        node->isFinal = true;
    }
}

static void
build_sequences(Node* root, const char* filename)
{
    ifstream is(filename);
    string line;
    while (!getline(is, line).fail()) {
        build_sequence(root, line);
    }
}

static void
build(FSTGeneric fst, int p, Node* n)
{
    if (n->isFinal) {
        fst->set_final(p);
    }
    for (Map::const_iterator i = n->children.begin(); i != n->children.end(); ++i) {
        int label = (*i).first;
        Node* c = (*i).second;
        int q = fst->add_state();
        // TODO: turn count into weight
        fst->add_arc(p, q, label, label);
        build(fst, q, c);
    }
}

static FSTGeneric
build(Node* root)
{
    FSTGeneric fst = FSTBasic();
    fst->add_state();
    fst->set_initial(0);
    cerr << "building..." << flush;
    build(fst, 0, root);
    cerr << "done\n";
    cerr << "minimizing..." << flush;
    fst = FSTMinimize(fst);
    cerr << "done\n";
#if 0
    fst->set_semiring(FSTMinPlus);
#endif
    return fst;
}

static int
fst_from_list(int argc, char** argv)
{
    if (argc < 0) {
        usage();
        return 1;
    }

    bool binary_flag = false;
    bool error = false;

    char c;
    while ((c = getopt(argc, argv, "B")) != EOF) {
        switch (c) {
        case 'B':
            binary_flag = true;
            break;
        default:
            error = true;
            break;
        }
    }
    argc -= optind;
    argv += optind;
    if (error || argc != 2) {
        usage();
        return 1;
    }

    const char* in_file  = argv[0];
    const char* out_file = argv[1];

    Node* root = new Node();
    build_sequences(root, in_file);
    FSTGeneric fst = build(root);
    delete root;
    root = NULL;

    if (binary_flag) {
        fst->dump_binary(out_file);
    } else {
        fst->dump_ascii(out_file);
    }

    return 0;
}

// Register as subcommand.
//
static FstTool from_list_tool("from_list", fst_from_list);

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 80
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
