////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2010, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

// TO DO
//
// *  Trim before dumping FSTNgram2.
//
// *  Implement remove unknown for ARPA files.
//
// *  Limit load depth of ARPA files.
//
// *  Reject count pruning and reverse for ARPA files.
//
// *  Allow different count thresholds at different levels of the tree.  I'm not
//    sure yet how this will interact with incremental n-grams.

// For compatibility with the probabilities computed by the old ngram_create
// tool.  However, I believe these are areas where the old tool was incorrect or
// suboptimal.
//
#undef COMPATIBILITY

// To enable EM training of lambdas for within-class unigram probabilities.
// COMPATIBILITY mode disables this feature.
//
#define WITHIN_CLASS_EM

// This was originally 1000, but empirically 5 gives slightly better perplexity
// and runs a lot faster.  Too many iterations leads to overtraining.
//
#define MAX_EM_ITERATIONS 5

#define EM_LAMBDA_TOLERANCE 1e-6

// To print out debugging information.
//
#undef DEBUG

// To verify that probabilities sum to 1.
//
#undef VERIFY_SUM
#define VERIFY_SUM_TOLERANCE 1e-4

#include <cstdlib>
#include <iostream>
#include <sstream>
#include <fstream>
#include <unistd.h>
#include <cmath>
#include "fst.h"
#include "fst/format.h"

using namespace std;

// Convenient macros:
//
#define BRANCH_NODE(NODE) \
    dynamic_cast<BranchNode*>(NODE)
#define FOR_EACH_CHILD(I, NODE) \
    for (BranchNode::iterator I = (NODE)->children.begin(); \
         I != (NODE)->children.end(); ++I)
#define FOR_EACH_UNIT(I, UNITS) \
    for (fst::set<int>::const_iterator I = (UNITS).begin(); \
         I != (UNITS).end(); ++I)

class NgramTree {

public: // methods

    NgramTree(const char* filename, bool reverse,
              int depth, int invert_depth,
              int threshold, bool exclude_unknown, bool det_flag, bool arpa_flag);
    ~NgramTree();

    void dump_FSTNgram2(const char* filename, const fst::set<int>& open_classes);
    void dump_FSTNgram2(ostream& os, const fst::set<int>& open_classes);

    void dump_FSTNgram3(const char* filename, const fst::set<int>& open_classes);
    void dump_FSTNgram3(ostream& os, const fst::set<int>& open_classes);

protected: // types

    // Leaf node and base class for branch node.
    //
    struct Node {
        int unit;
        int count;
        float prob;

        Node() {}
        Node(int c) : count(c) {}
        Node(const Node* n) : unit(n->unit), count(n->count) {}
        Node(float logprob) : count(-1), prob(logprob) {}
        virtual ~Node() {}
        virtual void set_count(int n) {
            count = n;
        }
        FSTAllocator_new_delete(Node);
    };

    // Comparison object for sorting and searching nodes by unit name.
    //
    struct NodeLess {
        bool operator()(const Node* a, const Node* b) const {
            return a->unit < b->unit;
        }
    };

    // Banch (non-terminal) node.
    //
    struct BranchNode : Node {
        int adj_count;                          // adjusted for excluded unknown
        float lambda;
        int state;
        BranchNode* backoff_node;
        fst::vector<Node*> children;

        typedef fst::vector<Node*>::iterator iterator;

        BranchNode(int c)
            : Node(c), adj_count(c), state(-1) {}
        BranchNode(const Node* n)
            : Node(n), adj_count(n->count), state(-1) {}
        BranchNode(float logprob, float bo_logprob)
            : Node(logprob), lambda(bo_logprob) {}
        virtual ~BranchNode() {
            FOR_EACH_CHILD(i, this)
                delete *i;
        }
        virtual void set_count(int n) {
            count = adj_count = n;
        }
        Node* find(int unit) {
            iterator i;
            return find(unit, i);
        }
        Node* find(int unit, iterator& i) {
            Node n; n.unit = unit;
            i = lower_bound(children.begin(), children.end(), &n, NodeLess());
            return (i != children.end() && (*i)->unit == unit) ? *i : 0;
        }
        void insert(int unit, Node* node, iterator i) {
            node->unit = unit;
            children.insert(i, node);
        }
        void insert(int unit, Node* node) {
            iterator i;
            Node* old_node = find(unit, i);
            assert(old_node == 0);
            insert(unit, node, i);
        }
        void append(int unit, Node* node) {
            node->unit = unit;
            children.push_back(node);
        }
        void erase(iterator i) {
            children.erase(i);
        }
        void sort() {
            ::sort(children.begin(), children.end(), NodeLess());
        }
        void reserve(int n) {
            children.reserve(n);
        }
        int n_children() {
            return children.size();
        }
        FSTAllocator_new_delete(BranchNode);
    };

protected: // members

    int         depth;
    int         invert_depth;
    fst::vector<int> thresholds;
    fst::vector<int> counts;
    BranchNode* root;
    BranchNode* class_root;
    fst::set<int>    units;
    bool        reverse;
    bool        exclude_unknown;
    bool        det_flag;
    bool        needs_trim;
    bool        arpa_flag;
    int         n_states;
    int         n_arcs;
    fst::vector<BranchNode*> states;
    int         unknown;                        // <unknown> index
    int         initial_final;                  // <> index
    int         cmu_initial;                    // <s> index
    int         cmu_final;                      // </s> index
    int         cmu_unknown1;                   // <unk> index
    int         cmu_unknown2;                   // <UNK> index
    
protected: // methods

    void        load_tree_file(istream& is, int threshold_);

    Node*       load(istream& is, const fst::map<int,int>& label_map,
                     int load_depth, int prune_depth, int d);
    void        load_reverse(istream& is, const fst::map<int,int>& label_map,
                             int load_depth, int prune_depth, int d,
                             fst::list<int>& path);
    void        build_reverse(const fst::list<int>& path, int count);

    void        arpa_load(istream& is, string header);
    void        add_logprob_path(const fst::vector<int>& path, Node* node);
    void        prune_worse_than_backoff();
    void        recompute_lambda(BranchNode* node);

    void        em_compute_logprobs(BranchNode* start);

    void        compute_within_class_logprobs();

    void        add_missing_unigrams();

    void        set_backoff();
    void        set_backoff(BranchNode* node, fst::vector<int>& path);

    void        remove_unknown();
    void        remove_unknown(BranchNode* node, int d);

    static float unit_prob(BranchNode* node, int unit, float uniform);
    static float unit_logprob(BranchNode* node, int unit);

    BranchNode* find_node(fst::vector<int>::const_iterator i,
                          fst::vector<int>::const_iterator end);

    void        assign_states(BranchNode* node=0);

    void        dump_FSTNgram2(ostream& os,
			       BranchNode* node, BranchNode* parent,
			       int final_unit, fst::vector<int>& context, int& p);

    void        dump_FSTNgram3(ostream& os,
			       BranchNode* node, BranchNode* parent,
			       int final_unit, fst::vector<int>& context, int& p);

#ifdef DEBUG
    void        print_lambdas();
    void        print_lambdas(Node*, fst::vector<int>&);
    void        print_logprobs();
    void        print_logprobs(Node*, fst::vector<int>&);
#endif
};

NgramTree::NgramTree(const char* filename, bool reverse_,
                     int depth_, int invert_depth_,
                     int threshold_, bool exclude_unknown_, bool det_flag_, bool arpa_flag_)
    : depth(depth_),
      invert_depth(invert_depth_),
      reverse(reverse_),
      exclude_unknown(exclude_unknown_),
      det_flag(det_flag_),
      needs_trim(false),
      arpa_flag(arpa_flag_)
{
    // Commonly used string indices.
    //
    initial_final = FSTstring_map["<>"];
    unknown       = FSTstring_map["<unknown>"];
    cmu_initial   = FSTstring_map["<s>"];
    cmu_final     = FSTstring_map["</s>"];
    cmu_unknown1  = FSTstring_map["<unk>"];
    cmu_unknown2  = FSTstring_map["<UNK>"];

    // this can be done better without the duplication of code here
    string name(filename);
    if (name.rfind(".gz") == name.size() - 3) {
      // has .gz suffix, use gzstream 
      
      ifstream is(filename);

      if (!is) {
        cerr << "cannot read file " << filename << endl;
        exit(1);
      }
      
      string header;
      getline(is, header);

      if (header == "NgramCountTree") {
        load_tree_file(is, threshold_);
      } else {
        arpa_load(is, header);
      }

    } else {
      // plain old file

      ifstream is(filename);

      if (!is) {
        cerr << "cannot read file " << filename << endl;
        exit(1);
      }
      
      string header;
      getline(is, header);

      if (header == "NgramCountTree") {
        load_tree_file(is, threshold_);
      } else {
        arpa_load(is, header);
      }
    }

}

NgramTree::~NgramTree()
{
    delete root;
    delete class_root;
}

void
NgramTree::load_tree_file(istream& is, int threshold_)
{
    int version = FSTLoadInt(is);
    assert(version == 1);

    int load_depth = FSTLoadInt(is);
    if (depth <= 0)
        depth = load_depth;
    if (depth > load_depth) {
        cerr << "tree depth " << load_depth
             << " cannot be increased to " << depth << endl;
        exit(1);
    }

    fst::vector<int> load_thresholds(load_depth + 1, 0);
    for (int d = 1; d <= load_depth; d++)
        load_thresholds[d] = FSTLoadInt(is);

    thresholds = load_thresholds;
    if (threshold_ > 0) {
        for (int d = 2; d <= depth; d++)
            thresholds[d] = threshold_;
    }

    //
    // label map
    //
    fst::map<int,int> label_map;
    int label_map_size = FSTLoadInt(is);
    string label;
    while (label_map_size-- > 0) {
        FSTLoadString(is, label);
        int index = FSTLoadInt(is);
        label_map[index] = FSTstring_map[label];
    }

    //
    // n-gram units
    //
    units.clear();
    int n_units = FSTLoadInt(is);
    while (n_units-- > 0)
        units.insert(label_map[FSTLoadInt(is)]);

    //
    // word class tree
    //

    counts.clear();
    counts.insert(counts.begin(), 3, 0);

    class_root = BRANCH_NODE(load(is, label_map, 2, 0, 0));

    //
    // n-gram tree
    //

    counts.clear();
    counts.insert(counts.begin(), depth + 1, 0);

    if (reverse) {
        cerr << "loading and reversing tree...";
        root = new BranchNode(-1);
        fst::list<int> scratch;
        load_reverse(is, label_map, load_depth, depth, 0, scratch);
        cerr << "done\n";
    } else {
        cerr << "loading tree...";
        root = BRANCH_NODE(load(is, label_map, load_depth, depth, 0));
        cerr << "done\n";
    }

    // Optionally remove all occurrences of <unknown> from tree.
    //
    if (exclude_unknown)
        remove_unknown();

    // Summarize number of nodes at various levels.
    //
    for (unsigned d = 1; d < counts.size(); d++)
        format(cerr, "%2d %8d\n", d, counts[d]);

    // Prepare tree by adding missing unigrams (with zero counts) and setting
    // the backoff pointer for every non-leaf node.
    //
    add_missing_unigrams();
    set_backoff();

    // Compute logprobs using EM.  I'll likely add alternatives here.
    //
    em_compute_logprobs(root);
    compute_within_class_logprobs();
}

// Load depth-first tree of depth load_depth from stream.  If prune_depth > 0,
// prune it to the specified depth using threshold-based pruning as well.
//
NgramTree::Node*
NgramTree::load(istream& is, const fst::map<int,int>& label_map,
                int load_depth, int prune_depth, int d)
{
    int count = FSTLoadCompressedUnsigned(is);

    if (prune_depth > 0
        && (d > prune_depth || count < thresholds[d])) {

        if (d < load_depth) {
            int n = FSTLoadCompressedUnsigned(is);
            while (n-- > 0) {
                FSTLoadCompressedUnsigned(is);
                load(is, label_map, load_depth, prune_depth, d + 1);
            }
        }
        return NULL;

    } else {

        if (d < load_depth) {
            BranchNode* branch = new BranchNode(count);
            int n = FSTLoadCompressedUnsigned(is);
            branch->reserve(n);
            while (n-- > 0) {
                int index = FSTLoadCompressedUnsigned(is);
                fst::map<int,int>::const_iterator i = label_map.find(index);
                assert(i != label_map.end());
                int label = (*i).second;
                Node* child = load(is, label_map,
                                   load_depth, prune_depth, d + 1);
                if (child)
                    branch->append(label, child);
            }
            branch->sort();
            counts[d + 1] += branch->n_children();
            if (branch->n_children() > 0 || d == 0) {
                return branch;
            } else {
                delete branch;
                return new Node(count);
            }
        } else {
            return new Node(count);
        }

    }
}

// Load depth-first tree of depth load_depth from stream, building a reversed
// tree.  If prune_depth > 0, prune it to the specified depth using
// threshold-based pruning as well.
//
void
NgramTree::load_reverse(istream& is, const fst::map<int,int>& label_map,
                                int load_depth, int prune_depth, int d,
                                fst::list<int>& path)
{
    int count = FSTLoadCompressedUnsigned(is);

    if (prune_depth > 0
        && (d > prune_depth || count < thresholds[d])) {

        if (d < load_depth) {
            int n = FSTLoadCompressedUnsigned(is);
            while (n-- > 0) {
                FSTLoadCompressedUnsigned(is);
                load_reverse(is, label_map, load_depth,
                             prune_depth, d + 1, path);
            }
        }

    } else {

        if (d < load_depth) {
            int n = FSTLoadCompressedUnsigned(is);
            while (n-- > 0) {
                fst::map<int,int>::const_iterator i =
                    label_map.find(FSTLoadCompressedUnsigned(is));
                assert(i != label_map.end());
                int label = (*i).second;
                path.push_front(label);
                load_reverse(is, label_map, load_depth,
                             prune_depth, d + 1, path);
                path.pop_front();
            }
        }
        build_reverse(path, count);
        counts[d]++;

    }
}

// Helper function for building reverse trees.  Add path and its count to tree,
// creating any necessary intermediate nodes.  Only nodes with children will be
// of type BranchNode.
//
void
NgramTree::build_reverse(const fst::list<int>& path, int count)
{
    int n = path.size();
    if (n == 0) {
        root->set_count(count);
    } else {
        BranchNode* branch = root;
        for (fst::list<int>::const_iterator i = path.begin(); i != path.end(); ++i) {
            int unit = *i;
            bool last = (--n == 0);
            BranchNode::iterator j;
            Node* node = branch->find(unit, j);
            if (!node) {
                node = last ? new Node(-1) : new BranchNode(-1);
                branch->insert(unit, node, j);
            }
            if (last) {
                node->set_count(count);
            } else {
                branch = BRANCH_NODE(node);
                if (!branch) {
                    branch = new BranchNode(node);
                    delete node;
                    *j = node = branch;
                }
            }
        }
    }
}

// Load an ARPA file containing logprobs.
//
void
NgramTree::arpa_load(istream& is, string header)
{
    string line;

    root = new BranchNode(-1);
    class_root = 0;

    // First find beginning of data.
    //
    line = header;
    while (!is.fail() && line != "\\data\\")
        getline(is, line);
    if (line != "\\data\\") {
        cerr << "could not find \\data\\" << endl;
        exit(1);
    }

    // Do each n-gram order, starting with 1-grams.
    //
    int max_depth = 0;
    for (int o = 1; depth <= 0 || o <= depth; ++o) {

        int count = 0;
        
        // Skip to beginning of each o-gram.
        //
        ostringstream marker;
        format(marker, "\\%d-grams:", o);
        while (!is.fail() && line != marker.str() && line != "\\end\\")
            getline(is, line);
        if (is.fail() || line == "\\end\\")
            break;
        cerr << "loading " << o << "-grams...";

        max_depth = max(max_depth, o);

        // Parse lines of the form: logprob w1 .. wo backoff.  Backoff will be
        // missing from the highest-order entries.
        //
        fst::vector<int> path;
        string word;
        while (getline(is, line), !is.fail()) {
            if (line.size() == 0)
                continue;
            if (line[0] == '\\')
                break;
            path.clear();

            istringstream lis(line.c_str());
            float lp;
            lis >> lp;
            lp *= M_LN10;

            for (int i = 0; i < o; ++i) {
                lis >> word;
                int wi = FSTstring_map[word];
                if (wi == cmu_unknown1 || wi == cmu_unknown2)
                    wi = unknown;
                path.push_back(wi);
                if (o == 1)
                    units.insert(wi);
            }

            float backoff_lp;
            lis >> backoff_lp;

            Node* node;
            if ((depth <= 0 || o < depth) && !lis.fail()) {
                backoff_lp *= M_LN10;
                node = new BranchNode(lp, backoff_lp);
            } else {
                node = new Node(lp);
            }

            add_logprob_path(path, node);

            ++count;
        }
        cerr << count << "...done" << endl;
    }
    if (depth <= 0)
        depth = max_depth;

    // Look for optional within-class probabilities.
    //
    while (!is.fail() && line != "\\end\\")
        getline(is, line);
    if (!is.fail()) {
        while (getline(is, line)) {
            if (line.size() > 0) {
                string class_name, word;
                float p = 0;
                istringstream lis(line.c_str());
                lis >> class_name >> p >> word;
                if (p == 0) {
                    cerr << "error: failed to interpret line: " << line << endl;
                    exit(1);
                } else {
                    if (!class_root) {
                        class_root = new BranchNode(-1);
                    }
                    int class_unit = FSTstring_map[class_name];
                    Node* class_node = class_root->find(class_unit);
                    if (!class_node) {
                        class_node = new BranchNode(-1);
                        class_root->insert(class_unit, class_node);
                    }
                    BranchNode* class_branch = BRANCH_NODE(class_node);
                    Node* word_node = new Node(log(p));
                    word_node->unit = FSTstring_map[word];
                    class_branch->insert(word_node->unit, word_node);
                    cerr << "added " << word << " to class " << class_name << endl;
                }
            }
        }
    }

    set_backoff();

    if (!arpa_flag) {
        prune_worse_than_backoff();
    }
}

// Add path to tree.
//
void
NgramTree::add_logprob_path(const fst::vector<int>& path, Node* new_node)
{
    int n = path.size();
    BranchNode* branch = root;
    for (fst::vector<int>::const_iterator i = path.begin(); i != path.end(); ++i) {
        int unit = *i;
        if (--n == 0) {
            // Last item in path: add new node.
            branch->insert(unit, new_node);
        } else {
            // Not last item, but it should exist since we do lower-order
            // n-grams first when loading ARPA file.
            Node* node = branch->find(unit);
            assert(node != 0);
            branch = BRANCH_NODE(node);
            assert(branch != 0);
        }
    }
}

// Recompute lambda because the ARPA files tend to have low precision.
//
void
NgramTree::recompute_lambda(BranchNode* node)
{
    double child_sum = 0;
    double child_bo_sum = 0;
    FOR_EACH_CHILD(i, node) {
        Node* child = *i;
        int unit = child->unit;
        if (child->prob <= 0) {
            child_sum += exp(child->prob);
            if (node->backoff_node)
                child_bo_sum += exp(unit_logprob(node->backoff_node, unit));
        }
    }
    node->lambda = log((1 - child_sum) / (1 - child_bo_sum));
}

// Prune away child logprobs that are worse than backoff.  For each node
// (history), we prune in decreasing order of (backoff - non-backoff) and stop
// when there are no more.  This monotonically increases the backoff weight
// lambda.  We process the nodes in a breadth-first order to assure that
// lower-order n-grams are finished before doing higher-order n-grams.
//
void
NgramTree::prune_worse_than_backoff()
{
    cerr << "pruning n-grams worse than backoff...";

    const float bad_logprob = 1e30;

    // For constructing priority queue of child logprobs to potentially prune.
    //
    typedef pair<float, Node*> Entry;
    typedef priority_queue<Entry> EntryQueue;

    int n_pruned = 0;

    fst::deque<BranchNode*> fifo;
    fifo.push_back(root);
    while (!fifo.empty()) {

        BranchNode* node = fifo.front();
        fifo.pop_front();

        recompute_lambda(node);

        // Go through children, computing the sum of child probs, adding an
        // entry for each child whose logprob is worse than backoff, and adding
        // non-leaf children to FIFO to continue breadth-first exploration.
        //
        EntryQueue entry_queue;
        double child_sum = 0;
        FOR_EACH_CHILD(i, node) {
            Node* child = *i;
            if (node->backoff_node) {
                int unit = child->unit;
                child_sum += exp(child->prob);
                float bo = unit_logprob(node->backoff_node, unit);
                if (node->lambda + bo >= child->prob)
                    entry_queue.push(Entry(bo - child->prob, child));
            }
            BranchNode* child_branch = BRANCH_NODE(child);
            if (child_branch)
                fifo.push_back(child_branch);
        }

        // The entry priority queue will be non-empty if there were any
        // problematic children.
        //
        if (!entry_queue.empty()) {

            double log_lambda = node->lambda;
            double non_child_sum = (1 - child_sum) / exp(log_lambda);

            while (!entry_queue.empty()) {

                // Pop element with biggest (backoff - non-backoff) difference.
                //
                Entry e = entry_queue.top();
                entry_queue.pop();

                Node* child = e.second;

                // Difference is backoff logprob - non-backoff logprob, but
                // backoff weight lambda is not included.
                //
                float difference = e.first;

                // If lambda has decreased enough to make this child no longer
                // worse than its backoff then we are finished.
                //
                if (difference + log_lambda < 0)
                    break;

                // Efficiently update lambda.
                //
                child_sum -= exp(child->prob);
                non_child_sum += exp(difference + child->prob);
                log_lambda = log((1 - child_sum) / non_child_sum);

                // Mark as pruned.
                //
                child->prob = bad_logprob;

                if (BRANCH_NODE(child))
                    needs_trim = true;

                n_pruned++;
            }
            node->lambda = log_lambda;

            // Unprune non-terminal children because they may be needed to
            // access higher-order n-grams.  Give them the same score they would
            // get through backoff.  We do not need to adjust lambda because we
            // are not really changing any probabilities.
            //
            // NOTE: This is currently commented out because it produces an
            // n-gram that cannot be properly inverted when building an
            // incremental n-gram.  So, for the time being this is not used.
            // Besides, it seemed to make very little difference in perplexity,
            // and without it the FST is a bit smaller as well.
            //
#if 1
            FOR_EACH_CHILD(i, node) {
                Node* child = *i;
                if (child->prob == bad_logprob && BRANCH_NODE(child)) {
                    child->prob = unit_logprob(node, child->unit);
                }
            }
#endif

#ifdef VERIFY_SUM
            double sum = 0;
            FOR_EACH_UNIT(ui, units) {
                int unit = *ui;
                sum += exp(unit_logprob(node, unit));
            }
            if (fabs(sum - 1) > VERIFY_SUM_TOLERANCE) {
                fst::vector<int> labels;
                for (BranchNode* n = node;
                     n->backoff_node;
                     n = n->backoff_node) {
                    labels.push_back(n->unit);
                }
                format(cerr, "%10.6f    ", sum);
                for (fst::vector<int>::reverse_iterator j = labels.rbegin();
                     j != labels.rend(); ++j)
                    cerr << ' ' << FSTstring_map[*j];
                cerr << endl;
            }
#endif
        }

    }
    cerr << n_pruned << "...done" << endl;
}            


static inline float
discounted_prob(int count, int discount, float normalization)
{
    if (count > discount)
        return float(count - discount) * normalization;
    else
        return 0;
}

static inline float
weight(float lambda, float x, float y)
{
    return lambda*x + (1 - lambda)*y;
}

// Compute probabilities for tree rooted at start using EM to estimate per-node
// lambda weighting node probabilities vs. backoff probabilities.  For EM we use
// discounted statistics to simulate leave-one-out jackknifing.  There are two
// breadth-first passes here: (1) to compute the lambdas and (2) to compute
// probabilities using the lambdas.  Breadth-first guarantees that lower-order
// n-grams are computed before higher-order n-grams.
//
void
NgramTree::em_compute_logprobs(BranchNode* start)
{
    fst::deque<BranchNode*> fifo;

    int n_units = start->n_children();
#ifdef COMPATIBILITY
    if (start == root && !start->find(unknown))
        n_units++;
#endif

    const int min_smooth = 1;
    const int max_smooth = n_units;

    const float uniform = float(1) / float(n_units);

    // First breadth-first pass: compute lambdas using discounted probabilities.
    //
    if (start == root)
        cerr << "computing lambdas using EM...";
    fifo.push_back(start);
    while (!fifo.empty()) {
        BranchNode* node = fifo.front();
        fifo.pop_front();

        // Compute the depth of this node's predictions (1..n).
        //
        int d = 1;
        for (BranchNode* n = node->backoff_node; n != 0; n = n->backoff_node)
            d++;

        // Set the discount value to the threshold value at this depth.  This
        // has not been checked for theoretical correctness yet, but it seems to
        // work and make sense intuitively.
        //
        int discount = thresholds[d];
        assert(discount > 0);

        int total = node->adj_count;
        if (total == 0)
            continue;
        float norm = float(1) / float(total - discount);

        // Compute lambda for node.
        //
        float lambda = 0.5;
        for (int k = 0; k < MAX_EM_ITERATIONS; k++) {
            float sum = 0;
            FOR_EACH_CHILD(i, node) {
                Node* child = *i;
                int unit = child->unit;
                float p = discounted_prob(child->count, discount, norm);
                float p_bo = unit_prob(node->backoff_node, unit, uniform);
                sum += child->count * lambda * p / weight(lambda, p, p_bo);
            }
            float old_lambda = lambda;
            lambda = sum / total;

            // Apply upper and lower bounds on lambda.
            //
#ifdef COMPATIBILITY
            if (node != root)
#endif
            {
                lambda = min(lambda, float(total) / float(total + min_smooth));
                lambda = max(lambda, float(total) / float(total + max_smooth));
            }
            if (fabs(lambda - old_lambda) < EM_LAMBDA_TOLERANCE)
                break;
        }

        // Fill in final discounted probabilities for children, and push
        // non-leaf children into FIFO to continue breadth-first computation.
        //
        node->lambda = lambda;
        FOR_EACH_CHILD(i, node) {
            Node* child = *i;
            int unit = child->unit;
            float p = discounted_prob(child->count, discount, norm);
            float p_bo = unit_prob(node->backoff_node, unit, uniform);
            child->prob = weight(lambda, p, p_bo);
            BranchNode* child_branch = BRANCH_NODE(child);
            if (child_branch)
                fifo.push_back(child_branch);
        }
        
    }
    if (start == root)
        cerr << "done" << endl;

#ifdef DEBUG
    print_lambdas();
#endif

    // Second breadth-first pass: compute probabilities using lambdas.
    //
    if (start == root)
        cerr << "computing probabilities...";
    fifo.push_back(start);
    while (!fifo.empty()) {
        BranchNode* node = fifo.front();
        fifo.pop_front();

        // Compute child probs and push branch children into FIFO to continue
        // breadth-first computation.
        //
        int total = node->adj_count;
        FOR_EACH_CHILD(i, node) {
            Node* child = *i;
            int unit = child->unit;
            float p = float(child->count) / float(total);
            float p_bo = unit_prob(node->backoff_node, unit, uniform);
            child->prob = weight(node->lambda, p, p_bo);
            BranchNode* child_branch = BRANCH_NODE(child);
            if (child_branch)
                fifo.push_back(child_branch);
        }

#ifdef VERIFY_SUM
        float sum = 0;
        for (fst::set<int>::const_iterator ui = units.begin();
             ui != units.end(); ++ui) {
            int unit = *ui;
            sum += unit_prob(node, unit, 0);
        }
        if (fabs(sum - 1) > VERIFY_SUM_TOLERANCE)
            format(cerr, "does not sum to 1: %.6f\n", sum);
#endif
    }
    if (start == root)
        cerr << "done" << endl;

    // Third breadth-first pass: convert to log probabilities.  lambda will be
    // replaced with log(1 - lambda), the backoff weight.  Unlike the first two
    // passes, this does not have to be breadth-first, but it is convenient to
    // do it this way here.  Note that it is slower to fold this into the second
    // pass because of weighted sums of log probs and the additional exp/logs
    // needed.
    //
    if (start == root)
        cerr << "computing log probabilities...";
    fifo.push_back(start);
    while (!fifo.empty()) {
        BranchNode* node = fifo.front();
        fifo.pop_front();

        node->lambda = log(1 - node->lambda);
        FOR_EACH_CHILD(i, node) {
            Node* child = *i;
            child->prob = log(child->prob);
            BranchNode* child_branch = BRANCH_NODE(child);
            if (child_branch)
                fifo.push_back(child_branch);
        }
    }
    if (start == root)
        cerr << "done" << endl;

#ifdef DEBUG
    print_logprobs();
#endif
}

static inline bool
class_is_uniform(int index)
{
    const char* label = FSTstring_map[index];
    return
        label[0] == '<' &&
        label[1] == 'U' &&
        label[2] == '>' &&
        label[3] == '_';
}

void
NgramTree::compute_within_class_logprobs()
{
    FOR_EACH_CHILD(i, class_root) {
        BranchNode* class_node = BRANCH_NODE(*i);
        int class_label = class_node->unit;
        if (class_is_uniform(class_label) || class_node->count == 0) {
            float uniform = float(1) / float(class_node->n_children());
            FOR_EACH_CHILD(j, class_node) {
                Node* child = *j;
                child->prob = log(uniform);
            }
        } else {
#if defined(COMPATIBILITY) || !defined(WITHIN_CLASS_EM)
            const int floor = 1;
            int total = 0;
            FOR_EACH_CHILD(j, class_node) {
                Node* child = *j;
                total += max(child->count, floor);
            }
            FOR_EACH_CHILD(j, class_node) {
                Node* child = *j;
                child->prob = float(max(child->count, floor)) / float(total);
            }
#else
            em_compute_logprobs(class_node);
#  ifdef DEBUG
            format(cerr, "class lambda: %.5f %s\n",
                   class_node->lambda, FSTstring_map[class_label]);
#  endif
#endif
        }
    }
}

#ifdef DEBUG
void
NgramTree::print_lambdas()
{
    fst::vector<int> path;
    print_lambdas(root, path);
}

void
NgramTree::print_lambdas(Node* node, fst::vector<int>& path)
{
    BranchNode* branch = BRANCH_NODE(node);
    if (branch) {
        format(cout, "lambda: %.5f %6d", branch->lambda, branch->adj_count);
        for (int i = 0; i < path.size(); ++i)
            cout << ' ' << FSTstring_map[path[i]];
        cout << endl;
        FOR_EACH_CHILD(i, branch) {
            Node* child = *i;
            path.push_back(child->unit);
            print_lambdas(child, path);
            path.pop_back();
        }
    }
}

void
NgramTree::print_logprobs()
{
    fst::vector<int> path;
    print_logprobs(root, path);
}

void
NgramTree::print_logprobs(Node* node, fst::vector<int>& path)
{
    if (path.size() > 0) {
        format(cout, "logprob: %8.4f", node->prob);
        for (int i = 0; i < path.size(); ++i)
            cout << ' ' << FSTstring_map[path[i]];
        cout << endl;
    }
    BranchNode* branch = BRANCH_NODE(node);
    if (branch) {
        FOR_EACH_CHILD(i, branch) {
            Node* child = *i;
            path.push_back(child->unit);
            print_logprobs(child, path);
            path.pop_back();
        }
    }
}
#endif // DEBUG

// Given a node, compute the probability of unit, backing off as far as
// necessary using recursion.  Recursion ends with the specified uniform
// probability.
//
float
NgramTree::unit_prob(BranchNode* node, int unit, float uniform)
{
    if (!node) return uniform;

    Node* child = node->find(unit);
    if (child) return child->prob;

    return (1 - node->lambda) * unit_prob(node->backoff_node, unit, uniform);
}

// Given a node, compute the log probability of unit, backing off as far as
// necessary using recursion.
//
float
NgramTree::unit_logprob(BranchNode* node, int unit)
{
    assert(node != 0);

    Node* child = node->find(unit);
    if (child && child->prob <= 0) return child->prob;

    return node->lambda + unit_logprob(node->backoff_node, unit);
}

// Add any missing children to root node for convenience.  We set the counts to
// zero, and the probabilities will eventually be filled in with the uniform
// probability.
//
void
NgramTree::add_missing_unigrams()
{
    for (fst::set<int>::const_iterator i = units.begin(); i != units.end(); ++i) {
        int unit = *i;
        BranchNode::iterator j;
        if (!root->find(unit, j))
            root->insert(unit, new Node(0), j);
    }
}

// For every non-leaf node in tree, store pointer to its backoff node and set
// its state number.
//
void
NgramTree::set_backoff()
{
    cerr << "setting backoff pointers...";
    fst::vector<int> path;
    path.reserve(depth);
    set_backoff(root, path);
    cerr << "done" << endl;
}

void
NgramTree::set_backoff(BranchNode* node, fst::vector<int>& path)
{
    if (!node) return;

    if (path.size() == 0) {
        node->backoff_node = 0;
    } else {
        node->backoff_node = find_node(path.begin() + 1, path.end());
        assert(node->backoff_node != 0);
    }

    FOR_EACH_CHILD(i, node) {
        Node* child = *i;
        int unit = child->unit;
        path.push_back(unit);
        set_backoff(BRANCH_NODE(child), path);
        path.pop_back();
    }
}

// Given a path, find the corresponding node in the tree.  The path is specified
// as a range of fst::vector<int> iterators [i, end).
//
NgramTree::BranchNode*
NgramTree::find_node(fst::vector<int>::const_iterator i,
                     fst::vector<int>::const_iterator end)
{
    BranchNode* node = root;
    for (; i != end && node != 0; ++i)
        node = BRANCH_NODE(node->find(*i));
    return node;
}

// Remove all occurrences of <unknown> from n-gram and class trees.  We use the
// adj_count member to hold the counts adjusted for removed children.
//

void
NgramTree::remove_unknown()
{
    remove_unknown(class_root, 0);
    
    counts.clear();
    counts.insert(counts.begin(), depth + 1, 0);

    remove_unknown(root, 0);

    units.erase(unknown);
}

void
NgramTree::remove_unknown(BranchNode* node, int d)
{
    if (!node) return;
    BranchNode::iterator i;
    Node* unknown_child = node->find(unknown, i);
    if (unknown_child) {
        node->adj_count -= unknown_child->count;
        delete unknown_child;
        node->erase(i);
    }
    FOR_EACH_CHILD(j, node)
        remove_unknown(BRANCH_NODE(*j), d + 1);
    counts[d + 1] += node->n_children();
}

// Eventually we might want to make this assign states only to those reachable
// from the initial state.  This would trim the resulting FST.
//
void
NgramTree::assign_states(BranchNode* node)
{
    if (!node) {
        n_states = n_arcs = 0;
        states.clear();
        node = root;
    }

    states.push_back(node);
    node->state = n_states++;

    FOR_EACH_CHILD(i, node) {
        BranchNode* child_branch = BRANCH_NODE(*i);
        if (child_branch) assign_states(child_branch);
    }
}

void
NgramTree::dump_FSTNgram2(const char* filename, const fst::set<int>& open_classes)
{
    if (!filename || !filename[0] || (filename[0] == '-' && !filename[1])) {
        dump_FSTNgram2(cout, open_classes);
    } else {
        ofstream os(filename);
        if (!os) FSTFatal("cannot write file %s", filename);
        dump_FSTNgram2(os, open_classes);
    }
}

void
NgramTree::dump_FSTNgram2(ostream& os, const fst::set<int>& open_classes)
{
    cerr << "dumping FSTNgram2...";

    assign_states();

    // Dump header.
    //
    os << "#FSTNgram2 Version=2";
    if (invert_depth || det_flag)
        os << " Deterministic";
    os << endl;

    // Dump order (n) of n-gram.
    //
    FSTDumpUnsignedShort(depth, os);

    // Dump list of (label, index) for all units.
    //
    assert(units.size() < (1<<16));
    FSTDumpUnsignedShort(units.size(), os);
    FOR_EACH_UNIT(i, units) {
        assert(*i < (1<<16));
        FSTDumpString(FSTstring_map[*i], os);
        FSTDumpUnsignedShort(*i, os);
    }

    // If any open classes were specified, remove them from the class tree and
    // dump out the within-class probabilities to FSTs, one per class.
    //
    if (class_root) {
        for (fst::set<int>::const_iterator ci = open_classes.begin();
             ci != open_classes.end(); ++ci) {
            BranchNode::iterator ni;
            Node* n = class_root->find(*ci, ni);
            if (!n) {
                cerr << endl << "warning: open class `"
                     << FSTstring_map[*ci] << "' not found -- skipping" << endl
                     << "...";
            } else {
                FSTGeneric class_fst = FSTBasic();
                class_fst->add_states(2);
                class_fst->set_initial(0);
                class_fst->set_final(1);
                BranchNode* class_node = BRANCH_NODE(n);
                FOR_EACH_CHILD(i, class_node) {
                    Node* word_node = *i;
                    class_fst->add_arc(0, 1, word_node->unit, word_node->unit,
                                       -word_node->prob);
                }
                class_fst->dump_ascii(string(FSTstring_map[*ci]) + ".wc.fst");
                class_root->erase(ni);
            }
        }
    }

    // Dump class information.
    //
    if (class_root) {
        FSTDumpUnsignedShort(class_root->n_children(), os);
        FOR_EACH_CHILD(i, class_root) {
            BranchNode* class_node = BRANCH_NODE(*i);
            FSTDumpUnsignedShort(class_node->unit, os);
            FSTDumpUnsignedShort(class_node->n_children(), os);
            FOR_EACH_CHILD(j, class_node) {
                Node* word_node = *j;
                FSTDumpString(FSTstring_map[word_node->unit], os);
                FSTDumpFloat(invert_depth ? 0 : -word_node->prob, os);
            }
        }
    } else {
        FSTDumpUnsignedShort(0, os);
    }

    // Dump initial state.
    //
    BranchNode* initial_node = BRANCH_NODE(root->find(initial_final));
    if (!initial_node)
        initial_node = BRANCH_NODE(root->find(cmu_initial));
    if (!initial_node && depth == 1)
        initial_node = root;
    assert(initial_node != 0);
    FSTDumpInt(initial_node->state, os);

    // Figure out which unit is final.
    //
    int final_unit;
    if (root->find(initial_final))
        final_unit = initial_final;
    else
        final_unit = cmu_final;

    // Dump states and their connections.
    //
    root->unit = 0;
    int state = 0;
    fst::vector<int> context;
    FSTDumpInt(n_states, os);
    dump_FSTNgram2(os, root, 0, final_unit, context, state);

    cerr << n_states << " states..."
         << n_arcs << " arcs...done" << endl;
}

// Recursively dump out state and its transitions.  We do a recursive,
// depth-first traversal so that we can keep track of the context (history).  We
// traverse the nodes in the same order as when the state numbers were assigned,
// but we verify this anyway just to be safe.
//
void
NgramTree::dump_FSTNgram2(ostream& os, BranchNode* node, BranchNode* parent,
			  int final_unit, fst::vector<int>& context, int& state)
{
    assert(node->state == state);

    BranchNode* backoff_node = node->backoff_node;

    // n of n-grams this node is predicting.
    //
    int n = context.size() + 1;

    // Get ready for possible incremental n-gram.
    //
    bool within_invert = (invert_depth && n <= invert_depth);
    BranchNode* invert_node = 0;
    if (invert_depth) {
        invert_node = node;
        for (int i = n - invert_depth; i > 0; --i)
            invert_node = invert_node->backoff_node;
        assert(invert_node != 0);
    }

    // Count number of transitions.
    //
    int n_transitions = 0;
    FOR_EACH_CHILD(i, node) {
        Node* child = *i;
        int unit = child->unit;
        if (child->prob > 0)       continue;
        if (unit == initial_final) continue;
        if (unit == cmu_initial)   continue;
        if (unit == cmu_final)     continue;
        ++n_transitions;
    }
    n_arcs += n_transitions;

    // Dump final weight.  We follow backoff here so that we can later lookup
    // the final weight quickly.
    //
    float final_weight = -unit_logprob(node, final_unit);
    if (invert_node)
        final_weight += unit_logprob(invert_node, final_unit);
    FSTDumpFloat(final_weight, os);

    // Dump backoff state (-1 for none) and score if backoff state exists.
    //
    if (backoff_node) {
        FSTDumpInt(backoff_node->state, os);
        FSTDumpFloat(within_invert ? 0 : -node->lambda, os);
        ++n_arcs;
    } else {
        FSTDumpInt(-1, os);
    }

    // Dump immediate context.
    //
    FSTDumpUnsignedShort(node->unit, os);

    // Dump parent.
    //
    FSTDumpInt(parent ? parent->state : -1, os);

    // Dump transitions.
    //
    assert(n_transitions < (1<<16));
    FSTDumpUnsignedShort(n_transitions, os);
    FOR_EACH_CHILD(i, node) {
        Node* child = *i;
        int unit = child->unit;
        if (child->prob > 0)       continue;
        if (unit == initial_final) continue;
        if (unit == cmu_initial)   continue;
        if (unit == cmu_final)     continue;
        --n_transitions;

        // The unit's weight, possibly incremental.
        //
        float w = -child->prob;
        if (invert_node) {
            if (invert_node == node)
                w = 0;
            else
                w += unit_logprob(invert_node, unit);
        }

#if 0
        // Compute next state by adding current unit to end of history and
        // backing off as much as necessary.
        //
        BranchNode* next_node = 0;
        context.push_back(unit);
        for (fst::vector<int>::iterator j = context.begin();
             !next_node && j != context.end(); ++j) {
            next_node = find_node(j, context.end());
        }
        context.pop_back();
        if (!next_node)
            next_node = root;
        assert(next_node->state >= 0);
#endif

        // Dump transition.
        //
#if 0
        FSTDumpInt(next_node->state, os);
#endif
        FSTDumpUnsignedShort(unit, os);
        FSTDumpFloat(w, os);
    }
    assert(n_transitions == 0);

    // Continue depth-first recursion.
    //
    FOR_EACH_CHILD(i, node) {
        BranchNode* child_branch = BRANCH_NODE(*i);
        if (child_branch) {
            context.push_back(child_branch->unit);
            dump_FSTNgram2(os, child_branch, node,
			   final_unit, context, ++state);
            context.pop_back();
        }
    }
}

void
NgramTree::dump_FSTNgram3(const char* filename, const fst::set<int>& open_classes)
{
    if (!filename || !filename[0] || (filename[0] == '-' && !filename[1])) {
        dump_FSTNgram3(cout, open_classes);
    } else {
        ofstream os(filename);
        if (!os) FSTFatal("cannot write file %s", filename);
        dump_FSTNgram3(os, open_classes);
    }
}

void
NgramTree::dump_FSTNgram3(ostream& os, const fst::set<int>& open_classes)
{
    cerr << "dumping FSTNgram3...";

    assign_states();

    // Dump header.
    //
    os << "#FSTNgram3 Version=2";
    if (invert_depth || det_flag)
        os << " Deterministic";
    os << endl;

    // Dump order (n) of n-gram.
    //
    FSTDumpUnsignedShort(depth, os);

    // Dump list of (label, index) for all units.
    //
    FSTDumpInt(units.size(), os);
    FOR_EACH_UNIT(i, units) {
        FSTDumpString(FSTstring_map[*i], os);
        FSTDumpInt(*i, os);
    }

    // If any open classes were specified, remove them from the class tree and
    // dump out the within-class probabilities to FSTs, one per class.
    //
    if (class_root) {
        for (fst::set<int>::const_iterator ci = open_classes.begin();
             ci != open_classes.end(); ++ci) {
            BranchNode::iterator ni;
            Node* n = class_root->find(*ci, ni);
            if (!n) {
                cerr << endl << "warning: open class `"
                     << FSTstring_map[*ci] << "' not found -- skipping" << endl
                     << "...";
            } else {
                FSTGeneric class_fst = FSTBasic();
                class_fst->add_states(2);
                class_fst->set_initial(0);
                class_fst->set_final(1);
                BranchNode* class_node = BRANCH_NODE(n);
                FOR_EACH_CHILD(i, class_node) {
                    Node* word_node = *i;
                    class_fst->add_arc(0, 1, word_node->unit, word_node->unit,
                                       -word_node->prob);
                }
                class_fst->dump_ascii(string(FSTstring_map[*ci]) + ".wc.fst");
                class_root->erase(ni);
            }
        }
    }

    // Dump class information.
    //
    if (class_root) {
        FSTDumpInt(class_root->n_children(), os);
        FOR_EACH_CHILD(i, class_root) {
            BranchNode* class_node = BRANCH_NODE(*i);
            FSTDumpInt(class_node->unit, os);
            FSTDumpInt(class_node->n_children(), os);
            FOR_EACH_CHILD(j, class_node) {
                Node* word_node = *j;
                FSTDumpString(FSTstring_map[word_node->unit], os);
                FSTDumpFloat(invert_depth ? 0 : -word_node->prob, os);
            }
        }
    } else {
        FSTDumpInt(0, os);
    }

    // Dump initial state.
    //
    BranchNode* initial_node = BRANCH_NODE(root->find(initial_final));
    if (!initial_node)
        initial_node = BRANCH_NODE(root->find(cmu_initial));
    if (!initial_node && depth == 1)
        initial_node = root;
    assert(initial_node != 0);
    FSTDumpInt(initial_node->state, os);

    // Figure out which unit is final.
    //
    int final_unit;
    if (root->find(initial_final))
        final_unit = initial_final;
    else
        final_unit = cmu_final;

    // Dump states and their connections.
    //
    root->unit = 0;
    int state = 0;
    fst::vector<int> context;
    FSTDumpInt(n_states, os);
    dump_FSTNgram3(os, root, 0, final_unit, context, state);

    cerr << n_states << " states..."
         << n_arcs << " arcs...done" << endl;
}

// Recursively dump out state and its transitions.  We do a recursive,
// depth-first traversal so that we can keep track of the context (history).  We
// traverse the nodes in the same order as when the state numbers were assigned,
// but we verify this anyway just to be safe.
//
void
NgramTree::dump_FSTNgram3(ostream& os, BranchNode* node, BranchNode* parent,
			  int final_unit, fst::vector<int>& context, int& state)
{
    assert(node->state == state);

    BranchNode* backoff_node = node->backoff_node;

    // n of n-grams this node is predicting.
    //
    int n = context.size() + 1;

    // Get ready for possible incremental n-gram.
    //
    bool within_invert = (invert_depth && n <= invert_depth);
    BranchNode* invert_node = 0;
    if (invert_depth) {
        invert_node = node;
        for (int i = n - invert_depth; i > 0; --i)
            invert_node = invert_node->backoff_node;
        assert(invert_node != 0);
    }

    // Count number of transitions.
    //
    int n_transitions = 0;
    FOR_EACH_CHILD(i, node) {
        Node* child = *i;
        int unit = child->unit;
        if (child->prob > 0)       continue;
        if (unit == initial_final) continue;
        if (unit == cmu_initial)   continue;
        if (unit == cmu_final)     continue;
        ++n_transitions;
    }
    n_arcs += n_transitions;

    // Dump final weight.  We follow backoff here so that we can later lookup
    // the final weight quickly.
    //
    float final_weight = -unit_logprob(node, final_unit);
    if (invert_node)
        final_weight += unit_logprob(invert_node, final_unit);
    FSTDumpFloat(final_weight, os);

    // Dump backoff state (-1 for none) and score if backoff state exists.
    //
    if (backoff_node) {
        FSTDumpInt(backoff_node->state, os);
        FSTDumpFloat(within_invert ? 0 : -node->lambda, os);
        ++n_arcs;
    } else {
        FSTDumpInt(-1, os);
    }

    // Dump immediate context.
    //
    FSTDumpInt(node->unit, os);

    // Dump parent.
    //
    FSTDumpInt(parent ? parent->state : -1, os);

    // Dump transitions.
    //
    FSTDumpInt(n_transitions, os);
    FOR_EACH_CHILD(i, node) {
        Node* child = *i;
        int unit = child->unit;
        if (child->prob > 0)       continue;
        if (unit == initial_final) continue;
        if (unit == cmu_initial)   continue;
        if (unit == cmu_final)     continue;
        --n_transitions;

        // The unit's weight, possibly incremental.
        //
        float w = -child->prob;
        if (invert_node) {
            if (invert_node == node)
                w = 0;
            else
                w += unit_logprob(invert_node, unit);
        }

#if 0
        // Compute next state by adding current unit to end of history and
        // backing off as much as necessary.
        //
        BranchNode* next_node = 0;
        context.push_back(unit);
        for (fst::vector<int>::iterator j = context.begin();
             !next_node && j != context.end(); ++j) {
            next_node = find_node(j, context.end());
        }
        context.pop_back();
        if (!next_node)
            next_node = root;
        assert(next_node->state >= 0);
#endif

        // Dump transition.
        //
#if 0
        FSTDumpInt(next_node->state, os);
#endif
        FSTDumpInt(unit, os);
        FSTDumpFloat(w, os);
    }
    assert(n_transitions == 0);

    // Continue depth-first recursion.
    //
    FOR_EACH_CHILD(i, node) {
        BranchNode* child_branch = BRANCH_NODE(*i);
        if (child_branch) {
            context.push_back(child_branch->unit);
            dump_FSTNgram3(os, child_branch, node,
			   final_unit, context, ++state);
            context.pop_back();
        }
    }
}

static void
usage()
{
    cerr <<
        "usage: ngram_fst [options] in.tree out.fst\n"
        "\n"
        "-a            maintain ARPA scores precisely (not generally recommended)\n"
        "-d            enable FST deterministic mode\n"
        "-i m          invert m-gram scores to make incremental n-gram\n"
        "-n n          prune tree to depth n [0 = no pruning]\n"
        "-o class      leave class open, dump within-class score FST to class.fst\n"
        "-r            reverse tree\n"
        "-t threshold  prune to count threshold [1 = no pruning]\n"
        "-u            exclude <unknown>\n"
        "-v            large vocabulary (> 64K words)\n"
        "\n";
    exit(1);
}

static int
fst_from_ngram(int argc, char** argv)
{
    if (argc < 0) {
        usage();
        return 1;
    }

    int  n = 0;
    int  invert = 0;
    bool reverse_flag = false;
    int  threshold = 1;
    bool exclude_unknown = false;
    bool det_flag = false;
    bool arpa_flag = false;
    bool large_vocabulary = false;
    fst::set<int> open_classes;

    bool error = false;
    int c;
    while ((c = getopt(argc, argv, "adi:n:o:rt:uv")) != EOF) {
        switch (c) {
        case 'a':
            arpa_flag = true;
            break;
        case 'd':
            det_flag = true;
            break;
        case 'i':
            invert = atoi(optarg);
            break;
        case 'n':
            n = atoi(optarg);
            break;
        case 'o':
            open_classes.insert(FSTstring_map[optarg]);
            break;
        case 'r':
            reverse_flag = true;
            break;
        case 't':
            threshold = atoi(optarg);
            break;
        case 'u':
            exclude_unknown = true;
            break;
        case 'v':
            large_vocabulary = true;
            break;
        default:
            error = true;
            break;
        }
    }
    argc -= optind;
    argv += optind;

    if (error || argc != 2)
        usage();

    const char* tree_file = *argv++;
    const char* fst_file = *argv++;

    NgramTree tree(tree_file, reverse_flag, n, invert,
                   threshold, exclude_unknown, det_flag, arpa_flag);

    if (large_vocabulary) {
        tree.dump_FSTNgram3(fst_file, open_classes);
    } else {
        tree.dump_FSTNgram2(fst_file, open_classes);
    }

    return 0;
}

// Register as subcommand.
//
static FstTool from_ngram_tool("from_ngram", fst_from_ngram);

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 80
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
