////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

/******************************************************************************
=head1 NAME

fst_info - output information about an FST

=head1 SYNOPSIS

B<fst_info> [I<OPTIONS>] I<fst-in>

=head1 DESCRIPTION

B<fst_info> outputs information on the size of the input and output alphabets,
the number of states, the number of arcs, and optionally the number of epsilon
input, output, and input/output arcs.

The specification of the input FST I<fst-in> can be a filename, an expression
(see fst_expression(1)), or "-" indicating standard input.

=head1 OPTIONS

=over 12

=item B<-e>

Output the number of epsilon input, output, and input/output arcs.

=back

=head1 SEE ALSO

fst_alphabet(1)

=head1 AUTHOR

I. Lee Hetherington <ilh@mit.edu>

=cut
******************************************************************************/

#include <iostream>
#include <string>
#include "fst.h"

using namespace std;

static void
usage()
{
    cerr <<
        "usage: fst info [options] in.fst\n"
        "\n"
        "-e     display information about epsilons\n"
        "\n";
}

static string
pad(int i)
{
    static char buf[16];
    sprintf(buf, "%8d", i);
    return string(buf);
}

static int
fst_info(int argc, char** argv)
{
    if (argc < 0) {
        usage();
        return 1;
    }

    bool epsilon_flag = false;
    bool error = false;

    char c;
    while ((c = getopt(argc, argv, "e")) != EOF) {
        switch (c) {
        case 'e':
            epsilon_flag = true;
            break;
        default:
            error = true;
            break;
        }
    }
    argc -= optind;
    argv += optind;
    if (error || argc != 1) {
        usage();
        return 1;
    }

    const char* in_file  = argv[0];

    FSTGeneric x = FSTLoad(in_file);

    int n_final = 0;
    int n_i_eps = 0;
    int n_o_eps = 0;
    int n_io_eps = 0;

    for (int p = 0; p < x->n_states(); p++) {
    if (x->is_final(p)) n_final++;
        if (epsilon_flag) {
            for (FSTArcIter a = x->arcs_out(p); a.valid(); a++) {
                n_o_eps  += (a->output == 0);
                n_i_eps  += (a->input == 0);
                n_io_eps += (a->input == 0 && a->output == 0);
            }
        }
    }

    cout <<
        "\n"
        "# input symbols:   " << pad(x->input_alphabet().size()) << "\n"
        "# output symbols:  " << pad(x->output_alphabet().size()) << "\n"
        "\n"
        "# states:          " << pad(x->n_states()) << "\n"
        "# final states:    " << pad(n_final) << "\n"
        "# arcs:            " << pad(x->n_arcs()) << "\n"
        "\n";
    if (epsilon_flag) {
        cout <<
            "# i eps arcs:      " << pad(n_i_eps) << "\n"
            "# o eps arcs:      " << pad(n_o_eps) << "\n"
            "# i/o eps arcs:    " << pad(n_io_eps) << "\n"
            "\n";
    }

    return 0;
}

// Register as subcommand.
//
static FstTool info_tool("info", fst_info);

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 80
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
