////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2008, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

/******************************************************************************
=head1 NAME

fst_letter_to_sound - generate word pronunciations

=head1 SYNOPSIS

B<fst_letter_to_sound> [I<OPTIONS>] I<letter-to-sound-fst>

=head1 DESCRIPTION

B<fst_letter_sound> generates baseform pronuncations given a letter-to-sound
FST.  The FST can have labels of the form letter:phone or letters/phonemes, the
latter representing multigram graphones.  Input words are read from standard
input.  Output is of the form I<word : pronuncation>, and there may be more than
one line for each word.

=head1 OPTIONS

=over 12

=item B<-i>

Ignore case (by converting to lower case).

=item B<-m>

Output missing entries as <???> instead of as empty pronunciations.

=item B<-n> I<N>

Output top I<N> pronunciations for each word.  Default is 10.

=item B<-s> I<S>

Limit I<N>-best output to pronunciations that score within I<S> of the
best-scoring pronunciation.  Default is 0.01.

=back

=head1 AUTHOR

I. Lee Hetherington <ilh@mit.edu>

=cut
******************************************************************************/

#include "fst.h"
#include <iostream>
#include <string>
#include <ctype.h>

using namespace std;

void
usage()
{
    cerr <<
        "usage: fst_letter_to_sound [options] l2s.fst\n"
        "\n"
        " -i          ignore case (convert to lowercase)\n"
        " -m          output missing entries as <\?\?\?>\n"
        " -n N   [10] output up to top N choices for each word\n"
        " -s S  [.01] output top choices scoring within S of best\n"
        "\n";
}

static string
to_lower_case(string word)
{
    for (string::size_type i = 0; i < word.length(); ++i) {
        word[i] = tolower(word[i]);
    }
    return word;
}

static int
fst_letter_to_sound(int argc, char** argv)
{
    if (argc < 0) {
        usage();
        return 1;
    }

    int N = 100;
    float S = float(.01);
    bool ignore = false;
    bool missing = false;
    bool error = false;

    char c;
    while ((c = getopt(argc, argv, "imn:s:")) != EOF) {
        switch (c) {
        case 'i':
            ignore = true;
            break;
        case 'm':
            missing = true;
            break;
        case 'n':
            N = atoi(optarg);
            break;
        case 's':
            S = atof(optarg);
            break;
        default:
            error = true;
            break;
        }
    }
    argc -= optind;
    argv += optind;
    if (error || argc != 1) {
        usage();
        return 1;
    }

    if (S == 0) {
        S = float(0.01);
        cerr << "warning: using -s = " << S << " instead of 0" << endl;
    }

    const char* l2s_file  = argv[0];

    FSTLetterToSound l2s(l2s_file, N, S);

    string word;
    while (getline(cin, word)) {

        string orig_word = word;
        if (ignore) {
            word = to_lower_case(word);
        }

        // call LTS
        //
        string result = l2s.generateString(word);

        if (result.size() == 0) {

            cout << word << " : ";
            if (missing)
                cout << "<\?\?\?>";
            cout << endl;

        }  else {

            // split at comma
            //
            string::size_type i = 0, j;
            do {
                j = result.find(',', i);
                string item = (j == string::npos) ? result.substr(i) : result.substr(i, j - i);
                cout << orig_word << " : " << item << endl;
                i = j + 1;
            } while (j != string::npos);
        }

    }

    return 0;
}

// Register as subcommand.
//
static FstTool letter_to_sound_tool("letter_to_sound", fst_letter_to_sound);

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 80
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
