////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

/******************************************************************************
=head1 NAME

fst_minimize - minimize an FST

=head1 SYNOPSIS

B<fst_minimize> [I<OPTIONS>] I<fst-in> I<fst-out>

=head1 DESCRIPTION

B<fst_minimize> performs minimization of an FST by merging equivalent states,
and thus the resulting I<fst-out> is never larger than I<fst-in>.  The FST is
treated as an FSA in that input:output labels are treated as a unit.

Note that there is no requirement that I<fst-in> be determinized or even have
epsilons removed.

If you do want to compute the minimal deterministic FST, then you would
generally apply fst_remove_epsilons(1), fst_determinize(1), fst_push_weights(1),
and then finally B<fst_minimize>.  (If fst_push_outputs(1) was implemented, you
would also run that as well.)  See fst_optimize(1) that wraps all of the above
operations into a single tool.

The specification of the input FST I<fst-in> can be a filename, an expression
(see fst_expression(1)), or "-" indicating standard input.  The specification of
the output FST I<fst-out> can be a filename or "-" indicating standard output.

=head1 OPTIONS

=over 12

=item B<-B>

Output I<fst-out> in binary format.

=item B<-e> I<thresh>

Use I<thresh> as the threshold for determining if two weights are to be
considered equal.  The default I<thresh> is 0.001.

=back

=head1 ALGORITHM

Hopcroft's O(N log N) algorithm for cyclic FSAs and Revuz's O(N) algorithm
for acyclic FSAs.

=head1 SEE ALSO

fst_determinize(1),
fst_optimize(1),
fst_push_weights(1),
fst_remove_epsilons(1)

=head1 AUTHOR

I. Lee Hetherington <ilh@mit.edu>

=cut
******************************************************************************/

#include <iostream>
#include "fst.h"

using namespace std;

static void
usage()
{
    cerr <<
        "usage: fst minimize [options] in.fst out.fst\n"
        "\n"
        "-B     binary output\n"
        "-e x   weight difference to consider equivalent [0.001]\n"
        "\n";
}

static int
fst_minimize(int argc, char** argv)
{
    if (argc < 0) {
        usage();
        return 1;
    }

    float eps = float(0.001);
    bool binary_flag = false;

    bool error = false;
    char c;
    while ((c = getopt(argc, argv, "Be:")) != EOF) {
        switch (c) {
        case 'B':
            binary_flag = true;
            break;
        case 'e':
            eps = atof(optarg);
            break;
        default:
            error = true;
            break;
        }
    }
    argc -= optind;
    argv += optind;

    if (argc != 2) {
    usage();
        return 1;
    }

    const char* in_file  = argv[0];
    const char* out_file = argv[1];

    FSTGeneric x = FSTLoad(in_file);
    x = FSTMinimize(x, eps);

    if (binary_flag)
        x->dump_binary(out_file);
    else
        x->dump_ascii(out_file);

    return 0;
}

// Register as subcommand.
//
static FstTool minimize_tool("minimize", fst_minimize);

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 80
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
