////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

/******************************************************************************
=head1 NAME

fst_nbest - compute N-best paths through FST or graph containing them

=head1 SYNOPSIS

B<fst_nbest> [B<-g> | B<-s>] [B<-i> | B<-o>] [B<-n> I<N>] [B<-t> I<T>] [B<-p>] I<fst-in> I<out>

=head1 DESCRIPTION

B<fst_nbest> computes the N-best paths through I<fst-in> or an FST graph
containing them, utilizing an A* search.  The state potentials are used as the
A* heuristic for each state, and if they are not computed, the A* search can
perform poorly.  Thus, if you have not already computed them in some other tool,
the B<-p> option is recommended.

The optional string output format (B<-s>), outputs the N-best outputs one per
line, with the total path score followed by the output sequence or the input
sequence if B<-i> is specified.

The specification of the input FST I<fst-in> can be a filename, an expression
(see fst_expression(1)), or "-" indicating standard input.  The specification of
the output I<out> can be a filename or "-" indicating standard output.

=head1 OPTIONS

=over 12

=item B<-g>

Generate graph instead of N-best.

=item B<-i>

Only output unique input-label sequences.  This also causes B<-s> to output
input sequences rather than output sequences.

=item B<-o>

Only output unique output-label sequences.

=item B<-n> I<N>

Find the I<N> best.  The default I<N> is 1.

=item B<-p>

Compute potential of each state, used for the A* heuristic.  This is recommended
if you have not already computed the potentials.

=item B<-s>

Generate string representation for I<out> instead of FST.

=item B<-t> I<T>

Use score threshold I<T> during A* search.  The default I<T> is 20.0.

=back

=head1 SEE ALSO

fst_potential(1)

=head1 AUTHOR

I. Lee Hetherington <ilh@mit.edu>

=cut
******************************************************************************/

#include <iostream>
#include <fstream>
#include "fst.h"
#include "fst/format.h"

using namespace std;

static void
usage()
{
    cerr <<
        "usage: fst nbest [-g | -s] [-i | -o] [-n N] [-t T] [-p] in.fst out.fst\n"
        "\n"
        "-g     generate graph instead of N-best\n"
        "-i     only output unique input-label sequences\n"
        "-o     only output unique output-label sequences\n"
        "-n N   find N best [1]\n"
        "-s     output string representation instead of FST\n"
        "-t T   use T as score threshold [20]\n"
        "-p     compute potential of each state (recommended)\n"
        "\n";
}

static int
fst_nbest(int argc, char** argv)
{
    if (argc < 0) {
        usage();
        return 1;
    }

    enum { Fst, Graph, Text } type = Fst;
    char c;
    bool error = false;
    int n = 1;
    float thresh = 20;
    bool potential = false;
    FSTAstarSide unique = FSTNone;

    while ((c = getopt(argc, argv, "gin:opst:")) != EOF) {
        switch (c) {
        case 'g':
            type = Graph;
            break;
        case 'i':
            unique = FSTInput;
            break;
        case 'n':
            n = atoi(optarg);
            if (n < 1)
                error = true;
            break;
        case 'o':
            unique = FSTOutput;
            break;
        case 'p':
            potential = true;
            break;
        case 's':
            type = Text;
            break;
        case 't':
            thresh = atof(optarg);
            break;
        case '?':
        default:
            error = true;
            break;
        }
    }
    argc -= optind;
    argv += optind;

    if (error || argc != 2) {
        usage();
        return 1;
    }

    const char* in_file  = argv[0];
    const char* out_file  = argv[1];

    FSTGeneric in = FSTLoad(in_file);
    if (potential)
        in = FSTPotential(in);

    if (type == Graph) {

        FSTAstarGraph(in, unique, thresh, 0)->dump_ascii(out_file);

    } else if (type == Fst) {

        FSTAstarNbest(in, unique, n, thresh)->dump_ascii(out_file);

    } else /* Text */ {

        ofstream os_file;
        ostream* osp = &cout;
        
        if (out_file[0] != '-' || out_file[1] != 0) {
            os_file.open(out_file);
            osp = &os_file;
        }
        
        FSTSemiring semiring = in->semiring();
        FSTGeneric fst = FSTAstarNbest(in, unique, n, thresh);

        // Iterate through N-best list.  They are in the order of the initial
        // arcs.
        //
        if (fst->initial() >= 0) {
            for (FSTArcIter i = fst->arcs_out(fst->initial()); i.valid(); i++) {
                
                // Iterate through arcs of an N-best entry.
                //
                float weight = semiring->one;
                string words;
                for (FSTArcIter a = i; a.valid(); a = fst->arcs_out(a->next)) {
                    weight = semiring->times(weight, a->weight);
                    if (fst->is_final(a->next))
                        weight = semiring->times(weight, fst->final_weight(a->next));
                    if (a->output) {
                        if (words.size() > 0) words += ' ';
                        words += FSTstring_map[a->output];
                    }
                }
                
                format(*osp, "%9.4f %s\n", weight, words.c_str());
                
            }
        }

    }

    return 0;
}

// Register as subcommand.
//
static FstTool nbest_tool("nbest", fst_nbest);

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 80
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
