////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

/******************************************************************************
=head1 NAME

fst_optimize - optimize FST

=head1 SYNOPSIS

B<fst_optimize> [I<OPTIONS>] I<fst-in> I<fst-out>

=head1 DESCRIPTION

B<fst_optimize> attempts to compute the minimal (i.e., fewest number of states)
deterministic FST I<fst-out> for the given I<fst-in>.  It performs epsilon
removal, determinization, weight pushing, and finally minimization.

B<fst_optimize> itself is really a wrapper around other tools.  For example,

    fst_optimize in.fst out.fst

is equivalent to

    fst_remove_epsilons in.fst - | fst_determinize - - | \
    fst_push_weights - - | fst_minimize - out.fst

but is more efficient because the intermediate FSTs are kept in memory.

The specification of the input FST I<fst-in> can be a filename, an expression
(see fst_expression(1)), or "-" indicating standard input.  The specification of
the output FST I<fst-out> can be a filename or "-" indicating standard output.

=head1 OPTIONS

=over 12

=item B<-A>

Output I<fst-out> in ASCII format.  The default is binary.

=item B<-a>

Treat I<fst-in> as an FSA with compound input:output labels.  No output pushing
will take place.

=item B<-d> I<N>

Allow nondeterminism of degree I<N> through.  For example, if there would be I<N>
or fewer arcs with identical input labels leaving a particular output state, they
would not be subject to determinization.  This can be used to make size/determinism
tradeoffs.  The default value of <N> is 1.

=item B<-e> I<thresh>

Treat two weights as equal if their difference is less than threshold I<thresh>.  The
default value of I<thresh> is 0.001.

=back

=head1 ALGORITHM

The algorithm is based on that published by Mehyar Mohri while he was at AT&T
[insert reference], and extended by the author.

=head1 SEE ALSO

fst_determinize(1),
fst_minimize(1),
fst_push_weights(1),
fst_remove_epsilons(1)

=head1 AUTHOR

I. Lee Hetherington <ilh@mit.edu>

=cut
******************************************************************************/

#include <iostream>
#include "fst.h"

using namespace std;

static void
usage()
{
    cerr <<
        "usage fst optimize [options] in.fst out.fst\n"
        "\n"
        "-A    output ASCII file\n"
        "-a    automata mode (treat input:output as an atomic unit)\n"
        "-d n  degree of non-determinism to allow [1]\n"
        "-e x  weight difference to consider equivalent [0.001]\n"
        "\n";
}

static int
fst_optimize(int argc, char** argv)
{
    if (argc < 0) {
        usage();
        return 1;
    }

    float eps = float(0.001);
    bool fsa = false;
    int degree = 1;
    bool ascii = false;

    bool error = false;
    char c;
    while ((c = getopt(argc, argv, "Aad:e:")) != EOF) {
        switch (c) {
        case 'A':
            ascii = true;
            break;
        case 'a':
            fsa = true;
            break;
        case 'd':
            degree = atoi(optarg);
            break;
        case 'e':
            eps = atof(optarg);
            break;
        default:
            error = true;
            break;
        }
    }
    argc -= optind;
    argv += optind;

    if (error || argc != 2) {
    usage();
        return 1;
    }

    const char* in_file  = argv[0];
    const char* out_file = argv[1];

    FSTGeneric x = FSTLoad(in_file);

    if (fsa)
        x = FSTOptimizeA(x, degree, eps);
    else
        x = FSTOptimize(x, degree, eps);

    if (ascii)
        x->dump_ascii(out_file);
    else
        x->dump_binary(out_file);

    return 0;
}

// Register as subcommand.
//
static FstTool optimize_tool("optimize", fst_optimize);

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 80
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
