/* -*- C++ -*-
 *
 * Copyright (c) 2007
 * Spoken Language Systems Group
 * MIT Computer Science and Artificial Intelligence Laboratory
 * Massachusetts Institute of Technology
 *
 * All Rights Reserved
 */

#define DUMP 1

#include <iostream>
#include <fstream>
#include "fst.h"
#include "fst/FSTPronRules.h"

// 2 and 3 supported, 0 means use latest
#define PFST_VERSION 3

using namespace std;

static void
usage()
{
    cerr <<
        "usage: (1) fst pocket_build [options] G.jsgf/fst MCP.fst L.fst L2S.fst(s) out.pfst\n"
        "       (2) fst pocket_build [options] -c n (-j G.jsgf | -l G.txt) MCP.fst L.fst L2S.fst(s) out.pfst\n"
        "       (3) fst pocket_build [options] -c n MCP.fst LG.fst out.pfst\n"
        "       (4) fst pocket_build [options] -p in.fst out.pfst\n"
        "\n"
        "(1) compiles a JSGF grammar file to a static FST or a dynamic main FST.\n"
        "\n"
        "-b N   number of weight bits [5]\n"
        "-f     G is an FST rather than a JSGF\n"
        "\n"
        "(2) compiles either a JSGF grammar file or phrase list file to dynamic class\n"
        "    name n (must begin with $).\n"
        "\n"
        "-b N   number of weight bits [5]\n"
        "-j F   compile JSGF file F\n"
        "-l F   compile phrase list file F\n"
        "\n"
        "(3) builds dynamic class FST from LG class contents (phoneme:word).\n"
        "\n"
        "-b N   number of weight bits [5]\n"
        "\n"
        "(4) packs an existing FST to PocketSUMMIT format.\n"
        "\n"
        "-a F   get input alphabet from FST file F, needed for dynamic splicing\n"
        "-b N   number of weight bits [5]\n"
        "-d F   debug FST output file F\n"
        "-v F   dump vocabulary to file F\n"
        "-V     increase verbosity\n"
        "\n"
        "L2S.fsts can be a comma-separated list of letter-to-sound FSTs with no whitespace\n"
        "(e.g., l2s-1.fst,l2s-2.fst).  L.fst is generated with fst from_baseforms -d.\n"
        "\n";
}

static string
loadFile(const char* filename)
{
    ifstream is(filename);
    string result;

    for (string line; getline(is, line);) {
        result += line;
        result += '\n';
    }

    return result;
}

static void
dumpFile(FSTMemory* pfst, const char* filename)
{
    ofstream os(filename, ios_base::binary);
    os.write(reinterpret_cast<const char*>(pfst->base()), pfst->size());
}

int
fst_pocket_build(int argc, char** argv)
{
    if (argc < 0) {
        usage();
        return 1;
    }

    fst::set<int> alphabet;

    int weightBits = 5;
    bool error = false;
    bool pack = false;
    bool fFlag = false;
    enum { Jsgf, List, None } gFileType = None;
    const char* aFile = NULL;
    const char* dFile = NULL;
    const char* gFile = NULL;
    const char* vFile = NULL;
    const char* className = NULL;
    int verbosity = 0;

    char c;
    while ((c = getopt(argc, argv, "a:b:c:d:fj:l:pv:V")) != EOF) {
        switch (c) {
        case 'a':
            aFile = optarg;
            break;
        case 'b':
            weightBits = atoi(optarg);
            break;
        case 'c':
            className = optarg;
            break;
        case 'd':
            dFile = optarg;
            break;
        case 'f':
            fFlag = true;
            break;
        case 'j':
            gFile = optarg;
            gFileType = Jsgf;
            break;
        case 'l':
            gFile = optarg;
            gFileType = List;
            break;
        case 'p':
            pack = true;
            break;
        case 'v':
            vFile = optarg;
            break;
        case 'V':
            ++verbosity;
            break;
        default:
            error = true;
            break;
        }
    }
    argc -= optind;
    argv += optind;
    if (error) {
        usage();
        return 1;
    }

    if (argc == 5) {
        //
        // (1) compile static FST or dynamic main FST.

        gFile                = *argv++;
        const char* mcpFile  = *argv++;
        const char* lFile    = *argv++;
        const char* l2sFiles = *argv++;
        const char* outFile  = *argv++;

        FSTPocketCompiler compiler(mcpFile, lFile, l2sFiles, weightBits);
        FSTMemory* pfst = NULL;
        if (fFlag) {
            pfst = compiler.fstCompile(FSTLoad(gFile));
        } else {
            pfst = compiler.jsgfCompile(loadFile(gFile).c_str());
        }
        dumpFile(pfst, outFile);
        pfst->decref();

    } else if (className && argc == 4) {
        //
        // (2) compile class from JSGF or phrase list.
        
        if (className[0] != '$') {
            cerr << "Class name '" << className << "' must begin with $.\n\n";
            return 1;
        }
        
        const char* mcpFile  = *argv++;
        const char* lFile    = *argv++;
        const char* l2sFiles = *argv++;
        const char* outFile  = *argv++;

        string gString = loadFile(gFile);
        FSTPocketCompiler compiler(mcpFile, lFile, l2sFiles, weightBits);
        FSTMemory* pfst = NULL;
        switch (gFileType) {
        case Jsgf:
            pfst = compiler.jsgfCompileClass(className, gString.c_str());
            break;
        case List:
            pfst = compiler.listCompileClass(className, gString.c_str());
            break;
        case None:
            cerr << "One of -j file.jsgf or -l file.txt must be specified.\n\n";
            return 1;
        }
        dumpFile(pfst, outFile);
        pfst->decref();

    } else if (className && argc == 3) {
        //
        // (3) compile class from LG

        if (className[0] != '$') {
            cerr << "Class name '" << className << "' must begin with $.\n\n";
            return 1;
        }
        
        const char* mcpFile  = *argv++;
        const char* lgFile   = *argv++;
        const char* outFile  = *argv++;

        FSTGeneric lg = FSTLoad(lgFile);
        FSTPocketCompiler compiler(mcpFile, "", "", weightBits);
        FSTMemory* pfst = compiler.fstCompileClass(className, lg);
        dumpFile(pfst, outFile);
        pfst->decref();

    } else if (pack && argc == 2) {
        //
        // (4) pack to file.

        const char* inFile = *argv++;
        const char* outFile = *argv++;

        if (aFile) {
            FSTGeneric aFst = FSTLoad(aFile);
            const fst::set<int>& inputs = aFst->input_alphabet();
            alphabet.insert(inputs.begin(), inputs.end());
        }
        
        FSTGeneric fst = FSTLoad(inFile);
        if (verbosity >= 1) cerr << "constructing packed FST file " << outFile << "...\n";
        fst = FSTPackToFile(outFile, vFile, fst, weightBits, alphabet, false, verbosity, PFST_VERSION);
        if (dFile) {
            fst->dump_ascii(dFile);
        }

    } else {
        //
        // error

        usage();
        return 1;

    }

    return 0;
}

// Register as subcommand.
//
static FstTool pocket_build_tool("pocket_build", fst_pocket_build);

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 80
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
