////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

/******************************************************************************
=head1 NAME

fst_project - manipulate input and output labels on each FST arc

=head1 SYNOPSIS

B<fst_project> I<in> I<out> I<fst-in> I<fst-out>

=head1 DESCRIPTION

B<fst_project> projects input and output labels of I<fst-in> to produce
I<fst-out>.  I<in> selects the source of the input side of each output arc, and
I<out> the source of the output side of each output arc.  I<in> and I<out> are
each one of:

=over 4

=item B<i>

Select the input symbol.

=item B<o>

Select the output symbol.

=item B<e>

Use epsilon.

=back

For example,

    fst_project o i in.fst out.fst

will transpose input and output labels, whereas

    fst_project i e in.fst out.fst

will replace the output labels with epsilon.

    fst_project i i in.fst out.fst

and

    fst_project o o in.fst out.fst

compute FSAs representing the input and output side of the input FST,
respectively.

The specification of the input FST I<fst-in> can be a filename, an expression
(see fst_expression(1)), or "-" indicating standard input.  The specification of
the output FST I<fst-out> can be a filename or "-" indicating standard output.

=head1 OPTIONS

None.

=head1 SEE ALSO

fst_invert(1)

=head1 AUTHOR

I. Lee Hetherington <ilh@mit.edu>

=cut
******************************************************************************/

#include <iostream>
#include "fst.h"

using namespace std;

static void
usage()
{
    cerr <<
        "usage: fst project (i|o|e) (i|o|e) in.fst out.fst\n"
        "\n"
        "       Project input and output labels.\n"
        "\n";
}

static bool
decode(FSTArcOrder& result, const string& x)
{
    if (x == "i") {
        result = FSTInput;
        return true;
    } else if (x == "o") {
        result = FSTOutput;
        return true;
    } else if (x == "e") {
        result = FSTNone;
        return true;
    } else {
        result = FSTNone;
        return false;
    }
}

static int
fst_project(int argc, char** argv)
{
    argc--;
    argv++;
    if (argc != 4) {
        usage();
        return 1;
    }

    FSTArcOrder input_spec, output_spec;
    if (!decode(input_spec, argv[0]))
        usage();
    if (!decode(output_spec, argv[1]))
        usage();

    const char* in_file = argv[2];
    const char* out_file = argv[3];

    FSTGeneric in = FSTLoad(in_file);
    FSTGeneric out = FSTProject(in, input_spec, output_spec);
    
    out->dump_ascii(out_file);

    return 0;
}

// Register as subcommand.
//
static FstTool project_tool("project", fst_project);

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 79
// indent-tabs-mode: nil
// c-basic-offset: 4
// End:
