////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

/******************************************************************************
=head1 NAME

fst_pron_rules - apply pronunciation rules to FST or compile to FST

=head1 SYNOPSIS

B<fst_pron_rules> I<rules-in> I<fst-in> I<fst-out>

B<fst_pron_rules> B<-d> I<rules-in> I<A-out> I<B-out> I<C-out>

B<fst_pron_rules> B<-s> I<rules-in> I<P-out>

=head1 DESCRIPTION

B<incomplete>

=head1 OPTIONS

B<incomplete>

=head1 SEE ALSO

fst_build_cplg(1)

=head1 AUTHOR

I. Lee Hetherington <ilh@mit.edu>

=cut
******************************************************************************/

#include <iostream>
#include "fst.h"
#include "fst/FSTPronRules.h"

using namespace std;

static void
usage()
{
    cerr <<
        "usage: fst pron_rules pron.rules in.fst out.fst         (apply mode)\n"
        "       fst pron_rules -d pron.rules A.fst B.fst C.fst   (dump multiple FST mode)\n"
        "       fst pron_fules -s pron.rules P.fst               (dump single FST mode)\n"
        "\n";
}

static int
fst_pron_rules(int argc, char** argv)
{
    if (argc < 0) {
        usage();
        return 1;
    }

    bool error = false;
    bool dump_flag = false;
    bool single_flag = false;

    char c;
    while ((c = getopt(argc, argv, "ds")) != EOF) {
        switch (c) {
        case 'd':
            dump_flag = true;
            break;
        case 's':
            single_flag = true;
            break;
        default:
            error = true;
            break;
        }
    }
    argc -= optind;
    argv += optind;
    if (error
        || (!dump_flag && !single_flag && argc != 3)
        || (dump_flag && argc != 4)
        || (single_flag && argc != 2)
        || dump_flag && single_flag) {

        usage();
        return 1;
    }

    const char* rules_file  = *argv++;

    FSTPronRules rules(rules_file);

    if (rules.bad())
        return 1;

    if (dump_flag) {
        const char* a_file = *argv++;
        const char* b_file = *argv++;
        const char* c_file = *argv++;
        rules.dump(a_file, b_file, c_file);
    } else if (single_flag) {
        const char* p_file = *argv++;
        FSTOptimizeA(FSTTrim(rules.single_fst()))->dump_binary(p_file);
    } else {
        const char* in_file  = *argv++;
        const char* out_file = *argv++;
        rules.apply(FSTLoad(in_file))->dump_binary(out_file);
    }

    return 0;
}

// Register as subcommand.
//
static FstTool pron_rules_tool("pron_rules", fst_pron_rules);

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 80
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
