////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

/******************************************************************************
=head1 NAME

fst_push_weights - push weights toward initial state

=head1 SYNOPSIS

B<fst_push_weights> [I<OPTIONS>] I<fst-in> I<fst-out>

=head1 DESCRIPTION

B<fst_push_weights> pushes final and arc weights as far as possible toward the
initial state.  Such weight pushing generally increases the number of states
that can be merged by fst_minimize(1).  Such weight pushing is also performed by
fst_optimize(1).

If the initial state is part of a cycle, the "residual weight" of the initial
state will be emitted when leaving the cycle.  That is, the weight will be
incorporated into final weights of states within the cycle and into arc weights
for all arcs exiting the initial cycle.

The specification of the input FST I<fst-in> can be a filename, an expression
(see fst_expression(1)), or "-" indicating standard input.  The specification of
the output FST I<fst-out> can be a filename or "-" indicating standard output.

=head1 OPTIONS

=over 12

=item B<-B>

Output I<fst-out> in binary format.

=back

=head1 CAVEATS

Problems with various semirings:

=over 12

=item (min,+)

Cycles with net negative weights not supported.

=item (max,+)

Cycles with net positive weights not supported.

=item (+,*)

Cycles with net weight > 1 not supported.

=back

=head1 SEE ALSO

fst_minimize(1), fst_optimize(1)

=head1 AUTHOR

I. Lee Hetherington <ilh@mit.edu>

=cut
******************************************************************************/

#include <iostream>
#include "fst.h"

using namespace std;

static void
usage()
{
    cerr <<
        "usage: fst push_weights [options] in.fst out.fst\n"
        "\n"
        "-B     output binary file\n"
        "\n";
}

static int
fst_push_weights(int argc, char** argv)
{
    if (argc < 0) {
        usage();
        return 1;
    }

    bool binary_flag = false;
    bool error = false;

    char c;
    while ((c = getopt(argc, argv, "B")) != EOF) {
        switch (c) {
        case 'B':
            binary_flag = true;
            break;
        default:
            error = true;
            break;
        }
    }

    argc -= optind;
    argv += optind;

    if (error || argc != 2) {
    usage();
        return 1;
    }

    const char* in_file  = argv[0];
    const char* out_file = argv[1];

    FSTGeneric x = FSTLoad(in_file);

    x = FSTPushWeights(x);
    if (binary_flag)
        x->dump_binary(out_file);
    else
        x->dump_ascii(out_file);

    return 0;
}

// Register as subcommand.
//
static FstTool push_weights_tool("push_weights", fst_push_weights);

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 80
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
