////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

// Maximally reduce sentences with an RTN.  If a given sentence has multiple
// unique minimal reductions, output all of them.  The RTN should most likely be
// created with fst_from_jsgf -p.
//
// Lee Hetherington
// March 2001

#include <iostream>
#include <fstream>
#include "fst.h"

#ifdef _MSC_VER
#  pragma warning(disable: 4065)
#endif

using namespace std;

// Convenient typedefs to STL types.
//
typedef fst::vector<int>  List;
typedef fst::map<int,int> Map;
typedef fst::set<int>     IntSet;

// Structure to hold a path.  Items contains the list of items (words), and
// n_terminals contains the count of terminals in the path.  A terminal is
// anything not contained in tag_map, namely not <foo>, </foo>, <@bar>, or
// </@bar>.
//
struct Path {
    List items;
    int n_terminals;

    Path() : n_terminals(0) {}
    Path(const Path& p) : items(p.items), n_terminals(p.n_terminals) {}
    bool operator>(const Path& p) const {
        return
            items.size() > p.items.size()  ||
            (items.size() == p.items.size() && n_terminals > p.n_terminals);
    }
    void push_back(int item, const Map& tag_map) {
        items.push_back(item);
        n_terminals += (tag_map.count(item) == 0);
    }
    void pop_back(const Map& tag_map) {
        int item = items.back();
        n_terminals -= (tag_map.count(item) == 0);
        items.pop_back();
    }
};

// For ordering paths within a PathSet.
//
struct PathCompare {
    bool operator()(const Path& a, const Path& b) const {
        return a.items < b.items;
    }
};

typedef fst::set<Path, PathCompare>    PathSet;

static void
usage()
{
    cerr <<
        "usage: fst reduce_text parser.fst in.txt\n"
        "\n";
}

// Convert a text string to a linear FST.
//
static FSTGeneric
text_to_fst(const string& text)
{
    FSTGeneric fst = FSTBasic();
    int p = fst->add_state();
    fst->set_initial(p);

    const char* space = " \t";

    string::size_type i, j;
    for (i = text.find_first_not_of(space);
         i != string::npos;
         i = text.find_first_not_of(space, j)) {
        j = text.find_first_of(space, i);
        int label = FSTstring_map[text.substr(i, j - i)];
        fst->add_state();
        fst->add_arc(p, p + 1, label, label);
        ++p;
    }
    
    fst->set_final(p);

    return fst;
}

// Given a text string, produce an FST containing all valid parses.
//
static FSTGeneric
all_parses(const string& text, const FSTGeneric& parser)
{
    return FSTTrim(FSTCompose(text_to_fst(text), FSTCache(parser)));
}

// Lookup item in the tag map.  If it doesn't exist, return -1. Valid entries
// are all >= 0.
//
static inline int
find(const Map& tag_map, int item)
{
    Map::const_iterator i = tag_map.find(item);
    return (i == tag_map.end()) ? -1 : (*i).second;
}

// Return whether path is worse than those in the best set, if any.  All paths
// in the set of best paths are equally good, so we only need to check the first
// one when making a comparison.
//
static bool
is_worse(const Path& path, const PathSet& best_paths)
{
    return !best_paths.empty() && path > *best_paths.begin();
}

// Store path into the best set.  We make it the only item in the set if it is
// beter than the set's contents.  We already know that path is not worse than
// those in the set of best paths.
//
static void
store(const Path& path, PathSet& best_paths)
{
    if (!best_paths.empty() && *best_paths.begin() > path)
        best_paths.clear();
    best_paths.insert(path);
}

// Recursive depth-first exploration of the FST containing all the parses.  It
// builds up a reduced path in path, detects cycles using the active set, and
// skips over balanced tags using the tag stack.  All unique minimal-length
// reduced paths go into the best set.
//
static void
recurse(const FSTGeneric& fst, int state, Path& path,
        IntSet& active, List& tag_stack, const Map& tag_map,
        PathSet& best_paths)
{
    // Prune if cycle detected.  Is this even possible?
    //
    if (active.count(state))
        return;

    // Prune if path already worse than those in the best set.
    //
    if (is_worse(path, best_paths))
        return;

    // Have a complete path and it is at least as good as those in the best set.
    //
    if (fst->is_final(state) && tag_stack.empty())
        store(path, best_paths);

    // Continue recursion.
    //
    active.insert(state);
    for (FSTArcIter a = fst->arcs_out(state); a.valid(); ++a) {
        bool pushed_path = false;
        bool pushed_stack = false;
        int  popped_item = 0;
        int item = a->output;

        if (!tag_stack.empty() && item == tag_stack.back()) {

            // We matched the closing tag on the top of the stack.  We'll just
            // eat this symbol and pop the stack.
            //
            tag_stack.pop_back();
            popped_item = item;

        } else {

            int mapped_item = find(tag_map, item);
            if (mapped_item != 0 && tag_stack.empty()) {

                // <foo> or ... and stack empty: push onto path.
                //
                path.push_back(item, tag_map);
                pushed_path = true;
            }
            if (mapped_item > 0) {

                // <foo>: push </foo> onto stack
                //
                tag_stack.push_back(mapped_item);
                pushed_stack = true;
            }
        }

        recurse(fst, a->next, path, active, tag_stack, tag_map, best_paths);

        // Clean up.
        //
        if (pushed_path)
            path.pop_back(tag_map);
        if (pushed_stack)
            tag_stack.pop_back();
        if (popped_item)
            tag_stack.push_back(popped_item);
    }    
    active.erase(state);
}

// Construct the tag map that is used to handle balanced parse tags.  The map is
// constructed as follows:
//
// <foo>   => </foo>
// <@bar>  => 0
// </@bar> => 0
//
// The reason for this is when we see <foo> we will skip input until we see a
// balanced </foo>, so we need to know what to look for.  If we see <@bar> or
// </@bar>, we will ignore them.
//
static void
construct_tag_map(const IntSet& alphabet, Map& tag_map)
{
    string start_label;
    for (IntSet::const_iterator i = alphabet.begin();
         i != alphabet.end(); ++i) {
        int end_index = *i;
        const char* end_label = FSTstring_map[end_index];
        if (end_label[0] == '<' && end_label[1] == '/') {
            start_label = '<';
            start_label += end_label + 2;
            int start_index = FSTstring_map[start_label];
            if (end_label[2] == '@')
                tag_map[end_index] = tag_map[start_index] = 0;
            else
                tag_map[start_index] = end_index;
        }
    }
}

// Compute and print out the best unique parses.
//
static void
print_best_parse(const string& text, const FSTGeneric& parser)
{
    FSTGeneric parses = all_parses(text, parser);

    if (parses->initial() < 0) {
        cerr << "no parse for: " << text << endl;
        return;
    }

    PathSet best_paths;
    IntSet active;
    Path path;
    List tag_stack;
    Map tag_map;

    construct_tag_map(parses->output_alphabet(), tag_map);

    recurse(parses, parses->initial(), path, active,
            tag_stack, tag_map, best_paths);

    assert(!best_paths.empty());

    PathSet::const_iterator i;
    List::const_iterator j;

    // Print out paths.
    //
    for (i = best_paths.begin(); i != best_paths.end(); ++i) {
        const Path& p = *i;
        if (!p.items.empty()) {
            j = p.items.begin();
            cout << FSTstring_map[*j];
            while (++j != p.items.end())
                cout << ' ' << FSTstring_map[*j];
        }
        cout << endl;
    }
}

static int
fst_reduce_text(int argc, char** argv)
{
    if (argc < 1) {
        usage();
        return 1;
    }

    bool error = false;

    char c;
    while ((c = getopt(argc, argv, "")) != EOF) {
        switch (c) {
        default:
            error = true;
            break;
        }
    }
    argc -= optind;
    argv += optind;

    if (error || argc != 2) {
        usage();
        return 1;
    }

    const char* parser_file = *argv++;
    const char* input_file = *argv++;

    FSTGeneric parser = FSTLoad(parser_file);

    ifstream is(input_file);
    if (!is) {
        cerr << "cannot read file " << input_file << endl;
        exit(1);
    }

    string line;
    while (getline(is, line), !is.fail())
        print_best_parse(line, parser);

    return 0;
}

// Register as subcommand.
//
static FstTool reduce_text_tool("reduce_text", fst_reduce_text);

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 80
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
