#include <iostream>
#include <limits>
#include "fst/FST.h"

// general implementation

template< class DAG, bool integral_state >
class _SCC_aux {
public:
    typedef typename DAG::State State;
    typedef fst::set<State> StateSet;
    typedef fst::list<StateSet> StateSetList;
    static void compute(const FSTGeneric& fst, StateSetList& components) {
        DAG dag(fst);
        components.clear();

        // DFS 1
        _StateListMap adjacencyMap;
        _StateList order;
        _dfs1(dag, adjacencyMap, order);

        // DFS 2
        StateSet visited;
        for (_StateListIterator i = order.begin(); i != order.end(); ++i) {
            const State& p = *i;
            if (visited.insert(p).second) {
                StateSet component;
                _dfs2(p, adjacencyMap, visited, component);
                components.push_back(component);
            }
        }
    }
private:
    typedef typename DAG::TransitionIterator _TransitionIterator;
    typedef typename StateSet::iterator _StateSetIterator;
    typedef fst::list<State> _StateList;
    typedef typename _StateList::iterator _StateListIterator;
    typedef fst::map<State,_StateList> _StateListMap;
    struct _StackEntry1 {
        State state;
        _TransitionIterator transitionIterator;
        _StackEntry1(const State& s, const _TransitionIterator& ti)
            : state(s), transitionIterator(ti) {
        }
    };
    static void _dfs1(const DAG& dag, _StateListMap& adjacencyMap, _StateList& order) {
        typedef fst::list<_StackEntry1> Stack;
        State p(dag.start());
        StateSet visited;
        Stack entryStack;
        visited.insert(p);
        entryStack.push_back(_StackEntry1(p, dag.transitions(p)));
        while (entryStack.size() > 0) {
            _StackEntry1& top = entryStack.back();
            if (top.transitionIterator.hasNext()) {
                State q(top.transitionIterator.next());
                adjacencyMap[q].push_back(top.state);
                if (visited.insert(q).second) {
                    entryStack.push_back(_StackEntry1(q, dag.transitions(q)));
                }
            } else {
                order.push_front(top.state);
                entryStack.pop_back();
            }
        }
    }
    struct _StackEntry2 {
        State state;
        _StateListIterator iter, end;
        _StackEntry2(const State& s, _StateList& l)
            : state(s), iter(l.begin()), end(l.end()) {
        }
    };
    static void _dfs2(const State& p, _StateListMap& adjacencyMap, StateSet& visited, StateSet& component)
    {
        typedef fst::list<_StackEntry2> Stack;
        Stack entryStack;
        component.insert(p);
        entryStack.push_back(_StackEntry2(p, adjacencyMap[p]));
        while (entryStack.size() > 0) {
            _StackEntry2& top = entryStack.back();
            if (top.iter != top.end) {
                State q(*top.iter++);
                if (visited.insert(q).second) {
                    component.insert(q);
                    entryStack.push_back(_StackEntry2(q, adjacencyMap[q]));
                }
            } else {
                entryStack.pop_back();
            }
        }
    }
};

// integral-state implementation (use vector instead of set and map)

template< class DAG >
class _SCC_aux<DAG, true> {
public:
    typedef typename DAG::State State;
    typedef fst::vector<State> StateSet;
    typedef fst::list<StateSet> StateSetList;
    static void compute(const FSTGeneric& fst, StateSetList& components) {
        DAG dag(fst);
        components.clear();

        // DFS 1
        _StateListVector adjacencyMap;
        _StateList order;
        order.reserve(dag.nStates());
        _dfs1(dag, adjacencyMap, order);

        // DFS 2
        _VectorBool visited;
        for (_StateListIterator i = order.begin(); i != order.end(); ++i) {
            const State& p = *i;
            if (int(visited.size()) < p + 1) {
                visited.resize(p + 1, false);
            }
            if (!visited[p]) {
                visited[p] = true;
                StateSet component;
                _dfs2(p, adjacencyMap, visited, component);
                components.push_back(component);
            }
        }
    }
private:
    typedef typename DAG::TransitionIterator _TransitionIterator;
    typedef fst::vector<State> _StateList;
    typedef typename _StateList::iterator _StateListIterator;
    typedef fst::vector<_StateList> _StateListVector;
    typedef fst::vector<bool> _VectorBool;
    struct _StackEntry1 {
        State state;
        _TransitionIterator transitionIterator;
        _StackEntry1(const State& s, const _TransitionIterator& ti)
            : state(s), transitionIterator(ti) {
        }
    };
    static void _dfs1(const DAG& dag, _StateListVector& adjacencyMap, _StateList& order) {
        typedef fst::list<_StackEntry1> Stack;
        State p(dag.start());
        _VectorBool visited(dag.nStates(), false);
        Stack entryStack;
        if (int(visited.size()) < p + 1) {
            visited.resize(p + 1, false);
        }
        visited[p] = true;
        entryStack.push_back(_StackEntry1(p, dag.transitions(p)));
        while (entryStack.size() > 0) {
            _StackEntry1& top = entryStack.back();
            if (top.transitionIterator.hasNext()) {
                State q(top.transitionIterator.next());
#if 1
                if (int(adjacencyMap.size()) < q + 1) {
                    adjacencyMap.resize(q + 1);
                }
                adjacencyMap[q].push_back(top.state);
#endif
                if (int(visited.size()) < q + 1) {
                    visited.resize(q + 1, false);
                }
                if (!visited[q]) {
                    visited[q] = true;
                    entryStack.push_back(_StackEntry1(q, dag.transitions(q)));
                }
            } else {
                order.push_back(top.state);
                entryStack.pop_back();
            }
        }
        reverse(order.begin(), order.end());
    }
    struct _StackEntry2 {
        State state;
        _StateListIterator iter, end;
        _StackEntry2(const State& s, _StateList& l)
            : state(s), iter(l.begin()), end(l.end()) {
        }
    };
    static void _dfs2(const State& p, _StateListVector& adjacencyMap, _VectorBool& visited, StateSet& component)
    {
        return;
        typedef fst::list<_StackEntry2> Stack;
        Stack entryStack;
        component.push_back(p);
        entryStack.push_back(_StackEntry2(p, adjacencyMap[p]));
        while (entryStack.size() > 0) {
            _StackEntry2& top = entryStack.back();
            if (top.iter != top.end) {
                State q(*top.iter++);
                if (int(visited.size()) < q + 1) {
                    visited.resize(q + 1, false);
                }
                if (!visited[q]) {
                    visited[q] = true;
                    component.push_back(q);
                    entryStack.push_back(_StackEntry2(q, adjacencyMap[q]));
                }
            } else {
                entryStack.pop_back();
            }
        }
    }
};

// dispatch the implementation based on whether or not DAG::State is integral

template< class DAG >
class SCC : public _SCC_aux<DAG, std::numeric_limits<typename DAG::State>::is_integer> {
};

using namespace std;

class SimpleDAG {
public:
    typedef int State;
    class TransitionIterator {
    public:
        TransitionIterator(const FSTArcIter arcs) : _arcs(arcs) {
        }
        bool hasNext() const {
            return _arcs.valid();
        }
        State next() {
            State state(_arcs->next);
            ++_arcs;
            return state;
        }
    private:
        FSTArcIter _arcs;
    };
    SimpleDAG(const FSTGeneric& fst) : _fst(fst) {
    }
    int nStates() const {
        return _fst->n_states();
    }
    State start() const {
        return State(_fst->initial());
    }
    TransitionIterator transitions(State state) const {
        return TransitionIterator(_fst->arcs_out(state));
    }
private:
    FSTGeneric _fst;
};

class EpsilonDAG {
public:
    struct State {
        int state, output;
        State(int s, int o) : state(s), output(o) {}
        bool operator<(const State& x) const {
            if (state < x.state) return true;
            if (state > x.state) return false;
            return output < x.output;
        }
    };
    class TransitionIterator {
    public:
        TransitionIterator(int output, const FSTArcIter arcs) : _output(output), _arcs(arcs) {
            while (_arcs.valid() && _output && _arcs->output)
                ++_arcs;
        }
        bool hasNext() const {
            return _arcs.valid();
        }
        State next() {
            State state(_arcs->next, _output ? _output : _arcs->output);
            ++_arcs;
            while (_arcs.valid() && _output && _arcs->output)
                ++_arcs;
            return state;
        }
    private:
        int _output;
        FSTArcIter _arcs;
    };
    EpsilonDAG(const FSTGeneric& fst) : _fst(fst) {}
    State start() const {
        return State(_fst->initial(), 0);
    }
    int nStates() const {
        return 1;
    }
    TransitionIterator transitions(const State& state) const {
        return TransitionIterator(state.output, _fst->arcs_out_with_input(state.state, 0));
    }
private:
    FSTGeneric _fst;
};

ostream& operator<<(ostream& os, const EpsilonDAG::State& s)
{
    return os << '(' << s.state << ',' << FSTstring_map[s.output] << ')';
}

using namespace std;

int main(int argc, char** argv)
{
    FSTGeneric x = FSTLoad(argv[1]);

    typedef SimpleDAG DAG;

    typedef SCC<DAG>::StateSet Component;
    typedef SCC<DAG>::StateSetList ComponentList;
    //typedef fst::vector<DAG::State> Component;
    //typedef fst::list<Component> ComponentList;

    //typedef fst::set<int> Component;            // same as SCC<DAG>::Component
    //typedef fst::list<Component> ComponentList; // same as SCC<DAG>::ComponentList

    ComponentList components;
    SCC<DAG>::compute(x, components);
    for (ComponentList::iterator i = components.begin(); i != components.end(); ++i) {
        Component& component = *i;
        cout << "component: (" << component.size() << "):";
        for (Component::iterator j = component.begin(); j != component.end(); ++j) {
            cout << ' ' << *j;
        }
        cout << '\n';
    }

    return 0;
}

