////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

/******************************************************************************
=head1 NAME

fst_search - perform a Viterbi-style search on two FSTs

=head1 SYNOPSIS

B<fst_search> [I<OPTIONS>] I<A-fst-in> I<B-fst-in> I<fst-out>

B<fst_search> [B<-t>] I<fst-in> I<fst-out>

=head1 DESCRIPTION

B<fst_search> performs a Viterbi-style dynamic programming search through the
composition of two FSTs I<A-fst-in> and I<B-fst-in>.  By default it produces an
FST representing the best path through the composition.  Optionally, it can
produce an FST representing the Viterbi trellis subject to beam pruning.

It is required that the states of I<A-fst-in> be topologically sorted.  A
warning message will be issued if this is not the case.  See
fst_topological_sort(1).

In general,

    fst_search A.fst B.fst C.fst

computes the same result as

    fst_compose A.fst B.fst - | fst_nbest -p - C.fst

provided that A.fst is topologically sorted.

The specification of each of the input FSTs I<A-fst-in>, I<B-fst-in>, and
I<fst-in> can be a filename, an expression (see fst_expression(1)), or "-"
indicating standard input.  The specification of the output FST I<fst-out> can
be a filename or "-" indicating standard output.

=head1 OPTIONS

=over 12

=item B<-g>

Produce a graph representing the Viterbi trellis rather than the single best
path.

=item B<-n> I<N>

The count-based beam width will be I<N>.  That is, at most I<N> active nodes
will be kept at each state of I<A-fst-in> during the search.

=item B<-s> I<S>

The score-based beam width will be I<S>.

=item B<-t>

Output textual representation rather than an FST.  This only applies if B<-g>
is not specified.

=item B<-v>

Be verbose during the search, displaying the number of active nodes at each
state of I<A-fst-in>.

=back

=head1 AUTHOR

I. Lee Hetherington <ilh@mit.edu>

=cut
******************************************************************************/

#include <iostream>

#include <fstream>
#include "fst.h"
#include "fst/FSTTimer.h"
#include "fst/format.h"

using namespace std;

static void
usage()
{
    cerr <<
        "usage: fst search [-n <nodes>] [-s <score>] [-g] [-v] a.fst b.fst out.fst\n"
        "       fst search [-t] in.fst out.fst\n"
        "\n";
}

static int
fst_search(int argc, char** argv)
{
    if (argc < 0) {
        usage();
        return 1;
    }

    bool error = false;
    int nodes = 0;
    bool graph = false;
    float score = 1e38F;
    bool text_flag = false;
    bool verbose = false;

    char c;
    while ((c = getopt(argc, argv, "n:s:gtv")) != EOF) {
        switch (c) {
        case 'n':
            nodes = atoi(optarg);
            break;
        case 's':
            score = atof(optarg);
            break;
        case 'g':
            graph = true;
            break;
        case 't':
            text_flag = true;
            break;
        case 'v':
            verbose = true;
            break;
        default:
            error = true;
            break;
        }
    }

    argc -= optind;
    argv += optind;

    assert((graph && !text_flag) || !graph);
    if (error ||
        argc < 2 ||
        (argc == 2 && (nodes != 0 || score != float(1e38) || graph)) ||
        argc > 3) {

        usage();
        return 1;
    }

    const char* out_file;
    FSTGeneric result;
    fst::vector<int> frontier_nodes;

    if (argc == 2) {

        out_file = argv[1];
        result = FSTSearch(FSTLoad(argv[0]));

    } else {

        out_file = argv[2];

        FSTGeneric b = FSTLoad(argv[1]);
        FSTGeneric a = FSTLoad(argv[0]);
        
        FSTTimer t;
        
        if (graph)
            result = FSTSearchGraph(a, b, score, nodes, verbose);
        else
            result = FSTSearch(a, b, score, nodes, verbose);
        
        format(cerr, "search time: %.3f\n", t.value());

    }
        
    if (text_flag) {
        
        ofstream os_file;
        ostream* osp = &cout;
        
        if (out_file[0] != '-' || out_file[1] != 0) {
            os_file.open(out_file);
            osp = &os_file;
        }
        
        FSTSemiring semiring = result->semiring();
        float weight = semiring->one;
        string words;
        for (FSTArcIter a = result->arcs_out(result->initial());
             a.valid();
             a = result->arcs_out(a->next)) {
            assert(a.n() == 1);
            weight = semiring->times(weight, a->weight);
            if (result->is_final(a->next))
                weight = semiring->times(weight, result->final_weight(a->next));
            if (a->output) {
                words += ' ';
                words += FSTstring_map[a->output];
            }
        }
        
        format(*osp, "%9.4f", weight) << words << endl;
        for(int i = 0; i < int(frontier_nodes.size()); i++)
            format(*osp, "%d ", frontier_nodes[i]);
        *osp << endl;

    } else {
        result->dump_ascii(out_file);
    }

    return 0;
}

// Register as subcommand.
//
static FstTool search_tool("search", fst_search);

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 80
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
