////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2007, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

/******************************************************************************
=head1 NAME

fst_semring - convert FST semiring

=head1 SYNOPSIS

B<fst_semiring> I<semiring> I<fst-in> I<fst-out>

=head1 DESCRIPTION

B<fst_semiring> convert FST I<fst-in> to specified I<semiring>.

The specification of the input FST I<fst-in> can be a filename, an expression
(see fst_expression(1)), or "-" indicating standard input.  The specification of
the output FST I<fst-out> can be a filename or "-" indicating standard output.

I<semiring> should be one of "minplus" (tropical semiring), "maxplus", or
"plustimes".

=head1 OPTIONS

None.

=head1 AUTHOR

I. Lee Hetherington <ilh@mit.edu>

=cut
******************************************************************************/

#include <iostream>
#include "fst.h"

using namespace std;

static void
usage()
{
    cerr <<
        "usage: fst semiring (minplus | maxplus | plustimes) in.fst out.fst\n"
        "\n";
}

static int
fst_semiring(int argc, char** argv)
{
    if (argc < 0) {
        usage();
        return 1;
    }

    argc--;
    argv++;

    if (argc != 3) {
        usage();
        return 1;
    }

    string semiring(*argv++);
    const char* in_file  = *argv++;
    const char* out_file = *argv++;

    FSTGeneric in = FSTLoad(in_file);
    FSTGeneric out;

    if (semiring == "minplus") {
        out = FSTConvert_to_MinPlus(in);
    } else if (semiring == "maxplus") {
        out = FSTConvert_to_MaxPlus(in);
    } else if (semiring == "plustimes") {
        out = FSTConvert_to_PlusTimes(in);
    } else {
        usage();
        return 1;
    }

    out->dump_ascii(out_file);

    return 0;
}

// Register as subcommand.
//
static FstTool semiring_tool("semiring", fst_semiring);

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 79
// indent-tabs-mode: nil
// c-basic-offset: 4
// End:
