////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

/******************************************************************************
=head1 NAME

fst_to_dot - produce DOT representation of FST for graphical output

=head1 SYNOPSIS

B<fst_to_dot> [I<OPTIONS>] I<fst-in> I<dot-out>

=head1 DESCRIPTION

B<fst_to_dot> produces output I<fst-out> that can be used by dot(1) to produce
graphical displays of I<fst-in>.

Arc labels are of the form I<input>:I<output>/I<weight>.  "@" is used for
epsilon I<input> and I<output>.

The specification of the input FST I<fst-in> can be a filename, an expression
(see fst_expression(1)), or "-" indicating standard input.  The specification of
the output FST I<dot-out> can be a filename or "-" indicating standard output.

=head1 OPTIONS

=over 12

=item B<-p>

Produce portrait mode.  The default is landscape mode.

=back

=head1 SEE ALSO

dot(1): part of Graphviz (http://www.graphviz.org/)

=head1 AUTHOR

I. Lee Hetherington <ilh@mit.edu>

=cut
******************************************************************************/

#include <iostream>
#include "fst.h"

using namespace std;

static void
usage()
{
    cerr <<
        "usage: fst to_dot [-p] in.fst out.dot\n"
        "\n"
        "-p        portrait mode\n";
}

static int
fst_to_dot(int argc, char **argv)
{
    if (argc < 0) {
        usage();
        return 1;
    }

    bool error = false;
    bool portrait = false;
    char c;

    while ((c = getopt(argc, argv, "p")) != EOF) {
        switch (c) {
        case 'p':
            portrait = true;
            break;
        default:
            error = true;
            break;
        }
    }
    argc -= optind;
    argv += optind;
    if (error || argc != 2) {
        usage();
        return 1;
    }

    const char* in_file  = argv[0];
    const char* out_file = argv[1];

    const char dot_init[] =
    "rankdir=\"LR\"; center=true; size=\"10,7.5\"; "
    "rotate=90; edge [fontsize=20]; node [shape=circle]";

    const char dot_portrait_init[] =
    "rankdir=\"LR\"; center=true; size=\"10,7.5\"; "
    "edge [fontsize=20]; node [shape=circle]";

    FSTGeneric x = FSTLoad(in_file);

    x->dump_dot(out_file, portrait ? dot_portrait_init : dot_init);

    return 0;
}

// Register as subcommand.
//
static FstTool to_dot_tool("to_dot", fst_to_dot);

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 80
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:



