////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

/******************************************************************************
=head1 NAME

fst_to_fsm - convert FST file to AT&T FSM file

=head1 SYNOPSIS

B<fst_to_fsm> [I<OPTIONS>] I<fst-in> I<fsm-out> I<fsm-symbols-out>

=head1 DESCRIPTION

B<fst_to_fsm> convert from FST file I<fst-in> to AT&T FSM files I<fsm-out> and
I<fsm-symbols-out>.  The I<fsm-out> file contains the transition and state
information with labels represented by indices.  The index to label mapping is
contained in the I<fsm-symbols-out> file.

The specification of the input FST I<fst-in> can be a filename, an expression
(see fst_expression(1)), or "-" indicating standard input.

=head1 OPTIONS

=over 12

=item B<-l> I<labels-in>

File I<labels-in> contains one symbol per line, in the order desired for the
FSM file.  Epsilon will always be assigned index 0 and should not appear in
I<labels-in>.

=back

=head1 EXAMPLE

Here is an example converting a simple FST file to an FSM file:

    % cat sample.fst
    #FSTBasic PlusTimes
    I 0
    T 0 1 a B 1
    T 0 1 a B 2
    T 0 1 b B 1
    T 0 2 b B 2
    F 1
    F 2

    % fst_to_fsm sample.fst sample.fsm sample.sym

    % cat sample.fsm
    0 1 1 2 1
    0 1 1 2 2
    0 1 3 2 1
    0 2 3 2 2
    1 1
    2 1

    % cat sample.sym
    , 0
    B 2
    a 1
    b 3

Now, if we want to control the mapping between indices and labels, we can
specify a labels file:

    % cat sample.labels
    a
    b
    B

    % fst_to_fsm -l sample.labels sample.fst sample.fsm sample.sym

    % cat sample.fsm
    0 1 1 3 1
    0 1 1 3 2
    0 1 2 3 1
    0 2 2 3 2
    1 1
    2 1

    % cat sample.sym
    , 0
    a 1
    b 2
    B 3

=head1 AUTHOR

I. Lee Hetherington <ilh@mit.edu>

=cut
******************************************************************************/
#include <cstring>
#include <cstdio>
#include "fst.h"
#include "fst/map"

using namespace std;

typedef fst::map<int, int> IntIntMap;

static int sym = 0;
static IntIntMap label_map;

static void
usage()
{
    fprintf(stderr, "usage: fst to_fsm [-l labels.txt] in.fst out.fsm out.sym\n\n");
}

static int
map_label(int x)
{
    if (label_map.count(x))
        return label_map[x];
    else
        return label_map[x] = sym++;
}

static void
load_labels(const char* labels_file)
{
    FILE* fp = fopen(labels_file, "r");
    assert(fp != 0);
 
    char buf[1024];

    map_label(0);
 
    while (fgets(buf, sizeof(buf), fp)) {
        size_t n = strcspn(buf, " \n\r\t");
        buf[n] = 0;
        if (strcmp(buf, ",") != 0)
            map_label(FSTstring_map[buf]);
    }
 
    fclose(fp);
}

static void
dump_transition(FILE* fp, const FSTArcIter& ai, int s)
{
    int i = map_label(ai->input);
    int o = map_label(ai->output);
    fprintf(fp, "%d %d %d %d %g\n", s, ai->next, i, o, ai->weight);
}

static void
dump_fsm(FSTGeneric fst, const char* fsm_file, const char* sym_file)
{
    FILE* fsm_fp = fopen(fsm_file, "w");
    assert(fsm_fp != 0);

    FILE* sym_fp = fopen(sym_file, "w");
    assert(sym_fp != 0);

    map_label(0);

    int initial = fst->initial();
    assert(initial >= 0);

    // Dump out arcs from initial state.  The initial state must be the
    // from state for the first transition.
    for (FSTArcIter ai = fst->arcs_out(initial); ai.valid(); ai++)
        dump_transition(fsm_fp, ai, initial);

    // Dump out rest of transitions.
    //
    for (FSTStateIter si = fst->states(); si.valid(); si++) {
        if (*si != initial)
            for (FSTArcIter ai = fst->arcs_out(*si); ai.valid(); ai++)
                dump_transition(fsm_fp, ai, *si);
        if (fst->is_final(*si))
            fprintf(fsm_fp, "%d %g\n", *si, fst->final_weight(*si));
    }

    // Output symbol file, the mappings for which are contained in label_map.
    //
    fprintf(sym_fp, ", 0\n");
    for (IntIntMap::iterator mi = label_map.begin();
         mi != label_map.end(); mi++) {
        int fst_index = (*mi).first;
        int fsm_index = (*mi).second;
        if (fst_index != 0)
            fprintf(sym_fp, "%s %d\n", FSTstring_map[fst_index], fsm_index);
    }

    fclose(fsm_fp);
    fclose(sym_fp);
}

static int
fst_to_fsm(int argc, char **argv)
{
    if (argc < 0) {
        usage();
        return 1;
    }

    const char* labels_file = 0;
    bool error = false;

    char c;
    while ((c = getopt(argc, argv, "l:")) != EOF) {
        switch (c) {
        case 'l':
            labels_file = optarg;
            break;
        default:
            error = true;
            break;
        }
    }
    argc -= optind;
    argv += optind;
    if (error || argc < 3) {
        usage();
        return 1;
    }

    if (labels_file)
        load_labels(labels_file);

    const char* fst_file = argv[0];
    const char* fsm_file = argv[1];
    const char* sym_file = argv[2];

    FSTGeneric fst = FSTLoad(fst_file);

    if (labels_file)
        load_labels(labels_file);

    dump_fsm(fst, fsm_file, sym_file);

    return 0;
}

// Register as subcommand.
//
static FstTool to_fsm_tool("to_fsm", fst_to_fsm);

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 80
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
