////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

#include <cstring>
#include <iostream>
#include <fstream>
#include <sstream>
#include "fst.h"

using namespace std;

static void
usage()
{
    cerr <<
        "usage: fst train [options] left.data right.data Ain.fst Bin.fst Cin.fst Bout.fst\n"
        "\n"
        "-f n   sets the floor for backoff\n"
        "-l     denotes left.data and right.data are lists of matched .phn and .wrd files,\n"
        "       otherwise they are matched input and output sequences, line by line\n"
        "-r     perform epsilon removal on B first before training\n"
        "-s d   saves the z*.fst's in directory d, and skips training\n"
        "-i d   loads the z*.fst's from directory d\n"
        "\n"
        "Perform EM training of the weights in B by generating sequences of input symbol of B.\n"
        "Assume we have Z = A o B o C, where Ain = A, Cin = C, and left.data and right.data\n"
        "contain matched left/right.  Left.data contains sequences within input alphabet of A\n"
        "and right.data contains sequences within output alphabet of C.\n"
        "\n";
}

// convert a file containing lists of forced transcription(e.g. .phn or .wrd files)
// to a file with one input sequence per line
static void 
forcedtrans_to_input(char* forced_trans_list, char* output_file)
{
    string forced_trans_file, tag;
    string forced_trans_line;
    string st, et, label, line;
    string trans;
    ifstream forced_trans_is;

    //    strcat_safe(output_file, forced_trans_list, 100);
    //    strcat_safe(output_file, ".input", 100);        
    strcpy(output_file, "");
    strcat(output_file, forced_trans_list);
    strcat(output_file, ".input");        

    // open input stream
    ifstream forced_trans_list_is(forced_trans_list);
    if (!forced_trans_list_is) {
        cerr << "can't read list file: " << forced_trans_list << endl;
        exit(1);
    }

    // open out stream
    ofstream os(output_file);
    if (!os) {
        cerr << "can't write file: " << output_file << endl;
        exit(1);
    }

    while (getline(forced_trans_list_is, forced_trans_line)) {
        // forced_trans_line is in the format: forced_trans_file (tag), parse tag
        forced_trans_file = forced_trans_line.substr(forced_trans_line.find_first_not_of(' '), 
                               forced_trans_line.rfind('(')-forced_trans_line.find_first_not_of(' '));
        forced_trans_file = forced_trans_file.substr(0, 
                                                     forced_trans_file.find_last_not_of(' ')+1);
        tag = forced_trans_line.substr(forced_trans_line.rfind('(')+1,
                                       forced_trans_line.rfind(')')-forced_trans_line.rfind('(')-1);

        forced_trans_is.open(forced_trans_file.c_str());
        if (!forced_trans_is)
            FSTFatal("can't read forced trans file |%s|", forced_trans_file.c_str());

        trans = "";
        while (getline(forced_trans_is, line)) {
            istringstream ls(line.c_str());
            ls >> st >> et >> label;
            trans += (label + " ");
        }
        os << trans << "(" << tag << ")" << endl;
        forced_trans_is.close();
    }
    os.close();

}

static int
fst_train(int argc, char** argv)
{
    if (argc < 0) {
        usage();
        return 1;
    }

    double floor = 0;
    bool error = false;
    bool list = false;
    string save_stats_dir, load_stats_dir;
    bool removeeps = false;
    ifstream left_is, right_is;

    char c;
    while ((c = getopt(argc, argv, "f:s:i:lr")) != EOF) {
        switch (c) {
        case 'f':
            floor = atof(optarg);
            break;
        case 'l':
            list = true;
            break;
        case 's':
            save_stats_dir = string(optarg);
            break;
        case 'i':
            load_stats_dir = string(optarg);
            break;
        case 'r':
            removeeps = true;
            break;
        default:
            error = true;
            break;
        }
    }
    argc -= optind;
    argv += optind;
    if (error || argc != 6) {
        usage();
        return 1;
    }

    char* data_left_file = *argv++;
    char* data_right_file = *argv++;
    const char* Ain_file  =  *argv++;
    const char* Bin_file  =  *argv++;
    const char* Cin_file  =  *argv++;
    const char* Bout_file =  *argv++;

    // cout << "double: " << DBL_MIN << " " << DBL_MAX << " " << 1.0e10 << " " << IS_NOT_TOO_SMALL(1.0e-300) << endl;

    if (list) {
        char output_file[100];

        forcedtrans_to_input(data_left_file, output_file);
        left_is.open(output_file);
        if (!left_is) {
            cerr << "can't read the produced input file: " << output_file << endl;
            exit(1);
        }
        forcedtrans_to_input(data_right_file, output_file);
        right_is.open(output_file);
        if (!right_is) {
            cerr << "can't read the produced input file: " << output_file << endl;
            exit(1);
        }
    } else {
        left_is.open(data_left_file);
        right_is.open(data_right_file);
        if (!left_is) {
            cerr << "can't read file: " << data_left_file << endl;
            exit(1);
        }
        if (!right_is) {
            cerr << "can't read file: " << data_right_file << endl;
            exit(1);
        }
    }

    cout << "Loading B...";
    cout.flush();
    FSTGeneric B = FSTLoad(Bin_file);
    if (B->n_states() == 0) {
    cerr << "B empty" << endl;
    exit(1);
    }
    cout << " DONE" << endl;

    if (removeeps) {
        cout << "Perform RemoveEpsilons on B...";
        cout.flush();
        B = FSTBasic(FSTRemoveEpsilons(B));
        B->dump_ascii("B.re2.fst");
        cout << " DONE" << endl;
    }

    // Algorithm is as follows:
    // for each pair of left and right input
    // 1) generate FST l_input from left_is
    // 2) generate FST r_input from right_is
    // 3) compute l_input o AB
    // 3a) make FSA from the input labels (l_input o AB), 
    //     also remove weights, result = Zl
    // 4) compute C o r_input
    // 4a) make FSA from the output labels (r_input o C)
    //     also remove weights, result = Zr
    // 5) compute Zl o Zr

    B->train_twosided(left_is, right_is, Ain_file, Cin_file, floor, 
                      save_stats_dir.c_str(), load_stats_dir.c_str());

    if (strlen(save_stats_dir.c_str()) == 0) {
        B->dump_ascii(Bout_file);
    }

    return 0;
}

// Register as subcommand.
//
static FstTool train_tool("train", fst_train);

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 80
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
