////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005, Massachusetts Institute of Technology              //
// All rights reserved.                                                   //
//                                                                        //
// Redistribution and use in source and binary forms, with or without     //
// modification, are permitted provided that the following conditions are //
// met:                                                                   //
//                                                                        //
//     * Redistributions of source code must retain the above copyright   //
//       notice, this list of conditions and the following disclaimer.    //
//                                                                        //
//     * Redistributions in binary form must reproduce the above          //
//       copyright notice, this list of conditions and the following      //
//       disclaimer in the documentation and/or other materials provided  //
//       with the distribution.                                           //
//                                                                        //
//     * Neither the name of the Massachusetts Institute of Technology    //
//       nor the names of its contributors may be used to endorse or      //
//       promote products derived from this software without specific     //
//       prior written permission.                                        //
//                                                                        //
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    //
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      //
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  //
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   //
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  //
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       //
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  //
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    //
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  //
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   //
////////////////////////////////////////////////////////////////////////////

#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
#include "fst.h"
#include "fst/FSTPronRules.h"

using namespace std;

typedef fst::set<int> IntSet;

static void
usage()
{
    cerr <<
        "usage: fst triphones [options] in.phones in.fst out.fst\n"
        "\n"
        "-B     output binary FST\n"
        "\n";
}

static void
load_phones(IntSet& phones, const char* file)
{
    ifstream is(file);
    string phone;
    while (is.good() && (is >> phone, !is.eof())) {
        phones.insert(FSTstring_map[phone]);
    }
    phones.erase(0);                            // epsilon
    phones.erase(1);                            // #
}

static int
diphone_label(int middle, int right)
{
    if (right == 0)
        right = FSTstring_map["<>"];
    ostringstream os;
    os << FSTstring_map[middle] << '('
       << "*,"
       << FSTstring_map[right] << ')';
    return FSTstring_map[os.str()];
}

static int
triphone_label(int left, int middle, int right)
{
    if (right == 0)
        right = FSTstring_map["<>"];
    if (left == 0)
        left = FSTstring_map["<>"];
    ostringstream os;
    os << FSTstring_map[middle] << '('
       << FSTstring_map[left] << ','
       << FSTstring_map[right] << ')';
    return FSTstring_map[os.str()];
}

static FSTGeneric
triphonify(const IntSet& phones, FSTGeneric in)
{
    IntSet::const_iterator pi, pj;
    int p;

    // Populate map pair_label_map from (middle, right) to negative integer and
    // map label_state_map from right to state.
    //
    typedef pair<int, int> Pair;
    fst::map<Pair, int> pair_label_map;
    fst::map<int, int> label_state_map;
    label_state_map[0] = 0;
    for (pi = phones.begin(), p = 1; pi != phones.end(); ++pi) {
        for (pj = phones.begin(); pj != phones.end(); ++pj) {
            pair_label_map[Pair(*pi, *pj)] = diphone_label(*pi, *pj);
        }
        pair_label_map[Pair(*pi, 0)] = diphone_label(*pi, 0);
        label_state_map[*pi] = p++;
    }

    // Build R, applied right-to-left to in, converting from phones to (middle,
    // right) temporary labels.
    //
    FSTGeneric R = FSTBasic();
    R->add_states(label_state_map.size());
    R->set_initial(0);
    R->set_final(0);
    for (fst::map<int, int>::const_iterator i = label_state_map.begin();
         i != label_state_map.end(); ++i) {
        int x = (*i).first;
        p = (*i).second;
        R->add_arc(p, p, 1, 1);                 // # self loop
        if (p == 0)
            continue;
        R->set_final(p);
        R->add_arc(0, p, pair_label_map[Pair(x, 0)], x);
        for (pi = phones.begin(); pi != phones.end(); ++pi) {
            int y = *pi;
            R->add_arc(p, label_state_map[y], pair_label_map[Pair(y, x)], y);
        }
    }
    //R->dump_ascii("R.fst");

    // Build L, applied left-to-right to the result of (R o in), outputting
    // triphone labels.
    //
    FSTGeneric L = FSTBasic();
    L->add_states(pair_label_map.size());
    L->set_initial(0);
    L->set_final(0);
    for (fst::map<Pair, int>::const_iterator i = pair_label_map.begin();
         i != pair_label_map.end(); ++i) {
        int m = (*i).first.first;
        int r = (*i).first.second;
        int label = (*i).second;
        L->add_arc(0, label_state_map[m], triphone_label(0, m, r), label);
    }
    for (fst::map<int, int>::const_iterator i = label_state_map.begin();
         i != label_state_map.end(); ++i) {
        int x = (*i).first;
        int p = (*i).second;
        if (p == 0)
            continue;
        L->set_final(p);
        L->add_arc(p, p, 1, 1);                 // # self loop
        for (fst::map<Pair, int>::const_iterator j = pair_label_map.begin();
             j != pair_label_map.end(); ++j) {
            int m = (*j).first.first;
            int r = (*j).first.second;
            int label = (*j).second;
            L->add_arc(p, label_state_map[m], triphone_label(x, m, r), label);
        }
    }
    //L->dump_ascii("L.fst");

    FSTGeneric x = FSTReverse(in);
    x = FSTReverse(FSTCompose(R, x));
    x = FSTBasic(FSTCompose(L, x));
    x = FSTBasic(FSTRemoveEpsilonsA(x));

    return x;
}

static int
fst_triphones(int argc, char** argv)
{
    if (argc < 0) {
        usage();
        return 1;
    }

    bool error = false;
    bool binary = false;

    char c;
    while ((c = getopt(argc, argv, "B")) != EOF) {
        switch (c) {
        case 'B':
            binary = true;
            break;
        default:
            error = true;
            break;
        }
    }
    argc -= optind;
    argv += optind;
    if (error || argc != 3) {
        usage();
        return 1;
    }

    const char* phones_file = *argv++;
    const char* in_file     = *argv++;
    const char* out_file    = *argv++;

    IntSet phones;
    load_phones(phones, phones_file);

    FSTGeneric out = triphonify(phones, FSTLoad(in_file));

    if (binary)
        out->dump_binary(out_file);
    else
        out->dump_ascii(out_file);

    return 0;
}

// Register as subcommand.
//
static FstTool triphones_tool("triphones", fst_triphones);

// for Emacs...
// Local Variables:
// mode: C++
// comment-column: 48
// fill-column: 80
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
