# Some subtle things:
#
# The SLS_FINISH should appear right before SLS_OUTPUT, so that is has
# a chance to add the sls_X variables to the X variables.
#
# GNU Autoconf/Automake specifies that the builder should be able to
# override CPPFLAGS, CFLAGS, CXXFLAGS, and LDFLAGS by putting values
# in the environment when running configure -- these get appended to
# whatever values are provided by the autoconf package definer in
# configure.ac and Makefile.am.  This is so that a platform-specific
# package maintainer (i.e. Debian, Red Hat, etc.) can force the flags
# in their package definition to the values they have determined are
# correct for that platform.
#
# The problem with this is that autoconf initializes CFLAGS and
# CXXFLAGS to '-O2 -g' if they have not been set in the environment,
# and the CFLAGS and CXXFLAGS follow any flags we may set; thuse -o2
# overrides anything we provide, so we need to set CFLAGS and CXXFLAGS
# ourselves if we want an optimization other than -O2.
#
# You may provide your own initializations for these variables:
#
#   SLS_CPPFLAGS
#      
#      Use this in Makefile.am's AM_CPPFLAGS.  It contains defines for
#      ARCH, OS, etc., as well as additional include directories.
#
#   SLS_LDFLAGS
#
#      Use this in Makefile.am's AM_LDFLAGS.  It contains the paths
#      for finding shared libraries at link and load time.
#
#   sls_CPPFLAGS
#   sls_noopt_CPPFLAGS
#   sls_opt_CPPFLAGS
#
#      SLS_CPPFLAGS will consist of sls_CPPFLAGS and one of
#      sls_opt_CPPFLAGS or sls_nopt_CPPFLAGS, depending on whether or
#      not optimization is enabled.
#
#   sls_CFLAGS
#   sls_noopt_CFLAGS
#   sls_opt_CFLAGS
#
#      These have no effect if CFLAGS was specified by the user.
#      Otherwise, CFLAGS will be set to sls_CFLAGS and one of
#      sls_opt_CFLAGS or sls_noopt_CFLAGS, depending on whether or not
#      optimization is enabled.
#
#   sls_CXXFLAGS
#   sls_noopt_CXXFLAGS
#   sls_opt_CXXFLAGS
#
#      These have no effect if CXXFLAGS was specified by the user.
#      Otherwise, CXXFLAGS will be set to sls_CXXFLAGS and one of
#      sls_opt_CXXFLAGS or sls_noopt_CXXFLAGS, depending on whether or
#      not optimization is enabled.
#
#
#
# NOTES:
#
# AC_REQUIRE does not know a macro is already expanded if it was seen
# at top-level, i.e. in configure.ac, so using a public macro that has
# been used a top-level in an AC_REQUIRE will cause it to expand
# twice.  Thus, if public macro X is needed in an AC_REQUIRE we define
# an _X and use AC_REQUIRE(_X) in X and the other macros that need X.
#

# Provide the usual SLS compilation flags
AC_DEFUN([_SLS_INIT],
[
AC_SUBST([SLS_CPPFLAGS])
AC_SUBST([SLS_LDFLAGS])
# Default settings for flags
sls_CFLAGS+='-Wall'
sls_noopt_CFLAGS+='-O0'
sls_CXXFLAGS+='-Wall'
sls_noopt_CXXFLAGS+='-O0'
])

# SLS_APPEND_LIBPATH(dir)
# Add dir to the directories the linker uses to find libraries
AC_DEFUN([SLS_APPEND_LIBPATH],
[AC_REQUIRE([_SLS_INIT])
_SLS_APPEND([SLS_LDFLAGS],[-L$1])
])

# SLS_BUILD_LIBPATH(dir)
# Add dir, a path relative to the build directory, to the directories the
# linker AND the shared library loader use to find libraries.  This lets
# you run from the build directory when some libraries were built by
# other Makefiles in the same project.
AC_DEFUN([SLS_BUILD_LIBPATH],
[AC_REQUIRE([_SLS_INIT])
SLS_APPEND_LIBPATH([\$(top_builddir)/$1])
# Need -R, not -rpath, since libtool uses -rpath as the installation
# directory
_SLS_APPEND([SLS_LDFLAGS],[-R\$(abs_top_builddir)/$1])
])

AC_DEFUN([_SLS_APPEND],[$1+=" $2"])

# SLS_INCLUDEPATH(dir)
# Add dir to the include path.
AC_DEFUN([SLS_INCLUDEPATH],
[AC_REQUIRE([_SLS_PROG_CPP])
_SLS_APPEND([sls_CPPFLAGS], [-I $1])
])

# _SLS_ARCHOS
# Set ARCH,OS, and ARCHOS to the "traditional" SLS values
AC_DEFUN([_SLS_ARCHOS],
[
AC_REQUIRE([AC_CANONICAL_TARGET])
AC_REQUIRE([_SLS_INIT])
AC_SUBST([ARCH])
AC_SUBST([OS])
AC_SUBST([ARCHOS])
AS_CASE([$target_cpu],
	[i?86*], [ARCH=x86],
	[i?86_64], [ARCH=x86_64],
	[arm], [ARCH=armel],
	[*], [ARCH=$target_cpu])
AS_CASE([$target_os],
	[linux*], [OS=linux],
	[darwin10.*], [OS=osx; ARCH=x86_64],
	[darwin*], [OS=osx; ARCH=x86],
	[*], [OS=$target_os])
ARCHOS="$ARCH-$OS"
])

# SLS_PROG_CPP
# AC_PROG_CPP and initialize SLS CPPFLAGS
AC_DEFUN([SLS_PROG_CPP],[AC_REQUIRE([_SLS_PROG_CPP])])

AC_DEFUN([_SLS_PROG_CPP],
[AC_REQUIRE([_SLS_ARCHOS])
AC_REQUIRE([AC_PROG_CPP])
_SLS_APPEND([sls_CPPFLAGS],[-DARCH=${ARCH} -DOS=${OS} -DARCHOS=${ARCHOS}])
])

# SLS_SUMMIT_PROG_CPP
# SLS_PROG_CPP and initialize additional summit/fst CPPFLAGS
AC_DEFUN([SLS_SUMMIT_PROG_CPP],
[AC_REQUIRE([_SLS_PROG_CPP])
sls_CPPFLAGS+=" -D__FST__ -D__${ARCH}__ -D__${OS}__"
])

# SLS_SUMMIT_PROG_CC
# AC_PROG_CC and initialize summit/fst CFLAGS
AC_DEFUN([SLS_SUMMIT_PROG_CC],
[AC_REQUIRE([_SLS_ARCHOS])
AC_REQUIRE([AC_PROG_CC])
sls_CFLAGS+=' -Wno-sign-compare'
AS_CASE([$ARCH],
	[x86], [ sls_CFLAGS+=' -march=i686 -msse'
	         sls_opt_CFLAGS+=' -O3 -fomit-frame-pointer -fno-strict-aliasing'],
	[x86_64], [ sls_CFLAGS+=' -march=nocona -m64 -mpreferred-stack-boundary=4' 
	          sls_opt_CFLAGS+=' -O2 -fomit-frame-pointer'])
])

# SLS_SUMMIT_PROG_CXX
# AC_PROG_CXX and initialize summit/fst CXXFLAGS
AC_DEFUN([SLS_SUMMIT_PROG_CXX],
[AC_REQUIRE([_SLS_ARCHOS])
AC_REQUIRE([AC_PROG_CXX])
sls_CXXFLAGS+=' -Wno-sign-compare -fno-exceptions'
AS_CASE([$ARCH],
	[x86], [ sls_opt_CXXFLAGS+=' -O3 -fomit-frame-pointer -fno-strict-aliasing' ],
	[x86_64], [ sls_opt_CXXFLAGS+=' -O2 -fomit-frame-pointer' ])
])
	
# SLS_PROG_LEX
# We are actively using several incompatible versions of flex.  This
# detects if we have a version that needs yywrap defined.
AC_DEFUN([SLS_PROG_LEX],
[AC_REQUIRE([AC_PROG_LEX])
AC_REQUIRE([AC_PROG_CXX])
# ylwrapper expects lex.yy.c, not lex.yy.cc
LEX='flex -o lex.yy.c'
# Newer versions (Debian/Lenny, Ubuntu 8.04) of <FlexLexer.h> require a
# definition for yywrap, while older versions (Debian/etch, n810s)
# cannot have a definition.  We define HAVE_YYFLEXER_YYWRAP here if
# <FlexLexer.h> is of the variety that needs a definition.
AC_LANG_PUSH([C++])
  AC_COMPILE_IFELSE([
#include <FlexLexer.h>
int yyFlexLexer::yywrap(){return 1;}
], 
[AC_DEFINE([HAVE_YYFLEXLEXER_YYWRAP],[1], [yywrap needs definition])],
[]
)
AC_LANG_POP([C++])
])

# SLS_CONFIGURE_DISABLE_OPT_FLAG
# Add --disable_sls_opt configure flag
#
# This flag lets yout turn off optimzation.  You would normally want
# to do this if you were developing.  The default is to build the
# package optimized, consistent with convention.
AC_DEFUN([SLS_CONFIGURE_DISABLE_SLS_OPT],
[AC_REQUIRE([_SLS_CONFIGURE_DISABLE_SLS_OPT])])

AC_DEFUN([_SLS_CONFIGURE_DISABLE_SLS_OPT],
[AC_REQUIRE([_SLS_INIT])
AC_MSG_CHECKING([whether to enable optimizations that make debugging hard])
AC_ARG_ENABLE([sls-opt],
 [AS_HELP_STRING([--disable-sls-opt],
                 [turn off optimizations that make debugging difficult])],
 [],
 [enable_sls_opt=yes]
)
])

# SLS_CONFIGURE_WITH_HOME
# Add --with-home configure option.
#
# This lets you specify a directory to use as a sandbox for one or more
# installed packages.  The headers and libraries will be added to the
# include and libpath.
AC_DEFUN([SLS_CONFIGURE_WITH_HOME],
[
AC_REQUIRE([_SLS_INIT])
AC_REQUIRE([_SLS_ARCHOS])
AC_ARG_WITH([home-prefix],
            AC_HELP_STRING([--with-home-prefix=dir],
                           [Install to specified home]),
	    [ sls_home_prefix=`eval withval=$withval ; mkdir -p $withval ; cd $withval ; pwd -P` ])

AC_ARG_WITH([home-root],
	    AC_HELP_STRING([--with-home-root=dir],
			   [Use definitions from home]),
            [ sls_home_root=`eval withval=$withval ; mkdir -p $withval ; cd $withval ; pwd -P` ])

AC_ARG_WITH([home],
            AC_HELP_STRING([--with-home=dir],
                           [Use home for install and definitions]),
	    [ sls_home_prefix=`eval withval=$withval ; mkdir -p $withval ; cd $withval ; pwd -P`
	      sls_home_root=`eval withval=$withval ; mkdir -p $withval ; cd $withval ; pwd -P`
	     ])

if test "$sls_home_prefix"
then
	prefix="${sls_home_prefix}/${ARCHOS}"
        bindir="${prefix}/bin"
        libdir="${prefix}/lib"
fi

if test "$sls_home_root"
then 
     SLS_INCLUDEPATH([$sls_home_root/$ARCHOS/include])
     _SLS_APPEND([SLS_LDFLAGS],[-L$sls_home_root/$ARCHOS/lib])
fi
])


# SLS_CONFIGURE_WITH_GALAXY
# Add --with-galaxy configure option
# OBSOLETE
#
# This is like --with-home except that it assumes the directories are
# in the format that was used with Galaxy/Communicator.
AC_DEFUN([SLS_CONFIGURE_WITH_GALAXY],
[
AC_REQUIRE([_SLS_INIT])
AC_REQUIRE([_SLS_ARCHOS])
AC_ARG_WITH([galaxy-prefix],
            AC_HELP_STRING([--with-galaxy-prefix=dir],
                           [Install to specified galaxy]),
	    [ sls_galaxy_prefix=`eval withval=$withval ; cd $withval ; pwd -P` ])

AC_ARG_WITH([galaxy-root],
	    AC_HELP_STRING([--with-galaxy-root=dir],
			   [Use Galaxy definitions from dir]),
            [ sls_galaxy_rootdir=`eval withval=$withval ; cd $withval ; pwd -P` ])

AC_ARG_WITH([galaxy],
            AC_HELP_STRING([--with-galaxy=dir],
                           [Use Galaxy for install and definitions]),
	    [ sls_galaxy_prefix=`eval withval=$withval ; cd $withval ; pwd -P`
	      sls_galaxy_rootdir=`eval withval=$withval ; cd $withval ; pwd -P`
	     ])

if test "$sls_galaxy_prefix"
then
	prefix="${sls_galaxy_prefix}"
        bindir="${prefix}/bin/${ARCHOS}"
        libdir="${prefix}/lib/${ARCHOS}"
fi

if test "$sls_galaxy_rootdir"
then 
     SLS_INCLUDEPATH([$sls_galaxy_rootdir/include])
     _SLS_APPEND([SLS_LDFLAGS],[-L$sls_galaxy_rootdir/lib/$ARCHOS])
fi])
])


# SLS_CONFIG_WITH_COMMUNICATOR
# Add --with-communicator configure processions
#
# This options lets you specify a directory with a built Communicator
# that should be used as the implementation of frames.
AC_DEFUN([SLS_CONFIGURE_WITH_COMMUNICATOR],
[
AC_REQUIRE([_SLS_INIT])
AC_REQUIRE([_SLS_ARCHOS])
AC_REQUIRE([_SLS_CONFIGURE_DISABLE_SLS_OPT])
AC_SUBST([SLS_FRAMES_LIBS])
AC_ARG_WITH([communicator],
	    AC_HELP_STRING([--with-communicator=dir],
	                   [Use Communicator for frames]),
            [ 
	      if test "x$enable_sls_opt" == "xyes" ; then
  	      	 SLS_FRAMES_LIBS='-lGalaxyEnhancements -lGalaxy'
              else
	         SLS_FRAMES_LIBS='-lGalaxyEnhancements_debug -lGalaxy_debug'
              fi
	      sls_communicator_root=`eval withval=$withval ; cd $withval ; pwd -P`
	      SLS_INCLUDEPATH([$sls_communicator_root/include])
	      SLS_INCLUDEPATH([$sls_communicator_root/contrib/MIT/galaxy/include])
	      SLS_BUILD_LIBPATH([$sls_communicator_root/contrib/MIT/galaxy/lib/$ARCHOS])
	      SLS_BUILD_LIBPATH([$sls_communicator_root/lib/$ARCHOS])
            ],
            [ SLS_FRAMES_LIBS='-lgalframes' ])
])

# SLS_FINISH
# Adjust variables before writing
#
# This should be called right before AC_OUTPUT.  It appends the sls_X
# variables to the X variables.
AC_DEFUN([SLS_FINISH],
[AC_REQUIRE([_SLS_INIT])
SLS_CPPFLAGS+=" $sls_CPPFLAGS"
if test "x$enable_sls_opt" == "xyes"; then
   SLS_CPPFLAGS+=" $sls_opt_CPPFLAGS"
else
   SLS_CPPFLAGS+=" $sls_noopt_CPPFLAGS"
fi

if test "$ac_test_CFLAGS" != set ; then
   CFLAGS+=" $sls_CFLAGS"
   if test "x$enable_sls_opt" == "xyes"; then
      CFLAGS+=" $sls_opt_CFLAGS"
   else
      CFLAGS+=" $sls_noopt_CFLAGS"
   fi
fi

if test "$ac_test_CXXFLAGS" != set ; then
   CXXFLAGS+=" $sls_CXXFLAGS"
   if test "x$enable_sls_opt" == "xyes"; then
      CXXFLAGS+=" $sls_opt_CXXFLAGS"
   else
      CXXFLAGS+=" $sls_noopt_CXXFLAGS"
   fi
fi
])
